/**
 * @file conf_obj.c
 *
 * @brief Functions for the manipulation of the various config objects.
 *
 * $Id: conf_obj.c,v 1.12.2.8 2003/04/30 20:03:05 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Mon Jun 17 16:15:19 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "strcompat.h"
#include "error.h"

#include "util/list.h"

#include "chaninfo.h"
#include "outputmod/outputmod.h"
#include "message/msg_frmt.h"
#include "config.h"
#include "conf_private.h"
#include "conf_obj.h"

static char const rcsid[] 
    = "$Id: conf_obj.c,v 1.12.2.8 2003/04/30 20:03:05 devink Exp $";

struct config_info *
CONFOBJ_createInfo(void)
{
	struct config_info *ret;

	ret=emalloc(sizeof(struct config_info));
    memset(ret, 0, sizeof(struct config_info));

	return ret;
}

struct inchan_info *
CONFOBJ_createInput(void)
{
	struct inchan_info *ret;

	ret=emalloc(sizeof(struct inchan_info));
	memset(ret,0,sizeof(struct inchan_info));

	ret->type=-1;
	ret->inPort=-1;

	return ret;
}

struct outchan_info *
CONFOBJ_createOutput(void)
{
	struct outchan_info *ret;

	ret=emalloc(sizeof(struct outchan_info));
	memset(ret,0,sizeof(struct outchan_info));

    ret->opts = emalloc(sizeof(struct OMOD_opt));
    memset(ret->opts, 0, sizeof(struct OMOD_opt));

	ret->type=-1;
	ret->opts->format.type = MFRMT_NOEXIST;
    ret->opts->format.options = 0;
	ret->opts->outPort=-1;

	return ret;
}

struct CONF_con *
CONFOBJ_createCon(void)
{
	struct CONF_con *ret;

	ret=emalloc(sizeof(struct CONF_con));
	memset(ret,0,sizeof(struct CONF_con));

	return ret;
}

struct CONF_acl *
CONFOBJ_createACL(void)
{
	fprintf(stderr,"**** CONFOBJ_createACL() NOT YET IMPLEMENTED ****\n");

	return NULL;
}

struct CONF_global *
CONFOBJ_createGlobal(void)
{
	struct CONF_global *ret;

	ret=emalloc(sizeof(struct CONF_global));
	memset(ret,0,sizeof(struct CONF_global));

	return ret;
}

void
CONFOBJ_destroyInfo(struct config_info **de)
{
	LLIterator it;
	struct outchan_info *out;
	struct inchan_info *in;
	struct CONF_acl *acl;
	struct CONF_con *con;

	if(NULL==de)
		return;
	if(NULL==*de)
		return;

	/* destroy the output list */
    if ((*de)->outputs != NULL) {
        assert((*de)->output_count==LL_size((*de)->outputs));

        it=LL_iterator((*de)->outputs);
        while(NULL!=LLI_next(it))
        {
            out=LLI_elem(it);
            CONFOBJ_destroyOutput(&out);
        }
        LLI_destroy(&it);
        LL_destroy(&((*de)->outputs));
    }

	/* destroy the input list */
    if ( (*de)->inputs != NULL) {
        assert((*de)->input_count==LL_size((*de)->inputs));

        it=LL_iterator((*de)->inputs);
        while(NULL!=LLI_next(it))
        {
            in=LLI_elem(it);
            CONFOBJ_destroyInput(&in);
        }
        LLI_destroy(&it);
        LL_destroy(&((*de)->inputs));
    }

	/* destroy the acl list */
    if ( (*de)->acls != NULL) {
        assert((*de)->acl_count==LL_size((*de)->acls));

        it=LL_iterator((*de)->acls);
        while(NULL!=LLI_next(it))
        {
            acl=LLI_elem(it);
            CONFOBJ_destroyACL(&acl);
        }
        LLI_destroy(&it);
        LL_destroy(&((*de)->acls));
    }

	/* destroy the connect list */
    if ( (*de)->connects != NULL) {
        assert((*de)->connect_count==LL_size((*de)->connects));

        it=LL_iterator((*de)->connects);
        while(NULL!=LLI_next(it))
        {
            con=LLI_elem(it);
            CONFOBJ_destroyCon(&con);
        }
        LLI_destroy(&it);
        LL_destroy(&((*de)->connects));
    }

	/* destroy the globals */
	if(NULL!=(*de)->global)
	{
		CONFOBJ_destroyGlobal(&((*de)->global));
	}

	free((*de));
	*de=NULL;
}

void
CONFOBJ_destroyInput(struct inchan_info **de)
{
	if(NULL==de)
		return;
	if(NULL==*de)
		return;

	if( NULL!=(*de)->inPath )
		free((*de)->inPath);
	if(NULL!=(*de)->name)
		free((*de)->name);
	(*de)->type=-1;
	(*de)->inPort=-1;
	NA_destroy((*de)->inAddr);

	free(*de);
	*de=NULL;
}

void
CONFOBJ_destroyOutput(struct outchan_info **de)
{
	if(NULL==de)
		return;
	if(NULL==(*de))
		return;

	if(NULL!=(*de)->name)
		free((*de)->name);

	if(NULL!=(*de)->opts)
	{
		if(NULL!=(*de)->opts->outPath)
			free((*de)->opts->outPath);
        if(NULL!=(*de)->opts->outList)
        {
            List    outList = (*de)->opts->outList;
            char    *str;
            while (LL_size(outList) > 0)
            {
                str = LL_first(outList);
                LL_rmFirst(outList);
                free(str);
            }
            LL_destroy(&((*de)->opts->outList));
        }
        /* ::pcache is not destroyed here because the controller
         * ownes it (and is thus responsible for cleaning it).
         */
        NA_destroy((*de)->opts->outAddr);

		free((*de)->opts);
		(*de)->opts=NULL;
	}

	free(*de);
	*de=NULL;
}

void
CONFOBJ_destroyCon(struct CONF_con **de)
{
	if(NULL==de)
		return;
	if(NULL==(*de))
		return;

	(*de)->input=NULL;
	(*de)->output=NULL;

	free((*de));
	*de=NULL;
}

void
CONFOBJ_destroyACL(struct CONF_acl **de)
{
	fprintf(stderr,"**** conf_sdsc_destroy_CONF_acl() NOT YET IMPLEMENTED ****\n");
}

void
CONFOBJ_destroyGlobal(struct CONF_global **de)
{
	if(NULL==de)
		return;
	if(NULL==(*de))
		return;

	if(NULL!=(*de)->user)
		free((*de)->user);

	if(NULL!=(*de)->group)
		free((*de)->group);

	if(NULL!=(*de)->stdredir)
		free((*de)->stdredir);

	free((*de));
	*de=NULL;
}

void
dump_config_info(struct config_info *in)
{
#ifdef DUMP_CONFIG_INFO
	printf("dumping conf input:\n");

	if(NULL==in)
	{
		printf("DUMP config: in is NULL\n");
		return;
	}

	if(0==in->input_count)
	{
		printf("DUMP config: input_count is 0\n");
	}
	else
	{
		LLIterator it;
		int i=0;

		it=LL_iterator(in->inputs);

		LLI_next(it);
		while(NULL!=LLI_elem(it))
		{
			printf("dumping input %d...\n",i++);
			dump_inchan_info( (struct inchan_info *)LLI_elem(it));

			LLI_next(it);
		}

		LLI_destroy(&it);
	}

	if(0==in->output_count)
	{
		printf("DUMP config: output_count is 0\n");
	}
	else
	{
		LLIterator it;
		int i=0;

		it=LL_iterator(in->outputs);

		LLI_next(it);
		while(NULL!=LLI_elem(it))
		{
			printf("dumping output %d...\n",i++);
			dump_outchan_info( (struct outchan_info *)LLI_elem(it));

			LLI_next(it);
		}

		LLI_destroy(&it);
	}

	if(NULL==in->global)
		printf("DUMP config: global is NULL\n");
	else
	{
		printf("DUMP config: user: %s\n",in->global->user);
		printf("DUMP config: group: %s\n",in->global->group);
		printf("DUMP config: debug-file: %s\n",in->global->stdredir);
	}
#endif /* DUMP_CONFIG_INFO */
}

void
dump_inchan_info(struct inchan_info *in)
{
#ifdef DUMP_CONFIG_INFO
	if(NULL==in)
	{
		printf("\tDUMP CONF: in is NULL\n");
		return;
	}

	printf("\tDUMP: name: %s\n",in->name);

	printf("\tDUMP: type: %d\n",in->type);

	if(NULL==in->inAddr)
	{
		printf("\tDUMP: inAddr: NULL\n");
	}
	else
		printf("\tDUMP: inAddr: %s\n",NA_getCharAddr(in->inAddr));

	printf("\tDUMP: inPath: %s\n",in->inPath);

	printf("\tDUMP: listen_port: %d\n",in->inPort);
#endif /* DUMP_CONFIG_INFO */
}

void
dump_outchan_info(struct outchan_info *in)
{
    /* Not Used any more */
#ifdef DUMP_CONFIG_INFO 
	int i;

	if(NULL==in)
	{
		printf("\tDUMP CONF: in is NULL\n");
		return;
	}

	if(NULL!=in->name)
		printf("\tDUMP: name: %s\n",in->name);

	printf("\tDUMP: type: %d\n",in->type);
	printf("\tDUMP: format: %d\n",in->opts->format.type);
    printf("\tDUMP: fopts: 0x%x\n", in->opts->format.options);

	printf("\tDUMP: outPort: %d\n",in->opts->outPort);

	printf("\tDUMP: outPath: %s\n",in->opts->outPath);

	printf("\tDUMP: priority: ");
	for(i=0;i<256;i++)
		if(in->pri[i])
			printf("%d ",i);
	printf("\n");
#endif /* DUMP_CONFIG_INFO */
}
