/**
 * @file conf_obj.h
 *
 * @brief Functions that create and destroy configuration (and controller)
 * objects
 *
 * @ingroup conf_obj
 * @todo Support ACLs and connections
 *
 * $Id: conf_obj.h,v 1.2.2.1 2003/04/30 20:03:05 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Jul 18 14:50:38 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CONF_OBJ_H
#define CONF_OBJ_H

/**@defgroup conf_obj Config Parser Objects
 * @ingroup conf
 * @{
 */

/** 
 * Create a new config_info struct
 * This function is responsible for creating the Lists, as well.  The 
 * destructor, CONFOBJ_destroyInfo() MUST be called on variables created 
 * with this function, or there WILL be memory lost.
 *
 * @return The initialized, but empty, struct config_info object
 */
struct config_info *CONFOBJ_createInfo(void);

/** 
 * Create a new inchan_info struct
 * This function is responsible for creating all sub-elements, as well, 
 * save for the name field.  The destructor, CONFOBJ_destroyInput() MUST 
 * be called to properly destroy the memory that is malloc()ed by 
 * this function.
 *
 * @return The initialized, but empty, struct inchan_info object
 */
struct inchan_info *CONFOBJ_createInput(void);

/** 
 * Create a new outchan_info struct
 * This function is responsible for creating all sub-elements, as well, 
 * save for any strings (which are malloc()ed to the exact size upon their 
 * creation.) Memory allocated by this function MUST be destroyed using the 
 * paired destructor, CONFOBJ_destroyOutput().
 *
 * @return The initialized, but empty, struct outchan_info object
 */
struct outchan_info *CONFOBJ_createOutput(void);

/** 
 * Create a new CONF_con struct
 * This function is not responsible for creating any sub-elements, only 
 * initializing them to NULL.
 * The object returned by this function MUST be destroyed using the paired
 * destructor, CONFOBJ_destroyCon().
 *
 * @return The initialized, but empty, struct CONF_con object
 */
struct CONF_con *CONFOBJ_createCon(void);

/** 
 * Create a new CONF_acl structure
 * This function is not responsible for creating any sub-elements, only
 * initializing them to NULL.
 * The object returned by this function MUST be destroyed using the paird
 * destructor, CONFOBJ_destroyACL()
 *
 * @return The initialized, but empty, struct CONF_acl object
 */
struct CONF_acl *CONFOBJ_createACL(void);

/** 
 * Create a new CONF_global structure
 * This function is not responsible for creating any sub-elements, only
 * initializing them to NULL.
 * The object returned by this function MUST be destroyed using the paird
 * destructor, CONFOBJ_destroyACL()
 *
 * @return THe initialized, but empty, struct CONF_global object
 */
struct CONF_global *CONFOBJ_createGlobal(void);

/** 
 * Destroy a config_info struct
 *
 * @pre de != NULL
 *
 * @param de The struct config_info to be destroyed
 */
void CONFOBJ_destroyInfo(struct config_info **de);

/** 
 * Destroy a inchan_info struct
 *
 * @pre de != NULL
 *
 * @param de The struct inchan_info to be destroyed
 */
void CONFOBJ_destroyInput(struct inchan_info **de);

/**
 * Destroy a outchan_info struct
 *
 * @pre de != NULL
 *
 * @param de The struct outchan_info to be destroyed
 */
void CONFOBJ_destroyOutput(struct outchan_info **de);

/** 
 * Destroy a CONF_con struct
 *
 * @pre de != NULL
 *
 * @param de The struct CONF_con to be destroyed
 */
void CONFOBJ_destroyCon(struct CONF_con **de);

/** 
 * Destroy a CONF_acl struct
 *
 * @pre de != NULL
 *
 * @param de The struct CONF_acl to be destroyed
 */
void CONFOBJ_destroyACL(struct CONF_acl **de);

/** 
 * Destroy a CONF_global structure
 *
 * @pre de != NULL
 *
 * @param de The struct CONF_global to be destroyed
 */
void CONFOBJ_destroyGlobal(struct CONF_global **de);

/*@}*/

#endif
