/**
 * @file conf_private.c
 *
 * @brief Private functions for the parser (symbol table, special prints 
 * for debugging, etc.)
 *
 *
 * $Id: conf_private.c,v 1.26.2.7 2003/04/30 20:03:06 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Apr 23 15:31:03 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "strcompat.h"
#include "error.h"
#include "util/list.h"
#include "util/netaddr.h"

#include "chaninfo.h"
#include "config.h"
#include "conf_private.h"
#include "conf_obj.h"

static char const rcsid[] = 
    "$Id: conf_private.c,v 1.26.2.7 2003/04/30 20:03:06 devink Exp $";

/** Global variable; holds the symbol table */
struct CONFSYM_str *CONFSYM_table;

void
CONFSYM_init(void)
{
	assert(NULL==CONFSYM_table);

	CONFSYM_table=emalloc(sizeof(struct CONFSYM_str));
	memset(CONFSYM_table,0,sizeof(struct CONFSYM_str));

	CONFSYM_table->acl_l=LL_create();
	CONFSYM_table->input_l=LL_create();
	CONFSYM_table->output_l=LL_create();
	CONFSYM_table->connect_l=LL_create();
}

void
CONFSYM_destroy(void)
{
	struct config_info *t;

	if(NULL==CONFSYM_table)
		return;

	t=CONFSYM_getInfo();
	CONFOBJ_destroyInfo(&t);
}

void
CONFSYM_addInput(struct inchan_info *in)
{
	LL_append(CONFSYM_table->input_l,in);
	CONFSYM_table->input_count++;
}

struct inchan_info *
CONFSYM_getInput(const char *name)
{
	LLIterator it;
	struct inchan_info *ret;

	/* check to make sure that we don't have a null or empty
	 * tree
	 */
	if(NULL==CONFSYM_table->input_l)
		return NULL;
	if(0==CONFSYM_table->input_count)
		return NULL;

	it=LL_iterator(CONFSYM_table->input_l);

	ret=LLI_next(it);
	while(NULL!=ret)
	{
		if(0==strncmp(ret->name,name,strlen(name)+1))
		{
			LLI_destroy(&it);
			return ret;
		}

		ret=LLI_next(it);
	}

	LLI_destroy(&it);
	return NULL;
}

void
CONFSYM_addOutput(struct outchan_info *in)
{
	LL_append(CONFSYM_table->output_l,in);
	CONFSYM_table->output_count++;
}

struct outchan_info *
CONFSYM_getOutput(const char *name)
{
	LLIterator it;
	struct outchan_info *ret;

	/* check to make sure that we don't have a null or empty
	 * tree
	 */
	if(NULL==CONFSYM_table->output_l)
		return NULL;
	if(0==CONFSYM_table->output_count)
		return NULL;

	it=LL_iterator(CONFSYM_table->output_l);

	ret=LLI_next(it);
	while(NULL!=ret)
	{
		if(0==strncmp(ret->name,name,strlen(name)+1))
		{
			LLI_destroy(&it);
			return ret;
		}

		ret=LLI_next(it);
	}

	LLI_destroy(&it);
	return NULL;
}

void
CONFSYM_addConnect(struct CONF_con *in)
{
	LL_append(CONFSYM_table->connect_l,in);
	CONFSYM_table->connect_count++;
}

struct CONF_con *
CONFSYM_getConnect(char *input,char *output)
{
	LLIterator it;
	struct CONF_con *ret;

	/* if they're both NULL, we don't have any scanning to do */
	if( (NULL==input) && (NULL==output) )
		return NULL;

	/* if neither's NULL, which would we match first?  Let's just
	 * return NULL */
	if( (NULL!=input) && (NULL!=output) )
		return NULL;

	it=LL_iterator(CONFSYM_table->connect_l);

	while(NULL!=LLI_next(it))
	{
		ret=LLI_elem(it);

		if( (NULL!=input) && 
				(0==strncmp(ret->input,input,strlen(input)+1)) )
		{
			LLI_destroy(&it);
			return ret;
		}

		if( (NULL!=output) &&
				(0==strncmp(ret->output,output,strlen(output)+1)) )
		{
			LLI_destroy(&it);
			return ret;
		}
	}

	LLI_destroy(&it);
	return NULL;
}

bool
CONFSYM_addGlobal(struct CONF_global *in)
{
	if(NULL!=CONFSYM_table->global)
		return FALSE;

	CONFSYM_table->global=in;

	return TRUE;
}

struct CONF_global *
CONFSYM_getGlobal(void)
{
	return CONFSYM_table->global;
}

struct config_info *
CONFSYM_getInfo(void)
{
	struct config_info *ret;
	LLIterator it;

	/* make sure the symbol table was initialized; if it wasn't, what are
	 * we doing in this function?
	 */
	if(NULL==CONFSYM_table)
		return NULL;

	/* malloc our new storage space */
	ret=CONFOBJ_createInfo();

	/* get the ACL stuff */
	ret->acls=LL_create();
	ret->acl_count=0;
	if(CONFSYM_table->acl_count>0)
	{
		it=LL_iterator(CONFSYM_table->acl_l);
		LLI_next(it);
		while(NULL!=LLI_next(it))
		{
			LL_append(ret->acls,LLI_elem(it));
			ret->acl_count++;
		}

		LLI_destroy(&it);
	}
	LL_destroy(&(CONFSYM_table->acl_l));

	/* get the outputs */
	ret->outputs=LL_create();
	ret->output_count=0;
	if(CONFSYM_table->output_count>0)
	{
		it=LL_iterator(CONFSYM_table->output_l);
		while(NULL!=LLI_next(it))
		{
			LL_append(ret->outputs,LLI_elem(it));
			ret->output_count++;
		}

		LLI_destroy(&it);
	}
	LL_destroy(&(CONFSYM_table->output_l));

	/* get the inputs */
	ret->inputs=LL_create();
	ret->input_count=0;
	if(CONFSYM_table->input_count>0)
	{
		it=LL_iterator(CONFSYM_table->input_l);
		while(NULL!=LLI_next(it))
		{
			LL_append(ret->inputs,LLI_elem(it));
			ret->input_count++;
		}

		LLI_destroy(&it);
	}
	LL_destroy(&(CONFSYM_table->input_l));

	/* get the connections */
	ret->connects=LL_create();
	ret->connect_count=0;
	if(CONFSYM_table->connect_count>0)
	{
		it=LL_iterator(CONFSYM_table->connect_l);
		while(NULL!=LLI_next(it))
		{
			LL_append(ret->connects,LLI_elem(it));
			ret->connect_count++;
		}

		LLI_destroy(&it);
	}
	LL_destroy(&(CONFSYM_table->connect_l));

	/* get the globals */
	ret->global=CONFSYM_table->global;

	free(CONFSYM_table);
	CONFSYM_table=NULL;

	return ret;
}

bool
CONFPRV_parseGlobal(char *keyword,char *value,struct CONF_global *global)
{
	if(0==strcmp(keyword,"user"))
	{
		global->user=value;
	}
	else if(0==strcmp(keyword,"group"))
	{
		global->group=value;
	}
	else if(0==strcmp(keyword,"socket-path"))
	{
		global->sockPath=value;
	}
	else if(0==strcmp(keyword,"pid-path"))
	{
		global->pidPath=value;
	}
    else if (0==strcmp(keyword, "mark-interval"))
    {
        /* 
         * Internally the mark interval must be in seconds, but is
         * specified in minutes in the config file.
         */
        global->mark_interval = atoi(value) * 60;
        /* We don't need the value any more */
        free(value);
    }
	else		/* error */
	{
		return FALSE;
	}

	return TRUE;
}

bool
CONFPRV_parseGlobalIntVal(char *keyword, int value, struct CONF_global *global)
{
    if (0==strcmp(keyword, "mark-interval"))
    {
        /* 
         * Internally the mark interval must be in seconds, but is
         * specified in minutes in the config file.
         */
        global->mark_interval = value * 60;
    }
    else 
    {
        return FALSE;
    }

    return TRUE;
}



netaddr_t
CONFPRV_parseIP(char *address)
{
	netaddr_t temp;

	temp = NA_newINetLookup(address);
	if(!NA_isError(temp))
		return temp;
	else NA_destroy(temp);

	/* we have an error */
	return NULL;
}

