/**
 * @file conf_private.h
 *
 * @brief Private functions for the parser (symbol table, special prints 
 * for debugging, etc.)
 *
 * @ingroup conf_parse
 *
 * @todo Support ACLs and connections
 *
 * $Id: conf_private.h,v 1.17.2.6 2003/04/30 20:03:06 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Apr 23 15:31:03 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CONF_CONF_PRIVATE_H
#define CONF_CONF_PRIVATE_H

#include <stdio.h>

#include "priority.h"

/**@defgroup conf_parse Configuration Parsing
 * @ingroup conf
 * @{
 */

/** Return both a path and a netaddr_t */
struct CONFPRV_addr
{
	netaddr_t addr;
	char *name;
};

/** Structure used for the version numbers; contains major & minor */
struct CONFPRV_version
{
	int major;		/**< The major version */
	int minor;		/**< The minor version */
};

/** Recognized major version */
#define CONFPRV_majver	1

/** Maximum minor version */
#define CONFPRV_minver	1

/** Get the current line number
 *
 * @return The current line number
 */
int CONFPRV_lineno(void);

/** Get the current file name
 *
 * @return The current file name
 */
char *CONFPRV_fname(void);

/** Parse a global option
 *
 * @param keyword The keyword (for instance, 'user' or 'group', without the ')
 * @param value The value to associate with it (with no quotation marks)
 * @param global The global struct (this must be previously allocated)
 *
 * @return TRUE if this was successful, FALSE if there was an error
 */
bool CONFPRV_parseGlobal(char *keyword,char *value,struct CONF_global *global);

/** Parse a global option With an integer value.
 *
 * @param kw The keyword (for instance, 'user' or 'group', without the ')
 * @param val The value to associate with it (with no quotation marks)
 * @param global The global struct (this must be previously allocated)
 *
 * @return TRUE if this was successful, FALSE if there was an error
 */
bool CONFPRV_parseGlobalIntVal(char *kw,int val,struct CONF_global *global);

/** Parse an address of some sort into some sort of IP (either v4 or v6)
 *
 * @param address The address to translate
 *
 * @return A new netaddr_t, or NULL on error
 */
netaddr_t CONFPRV_parseIP(char *address);

/** Define yy* to conf_* */
#define yyprep conf_prep
#define yyreset conf_reset
#define yystop conf_stop
#define yyparse conf_parse

/**
 *	Prepare the parser/lexer
 *	(#defined to conf_prep())
 *
 *	@param input_file The file to be read
 *	@param filename The filename
 */
void yyprep(FILE *input_file,char *filename);

/**
 *	Reset the parser/lexer
 *	(#defind to conf_reset())
 *
 * @param input_file The new file
 * @param filename The new filename
 */
void yyreset(FILE *input_file,char *filename);

/**
 *	Ends the parser/lexer (free()s a few variables
 *	(#defined to conf_stop())
 */
void yystop(void);

/**
 *	Runs the parser (created by Yacc/Lex from confparse.l and confparse.y)
 *	(#defined to conf_parse())
 *
 *	@return !0 on succees, 0 on failure
 */
int yyparse(void);

/** Maps facility and severity names to numbers */
struct CONFFS_str
{
	char 	*name;  /**< The name of the object in question */
	int 	fs;		/**< The severity or facility associated with the name */
};

/*@}*/


/**@defgroup conf_sym Config Parser Symbol Table
 * @ingroup conf
 * @{
 */

/** This data structure holds all information about the symbol table
 */
struct CONFSYM_str
{
	List acl_l;
	int acl_count;

	List input_l;
	int input_count;

	List output_l;
	int output_count;

	List connect_l;
	int connect_count;

	struct CONF_global *global;
};

/** Destroy the symbol table (held in the global CONFSYM_table)
 */
void CONFSYM_destroy(void);

/** Initialize the symbol table (held in the global CONFSYM_table)
 */
void CONFSYM_init(void);

/** Add an input structure to the symbol table
 *
 * @param in The input structure to add
 */
void CONFSYM_addInput(struct inchan_info *in);

/** Gets an input structure (with a specified name) from the symbol table
 *
 * @param name The name to search for
 *
 * @return The input structure that was matched, or NULL if none was found
 */
struct inchan_info *CONFSYM_getInput(const char *name);

/** Add an output structure to the symbol table
 *
 * @param in The output structure to add
 */
void CONFSYM_addOutput(struct outchan_info *in);

/** Gets an output structure (with the specified name) from the symbol table
 *
 * @param name The name to search for
 *
 * @return The output structure that was matched, or NULL if none was found
 */
struct outchan_info *CONFSYM_getOutput(const char *name);

/** Add a connect structure to the symbol table
 *
 * @param in The connect structure to add
 */
void CONFSYM_addConnect(struct CONF_con *in);

/** Gets a connection structure (with either a specified input name or output
 *  name) from the symbol table.
 *  When both parameters are NULL or non-NULL, NULL will be returned (in the 
 *  first case, no matches can be found to nothing; in the second, we're not
 *  sure which pattern actually needs to be matched.)
 *
 *  @param input The input name to search for (may be NULL)
 *  @param output The output name to search for (may be NULL)
 *
 *  @return The found connect struct (or a NULL on failure; if an entry couldn't
 *  be found, if both parameters were NULL)
 */
struct CONF_con *CONFSYM_getConnect(char *input,char *output);

/** Adds a global object to the symbol table
 * This will fail if an object has already been added
 *
 * @param in The pointer to the global object to add
 *
 * @return TRUE on success, FALSE on failure
 */
bool CONFSYM_addGlobal(struct CONF_global *in);

/** Removes a global object from the symbol table
 *
 * @return The global object, or NULL if none was present
 */
struct CONF_global *CONFSYM_getGlobal(void);

/** Gets a struct config_info from the symbol table.
 *
 * @pre If name is not NULL, all lists must have been initialized
 *
 * @post The global CONFSYM_table WILL be NULL
 * @warn This function destroys the symbol table
 *
 * @return a pointer config_info struct, containing all the information
 * 			read in from the configuration file; or NULL if the symbol table
 * 			was NULL
 */
struct config_info *CONFSYM_getInfo(void);

/*@}*/


/**@defgroup conf_debug Config Parser Debug Stuff
 * @ingroup conf
 * @{
 */

/** Dump a struct config_info to the screen
 * Recursively dumps any sub-objects of known type
 * 
 * @param in The struct config_info to dump
 */
void dump_config_info(struct config_info *in);

/** Dump a struct CONF_input to the screen
 *
 * @param in The struct CONF_input to dump
 */
void dump_inchan_info(struct inchan_info *in);

/** Dump a struct outchan_info to the screen
 *
 * @param in The struct outchan_info to dump
 */
void dump_outchan_info(struct outchan_info *in);

/* end conf_impl_debug */
/*@}*/

#endif
