/**
 * @file config.h
 *
 * @brief Define the data structure and functions to support the config
 * file.
 *
 * @ingroup conf
 *
 * $Id: config.h,v 1.30.2.6 2003/04/30 20:03:06 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Jan 25 13:28:45 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CONF_CONFIG_H
#define CONF_CONFIG_H

#include "util/list.h"
#include "util/acl.h"
#include "outputmod/outputmod.h"
#include "inputmod/inputmod.h"

#include "util/netaddr.h"

#include "const.h"

/* Some groups for documentation */
/**@defgroup conf Config File Library */

/**@addtogroup conf
 * @{
 */

/** A structure that holds connections between inputs and outputs.  */
struct CONF_con
{
    char  *input;			/**< Input name */
    char *output;			/**< Output name */
};

/** A structure that holds information about ACLs */
struct CONF_acl
{
    int     dummy;
};

/** A structure that holds global information about the configuration */
struct CONF_global
{
	char *user;			/**< The user to run as; stored as a string, must be free()ed */
	char *group;		/**< The group to run as; stored as a string, must be free()ed */

    char const  *sockPath;      /**< Path to put the IPC socket at */
	char const	*pidPath;		/**< Path to put the PID file at */

	char *stdredir;		/**< Where stdout and stderr are redirected */

    int mark_interval;  /**< The interval between marks. */
};

/** This data structure stores all the relevant configuration info
 */
struct config_info 
{
	int			input_count;	/**< Number of inputs */
	List		inputs;			/**< Syslog inputs (elem type: CONF_input) */

	int			output_count;	/**< Number of outputs */
	List		outputs;		/**< Syslog outputs (elem type: outchan_info) */

	int			connect_count;	/**< Number of connections */
	List		connects;		/**< input-output connections 
								 *   (elem type: CONF_con)
								 */

	int			acl_count;		/**< Number of ACLs in the system */
	List		acls;			/**< List of ACLs in the system (elem type: CONF_acl) */

    
    /**@name Global Options 
     * @{
     */
	struct CONF_global *global;	/**< Structure with misc. globals */
    /*@}*/
};

/** Read a config file and fill in a config_info structure. 
 *  @warning The returned pointer must be destroyed with 
 *      destroyConfigInfo(), this is the only way to assure proper 
 *      cleanup.  And make sure you are done with it first.
 *
 *  @param file The file name of the config file.
 *  @return a pointer to the filled in config_info structure, or NULL on
 *          error.
 */
struct config_info *getConfigInfo(const char *file);

/** Destroy a config_info object.  This is the only way to properly
 *  clean up a config_info pointer that was returned by getConfigInfo().
 *  
 *  @param obj The config_info* to destroy.
 */
void destroyConfigInfo(struct config_info **obj);

/*@}*/

#endif /* CONF_CONFIG_H */
