/**
 * @file conflex.l
 *
 * @brief Lex rules to lex the input for config.y
 * @ingroup conf_parse
 *
 *
 * $Id: conflex.l,v 1.29.2.6 2003/04/30 20:03:06 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Apr 11 13:47:21 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

%{
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <strcompat.h>
#include "error.h"

#include "util/list.h"
#include "util/netaddr.h"
#include "chaninfo.h"

#include "config.h"
#include "conf_private.h"
#include "confparse.h"

/* to match the decl in confparse.c (generated from confparse.y) */
#define	yylval conf_lval
#define yyprep conf_prep
#define yyreset conf_reset
#define yystop conf_stop
#define yyin conf_in
#define yyerror conf_error

/* prototypes of functions created by yacc/bison */
void yyerror(char *s);

/** Holds the current line number being parsed */
static int CONFPRV_int_lineno=1;

/** Global variable; holds the current file being parsed */
static char *CONFPRV_int_fname=NULL;
%}

/* either this, or actually define the function to return 1 */
%option noyywrap

/* disable yyunput(), which just errors */
%option nounput

/* specify yytext as a pointer; this prevents some buffer overflows */
%pointer

/* convenient macros */
DIGIT		[0-9]
ALPHA		[a-zA-Z]
ALPHANUM	{DIGIT}|{ALPHA}

%%

#					{
						/* trim out the comment; the rest of the line is garbage */
						int c;

						while(0!=(c=input()))
						{
							if(c=='\n')
							{
								CONFPRV_int_lineno++;
								break;
							}
						}
					}

[ \t]+			/* whitespace; do nothing */
\n					{	CONFPRV_int_lineno++;	}

\;					{	return T_SEMI;	}
\{					{	return T_OPENSCOPE;	}
\}					{	return T_CLOSESCOPE;	}
\.					{	return T_DOT; }
\,					{	return T_COMMA; }
\=					{	return T_EQUALS; }
\-					{	return T_DASH; }
\*					{	return T_STAR; }

"input"				{	return T_INPUT;	}
"output"			{	return T_OUTPUT; }
"listen"			{	return T_LISTEN; }
"port"				{	return T_PORT; }
"format-type"		{	return T_FORMAT; }
"format-options"    {   return T_FOPTS; }
"file-version"		{	return T_VERSION; }
"type"				{	return T_TYPE;	}
"output-path"		{	return T_OUTPATH; }
"connect"			{	return T_CONNECT; }
"to"				{	return T_TO; }
"priority"			{	return T_PRIORITY; }

"globals"			{	return T_GLOBAL; }

{DIGIT}+			{
						yylval.ival=atoi(yytext);

						return T_INT;
					}

{DIGIT}+\.{DIGIT}+\.({DIGIT}|\.)*		{
						/* match IPv4 addresses, not in quotes */
						char *tempstr;


						tempstr=emalloc(strlen(yytext)+1);
						strlcpy(tempstr,yytext,strlen(yytext)+1);

						yylval.sval=tempstr;
						return T_ID;
					}


\"[^ \t\n\"\{\}\;]+\"			{
						char *tempstr;

						tempstr=emalloc(strlen(yytext)+1-2);

						strlcpy(tempstr,&(yytext[1]),strlen(yytext)+1-2);

						yylval.sval=tempstr;
						return T_STRING;
					}

({ALPHA}|\/)[^ \t\n\"\{\}\;\.\,#]+	{
						/* the string is malloc()ed here; it MUST be free()ed later */
						char *tempstr;

						tempstr=emalloc(strlen(yytext)+1);

						strlcpy(tempstr,yytext,strlen(yytext)+1);

						yylval.sval=tempstr;
						return T_ID;
					}


<*>.|\\n			{
						/* the default rule, we only error */

						/* 
						 right now, we dont' do anything; it
						 would be possible to print out an error message here,
						 though
						*/
					}
%%

void
yyprep(FILE *input_file,char *filename)
{
	CONFPRV_int_lineno=1;

	if(CONFPRV_int_fname!=NULL) {
		free(CONFPRV_int_fname);
        CONFPRV_int_fname = NULL;
    }

	CONFPRV_int_fname=emalloc(strlen(filename)+1);
	strlcpy(CONFPRV_int_fname,filename,strlen(filename)+1);

	yyin=input_file;

	CONFSYM_init();
}

void
yyreset(FILE *input_file,char *filename)
{
	yyrestart(input_file);

	CONFPRV_int_lineno=1;
	free(CONFPRV_int_fname);
	CONFPRV_int_fname=emalloc(strlen(filename)+1);
	strlcpy(CONFPRV_int_fname,filename,strlen(filename)+1);

	CONFSYM_destroy();
	CONFSYM_init();
}

void
yystop(void)
{
	free(CONFPRV_int_fname);
    CONFPRV_int_fname = NULL;
	fclose(yyin);
}

int
CONFPRV_lineno(void)
{
	return CONFPRV_int_lineno;
}

char *
CONFPRV_fname(void)
{
	return CONFPRV_int_fname;
}
