/**
 * @file confparse.y
 *
 * @brief Bison/Yacc rules to parse a config file.
 * @ingroup conf_parse
 *
 *
 * $Id: confparse.y,v 1.45.2.7 2003/04/30 20:03:07 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Apr 11 13:47:21 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* top-of-the-file C decls */
%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "error.h"
#include "strcompat.h"
#include "log.h"
#include "util/netaddr.h"
#include "util/list.h"
#include "util/acl.h"
#include "util/bitvec.h"

#include "message/msg_frmt.h"

#include "chaninfo.h"
#include "config.h"
#include "conf_private.h"
#include "conf_obj.h"

#include "debug.h"

static char const rcsid[]
    = "$Id: confparse.y,v 1.45.2.7 2003/04/30 20:03:07 devink Exp $";

/* to match defines generated by lex/flex */
#define yylex conf_lex

/* prototypes for functions created by lex/flex */
int yylex(void);
void yyerror(char *s);

#if ENABLE_DEBUG
/* enable the possibility of debugging, turn on with 'yydebug=1' */
#define YYDEBUG 1

/* There are still some problems with YACC debugging, like where to turn
 * on yydebug (conf_debug here) and how to do it in such a way that
 * doesn't break unit tests without libconfig linked in.
 */
#if 0
#define YYFPRINTF(junk, fmt, ...)   \
    DBG_print(DBG_CNF, fmt, __FILE__, __LINE__, ##__VA_ARGS__)
#endif
#endif /* ENABLE_DEBUG */

/** Severity map; see CONFFS_sevlen for the length of the array */
struct CONFFS_str CONFFS_sev[] =
{
	{ "emerg", CONFFS_EMERG, },		/**< Emergency severity */
	{ "panic", CONFFS_EMERG, },		/**< Deprecated; Emergency severity */
	{ "alert", CONFFS_ALERT, },		/**< Alert severity */
	{ "crit", CONFFS_CRIT, },		/**< Critical severity */
	{ "error", CONFFS_ERR, },		/**< Deprecated; Error severity */
	{ "err", CONFFS_ERR, },			/**< Error severity */
	{ "warn", CONFFS_WARN, },		/**< Warning severity */
	{ "warning", CONFFS_WARN, },	/**< Deprecated; Warning severity */
	{ "notice", CONFFS_NOTICE, },	/**< Notice severity */
	{ "info", CONFFS_INFO, },		/**< Informational severity */
	{ "debug", CONFFS_DEBUG, },		/**< Debug severity */
	{ "none", -1, },				/**< Internal; no severity */
};

/** Length of the CONFFS_sev */
#define CONFFS_sevlen	(sizeof(CONFFS_sev)/sizeof(struct CONFFS_str))

/** Facility map; see CONFFS_faclen for the length of the array */
struct CONFFS_str CONFFS_fac[] =
{
    { "kern", CONFFS_KERN, },       /**< Map 'kern' to kernel facility */
	{ "kernel", CONFFS_KERN, },		/**< Alias: 'kernel' -> 'kern' */
	{ "user", CONFFS_USER, },		/**< Map 'user' to user facility */
	{ "mail", CONFFS_MAIL, },		/**< Map 'mail' to mail facility */
	{ "daemon", CONFFS_DAEMON, },	/**< Map 'daemon' to daemon facility */
	{ "auth", CONFFS_AUTH0, },		/**< Map 'auth' to auth0 facility */
	{ "auth0", CONFFS_AUTH0, },		/**< Map 'auth0' to auth0 facility */
	{ "syslog", CONFFS_SYSLOG, },	/**< Map 'syslog' to syslog facility */
	{ "lpr", CONFFS_LPR, },			/**< Map 'lpr' to LPR facility */
	{ "news", CONFFS_NEWS, },		/**< Map 'news' to news facility */
	{ "uucp", CONFFS_UUCP, },		/**< Map 'uucp' to UUCP facility */
	{ "cron", CONFFS_CRON0, },		/**< Map 'cron' to cron0 facility */
	{ "cron0", CONFFS_CRON0, },		/**< Map 'cron0' to cron0 facility */
    { "authpriv", CONFFS_AUTH1},    /**< Linux: 'authpriv' -> auth1 */
	{ "auth1", CONFFS_AUTH1, }	,	/**< Map 'auth1' to auth1 facility */
	{ "ftp", CONFFS_FTP, },			/**< Map 'ftp' to ftp facility */
	{ "ntp", CONFFS_NTP, },			/**< Map 'ntp' to ntp facility */
	{ "audit", CONFFS_AUDIT0, },	/**< Map 'audit' to audit0 facility */
	{ "audit0", CONFFS_AUDIT0, },	/**< Map 'audit0' to audit0 facility */
	{ "audit1", CONFFS_AUDIT1, },	/**< Map 'audit1' to audit1 facility */
	{ "cron1", CONFFS_CRON1, },		/**< Map 'cron1' to cron1 facility */
	{ "local0", CONFFS_LOCAL0, },	/**< Map 'local0' to local0 facility */
	{ "local1", CONFFS_LOCAL1, },	/**< Map 'local1' to local1 facility */
	{ "local2", CONFFS_LOCAL2, },	/**< Map 'local2' to local2 facility */
	{ "local3", CONFFS_LOCAL3, },	/**< Map 'local3' to local3 facility */
	{ "local4", CONFFS_LOCAL4, },	/**< Map 'local4' to local4 facility */
	{ "local5", CONFFS_LOCAL5, },	/**< Map 'local5' to local5 facility */
	{ "local6", CONFFS_LOCAL6, },	/**< Map 'local6' to local6 facility */
	{ "local7", CONFFS_LOCAL7, },	/**< Map 'local7' to local7 facility */
    { "mark", CONFFS_MARK},         /**< Map 'mark' to internal mark facility*/
};

/** Length of the CONFFS_fac array */
#define CONFFS_faclen	(sizeof(CONFFS_fac)/sizeof(struct CONFFS_str))

/**@name tri-state values used to build the priority_list. 
 * @{
 */
#define PRI_UNSET   0   /**< No value */
#define PRI_PASS    1   /**< This priority gets sent out */
#define PRI_DROP    2   /**< This priority will be dropped */
/*@}*/

/** Error string */
static char CONFPRV_errstr[1024];

/**
 * Function to parse Severity ranges.  This is needed in this file due
 * to needing to access CONFFS_sev and since T_ID can contain a '-' it
 * is a large chunck of code that I'd rather not just duplicate in the 
 * severity_list rule.
 *
 * @param rstr The right side severity.
 * @param lstr The Left side severity.
 * @param pris The priority array.
 * @return 0 on success, non-zero other wise.
 */
int parseSevRange(char const *rstr, char const *lstr, int* pris);

%}

/* Bison decls */
%union {
	char 	*sval;
	int		ival;
	int 	*ipntr;
	float	fval;
    List    list;
	struct	inchan_info *input_conf;
	struct	outchan_info *output_conf;
	struct	CONF_con *connect_conf;
	struct	CONF_global *global_conf;
	struct	CONFPRV_version ver;
}

%token <sval>	T_ID
%token <sval>	T_STRING
%token <fval>	T_FLOAT
%token <ival>	T_INT

%token T_OPENSCOPE
%token T_CLOSESCOPE
%token T_SEMI
%token T_QUOTE
%token T_COMMA
%token T_DOT
%token T_EQUALS
%token T_DASH
%token T_STAR

/* keywords */
%token T_OUTPUT
%token T_INPUT
%token T_ACL
%token T_VERSION

%token T_TYPE
%token T_LISTEN
%token T_PORT
%token T_FORMAT
%token T_FOPTS
%token T_OUTPATH
%token T_PRIORITY

%token T_CONNECT
%token T_TO

%token T_GLOBAL
%token T_USER
%token T_GROUP
%token T_DBGFILE
%token T_SOCKPATH

/* non-terminals with types */
%type <sval> type_decl
%type <sval> format_decl
%type <sval> listen_decl
%type <ival> port_decl
%type <sval> outpath_decl
%type <list> outpath_list_decl
%type <list> outpath_list
%type <ver> version
%type <input_conf> input_decls
%type <input_conf> input
%type <output_conf> output_decls
%type <output_conf> output
%type <connect_conf> connect
%type <global_conf> global
%type <global_conf> global_decls
%type <ipntr> priority_decl
%type <ipntr> facility_list
%type <ipntr> severity_list
%type <ipntr> priority_list
%type <ival> format_opts_decl
%type <ival> format_opt_list

/* grammar rules */
%start cfg
%%

/**@todo Add a catch so that a useful message is printed when the file
 * lacks a version keywork.
 */
cfg :
	  version 
		{
			/* if the version isn't recognized, exit */
            if ($1.major != CONFPRV_majver) {
                snprintf(CONFPRV_errstr, 1024, 
                    "Required Major Version %d was not found."
                    " Config file has version %d.%d. Config files are"
                    " not compatable across major versions", 
                    CONFPRV_majver, $1.major, $1.minor);
                yyerror(CONFPRV_errstr);
                return 1;
            }
            else if ($1.minor > CONFPRV_minver) {
                snprintf(CONFPRV_errstr, 1024, 
                    "Config file's minor version is greater than"
                    " expected.  Lastest parsable version is %d.%d,"
                    " config file is version %d.%d",
                    CONFPRV_majver, CONFPRV_minver, $1.major, $1.minor);
                yyerror(CONFPRV_errstr);
                return 1;
            }
            else if ($1.minor < CONFPRV_minver) {
                LOG_print(LOG_INFO, "Config file out of date but"
                    " compatable.  Note that some options might be"
                    " depreciated. Lastest parser version is %d.%d",
                    CONFPRV_majver, CONFPRV_minver);
            }
		}
	  definitions
	  	{
		}
	;

version :
	  T_VERSION T_INT T_DOT T_INT T_SEMI
		{
			$$.major=$2;
			$$.minor=$4;
		}
	;

definitions :
	  input 
	  	{
			CONFSYM_addInput($1);
		}
	  T_SEMI definitions
	| output
		{
			CONFSYM_addOutput($1);
		}
	  T_SEMI definitions
	| connect
		{
			CONFSYM_addConnect($1);
		}
 	  T_SEMI definitions
	| global
		{
			if(!CONFSYM_addGlobal($1))
			{
				yyerror("Multiple global declarations found");

				return 1;
			}
		}
 	  T_SEMI definitions
	|		/* empty */
	;

input :
	  T_INPUT T_ID T_OPENSCOPE input_decls T_CLOSESCOPE
	  	{
			$4->name=$2;

			/* we must have some sort of an input path; if there's no default,
			 * and one wasn't specified, we error. The expection is the
             * mark input module which has no input path.
             */
			if(NULL==$4->inPath && $4->type != IMOD_MARK)
			{
				snprintf(CONFPRV_errstr,1024,
						"listen option required in input %s",
						$4->name);
				yyerror(CONFPRV_errstr);
				return 1;
			}

			/* also, let's create the addr, depending on the type */
			switch ($4->type)
			{
				case IMOD_UDP:
				case IMOD_BEEP:
					if(NULL==($4->inAddr=CONFPRV_parseIP($4->inPath)))
					{
						snprintf(CONFPRV_errstr,1024,
							"invalid address (%s) in input %s",
							$4->inPath,$4->name);
						yyerror(CONFPRV_errstr);
						return 1;
					}

					break;

				case IMOD_SYSTEM:
					$4->inAddr=NA_newUnix($4->inPath);
					
					break;
                case IMOD_MARK:
                    /* NOTHING TO DO */
                    break;

				default:
					snprintf(CONFPRV_errstr,1024,
							"unrecognized input type (%d) in input %s",
							$4->type,$4->name);
					yyerror(CONFPRV_errstr);
					return 1;
					break;	/* never reached */
			}

			$$=$4;
		}
	;

input_decls :
	  type_decl T_SEMI input_decls
	  	{
			/* detect the type, and add any defaults */
			if(0==strncmp($1,"beep",4))
			{
				$3->type=IMOD_BEEP;

				if(-1==$3->inPort)
					$3->inPort=SYSLOG_CONN_PORT;
				if(NULL==$3->inPath)
				{
					$3->inPath=emalloc(sizeof(char)*(strlen("0.0.0.0")+1));
					strlcpy($3->inPath,"0.0.0.0",strlen("0.0.0.0")+1);
				}
			}
			else if(0==strncmp($1,"udp",4))
			{
				$3->type=IMOD_UDP;

				if(-1==$3->inPort)
					$3->inPort=SYSLOG_UDP_PORT;
				if(NULL==$3->inPath)
				{
					$3->inPath=emalloc(sizeof(char)*(strlen("0.0.0.0")+1));
					strlcpy($3->inPath,"0.0.0.0",strlen("0.0.0.0")+1);
				}
			}
			else if(0==strncmp($1,"system",7))
			{
				$3->type=IMOD_SYSTEM;

				if(NULL==$3->inPath)
				{
					$3->inPath=emalloc(sizeof(char)*
							(strlen(DEFAULT_SYS_SOCKET)+1));
					strlcpy($3->inPath,DEFAULT_SYS_SOCKET,
							strlen(DEFAULT_SYS_SOCKET)+1);
				}
			}
            else if (0==strncmp($1, "mark", 5))
            {
                $3->type=IMOD_MARK;
            }
			else	/* we need to report an error */
			{
				snprintf(CONFPRV_errstr,1024,"invalid input type: \"%s\"",$1);
				yyerror(CONFPRV_errstr);

				return 1;
			}

			free($1);
			$$=$3;
		} 
	| listen_decl T_SEMI input_decls
		{
			if(NULL!=$3->inPath)
				free($3->inPath);

			$3->inPath=$1;

			$$=$3;
		}
	| port_decl T_SEMI input_decls
		{
			$3->inPort=$1;

			$$=$3;
		}
	|		/* empty */
		{
			$$=CONFOBJ_createInput();
		}
	;

output :
	  T_OUTPUT T_ID T_OPENSCOPE output_decls T_CLOSESCOPE
	  	{
			$4->name=$2;

            /* Check for and log and clear invalid formatter options 
             * Bad options are not considerd fatal, so we just ignore
             * the return.
             */
            MFRMT_checkOptions(&($4->opts->format), 
                CONFPRV_fname(), CONFPRV_lineno());

            /* Do final output module specific validation of the
             * options.  A FALSE return indicates that there was a big
             * problem, and we are going to bail.
             */
            if (! OMOD_processOptions($4->type, $4->opts, 
                    CONFPRV_fname(), CONFPRV_lineno()))
            {
                snprintf(CONFPRV_errstr, 1024,
                    "Fatal error processing options for output module %s",
                    $4->name);
                yyerror(CONFPRV_errstr);
                return 1;
            }

			$$=$4;
		}
	;

output_decls :
	  type_decl T_SEMI output_decls
	  	{
            $3->type = OMOD_getTypeByName($1);

            if ($3->type == OMOD_COUNT) 
			{
				snprintf(CONFPRV_errstr,1024,"invalid output type: \"%s\"",$1);
				yyerror(CONFPRV_errstr);

				/* return with an error */
				return 1;
			}

			free($1);
			$$=$3;
		}
	| format_decl T_SEMI output_decls
		{
            MFRMT_TYPE  type;

            if ($3->opts->format.type != 0) 
            {
                yyerror("Can't specify two formatters in one output block");
                return 1;
            }

            type = MFRMT_getType($1);
            if (type == MFRMT_NOEXIST)
			{
				snprintf(CONFPRV_errstr,1024,"invalid format type: \"%s\"",$1);
				yyerror(CONFPRV_errstr);

				/* return with an error */
				return 1;
			}

            /* Set the type */
            $3->opts->format.type = type;

			free($1);
			$$=$3;
		}
    | format_opts_decl T_SEMI output_decls
        {
            $3->opts->format.options = BM_UNION($3->opts->format.options, $1);
            $$=$3;
        }
	| outpath_decl T_SEMI output_decls
		{
            DPRINT(DBG_CNF, "Reducing outpath_decl, String %s", $1);
			$3->opts->outPath=$1;

			$$=$3;
		}
    | outpath_list_decl T_SEMI output_decls
        {
            DPRINT(DBG_CNF, "Reducing outpath_list_decl", NO_VARG);
            $3->opts->outList=$1;

            $$=$3;
        }
	| priority_decl T_SEMI output_decls
		{
			int i;

			for(i=0;i<PRIORITY_COUNT;i++)
			{
				if($1[i])
					$3->pri[i]=1;
			}

			$$=$3;
            free($1);
		}
	| port_decl T_SEMI output_decls
		{
			$3->opts->outPort=$1;

			$$=$3;
		}
	|		/* empty */
		{
			$$=CONFOBJ_createOutput();
		}
	;

type_decl :
	  T_TYPE T_ID
		{
			$$=$2;
		}
	;

outpath_decl :
	  T_OUTPATH T_ID
	  	{
            DPRINT(DBG_CNF, "outpath_decl: Got ID %s", $2);
			$$=$2;
		}
	| T_OUTPATH T_STRING
		{
            DPRINT(DBG_CNF, "outpath_decl: Got string: %s", $2);
			$$=$2;
		}
    ;

outpath_list_decl :
      T_OUTPATH T_OPENSCOPE outpath_list T_CLOSESCOPE
        {
            $$=$3;
        }
	;

outpath_list :
      T_ID T_SEMI outpath_list
        {
            DPRINT(DBG_CNF, "outpath_list: Adding id to list: %s", $1);
            LL_append($3, $1);

            $$ = $3;
        }
    | T_STRING T_SEMI outpath_list
        {
            DPRINT(DBG_CNF, "outpath_list: Adding string to list: %s", $1);
            LL_append($3, $1);
            $$ = $3;
        }
    | /* EMPTY */
        {
            DPRINT(DBG_CNF, "outpath_list: new list", NO_VARG);
            $$ = LL_create();
        }
    ;


priority_decl :
	  T_PRIORITY T_OPENSCOPE priority_list T_CLOSESCOPE
	  	{
            int   i;
            /* this does final cleanup on the priority list, which is
             * supposed to only contain TRUE or FALSE.  
             */
            for (i=0; i < PRIORITY_COUNT; i++) {
                if ($3[i] == PRI_UNSET || $3[i] == PRI_DROP)
                    $3[i] = 0;  /* PRI_UNSET or PRI_DROP -> FALSE */
                else
                    $3[i] = 1;  /* PRI_PASS -> TRUE */
            }
	  		$$=$3;
		}
	;

priority_list :
	  facility_list T_DOT severity_list T_SEMI priority_list
		{
			int     sev, fac;

			for(sev=0;sev<CONFFS_SEVCOUNT;sev++)
			{
                for(fac=0;fac<CONFFS_FACCOUNT;fac++)
                {
                    /* only change the priority list if it contains a 
                     * value of 'UNSET'. 
                     */
                    if($1[fac] && $5[fac*8+sev] == PRI_UNSET)
                    {
                        /* The state of this priority can be found in
                         * the assocaited severity, which will be either
                         * PRI_UNSET, PRI_DROP, or PRI_PASS, The
                         * severity is left to figure this out b/c it
                         * has the special value of 'none'
                         */ 
                        $5[fac*8+sev] = $3[sev];
                        DPRINT(DBG_CNF, "Setting %d, %d to %d",
                            fac, sev, $3[sev]);
                    }
				}
			}

			free($1);
			free($3);

			$$=$5;
		}
	|		/* empty */
		{
			$$=emalloc(sizeof(int)*PRIORITY_COUNT);
			memset($$,0,sizeof(int)*PRIORITY_COUNT);
		}
	;

facility_list :
	  T_ID
	  	{
			int i=0;

			$$=emalloc(sizeof(int)*CONFFS_FACCOUNT);
			memset($$,0,CONFFS_FACCOUNT*sizeof(int));

			for(i=0;i<CONFFS_faclen;i++)
			{
				if(0==strncmp($1,CONFFS_fac[i].name,
						strlen(CONFFS_fac[i].name)+1))
				{
					break;
				}
			}

			/* did we not match anything? */
			if(CONFFS_faclen==i)
			{
				snprintf(CONFPRV_errstr,1024,"Unknown facility: %s",$1);

				yyerror(CONFPRV_errstr);

				free($$);
				return 1;		/* return an error */
			}
			/* add our new element */
			else
			{
				$$[CONFFS_fac[i].fs]=1;
			}

			free($1);
		}
	| T_STAR
		{
			int i;

			$$=emalloc(sizeof(int)*CONFFS_FACCOUNT);
			memset($$,0,CONFFS_FACCOUNT*sizeof(int));	

			for(i=0;i<CONFFS_FACCOUNT;i++)
				$$[i]=1;
		}
	| T_ID T_COMMA facility_list
	  	{
			int i=0;

			if(NULL==$3)
			{
				$3=emalloc(sizeof(int)*CONFFS_FACCOUNT);
				memset($$,0,CONFFS_FACCOUNT*sizeof(int));
			}

			for(i=0;i<CONFFS_faclen;i++)
			{
				if(0==strncmp($1,CONFFS_fac[i].name,
						strlen(CONFFS_fac[i].name)+1))
				{
					break;
				}
			}

			/* if we went to far, error */
			if(CONFFS_faclen==i)
			{
				snprintf(CONFPRV_errstr,1024,"Unknown facility: %s",$1);
				yyerror(CONFPRV_errstr);

				free($3);

				return 1;
			}
			/* add the single element */
			else
			{
				$3[CONFFS_fac[i].fs]=1;
			}

			free($1);
			$$=$3;
		}
	;

severity_list :
	  T_ID
	  	{
			int     i;
            char    *p;
            bool    done;

			$$=emalloc(sizeof(int)*CONFFS_SEVCOUNT);
			memset($$,0,CONFFS_SEVCOUNT*sizeof(int));

            /* Check for a dash which would make this a range */
            done = FALSE;
            for (p = $1; *p != '\0'; p++) 
            {
                if (*p == '-') 
                {
                    *p = '\0';
                    p++;
                    done = TRUE;    /* one way or another... */
                    if (parseSevRange($1, p, $$) != 0) 
                    {
                        free($$);
                        return 1;
                    }

                }
            }

            if (!done) 
            {

                for(i=0;i<CONFFS_sevlen;i++)
                {
                    if(0==strncmp($1,CONFFS_sev[i].name,
                            strlen(CONFFS_sev[i].name)+1))
                    {
                        break;
                    }
                }

                /* do we have an error? */
                if(CONFFS_sevlen==i)
                {	
                    snprintf(CONFPRV_errstr,1024,"Unknown severity: %s",$1);
                    yyerror(CONFPRV_errstr);
                    free($$);

                    return 1;
                }
                /* add everything less than this */
                else
                {
                    int j=CONFFS_sev[i].fs;
                    int val;

                    DPRINT(DBG_CNF, "sev: .%s is good, adding it", $1);

                    /* The 'none' severity tweaks all of them 
                     * Also For a severity of none, we want to flag that all
                     * the associated facilities should be droped, and
                     * not passed.
                     */
                    if (j == CONFFS_NONE) {
                        j = CONFFS_SEVCOUNT - 1;
                        val = PRI_DROP;
                    }
                    else {
                        val = PRI_PASS;
                    }

                    for(;j>=0;j--)
                    {
                        $$[j] = val;
                        DPRINT(DBG_CNF, "sev .%s: %d -> %d",
                            $1, j, $$[j]);
                    }
                }
            }
                
			free($1);
		}
	| T_EQUALS T_ID
	  	{
			int i;

			$$=emalloc(sizeof(int)*CONFFS_SEVCOUNT);
			memset($$,0,CONFFS_SEVCOUNT*sizeof(int));

			for(i=0;i<CONFFS_sevlen;i++)
			{
				if(0==strncmp($2,CONFFS_sev[i].name,
						strlen(CONFFS_sev[i].name)+1))
				{
					break;
				}
			}

			/* check for an error */
			if(CONFFS_sevlen==i)
			{
				snprintf(CONFPRV_errstr,1024,"Unknown severity: %s",$2);
				yyerror(CONFPRV_errstr);
				free($$);

				return 1;
			}
            /* A severity of none is considered an error here */
            else if (CONFFS_sev[i].fs == CONFFS_NONE)
            {
                yyerror("usage of '=none' makes no sense");
                free($$);

                return 1;
            }
			/* else, we just add this one */
			else
			{
				$$[CONFFS_sev[i].fs]=1;
			}

			free($2);
		}
	| T_ID T_DASH T_ID
		{
			$$=emalloc(sizeof(int)*8);
			memset($$,0,8*sizeof(int));

            if (parseSevRange($1, $3, $$) != 0) {
                free($$);
                return 1;
            }

			free($1);
			free($3);
		}
	| T_STAR
		{
			int j=0;

			$$=emalloc(sizeof(int)*8);
			memset($$,0,8*sizeof(int));

			for(j=CONFFS_DEBUG;j>CONFFS_NONE;j--)
				$$[j]=PRI_PASS;
		}
	;

format_decl :
	  T_FORMAT T_ID
	  	{
			$$=$2;
		}
	;

format_opts_decl :
      T_FOPTS T_OPENSCOPE format_opt_list T_CLOSESCOPE
        {
            $$=$3;
        }
    ;
format_opt_list :
      T_ID T_SEMI format_opt_list
        {
            enum MFOPT   opt;
            opt = MFRMT_getOption($1);
            if (opt == MFOPT_NOEXIST) 
            {
                snprintf(CONFPRV_errstr, 1024, 
                    "No such formatter option '%s'", $1);
                yyerror(CONFPRV_errstr);

                /* Not fatal */
            }
            else 
            {
                BM_SET($3, opt);
            }
            free($1);
            $$ = $3;
        }
    | /* empty */
        {
            $$ = 0;
        }
    ;

listen_decl :
	  T_LISTEN T_ID
	  	{
			$$=$2;
		}
	| T_LISTEN T_STRING
		{
			$$=$2;
		}
	;

port_decl :
	  T_PORT T_INT
		{
			$$=$2;
		}
	;

connect :
	  T_CONNECT T_ID T_TO T_ID
		{
			struct outchan_info *out;
			struct inchan_info *in;

			$$=CONFOBJ_createCon();

			out=CONFSYM_getOutput($4);
			if(NULL==out)
			{
				snprintf(CONFPRV_errstr,1024,"output type not defined: %s\n",$4);
				yyerror(CONFPRV_errstr);

				free($2);
				free($4);

				return 1;		/* return an error */
			}

			in=CONFSYM_getInput($2);
			if(NULL==in)
			{
				snprintf(CONFPRV_errstr,1024,"input type not defined: %s\n",$2);
				yyerror(CONFPRV_errstr);

				free($2);
				free($4);

				return 1;		/* return an error */
			}

			$$->input=in->name;
			$$->output=out->name;

			/* only you can prevent memory leaks */
			free($2);
			free($4);
		}
	;

global :
	  T_GLOBAL T_OPENSCOPE global_decls T_CLOSESCOPE
	    {
			$$=$3;
		}
	;

global_decls :
	  T_ID T_STRING T_SEMI global_decls
	  	{
			if(!CONFPRV_parseGlobal($1,$2,$4))
			{
				snprintf(CONFPRV_errstr,1024,"Unknown global: %s",$2);
				yyerror(CONFPRV_errstr);

				return 1;
			}

			$$=$4;
            free($1);
		}
	| T_ID T_ID T_SEMI global_decls
	  	{
			if(!CONFPRV_parseGlobal($1,$2,$4))
			{
				snprintf(CONFPRV_errstr,1024,"Unknown global: %s",$1);
				yyerror(CONFPRV_errstr);

				return 1;
			}

			$$=$4;
            free($1);
		}
    | T_ID T_INT T_SEMI global_decls
        {
            if (!CONFPRV_parseGlobalIntVal($1, $2, $4))
            {
                snprintf(CONFPRV_errstr,1024,"Unknown global: %s",$1);
                yyerror(CONFPRV_errstr);

                return 1;
            }
            $$ = $4;
            free($1);
        }
	|	/* NULL */
		{
			$$=CONFOBJ_createGlobal();
		}
    ;	  	

%%

#include <stdio.h>

void yyerror(char *s)
{
	if(CONFPRV_fname()!=NULL)
		printf("%s:%d: %s\n",CONFPRV_fname(),CONFPRV_lineno(),s);
	else
		printf("%d: %s\n",CONFPRV_lineno(),s);
}

/* This need access to data private to this file, but also needs to be a
 * function.
 */
int parseSevRange(char const *rstr, char const *lstr, int *pris)
{
    int lower, upper;

    DPRINT(DBG_CNF, "Entering parseSevRange (%s, %s)", rstr, lstr);

    /* Get the two edges of our range */
    for (lower=0; lower < CONFFS_sevlen; lower++) 
    {
        if (strcmp(rstr, CONFFS_sev[lower].name) == 0)
        {
            break;
        }
    }
    for (upper=0; upper < CONFFS_sevlen; upper++)
    {
        if (strcmp(lstr, CONFFS_sev[upper].name) == 0)
        {
            break;
        }
    }

    DPRINT(DBG_CNF, "parseSevRange: lower = %d, upper = %d", 
        lower, upper);

    DPRINT(DBG_CNF, "sev range: %d - %d",
        CONFFS_sev[lower], CONFFS_sev[upper]);

    /* check for an error */
    if(CONFFS_sevlen==lower)
    {   
        snprintf(CONFPRV_errstr,1024,"Unknown severity: %s",rstr);
        yyerror(CONFPRV_errstr);

        return 1;
    }   
    /* check for another error */
    else if(CONFFS_sevlen==upper)
    {   
        snprintf(CONFPRV_errstr,1024,"Unknown severity: %s",lstr);
        yyerror(CONFPRV_errstr);
        return 1;
    }
    /* using 'none' in a range is also an error */
    else if (CONFFS_sev[lower].fs == CONFFS_NONE
        || CONFFS_sev[upper].fs == CONFFS_NONE)
    {
        yyerror("Using 'none' in a range makes no sense");
        return 1;
    }
    /* else, we're good, so add things into the array */
    else
    {
        int i;

        /* swap if needed */
        if (CONFFS_sev[lower].fs > CONFFS_sev[upper].fs) 
        {
            i = upper;
            upper = lower;
            lower = i;
        }
        DPRINT(DBG_CNF, "sev range: %d - %d",
            CONFFS_sev[lower], CONFFS_sev[upper]);

        i = CONFFS_sev[lower].fs;


        for(;i<=CONFFS_sev[upper].fs;i++)
        {
            pris[i]=1;
        }
    }

    return 0;
}
