/**
 * @file controller.c
 *
 * @brief The main controller loop.
 *
 *
 * $Id: controller.c,v 1.37.2.8 2003/08/02 22:12:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Jan 25 15:25:20 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
# include "acconfig.h"
#endif

#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

/* for I/O */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdio.h>

#include "types.h"
#include "strcompat.h"
#include "error.h"
#include "log.h"
#include "privs.h"
#include "chaninfo.h"

#include "version.h"

#include "controller.h"
#include "ctlr_destroy.h"
#include "ctlr_config.h"

#include "config/config.h"
#include "util/acl.h"

#include "ipc/ipc.h"
#include "ipc/inputwait.h"
#include "ipc/ipc_err.h"

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif

static char const rcsid[] 
    = "$Id: controller.c,v 1.37.2.8 2003/08/02 22:12:03 devink Exp $";

/**@addtogroup controller
 * @{
 */

/* Some functions */
/**
 * Parse command line arguments and add the options to the CTLR object.
 *
 * @param ctlr The controller state to change.
 * @param argc The number of arguments
 * @param argv The argument vector (usual main style)
 */
static bool parseArgs(CTLR *ctlr, int argc, char *argv[]);

/**
 * create the PID file.
 *
 * @param pidPath A specified path to the PID file, if one was given in the
 * 					config file, or NULL
 *
 * @return TRUE on success, FALSE otherwise
 */
static bool create_pid_file(char *pidPath);

/**
 * Turn our self into a daemon. 
 *
 * @return TRUE on success, FALSE otherwise
 */
static bool daemonize(void);

/**
 * Deal with std I/O; usually, direct everything to /dev/null
 * If WITH_DMALLOC is defined, then it's directed to ./stdio.out
 * Also changes the logging to go to the log file, instead of going to stdout.
 *
 * @param ctlr The controller state
 */
static void redirStdIO(CTLR *ctlr);

/**@name Command line option support
 * @{
 */

/** The options string for getopt(3) */
#define OPTIONS "hc:pd:vPf"

#define OPT_USAGE		'h'		/**< Print usage information */
#define OPT_CONFFILE    'c'     /**< specify alternant config file */
#define OPT_DEBUG       'd'     /**< specify the debug mask */
#define OPT_NOPRIV		'p'     /**< Disable the PRIV stuff */
#define OPT_VERSION     'v'     /**< Print version and exit */
#define OPT_PARSEONLY   'P'     /**< only parse the config file and exit */
#define OPT_NOFORK      'f'     /**< keep the controller in the foreground */

/*@}*/

/** exit right after parsing the config file? */
static bool gParseOnly = FALSE;

/** Keep controler in foreground */
static bool gNofork = FALSE;

/** Store the debug mask (as read from the command line) */
static char gDebugMask[16];  /* magic const, means nothing really */

/* The main controller function */
int main(int argc, char *argv[])
{
    IPC                     *ipcHandle;

    CTLR                    state;

    INW                     *waiter;
    struct INW_eInfo        *event;

    int                     rtn;

    struct message          rmsg;
	struct path				path;
    struct IPC_addr         src;
    struct IPC_event        ipcevt;
#if WITH_DMALLOC
    unsigned long           dm_loop, dm_reconf = 0; 
#endif

    /* Initialize the logging system for the first time */
    LOG_init();

    /* Initialize the controler state */
    if (! CTLR_init(&state) ) {
        EXIT(EXIT_ERROR);
    }
    EXIT_setCTLR(&state);   /* set to be cleaned up */

    /* parse the command line arguments */
    if (! parseArgs(&state, argc, argv) ) {
        EXIT(EXIT_ERROR);
    }

	/* prepare for process name changing */
	DBG_initProcName(argc,argv);

    /* setup debugging */
    if (! DBG_init(gDebugMask, DEBUG_LOG_PATH) ) {
        LOG_eprint("Debug initilization failed", errno);
        EXIT(EXIT_ERROR);
    }
    /* Give us a nice name for debugging messages */
    DBG_newproc("CTLR");

    /* read the config file */
    if (! CTLRConf_readConfig(&state) ) {
        /* error already logged */
        EXIT(EXIT_ERROR);
    }

    /* Make sure that we have a good channel to log to */
    if (state.icSysID == IPCAC_GENERAL) {
        LOG_print(LOG_ERROR, 
            "No System channel defined. This means that syslogd will "
            "not be able to log it's own messages.  This is not allowed, "
            "good bye.");
        EXIT(EXIT_ERROR);
    }

    if (gParseOnly) {
        EXIT(EXIT_NORMAL);
    }
    /* Initialize the path cache */
    state.pcache = PCACHE_ctlrCreate();
    if (state.pcache == NULL) {
        LOG_print(LOG_ERROR, "Initializing path cache failed");
        EXIT(EXIT_ERROR);
    }

    /* Drop privs */
    if (! PRIV_dropTemp() ) {
        LOG_print(LOG_ERROR, "Can't drop privledges");
        EXIT(EXIT_ERROR);
    }

    /* turn ourselves into a daemon before continuing */
    if (! daemonize() ) {
        EXIT(EXIT_ERROR);
    }

    /* initilize the output modules and routes */
    if (! CTLRConf_OMInit(&state) ) {
        /* already logged */
        EXIT(EXIT_ERROR);
    }
    if (! CTLRConf_routeInit(&state) ) {
        /* already logged */
        EXIT(EXIT_ERROR);
    }

	/* close the standard out and all the rest; must be done after the output
	 * modules have a chance to start */
	redirStdIO(&state);

    LOG_print(LOG_INFO, "Daemon starting, version %s", VERSION_STR);

    /* start up our IPC channel */
    ipcHandle = IPC_initController(state.sockPath);
    if (IPC_geterrno(ipcHandle) != IPCE_SUCCESS) {
        LOG_print(LOG_ERROR, "Failed to initialize IPC channel (%s): %s",
                state.sockPath, IPC_geterrstr(ipcHandle));
        EXIT(EXIT_ERROR);
    }
    state.ipcHandle = ipcHandle;

    DPRINT(DBG_CTL, "Started IPC", NO_VARG);

    /* Set up our input waiter */
    waiter = INW_create();
    if (waiter == NULL || !INW_addIPC(waiter, ipcHandle) ) {
        LOG_print(LOG_ERROR, "Setting up input waiter failed");
        EXIT(EXIT_ERROR);
    }
    EXIT_setINW(waiter);

    /* Setup signal events */
    if (! INW_addSignal(waiter, SIGTERM) ) {
        LOG_print(LOG_ERROR, "Setting up sigterm handler");
        EXIT(EXIT_ERROR);
    }
    if (! INW_addSignal(waiter, SIGHUP) ) {
        LOG_print(LOG_ERROR, "Setting up sighup handler");
        EXIT(EXIT_ERROR);
    }
    if (! INW_addSignal(waiter, SIGUSR1) ) {
        LOG_print(LOG_ERROR, "Setting up sigusr1 handler");
        EXIT(EXIT_ERROR);
    }

    DPRINT(DBG_CTL, "Initialized Signal Handling & INW", NO_VARG);

    /* now we can start input modules */
    if (! CTLRConf_IMInit(&state) ) {
        /* already logged */
        EXIT(EXIT_ERROR);
    }

    DPRINT(DBG_CTL, "Initialized IMs", NO_VARG);

    /* The last part of a successful initilization */
    if (! create_pid_file(state.pidPath)) {
        /* should already be logged */
        LOG_print(LOG_DEBUG, "PID file creation failed");
        EXIT(EXIT_ERROR);
    }
    state.isrunning = TRUE;
    LOG_print(LOG_INFO, "Initialization complete, SDSC syslogd running");

#if WITH_DMALLOC
    dm_loop = dmalloc_mark();
#endif

    /* start the main loop */
    while (1) {
        event = INW_getEvent(waiter);

        if (event == NULL) {
            LOG_print(LOG_ERROR, 
                "Error in INW_getEvent, may be bad data on IPC socket.");
        }
        else if (INWE_IPC == event->event_id) {
            switch (event->info) {
                case IPCT_PATH:
                {
                    if (! IPC_cRecvPath(ipcHandle, &src, &path) ) {
                        LOG_print(LOG_ERROR, "IPC recvPath failed: %s",
                                IPC_geterrstr(ipcHandle));
                        break;      /* giveup */
                    }

                    if (! PCACHE_addPath(state.pcache, &path, FALSE) ) {
                        LOG_print(LOG_ERROR, "Failed to add path to cache");
                        break;  /* give up */
                    }
                    break;
                }
                case IPCT_MSG:
                {
                    DPRINT(DBG_CTL, "ipc msg waiting", NO_VARG);
                    /* log messages next */
                    if (! IPC_cRecvMessage(ipcHandle, &src, &rmsg) ) {
                        LOG_print(LOG_ERROR, "IPC recvMessage failed: %s",
                                IPC_geterrstr(ipcHandle));
                        MSG_free(&rmsg);
                        break;  /* exit switch */
                    }

                    DPRINT(DBG_CTL, "got ipc msg: '%s'", rmsg.msg);

                    CTLR_sendMessage(&state, &rmsg, src.channel);
                    /* cleanup the message */
                    MSG_free(&rmsg);
                    break;
                }
                case IPCT_EVENT:
                {
                    rtn = IPC_recvEvent(ipcHandle, &ipcevt);
                    if (rtn < 0) {
                        LOG_print(LOG_ERROR, "IPC recvEvent failed: %s",
                                IPC_geterrstr(ipcHandle));
                        break;
                    }

                    CTLR_processIPCEvent(&state, &ipcevt);

                    break;
                }
                default:
                {

                    /* No other message types should be comming into the 
                     * controller.  
                     */
                    throw_fatal("unexpected message type from IPC channel"
                            " in the controller");
                    /* NOT REACHED */
                    break;
                }

            }
        }
        else if (INWE_SIG == event->event_id) {
            switch (event->info) {
                case SIGTERM:
                {
                    DPRINT(DBG_CTL, "Got SIGTERM", NO_VARG);
                    LOG_print(LOG_INFO, "Starting shutdown");
                    TPRINT("SIGTERM: start shutdown");

                    /* stop new IPC connections */
                    IPC_stopConnections(ipcHandle);

                    /* tell all the input modules to die */
                    CTLR_killIMs(&state);

                    /* close all the sockets */
                    CTLR_shutdownSockets(&state);

                    /* flag that we should exit as soon as the IPC
                     * channel is clear
                     */
                    state.imClear = CTLR_IMC_EXIT;
                    TPRINT("waiting for IMs to exit");
                    break;
                }
                case SIGHUP:
                {
                    LOG_print(LOG_INFO, "Starting reconfig");
                    #if WITH_DMALLOC
                        dmalloc_message("%s:%d starting reconfig", 
                                __FILE__, __LINE__);
                        dmalloc_log_changed(dm_loop, 1, 0, 0);
                        dm_reconf = dmalloc_mark();
                    #endif

                    /* do as much of the reconf as we can */
                    if (! CTLRConf_readConfig(&state) ) {
                        /* already logged, cleanup now */
                        CTLRConf_destroy(state.newConf);
                        state.newConf = NULL;
                        break;  /* just stop now */
                    }

                    if (! CTLRConf_OMInit(&state) ) {
                        /* already logged, cleanup now */
                        CTLRConf_destroy(state.newConf);
                        state.newConf = NULL;
                        break;  /* just stop now */
                    }

                    if (! CTLRConf_routeInit(&state) ) {
                        /* already logged, cleanup now */
                        CTLRConf_destroy(state.newConf);
                        state.newConf = NULL;
                        break;  /* just stop now */
                    }

                    /* kill all the input modules, and close sockets */
                    CTLR_killIMs(&state);
                    CTLR_shutdownSockets(&state);

                    /* mark that we need to finish this when all the
                     * IMs are gone 
                     */
                    state.imClear = CTLR_IMC_RCONF;

                    break;
                }
                case SIGUSR1:
                {
                    /* reinitilize all OM's, for log rolling */
                    HTIterator          itr;
                    struct outchan_info *ocinfo;
                    #if WITH_DMALLOC
                        dmalloc_message("%s:%d starting reinit", 
                                __FILE__, __LINE__);
                        dmalloc_log_changed(dm_loop, 1, 0, 0);
                        dm_reconf = dmalloc_mark();
                        dm_loop = dmalloc_mark();
                    #endif

                    TPRINT("controller got SIGUSR1");

                    itr = HT_iterator(state.conf->ocMap);
                    while ( (ocinfo = HTI_next(itr)) != NULL) {
                        if (ocinfo->iface != NULL)
                            ocinfo->iface->REINIT(ocinfo->iface->state);
                    }
                    HTI_destroy(&itr);

                    LOG_print(LOG_INFO, "Reinitilized output modules");
                    #if WITH_DMALLOC
                        dmalloc_message("%s:%d finished reinit", 
                                __FILE__, __LINE__);
                        dmalloc_log_changed(dm_reconf, 0, 1, 0);
                    #endif

                    break;
                }
                default:
                {
                    LOG_print(LOG_INFO, "Caught unexpected signal %d", 
                            event->info);
                    break;
                }
            }
        }
        else {
            /* I'm confused */
            throw_fatal("unexpected event in the controller");
        }

        /* If we are trying to exit, see if we can */
        if (state.imClear) {
            int     i;
            /* check if all the input modules have exited */
            for (i = 0; i < IMOD_TYPECOUNT; i++) {
                DPRINT(DBG_CTL, "icType %d, has %d IMs running", i, 
                        state.conf->icTypes[i].imCount);
                if (state.conf->icTypes[i].imCount != 0)
                    break;
            }

            DPRINT(DBG_CTL, "waiting for IMs to clear, i = %d", i); 

            if (i >= IMOD_TYPECOUNT) {
                /* figure out why we wanted to clear the IMs and act
                 * accordingly.
                 */
                switch (state.imClear) {
                    case CTLR_IMC_EXIT:
                        TPRINT("shutdown finished, about to exit");
                        /* trying to exit, will be logged by EXIT */
                        EXIT(EXIT_NORMAL);
                        break;
                    case CTLR_IMC_RCONF:
                        #if WITH_DMALLOC
                            dmalloc_message("%s:%d recnofig finished",
                                    __FILE__, __LINE__);
                            dmalloc_log_changed(dm_reconf, 0, 1, 0);
                            dm_loop = dmalloc_mark();
                        #endif
                        /* trying to reconfigure 
                         * initilizing the IMs is the last step waiting
                         * for the old ones to die.
                         */
                        if (! CTLRConf_IMInit(&state) ) {
                            /* already logged, we have to bail now. */
                            EXIT(EXIT_ERROR);
                        }

                        DPRINT(DBG_CTL, "Reconfig completed", NO_VARG);
                        LOG_print(LOG_INFO, "Reconfig completed");

                        break;
                }
                state.imClear = CTLR_IMC_NONE;
            }
        }
    }
    
    /* NOT REACHED */
    EXIT(EXIT_NORMAL);  /* but just in case */

    return 0;   /* shut the compiler up */
}

static bool
parseArgs(CTLR *state, int argc, char *argv[]) 
{
    int         arg;

    /* clear the debug mask */
    memset(gDebugMask, 0, sizeof(gDebugMask));

    while ( (arg = getopt(argc, argv, OPTIONS)) >= 0) {
        switch (arg) {
            case OPT_PARSEONLY:
            {
                gParseOnly = TRUE;
                break;
            }
            case OPT_CONFFILE:
            {
                if (optarg == NULL) {
                    return FALSE;
                }
                strlcpy(state->confFile, optarg, sizeof(state->confFile)); 
                break;
            }
            case OPT_DEBUG:
            {
                if (optarg == NULL) {
                    return FALSE;
                }
                strlcpy(gDebugMask, optarg, sizeof(gDebugMask));
                break;
            }
            case OPT_NOFORK:
            {
                gNofork = TRUE;
                break;
            }
			case OPT_NOPRIV:
			{
                if (ENABLE_TESTING) {
				    PRIV_disableChanges();
                }
                else {
                    printf(
    "Attempting to use '-p' without enabling testing code at compile time\n"
    "may have nasty side effects. For your own sanity such a combination is\n"
    "not allowed. Please rerun configure with the '--enable-testing' argument.\n"
                    );
                    EXIT(EXIT_ERROR);
                }

				break;
			}
			case OPT_USAGE:
			{
				printf(
	"Usage: syslogd [-fhpPv] [-d debugmask] [-c configfile]\n"
	"   -f              Keep the controller in the foreground\n"
	"   -h              Print usage information\n"
	"   -p              Disable privelege drops\n"
	"   -P              Only parse config file, then exit\n"
	"   -v              Print version information\n"
	"   -d debugmask    Enable debug info, with specified mask (see debug.h)\n"
	"   -c configfile   Specify a config file\n"
					  );

				EXIT(EXIT_NORMAL);
				break;	/* never reached */
			}
            case OPT_VERSION:
            {
                printf("%s %s\n\n", PACKAGE, VERSION_STR);

                printf("Compile Time Options:\n");
                printf("\tENABLE_DEBUG = %d\n", ENABLE_DEBUG);
                printf("\tENABLE_TESTING = %d\n", ENABLE_TESTING);
                printf("\tDebugging output dir is '%s'\n", DEBUG_OUT_DIR);
                printf("\tDefault config file is '%s'\n", DEFAULT_CONF_PATH);

                printf("\n%s\n", COPYRIGHT);
                EXIT(EXIT_NORMAL);
            }
            case '?':
            {
                /* undefined option */
                return FALSE;
            }
            default:
            {
                /* Un-reconized option */
                LOG_print(LOG_ERROR, 
                    "Option '%c' is unreconzied, but listed in OPTIONS."
                    "This is a programmer error", arg);
                return FALSE;
            }

        }
    }

    return TRUE;
}

static bool
create_pid_file(char *pidPath)
{
    int     pidfd;
    bool    success = TRUE;
    char    buf[16];

    /* put the ascii version of our pid into buf */
    snprintf(buf, sizeof(buf), "%d", getpid());

    /**@todo is the correct way to deal with stale pid files? */ 
    PRIV_unlink(pidPath);

    /* open and write to the file */
    pidfd = PRIV_openWriteFile(pidPath, O_WRONLY|O_CREAT|O_EXCL);
    if (pidfd < 0) {
        LOG_print(LOG_ERROR, "Failed to open the PID file (%s)",pidPath);
        return FALSE;
    }

    if (write(pidfd, buf, strlen(buf)) < 0) {
        LOG_eprint("Writing PID file", errno);
        success = FALSE;
    }


    /* the PRIV_ function loggs an error and we don't treat it as a
     * significant failure.
     */
    PRIV_fchmod(pidfd, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

    close(pidfd);
    return success;
}

static bool
daemonize(void)
{
    pid_t       pid;
    int         tty;

    /* Change to the root directory, iff we are not testing */
    if (! ENABLE_TESTING) {
        if (chdir("/") < 0)  {
            LOG_eprint("Can't chdir(\"/\")", errno);
            return FALSE;
        }
    }

    /* background ourself, unless a command line option has said not to */
    if (! gNofork ) {
        pid = fork();
        if (pid < 0) {
            /* Error */
            LOG_eprint("Can't fork()", errno);
            return FALSE;
        }
        else if (pid != 0) {
            /* parent ... here we just exit normally and let the child keep
             * running.
             */
            EXIT(EXIT_NORMAL | EXIT_NOCLEAN);
        }
    }

    /* Child ... this is the process that continues */

    /* Detach from controlling TTY */
    tty = open("/dev/tty", O_RDWR);
    if (tty >= 0) {
        /* if the open failed then there is no controling TTY */
        if (ioctl(tty, TIOCNOTTY, 0) < 0) {
            LOG_eprint("Detaching from controlling tty failed", errno);
            /* It's an error, but not the end of the world */
        }
        close(tty);
    }

	/* change the name */
	DBG_setProcName("syslogd (control)");

    return TRUE;
}

static void
redirStdIO(CTLR *ctlr)
{
	int temp;

    /* start logging to the log file now */
    if (! LOG_initCTLR(ctlr) ) {
        LOG_print(LOG_ERROR, "Can't shift logging to a backing file");
        EXIT(EXIT_ERROR);
    }

	if( -1==(temp=open("/dev/null",O_RDONLY|O_NONBLOCK)) )
	{
		LOG_eprint("Failed opening \"/dev/null\"",errno);
		return;
	}
	if(-1==dup2(temp,STDIN_FILENO))
	{
		LOG_eprint("STDIN redirect failed",errno);
		return;
	}

	close(temp);

#ifdef WITH_DMALLOC
    if( -1==(temp=open(STDIO_SVPATH,O_WRONLY|O_SYNC|O_CREAT)) )
    {
        LOG_eprint("Failed opening \"./stdio.out\"",errno);
        return;
    }
#else
    if( -1==(temp=open("/dev/null",O_WRONLY|O_NONBLOCK)) )
    {
        LOG_eprint("Failed opening \"/dev/null\"",errno);
        return;
    }
#endif
	if(-1==dup2(temp,STDOUT_FILENO))
	{
		LOG_eprint("STDOUT redirect failed",errno);
		return;
	}
	if(-1==dup2(temp,STDERR_FILENO))
	{
		LOG_eprint("STDERR redirect failed",errno);
		return;
	}

	close(temp);
}
