/**
 * @file controller.h
 *
 * @brief Define Internal controller functions and data structures.
 *
 * @internal
 *
 * $Id: controller.h,v 1.16.2.6 2003/04/30 20:03:09 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Jan 25 15:48:15 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CTLR_CONTROLLER_H
#define CTLR_CONTROLLER_H

#include "util/list.h"
#include "util/acl.h"
#include "config/config.h"
#include "ipc/ipc.h"
#include "outputmod/outputmod.h"
#include "inputmod/inputmod.h"
#include "util/netaddr.h"
#include "const.h"
#include "util/pathcache.h"

/**@addtogroup controller Controller 
 * @{
 */

/**@name Reasons to wait for all the IMs to finish.
 * @{
 */
#define CTLR_IMC_NONE   0   /**< Don't wait */
#define CTLR_IMC_EXIT   1   /**< Trying to exit the controller */
#define CTLR_IMC_RCONF  2   /**< Trying to restart IMs */
/*@}*/

/** Input Module Inforamion */
struct CTLR_imInfo
{
    IMOD_TYPE       type;           /**< IM type */
    pid_t           pid;            /**< process ID */
    int             ipcproc;        /**< Process address for IPC */
};

/** 
 * The contains the entire state of the controller.  Basically rather
 * than using global (even in limited compilation units) varibales, I'm
 * putting all of them in this structure.
 * @internal
 */
typedef struct 
{
    char    confFile[PATH_MAX];             /**< config file name */
    char    sockPath[UNIX_PATH_MAX];        /**< IPC socket name */
	char	pidPath[PATH_MAX];			    /**< Path to PID file */

    char    *icSysName;     /**< name of the real system input module */

    List    unInitChild;    /**< Uninitilized children */

    IPC     *ipcHandle;     /**< ipcHandle */

    /* Data for all initilized input modules */
    struct CTLR_imInfo  *ims[IPC_MAX_IMPROCS];

    PCache  *pcache;        /**< path cache */

    int     imClear;        /**< are we waiting for IMs to finish, and why */
    bool    isrunning;      /**< are we in a running state ? */

    struct CTLR_conf    *conf;      /**< information from the config file */
    struct CTLR_conf    *newConf;   /**< partially built replacement \c conf */

    uint32_t    nextICID;   /**< The ID for the next input channel */
    uint32_t    nextOCID;   /**< The ID for the next output channel */

    uint32_t    icSysID;    /**< ID of system input channel */

    int     mark_interval;  /**< Interval between marks in seconds */
} CTLR;


/**
 * Initilize the controller state.
 *
 * @param ctlr The CTLR object to initilized
 * @return TRUE on success, FALSE otherwise
 *
 * @pre ctlr != NULL
 */
bool CTLR_init(CTLR *ctlr);

/**
 * Increment the next input channel ID.
 *
 * @param ctlr The CTLR object to work on.
 */
void CTLR_incNextICID(CTLR *ctlr);

/**
 * Send a message to the configured output modules.
 *
 * @param ctlr The controller state
 * @param msg The Message we are trying to send
 * @param channel The source channel for the message
 */
void CTLR_sendMessage(CTLR *ctlr, const struct message *msg, int channel);

/**
 * Process an IPC event.
 *
 * @param ctlr The CTLR object for the current state.
 * @param evt The IPC event to process
 * @return TRUE on success, FALSE on error.
 */
bool CTLR_processIPCEvent(CTLR *ctlr, struct IPC_event *evt);

/**
 * Fork a new input module process and start it.  This function will
 * only return to the parent.
 *
 * @param ctlr The state info needed to setup the input module
 * @param cit The type of the input module
 * @return TRUE on success, FALSE on error.
 */
bool CTLR_forkNewIM(CTLR  *ctlr, IMOD_TYPE cit);

/**
 * Set routing info.  This also does bounds checking on \c icid 
 * and \c pri.  In the event of a bounds violation, a long message will
 * be printed (via WriteLog()).
 *
 * @param rts The routing tables.
 * @param icid Input Channel ID
 * @param pri Priority
 * @param omiface Output Module Iface which is the destination.
 */
inline void CTLR_setRoute(
        struct CTLR_conf *rts, int icid, int pri, OMOD_interface *omiface);

/*@}*/

#endif /* CTLR_CONTROLLER_H */
