/**
 * @file ctlr_config.c
 *
 * @brief This defines all the functions used to configure the
 *        controller
 *
 * $Id: ctlr_config.c,v 1.28.2.7 2003/04/30 20:03:09 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed May 15 12:27:22 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#if HAVE_CONFIG_H
# include "acconfig.h"
#endif

#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "strcompat.h"

#include "types.h"
#include "chaninfo.h"
#include "error.h"
#include "debug.h"
#include "log.h"
#include "privs.h"
#include "const.h"

#include "config/config.h"
#include "config/conf_obj.h"

#include "controller.h"
#include "ctlr_config.h"


static char const rcsid[] 
    = "$Id: ctlr_config.c,v 1.28.2.7 2003/04/30 20:03:09 devink Exp $";

/** The mode for a SYSTEM input channel socket */
#define SYS_SOCKET_MODE \
    (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)

/**
 * Save the global options into the controller object.  This assumes
 * that ctlr->newConf->cinfo is initilized and contains the config to
 * transfer into the controller.
 *
 * @param ctlr The Controller object to change.
 * @param conf The new CTLR_conf object to work from.
 * @return TRUE on sucecss, FALSE otherwise
 */
static bool saveGlobalOpts(CTLR *ctlr, struct CTLR_conf *conf);

/**
 * Create a new CTLR_conf object and initlize it.
 *
 * @param cinfo The config_info object to use for initlization.
 * @return a pointer to the new object or NULL on error.
 */
static struct CTLR_conf * newConf(struct config_info *cinfo);

/**
 * Open an input channel.  This will create the socket, bind it to the
 * correct address, listen(2) if needed, and finally update \c chan with
 * the information.
 *
 * @param chan The channel to open
 * @return TRUE on success
 */
static bool openChannel(struct inchan_info *chan);

/* 1 */
bool
CTLRConf_readConfig(CTLR *ctlr)
{
    struct config_info  *cinfo;
    struct CTLR_conf    *conf;
    struct inchan_info  *icinfo;
    struct outchan_info *ocinfo;
    LLIterator          itr;

    /* make sure that we aren't currently in the middle of this process */
    assert(ctlr->newConf == NULL);

    /* read the config file */
    cinfo = getConfigInfo(ctlr->confFile);
    if (cinfo == NULL) {
        LOG_print(LOG_ERROR, "Reading config file failed");
        return FALSE;
    }
    
    /* Create a new CTLR_conf object */
    conf = newConf(cinfo);
    if (conf == NULL) {
        LOG_print(LOG_ERROR, "Error initilizing controller config info");
        return FALSE;
    }
    conf->cinfo = cinfo;    /* save for later */

    /* save the global options */
    if (! saveGlobalOpts(ctlr, conf) ) {
        /* this is very unlikely */
        LOG_print(LOG_ERROR, "Failed to process global options");
        return FALSE;
    }

    /* save input channel information */
    conf->baseICID = ctlr->nextICID;    /* save the first IC ID */

    itr = LL_iterator(cinfo->inputs);
    while ( (icinfo = LLI_next(itr)) != NULL) {
        /* assign an id */
        icinfo->id = ctlr->nextICID;
        CTLR_incNextICID(ctlr);

        /* save the IC info */
        HT_insert(conf->icMap, icinfo->name, icinfo);
        conf->icTypes[icinfo->type].count++;
        /* chans will be updated later */

        /* handle special system IC */
        if (icinfo->type == IMOD_SYSTEM
            && strcmp(icinfo->name, ctlr->icSysName) == 0) 
        {
            ctlr->icSysID = icinfo->id;
        }
    }
    LLI_destroy(&itr);
    
    /* 
     * free the cinfo->inputs list w/o free()ing any inchan_info
     * objects.  This prevents them from being freed when cinfo is
     * destroyed..
     */ 
    LL_destroy(&(cinfo->inputs));

    /* save output channel information */
    itr = LL_iterator(cinfo->outputs);
    while ( (ocinfo = LLI_next(itr)) != NULL) {

        /* assign an id */
        ocinfo->id = ctlr->nextOCID;
        ctlr->nextOCID++;

        /* save the OC info */
        HT_insert(conf->ocMap, ocinfo->name, ocinfo);

    }
    LLI_destroy(&itr);

    /* 
     * free the cinfo->outputs list w/o free()ing any outchan_info
     * objects.  This prevents them from being freed when cinfo is
     * destroyed.
     */ 
    LL_destroy(&(cinfo->outputs));

    ctlr->newConf = conf;

    return TRUE;
}

/* 2
 * PCACHE init 
 */

/* 3 */
bool
CTLRConf_OMInit(CTLR *ctlr)
{
    struct outchan_info *ocinfo;
    struct CTLR_conf    *conf;
    HTIterator          itr;

    assert(ctlr->pcache != NULL);
    assert(ctlr->newConf != NULL);

    conf = ctlr->newConf;

    itr = HT_iterator(conf->ocMap);
    while ( (ocinfo = HTI_next(itr) ) != NULL) {
        ocinfo->opts->pcache = ctlr->pcache;

        /* Create a new output module for this one */
        ocinfo->iface = OMOD_getInterface(ocinfo->type);
        ocinfo->iface->state = ocinfo->iface->INIT(ocinfo->opts);
        if (ocinfo->iface->state == NULL) {
            LOG_print(LOG_INFO, "Ouput module %s failed to initilize.",
                    ocinfo->name);
            OMOD_freeInterface(ocinfo->iface);
            ocinfo->iface = NULL;
        }
    }
    HTI_destroy(&itr);

    return TRUE;
}

/* 4 */
bool
CTLRConf_routeInit(CTLR *ctlr)
{
    /* setup all the routes */
    struct CONF_con             *confcon;
    struct outchan_info         *ocinfo;
    struct inchan_info          *icinfo;
    struct CTLR_conf            *conf;
    size_t                      tblSize;
    int                         i;
    LLIterator                  itr;

    assert(ctlr->newConf != NULL);

    conf = ctlr->newConf;

    tblSize = sizeof(CTLR_chanRoutes) * conf->icCount;
    conf->routes = emalloc(tblSize);
    memset(conf->routes, 0, tblSize);


    itr = LL_iterator(conf->cinfo->connects);
    while ( (confcon = LLI_next(itr)) != NULL) {
        icinfo = HT_find(conf->icMap, confcon->input);
        ocinfo = HT_find(conf->ocMap, confcon->output);

        if (ocinfo->iface == NULL) { 
            LOG_print(LOG_INFO, 
                "Not adding unitilizied output %s to routing table",
                ocinfo->name);
            continue;   /* skip */
        }

        for (i = 0; i < PRIORITY_COUNT; i++) {
            if (ocinfo->pri[i]) {
                CTLR_setRoute(conf, icinfo->id, i, ocinfo->iface);
            }
        }
    }
    LLI_destroy(&itr);

    /* We don't need the config_info object anymore */
    destroyConfigInfo(&(conf->cinfo));

    /* Special case, if there is no old config then we set this as the
     * real config now.
     */
    if (ctlr->conf == NULL) {
        ctlr->conf = ctlr->newConf;
        ctlr->newConf = NULL;
    }

    return TRUE;
}

/* 5 
 * daemonize 
 * IPC init
 * signal (INW) init
 */

/* 6 */
bool
CTLRConf_IMInit(CTLR *ctlr)
{
    HTIterator          itr;
    struct inchan_info  *icinfo;
    struct CTLR_conf    *conf;
    size_t              tblSize;
    int                 i;
    int                 idx[IMOD_TYPECOUNT];

    conf = ctlr->newConf;
    if (conf == NULL)
        conf = ctlr->conf;  /* there was no old config after route init */

    assert(conf != NULL);

    /* Setup icTypes.chans */
    for (i = 0; i < IMOD_TYPECOUNT; i++) {
        tblSize = sizeof(struct inchan_info) * conf->icTypes[i].count;
        if (tblSize > 0) 
            conf->icTypes[i].chans = emalloc(tblSize);
    }
    memset(idx, 0, sizeof(idx));

    itr = HT_iterator(conf->icMap);
    while ( (icinfo = HTI_next(itr)) != NULL) {
        DPRINT(DBG_CTL, "Starting IC init loop", NO_VARG);
        /* Open the input channel */
        if (! openChannel(icinfo) ) {
            LOG_print(LOG_INFO, "Not listening on channel %s.", icinfo->name);

            /* one less channel to worry about */
            conf->icTypes[icinfo->type].count--;

            /* clean up the chaninfo object */
            CONFOBJ_destroyInput(&icinfo);
        }
        else {
            DPRINT(DBG_CTL, "Adding channel to controller IC list", NO_VARG);
            /* open successful, set the channel in our IC list */
            conf->icTypes[icinfo->type].chans[idx[icinfo->type]] = icinfo;
            idx[icinfo->type]++;
        }
        DPRINT(DBG_CTL, "Ending IC init loop", NO_VARG);
    }
    DPRINT(DBG_CTL, "done setting up channels", NO_VARG);
    HTI_destroy(&itr);


    /* we are done with icMap ... */
    HT_destroy(conf->icMap);
    conf->icMap = NULL;

    /* move this into position as the real conf info, if not already done */
    if (ctlr->newConf != NULL) {
        if (ctlr->conf != NULL)
            CTLRConf_destroy(ctlr->conf);
        ctlr->conf = ctlr->newConf;
        ctlr->newConf = NULL;
    }

    /* Now start all IMs */
    for (i = 0; i < IMOD_TYPECOUNT; i++) {
        if (conf->icTypes[i].count > 0) {
            if (! CTLR_forkNewIM(ctlr, i) ) {
                LOG_print(LOG_ERROR, "Failed to start %s Input Module",
                    IMOD_typeNames[i]);
            }
        }
    }

    DPRINT(DBG_CTL, "Done starting IMs", NO_VARG);

    return TRUE;
}

void
CTLRConf_destroy(struct CTLR_conf *conf)
{
    HTIterator          itr;
    struct outchan_info *ocinfo;
    int                 i, j;

    /* Kill all the output channel info objects */
    itr = HT_iterator(conf->ocMap);
    while ( (ocinfo = HTI_next(itr)) != NULL) {
        CONFOBJ_destroyOutput(&ocinfo);
    }
    HTI_destroy(&itr);

    /* Kill all the input channel info objects */
    for (i = 0; i < IMOD_TYPECOUNT; i++) {
        for (j = 0; j < conf->icTypes[i].count; j++) {
            CONFOBJ_destroyInput(&(conf->icTypes[i].chans[j]));
        }
        free(conf->icTypes[i].chans);
    }

    /* destroy all the routes */
    for (i = 0; i < conf->icCount; i++) {
        for (j = 0; j < PRIORITY_COUNT; j++) {
            if (conf->routes[i][j] != NULL) 
                LL_destroy(&(conf->routes[i][j]));
        }
    }
    free(conf->routes);
}

static struct CTLR_conf *
newConf(struct config_info *cinfo)
{
    uint32_t            ocmap_size;
    uint32_t            icmap_size;
    int                 i;
    struct CTLR_conf    *conf;

    ocmap_size = 1 << (sizeof(uint32_t) * CHAR_BIT - 1);
    icmap_size = 1 << (sizeof(uint32_t) * CHAR_BIT - 1);

    /* hash table sizes need to be a power of 2, so find the smallest 
     * power of 2 that is larger than the number of input/output channels.
     *
     * This is done by finding the highest bit that is set in _count,
     * then multiplying it by 2 (left shift 1)
     */
    for (i = sizeof(uint32_t) * CHAR_BIT - 1; i >= 0; i--) {

        /* 
         * if this bit is not set right shift and try again 
         * if it is set, then do nothing, we have what we need
         */
        if ( (cinfo->input_count & icmap_size) == 0) 
            icmap_size >>= 1;

        if ( (cinfo->output_count & ocmap_size) == 0) 
            ocmap_size >>= 1;
    }
    icmap_size <<= 1;
    ocmap_size <<= 1;

    
    conf = emalloc(sizeof(struct CTLR_conf));
    memset(conf, 0, sizeof(struct CTLR_conf));

    conf->icMap = HTStr_create(icmap_size);
    conf->ocMap = HTStr_create(ocmap_size);

    conf->icCount = cinfo->input_count;
    conf->ocCount = cinfo->output_count;

    return conf;
}
     
static bool
saveGlobalOpts(CTLR *ctlr, struct CTLR_conf *conf)
{
    struct CONF_global  *globals;

    globals = conf->cinfo->global;

	if(globals == NULL)
	{
		LOG_print(LOG_ERROR, "No global section configured");
		return FALSE;
	}

    /* check for required options */
    if (globals->user == NULL || globals->group == NULL)
	{
		LOG_print(LOG_ERROR, "Unprivileged user/group not defined");
        /* only fatal if we aren't running yet */
		if (! ctlr->isrunning) 
            return FALSE;
	}

    /* process options that can't change after the system starts */
    if (! ctlr->isrunning) {
        if(! PRIV_setUnprivUser(globals->user, globals->group))
            return FALSE;

        if (globals->sockPath != NULL) {
            strlcpy(ctlr->sockPath, globals->sockPath, UNIX_PATH_MAX);
        }

		if (globals->pidPath != NULL) {
			strlcpy(ctlr->pidPath, globals->pidPath, PATH_MAX);
		}
    }

    /* process options that can change whenever */
    if (globals->mark_interval != 0) {
        ctlr->mark_interval = globals->mark_interval;
    }
    else {
        ctlr->mark_interval = DEFAULT_MARK_INTERVAL;
    }

	return TRUE;
}

static bool
openChannel(struct inchan_info *chan)
{
    union NA_sockaddr   addr;
    int                 protType;
    int                 success;
    int                 sock;

    DPRINT(DBG_CTL, "openChannel: start (%s)", chan->name);

    protType = 0;
    
    /* Do we have anything to open ? */
    if (NA_isEmpty(chan->inAddr) ) {
        return TRUE;
    }

    switch (chan->type) {
        case IMOD_UDP:
            /* UDP and SYSTEM share the same protocol type */
        case IMOD_SYSTEM: 
            protType = SOCK_DGRAM; break;

        case IMOD_BEEP:  
            protType = SOCK_STREAM; break;

        default:
            throw_fatal("Unknown Input Module Type"); break;
    }

    sock = socket(NA_getProtFamily(chan->inAddr), protType, 0);
    if (sock < 0) {
        LOG_eprint("creating socket", errno);
        goto error;
    }

    DPRINT(DBG_CTL, "socket opened, fd = %d", sock);


    if (! NA_mkSockaddr(chan->inAddr, chan->inPort, &addr) ) {
        /* This is probably a programmer issue, specifically with the
         * sizeof sockaddr.  This should work, but the way sockaddrs are
         * handled is beyond strange (see sys/socket.h) (for real fun see 
         * linux's SUN_LEN macro in sys/un.h).
         */
        throw_fatal("Failed to turn netaddr into sockaddr");
    }

    /* if this is a udp input module listening on the default port we
     * need to set SO_REUSEADDR incase there is a UDP output module that
     * will try to send from the same address.
     *
     * For BEEP modules, there are cases where there is no listener, but
     * not all the connections have closed completely, so we use
     * SO_REUSEADDR to get around that and bind anyway.
     */
    if ( (chan->type == IMOD_UDP && chan->inPort == SYSLOG_UDP_PORT)
        || chan->type == IMOD_BEEP) 
    {
        int     val = 1;
        if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) < 0) 
        {
            LOG_eprint("setsockopt(SO_REUSEADDR)", errno);
            goto error;
        }
    }

    if (NA_getProtFamily(chan->inAddr) == PF_UNIX) {
        success = PRIV_bindUnixSocket(sock,&(addr.sun),FALSE,SYS_SOCKET_MODE);
    }
    else {
        success = PRIV_bindSocket(sock, &(addr.sa), sizeof(addr));
    }

    if (!success) {
        LOG_print(LOG_ERROR, "Failed to bind to listen socket");
        goto error;
    }

    DPRINT(DBG_CTL, "bind complete", NO_VARG);

    if (protType == SOCK_STREAM) {
        if (listen(sock, 0) < 0) {
            LOG_eprint("listing on input socket", errno);
            goto error;
        }
    }

	/* save the socket, we'll need it */
	chan->fd=sock;

    DPRINT(DBG_CTL, "openChannel: end (TRUE)", NO_VARG);

    return TRUE;

    /* Error handler for the function.  If we reanch this part of the
     * code then there was definantly an error.
     */
    error:

        if (sock >= 0) close(sock);

        DPRINT(DBG_CTL, "openChannel: end (FALSE)", NO_VARG);

        return FALSE;
}
