/**
 * @file ctlr_config.h
 *
 * @brief Functions used by the controller to process config information.
 * @par
 * These are the functions that are used to take config file information
 * (found in a struct config_info) and convert it into a running and
 * configured syslog daemon.
 *
 * @ingroup ctlr_conf
 *
 * $Id: ctlr_config.h,v 1.3.2.1 2003/04/30 20:03:09 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jul 17 17:50:30 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CTLR_CTLR_CONFIG_H
#define CTLR_CTLR_CONFIG_H

#include "util/hashtable.h"
#include "util/list.h"

#include "controller.h"

/**@defgroup ctlr_conf Controller Configuration
 * @ingroup ctlr
 *
 * These data structures and functions are used to store the config
 * information in the controller.  The configuration process is a multi
 * stage process consisting of:
 * \li 1. Read the config file and to initial processing 
 *        (via CTLRConf_readConfig() )
 * \li 2. Initilize the Path Cache
 * \li 3. Initilize the output modules (via CTLRConf_OMInit() )
 * \li 4. Initilize the routing tables (via CTLRConf_routeInit() )
 * \li 5. daemonize, initilize the IPC layer, and initilize signal
 *        handling.
 * \li 6. start the input modules (via CTLRConf_IMInit() )
 *
 * The first step sets CTLR::newConf to the newly created CTLR_conf
 * object.  CTLR::newConf is worked on until everything is done, and in
 * CTLRConf_IMInit() it is made active in the controller (by setting
 * CTLR::conf to the new CTLR_conf object).  Also this takes care of
 * destroying the old conf info.
 *
 * @{
 */

/** Information about inpuct channels on a per type bases */
struct CTLR_icType
{
    size_t              count;      /**< channel count */
    size_t              imCount;    /**< Count of Input Modules of this type */
    struct inchan_info  **chans;    /**< Input Channels */
};

/** Per Channel routing information data structure. */
typedef List    CTLR_chanRoutes[PRIORITY_COUNT];

/**
 * This structure contains all the information needed to take the
 * controller through a multi-phase configuration process.  
 */
struct CTLR_conf {
    HashTable   icMap;      /**< map input channel names -> id */
    HashTable   ocMap;      /**< map ouput channel names -> id */

    int         icCount;    /**< number of input channels */
    int         ocCount;    /**< number of output channels */

    int         baseICID;   /**< first IC ID */

    /** IC type info */
    struct CTLR_icType  icTypes[IMOD_TYPECOUNT];

    /** routing table */
    CTLR_chanRoutes *routes;

    /** info from the config file */
    struct config_info  *cinfo; 
};

/**
 * Read the config file and do initial processing on it.
 * This also creates \c ctlr->newConf
 *
 * @param ctlr The Controller state object.
 * @return TRUE on success, FALSE otherwise
 *
 * @pre ctlr != NULL
 * @pre ctlr->newConf == NULL (not partially initilized already)
 */
bool CTLRConf_readConfig(CTLR *ctlr);

/**
 * This will start the configured output modules. 
 *
 * @param ctlr The Controler state object. 
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre ctlr != NULL
 * @pre ctlr->newConf != NULL
 */
bool CTLRConf_OMInit(CTLR *ctlr);

/**
 * This will setup routing based on new config info.
 *
 * @param ctlr The Controler state object. 
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre ctlr != NULL
 * @pre ctlr->newConf != NULL
 */
bool CTLRConf_routeInit(CTLR *ctlr);

/**
 * Start Input modules.
 * This also destroys \c ctlr->conf and sets \c ctlr->conf=ctlr->newConf
 * 
 * @param ctlr The Controler state object. 
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre ctlr != NULL
 * @pre ctlr->newConf != NULL
 */
bool CTLRConf_IMInit(CTLR *ctlr);

/**
 * Destroy the given CTLR_conf object.
 *
 * @param conf The object to destroy.
 */
void CTLRConf_destroy(struct CTLR_conf *conf);

/*@}*/

#endif /* CTLR_CTLR_CONFIG_H */
