/**
 * @file ctlr_destroy.c
 *
 * @brief Implement functions for destroying the controller.
 *
 * $Id: ctlr_destroy.c,v 1.9.2.3 2003/04/30 20:03:09 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Jul 02 13:45:18 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef HAVE_CONFIG_H
#   include "acconfig.h"
#endif 

#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif

#include "ctlr_destroy.h"
#include "ctlr_config.h"
#include "chaninfo.h"

#include "config/conf_obj.h"

#include "util/acl.h"
#include "util/netaddr.h"
#include "util/list.h"
#include "util/hashtable.h"

static char const rcsid[] = 
    "$Id: ctlr_destroy.c,v 1.9.2.3 2003/04/30 20:03:09 devink Exp $";

void
CTLR_killIMs(CTLR *ctlr)
{
    int                 i;
    LLIterator          itr;
    struct CTLR_imInfo  *im;

    /* make sure we have somthing to do */
    if (ctlr->ims == NULL) 
        return;

    for (i = 0; i < IPC_MAX_IMPROCS; i++) {
        if (ctlr->ims[i] != NULL) 
		{
			DPRINT(DBG_CTL,"CTLR_killIMs() killing %d because in proc table",
					ctlr->ims[i]->pid);
            kill(ctlr->ims[i]->pid, SIGTERM);
		}
    }

	itr = LL_iterator(ctlr->unInitChild);
	while ( (im = LLI_next(itr)) != NULL) {
		DPRINT(DBG_CTL,"CTLR_killIMs() killing %d because unInitChild",im->pid);
		kill(im->pid, SIGTERM);
	}
	LLI_destroy(&itr);
}

void
CTLR_destroyOMs(CTLR *ctlr)
{
    HTIterator          itr;
    struct outchan_info *ocinfo;

    /* make sure we have something to do */
    if (ctlr->conf == NULL || ctlr->conf->ocMap == NULL)
        return;

    itr = HT_iterator(ctlr->conf->ocMap);
    while ( (ocinfo = HTI_next(itr)) != NULL) {
        if (ocinfo->iface != NULL) {
            OMOD_freeInterface(ocinfo->iface);
        }
    }
    HTI_destroy(&itr);
}

void
CTLR_shutdownIPC(CTLR *ctlr)
{
    if (ctlr->ipcHandle == NULL) 
        return;

    IPC_destroy(ctlr->ipcHandle);

    ctlr->ipcHandle = NULL;
}

void
CTLR_shutdownSockets(CTLR *ctlr)
{
    int                 i,j;
    struct CTLR_icType  *type;

    if (ctlr->conf == NULL)
        return;

    for (i = 0; i < IMOD_TYPECOUNT; i++) {
        type = &(ctlr->conf->icTypes[i]);

        if (type->chans == NULL)
            continue;

        for (j = 0; j < type->count; j++) {
            close(type->chans[j]->fd);
        }
    }
}

void 
CTLR_freeMemory(CTLR *ctlr)
{
    int                 i, j;
    struct CTLR_icType  *type;
    struct outchan_info *ocinfo;
    HTIterator          itr;

    /* Input module info */
    if (ctlr->ims != NULL) {
        for (i = 0; i < IPC_MAX_IMPROCS; i++) {
            if (ctlr->ims[i] != NULL) 
                free(ctlr->ims[i]);
        }
    }

    /* Uninitilized child list */
    LL_destroy(&(ctlr->unInitChild));

    /* Path Cache */
    if (ctlr->pcache != NULL) 
        PCACHE_destroy(ctlr->pcache);

    /* from here on everything is in ctlr->conf */ 
    if (ctlr->conf == NULL) 
         return;

    /* routes */
    for (i = 0; i < ctlr->conf->icCount; i++) {
        for (j = 0; j < PRIORITY_COUNT; j++) {
            if (ctlr->conf->routes[i][j] != NULL) 
                LL_destroy(&(ctlr->conf->routes[i][j]));
        }
    }
    free(ctlr->conf->routes);

    /* Input Channels */
    for (i = 0; i < IMOD_TYPECOUNT; i++) {
        type = &(ctlr->conf->icTypes[i]);
        for (j = 0; j < type->count; j++) {
            CONFOBJ_destroyInput(&(type->chans[j]));
        }
        if (type->chans != NULL) 
            free(type->chans);
    }

    /* Output Modules */
    itr = HT_iterator(ctlr->conf->ocMap);
    while ( (ocinfo = HTI_next(itr)) != NULL) {
        CONFOBJ_destroyOutput(&ocinfo);
    }
    HTI_destroy(&itr);
    HT_destroy(ctlr->conf->ocMap);

    /* Other misc cleanup */
    free(ctlr->icSysName);
    free(ctlr->conf);

}

