/**
 * @file ctlr_util.c
 *
 * @brief Misc. functions that are used by the controller
 *
 * $Id: ctlr_util.c,v 1.27.2.7 2003/04/30 20:03:10 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon May 20 12:42:08 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "controller.h"
#include "ctlr_config.h"
#include "chaninfo.h"

#include "privs.h"
#include "const.h"

#include "inputmod/inputmod.h"

#include "error.h"
#include "strcompat.h"
#include "log.h"

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif


static char const rcsid[] 
    = "$Id: ctlr_util.c,v 1.27.2.7 2003/04/30 20:03:10 devink Exp $";

/**@name Misc Controller Functions
 * @ingroup controller
 * @{
 */

/**
 * Start an input module's main function after setting up the
 * environment properly.  This handles things like closing all unneeded
 * FDs, and setting up the IMOD_info object for the input module start
 * function.
 *
 * @param ctlr The controller state to work with
 * @param type The input module type to start
 * @return TRUE on success, FALSE on error.
 */
static void startIM(CTLR *ctlr, IMOD_TYPE type);

/**
 * Check the two indexs used in accessing a route.  This will write an
 * error to the log (via LOG_print()) if either index is out of range.
 * \c icid is not a good index into the routing table, so this function
 * will return the adjusted value of \c icid, which can be used to index
 * the table,
 *
 * @param conf The Config Info
 * @param icid The input channel ID
 * @param pri The priority
 * @return < 0 on out of range, otherwise the adjusted channel index.
 */
static inline int checkRouteIndex(struct CTLR_conf *conf, int icid, int pri);

/**
 * Update the count and other status info for a particular IM Type.
 * This will also start a new IM if the correct type if appropriate and
 * necessary.
 *
 * This is meant to be used whenever we figure out that an IM has died.
 *
 * @param ctlr The Controler object with state info to update.
 * @param iminfo The CTLR_imInfo object for the dead IM
 */
static void updateIMTypeStatus(CTLR *ctlr, struct CTLR_imInfo *iminfo);

bool
CTLR_init(CTLR *ctlr)
{
    size_t      len;
    assert(ctlr != NULL);

    memset(ctlr, 0, sizeof(CTLR));

    strlcpy(ctlr->confFile, DEFAULT_CONF_PATH, PATH_MAX);
    strlcpy(ctlr->sockPath, DEFAULT_IPCSOCK_PATH, UNIX_PATH_MAX);
    strlcpy(ctlr->pidPath, DEFAULT_PID_FILE_PATH, PATH_MAX);
    
    len = strlen(DEFAULT_SYSCHAN_NAME) + 1;
    ctlr->icSysName = emalloc(len);
    strlcpy(ctlr->icSysName, DEFAULT_SYSCHAN_NAME, len);

    /* This either works or it doesn't, IPAC_GENERAL will prevent it from only
     * working until the first SIGHUP, when no "system" IC is defined.
     */
    ctlr->icSysID = IPCAC_GENERAL;
    ctlr->nextICID = IPCAC_GENERAL + 1; /* skip the general channel */

    ctlr->unInitChild = LL_create();

    return TRUE;
}

void
CTLR_incNextICID(CTLR *ctlr)
{
    ctlr->nextICID++;
    if (ctlr->nextICID == IPCAC_GENERAL) 
        ctlr->nextICID++;
}


bool
CTLR_forkNewIM(CTLR *ctlr, IMOD_TYPE type)
{
    pid_t               pid;   
    struct CTLR_imInfo  *info;

    pid = fork();

    if (pid < 0) {
        LOG_eprint("fork()", errno);
        return FALSE;
    }
    else if (pid == 0) {
        /* child */
        DPRINT(DBG_IM, "New Input module started, pid %u, ppid %u, type %d",
                getpid(), getppid(), type);
        startIM(ctlr, type);
        EXIT(EXIT_NORMAL);
        /* NOT REACHED */
    }

    /* parent */

    info = emalloc(sizeof(struct CTLR_imInfo) );
    info->type = type;
    info->pid = pid;
    info->ipcproc = IPCAP_UNSET;

    LL_append(ctlr->unInitChild, info);

    ctlr->conf->icTypes[type].imCount++;

    return TRUE;
}

void
CTLR_sendMessage(CTLR *ctlr, const struct message *msg, int channel)
{
    int             priority;
    int             chanIndex;
    List            omList;
    LLIterator      itr;
    OMOD_interface  *omiface;
#if WITH_DMALLOC
    unsigned long   dm_mark;

    dm_mark = dmalloc_mark();
#endif

    priority = MSG_getPriority(msg);
    chanIndex = checkRouteIndex(ctlr->conf, channel, priority);

    if (chanIndex < 0) return;

    omList = ctlr->conf->routes[chanIndex][priority];

    DPRINT(DBG_CTL, 
        "CTLR_sendMessage: sending message from chan %d, to %d oms",
        channel, omList == NULL ? -1 : LL_size(omList));

    /* see if there are any routes here */
    if (omList == NULL || LL_isEmpty(omList)) 
        return;

    itr = LL_iterator(omList);
    while ( (omiface = LLI_next(itr)) != NULL) {
        omiface->SEND_MSG(omiface->state, msg);
    }
    LLI_destroy(&itr);
#if WITH_DMALLOC
    dmalloc_message("%s:%d checking output modules for leaks", 
            __FILE__, __LINE__);
    dmalloc_log_changed(dm_mark, 1, 0, 0);
#endif

}

bool
CTLR_processIPCEvent(CTLR *ctlr, struct IPC_event *evt)
{
    int                 index;

    assert(ctlr != NULL);
    assert(evt != NULL);

    index = IPC_PAToIndex(evt->addr.proc);

    switch (evt->type) {
        case IPCEVT_IMSTART:
        {
            /* first task is to find the pid on the unilitized child
             * list. 
             */
            LLIterator          itr;
            struct CTLR_imInfo  *iminfo;

            DPRINT(DBG_CTL, "got IPC Event: IMSTART", NO_VARG);

            itr = LL_iterator(ctlr->unInitChild);
            while ( (iminfo = (struct CTLR_imInfo*)LLI_next(itr)) != NULL) {
                if (iminfo->pid == evt->pid) {
                    LLI_remove(itr);
                    break;
                }
            }
            LLI_destroy(&itr);
            if (iminfo == NULL) {
                /* not sure who this his, ignore it */
                /**@todo add IPC call to drop a client w/ extreme predjuice */
                break;
            }

            ctlr->ims[index] = iminfo;
            iminfo->ipcproc = evt->addr.proc;

            /* everything is done here */
            break;
        }
        case IPCEVT_CONEND:
        {
            DPRINT(DBG_CTL, "got IPC Event: CONNEND", NO_VARG);

            if (! PCACHE_clearConn(ctlr->pcache, evt->data) ) {
                LOG_print(LOG_ERROR, "Clearing paths from cache failed");
            }
        }
        case IPCEVT_IMDEATH:
        {
            /* log abnormal termination */
            DPRINT(DBG_CTL, "got IPC Event: IMDEATH", NO_VARG);

            LOG_print(LOG_ERROR, 
                "Input module (pid %u) died unexpectedly with status %d",
                evt->pid, evt->data);

            /* handled the same as a normal death from here on.*/
            /* FALL THROUGH */
        }
        case IPCEVT_IMSTOP:
        {
            struct CTLR_imInfo  *iminfo;

            DPRINT(DBG_CTL, "got IPC Event: IMSTOP", NO_VARG);

            iminfo = ctlr->ims[index];

            /* if there is no registered proc then just give up */
            if (iminfo == NULL)  {
                DPRINT(DBG_CTL, "IMSTOP handler: no registered proc", NO_VARG);
                break;
            }

            /* clear it */
            ctlr->ims[index] = NULL;

            /* sanity check */
            assert(iminfo->ipcproc == evt->addr.proc);

            /* Update the various status bits for this im type.  will
             * also start a new IM if needed.
             */
            updateIMTypeStatus(ctlr, iminfo);

            /* cleanup the old info */
            free(iminfo);
            
            break;
        }
        case IPCEVT_CHLDDEATH:
        {
            LLIterator              itr;
            struct CTLR_imInfo      *iminfo;

            /* log and remove it from unInitChild */
            DPRINT(DBG_CTL, "got IPC Event: CHLDDEATH", NO_VARG);
            LOG_print(LOG_ERROR, 
                "Child (pid %u) died unexpectedly with status %d",
                evt->pid, evt->data);


            itr = LL_iterator(ctlr->unInitChild);
            while ( (iminfo = (struct CTLR_imInfo*)LLI_next(itr)) != NULL) {
                if (iminfo->pid == evt->pid) {
                    LLI_remove(itr);
                    break;
                }
            }
            if (iminfo != NULL) {
                /* Update the various status bits for this im type.  will
                 * also start a new IM if needed.
                 */
                updateIMTypeStatus(ctlr, iminfo);

                /* Done with it */
                free(iminfo);
            }

            break;
        }
        default:
        {
            throw_fatal("Unknown Event type recieved");
            break;
        }
    }

    return TRUE;
}

#if 0 /* unused code */
bool
CTLR_sendAllACLs(CTLR *ctlr, struct CTLR_ichan *icinfo, int proc)
{
    bool            success;
    struct IPC_addr dst;
    IPC             *ipc;

    assert(ctlr != NULL);
    assert(ctlr->ipcHandle != NULL);
    assert(icinfo != NULL);

    ipc = ctlr->ipcHandle;

    dst.proc = proc;
    dst.channel = icinfo->id;

    /**@todo figure out if better error handling is needed here */
    success = FALSE;

    if( (NULL!=icinfo->path_acl) &&
			(! IPC_cSendACL(ipc, &dst, icinfo->path_acl)) ) {
        goto cleanup;    /* Error sending, skip the rest */
    }

    if( (NULL!=icinfo->psec_acl) &&
			(! IPC_cSendACL(ipc, &dst, icinfo->psec_acl)) ) {
        goto cleanup;    /* Error sending, skip the rest */
    }

    if( (NULL!=icinfo->hsrc_acl) &&
			(! IPC_cSendACL(ipc, &dst, icinfo->hsrc_acl)) ) {
        goto cleanup;    /* Error sending, skip the rest */
    }

    if( (NULL!=icinfo->hcon_acl) &&
			(! IPC_cSendACL(ipc, &dst, icinfo->hcon_acl)) ) {
        goto cleanup;    /* Error sending, skip the rest */
    }

    /* all of them sent successfuly */
    success = TRUE;

    /* 
     * This is the end of the attempt to send ACLs.  
     * Any cleanup code goes here.  As does the return.
     */
    cleanup:
        return success;
}
#endif 

void
CTLR_setRoute(struct CTLR_conf *conf, int icid, int pri, OMOD_interface *omi)
{
    int     chanidx;
    List    omList;

    chanidx = checkRouteIndex(conf, icid, pri);

    /* check for out of range */
    if (chanidx < 0) return;

    omList = conf->routes[chanidx][pri];
    if (omList == NULL) {
        omList = LL_create();
        conf->routes[chanidx][pri] = omList;
    }
    LL_append(omList, omi);
}

static int 
checkRouteIndex(struct CTLR_conf *conf, int icid, int pri)
{
    int     chanidx;

    chanidx = icid - conf->baseICID;

    if ( chanidx < 0 || chanidx >= conf->icCount) {
        LOG_print(LOG_ERROR, "Routes: Channel ID (%d) out of range %d .. %d",
                icid, conf->baseICID, conf->baseICID + conf->icCount - 1);
        /* we are done here */
        return -1;
    }

    if (pri < 0 || pri >= PRIORITY_COUNT) {
        LOG_print(LOG_ERROR, "Routes: priority (%d) out of range %d .. %d ", 
                pri, 0, PRIORITY_COUNT - 1);
        /* nothing more to do */
        return -1; 
    }

    return chanidx;
}

static void
startIM(CTLR *ctlr, IMOD_TYPE type)
{
    struct IMOD_info    iminfo;
    struct CTLR_conf    *conf;
    int                 i, j;

    /* Build info to pass into IM init function */
    conf = ctlr->conf;

    iminfo.fd_len = conf->icTypes[type].count;
    iminfo.fd = emalloc(sizeof(int) * 2 * iminfo.fd_len);
    iminfo.chans = iminfo.fd + iminfo.fd_len;

    iminfo.sysChan = ctlr->icSysID;
    iminfo.mark_interval = ctlr->mark_interval;

    iminfo.path_acls = emalloc(sizeof(ACL*) *  iminfo.fd_len);
    iminfo.psec_acls = emalloc(sizeof(ACL*) *  iminfo.fd_len);
    iminfo.hsrc_acls = emalloc(sizeof(ACL*) *  iminfo.fd_len);
    iminfo.hcon_acls = emalloc(sizeof(ACL*) *  iminfo.fd_len);

    for (i = 0; i < iminfo.fd_len; i++) {
        iminfo.fd[i] = conf->icTypes[type].chans[i]->fd;
        iminfo.chans[i] = conf->icTypes[type].chans[i]->id;

        iminfo.path_acls[i] = conf->icTypes[type].chans[i]->path_acl;
        iminfo.psec_acls[i] = conf->icTypes[type].chans[i]->psec_acl;
        iminfo.hsrc_acls[i] = conf->icTypes[type].chans[i]->hsrc_acl;
        iminfo.hcon_acls[i] = conf->icTypes[type].chans[i]->hcon_acl;
    }

	iminfo.ipc_path=emalloc(sizeof(char)*(strlen(ctlr->sockPath)+1));
	strlcpy(iminfo.ipc_path,ctlr->sockPath,strlen(ctlr->sockPath)+1);

    /* now close all uneeded FDs
     * except STD{IN,OUT,ERR} which should be properly dealt with already. 
     */ 
    for (i = 3; i < FD_SETSIZE; i++) {
        /* avoid closing the debugging output FD */
        if (DBG_isOutputFD(i) ) continue;

        /* don't close our input FDs */
        for (j = 0; j < iminfo.fd_len; j++) {
            if (iminfo.fd[j] == i) break;
        }
        if (j >= iminfo.fd_len) {
            close(i);
        }
    }

    /**@todo consider freeing the CTLR object if it is very large */

    /* Tell the logging system that we are in limbo so that it can DTRT,
     * and give people a chance at finding the log messages (DPRINT)
     */
    LOG_initLimbo();

    /* drop privs */
    PRIV_dropPerm();    /* never to regain them again */

    /* prepare for exit handlers */
    EXIT_initIM();  


    /* re-initilize the input waiter, so that we can install our own
     * signal handlers. 
     */
    INW_initialize();

	/* sleep while the controller finishes setup */
	sleep(5);

    switch (type) {
        case IMOD_UDP:
        {
            DBG_newproc("IM-UDP");
			DBG_setProcName("syslogd (UDP input module)");

			IMOD_UDP_start(&iminfo,IMOD_UDP);
            break;
        }
#if HAVE_LIBRR
        case IMOD_BEEP: 
        {
            DBG_newproc("IM-BEEP");
			DBG_setProcName("syslogd (BEEP input module)");

            /* call correct function */ 
            IMOD_BEEP_start(&iminfo, IMOD_BEEP);
			break;
        }
#endif
        case IMOD_SYSTEM:
        {
            DBG_newproc("IM-SYS");
			DBG_setProcName("syslogd (System input module)");

			IMOD_UDP_start(&iminfo,IMOD_SYSTEM);
			break;
        }
        case IMOD_MARK:
        {
            DBG_newproc("IM-MARK");
            DBG_setProcName("syslogd (MARK input module)");

            IMOD_MARK_start(&iminfo, IMOD_MARK);
            break;
        }
        default:
        {
            throw_fatal("Attempt to fork unimplemented input module type");
            break;
        }
    }
}

static void
updateIMTypeStatus(CTLR *ctlr, struct CTLR_imInfo *iminfo)
{

    /* update the IM count */
    ctlr->conf->icTypes[iminfo->type].imCount--;

    DPRINT(DBG_CTL, "updateIMTypeStatus: imtype = %d (%d)", 
            iminfo->type, ctlr->conf->icTypes[iminfo->type].imCount);

    /* 
     * If we are not waiting for input modules to exit 
     * And we don't have any of this IM type running but are
     * supposed to ... then we need to start a new one.
     */
    if (! ctlr->imClear
        && ctlr->conf->icTypes[iminfo->type].imCount <= 0
        && ctlr->conf->icTypes[iminfo->type].count > 0)
    {
        LOG_print(LOG_INFO, 
            "Last %s IM died, trying to start a new one.",
            IMOD_typeNames[iminfo->type]);
        /* start a new IM of the correct type. */
        if (! CTLR_forkNewIM(ctlr, iminfo->type) ) {
            LOG_print(LOG_ERROR, 
                "im start failed: No more %s IMs running.",
                IMOD_typeNames[iminfo->type]); 
        }
    }
}
/*@}*/
