/**
 * @file debug.h
 *
 * @brief Functions to support printf debugging
 * @par
 *
 *
 * $Id: debug.h,v 1.19.2.7 2003/04/30 20:03:12 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Apr 08 11:57:43 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INC_DEBUG_H
#define INC_DEBUG_H

#ifdef HAVE_CONFIG_H
# include "acconfig.h"
#endif

#include "types.h"
#include "ipc/buffer.h"

/**@defgroup debug Debugging Support Functions
 * @{
 */ 

/**
 * Macro used to print a debugging message if compiled to do such.
 * @note You must have at least one variable arugemnt.  If one is not
 * needed (besides the $#%!ing C99 brain damage) just use \c NO_VARG.
 *
 * @warning This uses a C99 specific feature.
 * @todo This breaks older compilers, and in general this macro really
 *       fucking sucks.  Something needs to be done about the older
 *       compiler issue, and about the warnings.  In fact, as DBG_print
 *       will check for required args anyways, we can kill all the
 *       required args and simply use 'args...' or '...'.  This also
 *       allows removal of all the VARG shit.  Consider doing the same
 *       for FlexML varargic macro.
 *
 * @param mask Debug mask
 * @param msg Debug message (printf style format)
 * @param ... Variable args
 */
#if ENABLE_DEBUG
# define DPRINT(mask, msg, ...) \
    DBG_print(mask, msg, __FILE__, __LINE__, __VA_ARGS__)
#else   /* !ENABLE_DEBUG */
# define DPRINT(mask, msg, ...) /* nothing */
#endif  /* ENABLE_DEBUG */

/**
 * Macro used to print a timestamp message.  This is primarily used when
 * doing wall clock timing of various parts of the system.
 * 
 * @param loc A name for the location of the measurement.
 */
#if ENABLE_DEBUG
# define TPRINT(loc) \
    DBG_print(DBG_TM, "timing: %s: %u", __FILE__, __LINE__, loc, time(NULL))
#else
# define TPRINT(loc) /* nothing */
#endif

/**
 * Macro used to print an IPC buffer for debugging purposes 
 * @param mask Debug mask
 * @param msg Debug Message
 * @param buf THe IPC_buffer to print
 * @param len How much of \c buf to print, or -1 for all of it.
 */
#if ENABLE_DEBUG
# define DPRINT_BUF(mask, msg, buf, len) \
    DBG_printBuf(mask, msg, __FILE__, __LINE__, buf, len);
#else
# define DPRINT_BUF(mask, msg, buf, len) /* nothing */
#endif

/** 
 * No Variable Argumens for the DPRINT macro place holder.
 * C99 macro Varargs are quite silly.
 */
#define NO_VARG     0

/**
 * Print a debugging message to the debug output if the mask matches.
 * This is the real function that does the work of the DPRINT macro.
 *
 * @param mask Debug mask
 * @param msg Debug message (printf style formating)
 * @param f current file (__FILE__)
 * @param l current line (__LINE__)
 * @param ... variable args
 */
void DBG_print(char mask, char const *msg, char const *f, int l, ...);

/**
 * Print an IPCBuffer to the debugging output.  This also prints msg
 * before the printout of the buffer.
 *
 * @param mask Debug mask
 * @param msg The prefix message.
 * @param f current file
 * @param l current line
 * @param buf The buffer to print
 * @param len How many bytes from \c buf to print, -1 means all
 */
void DBG_printBuf(char mask, char const *msg, char const *f, int l, 
        struct IPC_buffer const *buf, int len);

/**
 * Print a fatal error message.
 *
 * @param msg The Message to print
 * @param f The file that the error occured in.
 * @param l The line that the error occured on.
 */
void DBG_printfatal(char const *msg, char const *f, int l);


/**
 * Initilize the debug library.  The mask string consists of a series of
 * masks which should be printed.  
 * On error, your best bet is to check errno, as the error probably
 * occured when opening \c ofile.
 *
 * @param mask The types of debug messages to print.
 * @param ofile The file to write messages to, "-" == STDOUT.
 * @return TRUE on success, FALSE otherwise.
 */ 
bool DBG_init(char const *mask, char const *ofile);

/**
 * Set a name for the current process, this helps in identifing which
 * process a message was generated by.  
 *
 * @param name The name to use
 */
void DBG_newproc(char const *name);

/**
 * Check an FD against the debugging output FD, return true if they
 * match.  
 *
 * @param fd The FD to check
 * @return TRUE if fd is the debug output FD.
 */
bool DBG_isOutputFD(int fd);

/** 
 * Initialize the static globals (inside of debug.c) that let us change the
 * process name
 * @note Clobbers the environment and argv
 *
 * @param argc_p From main
 * @param argv_p From main
 */
void DBG_initProcName(int argc_p,char **argv_p);

/** 
 * Change the name of the current process
 * First, to be safe, this overwrites all the possible space for the name with
 * 0's before writing the new process name.
 * Note that DBG_initProcName() must have been run at program start for this to 
 * work.
 *
 * @param name The new name
 */
void DBG_setProcName(char *name);

/** 
 * Loop indefinitely; this WILL hang your program, unless GDB or similar is
 * used to break the loop.
 * Use 'set loopinfo=0' in GDB.
 */
#define DBG_LOOPINF() \
	{ \
		int loopinf=1; \
		while(loopinf) \
			;	/* do nothing */ \
	} \

/*@}*/

/**@name Debugging Masks
 * @ingroup debug
 * @{
 */
#define DBG_IPC     'i'     /**< IPC module debugging */
#define DBG_DBG     'd'     /**< Debug module Messages */
#define DBG_BUF     'b'     /**< IPC Buffer module debugging */
#define DBG_BEEP	'B'		/**< BEEP profile debugging */
#define DBG_ACL     'a'     /**< ACL module debugging */
#define DBG_OMD		'o'		/**< Output modudel debugging */
#define DBG_PRIV    'p'     /**< PRIV messages */
#define DBG_PC      'P'     /**< Path Cache */
#define DBG_AVL     'A'     /**< AVL Tree */
#define DBG_CTL		'c'		/**< Controller messages */
#define DBG_HT      'h'     /**< Hash Table */
#define DBG_IM      'I'     /**< Input Module debugging */
#define DBG_NA      'n'     /**< Net address debugging */
#define DBG_CNF     'C'     /**< Config module debugging */
#define DBG_TM      't'     /**< Timing messages */
#define DBG_MSG     'm'     /**< Message debbugging (parsers/formatters) */
/*@}*/

#endif /* INC_DEBUG_H */
