/**
 * @file error.h
 *
 * @brief This defines some functions for coping with fatal errors.  
 * @par
 * It includes emalloc() for error checked malloc()s.  Functions for
 * registering cleanup functions. And functions for handling fatal
 * errors.
 *
 * $Id: error.h,v 1.10.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Jan 28 15:32:26 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INC_ERROR_H
#define INC_ERROR_H

#ifdef HAVE_CONFIG_H
# include "acconfig.h"
#endif

/* assert() is your friend, USE IT */
#if !ENABLE_DEBUG
# define NDEBUG 1   /* remove assert statements if !ENABLE_DEBUG */
#endif
#include <assert.h>

#include "controller/controller.h"
#include "ipc/inputwait.h"
#include "ipc/ipc.h"

/**@defgroup exit Functions to Exit in various ways
 * @{
 */

/** 
 * An error checked malloc. This will try to malloc() some memory, and
 * throw a fatal error condition if that fails.
 *
 * @note This doesn't really belong in this doc group ... but until I
 * find a better one.
 *
 * @param size The number of bytes to allocate.
 * @return a pointer to the allocated memory.
 */
#define emalloc(size)   errorMalloc(__FILE__, __LINE__, (size))


/** 
 * Signal a fatal error. This makes an attempt to clean up the
 * important stuff, print a message and exits().
 * @see throwFatal()
 * @param msg The message to print
 */
#define throw_fatal(msg) throwFatal(__FILE__, __LINE__, (msg) )

/**@defgroup exit_codes Exit Status Codes
 * @ingroup exit
 * @{
 */
#define EXIT_NORMAL     0       /**< Successful exit */
#define EXIT_ERROR      1       /**< Error, but try to clean up */
#define EXIT_IMMED      2       /**< Error, quit now without a lot of cleanup */
#define EXIT_NOCLEAN    0x80	/**< flag, no cleanup */
/*@}*/


/**
 * Perform an exit in response to a fatal error.  This will _attempt_
 * to cleanup what it can/needs to ... but operates on the assumption
 * that the system is pretty throughly hosed.
 *
 * To avoid doing any cleanup use the EXIT_NOCLEAN flag.
 *
 * This will never return.
 *
 * @param code exit code (\ref exit_codes)
 */
void EXIT(int code);

/**
 * Set the CTLR object that will have the information needed for
 * cleaning things up.
 *
 * @param ctlr The CTLR object with the needed info.
 */
void EXIT_setCTLR(CTLR *ctlr);

/**
 * Do some initilization for a newly forked IM process so that it can
 * use the EXIT_ functions to perform a clean exit.
 *
 */
void EXIT_initIM(void);

/**
 * Set the IPC handle to cleanup when an input module exits
 * @param ipc The IPC object to cleanup
 */
void EXIT_setIM_IPC(IPC *ipc);

/**
 * Set the INW object to cleanup on exit.
 * @param inw The INW object to cleanup.
 */
void EXIT_setINW(INW *inw);

/*@}*/

/** 
 * The actual function that implements emalloc().
 * @internal
 * @param file The name of the calling file.
 * @param line The line number where we were called
 * @param size The number of bytes to allocate.
 * @return a pointer to the allocated memory
 */
void* errorMalloc(char const *file, int line, size_t size);

/** 
 * Signal a fatal error.  This actually implements the throw_fatal()
 * macro, and can be used if the file/line to report differes from where
 * this is called.
 * @internal
 *
 * @param file The file the error occured in.
 * @param line The line the error occured on.
 * @param msg The error message.
 */
void throwFatal(char const *file, int line, char const *msg);

#endif /* INC_ERROR_H */
