/**
 * @file log.h
 *
 * @brief Internal Logging functions.  
 * @par
 * Once initilized the LOG_print()/LOG_eprint function will simply DTRT.  
 * This meas sending an ipc message for input modules, going direct to an 
 * output module for the controller, or printing to screen if no OMs have 
 * been setup yet.
 *
 * $Id: log.h,v 1.5.2.5 2003/04/30 20:03:12 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed May 15 12:39:05 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INC_LOG_H
#define INC_LOG_H

#include "controller/controller.h"
#include "ipc/ipc.h"
#include "util/tsqueue.h"

/**@defgroup log Logging Functions
 * @{
 */

/**
 * Write a log message to correct output device.
 *
 * @param sev The severity of the message.
 * @param fmt The printf like format
 * @param ... Arguments for the format.
 */
void LOG_print(int sev, char const *fmt, ...);

/**
 * Write an error message that uses errno to figure out the message.
 *
 * The output text will be in the form:
 *      Error: \c fmt: strerror(\c err)
 * It is always logged at severity LOG_ERROR.
 * @warning This has limited internal buffer space... exceeding that
 * buffer space will lead to truncation of \c fmt.  This does not try to
 * expand the arguments in \c fmt.  Do try to keep \c fmt from being too long. 
 *
 * @param fmt The printf format string.
 * @param err The error code to print (usually errno)
 * @param ... Arguments for the format
 */
void LOG_eprint(char const *fmt, int err, ...);

/**
 * Initilize the logging system for the controller
 *
 * @param ctlr The controller state that is needed to figure out where
 *             the output goes
 * @return TRUE on success, FALSE otherwise
 */
bool LOG_initCTLR(CTLR *ctlr);

/**
 * Initilize the logging system for input modules
 *
 * @param ipc The IPC handle that will send messages to the right palce.
 * @param icSysId The input channel that messages should come from.
 *
 * @return TRUE on success, FALSE otherwise
 */
bool LOG_initIM(IPC *ipc, uint32_t icSysId);

/**
 * Initilize the logging system to put messages in a queue.
 * This is used by threaded input modules.
 *
 * @param queue The TSQueue object to put messages in.
 * @param sysid The input channel that mesasges should come from.
 *
 * @return TRUE on success, FALSE otherwise.
 */
bool LOG_initTSQ(TSQueue *queue, uint32_t sysid);

/**
 * Initilize the logging system for that limbo period between forking
 * a new child process and starting the Input Module IPC layer.
 *
 * This uses DPRINT to log things.
 * @return TRUE on success
 */
bool LOG_initLimbo(void);

/**
 * Initilize the logging system for use prior to read the config for the
 * first time.
 *
 */
void LOG_init(void);

/*@}*/

/**@name Log Severities
 * @ingroup log
 * @{
 */
#define LOG_DEBUG   1   /**< Debug messages */
#define LOG_INFO    2   /**< Informational messages */
#define LOG_ERROR   3   /**< Error messages */
#define LOG_CRIT    4   /**< Critical/Fatal Error messages */
/*@}*/

#endif /* INC_LOG_H */
