/**
 * @file privs.h
 *
 * @brief This file defines the functions to manipulate privledge levels
 * and performe operations that need privledges.
 * @par
 * Once the Main process (Controller) has initilized it's self, it
 * should then drop privileges via PRIV_dropTemp().  Once this is done
 * than any operation which needs to be done as root must be done by
 * calling a function in this header.  This helps place all the
 * functions that are run with privileges in one place.
 * @par
 * Input modules must call PRIV_dropPerm() \b immedately after forking.
 * this will permantly prevent the input modules from re-gaining root.
 *
 * @note The function to regain root privilege is not exported to the rest of
 * the program for a reason (it's static in privs.c). <b>DO NOT EXPORT IT</b>
 *
 *
 * $Id: privs.h,v 1.10.2.4 2003/04/30 20:03:13 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Mar 08 11:09:12 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CTRL_PRIVS_H
#define CTRL_PRIVS_H

#include <sys/socket.h>     /* for 'struct sockaddr' def */
#include <sys/un.h>         /* for 'struct sockaddr_un' def */

/**@defgroup privs Privileged Functions
 * These are functions that handle privileges and operations which need
 * privileges.
 * @see privs.h for more information
 * @{
 */

/** 
 * Set the user and group to change to when privilges are dropped.
 *
 * @param user The user name to change to
 * @param group The group name to change to
 * @return TRUE on success, FALSE otherwise
 * @post On error PRIV_errno will be set to the error value.
 */
bool PRIV_setUnprivUser(char const *user, char const *group);

/**
 * Disable use of the privilges code.
 * @warning This will only cooperate if the program is not running as
 *          root.  If it is running as root then this function will
 *          exit.
 */
void PRIV_disableChanges(void);

/** 
 * Drop privileges in a manner that allows them to be regained.
 *
 * @return TRUE on success, FALSE on error.
 * @pre PRIV_setUnprintUser() must have completed successfully.
 * @post On error PRIV_errno will be set to the error value.
 */
bool PRIV_dropTemp(void);

/** 
 * Drop privileges perminantly, they can not be regained later.
 *
 * @return TRUE on success, FALSE on error.
 * @pre PRIV_setUnprintUser() must have completed successfully.
 * @post On error PRIV_errno will be set to the error value.
 */
bool PRIV_dropPerm(void);

/** 
 * Open a file for reading which may be readable only to root.  This
 * makes several checks:
 * \li The file must be writable only by root and group 0.
 *
 * Also if this function is called when the program is root, it will not
 * drop privileges when it's done.
 *
 * If there is an error an appropriate message will be written via
 * WriteLog().
 *
 * @note Only certain option flags will be allowed in \c flags:
 * \li O_RDONLY
 * \li O_NOCTTY
 * \li O_NONBLOCK or O_NDELAY
 * \li O_LARGEFILE
 *
 * @param name The name of the file to open
 * @param flags Option flags for open.
 * @return a FD for the file, or < 0 on error
 */
int PRIV_openReadFile(char const *name, int flags);

/** 
 * Open a file for writing. These conditions must be meet for the file
 * to be successfully opened:
 * \li It must not be a regular file, character device, or fifo (no
 *     symlinks, directories, etc...).  
 * \li If it exists it will be appended to
 *
 * If there is an error, the appropriate message will be written via
 * WriteLog().
 *
 * @note Only certain flags will be allowed for \c flags:
 * \li O_WRONLY
 * \li O_NONBLOCK or O_NDELAY
 * \li O_SYNC
 * \li O_LARGEFILE
 * \li O_CREAT
 * \li O_EXCL
 *
 * And the following flags will be added to \c flags:
 * \li O_WRONLY
 * \li O_NOFOLLOW (if present)
 * \li O_APPEND
 *
 * @param name The name of the file to open.
 * @param flags The open flags
 * @return a FD for the file, or < 0 on error.
 */
int PRIV_openWriteFile(char const *name, int flags);

/** 
 * Bind a socket to a particular address.  This may be any address
 * family.  If privileges are possesed on entering the function then
 * they will remaine on exit.  It allows binding to ports < 1024 because
 * the bind(2) will always be executed with privileges.
 *
 * In case \c sa->sa_family==PF_UNIX, the umask is set to 077 before
 * bind(2).
 *
 * On error the appropriate message will be printed and an error
 * condition will be returned.
 *
 * @param sock The socket to bind to an address
 * @param sa The sockaddr structure that holds the address
 * @param salen The size of the sockaddr structure.
 * @return TRUE on success, FALSE on error.
 */
bool PRIV_bindSocket(int sock, struct sockaddr *sa, socklen_t salen);

/**
 * This is a special bind function that specifically binds a unix domain
 * socket with some special processing.  
 *
 * This allows changing the owner to be the unprivleged user, or leaving
 * it as root:root.  Also it allows changing the permissions after
 * creation.  The default permissions are to bind to the socket with a
 * umask of 077.
 *
 * This will unlink(2) the socket, bind(2) to it, and then
 * chown(2)/chmod(2).  
 *
 * @param sock The socket to bind to
 * @param sun The sockaddr for the bind call.
 * @param unprivOwner Change owner to be unprivleged user if TRUE
 * @param mode The mode to change the socket to, if 0 use default. 
 * @return TRUE on sucecss, FALSE otherwise.
 */
bool PRIV_bindUnixSocket(int sock, struct sockaddr_un *sun, 
                         bool unprivOwner, mode_t mode);

/**
 * Delete a file that is in a root owned directory.  This does a simple
 * unlink(2).
 *
 * This will ignore errors created by tring to unlink non-existent files.
 *
 * @param file The full path to the file
 * @return TRUE on success, FALSE on error.
 */
bool PRIV_unlink(char const *file);

/**
 * Change the mode (permissions) of a file.  This requires that the file
 * be opened first.
 *
 * @param fd The file descriptor to change the perms on.
 * @param mode The new mode of the file.
 */
bool PRIV_fchmod(int fd, mode_t mode);

/*@}*/

#endif /* CTRL_PRIVS_H */
