/**
 * @file test.h
 *
 * @brief Stuff to support the minor differences between testing an not
 * testing builds.
 * @par
 * The unfortunate reality is that inorder to do some of the unit
 * testing it is necessary to change some small aspects of runtime
 * behavior.  I *HATE* having to do this but there is no way around it
 * at the moment, this file makes it as *ugly* as possible to do.  All
 * unit test special case code should be contained in either the
 * NOT_IN_UNIT_TEST macro, or IN_UNIT_TEST macro.
 *
 * The current list of situations which force use of this is:
 *
 * \li IPC Unit tests don't get SIGCHLD.
 *
 * $Id: test.h,v 1.3.2.6 2003/04/30 20:03:13 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Jun 24 17:40:25 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INC_TEST_H
#define INC_TEST_H

#ifdef HAVE_CONFIG_H
# include "acconfig.h"
#endif

#include "types.h"

#if !ENABLE_TESTING
/* all unit tests are run with ENABLE_TESTING */
# define NOT_IN_UNIT_TEST(stmt)  stmt
# define IN_UNIT_TEST(stmt)  

/* Make sure this is false */
#undef IN_UNIT_TEST_CT
#define IN_UNIT_TEST_CT 0

#else /* ENABLE_TESTING */
/** Are we doing a unit test? NEVER use this directly for conditionals. */
extern bool unit_test;

/** A code block that is not executed when unit testing */
#define NOT_IN_UNIT_TEST(stmt)    do {if (!unit_test) {stmt;}}while (0)

/** A code block that is only executed when unit testing */
#define IN_UNIT_TEST(stmt)        do {if (unit_test) {stmt;}}while (0)

#endif /* ENABLE_TESTING */

/** 
 * Compile time variable for determining if we are doing unit tests.
 * This is needed so that parts of the exiterr.c and log.c files don't
 * pull in functions from the controller.
 */
#ifndef IN_UNIT_TEST_CT
# define IN_UNIT_TEST_CT 0
#endif

/**@define UT_FAIL
 * Defined to be fail() from libcheck IFF we are complinig unit tests
 * (the only time libcheck is linked in.  Otherwise it's just empty
 */
#if IN_UNIT_TEST_CT
# define UT_FAIL(msg) fail(msg)
#else
# define UT_FAIL(msg) 
#endif

#endif /* INC_TEST_H */
