/**
 * @file imod_beep.c
 *
 * @brief BEEP Input
 * @par
 * Provides an interface to reading from a BEEP socket
 *
 * @design d2_3
 *
 * $Id: imod_beep.c,v 1.9.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author  Matt Small <small@sdsc.edu>
 * @date    Fri Sep  6 10:53:24 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>

#include "error.h"
#include "log.h"
#include "ipc/inputwait.h"
#include "ipc/ipc.h"
#include "ipc/ipc_err.h"

#include "librr/rr.h"
#include "util/profile-beep.h"
#include "util/hashtable.h"

#include "inputmod.h"

static char const rcsid[]
	= "$Id: imod_beep.c,v 1.9.2.7 2003/10/22 04:09:03 devink Exp $";

/** How long to wait before we stop waiting for connections to close, in
 * seconds */
static int exit_wait = 60;

/** Send all messages on the queue to the controller, through the IPC layer
 *
 * @param ipc_socket The IPC socket to use
 * @param msgqueue The message queue to use
 * @param pathcache The pathcache to add paths to
 */
static void IMOD_BEEP_sendAll(IPC *ipc_socket, TSQueue *msgqueue, 
                                PCache *pathcache, HashTable connections);

static void closeAllConnections(HashTable connections);

static void finishConnection(HashTable connections, uint32_t connid);

/* Must be a power of 2 */
#define CON_TBL_SZ  1024

static HT_key_t
contbl_hash(void *keydata) {
    /* keydata is just the pointer to the connection object, we just use
     * that value (as an uint to key the hash)
     */
    return (HT_key_t)keydata;
}

static int
contbl_cmp(void *lhs, void *rhs)
{
    /* Done this way b/c I'm too lazy to figure out what would happen
     * with sign/unsign conversions 
     */
    if (lhs < rhs)
        return -1;
    if (lhs > rhs)
        return 1;
    return 0;
}

void
IMOD_BEEP_start(struct IMOD_info *info, IMOD_TYPE type)
{
	GError *error=NULL;
	RRProfileRegistry *profreg;
	BEEP_conn *beepconn;

	INW *waiter;
	IPC *ipc_socket;
	struct INW_eInfo *event;

	int i;
	int first_id,chan_id;
	int accept_fd;

	struct sockaddr_in addr;
	unsigned int addrlen;

	PCache *pcache;

	TSQueue *msgqueue;
    bool acceptConnections = TRUE;
    HashTable connections;

	if( IMOD_BEEP != type)
		throw_fatal("IMOD_BEEP_start() used on non-recognized type");

	DPRINT(DBG_IM, "Starting BEEP input module", NO_VARG);

	/* start the RR library */
	BEEP_LI_init();

    connections = HT_create(CON_TBL_SZ, contbl_hash, contbl_cmp);

	/* start the IPC layer */
	DPRINT(DBG_IM, "Starting IPC", NO_VARG);

	ipc_socket=IPC_initInputModule(info->ipc_path);
	if(IPCE_SUCCESS!=IPC_geterrno(ipc_socket))
	{
		LOG_print(LOG_ERROR, "Failed to initilize the IPC socket: (%d) %s",
				IPC_geterrno(ipc_socket), IPC_geterrstr(ipc_socket));

		EXIT(EXIT_ERROR);
	}

	/* signal the controller that we're ready */
	if(!IPC_inputmoduleStart(ipc_socket))
	{
		LOG_print(LOG_ERROR,
				"BEEP IM: Failed to communicate with controller: (%d) %s",
				IPC_geterrno(ipc_socket), IPC_geterrstr(ipc_socket));

		EXIT(EXIT_ERROR);
	}
	EXIT_setIM_IPC(ipc_socket);

	/* begin real logging */
	LOG_initIM(ipc_socket, info->sysChan); 


	DPRINT(DBG_IM, "Creating input waiter", NO_VARG);
	waiter=INW_create();
	if(!INW_addIPC(waiter,ipc_socket))
	{
		LOG_print(LOG_ERROR,"BEEP IM: Failed to add IPC to input waiter");
		EXIT(EXIT_ERROR);
	}
	EXIT_setINW(waiter);

	if (! INW_addSignal(waiter, SIGTERM) )
	{
		LOG_print(LOG_ERROR, "UDP BEEP: Failed to add SIGTERM handler");
		EXIT(EXIT_ERROR);
	}

	first_id=INW_getNextEventID(waiter);
	for(i=0;i<info->fd_len;i++)
	{
		if(!INW_addFD(waiter,info->fd[i],first_id+i))
		{
			LOG_print(LOG_ERROR,"BEEP IM: Failed to add FD to input waiter");
			EXIT(EXIT_ERROR);
		}
	}

	/* finish the RR layer; create the profile registry */
	profreg = rr_profile_registry_new();
	rr_profile_registry_add_profile(profreg, BEEPRAW_getType(), NULL);
	rr_profile_registry_add_profile(profreg, BEEPCOOK_getType(), NULL);

	/* init the path cache and message queue */
	pcache=PCACHE_imCreate();

	msgqueue=TSQ_create();
	if(!INW_addQueue(waiter,msgqueue))
	{
		LOG_print(LOG_ERROR, "BEEP IM: Failed to add TSQueue to input waiter");
		EXIT(EXIT_ERROR);
	}


	DPRINT(DBG_IM, "Starting the BEEP IM main loop, type: %d", type);

    /* Loop while we are accepting connections or still have some
     * connections waiting to die.
     */
	while(acceptConnections || HT_size(connections) > 0)
	{
		/* get new event */
		event=INW_getEvent(waiter);

        DPRINT(DBG_IM, "Open connections: %d; event = %p", 
                HT_size(connections), event);
                

		/* check for errors */
        /*
		if(0!=IPC_geterrno(ipc_socket))
		{
			throw_fatal(IPC_geterrstr(ipc_socket));
		}
		else 
        */
        if(NULL==event)
		{
			throw_fatal("Input waiter return NULL (controller death)");
		}

        DPRINT(DBG_IM, "Got Event id = %d, info = %d", 
            event->event_id, event->info);

		/* IPC event waiting */
		if(INWE_IPC==event->event_id)
		{
			DPRINT(DBG_IM, "Received IPC event %d", event->info);
		}
		/* signal event */
		else if(event->event_id == INWE_SIG) {
			if (event->info == SIGTERM) {
				DPRINT(DBG_IM, "Received signal: SIGTERM", NO_VARG);
                if (! INW_addSignal(waiter, SIGALRM) ) {
                    LOG_print(LOG_ERROR, "Failed to add SIGALRM handler");
                    break;
                }
                /* Set an alarm so that we don't sit around for too
                 * long. 
                 */
                alarm(exit_wait);

                /* blow the listen sockets */
                DPRINT(DBG_IM, "closing listening sockets (%d)", info->fd_len);
                for(i=0; i < info->fd_len; i++)
                {
                    close(info->fd[i]);
                }
                /* Tell the Input watier to only look at the queue */
                INW_stopFDWaiting(waiter);

                /* need to close channels */
                closeAllConnections(connections);
                /* stop accepting connections */
                acceptConnections = FALSE;
			}
            else if (event->info == SIGALRM) {
                /* given up waiting ... just close */
                DPRINT(DBG_IM, "Recieve signal: SIGALRM", NO_VARG);
                break;
            }
			else
			{
				DPRINT(DBG_IM, "Received signal %d", event->info);
			}
		}
		/* Thread safe queue event */
		else if(event->event_id == INWE_TSQ) {
			/* remove everything from the message queue */
			IMOD_BEEP_sendAll(ipc_socket,(TSQueue *)event->info,pcache, 
                    connections);
		}
		/* FD event */
		else if (acceptConnections)
		{
			DPRINT(DBG_IM, "Received event %d, info %d",
					event->event_id, event->info);

			/* calculate the channel id through brute forece */
			chan_id=-1;
			for(i=0;i<info->fd_len;i++)
			{
				if(event->info==info->fd[i])
				{
					chan_id=info->chans[i];
					break;
				}
			}
			if(-1==chan_id)
			{
				LOG_print(LOG_ERROR,"Error: no channel ID found for FD: %d",
						event->info);
				continue;
			}

			/* accept the connection */
            DPRINT(DBG_IM, "Accepting new connection", NO_VARG);
			addrlen=sizeof(struct sockaddr_in);
			accept_fd=accept(event->info,(struct sockaddr *)&addr,&addrlen);
			if(-1==accept_fd)
			{
				LOG_eprint("Error in accept()",errno);
				continue;
			}

			DPRINT(DBG_IM, 
                "Calculated channel to be: %d, fd: %d; connecting to RR", 
                chan_id, accept_fd);

			/* pass the connection into the RR layer */
            DPRINT(DBG_IM, "Passing new connection to RR", NO_VARG);
			beepconn = BEEP_conn_new(profreg, NULL);

			beepconn->channel_id = chan_id;
			beepconn->inmsgs = msgqueue;

			rr_tcp_connection_connect_fd(
					RR_TCP_CONNECTION(beepconn), accept_fd, RR_ROLE_LISTENER,
					&error);
            /* keep our own list of connections */
            HT_insert(connections, beepconn, beepconn);

			if(NULL!=error)
			{
				LOG_print(LOG_ERROR, "rr_tcp_connection_connect_fd() error: %s",
						error->message);

				EXIT(EXIT_ERROR);
			}
		}
	} /* end while */

	DPRINT(DBG_IM, "Out of the while loop, attempting cleanup; %d conn open", 
			HT_size(connections));

	DPRINT(DBG_IM, "Trying rr_exit()", NO_VARG);

	rr_exit(&error);

	/* make sure we don't have any thing more on the queue */
	IMOD_BEEP_sendAll(ipc_socket, msgqueue, pcache, connections);

	/* do some clean up */
	/*free(info);*/
	TSQ_destroy(&msgqueue);

	/* destroy the path cache and its mutex */
	PCACHE_destroy(pcache);

    TPRINT("BEEP IM Exit");

	EXIT(EXIT_NORMAL);
}

static void
closeAllConnections(HashTable connections) 
{
    HTIterator  itr;
    BEEP_conn   *conn;

    itr = HT_iterator(connections);

    while ( NULL != (conn = (BEEP_conn*)HTI_next(itr))) {
        BEEP_LI_shutdown(conn);
    }
}

static void
finishConnection(HashTable connections, uint32_t key)
{
    BEEP_conn   *conn;

    conn = (BEEP_conn*)HT_remove(connections, (void*)key);

    if (conn != NULL) {
        BEEP_LI_shutdown_final(conn);
        /*g_object_unref(G_OBJECT(conn));*/
    }
}


static void
IMOD_BEEP_sendAll(IPC *ipc_socket, TSQueue *msgqueue, PCache *pathcache, 
                    HashTable connections)
{
	struct MSG_qelem *msgelem;

	DPRINT(DBG_IM, "We have %d messages on msgqueue (pc: %X)",
			TSQ_size(msgqueue),pathcache);
	while(!TSQ_isEmpty(msgqueue))
	{
		msgelem=(struct MSG_qelem *)TSQ_dequeue(msgqueue);

		switch(msgelem->type)
		{
			case MSGQE_MSG:
				DPRINT(DBG_IM, "Found message: %s on the queue",
						msgelem->data.msg->msg);
				if(!IPC_imSendMessage(ipc_socket, msgelem->chan,
							msgelem->data.msg))
					throw_fatal("Failed to send message to the controller"
                            " (controller death)");

				MSG_free(msgelem->data.msg);
				free(msgelem->data.msg);	/* XXX remove this; see message/
											   message.c:133 */
				free(msgelem);
				break;

			case MSGQE_PATH:
				DPRINT(DBG_IM, "Found path on queue", NO_VARG);
				/* make sure the path ID is not 0 */
				if(msgelem->data.path->id==PATH_UNSET_ID)
				{
					LOG_print(LOG_ERROR, 
                        "Received path with id: %d (invalid id)",
                        msgelem->data.path->id);
					break;
				}

				/* If we can add the path to our path cache, then there's no
				 * conflict, so send it to the controller */
				/** @todo ACL */
				if(PCACHE_addPath(pathcache, msgelem->data.path, TRUE))
				{
					if(!IPC_imSendPath(ipc_socket, msgelem->chan,
								msgelem->data.path))
						throw_fatal("Failed to send message to the "
                                "controller (controller death)");
				}

				PATH_free(msgelem->data.path);
				free(msgelem);
				break;
            case MSGQE_CLOSE:
                DPRINT(DBG_IM, "Found close event on queue: %d",  
                        msgelem->data.connid);
                finishConnection(connections, msgelem->data.connid);
                free(msgelem);
                break;
			default:
				DPRINT(DBG_IM, 
                    "Got a message with unknown type off of the message queue",
                    NO_VARG);
				free(msgelem);
				break;
		}
	}

	DPRINT(DBG_IM, "Done flushing msgqueue", NO_VARG);
}
