/**
 * @file imod_mark.c
 *
 * @brief An input module that generates 'mark' messages.
 *
 * $Id: imod_mark.c,v 1.1.2.7 2003/04/30 20:03:15 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Sep 03 15:46:33 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>

#include "strcompat.h"
#include "log.h"
#include "error.h"

#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/inputwait.h"

#include "config/priority.h"

#include "const.h"

static char const rcsid[]
    = "$Id: imod_mark.c,v 1.1.2.7 2003/04/30 20:03:15 devink Exp $";

/** The template for mark messages */
#define MARK_MSG_TMPL "[%d]: -- MARK --"

/**
 * Create a message object for the mark message.
 *
 * @param msg The struct to fill in.
 */
void makeMarkMsg(struct message *msg);

void IMOD_MARK_start(struct IMOD_info *info, IMOD_TYPE type)
{
    INW                 *waiter;
    IPC                 *ipc;
    struct INW_eInfo    *event;
    struct message      msg_mark;
    int                 alrm_sec;

    if (type != IMOD_MARK) 
        throw_fatal("IMOD_MARK_start() used on unreconized type");
    
    ipc = IPC_initInputModule(info->ipc_path);
    if (IPC_geterrno(ipc) != IPCE_SUCCESS) {
        LOG_print(LOG_ERROR, "Failed to initilize the IPC socket: (%d) %s",
                IPC_geterrno(ipc), IPC_geterrstr(ipc));
        EXIT(EXIT_ERROR);
    }

    waiter = INW_create();
    if (!INW_addIPC(waiter, ipc)) {
        LOG_print(LOG_ERROR, "MARK IM: failed to add IPC to input waiter");
        EXIT(EXIT_ERROR);
    }
    EXIT_setINW(waiter);

    if (!INW_addSignal(waiter, SIGALRM) ) {
        LOG_print(LOG_ERROR, "MARK IM: failed to add SIGALARM event");
        EXIT(EXIT_ERROR);
    }

    if (!INW_addSignal(waiter, SIGTERM) ) {
        LOG_print(LOG_ERROR, "MARK IM: failed to add SIGTERM event");
        EXIT(EXIT_ERROR);
    }

    if (! IPC_inputmoduleStart(ipc) ) {
        LOG_print(LOG_ERROR, 
            "MARK IM: Failed to communicate with controller: (%d) %s",
            IPC_geterrno(ipc), IPC_geterrstr(ipc));
        EXIT(EXIT_ERROR);
    }
    EXIT_setIM_IPC(ipc);

    LOG_initIM(ipc, info->sysChan);

    /* ready to go */

    /* Prep the message structure */
    makeMarkMsg(&msg_mark);


    /* calculate the first alarm time */
    alrm_sec = time(NULL) % info->mark_interval;
    if (alrm_sec == 0) alrm_sec = info->mark_interval;
    alarm(alrm_sec);
    DPRINT(DBG_IM, "First alarm interval: %d", alrm_sec);
    alrm_sec = info->mark_interval;

    while (TRUE) {
        event = INW_getEvent(waiter);

        if (event == NULL) {
            LOG_print(LOG_ERROR, "Error in INW_getEvent()");
            EXIT(EXIT_ERROR);
        }

        if (event->event_id == INWE_SIG) { 
            if (event->info == SIGALRM) {
                /* Make the timestamp right */
                msg_mark.timestamp = time(NULL);

                /* Set a new alarm */
                DPRINT(DBG_IM, "Setting Alarm for %d sec.", alrm_sec);
                alarm(alrm_sec); 

				DPRINT(DBG_IM, "Sending mark message", NO_VARG);

                /* Send the message */
                if (! IPC_imSendMessage(ipc, info->chans[0], &msg_mark) ) {
                    DPRINT(DBG_IM, "Error Sending message", NO_VARG);
                    LOG_print(LOG_ERROR, 
                        "MARK: error sending message (%d) %s",
                        IPC_geterrno(ipc), IPC_geterrstr(ipc));
                }
            }
            else if (event->info == SIGTERM) {
                EXIT(EXIT_NORMAL);
            }
            else {
                LOG_print(LOG_INFO, "Recieved Unexpected signal %d", 
                        event->info);
            }
        }
        else {
            LOG_print(LOG_INFO, "Unexpected INW event id = %d, info = %d", 
                    event->event_id, event->info);
        }
    }

    /* NOT REACHED */
}

void makeMarkMsg(struct message *msg_mark)
{
    struct timeval  tv;
    struct timezone tz;

    memset(msg_mark, 0, sizeof(struct message));

    msg_mark->sev = CONFFS_INFO;
    msg_mark->fac = CONFFS_MARK;

    IMOD_setLocalhostInfo(msg_mark);

    strlcpy(msg_mark->tag, "syslog-im", MSGTAG_LEN);

    msg_mark->timestamp = 0;
    gettimeofday(&tv, &tz);
    msg_mark->tz = tz.tz_minuteswest * -1;

    msg_mark->msglen = strlen(MARK_MSG_TMPL) + 32;
    msg_mark->msg = emalloc(msg_mark->msglen);
    snprintf(msg_mark->msg, msg_mark->msglen, MARK_MSG_TMPL, getpid());
}
