/**
 * @file imod_udp.c
 *
 * @brief UDP Input
 * @par
 * Provides an interface to reading from a UDP socket
 *
 * @design d2_4
 *
 * $Id: imod_udp.c,v 1.21.2.8 2003/10/22 04:09:03 devink Exp $
 *
 * @author	Matt Small <small@sdsc.edu>
 * @date	Thu May  9 12:16:59 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "strcompat.h"
#include "acconfig.h"
#include "error.h"
#include "log.h"
#include "const.h"

#include "util/bitvec.h"

#include "ipc/ipc.h"
#include "ipc/inputwait.h"
#include "ipc/ipc_err.h"
#include "config/config.h"
#include "message/msg_parse_rfc3164.h"

#include "inputmod.h"

static char const rcsid[] 
    = "$Id: imod_udp.c,v 1.21.2.8 2003/10/22 04:09:03 devink Exp $";
/* Buffer length to recv strings into, add 1 for a trailing \0 */
#define RECVSTR_LEN 1025

void
IMOD_UDP_start(struct IMOD_info *info,IMOD_TYPE type)
{
	INW *waiter;
	IPC *ipc_socket;
	struct INW_eInfo *event;	/* the incoming event */
	struct message *recvd_msg;	/* the outgoing message */
	char *recvd_str=NULL;	/* the raw, received string */
	int i;
	int first_id;
	int chan_id=-1;		/* the channel ID of the recent connection */

	if( IMOD_UDP!=type && IMOD_SYSTEM!=type )
		throw_fatal("IMOD_UDP_start() used on non-recognized type");

	recvd_msg=emalloc(sizeof(struct message));
	recvd_str=emalloc(sizeof(char)*RECVSTR_LEN);	

	DPRINT(DBG_IM,"Starting UDP/System input module", NO_VARG);

	/* start the IPC socket */
	ipc_socket=IPC_initInputModule(info->ipc_path);
	if(IPCE_SUCCESS!=IPC_geterrno(ipc_socket))
	{
        LOG_print(LOG_ERROR, "Failed to initilize the IPC socket: (%d) %s",
                IPC_geterrno(ipc_socket), IPC_geterrstr(ipc_socket));
        EXIT(EXIT_ERROR);
	}

	/* init the input waiter */
	waiter=INW_create();
	if(!INW_addIPC(waiter,ipc_socket))
	{
		LOG_print(LOG_ERROR,"UDP IM: Failed to add IPC to input waiter");
		EXIT(EXIT_ERROR);
	}
    EXIT_setINW(waiter);

    if (! INW_addSignal(waiter, SIGTERM) )
    {
        LOG_print(LOG_ERROR, "UDP IM: Failed to add SIGTERM handler");
        EXIT(EXIT_ERROR);
    }

	first_id=INW_getNextEventID(waiter);
	for(i=0;i<info->fd_len;i++)
	{
		if(!INW_addFD(waiter,info->fd[i],first_id+i))
		{
			LOG_print(LOG_ERROR,"UDP IM: Failed to add FD to input waiter");
			EXIT(EXIT_ERROR);
		}
	}

	/* signal the controller that we're ready */
	if(!IPC_inputmoduleStart(ipc_socket))
	{
		LOG_print(LOG_ERROR,
            "UDP IM: Failed to communicate with controller: (%d) %s",
            IPC_geterrno(ipc_socket), IPC_geterrstr(ipc_socket));
		EXIT(EXIT_ERROR);
	}
    EXIT_setIM_IPC(ipc_socket);

	/* begin real logging */
	LOG_initIM(ipc_socket, info->sysChan);


	DPRINT(DBG_IM,"Starting UDP IMOD main loop, type: %d", type);

	while(1)
	{
		/* get new event */
		event=INW_getEvent(waiter);

		/* check for errors */
		if (0!=IPC_geterrno(ipc_socket))
		{
			throw_fatal(IPC_geterrstr(ipc_socket));
		}
		else if(NULL==event)
		{
			throw_fatal("Input waiter return NULL (controller death)");
		}

		/* IPC event waiting */
		if(INWE_IPC==event->event_id)
		{
			DPRINT(DBG_IM,"Received IPC event %d", event->info);
		}
        /* signal event */
        else if (event->event_id == INWE_SIG) {
            if (event->info == SIGTERM) {
                EXIT(EXIT_NORMAL);
            }
            else {
                DPRINT(DBG_IM, "Received signal %d", event->info);
            }
        }
		/* FD event */
		else
		{
			int recbytes=0;
			struct sockaddr_in peer;
			socklen_t peer_len=sizeof(struct sockaddr_in);

            DPRINT(DBG_IM, "Received event %d, info %d",
                    event->event_id, event->info);

			/* calculate the channel ID; brute force */
			chan_id=-1;
			for(i=0;i<info->fd_len;i++)
			{
				if(event->info==info->fd[i])
				{
					chan_id=info->chans[i];
					break;
				}
			}
			if(-1==chan_id)
			{
				LOG_print(LOG_ERROR,"Error: no channel ID found for FD: %d",
						event->info);
				continue;
			}

			recbytes=recvfrom(event->info,recvd_str,RECVSTR_LEN-1,0,
					(struct sockaddr *)&peer,&peer_len);
			if(-1==recbytes)
			{
                /* Linux 2.2 returns icmp port unreach as an error from
                 * recvfrom() with errno == ECONNREFUSED.  These get
                 * generated if the UDP output module is sending
                 * messages to a host that is not listening, We don't
                 * really care, and it generates _way_ too much log
                 * traffic.  This doesn't happen in 2.4.
                 */
                /**@todo This loop can spin alot due to the icmp
                 * unreach, might consider setting SO_BSDCOMPAT to
                 * avoid.
                 */
                if (errno != ECONNREFUSED && errno != EINTR) {
				    LOG_eprint("imod_udp recv() failed", errno);
                }
				continue;
			}
           

			if(0==recbytes)
				continue;

			recvd_str[recbytes]='\0';

			DPRINT(DBG_IM,"Received message: (%d) \"%s\"", recbytes, recvd_str);

			/* transform data to internal message format */
			if(IMOD_UDP==type)
			{
				if(MPOLD_parsein(recvd_msg,recvd_str,recbytes+1,FALSE, TRUE)<0)
				{
					DPRINT(DBG_IM,"Received invalid message: \"%s\"",recvd_str);
				}
			}
			else if(IMOD_SYSTEM==type)
			{
				if(MPOLD_parsein(recvd_msg,recvd_str,recbytes+1,TRUE, FALSE)<0)
				{
					DPRINT(DBG_IM,"Received invalid message: \"%s\"",recvd_str);
				}
			}

            /* Set the DeviceIP unless something has set it already,
             * skip for local messages, it's dealt with elsewhere
             */
            if (IMOD_UDP == type && NA_isEmpty(recvd_msg->dip) ) {
                recvd_msg->dip = NA_newINet4(peer.sin_addr);
            }

			if(recvd_msg->host == NULL && type == IMOD_UDP)
			{
                struct hostent  *peer_host;
                char            *p;

				DPRINT(DBG_IM,
                    "Adding peer address for UDP with no other hostname",
                    NO_VARG);

                peer_host=gethostbyaddr(
                        (char *)&(peer.sin_addr.s_addr),
                        sizeof(peer.sin_addr.s_addr),AF_INET);

                if (peer_host != NULL) {
                    /* Set the FQDN */
                    recvd_msg->dfqdnlen=strlen(peer_host->h_name)+1;
                    recvd_msg->dfqdn=emalloc(sizeof(char)*recvd_msg->dfqdnlen);
                    strlcpy(recvd_msg->dfqdn,peer_host->h_name, 
                            recvd_msg->dfqdnlen);
                    DPRINT(DBG_IM, "Set dfqdn to '%s'", recvd_msg->dfqdn);

                    /* Set the HOSTNAME */
                    p = recvd_msg->dfqdn;
                    recvd_msg->hostlen = 0;
                    while (*p != '\0' && *p != '.') {
                        p++;
                        recvd_msg->hostlen++;
                    }
                    recvd_msg->hostlen++;   /* for terminating NUL */
                    recvd_msg->host = emalloc(sizeof(char)*recvd_msg->hostlen);
                    strlcpy(recvd_msg->host, recvd_msg->dfqdn, 
                            recvd_msg->hostlen);
                    DPRINT(DBG_IM, "Set host to '%s'", recvd_msg->host);
                }
                else {
                    /* Use the peer IP as the host */
                    char        *addrStr = inet_ntoa(peer.sin_addr);

                    recvd_msg->hostlen = strlen(addrStr);
                    recvd_msg->host = emalloc(recvd_msg->hostlen);
                    strlcpy(recvd_msg->host, addrStr, recvd_msg->hostlen);
                }
            }

            /* Finish filling in info about the local host */
            if(IMOD_SYSTEM==type) {
                IMOD_setLocalhostInfo(recvd_msg);

                /* Set a timezone if none was recieved */
                MSG_setTimeZone(recvd_msg);
            }

			/* send the message back to the controller */
			if(!IPC_imSendMessage(ipc_socket,chan_id,recvd_msg))
			{
				throw_fatal("Failed to send message to the controller (controller death)");
			}

			/* Since we use recvd_msg each time, we don't need to free it; but
			 * the fields which are created each time must be free()ed */
            MSG_free(recvd_msg);
		}
	}	/* end while */

	/* clean up; this is presumably after a successful run, so we need
	 * to shutdown cleanly */
	free(info);
	free(recvd_msg);

    EXIT(EXIT_NORMAL);
}
