/**
 * @file inputmod.h
 *
 * @brief This provides general input module structures and functions
 *
 * $Id: inputmod.h,v 1.7.2.6 2003/04/30 20:03:16 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue May 14 11:43:09 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef IMOD_H
#define IMOD_H

#include "util/acl.h"

/**@defgroup imod Input module
 * @{
 */

/** Provides names for the types of possible input modules */
typedef enum {
	IMOD_BEEP,		/**< BEEP Input Module */
	IMOD_UDP,		/**< UDP Input Module */
	IMOD_SYSTEM,	/**< System Input Module */
    IMOD_MARK,      /**< MARK message generator */
	IMOD_TYPECOUNT	/**< Count of input types */
} IMOD_TYPE;

/** Human readable IMOD type names */
extern char const *IMOD_typeNames[];

/**
 * Provides a structure to hold file descriptors and any other generic 
 * information need by the input modules
 */
struct IMOD_info
{
	int fd_len;		/**< Length of list of file descriptors */
	int *fd;		/**< List of file descriptors; must be free()ed */
    int *chans;     /**< Channels corresponding to fds. allocated with \c fds*/

    uint32_t sysChan;   /**< System channel ID for internal logging */ 

    int mark_interval;  /**< Seconds to the next mark message (MARK IM only) */

    ACL **path_acls;    /**< path acls */
    ACL **psec_acls;    /**< path security acls */
    ACL **hsrc_acls;    /**< source host acls */
    ACL **hcon_acls;    /**< connection acls */

	char *ipc_path;	/**< The path to the IPC descriptor */
};

/** 
 * Set the message::host and message::dfqdn fields to the localhost.
 * 
 * @param msg The message to set the dfqdn in
 */
void IMOD_setLocalhostInfo(struct message *msg);

/*@}*/


/**@defgroup imod_udp UDP Input Module
 * @ingroup imod
 * @{
 */

/** Start a new UDP Input Module; this is expected to be forked off into
 *  its own process.
 *
 *  @param info Contains file descriptors and IPC information
 *  @param type Contains the type of the input module being started (must be
 *  			either IMOD_UDP or IMOD_SYSTEM)
 */
void IMOD_UDP_start(struct IMOD_info *info,IMOD_TYPE type);

/**
 * Start a new MARK input module.  This is expected to be forked off
 * into it's own process.
 *
 * @param info Contains information needed by the IM.
 * @param type The type of the input module (should always be IMOD_MARK).
 */
void IMOD_MARK_start(struct IMOD_info *info, IMOD_TYPE type);

/** Start a new BEEP input module.  This is expected to be forked off into its
 * own process.
 *
 * @note FDs passed in in info should be O_NONBLOCK to rpevent select() from
 * 		returning an FD that isn't ready, and having accept() block on that
 *
 * @param info Contains file descriptors and IPC information
 * @param type The type of the input module (should always be OMOD_BEEP).
 */
void IMOD_BEEP_start(struct IMOD_info *info, IMOD_TYPE type);

/*@}*/

#endif /* IMOD_H */
