/**
 * @file buffer.c
 *
 * @brief Implement the buffer manipulation functions for the IPC
 * module.
 *
 * $Id: buffer.c,v 1.12.2.5 2003/04/30 20:03:17 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Feb 28 17:31:09 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <sys/time.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <errno.h>

/* DEBUG */
#include <ctype.h>

#include "util/acl.h"
#include "util/netaddr.h"

#include "buffer.h"
#include "ipc_types.h"

#include "types.h"
#include "strcompat.h"
#include "error.h"
#include "debug.h"
#include "dmalloc_sup.h"

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif


static char const rcsid[] = 
    "$Id: buffer.c,v 1.12.2.5 2003/04/30 20:03:17 devink Exp $";

/** 
 * Check the space available in buf 
 * @param buf The buffer to check (IPC_buffer *)
 * @return The number of bytes available in the buffer.
 *
 * @pre buf != NULL
 */
static inline int availSpace(struct IPC_buffer const *buf);

/** 
 * Check that an address is within the bounds of a buffer
 *
 * @param buf The buffer to check against
 * @param addr The address to check
 * @return TRUE if \c addr is in range, false otherwise.
 *
 * @pre buf != NULL
 */
static inline bool isValid(struct IPC_buffer *buf, IPC_bufMark addr);

void
BUF_alloc_priv(DMS_PARAMS(struct IPC_buffer *buf))
{
    assert(buf != NULL);
    assert(buf->size > 0);

    buf->data = DMSmalloc(buf->size);
    buf->pos = buf->data;
}

void
BUF_free(struct IPC_buffer *buf) 
{
    if (buf == NULL) return;

    /* free data */
    /* only free the data */
    free(buf->data);

    /* zero fields */
    buf->data = buf->pos = NULL;
    buf->size = 0;
}

void
BUF_init(struct IPC_buffer *buf)
{
    buf->data = buf->pos = NULL;
    buf->size = 0;
}

bool
BUF_setByte(struct IPC_buffer *buf, uint32_t val)
{
    assert(buf != NULL);
    assert(buf->data != NULL);
    assert(buf->pos >= buf->data);

    /* check avail space */
    if (availSpace(buf) < BUF_BYTESZ) {
        return FALSE;
    }
    *(buf->pos) = val & 0xFF;
    buf->pos += BUF_BYTESZ;

    assert(buf->pos <= buf->data + buf->size);

    return TRUE;
}

bool
BUF_setInteger(struct IPC_buffer *buf, uint32_t val)
{
    uint32_t    nval;
    int         i;

    assert(buf != NULL);
    assert(buf->data != NULL);
    assert(buf->pos >= buf->data);

    if (availSpace(buf) < BUF_INTSZ) {
        return FALSE;
    }

    /* this tries to keep things in network byte order */
    /* this assumes there are 8 bits in a byte */
    nval = htonl(val);
    for (i = 0; i < BUF_INTSZ; i++) {
        /* the SLR puts the bits we want in 0-8, & zeros other bits */
        buf->pos[i] = (nval >> ((BUF_INTSZ - 1 - i)*8)) & 0xFF;
    }

    buf->pos += BUF_INTSZ;  /* leave pos on writable byte */

    assert(buf->pos <= buf->data + buf->size);

    return TRUE;
}

bool
BUF_setTimestamp(struct IPC_buffer *buf, time_t val)
{
    time_t      nval;
    int         i;

    assert(buf != NULL);
    assert(buf->data != NULL);
    assert(buf->pos >= buf->data);

    if (availSpace(buf) < BUF_TIMESZ) {
        return FALSE;
    }

    nval = htonl(val);
    for (i = 0; i < BUF_TIMESZ; i++) {
        buf->pos[i] = (nval >> (BUF_TIMESZ - 1 - i)*8) & 0xFF;
    }

    buf->pos += BUF_TIMESZ;

    assert(buf->pos <= buf->data + buf->size);

    return TRUE;
}

bool
BUF_setIntegerAt(struct IPC_buffer *buf, IPC_bufMark pos, uint32_t val)
{
    IPC_bufMark  origPos;

    assert(buf != NULL);
    assert(buf->data != NULL);
    assert(buf->pos >= buf->data);

    if (!isValid(buf, pos) ) return FALSE;

    origPos = BUF_markPos(buf);

    BUF_resetToPos(buf, pos);


    if (availSpace(buf) < BUF_INTSZ) goto error; 

    if (! BUF_setInteger(buf, val) ) goto error;

    /* make sure we weren't so close to the end of the buffer that we
     * actually went past the original position. dispite the fact that
     * this means that someone forgot to allocate space for this
     * integer.
     */
    if (origPos > buf->pos) {
        BUF_resetToPos(buf, origPos);
    }

    return TRUE;

    /* Error handler */
    error:
        BUF_resetToPos(buf, origPos);
        return FALSE;
}

bool
BUF_setFixString(struct IPC_buffer *buf, char const *str, size_t len)
{
    size_t      wrlen;

    assert(buf != NULL);
    assert(buf->data != NULL);
    assert(buf->pos >= buf->data);

    /*
    DPRINT(DBG_BUF, "buf set: fixstr: start, str = \"%s\", len = %d, pos = %p", 
            __FILE__, __LINE__, str, len, buf->pos);
    */

    /* fix up str if it's NULL */
    if (str == NULL) str = "";

    if (availSpace(buf) < len) {
        return FALSE;
    }

    /* we verified above that there is at least len+1 bytes avaialble in
     * buf, so don't copy more than that.
     */
    wrlen = strlcpy(buf->pos, str, len);
    if (wrlen > len) {
        /* truncation occured ... do we care? */
        return FALSE;
    }

    /* strlcpy() doesn't count the NIL in its return */
    wrlen++;

    /*
    DPRINT(DBG_BUF, "buf set: fixstr: wrote string \"%s\", len = %d, pos = %p", 
            __FILE__, __LINE__, buf->pos, wrlen, buf->pos);
    */

    buf->pos += wrlen;    

    assert(buf->pos <= buf->data + buf->size);

    if (wrlen < len) {
        /* str was short, fill in rest with NIL */
        for (wrlen += 2; wrlen <= len; wrlen++)  {
            *(buf->pos) = '\0';
            buf->pos++;
        }
        buf->pos++;     /* needs to be one past the end of the buffer */
    }

    /*
    DPRINT(DBG_BUF, "buf set: fixstr: end, pos = %p", __FILE__, __LINE__,
            buf->pos);
    */

    assert(buf->pos <= buf->data + buf->size);
    assert(buf->pos >= buf->data);

    return TRUE;
}

int
BUF_setVarString(struct IPC_buffer *buf, char const *val, size_t len)
{
    IPC_bufMark  startPos;

    /* save the starting postion */
    startPos = BUF_markPos(buf);

    /* fix up str if it's null */
    if (val == NULL) val = "";

    /* fix up len if it's zero */
    if (len <= 0) {
        len = strlen(val)+1;
    }
    
    /* write the length */
    if (! BUF_setInteger(buf, len) ) goto error;

    /* write the string */
    if (! BUF_setFixString(buf, val, len) ) goto error;


    return len;

    /* Error handler */
    error:
        BUF_resetToPos(buf, startPos);
        return -1;
}

bool
BUF_setEmptyString(struct IPC_buffer *buf)
{
    /* asserts handled by setVarString() */
    return BUF_setVarString(buf, "", 0);
}

bool
BUF_getByte(struct IPC_buffer *buf, uint8_t *out)
{
    assert(buf != NULL);
    assert(buf->data != NULL);

    if (availSpace(buf) < BUF_BYTESZ) 
        return FALSE;

    *out = *(uint8_t*)buf->pos;

    buf->pos += BUF_BYTESZ;

    return TRUE;
}

bool
BUF_getInteger(struct IPC_buffer *buf, uint32_t *out)
{
    int     i;

    assert(buf != NULL);
    assert(buf->data != NULL);

    if (availSpace(buf) < BUF_INTSZ) 
        return FALSE;

    /*
    DPRINT(DBG_BUF, "buf get: int: bytes = {%x, %x, %x, %x}",
            __FILE__, __LINE__, 
            buf->pos[0], buf->pos[1], buf->pos[2], buf->pos[3]);
    */

    *out = 0; 
    for (i = 0; i < BUF_INTSZ; i++) {
        *out |= (uint32_t)(buf->pos[i] & 0xFF) << (BUF_INTSZ - i - 1)*8;
    }

    /*
    DPRINT(DBG_BUF, "buf get: int: val = 0x%x", __FILE__, __LINE__, *out);
    */
    *out = ntohl(*out);

    buf->pos += BUF_INTSZ;

    return TRUE;
}

bool
BUF_getTimestamp(struct IPC_buffer *buf, time_t *out)
{
    int     i;

    assert(buf != NULL);
    assert(buf->data != NULL);

    if (availSpace(buf) < BUF_TIMESZ) 
        return FALSE;

    *out = 0;
    for (i = 0; i < BUF_TIMESZ; i++) {
        *out |= (uint32_t)(buf->pos[i] & 0xFF) << (BUF_TIMESZ - i - 1)*8;
    }
    /*
    DPRINT(DBG_BUF, "buf get: ts: out = 0x%x, ntohl(out) = 0x%x",
            __FILE__, __LINE__, *out, ntohl(*out));
    */
    *out = ntohl(*out);


    buf->pos += BUF_TIMESZ;

    return TRUE;
}

bool
BUF_getVarString(struct IPC_buffer *buf,size_t *oLen,char **oStr,size_t ossz)
{
    uint32_t    len;
    bool        rtn;
    char        empty[1];   /* buffer to put an empty string in */

    /* preconditions checked in getInteger() */
    rtn = BUF_getInteger(buf, &len);
    if (!rtn) {
        return FALSE;
    }


    if (len == 1 && ossz == 0) {
        /* Special case for a malloc()ed string which is actually empty.*/
        *oStr = NULL;
        rtn = BUF_getString(buf, empty, len);
        len = 0; 
    }
    else {  
        /* len > 1 */

        /* Malloc space if we need to */
        if (ossz == 0) {
            *oStr = emalloc(len);
        }

        rtn = BUF_getString(buf, *oStr, len);
    }

    /* Check result of BUF_getString() */
    if (!rtn) {
        if (ossz == 0) free(*oStr);
        return FALSE;
    }

    /* Set the length return */
    if (oLen != NULL) *oLen = len;

    return TRUE;
}

bool
BUF_getString(struct IPC_buffer *buf, char *out, size_t len)
{
    size_t  wrlen;

    if (availSpace(buf) < len)
        return FALSE;

    wrlen = strlcpy(out, buf->pos, len);

    wrlen++;    /* strlcpy doesn't include NIL in return */

    if (wrlen > len) {
        /* trucation occurred */
        return FALSE;
    }

    buf->pos += len; 

    return TRUE;
}

IPC_bufMark
BUF_markPos(struct IPC_buffer const *buf)
{
    assert(buf != NULL);

    return buf->pos;
}

bool
BUF_resetToPos(struct IPC_buffer *buf, IPC_bufMark pos)
{
    assert(buf != NULL);
    assert(buf->data != NULL);

    if (! isValid(buf, pos) ) return FALSE;

    buf->pos = pos;

    /* post assert: not actually done b/c it would be silly
     * assert(buf->pos == pos);
     */
    return TRUE;
}

IPC_readState
BUF_read(struct IPC_buffer *buf, int sock, bool fheader)
{
    int             bytesToRead = 0;
    int             rlen;
    IPC_readState   rdState;
    char            *rpos = NULL;
#if USE_IPC_FRAME_HEADER
    char            header[IPC_FHEAD_LEN];
#endif

    assert(buf->pos != NULL);
    assert(buf->data != NULL);
    assert(buf->pos == buf->data);
    assert(sock >= 0);

#if USE_IPC_FRAME_HEADER
    memset(header, 0, sizeof(header));
    if (fheader) {
        rpos = &(header[0]);
        bytesToRead = IPC_FHEAD_LEN;
    }
    else 
#endif
    {
        bytesToRead = buf->size;
        rpos = buf->data;
    }

    DPRINT(DBG_BUF, "BUF_read: start: read %d bytes (fheader = %d)", 
            bytesToRead, fheader);

    /**
     * @todo Add a timeout to this loop, otherwise it could loop forever
     *       if the expected data never arrives.
     */
    rdState = IPCRS_SUCCESS;
    while (bytesToRead > 0) { 
        rlen = read(sock, rpos, bytesToRead);
        DPRINT(DBG_BUF, "BUF_read: got %d bytes in read", rlen);
        if (rlen < 0) {
            DPRINT(DBG_BUF, "recv: read(): errno = %d", errno);

            if (errno == EAGAIN) {
                rdState = IPCRS_AGAIN;
            }
            else {
                rdState = IPCRS_ERROR;
            }
            break;
        }
        else if (rlen == 0) {
            /* discard the packet ... the other end is probably dead */
            DPRINT(DBG_BUF, "recv: read(): EOF on sock", NO_VARG);
            rdState = IPCRS_EOF;
            break;
        }

#if USE_IPC_FRAME_HEADER
        if (fheader) {
            /* Try to find the frame header first */
            DPRINT(DBG_IPC, "recv: fheader = %.2x %.2x %.2x %.2x",
                    (int)header[0] & 0xFF, (int)header[1] & 0xFF, 
                    (int)header[2] & 0xFF, (int)header[3] & 0xFF);

            bytesToRead -= rlen;

            if (strncmp(header, IPC_FRAME_HEADER, IPC_FHEAD_LEN) == 0) {
                /* found it */
                bytesToRead = buf->size;
                rpos = buf->data;
                fheader = FALSE;
                DPRINT(DBG_BUF, "BUF_read: got fheader, reading %d bytes",
                        bytesToRead);
            }
            else if (bytesToRead > 0) {
                /* still reading a full header */
                rpos += rlen;
            }
            else {
                /* still looking, shift and read one byte */
                int j;
                for (j = 0; j < IPC_FHEAD_LEN - 1; j++) {
                    header[j] = header[j+1];
                }
                bytesToRead = 1;
                rpos = &(header[IPC_FHEAD_LEN - 1]);
            }

        }
        else 
#endif 
        {
            bytesToRead -= rlen;
            rpos += rlen;
        }
    }
    DPRINT(DBG_BUF, "BUF_read: finish read with %d bytes left", bytesToRead);

    /* Dump the buffer */
    DPRINT_BUF(DBG_BUF, "BUF_read", buf, -1);

    return rdState;
}

bool
BUF_send(struct IPC_buffer *buf, int sock, int len, bool fheader)
{
    int             slen;
    int             bytesToWrite;
    bool            rtn;
    fd_set          wfds;
    char const      *wpos;

    assert(buf != NULL);
    assert(buf->data != NULL);
    assert(len <= buf->size);

    /* never try to write more than we have */
    if (len > buf->size)
        return FALSE;

    DPRINT(DBG_BUF, "BUF_send: len = %d; fheader = %d", len, fheader); 

#if USE_IPC_FRAME_HEADER
    if (fheader) {
        wpos = IPC_FRAME_HEADER;
        bytesToWrite = IPC_FHEAD_LEN;
    }
    else 
#endif
    {
        wpos = buf->data;
        bytesToWrite = len;
    }

    /* Dump the sending buffer */
    DPRINT_BUF(DBG_BUF, "BUF_send", buf, len);

    rtn = TRUE;
    while (bytesToWrite > 0) {
        slen = send(sock, wpos, bytesToWrite, 0);
        DPRINT(DBG_BUF, "BUF_send: wrote %d bytes in send", slen);
        if (slen < 0) {
            if (errno == EAGAIN) {
                DPRINT(DBG_BUF, 
                    "BUF_send: EAGAIN, waiting for socket", NO_VARG);
        
                /* Wait for the socket to be ready */
                FD_ZERO(&wfds);
                FD_SET(sock, &wfds);

                if (select(sock + 1, NULL, &wfds, NULL, NULL) < 0) {
                    rtn = FALSE;
                    break;
                }
                /* We are just going to assume that we can write now, as
                 * only one FD was selected on.
                 */
            }
            else {
                rtn = FALSE;
                break;
            }
        }
        else {
            bytesToWrite -= slen;
            wpos += slen;
        }
        
#if USE_IPC_FRAME_HEADER
        if (fheader && bytesToWrite <= 0) {
            wpos = buf->data;
            bytesToWrite = len;
            fheader = FALSE;
        }
#endif /* USE_IPC_FRAME_HEADER */

    }

    DPRINT(DBG_BUF, "BUF_send: done (%d), bytesToWrite = %d, errno = %d", 
            rtn, bytesToWrite, errno);

    return rtn;
}

static inline int 
availSpace(struct IPC_buffer const *buf) 
{
    ptrdiff_t   used;

    assert(buf != NULL);

    used = buf->pos - buf->data;

    return buf->size - used;
}

static inline bool 
isValid(struct IPC_buffer *buf, IPC_bufMark addr)
{
    assert(buf != NULL);

    return buf->pos != NULL 
           && addr >= buf->data 
           && addr < (buf->data + buf->size);
}
