/**
 * @file buffer.h
 *
 * @brief This provides a set of functions that can be used for low level
 * IPC packet assembly.
 * @par
 * These functions all operate on a IPC_buffer structure, which holds the 
 * buffer data and size, and acts an an iterator by holding the current
 * position in the buffer.  The functions insert the next field (as a
 * particular type) into the current location in the packet. These
 * functions are geared towards the IPC protocol used by SDSCSyslog, in
 * that they make assuptions about data types based on the IPC Protocol
 * spec \ref i4_1
 *
 * @todo Consider using data buffers from preallocated pools.
 * @see \ref ipc_proto
 *
 *
 * $Id: buffer.h,v 1.6.2.5 2003/04/30 20:03:17 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Feb 26 16:49:16 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IPC_BUFFER_H
#define IPC_BUFFER_H

#ifdef HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <stddef.h>
#include <time.h>

#include "ipc_types.h"

#include "types.h"
#include "dmalloc_sup.h"

/**@defgroup ipc_buffer IPC Buffer Manipulation Functions
 * @ingroup ipc
 */

/** 
 * Contain all the data needed for buffer manipulations. 
 * @ingroup ipc_buffer
 */
struct IPC_buffer {
    size_t  size;
    char    *data;
    char    *pos;
};

/** 
 * A Position marker in an IPCBuffer 
 * @ingroup ipc_buffer
 */
typedef char * IPC_bufMark;

/**@name Buffer Type Lengths.
 * These are the same as the types in the IPC spec. They only cover the
 * constant size types.
 * @see \ref i4_1
 * @ingroup ipc_buffer
 * @{
 */
#define BUF_BYTESZ      1               /**< BYTE */
#define BUF_INTSZ       4               /**< INTEGER */
#define BUF_TIMESZ      sizeof(time_t)  /**< TIMESTAMP */
#define BUF_EMPTYSTRSZ  5               /**< The size of an empty string */
#define BUF_MAXINETSZ   NA_MAX_BUFLEN   /**< The max size of an INET type */
/*@}*/

/**@name Functions
 * @ingroup ipc_buffer
 * @{
 */

/** 
 * Initilize a new buffer.
 * @param buf The buffer to work on.
 *
 * @pre buf != NULL
 * @post buf->data == buf->pos == NULL
 * @post buf->size == 0
 */
inline void BUF_init(struct IPC_buffer *buf); 

/** 
 * Allocate the data space for a buffer.  
 * This assumes that you have set buf->size correctly.
 *
 * @param buf The buf to allocate space in.
 *
 * @pre buf != NULL
 * @pre buf->size > 0
 */
#define BUF_alloc(buf)  BUF_alloc_priv(DMS_ARGS(__FILE__, __LINE__, buf))

/** private function to do the actual allocation */
void BUF_alloc_priv(DMS_PARAMS(struct IPC_buffer *buf));

/** 
 * Free a buffer.
 *
 * @param buf The buffer to free.
 */
void BUF_free(struct IPC_buffer *buf);

/** 
 * Insert a BYTE in the buffer.  
 *
 * @param buf the buffer to work on
 * @param val The value to insert.  Only uses the low byte.
 * @return TRUE on success, FALSE otherwise (probably an overflow)
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
bool BUF_setByte(struct IPC_buffer *buf, uint32_t val);

/** 
 * Insert an INTEGER in the buffer.
 *
 * @param buf the buffer to work on.
 * @param val The value to insert.  Only uses the low 4 bytes.
 * @return TRUE on success, FALSE otherwise (probably an overflow)
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
bool BUF_setInteger(struct IPC_buffer *buf, uint32_t val);

/**
 * Insert a TIMESTAMP into the buffer.
 *
 * @param buf The buffer to work on.
 * @param val The value to insert
 * @return TRUE on success, FALSE otherwise (probably an overflow)
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
bool BUF_setTimestamp(struct IPC_buffer *buf, time_t val);

/** 
 * Insert an INTEGER at a specific (saved) point in the buffer.  
 * This is only available for integers because they are the most commonly
 * used type for lengths (which may not be known right away).  Further
 * it is upto the programmer to remember to save room for that INTEGER.
 * @par
 * Example:
 * @verbatim
 IPC_bufMark pos = BUF_markPos(buf);
 BUF_setInteger(buf, 0);    // reserve space for integer 
 // do some stuff, and figure out data_length
 
 // set the integer to the real vaule.
 BUF_setIntegerAt(buf, pos, data_length);   
 @endverbatim
 *
 * @param buf The buffer to work on.
 * @param pos The position to insert at.
 * @param val The vaule to instert.
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
bool BUF_setIntegerAt(struct IPC_buffer *buf, IPC_bufMark pos, uint32_t val);

/** 
 * Insert a fixed length string in the buffer.  
 * Because the string is fixed length, there is no need for a length 
 * parameter.  If the string has a length paramer use setVarString().  
 * Extra bytes will be set to NIL. Further, if <code>strlen(str) > len</code> 
 * then FALSE will be returned.
 * Note that \c len should include the terminating NIL
 *
 * @param buf The buffer to work on.
 * @param str The string to insert.
 * @param len The length of the string.
 * @return TRUE on success, FALSE otherwise (probably an overflow)
 *  
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
bool BUF_setFixString(struct IPC_buffer *buf, char const *str, size_t len);

/** 
 * Insert a variable length string in the buffer.  This will also
 * insert the length parameter for the string as an INTEGER directly
 * before the actual string data.  Also you can pass <code>len = 0</code>
 * and this function will figure out the length on it's own. Also note
 * that \c len is the length including the terminating NIL, and \c str MUST
 * HAVE a terminating NIL.
 *
 * @param buf The buffer to work on.
 * @param str The string to insert.
 * @param len The length of the string.
 * @return the number of bytes (including NIL) written, otherwise < 0
 *  
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
int BUF_setVarString(struct IPC_buffer *buf, char const *str, size_t len);

/** 
 * Insert an empty string in the buffer.  
 * This will insert an empty variable length string in the buffer.  
 *
 * @param buf The buffer to work on.
 * @return TRUE on success, FALSE otherwise (probably an overflow)
 *  
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
bool BUF_setEmptyString(struct IPC_buffer *buf);

/** 
 * Get a BYTE from the buffer.
 * @param buf The buffer to work on.
 * @param valout The vairaible to store the data from that byte in.
 * @return TRUE on success, FALSE otherwise
 *  
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post \c buf->pos is no more than one past the end of \c data.
 */
bool BUF_getByte(struct IPC_buffer *buf, uint8_t *valout);

/** 
 * Get an INTEGER from the buffer.
 *
 * @param buf The buffer to work on.
 * @param valout The place to store the retrieved value
 * @return TRUE on success, FALSE otherwise
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 */
bool BUF_getInteger(struct IPC_buffer *buf, uint32_t *valout);

/** 
 * Get a TIMESTAMP from the buffer.
 *
 * @param buf The buffer to work on.
 * @param valout The place to store the retrieved value
 * @return TRUE on success, FALSE otherwise
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 */
bool BUF_getTimestamp(struct IPC_buffer *buf, time_t *valout);

/** 
 * Get the a variable length string from the buffer.  This will get both
 * the length and the string from the buffer.  It will malloc space in
 * \c oStr if \c ossz==0, otherwise it will assume that \c *oStr has 
 * \c ossz bytes in it.  If \c *oLen>ossz then an error will be
 * returned.  If <code> *oLen==0 && ossz==0 </code> then \c *oStr==NULL
 * on return. Otherwise if <code> *oLen==0 && ossz>0 </code> then 
 * \c * **oStr==0 on return.
 *
 * @param buf The buffer to work on
 * @param oLen The length of the string (may be null to ignore) (output)
 * @param oStr The variable length string (output)
 * @param ossz Size of \c *oStr or 0 if \c oStr should be malloc()ed.
 * @retrun TRUE on success.
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 */
bool BUF_getVarString(struct IPC_buffer *buf, size_t *oLen, char **oStr, 
                      size_t ossz);

/** 
 * Get a string from the buffer.  
 * Note that \c len is dual use, it is both the assumed length of 
 * the string to retrieve, and the length (or < the length) of the buffer 
 * to store the retrieved string in.  Also \c len includes the
 * terminating NIL. 
 *
 * @param buf The buffer to work on.
 * @param out The place to store the retrieved value
 * @param len The length of the output buffer
 *
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 */
bool BUF_getString(struct IPC_buffer *buf, char *out, size_t len);

/** 
 * Mark the current position in a buffer.  
 * This generates an IPC_bufMark that references the current position in 
 * the buffer and can be returned to.
 *
 * @param buf The buffer to mark
 * @return A marker for the current position.
 *
 * @pre buf != NULL
 * @post buf fields are unmodified
 */
IPC_bufMark BUF_markPos(struct IPC_buffer const *buf);

/** 
 * Return to the marked position in the buffer.  
 *
 * @param buf The buffer to work on.
 * @param pos The position to return to.
 *
 * @return TRUE on success, FALSE otherwise (probably pos out of bounds)
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @post buf->post == marked position (on success)
 */
bool BUF_resetToPos(struct IPC_buffer *buf, IPC_bufMark pos);

/**
 * Read data off of the given fd into the given buffer.
 * On error, errno will remain set correctly.  When this returns \c buf
 * will be set to the begining of the data.
 *
 * @param buf The buffer to read into
 * @param sock The file descriptor (socket) to read from.
 * @param fheader TRUE when we should look for a frame header first.
 * @return an IPC_readState value
 *
 * @pre buf->data != NULL
 * @pre buf->pos != NULL
 * @pre buf->pos == buf->data
 * @pre sock >= 0
 * @post buf->pos == buf->data
 *
 * @todo add BUF_send() function to buffer sub-module
 */
IPC_readState BUF_read(struct IPC_buffer *buf, int sock, bool fheader);

/**
 * Send data over a socket.  On error, errno will remain set correctly.
 * This will block until data is written.  The data sent will always
 * start at the begining of the buffer. The current position in the
 * buffer will be one past the last byte sent.
 * 
 *
 * @param buf The buffer to send out
 * @param sock The socket to send on.
 * @param len The number of bytes to send.
 * @param fheader TRUE when we should sent a frame header.
 * @return TRUE on sucess, FALSE otherwise.
 *
 * @pre buf != NULL
 * @pre buf->data != NULL
 * @pre buf->size >= len
 *
 * @post buf->pos will be set to one past the last byte sent.
 */
bool BUF_send(struct IPC_buffer *buf, int sock, int len, bool fheader);

/*@}*/

#endif /* IPC_BUFFER_H */
