/**
 * @file inputwait.c
 *
 * @brief A library that handles blocking for input.
 * @par
 * These functions abstract blocking until input arrives.
 * @note This is closly tied to the IPC mechanisms.
 * @ingroup inwait
 *
 * $Id: inputwait.c,v 1.15.2.7 2003/04/30 20:03:18 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue May  7 13:04:23 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>

#include "strcompat.h"
#include "error.h"

#include "util/list.h"
#include "util/tsqueue.h"

#include "ipc.h"
#include "ipcll.h"
#include "inputwait.h"

static char const rcsid[] 
    = "$Id: inputwait.c,v 1.15.2.7 2003/04/30 20:03:18 devink Exp $";

/** 
 * Select on all the FDs set in the INW handle (passed in via \c data)
 * and all the FDs passed in via \c rfds.  This also checks for pending
 * signals.
 *
 * @param sockMax The maximum valued FD in \c rfds.
 * @param rfds A Set of additional FDs to wait on.
 * @param data The INW handle to work on.
 * @return The number of ready FDs in the original passed in \c rfds
 * @post \c data->events will have a list of all our FDs that are ready. 
 *
 * @pre data != NULL
 * @pre rfds != NULL
 */
static int INWSelect(int sockMax, fd_set *rfds, void *data);

/**
 * Check for registered signals, the events queue in handle, and the IPC 
 * queues if configured to wait on IPC events.
 *
 * @param handle The INW handle to work on.
 * @return An event if one is present, or NULL if not.
 *
 * @pre handle != NULL
 */
static struct INW_eInfo * checkQueues(INW *handle);


/** 
 * Maximum number of signals on a system. More of a bounds check for a
 * sigset_t object really. 
 */
#define MAX_SIGNALS     (sizeof(sigset_t) * 8)

/** 
 * Track raised signals.
 */
static sigset_t raisedSignals;

/**
 * Track registered signals.
 */
static sigset_t regSignals;

/**
 * Move raised signals into the events queue.  Only signals which are
 * registered to the given INW object will be queueed.
 *
 * @param handle The INW handle to work with.
 * @return TRUE of any signals were queued, FALSE otherwise.
 *
 * @pre handle != NULL
 */
static bool queueSignals(INW *handle);

/**
 * The generic input event signal handler 
 * @param signo The signal number this is handling.
 */
void INW_sighandler(int signo);

void
INW_initialize(void)
{
    /* make sure signal tracking is zeroed */
    sigemptyset(&regSignals);
    sigemptyset(&raisedSignals);


    /* mark the uncatchable signals as valid so no one tries to register
     * them.
     */
    sigaddset(&regSignals, SIGKILL);
    sigaddset(&regSignals, SIGSTOP);
}

void
INW_shutdown(void)
{
    /* Nothing to do */
}

INW *
INW_create(void)
{
	INW *ret;

	ret=emalloc(sizeof(INW));

	/* create the list */
	ret->fds=LL_create();
    ret->sigs = LL_create();
    ret->ipc = NULL;
    ret->events = LL_create();
	ret->queue = NULL;
	ret->next_evtid = INWE_FDSTART;
    ret->ipcEInfo.event_id = INWE_IPC;
    ret->qEInfo.event_id = INWE_TSQ;
    ret->noFDWait = FALSE;

	return ret;
}

void
INW_destroy(INW **handle)
{
	/* destroy the list */
    INW_clearFDs(*handle);
    INW_clearSignals(*handle);

    LL_destroy(&((*handle)->sigs));
	LL_destroy(&((*handle)->fds));

    LL_destroy(&((*handle)->events));

	free(*handle);
	*handle=NULL;
}

void
INW_clearFDs(INW *handle)
{
	if(handle->fds != NULL)
	{
		/* go through every node, destroy the element, then destroy
		 * the node */
		while(LL_size(handle->fds) > 0)
		{
			free(LL_first(handle->fds));
			LL_rmFirst(handle->fds);
		}
	}
}

void
INW_clearSignals(INW *handle)
{
    struct INW_eInfo    *info;
    struct sigaction    act;

    act.sa_handler = SIG_DFL;

    if (handle->sigs != NULL) {
        while (LL_size(handle->sigs) > 0) {
            info = LL_first(handle->sigs);
            
            sigaction(info->info, &act, NULL);
            sigdelset(&regSignals, info->info);

            free(info);
            LL_rmFirst(handle->sigs);
        }
    }
}

int
INW_getNextEventID(INW *handle)
{
	return (handle->next_evtid)++;
}

bool
INW_addIPC(INW *handle,IPC *ipc)
{
    assert(handle != NULL);

    if (handle->ipc != NULL) {
        return FALSE;
    }

    handle->ipc = ipc;
    return TRUE;
}

bool
INW_addFD(INW *handle, int fd, int eventid)
{
	struct INW_eInfo *container;

	/* check the eventid */
	if(eventid < INWE_FDSTART)
		return FALSE;

	/* set up our container */
	container=emalloc(sizeof(struct INW_eInfo));
	container->info=fd;
	container->event_id=eventid;

	LL_append(handle->fds, container);

	return TRUE;
}

bool
INW_addSignal(INW *handle, int signo)
{
    struct sigaction    act;
    struct INW_eInfo    *evt;


    assert(handle != NULL);
    assert(signo < MAX_SIGNALS);

    if (sigismember(&regSignals, signo)) {
        DPRINT(DBG_IPC, "INW addSignal: signal %d already monitored", signo);
        return FALSE;   /* can't share signals between INW objects */
    }

    /* prep the event info now */
    evt = emalloc(sizeof(struct INW_eInfo));
    evt->event_id = INWE_SIG;
    evt->info = signo;

    /* setup the sigaction struct */
    memset(&act, 0, sizeof(act));
    act.sa_handler = INW_sighandler;

    if (sigaction(signo, &act, NULL) < 0) 
        return FALSE;   /* sigaction failed */

    /* mark it as registered, and add it to the signal event list */
    sigaddset(&regSignals, signo);
    LL_append(handle->sigs, evt);

    return TRUE;
}

bool
INW_addQueue(INW *handle, TSQueue *queue)
{
    assert(handle != NULL);
    assert(handle->queue == NULL);

    handle->queue = queue;
    /* event_id should alread be set */
    handle->qEInfo.info = (int)queue;

	return TRUE;
}

bool
INWLL_regSignal(int signo)
{
    assert(signo < MAX_SIGNALS);

    if (sigismember(&regSignals, signo)) 
        return FALSE;

    sigaddset(&regSignals, signo);

    return TRUE;
}

struct INW_eInfo *
INW_getEvent(INW *handle)
{
	fd_set              rfds;
    int                 n;
    struct INW_eInfo    *evt;

    TPRINT("INW_getEvent start");

    /* check our queues first */
    evt = checkQueues(handle);
    n = 0;

    DPRINT(DBG_IPC, "getEvent: start", NO_VARG);

    /* If there is nothing in the queues, then check the FDs */
    while (evt == NULL) {
        
        if (handle->noFDWait) { 
            /* blocking wait on the TSQ, any elements present will turn
             * up in the checkQueues function anyway.
             */
            DPRINT(DBG_IPC, "getEvent: TSQ wait", NO_VARG);
            TSQ_wait(handle->queue, IPC_SELECT_TIMEOUT);
        }
        else if (handle->ipc == NULL) {
            /* create the fd_set */
            DPRINT(DBG_IPC, "getEvent: INWSelect wait", NO_VARG);
            FD_ZERO(&rfds);
            n = INWSelect(0, &rfds, handle);
            if (n < 0) break;
        }
        else {
            DPRINT(DBG_IPC, "getEvent: IPC wait", NO_VARG);
            if (! IPCLL_readAllSocks(handle->ipc, INWSelect, handle) ) {
                n = -1;
                break;
            }
        }

        /* now we need to check the queues agin for waiting data */
        evt = checkQueues(handle);
    } 

    if (n < 0) {
        DPRINT(DBG_IPC, "getEvent: error returning NULL", NO_VARG);
        return NULL;
    }

    /* evt should be non-null here */
    DPRINT(DBG_IPC, "getEvent: end (info = %d, id = %d)",
            evt->info, evt->event_id);

    TPRINT("INW_getEvent end");

	return evt;
}

bool
INW_stopFDWaiting(INW *handle)
{
    assert(handle != NULL);

    if (handle->queue != NULL) {
        handle->noFDWait = TRUE;
        return TRUE;
    }

    return FALSE;
}

static int
INWSelect(int sockMax, fd_set *rfds, void *data)
{
	LLIterator          it;
    INW                 *handle;
	struct INW_eInfo    *temp_elem;
    struct timeval      timeout;
    int                 fdcount;

    assert(data != NULL);
    assert(rfds != NULL);

    handle = (INW*)data;

    DPRINT(DBG_IPC, "INWSelect: start (sockMax = %d; rfds = %p; data = %p)",
            sockMax, rfds, data);

    /* Add the FDs to the select set */
    it = LL_iterator(handle->fds);
    while(NULL!=LLI_next(it))
    {
        temp_elem = (struct INW_eInfo*)LLI_elem(it);
        if (temp_elem->event_id != INWE_SIG) {
            DPRINT(DBG_IPC, "INWSelect: adding fd #%d to select list", 
                    temp_elem->info);

            FD_SET(temp_elem->info, rfds);
            if (sockMax < temp_elem->info) 
                sockMax = temp_elem->info;
        }
    }
    LLI_destroy(&it);

    DPRINT(DBG_IPC, "INWSelect: after adding FDs, sockMax = %d", sockMax);

    sockMax++;  /* greatest FD + 1 */

    /* Check for waiting signals and queue them.
     * If there are signal waiting skip the select so that the signal
     * will get passed up without the 10s delay of the select.
     */
    if (queueSignals(handle) ) {
        /* Not an error, but no waiting FDs (as select() wasn't called). */
        DPRINT(DBG_IPC, "INWSelect: end (got a signal)", NO_VARG);
        return 0;
    }

    /* This will make us active every so often to deal with signals
     * and what not.
     */
    timeout.tv_sec = IPC_SELECT_TIMEOUT;
    timeout.tv_usec = 0;
    fdcount = select(sockMax,rfds,NULL,NULL,&timeout);

    DPRINT(DBG_IPC, "INWSelect: select() returned = %d",
            fdcount);


    if (fdcount < 0) {
        if (errno == ENOMEM) {
            /* oops ... big problem */
            throw_fatal("Select Returned ENOMEM");
        }

        /* if interrupted by a signal, just pretend that no fds were
         * ready. it's not an error for our purposes
         */
        if (errno == EINTR) fdcount = 0;
        else return -1;
    }

    /* Check for FD ready events. */
    it=LL_iterator(handle->fds);
    while(fdcount > 0 && NULL!=LLI_next(it))
    {
        temp_elem=LLI_elem(it);
        if (temp_elem->event_id != INWE_SIG && FD_ISSET(temp_elem->info,rfds) )
        {
            /* Queue the FD */
            DPRINT(DBG_IPC, "INWSelect: adding fd (= %d) to queue", 
                    temp_elem->info);
            LL_append(handle->events, temp_elem);
            fdcount--;    
        }
    }


    /* clean up */
    LLI_destroy(&it);

    DPRINT(DBG_IPC, "INWSelect: end (%d)", fdcount);

    return fdcount;
}

static struct INW_eInfo *
checkQueues(INW *handle)
{
    struct INW_eInfo    *evt;
    int                 n;

    assert(handle != NULL);
    evt = NULL;

    DPRINT(DBG_IPC, "checkQueues: handle->queue = %p", handle->queue);

    if (! LL_isEmpty(handle->events) ) {
        evt = (struct INW_eInfo*)LL_first(handle->events);
        LL_rmFirst(handle->events);
    }
    else if (NULL != handle->ipc && (n = IPCLL_hasQueued(handle->ipc) )) {
        evt = &(handle->ipcEInfo);
        evt->info = n;
    }
	else if (NULL != handle->queue && !TSQ_isEmpty(handle->queue) ) {
        DPRINT(DBG_IPC, "checkQueues: TSQ has elements", NO_VARG);
        evt = &(handle->qEInfo);
	}

    return evt;
}

static bool
queueSignals(INW *handle)
{
    LLIterator          itr;
    struct INW_eInfo    *evt;
    bool                have_queued;

    assert(handle != NULL);

    TPRINT("queueSignals");

    /* do this before we block signals */
    itr = LL_iterator(handle->sigs);

    /* block signals that we handle */
    sigprocmask(SIG_BLOCK, &regSignals, NULL);

    have_queued = FALSE;

    while ( (evt = LLI_next(itr)) != NULL) {
        if (sigismember(&raisedSignals, evt->info) ) {
            sigdelset(&raisedSignals, evt->info);

            /* queue it */
            LL_append(handle->events, evt);
            have_queued = TRUE;
        }
    }
    /* unblock signals and clean up */
    sigprocmask(SIG_UNBLOCK, &regSignals, NULL);

    LLI_destroy(&itr);

    return have_queued;
}

void
INW_sighandler(int signo)
{
    /* add this to the list of raised signals IFF there is a registered
     * INW object that is expecting it.
     */
    if (sigismember(&regSignals, signo) )
        sigaddset(&raisedSignals, signo);

    TPRINT("INW_sighandler");
    DPRINT(DBG_IPC, "INW_sighandler: got signal: %d", signo);
}

