/**
 * @file inputwait.h
 *
 * @brief A library that handles blocking for input.
 * @par
 * These functions abstract blocking until input arrives.
 * @note This is closly tied to the IPC mechanisms.
 * @ingroup inwait
 *
 * $Id: inputwait.h,v 1.9.2.7 2003/04/30 20:03:18 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Jan 28 12:42:47 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "util/tsqueue.h"

#ifndef INC_INPUTWAIT_H
#define INC_INPUTWAIT_H

/**@defgroup inwait Input Waiter 
 * @ingroup ipc
 */

/**@name Predefined event numbers 
 * @ingroup inwait
 * @{
 */
#define INWE_IPC        1           /**< IPC Event */
#define INWE_SIG        2           /**< Signal Event */
#define	INWE_TSQ		3			/**< TSQueue event */
#define INWE_FDSTART    4           /**< first allowable fd event */
/*@}*/

/**@addtogroup inwait
 * @{
 */

/** Holds data about an event */
struct INW_eInfo
{
    /** 
     * Info about the event. One of the following possibilities: 
     * \li a file descriptor number.
     * \li if \c event_id==INWE_IPC, it's a bit map of waiting IPC events. 
     * \li if \c event_id==INWE_SIG, it's a signal number.
	 * \li if \c event_id==INWE_TSQ, it's a TSQueue * object that points
	 * 		directly to the list; THIS MUST BE CAST BACK.
     */
	int         info;			

	int         event_id;   /**< Event ID */
};

/** The input waiter struct */
typedef struct 
{
	List        fds;		/**< FD event list, of INW_eInfo objects */ 
    List        sigs;       /**< Signal event list, of INW_eInfo objects */
    List        events;     /**< Queued Events, of INW_eInfo objects  */

    TSQueue             *queue;     /**< Thread Safe Queue */
    struct INW_eInfo    qEInfo;     /**< Object to return TSQueue event in */

    IPC                 *ipc;       /**< IPC handle to wait on */
    struct INW_eInfo    ipcEInfo;   /**< Object to return IPC event in.*/  

	uint32_t    next_evtid; /**< ID of the next event */
    bool        noFDWait;   /**< Don't wait on FDs, @see INW_stopFDWaiting() */
} INW;


/**
 * Initilize the Input Waiter module 
 */
void INW_initialize(void);

/**
 * Shutdown the Input Waiter module
 */
void INW_shutdown(void);

/** 
 * Create a new INW 
 *  @return the new INW struct.
 */
INW * INW_create(void);

/** Destroy an INW structure 
 *  @param handle The INW structure to destory
 */
void INW_destroy(INW **handle);

/** 
 * Flush the FD wait list.  This empties the list of FDs to wait for
 * input on.
 * @param handle The INW handle
 */
void INW_clearFDs(INW *handle);

/**
 * Flush the Signal wait list.  This empties the list of signals to wait
 * for.
 *
 * @param handle The INW handle.
 */
void INW_clearSignals(INW *handle);

/** Get a new eventid to use with _this_ INW object
 *  Can only provide 1 through the maximum value in an unsigned int before
 *  undefined activity occurs.  This can only manage the event IDs if this
 *  is the only way they are generated.
 *
 *  @param handle The handle to get a new eventid for
 *  @return A new eventid
 */
int INW_getNextEventID(INW *handle);

/** Add an IPC handle to wait on.
 *  @param handle the INW handle
 *  @param ipc The IPC handle
 *  @return TRUE on success, FALSE otherwise.
 */
bool INW_addIPC(INW *handle, IPC *ipc);

/** Add a File Descriptor to wait on.
 *  The event id given to this function must be greater than
 *  INWE_FDSTART for it to be accepted.
 *
 *  @param handle The INW handle.
 *  @param fd The File Descriptor to wait on.
 *  @param eventid The event ID to return when input is waiting on fd
 *  @return TRUE on success, FALSE otherwise
 */
bool INW_addFD(INW *handle, int fd, int eventid);

/**
 * Add a signal to wait on.
 * Signal Events all used the fixed event_id \c INWE_SIG, and the info
 * field will contain the signal number.  Duplicate signals are ignored,
 * either a signal has been raised since the last queuing or it hasn't.
 *
 * Only one INW handle can watch any given signal.  So watching a signal
 * registered by another handle will fail.
 *
 * @param handle The INW handle
 * @param signo The signal number to watch
 * @return TRUE on sucess, FALSE otherwise.
 */
bool INW_addSignal(INW *handle, int signo);

/**
 * Add a queue to wait on.
 * Queue events use the fixed event_id \c INWE_TSQ, and the info field will
 * contain a pointer to the queue (this will be cast as an integer while stored;
 * you MUST typecast it back to a (TSQueue *)). 
 *
 * @note There can be only one TSQueue waited on.
 *
 * @param handle The INW handle
 * @param queue The queue to wait on
 * @return TRUE on sucess, FALSE otherwise.
 */
bool INW_addQueue(INW *handle, TSQueue *queue);

/** 
 * Wait for input.
 * This will block _indefinitely_.
 *
 * @param handle The INW handle
 * @return the event that has input waiting, or NULL on error.
 */
struct INW_eInfo * INW_getEvent(INW *handle);

/**
 * Stop waiting on FDs.  If this is called, and there is a TSQ to wait
 * on then INW_getEvent() will only wait on that TSQ ... this means that
 * FDs will not be check any longer. Signals will still be checked.
 * This also means that you can close the FDs registered with this input
 * waiter.
 *
 * @note Yes this is part of a hack to make BEEP IM shutdown better.
 *
 * @param handle The INW object.
 * @return TRUE if we can only wait on the TSQ (i.e. there is one
 *  registered).
 */
bool INW_stopFDWaiting(INW *handle);

/*@}*/

#endif /* INC_INPUTWAIT_H */
