/**
 * @file ipc.c
 *
 * @brief Implementation of the IPC module.
 * @ingroup ipc
 *
 * $Id: ipc.c,v 1.27.2.7 2003/04/30 20:03:18 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Feb 26 12:01:34 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include "ipc_private.h"
#include "ipc.h"
#include "ipc_err.h"
#include "buffer.h"
#include "ipcll.h"
#include "ipc_ctl.h"
#include "ipc_pkt.h"
#include "ipc_sig.h"
#include "ipc_cstate.h"
#include "ipc_lock.h"

#include "util/acl.h"
#include "util/netaddr.h"
#include "privs.h"

#include "strcompat.h"
#include "error.h"
#include "debug.h"
#include "test.h"

/* ******** File Globals **********/

static char const rcsid[] 
    = "$Id: ipc.c,v 1.27.2.7 2003/04/30 20:03:18 devink Exp $";

/** 
 * The buffer size to start with when sending paths. The buffer for
 * constructing path messages is reused for all path blocks, and this is
 * just a best guess at initial sizing, any message larger than this
 * will cause the buffer to be reallocated to the larger size. The extra
 * 128 bytes is for the two host names (assumes host names are usually
 * less that 64 bytes).
 */
#define DFLT_PATH_BUFSZ     IPCT_PATH_LEN + 128

/* ******** Private File Functions *********/
/**
 * Allocate space for a new IPC struct.  This will do common
 * initilization.  specifically:
 * \li zero all the memory
 * \li create new lists for queue
 *
 * @return A pointer to the new memory.
 */
static IPC* IPC_new(void);

/* ******* Public Functions *********/

/* initilize the controller */
IPC*
IPC_initController(char const *sock_path)
{
    IPC                 *ipc;
    struct sockaddr_un  addr;
    int                 i;
    int                 sock;

    ipc = IPC_new();

    ipc->mode = IPCM_SERVER;
    ipc->paddr = IPCAP_CTLR;

    if (sock_path == NULL || *sock_path == '\0') {
        ipc->errcode = IPCE_BADPATH;
        goto end;   /* error, no cleanup needed */
    }

    IPCLL_clearError(ipc);

    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock < 0) {
        ipc->errnocode = errno;
        ipc->errcode = IPCE_BINDERR;
        goto end;   /* error, no real clean up is needed yet */
    }
    ipc->mainState.sock = sock;

    addr.sun_family = PF_UNIX;
    strlcpy(addr.sun_path, sock_path, UNIX_PATH_MAX);

    /* Bind the socket ... we need privs to do this */
    if (! PRIV_bindUnixSocket(sock, &addr, TRUE, 0) ) {
        /* errors should have been logged by now */
        ipc->errcode = IPCE_BINDERR;
        goto error_sock;    /* error, cleanup the socket */
    }

    if (listen(sock, IPC_MAX_BACKLOG) ) {
        ipc->errnocode = errno;
        ipc->errcode = IPCE_BINDERR;
        goto error_sock;   /* error, cleanup the socket */

    }

    /* make sock non-blocking from here on out */
    if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
        ipc->errnocode = errno;
        ipc->errcode = IPCE_BINDERR;
        goto error_sock;   /* error, cleanup the socket */
    }

    /* okay the socket is ready to roll */
    ipc->mainState.state = IPCST_READY;

    /* okay, now allocate the cilConns data */
    ipc->cliConns = emalloc(IPC_MAX_IMPROCS * sizeof(struct IPC_cstate));
    memset(ipc->cliConns, 0, IPC_MAX_IMPROCS * sizeof(struct IPC_cstate));
    /* set sockets to -1 */
    for (i = 0; i < IPC_MAX_IMPROCS; i++) {
        IPCCST_init(ipc, i, -1);
    }


    /* and save the path for later */
    strlcpy(ipc->sock_path, sock_path, UNIX_PATH_MAX);

    /* Setup signal handling. */
    if (! IPCSig_setupController(ipc) ) goto error_sock;

    /* the end of the function, ipc must be in the proper return state
     * when we get here.
     */
    end:
        return ipc;

    /* Error handlers */


    error_sock:     /* need to cleanup the socket */
        close(sock);
        ipc->mainState.sock = -1;

        return ipc;
    
}

/* initilize the input module */
IPC*
IPC_initInputModule(char const *sock_path)
{
    IPC     *ipc;

    ipc = IPC_new();

    ipc->mode = IPCM_CLIENT;
    ipc->paddr = IPCAP_UNSET;

    if (sock_path == NULL || *sock_path == '\0')  {
        ipc->errcode = IPCE_BADPATH;
        return ipc;
    }

    /* connection is done elsewhere, but we need to save the path */
    strlcpy(ipc->sock_path, sock_path, UNIX_PATH_MAX);

    /* setup signal handling. Any error will be communicated via the new
     * ipc handle.
     */ 
    IPCSig_setupInputModule(ipc);

    /* done */
    return ipc;
}

/* get sock_path */
char const *
IPC_getSockPath(IPC *handle)
{
    assert(handle != NULL);
    return handle->sock_path;
}

/* Cleanup an IPC object */
void
IPC_destroy(IPC *handle)
{
    int         i;

    assert(handle != NULL);

    DPRINT(DBG_IPC, "IPC_destroy: start", NO_VARG);


    /* shut down the connection if needed */
    if (handle->cliConns != NULL) {
        for (i = 0; i < IPC_MAX_IMPROCS; i++) {
            /* 
             * This forces the free, which means it will be cleaned if
             * needed.
             */
            IPCCST_free(handle, i, TRUE);
        }
        free(handle->cliConns);
    }

    if (handle->mainState.sock >= 0) {
        close(handle->mainState.sock);
    }

    /* The server created the scoket, so only the server can delete it */
    if (handle->mode == IPCM_SERVER) {
        PRIV_unlink(handle->sock_path);
    }

    for (i = 0; i < IPCQ_COUNT; i++) {
        DPRINT(DBG_IPC, "IPC_destroy: queue %d has %d events.",
                i, LL_size(handle->queues[i]));
        LL_destroy(&(handle->queues[i]));
    }

    free(handle);
}

void
IPC_stopConnections(IPC *handle)
{
    /* unlink the socket so that no one can connect */
    PRIV_unlink(handle->sock_path);
}


/* Send a syslog message */
bool
IPC_imSendMessage(IPC *handle, int channel, struct message *msg)
{
    struct IPC_buffer    buf;
    struct IPC_packet   packet;
    bool                success;
    
    assert(handle != NULL);

    if (msg == NULL) return FALSE;
    /* XXX check channel ? */

    DPRINT(DBG_IPC, "send msg: begin", NO_VARG);

    /* fill in the header */
    IPCPacket_init(&packet, handle, IPCT_MSG, channel);
    IPCPacket_setDstAddr(&packet, IPCAP_CTLR, IPCAC_GENERAL);

    BUF_init(&buf);

    /* first figure out how much space we need */
    buf.size = IPCT_MSG_LEN;
    if (msg->msglen == 0) 
        buf.size += BUF_EMPTYSTRSZ;
    else
        buf.size += msg->msglen;

    if (msg->hostlen == 0) 
        buf.size += BUF_EMPTYSTRSZ;
    else
        buf.size += msg->hostlen;

    if (msg->dfqdnlen == 0) 
        buf.size += BUF_EMPTYSTRSZ;
    else
        buf.size += msg->dfqdnlen;

    BUF_alloc(&buf);

    /* Just skip to the final cleanup and return code on error */
    success = FALSE;    /* haven't done anything yet */

    /* now we start adding the other fields */
    if (! BUF_setByte(&buf, msg->sev) ) goto cleanup;
    if (! BUF_setByte(&buf, msg->fac) ) goto cleanup;
    if (! BUF_setVarString(&buf, msg->host, msg->hostlen) ) goto cleanup;
    if (! BUF_setInteger(&buf, msg->timestamp) ) goto cleanup;
    if (! BUF_setInteger(&buf, msg->tz) ) goto cleanup;
    if (! BUF_setFixString(&buf, msg->tag, MSGTAG_LEN) ) goto cleanup;
    if (! BUF_setVarString(&buf, msg->msg, msg->msglen) ) goto cleanup;


    /* TCP cooked stuff */
    if (! BUF_setInteger(&buf, msg->pathid) ) goto cleanup;
    if (! BUF_setInteger(&buf, msg->connid) ) goto cleanup;
    if (NA_toBuffer(msg->dip, &buf) < 0) goto cleanup;
    if (msg->dfqdnlen > 0) {
        if (! BUF_setVarString(&buf, msg->dfqdn, msg->dfqdnlen) ) goto cleanup;
    }
    else {
        if (! BUF_setEmptyString(&buf) ) goto cleanup;
    }

    /* if we make it here then everything worked fine */
    IPCPacket_setData(&packet, &buf);

    IPCLL_send(handle, &packet);
    success = TRUE;


    /* Final cleanup and return code */
    cleanup:
        BUF_free(&buf);
        DPRINT(DBG_IPC, "send msg: end (%d)", success);

        return success;
} 

/* send a path */
bool
IPC_imSendPath(IPC *handle, int channel, struct path *pathHead)
{
    struct path         *curPath;
    uint32_t            last_id, first_id;
    struct IPC_buffer   ctrlBuf, pathBuf;
    IPC_bufMark         csPos, pbPos;
    size_t              pbSize, curSize;
    struct IPC_packet   packet;
    struct IPC_control  ctrl;
    int                 wlen;

    assert(handle != NULL);

    if (pathHead == NULL) return FALSE;

    BUF_init(&ctrlBuf);
    BUF_init(&pathBuf);

    /* initilize the packet header*/
    IPCPacket_init(&packet, handle, IPCT_CTRL, channel);
    IPCPacket_setDstAddr(&packet, IPCAP_CTLR, IPCAC_GENERAL);

    first_id = pathHead->id;    /* save this for later */
    last_id = first_id;         /* first message's last_id field */

    /* send the control message first */
    ctrl.type = IPCC_PATHST;
    ctrl.datalen = 1;
    ctrl.data = &first_id;

    if (! IPCLL_makeControlBuf(&ctrlBuf, &ctrl) ) {
        return FALSE;   /* big problem, bail now */
    }

    packet.type = IPCT_CTRL;
    IPCPacket_setData(&packet, &ctrlBuf);

    IPCLL_send(handle, &packet);

    BUF_free(&ctrlBuf);         /* will be reusing */

    pbSize = DFLT_PATH_BUFSZ;
    pathBuf.size = pbSize;
    BUF_alloc(&pathBuf);        /* probably big enough */
    pbPos = BUF_markPos(&pathBuf);  



    /* allocate space for the control packet buffers */
    csPos = BUF_markPos(&ctrlBuf);  /* save for later. */

    /* switch the packet to a PATH type for use in the loop */
    packet.type = IPCT_PATH;

    /* now start sending path messages */
    curPath = pathHead;
    while (curPath != NULL) {
        curSize = IPCT_PATH_LEN;
        curSize += curPath->ffqdnlen;
        curSize += curPath->tfqdnlen;

        if (pbSize < curSize) {
            /* allocate a new buffer with more space */
            BUF_free(&pathBuf);
            pathBuf.size = pbSize = curSize;
            BUF_alloc(&pathBuf);
        }
        else {
            BUF_resetToPos(&pathBuf, pbPos);
        }
        
        if (! BUF_setInteger(&pathBuf, last_id) ) break;
        if (! BUF_setInteger(&pathBuf, curPath->id) ) break;
        if (! BUF_setInteger(&pathBuf, curPath->connid) ) break;

        wlen = BUF_setVarString(&pathBuf, curPath->fromFQDN, curPath->ffqdnlen);
        if (wlen < 0) break;

        wlen = NA_toBuffer(curPath->fromIP, &pathBuf); 
        if (wlen < 0) break;

        wlen = BUF_setVarString(&pathBuf, curPath->toFQDN, curPath->tfqdnlen);
        if (wlen < 0) break;

        wlen = NA_toBuffer(curPath->toIP, &pathBuf);
        if (wlen < 0) break;

        if (! BUF_setInteger(&pathBuf, curPath->props) ) break;

        last_id = curPath->id;
        curPath = curPath->next;

        IPCPacket_setData(&packet, &pathBuf); 

        if (! IPCLL_send(handle, &packet) ) break;
    }

    /* Only sending control messages from here on out */
    packet.type = IPCT_CTRL;

    if (curPath != NULL) {
        /* Error sending path, cancle */
        ctrl.type = IPCC_PATHCNCL;
        ctrl.datalen = 1;
        ctrl.data = &first_id;
        if (IPCLL_makeControlBuf(&ctrlBuf, &ctrl) ) {
            IPCPacket_setData(&packet, &ctrlBuf);
            IPCLL_send(handle, &packet);
            BUF_free(&ctrlBuf);
        }
        /* XXX we have a problem if that didn't work */
        return FALSE;
    }

    ctrl.type = IPCC_PATHEND;
    ctrl.datalen = 1;
    ctrl.data = &last_id;

    if (IPCLL_makeControlBuf(&ctrlBuf, &ctrl) ) {
        IPCPacket_setData(&packet, &ctrlBuf);
        IPCLL_send(handle, &packet);
        BUF_free(&ctrlBuf);
    }
    /* XXX we have a problem if that didn't work */

    return TRUE;
}

/* recieve a syslog message */
bool
IPC_cRecvMessage(IPC *handle, struct IPC_addr *src, struct message *msg)
{
    struct IPC_packet   *packet;
    struct IPC_buffer    *buf;
    uint8_t             tbyte;

    assert(handle != NULL);
    assert(msg != NULL);

    DPRINT(DBG_IPC, "recv msg: start", NO_VARG); 
    
    packet = IPCLL_recvPacket(handle, IPCT_MSG); 
    if (packet == NULL) return FALSE;      /* errcode set by IPCLL_recv() */

    /* fill in the address */
    if (src != NULL) {
        src->proc = packet->sp_addr;
        src->channel = packet->sc_addr;
    }

    /* fill in the message */
    buf = packet->data;

    if (! BUF_getByte(buf, &tbyte) ) goto error;
    msg->sev = tbyte;

    if (! BUF_getByte(buf, &tbyte) ) goto error;
    msg->fac = tbyte;

    if (! BUF_getVarString(buf, &(msg->hostlen), &(msg->host), 0) ) 
        goto error;

    if (! BUF_getTimestamp(buf, &(msg->timestamp)) ) goto error;
    if (! BUF_getInteger(buf, (uint32_t*)&(msg->tz)) ) goto error;

    if (! BUF_getString(buf, msg->tag, MSGTAG_LEN) ) goto error;

    if (! BUF_getVarString(buf, &(msg->msglen), &(msg->msg), 0) ) 
        goto error;

    if (! BUF_getInteger(buf, (uint32_t*)&(msg->pathid)) ) goto error;
    if (! BUF_getInteger(buf, (uint32_t*)&(msg->connid)) ) goto error;

    msg->dip = NA_newBuffer(buf);
    if (NA_isError(msg->dip) ) {
        NA_destroy(msg->dip);
        goto error;
    }

    if (! BUF_getVarString(buf, &(msg->dfqdnlen), &(msg->dfqdn), 0) ) 
        goto error;

    DPRINT(DBG_IPC, "recv msg: done (TRUE)", NO_VARG);

    IPCPacket_free(packet);
    return TRUE;

    /* Cleanup after an error and return the proper error condition */
    error:
        IPCPacket_free(packet);
        DPRINT(DBG_IPC, "recv msg: error", NO_VARG); 
        handle->errcode = IPCE_BADFMT;
        return FALSE;
}

/* recieve a path message */
bool
IPC_cRecvPath(IPC *handle, struct IPC_addr *src, struct path *pathHead)
{
    struct IPC_buffer    *buf;
    struct IPC_pathList *pthList;
    struct path         *curPath;
    List                blist;
    uint32_t            lastid;

    IPCLL_clearError(handle);

    pthList = IPCLL_recvPath(handle);
    if (pthList == NULL) return FALSE;  /* errcode set by IPCLL_recv() */

    if (src != NULL) {
        src->proc = pthList->src.proc;
        src->channel = pthList->src.channel;
    }


    curPath = pathHead;
    blist = pthList->bufList;
    buf = (struct IPC_buffer *)LL_first(blist);
    while (buf != NULL) {
        curPath->next = NULL;

        /* this remove curPkt from the list */
        if (! LL_rmFirst(blist) ) break;

        /* de-serialize the packet */

        /* this is the last id, we can just toss it */
        if (! BUF_getInteger(buf, &lastid) ) break; 

        if (! BUF_getInteger(buf, (uint32_t*)&(curPath->id)) ) break;
        if (! BUF_getInteger(buf, (uint32_t*)&(curPath->connid)) ) break;

        if (!BUF_getVarString(buf,&(curPath->ffqdnlen),&(curPath->fromFQDN),0)) 
            break;

        curPath->fromIP = NA_newBuffer(buf);
        if (NA_isError(curPath->fromIP) ) {
            NA_destroy(curPath->fromIP);
            break;
        }

        if (! BUF_getVarString(buf,&(curPath->tfqdnlen),&(curPath->toFQDN),0)) 
                break;

        curPath->toIP = NA_newBuffer(buf);
        if (NA_isError(curPath->toIP) ) {
            NA_destroy(curPath->toIP);
            break;
        }

        if (! BUF_getInteger(buf, &(curPath->props)) ) break;

        /* free the current buffer */
        BUF_free(buf);
        free(buf);

        /* get the next path packet */
        buf = (struct IPC_buffer *)LL_first(blist);
        
        /* setup the next path struct */
        if (buf != NULL) {
            curPath->next = emalloc(sizeof(struct path));
            curPath = curPath->next;
        }

    }


    if (buf != NULL) {
        struct path *lstPath;
        /* error */

        /* first free all thoes paths that we just alloc'ed */
        curPath = pathHead->next;
        while (curPath != NULL) {
            lstPath = curPath;
            curPath = curPath->next;

            free(lstPath);
        } 

        /* free the current buffer */
        BUF_free(buf);
        free(buf);

        /* free the rest of the buffers in the list */
        buf = (struct IPC_buffer *)LL_first(blist);
        while (buf != NULL) {
            LL_rmFirst(blist);  /* rm from list */
            BUF_free(buf);      /* free memory */
            free(buf);

            /* next */
            buf = (struct IPC_buffer *)LL_first(blist);
        }
        /* We can safely free the the pathList struct now */
        LL_destroy(&(pthList->bufList));
        free(pthList);


        if (handle->errcode != IPCE_SUCCESS) {
            handle->errcode = IPCE_BADFMT;
        }
        return FALSE;
    }

    /* success */ 

    /* free the the pathList struct now, it should be empty */
    LL_destroy(&(pthList->bufList));
    free(pthList);


    return TRUE;
}


void
IPC_inputmoduleExit(IPC *handle)
{
    struct IPC_buffer    buf;
    struct IPC_control  ctrl;
    struct IPC_packet   pkt;
    
    assert(handle != NULL);

    DPRINT(DBG_IPC, "IPC_inputmoduleExit: start", NO_VARG);

    BUF_init(&buf);
    IPCPacket_init(&pkt, handle, IPCT_CTRL, IPCAC_GENERAL);
    IPCPacket_setDstAddr(&pkt, IPCAP_CTLR, IPCAC_GENERAL);

    ctrl.type = IPCC_IMEXIT;
    ctrl.datalen = 1;
    ctrl.data = &(handle->paddr);

    if (! IPCLL_makeControlBuf(&buf, &ctrl) ) {
        DPRINT(DBG_IPC, "IPC_inputmoduleExit: makeControlBuf failed", NO_VARG);
        /* XXX Big problem, not quite sure what to do about it though */
        return;
    }

    IPCPacket_setData(&pkt, &buf); 

    if (! IPCLL_send(handle, &pkt) ) {
        DPRINT(DBG_IPC, "IPC_inputmoduleExit: send failed", NO_VARG);
        /* XXX Big problem, not quite sure what to do about it though */
        return;
    }

    /* update our state */
    handle->mainState.state = IPCST_SENTEXIT;

    /* an ACK is needed before this connection is closed. 
     * So this will block until such an ack is recieved
     */
    /** @todo put a timeout on this so that it can't get stuck waiting
     * forever.
     */
    while (handle->mainState.state != IPCST_EXIT) {
        IPCLL_recvPacket(handle, IPCT_NONE);
    }

    DPRINT(DBG_IPC, "IPC_inputmoduleExit: shutdown complete", NO_VARG);

    /* now we shutdown the connection */
    close(handle->mainState.sock);
    handle->mainState.sock = -1;

    DPRINT(DBG_IPC, "IPC_inputmoduleExit: done", NO_VARG);
}

bool 
IPC_inputmoduleStart(IPC *handle)
{
    struct IPC_buffer    buf;
    struct IPC_control  ctrl;
    struct IPC_packet   pkt;
    struct sockaddr_un  addr;
    int                 sock;
    pid_t               pid;

    assert(handle != NULL);

    pid = getpid();

    BUF_init(&buf);
    IPCPacket_init(&pkt, handle, IPCT_CTRL, IPCAC_GENERAL);
    IPCPacket_setDstAddr(&pkt, IPCAP_CTLR, IPCAC_GENERAL);

    ctrl.type = IPCC_IMSTART;
    ctrl.datalen = 1;
    ctrl.data = (uint32_t*)&pid;

    IPCLL_clearError(handle);


    if (! IPCLL_makeControlBuf(&buf, &ctrl) ) {
        handle->errcode = IPCE_INITFAIL;
        return FALSE;
    }

    IPCPacket_setData(&pkt, &buf);

    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock < 0) {
        handle->errcode = IPCE_INITFAIL;
        handle->errnocode = errno;
        return FALSE;
    }
    handle->mainState.sock = sock;

    addr.sun_family = PF_UNIX;
    strlcpy(addr.sun_path, handle->sock_path, UNIX_PATH_MAX);

    if (connect(sock, (struct sockaddr*)&addr, sizeof(addr) ) < 0) {
        handle->errcode = IPCE_INITFAIL;
        handle->errnocode = errno;
        return FALSE;
    }

    /* make sock non-blocking from here on out */
    if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
        handle->errnocode = errno;
        handle->errcode = IPCE_BINDERR;
        return FALSE;
    }

    handle->mainState.state = IPCST_NEW;

    if (! IPCLL_send(handle, &pkt) ) {
        return FALSE;
    }

    while (handle->mainState.state != IPCST_READY) {
        /** @todo Add a timeout so this won't loop forever. */
        (void) IPCLL_recvPacket(handle, IPCT_NONE);

        /* make sure there wasn't an error */
        if (handle->errcode != IPCE_SUCCESS) {
            break;
        }
    }

    /* Check for an error completing the hand shake */
    if (handle->errcode != IPCE_SUCCESS) {
        return FALSE;
    }

    return TRUE;
}

bool
IPC_recvEvent(IPC *handle, struct IPC_event *evt)
{
    struct IPC_event    *tevt;

    DPRINT(DBG_IPC, "recv Event: start", NO_VARG);

    if (handle->mode != IPCM_SERVER) {
        handle->errcode = IPCE_BADMODE;
        return FALSE;
    }

    /* good opp, to get more data off the sockets */
    tevt = IPCLL_recvEvent(handle);

    if (tevt == NULL) {
        DPRINT(DBG_IPC, "recv Event: IPCLL_recvEvent failed", NO_VARG);
        return FALSE;
    }

    if (tevt != NULL) {
        /* copy tevt to evt */    
        *evt = *tevt;
        free(tevt);     /* clean that up */
    }

    /* if this is the last event for a client (IMSTOP or IMDEATH) then
     * we can safely free the client state.
     */
    if (evt->type == IPCEVT_IMSTOP || evt->type == IPCEVT_IMDEATH) {
        DPRINT(DBG_IPC, "recv Event: got EXIT event.", NO_VARG);

        IPCCST_free(handle, IPC_PAToIndex(evt->addr.proc), FALSE);
    }

    DPRINT(DBG_IPC, "recv Event: end (event %d From %d)", 
            evt->type, evt->addr.proc); 

    return TRUE;
}

bool
IPC_sendEvent(IPC *handle, struct IPC_addr *dst, int channel, 
        IPC_eventType evt, uint32_t data)
{
    struct IPC_packet   pkt;
    struct IPC_buffer    buf;
    bool                success;

    
    /* Make sure the caller cand send this message */
    if (handle->mode == IPCM_SERVER) {
        if ( (evt & IPCEVT_SRVSEND) != evt) return FALSE;
    }
    else if (handle->mode == IPCM_CLIENT ) {
        if ( (evt & IPCEVT_CLISEND) != evt) return FALSE;
    }
    else {
        throw_fatal("Unknowen IPC mode");
    }

    IPCPacket_init(&pkt, handle, IPCT_EVENT, channel);
    IPCPacket_setDstAddr(&pkt, dst->proc, dst->channel);

    BUF_init(&buf);
    buf.size = IPCT_EVENT_LEN;
    BUF_alloc(&buf);

    success = FALSE;

    /* build the data portion of the packet */
    if (! BUF_setInteger(&buf, evt) ) goto cleanup; /* type */
    if (! BUF_setInteger(&buf, 0) ) goto cleanup;   /* pid */
    if (! BUF_setInteger(&buf, data) ) goto cleanup;/* data */

    /* Send the packet */
    IPCPacket_setData(&pkt, &buf);
    IPCLL_send(handle, &pkt);

    success = TRUE;     /* done */

    /* Cleanup and exit, may or may not be success */
cleanup:
    BUF_free(&buf);

    return success;
}
    
bool
IPC_lockRequest(IPC *handle, IPC_lock lockid)
{
    assert(handle != NULL);
    assert(handle->mode == IPCM_CLIENT);

    return IPCLock_send(handle, IPCLT_REQ, lockid, IPCAP_CTLR);
}

void
IPC_lockReqCancle(IPC *handle, IPC_lock lockid)
{
    assert(handle != NULL);
    assert(handle->mode == IPCM_CLIENT);

    (void) IPCLock_send(handle, IPCLT_REQCNCL, lockid, IPCAP_CTLR);
}

bool
IPC_imRecvLock(IPC *handle, IPC_lock *lockid)
{
    IPC_lockType    type;

    if (! IPCLock_recv(handle, NULL, &type, lockid) ) 
        return FALSE;

    /* Make sure that the subtype is correct */
    if (type != IPCLT_LOCK) {
        handle->errcode = IPCE_BADLT;
        return FALSE;
    }

    return TRUE;
}

void
IPC_lockRelease(IPC *handle, IPC_lock lockid)
{
    assert(handle != NULL);
    assert(handle->mode == IPCM_CLIENT);

    (void) IPCLock_send(handle, IPCLT_LOCK, lockid, IPCAP_CTLR); 
}

bool 
IPC_cRecvLock(IPC *handle,struct IPC_addr *src,IPC_lockType *ty,IPC_lock *lk)
{
    assert(handle != NULL);
    assert(handle->mode == IPCM_SERVER);

    return IPCLock_recv(handle, src, ty, lk);
}

bool
IPC_cSendLock(IPC *handle, struct IPC_addr *dst, IPC_lock lock)
{
    assert(handle != NULL);
    assert(handle->mode == IPCM_SERVER);

    return IPCLock_send(handle, IPCLT_LOCK, lock, dst->proc);
}

/* ACL Passing Stuff */
bool
IPC_cSendACL(IPC *handle, struct IPC_addr *dst, ACL *acl)
{
    struct IPC_packet   pkt;
    struct IPC_buffer   buf;
    bool                success;

    assert(handle != NULL);
    assert(dst != NULL);
    assert(acl != NULL);

    BUF_init(&buf);
    IPCPacket_init(&pkt, handle, IPCT_LOCK, IPCAC_GENERAL);
    IPCPacket_setDstAddr(&pkt, dst->proc, dst->channel);

    /* first pass to figure out the space needs */
    buf.size = ACL_getBufferSize(acl);
    if (buf.size < 0) {
        return FALSE;
    }

    /* skip to cleanup and return code on error */
    success = FALSE;

    BUF_alloc(&buf);
    if (ACL_toBuffer(acl, &buf) < 0) goto cleanup;

    IPCPacket_setData(&pkt, &buf); 

    if (! IPCLL_send(handle, &pkt) ) goto cleanup;

    success = TRUE;

    /* cleanup and return */
    cleanup:
        BUF_free(&buf);

        return success;
}

bool
IPC_imRecvACL(IPC *handle, int *channel, ACL **acl)
{
    struct IPC_packet   *pkt;

    assert(handle != NULL);
    assert(acl != NULL);
    assert(channel != NULL);

    /* will set handle::errcode properly */
    pkt = IPCLL_recvPacket(handle, IPCT_ACL); 
    if (pkt == NULL) return FALSE;

    *channel = pkt->dc_addr;

    *acl = ACL_fromBuffer(pkt->data);
    IPCPacket_free(pkt);                
    /* Error check is delayed, because pkt needs to be free regardless */
    if (*acl == NULL) {
        handle->errcode = IPCE_BADFMT;
        return -1;
    }

    return TRUE;
}

/*************** Private Functions ****************/
static IPC*
IPC_new()
{
    IPC     *ipc;
    int     i;

    ipc = emalloc(sizeof(IPC));
    memset(ipc, 0, sizeof(IPC));

    for (i = 0; i < IPCQ_COUNT; i++) {
        ipc->queues[i] = LL_create();
    }

    ipc->mainState.curPath = NULL;
    ipc->mainState.curPathId = -1;
    ipc->mainState.sock = -1;
    ipc->mainState.open_pkt = NULL;
    ipc->mainState.state = IPCST_FREE;

    return ipc;
}
