/**
 * @file ipc.h
 *
 * @brief InterProcess Communication top level interface.  
 * @par
 * The input modules and controlers use this interface to handle all
 * their IPC needs.  It hides all the details of marshalling data across
 * the communication link, and allows the actual mechanism to be changed
 * without affecting the rest of the program.
 *
 * @design d3_2
 *
 * @ingroup ipc
 *
 * $Id: ipc.h,v 1.18.2.4 2003/04/30 20:03:18 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Jan 24 11:30:23 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IPC_IPC_H
#define IPC_IPC_H

/* some document groups */
/**@defgroup ipc IPC Library */

/**@defgroup ipc_msgtypes IPC Message Type 
 * @ingroup ipc
 */

/**@defgroup ipc_locks IPC Locks 
 * @ingroup ipc
 */

#include "types.h"

#include "message/message.h"
#include "util/acl.h"

/**@name IPC Message Types 
 * @ingroup ipc_msgtypes
 * @{
 */

/** Message Type Codes */
typedef enum {
    IPCT_NONE = 0,      /**< No type. @internal */

    IPCT_LOCK,          /**< Lock */
    IPCT_PATH,          /**< Path packet */
    IPCT_MSG,           /**< Sysolg packet */
    IPCT_ACL,           /**< ACL packet */
    IPCT_EVENT,         /**< Event (i.e. start/stop) */
    IPCT_COUNT,         /**< Number of IPC packet types */

    IPCT_CTRL = 100     /**< Control packet, @internal */
} IPC_type;
/*@}*/

/**
 * IPC Lock Packet Sub-Types.
 * These are the sub types of the lock packet.  They use to be control
 * types ... but controls are internal to the IPC layer only.
 * @todo Update ipc_proto and design documents to reflect this.
 *
 * @ingroup ipc_msgtypes
 */
typedef enum {
    IPCLT_REQ,      /**< Lock Request */
    IPCLT_REQCNCL,  /**< Lock Request Cancle */
    IPCLT_LOCK,     /**< Actual lock */
    IPCLT_COUNT     /**< Number of lock sub-types */
} IPC_lockType;

/** This enumerates the various locks which can be given to the input
 *  modules.
 * @ingroup ipc_locks
 */
typedef enum {
    IPCL_ACCEPT,        /**< Lock accept() on the listening socket */
    IPCL_COUNT          /**< Number of defined locks */
} IPC_lock;



/** This structure contains information for an address in the IPC
 * system.  It consists of a process, channel pairing 
 * @todo reference design document here.
 * @ingroup ipc
 */
struct IPC_addr
{
    uint32_t  proc;     /**< process address id */
    uint32_t  channel;  /**< address input channel */
};

/**
 * This enumerates all the possible input module events
 * @ingroup ipc
 */
typedef enum
{

    /* Input module & child process related events */
    IPCEVT_IMSTART,     /**< Input module started */

    /* Stop/Death events */
    IPCEVT_CTLRDEATH,   /**< Controller died unexpectedly */
    IPCEVT_CHLDDEATH,   /**< unregistered child died */
    IPCEVT_IMSTOP,      /**< Input moudle ended */
    IPCEVT_IMDEATH,     /**< Input module died unexpectedly */

    /* events that can be sent */
    /** Connection Terminated.  Input modules need to send these events
     * to the controller so that the path cache will remain consistent.
     * It's expected that IPC_event::data will contain the connID
     */
    IPCEVT_CONEND = 0x100,      

    /* internal masks */
    IPCEVT_CLISEND = 0x0F00,    /**< Mask, event a client can send. @internal */
    IPCEVT_SRVSEND = 0xF000     /**< Mask, event a server can send. @internal */
} IPC_eventType;

/**
 * This stores information about an input module event
 * @ingroup ipc
 */
struct IPC_event
{
    IPC_eventType       type;   /**< Event Code */
    struct IPC_addr     addr;   /**< Address of sender */
    pid_t               pid;    /**< Process ID */
    uint32_t            data;   /**< Generic data */
};

/**@name Message Addresses 
 * @see IPC_addr
 * @ingroup ipc 
 * @{
 */
#define IPCAP_CTLR      1   /**< Address Node for the the Controller */
#define IPCAP_UNSET     0   /**< Initial Address for Input Modules */

#define IPCAC_GENERAL   0   /**< Address Channel for entire node */

/** 
 * Convert a process address to a proc index 
 * @param pa A uint32_t containing the process address
 * @return the proc index for that process address
 */
#define IPC_PAToIndex(pa)   ( ( (pa) - IPCAP_CTLR) - 1) 

/** 
 * Convert a proc index to a process address.
 * @param idx the proc index to convert
 * @return The process address for that proc index
 */
#define IPC_indexToPA(idx)  ( (idx) + IPCAP_CTLR + 1)
/*@}*/

/** 
 * Maximum number of separate input module processes 
 * @ingroup ipc
 */
#define IPC_MAX_IMPROCS     1024

/** IPC object type.  This can be treated as opaque. 
 * @ingroup ipc
 */
typedef struct IPC_object IPC;

/**@name IPC Initilization 
 * @ingroup ipc
 * @{ 
 */
/** 
 * Initialize the %message passing interface.  
 * Input modules and controlers have separate intilization processes.  
 * In the event of an errror a partially initilized IPC structure will
 * be returned, from this it is only saft to call the IPC_geterrno() and
 * IPC_geterrstr() functions.  It must be destroyed via IPC_destroy
 * still.
 *
 * @param sock_path The path of the UNIX domain socket to use, or NULL
 *        for the default (IPC_DFLT_SOCKET_PATH)
 * @return A pointer to the IPC structure
 */
IPC* IPC_initInputModule(char const *sock_path);     /* for an input module */
IPC* IPC_initController(char const *sock_path);      /* for the controller  */
/*@}*/

/**@name General IPC Functions
 * @ingroup ipc
 * @{
 */
/** 
 * Get the path that this IPC handle is connected to.
 * @return A poniter to  the string containing the socket path
 *
 * @pre handle != NULL
 */
char const *IPC_getSockPath(IPC *handle);

/** 
 * Destroy the %message passing handle.
 * This must be called when the process is done.
 * @param handle The IPC Handle
 * @param handle The handle obtained by one of the IPC_init* functions
 *
 * @pre handle != NULL
 */
void IPC_destroy(IPC *handle);

/**
 * Stop accepting connections on the IPC socket.
 *
 * @param handle The handle to work on.
 */
void IPC_stopConnections(IPC *handle);

/** 
 * Send a syslog message over the channel. 
 * This sends the message from an input module to the controller
 * @param handle The IPC handle
 * @param channel The input channel this message was recieved on.
 * @param msg The syslog message to send
 * @return return TRUE on success, FALSE on error
 *
 * @pre handle != NULL
 */
bool IPC_imSendMessage(IPC *handle, int channel, struct message *msg);

/** 
 * Send a path structure over the channel.  
 * This sends a message from the input module to the controller.
 * This will assume that the last element of the Path list is NULL,
 * and will send the entire linked structure.
 *
 * @param handle The IPC handle
 * @param channel The input channel this message was recieved on.
 * @param pathHead Pointer to the head of the path list
 * @return return TRUE on success, FALSE on error
 *
 * @pre handle != NULL
 */
bool IPC_imSendPath(IPC *handle, int channel, struct path *pathHead);

/** 
 * Receive a syslog message which is waiting.
 * @param handle The IPC handle
 * @param src An address struct to hold the source address.
 * @param msg message structure to hold the received data
 * @return TRUE on success, FALSE on error.
 *
 * @pre handle != NULL
 * @pre msg != NULL
 */
bool IPC_cRecvMessage(IPC *handle, struct IPC_addr *src, struct message *msg);

/** 
 * Receive a path which is waiting. You can allocate pathHead however
 * you choose, however, all of the struct paths that are linked into th
 * chain will be allocated via emalloc(), and must be cleaned up
 * accordingly.
 *
 * @param handle The IPC handle
 * @param src An address struct to hold the source address.
 * @param pathHead Structure to hold the received data
 * @return TRUE on success, FALSE on error.
 *
 * @pre handle != NULL;
 * @pre pathHead != NULL;
 */
bool IPC_cRecvPath(IPC *handle, struct IPC_addr *src, struct path *pathHead);


/** 
 * Notify the controller that this input module is exiting.  
 * @param handle The IPC handle
 *
 * @pre handle != NULL
 */
void IPC_inputmoduleExit(IPC *handle);

/** 
 * Notify the controller that this input module is ready.
 * @param handle The IPC handle
 * @return TRUE on success.
 *
 * @pre handle != NULL
 */
bool IPC_inputmoduleStart(IPC *handle);

/**
 * Recieve notification of events.
 * These events currently include start/stop/death events.
 *
 * @note 
 * no IPCEVT_END event will be passed up if there are
 * still messages from that client queued.  Instead an IPCEVT_QUEUED
 * event will be passed up. Conversly when an IPCEVT_END
 * event arrives it is safe to clean up resources 
 * associated with that child.  
 * @warning
 * It is possible to go into an infinate loop trying to clear this event
 * queue.  The reason is that an inputmodule that has died but still has
 * other queued messages will return an IPCEVT_QUEUED event, but the
 * IPCEVT_END event will still be on the queue.
 *
 * @param handle The IPC handle.
 * @param evt The variable to place the event in.
 * @return TRUE on success, FALSE on error.
 */
bool IPC_recvEvent(IPC *handle, struct IPC_event *evt);

/**
 * Send an IPC event.  
 * There are only a few types of events that can be sent, and this will
 * error if the event is not of the allowed types, see IPC_eventType.
 * 
 * @param handle The IPC handle.
 * @param dst The destination of the event.
 * @param channel The source channel.
 * @param evt The IPC_eventType to send.
 * @param data The data that goes with the IPC_event
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre handle != NULL
 */
bool IPC_sendEvent(IPC *handle, struct IPC_addr *dst, int channel, 
                    IPC_eventType evt, uint32_t data);
/*@}*/

/**@name Lock Passing Interface
 * Locks are general, thus the lack of channel info
 * @ingroup ipc
 * @{
 */
/** 
 * Request a lock.
 * This sends a lock request %message to the controller, however, the
 * lock is not actully granted until IPC_recvLock() is called.  The
 * return value of this function only represents the success or failure
 * of sending a lock request %message.
 * 
 * @param handle The IPC handle
 * @param lockid one of the \ref ipc_locks variables specifying
 *        which lock to acquire
 * @return TRUE if the lock request was sent, FALSE otherwise
 */
bool IPC_lockRequest(IPC *handle, IPC_lock lockid);

/** 
 * Cancle a lock request.
 * This function will cancle a previously made lock request.  
 *
 * @param handle The IPC handle
 * @param lockid The \ref ipc_locks for which we want to cancle our
 *      request
 */
void IPC_lockReqCancle(IPC *handle, IPC_lock lockid);

/** 
 * Receive a lock.
 * The lock is assumed to be held from the point of return of this
 * function.  This function is for the input modules to use.
 *
 * @param handle The IPC handle
 * @param lockid A place to store the \ref ipc_locks that was
 *        recieved.
 * @return TRUE on success, FALSE on error.
 */
bool IPC_imRecvLock(IPC *handle, IPC_lock *lockid);

/** Release the lock.
 *  This function is used by input modules to inform the controller that
 *  they have release a lock.  The lock is considered to be released the
 *  moment this function is entered.
 *
 *  @param handle The IPC handle
 *  @param lockid one of the \ref ipc_locks variables specifying
 *      which lock to acquire
 */
void IPC_lockRelease(IPC *handle, IPC_lock lockid);

/** Recieve a lock Request (cancle).
 *  This function is used by the controller to get a lock request or a
 *  request canclation.
 *
 *  @param handle The IPC handle.
 *  @param src The source of this packet
 *  @param type The specific request type this is.
 *  @param lockid The lock being requested is put in here.
 *  @return TRUE on success, FALSE otherwise.
 */
bool IPC_cRecvLock(IPC *handle, struct IPC_addr *src, IPC_lockType *type, 
                    IPC_lock *lockid);

/** Send a lock.
 *  This is used by the controller to send locks to the input modules.
 *
 *  @param handle The IPC handle
 *  @param dest The destination of the lock.
 *  @param lockid The lock which is being sent to the specified input
 *      module.
 *  @return TRUE on success, FALSE on error
 */
bool IPC_cSendLock(IPC *handle, struct IPC_addr *dest, IPC_lock lockid); 

/*@}*/

/**@name ACL Passing Interface
 * @ingroup ipc
 * @{
 */
/** Send an ACL.
 *  This will be used by the controller to pass an ACL to an input
 *  module.
 *  @param handle The IPC handle
 *  @param dst The input module id to send this acl to.
 *  @param acl The ACL structure to send to the input
 *      module.
 *  @return TRUE on success, FALSE otherwise.
 *
 *  @pre handle != NULL
 *  @pre dst != NULL
 *  @pre acl != NULL
 */                 
bool IPC_cSendACL(IPC *handle, struct IPC_addr *dst, ACL *acl);

/** 
 * Receive an ACL.
 * This will be used by the input modules to receive an ACL. The acl
 * parameter will point to the recieved ACL structure when this is over.
 * This structure will have been created via ACL_new() and should be
 * destroyed accordingly.
 *
 * @param handle The IPC handle to work with.
 * @param acl An ACL structure to hold the data.  
 * @param channel A palce to hold the channel this ACL applies to.
 * @return TRUE on success, FALSE on error.
 *
 * @pre handle != NULL;
 * @pre acl != NULL;
 * @pre channel != NULL;
 */
bool IPC_imRecvACL(IPC *handle, int *channel, ACL **acl);
/*@}*/

#endif /* IPC_IPC_H */

