/**
 * @file ipc_cstate.c
 *
 * @brief Define the functions that work on IPC_cstate objects.
 *
 * $Id: ipc_cstate.c,v 1.7.2.1 2003/04/30 20:03:19 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jun 26 18:49:51 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <unistd.h>
#include <stdlib.h>

#include "ipc_private.h"
#include "ipc.h"
#include "ipc_cstate.h"
#include "ipc_pkt.h"
#include "ipcll.h"
#include "ipc_ctl.h"

#include "debug.h"
#include "error.h"
#include "test.h"

static char const rcsid[] = "$Id: ipc_cstate.c,v 1.7.2.1 2003/04/30 20:03:19 devink Exp $";

/** 
 * Can we queue an event for the exit of this client?
 * @param cst The object to check.
 */
inline static bool IPCCST_canQueue(struct IPC_cstate *cst);

/**
 * Cleanup an IPC_cstate structure.  
 *
 * @param cst The IPC_cstate to cleanup.
 */
static void IPCCST_cleanup(struct IPC_cstate *cst);

/**
 * Queue an event for the given client's exit.
 *
 * @param handle The IPC handle to work with
 * @param pidx The proc index of the cstate in question.
 */
static void IPCCST_queueEvent(IPC *handle, int pidx);
/**@name CState Partial exit states
 * @{
 */
#define CST_PIPECLR     0x01    /**< All data is out of the pipe */
#define CST_REAPED      0x02    /**< waitpid() has been called */
#define CST_PROTSHUT    0x04    /**< Clean exit via IMEXIT packet */

/** Ready to be cleaned */
#define CST_CANCLEAN    (CST_PIPECLR | CST_REAPED)
/*@}*/

void
IPCCST_init(IPC *handle, int pidx, int sock)
{
    struct IPC_cstate  *cst;

    assert(handle != NULL);
    assert(pidx >= 0);
    assert(pidx < IPC_MAX_IMPROCS);

    cst = &(handle->cliConns[pidx]);

    memset(cst, 0, sizeof(struct IPC_cstate));


    if (sock >= 0) {
        /* Only do this is we are really initilizing for use */
        assert(cst->state == IPCST_FREE);

        cst->state = IPCST_NEW;
        handle->cliCount++;
    }
    else {
        cst->state = IPCST_FREE;
    }

    cst->sock = sock;
    cst->curPathId = -1;

    /* SIGCHLD is not given to the controller in unit_tests */
    IN_UNIT_TEST(cst->exitState |= CST_REAPED);
}

bool
IPCCST_canFree(struct IPC_cstate *cst)
{
    return cst->state == IPCST_CLEAN; 
}

void
IPCCST_pipeClear(IPC *handle, int pidx, bool protoshut)
{
    struct IPC_cstate   *cst;

    assert(handle != NULL);
    assert(pidx < IPC_MAX_IMPROCS);
    assert(pidx >= 0);

    DPRINT(DBG_IPC, "cst: pipeClear: Client %d, normal proto. shutdown = %s",
            IPC_indexToPA(pidx), protoshut ? "TRUE" : "FALSE");

    cst = &(handle->cliConns[pidx]);

    if (protoshut) {
        cst->exitState |= CST_PROTSHUT;
    }

    cst->exitState |= CST_PIPECLR;

    IPCCST_cleanup(cst);

    if (IPCCST_canQueue(cst) ) {
        IPCCST_queueEvent(handle, pidx);
    }

    DPRINT(DBG_IPC, "cst: pipeClear: end", NO_VARG);
}

void
IPCCST_reaped(IPC *handle, int pidx, int status)
{
    struct IPC_cstate   *cst;

    assert(handle != NULL);
    assert(pidx >= 0);
    assert(pidx < IPC_MAX_IMPROCS);

    DPRINT(DBG_IPC, "cst: reaped: Client %d, exit status = %x",
            IPC_indexToPA(pidx), status);

    cst = &(handle->cliConns[pidx]);

    cst->exitState |= CST_REAPED;

    if (IPCCST_canQueue(cst)) {
        IPCCST_queueEvent(handle, pidx);
    }

    DPRINT(DBG_IPC, "cst: reaped: end", NO_VARG);
}

void
IPCCST_free(IPC *handle, uint32_t pidx, bool force)
{
    struct IPC_cstate   *cst;

    assert(handle != NULL);
    assert(pidx >= 0);
    assert(pidx < IPC_MAX_IMPROCS);

    cst = &(handle->cliConns[pidx]);

    /* do we have anything to do? */
    if (cst->state == IPCST_FREE)  return;

    /* cleanup if needed, and if we are being forced to */
    if (force && IPCST_isActive(cst->state) ) {
        IPCCST_cleanup(cst);
        /* no event is queued, b/c this may be reused before it:
         * surfaces.
         */
    }

    /* make sure that we can free this */
    assert(IPCCST_canFree(cst));

    /* final reset of some variables so that it can be reused. */
    cst->state = IPCST_FREE;

    /* one less client */
    handle->cliCount--;
}

void
IPCCST_clearCurPath(struct IPC_cstate *cst)
{
    struct IPC_buffer    *oldbuf;

    assert(cst != NULL);

    if (cst->curPath == NULL) 
        return;     /* nothing to do */

    oldbuf = LL_first(cst->curPath->bufList);
    while (oldbuf != NULL) {
        LL_rmFirst(cst->curPath->bufList);
        BUF_free(oldbuf);
        free(oldbuf);

        oldbuf = LL_first(cst->curPath->bufList);
    }

    LL_destroy(&cst->curPath->bufList);
    free(cst->curPath);
    cst->curPath = NULL;
    cst->curPathId = -1;
}

static void
IPCCST_cleanup(struct IPC_cstate *cst)
{

    DPRINT(DBG_IPC, "cst: Cleaning client", NO_VARG);

    if (cst->curPath != NULL) {
        IPCCST_clearCurPath(cst);
        cst->curPath = NULL;
    }

    if (cst->sock >= 0) {
        close(cst->sock);
        cst->sock = -1;
    }

    if (cst->open_pkt != NULL) {
        IPCPacket_free(cst->open_pkt);
        cst->open_pkt = NULL;
    }

    cst->state = IPCST_CLEAN;
}

static void
IPCCST_queueEvent(IPC *handle, int pidx) 
{
    struct IPC_event    *evt;
    struct IPC_cstate   *cst;

    cst = &(handle->cliConns[pidx]);

    /* setup the event */
    evt = emalloc(sizeof(struct IPC_event));
    evt->addr.proc = IPC_indexToPA(pidx);
    evt->addr.channel = IPCAC_GENERAL;
    evt->pid = cst->pid;
    evt->data = cst->exitStatus;
    if ( (cst->exitState & CST_PROTSHUT) != 0) {
        evt->type = IPCEVT_IMSTOP;
    }
    else {
        evt->type = IPCEVT_IMDEATH;
    }

    DPRINT(DBG_IPC, "cst: Queueing Event exit for client %d: type = %s",
            evt->addr.proc, 
            evt->type == IPCEVT_IMSTOP ? "IMSTOP" : "IMDEATH");

    /* put the event in the queue */
    IPCLL_queueEvent(handle, evt);
}

static bool
IPCCST_canQueue(struct IPC_cstate *cst)
{
    return (cst->exitState & CST_CANCLEAN) == CST_CANCLEAN;
}

