/**
 * @file ipc_cstate.h
 *
 * @brief Functions/Constants for working with IPC clients.
 * @par
 * This does not contain the actual IPC_cstate data structure becasue
 * that would cause circular include problems.
 *
 * $Id: ipc_cstate.h,v 1.4.2.1 2003/04/30 20:03:19 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jun 26 18:20:43 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IPC_IPC_CSTATE_H
#define IPC_IPC_CSTATE_H

#include "ipc.h"
#include "ipc_pkt.h"
#include "ipcll.h"

/**@defgroup ipc_cstate IPC Client State Handling
 * @ingroup ipc
 */

/**@defgroup ipc_states IPC Client States
 * @ingroup ipc_cstate
 */

/**@addtogroup ipc_states
 * @{
 */
#define IPCST_FREE      0x00    /**< No Client Attached */
#define IPCST_EXIT      0x01    /**< Protocol shutdown complete (Client) */
#define IPCST_CLEAN     0x02    /**< cleaned, with queue messages */
#define IPCST_BRKPIPE   0x03    /**< Broken Pipe (Client) */

#define IPCST_NEW       0x10    /**< just accepted on */
#define IPCST_GIVENADDR 0x20    /**< just set IPCC_PADDR */
#define IPCST_READY     0x30    /**< ready to be used */
#define IPCST_SENTEXIT  0x40    /**< IPCC_IMEXIT was sent, wait for response*/

/**
 * Check to see if this state is an active one.
 * @param st the state to check
 * @return TRUE if it's active, FALSE if not.
 */
#define IPCST_isActive(st)  (((st) & 0xF0) != 0) 

/**
 * Check to see if this slot in handle->cliConns is available.
 * @param st the state to check
 * @return TRUE if it's available, FALSE if not.
 */
#define IPCST_isFree(st)    ((st) == IPCST_FREE)

/*@}*/

/**@addtogroup ipc_cstate 
 * @{
 */

/** A struct to hold client state.  */
struct IPC_cstate {
    int         sock;       /**< socket to that client */
    pid_t       pid;        /**< Cilent's process ID */
    int         state;      /**< current client state (\ref ipc_states) */

    uint32_t            curPathId;  /**< \c path_id last path message */
    struct IPC_pathList *curPath;   /**< incomming partial path */
    struct IPC_packet   *open_pkt;  /**< have headers but no data yet */

    int         exitStatus;         /**< Exit stats from waitpid() */
    uint32_t    exitState;          /**< Partial exit state */
};

/**
 * Cleanup for an IPC_cstate object that is done with it's socket.  This
 * should only be called once the pipe is clear of incomming data.  Such
 * an event occures when either an IMEXIT packet comes in, or EOF.  
 *
 * This will cleanup the client state, and close the socket.  The clean 
 * up will leave the client state in a used-but-clean state until the 
 * child is reaped (which it may already have been).
 *
 * If this is the last exit condition that needs to be meet then the
 * exit event will be queued (either IMDEATH or IMSTOP).
 *
 * The \c protoshut parameter if TRUE means the pipe was clear via a proper
 * protocol shutdown (i.e. IMEXIT, ACK). If FALSE it means that no
 * proper protocol shutdown was done (i.e. EOF).
 *
 * @param handle The IPC handle that we are working on.
 * @param sockid The index into IPC::cliConns for the target IPC_cstate
 * @param protoshut Was a proper protocol shutdown done?
 */
void IPCCST_pipeClear(IPC *handle, int sockid, bool protoshut);

/**
 * Cleanup for an IPC_cstate object that has been reaped.  This should
 * be called after waitpid() has been called on the cstate.
 *
 * If this is the last exit condition that needs to be meet then the
 * exit event will be queued (either IMDEATH or IMSTOP).
 *
 * @param handle The IPC handle that we are working on.
 * @param pidx The index into IPC::cliConns for the target IPC_cstate
 * @param status The Exit status of the process.
 */
void IPCCST_reaped(IPC *handle, int pidx, int status);

/**
 * Initilize an IPC_cstate structure.  If \c sock==-1 then this is
 * assumed to be the first initilization, so a lot less is done.
 *
 * @param handle The IPC handle to work on
 * @param pidx The proc index of the IPC_cstate to init.
 * @param sock The socket that is attached to this client.
 */
void IPCCST_init(IPC *handle, int pidx, int sock);

/**
 * Free an IPC_cstate object in the given IPC handle.
 * This finishes elminating all the associated state in the handle.
 * Normally this requires that the client state be cleaned and therefore
 * ready to free.  However, if \c force==TRUE then the client state will
 * be cleanedup as part of the free.  
 * @warning setting \c force==TRUE is not safe if the IPC handle will be
 * used after this.  It will be possible for the client's process
 * address to be reused before all this client's messages have left the
 * queue.  In short do not use \c force==TRUE unless you are destroying
 * the handle.
 *
 * @param handle The IPC handle to work on.
 * @param pidx The proc index of the IPC_cstate object to clear.
 * @param force Cleanup and free regardless of current state.
 *
 * @pre force || IPCCST_canFree() == TRUE
 */
void IPCCST_free(IPC *handle, uint32_t pidx, bool force);

/**
 * Destroy a partially constructed IPC_cstate::curPath structure.
 *
 * @param cst The IPC_cstate to clean up.
 */
void IPCCST_clearCurPath(struct IPC_cstate *cst);

/*@}*/
#endif /* IPC_IPC_CSTATE_H */
