/**
 * @file ipc_ctl.c
 *
 * @brief Implement the IPC control functions.
 * @par
 *
 *
 * $Id: ipc_ctl.c,v 1.12.2.6 2003/04/30 20:03:19 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Mar 15 15:21:45 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <unistd.h>
#include <stdlib.h>

#include "ipc_private.h"
#include "ipc.h"
#include "ipc_ctl.h"
#include "ipc_err.h"
#include "ipc_pkt.h"
#include "ipcll.h"
#include "ipc_cstate.h"

#include "util/list.h"

#include "error.h"
#include "debug.h"
#include "test.h"

static char const rcsid[] 
    = "$Id: ipc_ctl.c,v 1.12.2.6 2003/04/30 20:03:19 devink Exp $";

/**
 * Get the last and current Path Id from a packet containing a path element.
 *
 * @param pkt The packet to pull data out of.
 * @param lastid Where the \c last_id field will be put.
 * @param curid Where the \c path_id field will be put.
 * @return TRUE on success, FALSE otherwise.
 */
static bool getPathIds(struct IPC_packet *pkt, 
                          uint32_t *lastid, uint32_t *curid);

IPC_readState
IPCLL_processControl(IPC *handle, struct IPC_packet *pkt, int pidx)
{

    struct IPC_control  ctrl;
    struct IPC_cstate   *cst;
    IPC_readState       rdState;
    uint32_t            cliPaddr;


    IPCLL_getControlBuf(&ctrl, pkt->data);

    if (pidx == IPC_MAINSOCK) {
        cst = NULL;
        cliPaddr = IPCAP_UNSET;
    }
    else {
        cst = &(handle->cliConns[pidx]);
        cliPaddr = IPC_indexToPA(pidx);
    }

    rdState = IPCRS_SUCCESS;

    switch (ctrl.type) {
        case IPCC_PATHST:
        {
            DPRINT(DBG_IPC, "recv: PATHST: %d -> %d (%d)", 
                    pkt->sp_addr, pkt->dp_addr, handle->paddr);

            /* setup for path factory mode */
            if (cst == NULL) {
                handle->errcode = IPCE_PROTERR;
                rdState = IPCRS_ERROR;      
                break;
            }
            if (cst->state !=  IPCST_READY) {
                handle->errcode = IPCE_PROTERR;
                rdState = IPCRS_ERROR;
                break;
            }

            if (ctrl.datalen < 1) {
                /* error, skip starting the path */
                handle->errcode = IPCE_BADCTRLFMT;
                rdState = IPCRS_ERROR;
                break;
            }

            if (cst->curPath != NULL) {
                handle->errcode = IPCE_TWOPATHS;
                rdState = IPCRS_ERROR;
                break;
            }
            cst->curPath = emalloc(sizeof(struct IPC_pathList));
            cst->curPath->bufList = LL_create();
            cst->curPath->src.proc = pkt->sp_addr;
            cst->curPath->src.channel = pkt->sc_addr;

            cst->curPathId = ctrl.data[0];
            DPRINT(DBG_IPC, "recv: PATHST: new path, id %d", cst->curPathId);

            /* keep reading, as there is probably more */
            rdState = IPCRS_AGAIN;

            break;
        }
        case IPCC_PATHEND:
        {
            DPRINT(DBG_IPC, "recv: PATHEND: %d -> %d (%d)", 
                    pkt->sp_addr, pkt->dp_addr, handle->paddr);
            /* save path */
            if (cst == NULL) {
                handle->errcode = IPCE_PROTERR;
                rdState = IPCRS_ERROR;      
                break;  /* nothing we can do */
            }
            if (cst->state != IPCST_READY) {
                handle->errcode = IPCE_PROTERR;
                rdState = IPCRS_ERROR;
                break;
            }

            DPRINT(DBG_IPC, "recv: PATHEND: ending path %d", ctrl.data[0]);

            /* make sure we have the data we need */
            if (ctrl.datalen < 1) {
                handle->errcode = IPCE_BADCTRLFMT;
                rdState = IPCRS_ERROR;
            }
            /* final sanity check */
            else if (cst->curPathId != ctrl.data[0]) {
                handle->errcode = IPCE_BADPATHFMT;
                rdState = IPCRS_ERROR;
            }
            /* okay, all checks out */
            else {
                DPRINT(DBG_IPC, "recv: PATHEND: path %d appeneded to queue",
                        cst->curPathId);

                /* Append it to the queue */
                IPCLL_queuePath(handle, cst->curPath);

                /* Mark the current path as unused */
                cst->curPath = NULL;
                cst->curPathId = -1;
            }

            /* free the path data on error */
            if (rdState == IPCRS_ERROR) {
                IPCCST_clearCurPath(cst);
            }

            /* Post condition checks */
            assert(cst->curPathId == -1);
            assert(cst->curPath == NULL);

            /* we are done here */
            break;

        }
        case IPCC_PATHCNCL:
        {
            DPRINT(DBG_IPC, "recv: PATHCNCL: %d -> %d (%d)", 
                    pkt->sp_addr, pkt->dp_addr, handle->paddr);
            if (cst == NULL) {
                handle->errcode = IPCE_PROTERR;
                rdState = IPCRS_ERROR;
                break;
            }
            if (cst->state != IPCST_READY)  {
                handle->errcode = IPCE_PROTERR;
                rdState = IPCRS_ERROR;
                break;
            }

            IPCCST_clearCurPath(cst);

            break;
        }
        case IPCC_IMEXIT:
        {
            /* start exit procedure, run in controller */
            struct IPC_packet   spkt;   /* packet to send */
            struct IPC_control  sctl;   /* Control data to send */
            struct IPC_buffer    sbuf;   /* buffer to send */

            DPRINT(DBG_IPC, "recv: IMEXIT: %d -> %d (%d)", 
                    pkt->sp_addr, pkt->dp_addr, handle->paddr);

            BUF_init(&sbuf);
            IPCPacket_init(&spkt, handle, IPCT_CTRL, IPCAC_GENERAL);
            IPCPacket_setDstAddr(&spkt, cliPaddr, IPCAC_GENERAL);

            sctl.type = IPCC_ACK;
            sctl.datalen = 0;
            sctl.data = NULL;

            if (! IPCLL_makeControlBuf(&sbuf, &sctl) ) {
                handle->errcode = IPCE_ERROR;
                rdState = IPCRS_ERROR;
                BUF_free(&sbuf);
                break;
            }

            IPCPacket_setData(&spkt, &sbuf); 

            if (! IPCLL_send(handle, &spkt) ) {
                rdState = IPCRS_ERROR;
                BUF_free(&sbuf);
                break;
            }

            /* if we don't make it here then we have problems, like 
             * a client who wants to exit really badly.
             */

            /* mark the pipe as clear for this client */
            IPCCST_pipeClear(handle, pidx, TRUE);

            BUF_free(&sbuf);

            DPRINT(DBG_IPC, "recv: IMEXIT: handler done", NO_VARG);

            break;
        }
        case IPCC_IMSTART:
        {
            struct IPC_packet   spkt;   /* packet to send */
            struct IPC_control  sctl;   /* Control data to send */
            struct IPC_buffer    sbuf;   /* buffer to send */

            DPRINT(DBG_IPC, "recv: IMSTART: %d -> %d (%d)", 
                    pkt->sp_addr, pkt->dp_addr, handle->paddr);

            /* start init procedure */
            if (cst->state != IPCST_NEW) {
                handle->errcode = IPCE_INITFAIL;
                rdState = IPCRS_ERROR;
                break;  /* nothing more can be done */
            }

            BUF_init(&sbuf);
            IPCPacket_init(&spkt, handle, IPCT_CTRL, IPCAC_GENERAL); 
            IPCPacket_setDstAddr(&spkt, cliPaddr, IPCAC_GENERAL);

            sctl.type = IPCC_PADDR;
            sctl.datalen = 1;
            sctl.data = &cliPaddr;

            if (! IPCLL_makeControlBuf(&sbuf, &sctl) ) {
                handle->errcode = IPCE_ERROR;
                rdState = IPCRS_ERROR;
                BUF_free(&sbuf);
                break;
            }

            IPCPacket_setData(&spkt, &sbuf); 

            if (! IPCLL_send(handle, &spkt) ) {
                rdState = IPCRS_ERROR;
                BUF_free(&sbuf);
                break;
            }

            cst->state = IPCST_GIVENADDR;
            cst->pid = (pid_t)(*ctrl.data);

            /* cleanup */
            BUF_free(&sbuf);
            break;
        }
        case IPCC_PADDR:
        {
            struct IPC_packet   spkt;
            struct IPC_buffer    sbuf;
            struct IPC_control  sctl;

            DPRINT(DBG_IPC, "recv: PADDR: %d -> %d (%d)", 
                    pkt->sp_addr, pkt->dp_addr, handle->paddr);

            /* process address, this is done in the input modules */
            if (handle->mode != IPCM_CLIENT) {
                handle->errcode = IPCE_PROTERR;
                rdState = IPCRS_ERROR;
                break;
            }

            if (ctrl.datalen < 1) {
                handle->errcode = IPCE_BADCTRLFMT;
                rdState = IPCRS_ERROR;
                break;
            }

            if (handle->mainState.state != IPCST_NEW) {
                /* just drop it */
                break;
            }

            handle->paddr = ctrl.data[0];

            /* incase something below needs this set correctly */
            handle->mainState.state = IPCST_GIVENADDR;

            /* now we send an ack */
            BUF_init(&sbuf);

            IPCPacket_init(&spkt, handle, IPCT_CTRL, IPCAC_GENERAL);
            IPCPacket_setDstAddr(&spkt, IPCAP_CTLR, IPCAC_GENERAL);

            sctl.type = IPCC_ACK;
            sctl.datalen = 1;
            sctl.data = &handle->paddr;

            if (! IPCLL_makeControlBuf(&sbuf, &sctl) ) {
                handle->errcode = IPCE_ERROR;
                rdState = IPCRS_ERROR;
                BUF_free(&sbuf);
                break;
            }

            IPCPacket_setData(&spkt, &sbuf);

            if (! IPCLL_send(handle, &spkt) ) {
                rdState = IPCRS_ERROR;
                BUF_free(&sbuf);
                break;  /* in case more is ever added below this */
            }

            handle->mainState.state = IPCST_READY;

            BUF_free(&sbuf);

            break;
        }
        case IPCC_ACK:
        {
            DPRINT(DBG_IPC, "recv: ACK: %d -> %d (%d)",
                    pkt->sp_addr, pkt->dp_addr, handle->paddr);

            /* ack */
            if (handle->mode == IPCM_CLIENT) {
                if (handle->mainState.state == IPCST_SENTEXIT) {
                    handle->mainState.state = IPCST_EXIT;
                }
                /* else just drop it on the floor */
            }
            else if (handle->mode == IPCM_SERVER) {
                if (cst->state == IPCST_GIVENADDR) {
                    struct IPC_event *evt;

                    cst->state = IPCST_READY;

                    evt = emalloc(sizeof(struct IPC_event));
                    evt->type = IPCEVT_IMSTART;
                    evt->addr.proc = cliPaddr;
                    evt->addr.channel = pkt->sc_addr;
                    evt->pid = cst->pid;
                    IPCLL_queueEvent(handle, evt);
                }
                /* else just drop it on the floor */
            }

            break;
        }
        default:
        {
            throw_fatal("Illegal Control Packet Sub-Type");
            rdState = IPCRS_ERROR;
        }

    }


    /* we don't need the control data either */
    free(ctrl.data);

    return rdState;
}

IPC_readState
IPCLL_processPath(IPC *handle, struct IPC_packet *pkt, int pidx)
{
    struct IPC_cstate   *cst;
    uint32_t            lastid, curid;

    assert(pkt != NULL);
    assert(handle != NULL);

    cst = &(handle->cliConns[pidx]);

    if (! getPathIds(pkt, &lastid, &curid) || lastid != cst->curPathId) {
        handle->errcode = IPCE_BADPATHFMT;
        return IPCRS_ERROR;
    }

    LL_append(cst->curPath->bufList, pkt->data);
    cst->curPathId = curid;

    pkt->data = NULL;   /* make sure this does stay */

    return IPCRS_SUCCESS;
}

bool
IPCLL_makeControlBuf(struct IPC_buffer *buf, struct IPC_control *ctl)
{
    int     i;

    assert(buf != NULL);
    assert(ctl != NULL);

    buf->size = 2*BUF_INTSZ + ctl->datalen * BUF_INTSZ;
    BUF_alloc(buf);

    if (! BUF_setInteger(buf, ctl->type) ) goto error;
    if (! BUF_setInteger(buf, ctl->datalen) ) goto error;

    for (i = 0; i < ctl->datalen; i++) {
        if (! BUF_setInteger(buf, ctl->data[i]) ) goto error;
    }
    if (i < ctl->datalen) goto error;

    /* if we are here then it worked */
    return TRUE;

    /* Cleanup after an error, then return FALSE */
    error:
        BUF_free(buf);
        return FALSE;
}

bool
IPCLL_getControlBuf(struct IPC_control *ctl, struct IPC_buffer *buf)
{
    uint32_t    tmp;
    bool        success;
    int         i;
    IPC_bufMark start;

    assert(buf != NULL);
    assert(ctl != NULL);

    start = BUF_markPos(buf);

    success = FALSE;

    if (! BUF_getInteger(buf, &tmp) ) goto error;
    ctl->type = tmp;

    if (! BUF_getInteger(buf, &(ctl->datalen) ) ) goto error;

    /**@todo check ctl fields here for sanity*/


    ctl->data = emalloc(ctl->datalen * sizeof(uint32_t));
    memset(ctl->data, 0, ctl->datalen);
    for (i = 0; i < ctl->datalen; i++) {
        if (! BUF_getInteger(buf, &(ctl->data[i]) ) ) goto error;
    }
    if (i < ctl->datalen) goto error;

    /* if we are here then all is well */
    success = TRUE;
    goto end;       /* jump to the return */

    /* Cleanup after an error and exit FALSE */
    error:
        success = FALSE;
        BUF_resetToPos(buf, start);

    end:
        return success;
}



static bool
getPathIds(struct IPC_packet *pkt, uint32_t *lastid, uint32_t *curid)
{
    IPC_bufMark sPos;
    bool        success = FALSE;

    if (pkt->type != IPCT_PATH) {
        return -1;
    }

    sPos = BUF_markPos(pkt->data);

    /* last_id is the first field */
    if (lastid != NULL && !BUF_getInteger(pkt->data, lastid) ) goto cleanup;
    if (curid != NULL && !BUF_getInteger(pkt->data, curid) ) goto cleanup;

    success = TRUE;

    /* general Cleanup and return code */
    cleanup:    
        /* we just wanted to _peek_ at the buffer, not change our
         * location in it.
         */
        BUF_resetToPos(pkt->data, sPos);
        return success;
}



