/**
 * @file ipc_ctl.h
 *
 * @brief This defines data structions and functions for dealing with
 * control packets and channel control (state info, etc).
 * @par
 *
 * @internal
 *
 *
 * $Id: ipc_ctl.h,v 1.7.2.2 2003/04/30 20:03:19 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Mar 15 15:13:40 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef IPC_IPC_CTL_H
#define IPC_IPC_CTL_H

#include "ipc_types.h"
#include "buffer.h"
#include "ipc_pkt.h"

/**@defgroup ipc_ctl IPC Controll Packet Handling
 * @ingroup ipc
 */

/**@defgroup ipc_ctltypes IPC Control Message Type 
 * @ingroup ipc_ctl
 */

/** @name Control Message Types
 * @ingroup ipc_ctltypes
 * @{
 */
typedef enum 
{

    IPCC_IMEXIT,    /**< Input Module exit event */
    IPCC_IMSTART,   /**< Input Madule start event */
    IPCC_PADDR,     /**< Providing a process address */
    IPCC_ACK,       /**< ACKnowlage last packet */
    IPCC_PATHST,    /**< Start of a series of path messages */
    IPCC_PATHEND,   /**< End of a series of path messages */
    IPCC_PATHCNCL,  /**< Cancle Path series (error while sending) */

    /** number of control types. mostly to allow ',' after all control types */
    IPCC_COUNT      
} IPC_ctlTypes;
/*@}*/

/** 
 * This structure contains the information for a control packet.
 * Control messages are sent via special functions in the interface,
 * but can be received like most other messages
 * @ingroup ipc_ctl
 */
struct IPC_control
{
    IPC_ctlTypes    type;       /**< \ref ipc_ctltypes */
    uint32_t        datalen;    /**< length of data */
    uint32_t        *data;      /**< actual data */
}; 


/**@name IPC Control Packet Handling Functions
 * @ingroup ipc_ctl
 * @{
 */

/** 
 * process control messages.
 *
 * @param handle The handle this control message applies to.
 * @param pkt The packet containing the control message.
 * @param pidx The proc index of the sender.
 * @return an IPC_readState value.
 *
 * @post Done with pkt, it can be free()ed
 */
IPC_readState IPCLL_processControl(IPC *handle, struct IPC_packet *pkt, 
                                   int pidx);
/** 
 * Create a control packet buffer.  
 * This works by converting an IPC_control to an IPC_buffer.  It will take 
 * care of allocating space in \c buf.  Do not try to allocate space 
 * in buf first.
 *
 * @param buf The IPC_buffer to store the data in
 * @param ctl The data to use in the control packet
 * @return TRUE on success, FALSE otheriwes.
 *
 * @pre buf != NULL;
 * @pre ctl != NULL;
 */
bool IPCLL_makeControlBuf(struct IPC_buffer *buf, struct IPC_control *ctl);

/** Convert a buffer to a control packet.  
 * @param ctl The place to put the retrieved control packet data
 * @param buf The buffer to parse data out of.
 * @return TRUE on success, FALSE otherwise.
 */
bool IPCLL_getControlBuf(struct IPC_control *ctl, struct IPC_buffer *buf);

/**
 * Process a PATH packet.  This is here, because sending paths requires
 * control messages.  So putting it here keeps it all in the same palce.
 *
 * When this returns there is no need to keep pkt around any longer.
 *
 * @param handle The IPC handle
 * @param pkt The path pkt
 * @param pidx The proc index (to get the client info).
 *
 * @return an IPC_readState value.
 *
 * @post Done with pkt, it can be free()ed
 */
IPC_readState IPCLL_processPath(IPC *handle, struct IPC_packet *pkt, int pidx);

/*@}*/

#endif /* IPC_IPC_CTL_H */
