/**
 * @file ipc_lock.c
 *
 * @brief Functions for handling IPC lock passing.
 *
 * $Id: ipc_lock.c,v 1.4.2.1 2003/04/30 20:03:20 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Jul 12 17:46:31 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ipc_private.h"
#include "ipc.h"
#include "ipc_err.h"
#include "ipc_lock.h"
#include "ipc_pkt.h"
#include "buffer.h"
#include "ipcll.h"

#include "debug.h"
#include "error.h"

static char const rcsid[] = "$Id: ipc_lock.c,v 1.4.2.1 2003/04/30 20:03:20 devink Exp $";

bool
IPCLock_send(IPC *handle, IPC_lockType type, IPC_lock lk, uint32_t dst)
{
    struct IPC_packet   pkt;
    struct IPC_buffer   buf;
    bool                success;

    assert(handle != NULL);

    success = FALSE;

    /* build a lock packet w/ type == IPCLT_REQCNCL and send it */
    IPCPacket_init(&pkt, handle, IPCT_LOCK, IPCAC_GENERAL);
    IPCPacket_setDstAddr(&pkt, dst, IPCAC_GENERAL);
    BUF_init(&buf);

    buf.size = IPCT_LOCK_LEN;
    BUF_alloc(&buf);
    
    if (! BUF_setInteger(&buf, type) ) goto error;
    if (! BUF_setInteger(&buf, lk) ) goto error;

    IPCPacket_setData(&pkt, &buf);
    if (! IPCLL_send(handle, &pkt) ) goto error;

    success = TRUE;

    /* Cleanup and return */
error:
    BUF_free(&buf);

    return success;
}

bool
IPCLock_recv(IPC *handle, struct IPC_addr *src, IPC_lockType *ty, IPC_lock *lk)
{
    struct IPC_packet   *pkt;
    bool                success;

    assert(handle != NULL);
    assert(ty != NULL);
    assert(lk != NULL);

    success = FALSE;

    pkt = IPCLL_recvPacket(handle, IPCT_LOCK);
    if (pkt == NULL) return FALSE;  /* errcode set by IPCLL_recvPacket() */

    if (! BUF_getInteger(pkt->data, ty)) goto cleanup;
    if (! BUF_getInteger(pkt->data, lk)) goto cleanup;

    if (*ty < 0 || *ty >= IPCLT_COUNT) {
        handle->errcode = IPCE_BADFMT;
        goto cleanup;
    }

    if (*lk < 0 || *lk >= IPCL_COUNT) {
        handle->errcode = IPCE_BADFMT;
        goto cleanup;
    }

    if (src != NULL) {
        src->proc = pkt->sp_addr;
        src->channel = pkt->sc_addr;
    }

    success = TRUE;
    
    /* Cleanup and return */
cleanup:
    IPCPacket_free(pkt);

    return success;
}
