/**
 * @file ipc_pkt.c
 *
 * @brief Implementation of IPC packet manipulation functions.
 * @par
 *
 *
 * $Id: ipc_pkt.c,v 1.6.2.2 2003/04/30 20:03:20 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Mar 29 16:30:54 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>

#include "ipc_private.h"
#include "ipc.h"
#include "ipc_pkt.h"
#include "ipc_err.h"
#include "buffer.h"

#include "error.h"
#include "debug.h"

static char const rcsid[] = "$Id: ipc_pkt.c,v 1.6.2.2 2003/04/30 20:03:20 devink Exp $";

struct IPC_packet*
IPCPacket_new(IPC *handle, IPC_type type, uint32_t schan)
{
    struct IPC_packet   *pkt;

    pkt = emalloc(sizeof(struct IPC_packet));

    IPCPacket_init(pkt, handle, type, schan); 

    pkt->data = emalloc(sizeof(struct IPC_buffer));
    BUF_init(pkt->data);

    return pkt;
}

struct IPC_packet*
IPCPacket_fromBuffer(IPC *handle, struct IPC_buffer *buf)
{
    struct IPC_packet   *pkt = NULL;
    uint8_t             majVer, minVer;
    uint32_t            val;

    assert(handle != NULL);
    assert(buf != NULL);


    pkt = emalloc(sizeof(struct IPC_packet));
    memset(pkt, 0, sizeof(struct IPC_packet));

    if (! BUF_getByte(buf, &majVer) ) goto error;
    if (! BUF_getByte(buf, &minVer) ) goto error;

    /* consistency check */
    if ( ((majVer << 8) | minVer) != IPC_PROTOVER) {
        handle->errcode = IPCE_BADVER;
        DPRINT(DBG_IPC, "bad protocol version (%d, %d)", majVer, minVer);
        goto error;
    }
    pkt->majorVer = majVer;
    pkt->minorVer = minVer;

    if (! BUF_getInteger(buf, &val) ) goto error;
    pkt->sp_addr = val;

    if (! BUF_getInteger(buf, &val) ) goto error;
    pkt->sc_addr = val;

    if (! BUF_getInteger(buf, &val) ) goto error;
    pkt->dp_addr = val;

    if (! BUF_getInteger(buf, &val) ) goto error;
    pkt->dc_addr = val;

    if (! BUF_getInteger(buf, &val) ) goto error;
    pkt->type = (int)val;

    if (! BUF_getInteger(buf, &val) ) goto error;
    pkt->datalen = val;


    pkt->data = emalloc(sizeof(struct IPC_buffer));
    BUF_init(pkt->data);

    return pkt;

    /* Cleanup after an error and return */
    error:
        /* make sure some error code was set */
        if (handle->errcode == IPCE_SUCCESS) {
            handle->errcode = IPCE_BADFMT;
        }

        free(pkt);

        return NULL;
}

void
IPCPacket_init(struct IPC_packet *pkt, IPC *handle, IPC_type ty, uint32_t sc)
{
    assert(pkt != NULL);
    assert(handle != NULL);

    memset(pkt, 0, sizeof(struct IPC_packet));

    pkt->majorVer = (IPC_PROTOVER >> 8) & 0xFF;
    pkt->minorVer = IPC_PROTOVER & 0xFF;

    pkt->sp_addr = handle->paddr;
    pkt->sc_addr = sc;
    pkt->dp_addr = IPCAP_UNSET;
    pkt->dc_addr = IPCAC_GENERAL;

    pkt->type = ty;
}

void
IPCPacket_free(struct IPC_packet *pkt)
{
    if (pkt->data != NULL) {
        BUF_free(pkt->data);
        free(pkt->data);
    }

    if (pkt != NULL) {
        free(pkt);
    }
}

void
IPCPacket_setDstAddr(struct IPC_packet *pkt, uint32_t proc, uint32_t chan)
{
    assert(pkt != NULL);

    pkt->dp_addr = proc;
    pkt->dc_addr = chan;
}

void
IPCPacket_setData(struct IPC_packet *pkt, struct IPC_buffer *buf)
{

    assert(pkt != NULL);
    assert(buf != NULL);

    pkt->data = buf;
    pkt->datalen = buf->pos - buf->data;
}

bool
IPCPacket_toBuffer(struct IPC_buffer *buf, struct IPC_packet *pkt)
{
    assert(buf != NULL);
    assert(pkt != NULL);

    buf->size = IPCHEADER_LEN;
    BUF_alloc(buf);

    if (! BUF_setByte(buf, pkt->majorVer) ) goto error;
    if (! BUF_setByte(buf, pkt->minorVer) ) goto error;

    if (! BUF_setInteger(buf, pkt->sp_addr) ) goto error;
    if (! BUF_setInteger(buf, pkt->sc_addr) ) goto error;
    if (! BUF_setInteger(buf, pkt->dp_addr) ) goto error;
    if (! BUF_setInteger(buf, pkt->dc_addr) ) goto error;

    if (! BUF_setInteger(buf, pkt->type) ) goto error;
    if (! BUF_setInteger(buf, pkt->datalen) ) goto error;

    /* If we make it here then all is well */
    return TRUE;

    /* post-error cleanup and return */
    error:
        BUF_free(buf);
        return FALSE;
}


void IPCPacket_ntoh(struct IPC_packet *pkt)
{
    /*
    pkt->sp_addr = ntohl(pkt->sp_addr);
    pkt->sc_addr = ntohl(pkt->sc_addr);
    pkt->dp_addr = ntohl(pkt->dp_addr);
    pkt->dc_addr = ntohl(pkt->dc_addr);

    pkt->type = ntohl(pkt->type);
    pkt->datalen = ntohl(pkt->datalen);
    */
}
