/**
 * @file ipc_pkt.h
 *
 * @brief Functions to manipulate IPC Packets.  
 * @par
 * These mostly deal with the headers, but it includes things like
 * packet size and other such goodies. 
 *
 * $Id: ipc_pkt.h,v 1.6.2.1 2003/04/30 20:03:20 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Mar 15 15:11:32 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef IPC_IPC_PKT_H
#define IPC_IPC_PKT_H

#ifdef HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include "ipc.h"
#include "buffer.h"

#include "types.h"

/**@defgroup ipc_pkt IPC Packet Manipulation Functions
 * @ingroup ipc
 */

/** 
 * Current IPC protocol version. The high byte is the major version,
 * the low byte is the minor version.
 * @ingroup ipc
 */
#define IPC_PROTOVER    0x0100


/**@name IPC Packet Lengths.  
 * @note These are the fixed size portion of the
 * packet only, and any variable length fields need to be added in
 * later.
 * @ingroup ipc_pkt
 * @{
 */

/** Packet Header length.  This has to be known, using sizeof will not
 * work because there may be padding added to the struct to allow for
 * data alignment. 
 */
#define IPCHEADER_LEN   (sizeof(uint32_t)*6 + sizeof(uint8_t)*2)

/** Syslog Message Length (statically sized fields only) */
#define IPCT_MSG_LEN    (5*BUF_INTSZ + 2*BUF_BYTESZ + MSGTAG_LEN \
                        + NA_MAX_BUFLEN+BUF_TIMESZ)

/** Path Message Length (statically sized fields only) */
#define IPCT_PATH_LEN   (5*BUF_INTSZ + 2*NA_MAX_BUFLEN)

/** Event Packet Length (only has static sized fields) */
#define IPCT_EVENT_LEN  (3*BUF_INTSZ)

/** Lock Packet Length (only has static sized fields) */
#define IPCT_LOCK_LEN   (2*BUF_INTSZ)
/*@}*/

/** 
 * A struct for the header information 
 * @ingroup ipc_pkt
 */
struct IPC_packet {
    uint8_t     majorVer;   /**< Protocol Major Version */
    uint8_t     minorVer;   /**< Protocol Minor Version */
    uint32_t    sp_addr;    /**< source process address */
    uint32_t    sc_addr;    /**< source channle address */
    uint32_t    dp_addr;    /**< destination process address */
    uint32_t    dc_addr;    /**< destination channel address */
    int32_t     type;       /**< packet type (\ref ipc_msgtypes) */
    uint32_t    datalen;    /**< Length of the data fields */

    struct IPC_buffer    *data;      /**< data fields */
};

/**@name IPC_packet manipulation functions
 * @ingroup ipc_pkt
 * @{
 */

/** 
 * Get an IPC_packet that can be returned by one of the low-level
 * functions.
 *
 * @param handle The IPC handle for this connection.
 * @param type The packet type.
 * @param schan The source channel address for this packet.
 * @return a pointer to the new IPC_packet.
 *
 * @pre handle != NULL
 */
struct IPC_packet* IPCPacket_new(IPC *handle, IPC_type type, uint32_t schan);

/** 
 * Get an IPC_packet that can be returned by one of the low-level
 * functions.  This packet must be freed via IPCPacket_free.  Further,
 * the contents of this packet come from the provided IPC_buffer.
 *
 * @param handle The IPC handle for this connection.
 * @param buf The buffer to get data from.
 * @return a pointe to the new IPC_packet, or NULL on error.
 *
 * @pre handle != NULL
 * @pre buf != NULL
 */
struct IPC_packet* IPCPacket_fromBuffer(IPC *handle, struct IPC_buffer *buf);

/** 
 * Initilize an IPC_packet.  
 * This serves the same purpose as IPCPacket_new(), except that it 
 * doesn't malloc space for the packet.
 *
 * @param pkt The packet to work on
 * @param handle The IPC handle for this connection.
 * @param type The packet type.
 * @param schan The source channel address for this packet.
 *
 * @pre pkt != NULL
 * @pre handle != NULL
 */
void IPCPacket_init(struct IPC_packet *pkt, IPC *handle, 
        IPC_type type, uint32_t schan);

/** 
 * Free an IPC_packet that was passed back by one of the low-level
 * functions.  This should only be used to free packet created via
 * IPCPacket_new()
 *
 * @param pkt The packet to free
 */
void IPCPacket_free(struct IPC_packet *pkt);

/** 
 * Set the data field of a packet.  
 * This should only be used after the buffer has been filled in with the data.
 * @param packet The IPC_packet to work on.
 * @param buf The IPC_buffer to use as the data.
 *
 * @pre packet != NULL
 */
void IPCPacket_setData(struct IPC_packet *packet, struct IPC_buffer *buf);

/** 
 * Set the destination address on a packet.
 *
 * @param packet The packet to work on.
 * @param proc The process address
 * @param chan The channel address
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre packet != NULL
 */
void IPCPacket_setDstAddr(struct IPC_packet *packet, uint32_t proc, 
                          uint32_t chan);

/**
 * Put an IPC_packet struct into an IPC_buffer.  This does not add the
 * contents of IPC_packet::data to the buffer, only the header portion.
 * The buffer will have space allocated for it.
 *
 * @param buf The buffer to put data into. 
 * @param pkt The packet to serialize
 * @return TRUE on success, FALSE otherwise.
 *
 * @pre buf != NULL
 * @pre pkt != NULL
 */
bool IPCPacket_toBuffer(struct IPC_buffer *buf, struct IPC_packet *pkt);

/**
 * Convert a packet with fields in network byte order to host byte
 * order.  Useful right after reading it off the socket.
 *
 * @param pkt The packet to work on.
 *
 * @pre pkt != NULL
 */
void IPCPacket_ntoh(struct IPC_packet *pkt);


/*@}*/

#endif /* IPC_IPC_PKT_H */

