/**
 * @file ipc_queue.c
 *
 * @brief Functions that manage the IPC packet queue.
 *
 * $Id: ipc_queue.c,v 1.3.2.1 2003/04/30 20:03:21 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jul 10 13:47:06 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>

#include "ipc_private.h"
#include "ipc.h"
#include "ipcll.h"
#include "ipc_pkt.h"
#include "ipc_err.h"

#include "error.h"
#include "debug.h"

static char const rcsid[] = "$Id: ipc_queue.c,v 1.3.2.1 2003/04/30 20:03:21 devink Exp $";

/** Data structure which is placed on the IPC queue */
struct IPC_qent
{
    IPC_type    type;   /**< dat type in this queue ent (\ref ipc_msgtypes). */

    /** The data stored in this queue element is one of these possible
     * types.
     */
    union  
    {
        struct IPC_packet   *pkt;   /**< A queued packet */
        struct IPC_event    *evt;   /**< A queued event */
        struct IPC_pathList *pth;   /**< A queued path */
    } data;
};

/**
 * Create a new IPC_qent object.
 *
 * @return A Pointer to the new IPC_qent.
 */
static struct IPC_qent *QEnt_create(void);

/**
 * Destroy an IPC_qent Object.
 *
 * @param qe The IPC_qent object to destroy.
 */
static void QEnt_destroy(struct IPC_qent *qe);

/**
 * do the work for the IPCLL_recv*() functions.  This is here so that I
 * can have special IPCLL_recv*() functions for each data type.
 *
 * As there are several priority levels to the queueing, this will look
 * for a packet of type \c type in descending proirity order.
 *
 * @param handle The IPC handle to work on.
 * @param type The type needed from the next queue entry.
 * @return The data from the queue specified by \c type
 *
 * @pre handle != NULL
 */
static struct IPC_qent* doRecv(IPC *handle, IPC_type type);

/**
 * Actual queueing function.  This does the work for the IPCLL_queue*()
 * functions.  The IPCLL_queue*() function are present for type safty
 * and because they know how to get the required information.
 *
 * There are several queues, each assigned a priority.  This will put 
 * \c data into the queue with priority \c pri.
 *
 * @param handle The IPC handle to work on.
 * @param data The data to queue.
 * @param high Does this go on the high priority queue.
 */
static void doQueue(IPC *handle, struct IPC_qent *data, IPC_queuePri pri);

void
IPCLL_queuePacket(IPC *handle, struct IPC_packet *pkt)
{
    struct IPC_qent     *ent;
    IPC_queuePri        pri;

    assert(handle != NULL);
    assert(pkt != NULL);

    ent = QEnt_create();
    ent->type = pkt->type;
    ent->data.pkt = pkt;

    /* figure out which queue this goes on, high or normal pority. */
    pri = IPCQ_NORM;
    if (pkt->type == IPCT_LOCK)
        pri = IPCQ_HIGH;

    doQueue(handle, ent, pri);
}

void
IPCLL_queuePath(IPC *handle, struct IPC_pathList *pth)
{
    struct IPC_qent     *ent;

    assert(handle != NULL);
    assert(pth != NULL);

    ent = QEnt_create();
    ent->type = IPCT_PATH;
    ent->data.pth = pth;

    doQueue(handle, ent, IPCQ_NORM);
}

void
IPCLL_queueEvent(IPC *handle, struct IPC_event *evt)
{
    struct IPC_qent     *ent;
    IPC_queuePri        pri;

    assert(handle != NULL);
    assert(evt != NULL);

    ent = QEnt_create();
    ent->type = IPCT_EVENT;
    ent->data.evt = evt;


    /* Which queue, high or normal prority */
    pri = IPCQ_NORM;
    if (evt->type == IPCEVT_IMSTART)
        pri = IPCQ_HIGH;

    DPRINT(DBG_IPC, "queueEvent: queueing event %d in queue %d from %d",
            evt->type, pri, evt->addr.proc);

    doQueue(handle, ent, pri);
}
    
/* figure out what is waiting in the input queue */
int
IPCLL_hasQueued(IPC *handle)
{
    int             rtn, i;
    struct IPC_qent *qent;

    assert(handle != NULL);

    DPRINT(DBG_IPC, "hasQueued: start", NO_VARG);

    /* Check queues from highest priority to lowest priority */
    rtn = IPCT_NONE;
    for (i = 0; i < IPCQ_COUNT; i++) {
        if (! LL_isEmpty(handle->queues[i]) ) {
            DPRINT(DBG_IPC, "hasQueued: data on queue %d", i);
            qent = LL_first(handle->queues[i]);
            rtn = qent->type;
            break;  /* found one ... we are done */
        }
    }

    DPRINT(DBG_IPC, "hasQueued: end (%d)", rtn);


    return rtn;
}

struct IPC_packet *
IPCLL_recvPacket(IPC *handle, IPC_type type)
{
    struct IPC_packet   *pkt;
    struct IPC_qent     *ent;

    assert(handle != NULL);

    pkt = NULL;

    if (type < IPCT_NONE || type >= IPCT_COUNT           /* range check */
            || type == IPCT_PATH || type == IPCT_EVENT) /* not here check */
    {
        type = IPCT_NONE;   /* don't recv anything */
    }

    ent = doRecv(handle, type);

    if (ent != NULL ) {
        pkt = ent->data.pkt;
        QEnt_destroy(ent);
    }

    return pkt;
}

struct IPC_pathList *
IPCLL_recvPath(IPC *handle)
{
    struct IPC_pathList     *pth;
    struct IPC_qent         *ent;

    assert(handle != NULL);

    pth = NULL;
    ent = doRecv(handle, IPCT_PATH);

    if (ent != NULL) {
        pth = ent->data.pth;
        QEnt_destroy(ent);
    }

    return pth;
}

struct IPC_event *
IPCLL_recvEvent(IPC *handle)
{
    struct IPC_event    *evt;
    struct IPC_qent     *ent;

    assert(handle != NULL);

    evt = NULL;
    ent = doRecv(handle, IPCT_EVENT);

    if (ent != NULL) {
        evt = ent->data.evt;
        QEnt_destroy(ent);
    }

    return evt;
}

static struct IPC_qent*
doRecv(IPC *handle, IPC_type type)
{
    struct IPC_qent     *data;
    int                 i;

    assert(handle != NULL);

    /* first we are here so we try to gather all data off the socket */
    if (! IPCLL_readAllSocks(handle, NULL, NULL) ) {
        return NULL;
    }

    /* if type is invalid then skip checking the queues. */
    if (type == IPCT_NONE) return NULL;
    assert(type > IPCT_NONE && type < IPCT_COUNT);
    
    DPRINT(DBG_IPC, "doRecv: start; looking for type %d", type);

    /* setup some defaults */
    IPCLL_clearError(handle);
    data = NULL;

    /* okay, now we need to check the wait queues, highest to lowestclear  */
    for (i = 0; data == NULL && i < IPCQ_COUNT; i++) {
        if (! LL_isEmpty(handle->queues[i]) ) {
            data = LL_first(handle->queues[i]);
            DPRINT(DBG_IPC, "doRecv: have data on queue %d, type %d", 
                    i, data->type);

            if (data->type != type) {
                data = NULL;
            }
            else if (! LL_rmFirst(handle->queues[i]) ) {
                /* XXX I'm not sure this is a good way to deal with it 
                 * no free b/c I expect the packet to be returned next time*/ 
                handle->errcode = IPCE_ERROR;
                data = NULL;
                break;  /* stop trying queues, this was an error */
            }
        }
    }

    /* set the default error code if nothing else claims an error */
    if (data == NULL && handle->errcode == IPCE_SUCCESS) {
        handle->errcode = IPCE_NOSUCHPKT;
        DPRINT(DBG_IPC, "doRecV: No packet of type %d found", type);
    }

    DPRINT(DBG_IPC, "doRecv: end; data = %p", data);
    return data;
}

static void
doQueue(IPC *handle, struct IPC_qent *data, IPC_queuePri pri)
{
    assert(handle != NULL);
    
    DPRINT(DBG_IPC, "doQueue: queueing type %d in queue %d", data->type, pri);
    LL_append(handle->queues[pri], data);
}

static struct IPC_qent *
QEnt_create(void)
{
    struct IPC_qent     *qe;

    qe = emalloc(sizeof(struct IPC_qent));
    memset(qe, 0, sizeof(struct IPC_qent));

    return qe;
}

static void
QEnt_destroy(struct IPC_qent *qe)
{
    free(qe);
}
