/**
 * @file ipc_sig.c
 *
 * @brief Implement the IPC Signal handling stuff.
 *
 * $Id: ipc_sig.c,v 1.6.2.2 2003/04/30 20:03:21 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Jun 21 17:26:18 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#include "ipc_private.h"
#include "ipc.h"
#include "ipc_sig.h"
#include "ipc_err.h"
#include "ipcll.h"
#include "ipc_ctl.h"
#include "ipc_cstate.h"

#include "log.h"
#include "error.h"

static char const rcsid[] = "$Id: ipc_sig.c,v 1.6.2.2 2003/04/30 20:03:21 devink Exp $";

/**
 * Find a IM proc id given a PID
 *
 * @param handle the IPC to look in.
 * @param pid The PID to find.
 * @return the proc index if found, < 0 if not.
 *
 * @pre handle != NULL
 */
static int findPIndex(IPC *handle, pid_t pid);

/**
 * The signal handler for the IPC layer
 *
 * @param signo The signal number.
 */
static void ipc_sighandler(int signo, siginfo_t *info, void *v);

/** Count the SIGCHLD events that have arrived since last checking. */
static uint32_t sigchld_count = 0;

/** Count the SIGPIPE events that have arrived since last checking. */
static uint32_t sigpipe_count = 0;


bool
IPCSig_setupController(IPC *handle)
{
    struct sigaction    act;

    /* we don't care if this fails, it's mostly to keep others away from
     * them.
     */
    INWLL_regSignal(SIGCHLD);
    INWLL_regSignal(SIGPIPE);

    act.sa_handler = SIG_IGN;
    
    if (sigaction(SIGPIPE, &act, NULL) < 0) {
        handle->errcode = IPCE_SIGINST;
        handle->errnocode = errno;
        return FALSE;
    }

    act.sa_sigaction = ipc_sighandler;
    act.sa_flags = SA_NOCLDSTOP | SA_SIGINFO;
    if (sigaction(SIGCHLD, &act, NULL) < 0) {
        handle->errcode = IPCE_SIGINST;
        handle->errnocode = errno;
        return FALSE;
    }

    sigchld_count = 0;
    sigpipe_count = 0;

    DPRINT(DBG_IPC, "sig: Controller Handlers installed", NO_VARG);

    return TRUE;
}

bool
IPCSig_setupInputModule(IPC *handle)
{
    struct sigaction    act;
    bool                error;

    memset(&act, 0, sizeof(act));

    error = FALSE;

    /* We are still reserving these signals for the IPC layer, but don't
     * care about them.
     */
    INWLL_regSignal(SIGPIPE);
    INWLL_regSignal(SIGCHLD);

    act.sa_sigaction = ipc_sighandler;
    act.sa_flags = SA_SIGINFO;
    if (sigaction(SIGPIPE, &act, NULL) < 0) {
        handle->errcode = IPCE_SIGINST;
        handle->errnocode = errno;
        error = TRUE;
    }

    act.sa_handler = SIG_DFL;
    act.sa_flags = 0;
    if (sigaction(SIGCHLD, &act, NULL) < 0) {
        handle->errcode = IPCE_SIGINST;
        handle->errnocode = errno;
        error = TRUE;
    }

    DPRINT(DBG_IPC, "sig: Inputmodule Handlers installed", NO_VARG);
    return !error;
}

void
IPCSig_check(IPC *handle)
{
    sigset_t            sigmask;
    int                 status;
    pid_t               pid;
    int                 pidx;
    struct IPC_event    *evt;
    bool                done;

    DPRINT(DBG_IPC, "sig_check: start", NO_VARG);

    /* block incomming signals while we clear up the dead babies */
    sigemptyset(&sigmask);
    sigaddset(&sigmask, SIGCHLD);
    sigaddset(&sigmask, SIGPIPE);
    sigprocmask(SIG_BLOCK, &sigmask, NULL);

    if (handle->mode == IPCM_SERVER) {

        /*
         * If there are any children to reap, then we need to keep
         * reaping until we run out.  Linux will not deliever signals
         * for reaped children, which means that sigchld_count may not
         * accuratly reflect the number of children which have died.
         */
        done = sigchld_count == 0;
        sigchld_count = 0;

        while (! done) {
            pid = waitpid(-1, &status, WNOHANG);
            if (pid == 0) {
                /* No more children to reap. */
                done = TRUE;
            }
            else if (pid == -1) {
                /* error, maybe
                 * The Linux man page _claims_ that waitpid() returns 0
                 * when there are no more children to reap.  However, it
                 * lies ... and waitpid() has been seen to return -1
                 * with errno == ECHILD.
                 */
                if (errno == ECHILD) 
                    done = TRUE;
                else 
                    throw_fatal("waitpid() failed in IPC layer");
            }
            else {
                /* reaped a child */
                DPRINT(DBG_IPC, "sig_check: reaped PID %d (status = %x)", 
                        pid, status);

                pidx = findPIndex(handle, pid);


                /**@bug Assuming that it is in *this* handle or is not an
                 * IPC client will cause problems if we ever tried to use
                 * more than one IPC handle in a process.
                 */
                if (pidx < 0) {
                    /* not initilized yet */
                    DPRINT(DBG_IPC, "sig_check: queuing child death", NO_VARG);

                    evt = emalloc(sizeof(struct IPC_event));
                    evt->type = IPCEVT_CHLDDEATH;
                    evt->addr.proc = IPCAP_UNSET;
                    evt->addr.channel = IPCAC_GENERAL;
                    evt->pid = pid;
                    evt->data = status;
                    IPCLL_queueEvent(handle, evt);
                }
                else {
                    /* mark that we have called waitpid() on it */
                    DPRINT(DBG_IPC, "sig_check: marking client (%d) as reaped",
                            IPC_indexToPA(pidx));

                    IPCCST_reaped(handle, pidx, status);
                }
            }
        }
    }
    else if (handle->mode == IPCM_CLIENT) {
        if (sigpipe_count > 0) {
            /* looks like our controller is dead */
            DPRINT(DBG_IPC, "sig_check: broken pipe", NO_VARG);
            IPCLL_handleBrokenPipe(handle);
        }
        sigpipe_count = 0;
    }

    /* unblock the signals now, just ignore any that came in while we
     * were busy ... we'll get them next time. 
     */
    sigprocmask(SIG_UNBLOCK, &sigmask, NULL);

    DPRINT(DBG_IPC, "sig_check: end", NO_VARG);
}
    

static int
findPIndex(IPC *handle, pid_t pid)
{
    struct IPC_cstate   *cst;
    int                 pidx;
    int                 i;

    pidx = -1;

    /* A linear search is the best solution that I have at the moment.
     * Until we see a performance problem here, using an additional data
     * strutrue to index on PID is not worth it.
     */
    for (i = 0, cst = handle->cliConns; i < IPC_MAX_IMPROCS; cst++, i++) {
        if ( (IPCST_isActive(cst->state) || cst->state == IPCST_CLEAN)
            && cst->pid == pid) 
        {
            pidx = i;
            break;
        }
    }

    return pidx;
}

static void
ipc_sighandler(int signo, siginfo_t *info, void *v)
{
    DPRINT(DBG_IPC, "signal: number = %d, from PID %d", 
            signo, info->si_pid);

    switch (signo) {
        case SIGCHLD: 
            sigchld_count++; break;

        case SIGPIPE: 
            sigpipe_count++; break;
    }
}
