/**
 * @file ipc_types.h
 *
 * @brief This file defines several of the types used in the IPC
 * libarary.
 * @par
 * This is here to reduce circular dependencies ... header files often
 * only need the types, and the source files can include the actual
 * functions as needed.  This should not ever be used outside of the IPC
 * module.
 * @internal
 *
 * $Id: ipc_types.h,v 1.4.2.2 2003/04/30 20:03:21 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Apr 26 17:59:13 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IPC_IPC_TYPES_H
#define IPC_IPC_TYPES_H

#include <sys/types.h>
#include "types.h"

#include "util/list.h"

/** 
 * The return codes for the various reading/packet processing
 * functions.  This is because such functions need to return a quad
 * state: SUCCESS, ERROR, EOF ERROR, READ MORE.
 * @ingroup ipc_ll;
 */
typedef enum {
    IPCRS_SUCCESS,  /**< Successful return */
    IPCRS_ERROR,    /**< Error */
    IPCRS_AGAIN,    /**< Read more data from that socket */
    IPCRS_EOF       /**< Unexpected EOF on read */
} IPC_readState;

/**
 * The function type used to do a select() on FDs in IPCLL_recvData().
 *
 * @internal
 * @ingroup ipc_ll
 *
 * @param sockMax The highest numbered socket in readfds
 * @param readfds The FDs to read from
 * @param data Any additional data that might be needed for the function
 * @retrun The number of FDs with data, or < 0 on error.
 */
typedef int (*IPC_select_t)(int sockMax, fd_set *readfds, void *data);

/** 
 * Enumerate the possible modes that this can be in.
 * @ingroup ipc_ll
 * @internal
 */
typedef enum {
    IPCM_SERVER,
    IPCM_CLIENT
} IPC_mode;

#define USE_IPC_FRAME_HEADER 0

#if USE_IPC_FRAME_HEADER
/**
 * A frame header value for IPC packets.  And yes this is a bit of a
 * poorly thought out hack.
 */
#define IPC_FRAME_HEADER    "\xFF\x11\xEE\x22"
#define IPC_FHEAD_LEN       4
#endif


#endif /* IPC_IPC_TYPES_H */
