/**
 * @file ipcll.c
 *
 * @brief Implement Low-Level IPC functions
 *
 * $Id: ipcll.c,v 1.19.2.2 2003/04/30 20:03:22 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Mar 08 16:42:33 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "ipc_private.h"
#include "ipc.h"
#include "ipcll.h"
#include "ipc_err.h"
#include "ipc_ctl.h"
#include "ipc_pkt.h"
#include "ipc_sig.h"
#include "ipc_cstate.h"

#include "error.h"
#include "debug.h"


static char const rcsid[] 
    = "$Id: ipcll.c,v 1.19.2.2 2003/04/30 20:03:22 devink Exp $";

/** 
 * Accept an incomming connection.  
 * This will not block, if there is no connection waiting to be accepted 
 * then, this assumes that it's done and retruns TRUE.
 *
 * @internal
 * @param handle The handle that \c sock belongs to.
 * @param sock The socket to accept the connection on.
 * @param outpidx The proc index of the new client, may be NULL
 * @return TRUE if there were no problems, FALSE otherwise.
 *
 * @pre handle != NULL
 * @pre sock >= 0
 */
static bool acceptConn(IPC *handle, int sock, int *outpidx);

/** 
 * Read data from a socket and do any processing needed.
 *
 * @internal
 * @param handle The IPC structure \c sock belongs to.
 * @param sock The socket to read data from
 * @return an IPC_readState value.
 *
 * @pre handle != NULL
 */
static IPC_readState readSock(IPC *handle, int sock);

/**
 * Process an event packet.
 * This will drop any event type that was not supposed to be send over
 * the wire.  If the event type is legal then it will be queued for
 * later pickup.
 *
 * @param handle The IPC handle to work on.
 * @param pkt The IPCT_EVENT packet.
 * @return an IPC_readState value.
 *
 * @pre handle != NULL
 * @pre pkt != NULL
 * @pre pkt->type == IPCT_EVENT
 * @post Done with pkt, it can be relased.
 */
static IPC_readState processEvent(IPC *handle, struct IPC_packet *pkt);

/**
 * Do a select call to wait for data to read off the given socket set.  
 * Handle will have error codes set correctly on error.
 *
 * @param sockMax The highest numberd socket in the set
 * @param readSet The FDs to wait for data on
 * @param data Not used.
 *
 * @return The number of FDs with data, or -1 on error.
 */
static int IPCSelect(int sockMax, fd_set *readSet, void *data);

/**
 * Add the given socket FD to a file descriptor set.  This also checks
 * error conditions.  And it will update sockMax if need be.
 *
 * @param sockMax The highest number FD added to \c sockSet
 * @param sockSet The set to add \c sock to.
 * @param sock The socket FD to add.
 *
 * @pre sockMax != NULL
 * @pre sockSet != NULL
 * @pre sock >= 0
 */ 
inline static void addToSSet(int *sockMax, fd_set *sockSet, int sock);


bool
IPCLL_send(IPC *handle, struct IPC_packet *packet)
{
    struct IPC_cstate   *dst = NULL;
    int                 sock = -1;
    struct IPC_buffer   headBuf;

    assert(handle != NULL);
    assert(packet != NULL);

    DPRINT(DBG_IPC, "send: %d: %d -> %d",
            packet->type, packet->sp_addr, packet->dp_addr, handle->paddr);


    IPCLL_clearError(handle);

    BUF_init(&headBuf);

    /* figure out where we are sending this */
    if (handle->mode == IPCM_SERVER) {
        dst = &handle->cliConns[IPC_PAToIndex(packet->dp_addr)];
        if (! IPCST_isActive(dst->state) ) {
            handle->errcode = IPCE_NOSUCHPROC;
            return FALSE;
        }
        sock = dst->sock;
    }
    else if (handle->mode == IPCM_CLIENT) {
        if (! IPCST_isActive(handle->mainState.state) ) {
            handle->errcode = IPCE_NOSUCHPROC;
            return FALSE;
        }
        /* these can only be sent to the controller */
        if (! packet->dp_addr == IPCAP_CTLR) {
            handle->errcode = IPCE_NOSUCHPROC;
            return FALSE;
        }
        sock = handle->mainState.sock;
    }

    if (sock < 0) {
        DPRINT(DBG_IPC, "IPC mode = %d", handle->mode);
        throw_fatal("Unknown IPC mode");
        /* Not Reached */
        return FALSE;
    }



    /* we do this in two blocks */
    DPRINT(DBG_IPC, "send: packet = {%d, %d, %d, %d, %d, %d, %d, %d, %p}",
            packet->majorVer, packet->minorVer,
            packet->sp_addr, packet->sc_addr, packet->dp_addr, packet->dc_addr,
            packet->type, packet->datalen, packet->data);

    if (! IPCPacket_toBuffer(&headBuf, packet) ) {
        handle->errcode = IPCE_ERROR;
        return FALSE;
    }


    if (! BUF_send(&headBuf, sock, IPCHEADER_LEN, TRUE) ) {
        /* error */
        handle->errcode = IPCE_IOERROR;
        handle->errnocode = errno;
        DPRINT(DBG_IPC, "send failed: errno = %d", errno);
        return FALSE;
    }

    BUF_free(&headBuf);

    if (! BUF_send(packet->data, sock, packet->datalen, FALSE) ) {
        /* error */
        handle->errcode = IPCE_IOERROR;
        handle->errnocode = errno;
        DPRINT(DBG_IPC, "send failed: errno = %d", errno);
        return FALSE;
    }

    DPRINT(DBG_IPC, "send: end (success)", NO_VARG);

    return TRUE;
}


void
IPCLL_clearError(IPC *handle)
{
    handle->errcode = IPCE_SUCCESS;
    handle->errnocode = 0;
}

bool
IPCLL_readAllSocks(IPC *handle, IPC_select_t sel, void *selData)
{
    fd_set              *curSSet;
    fd_set              *nextSSet; 
    fd_set              *freeSSet;   /* tmp for swap */
    fd_set              sockSets[2];
    int                 sockMax;
    int                 fdcount;
    bool                error;
    int                 i;
    IPC_readState       rdState;

    assert(handle != NULL);


    DPRINT(DBG_IPC, "recv: start", NO_VARG);


    /* default to no errors */
    error = FALSE;

    /* stop compiler complaints */
    i = 0;
    fdcount = 0;

    /* See which select function we are using */
    if (sel == NULL) 
        sel = IPCSelect;

    /* build sockSet for the select call */
    curSSet = &(sockSets[0]);
    nextSSet = &(sockSets[1]);
    freeSSet = NULL;
    sockMax = 0;
    FD_ZERO(curSSet);
    FD_ZERO(nextSSet);

    /* check & handle waiting signals. Done early so that an clean up
     * that is done will not cause unexpected results
     */
    IPCSig_check(handle);

    /* Don't do anything if we are not in an active state, but don't
     * make our caller think there was an actual error. 
     */
    if (! IPCST_isActive(handle->mainState.state) ) {
        DPRINT(DBG_IPC, "recv: skipping, not in an active state", NO_VARG); 
        return TRUE;
    }

    /* Add main handle */
    addToSSet(&sockMax, nextSSet, handle->mainState.sock);

    /* add all active clients */
    if (handle->mode == IPCM_SERVER) {
        fdcount = handle->cliCount;
        for (i = 0; i < IPC_MAX_IMPROCS && fdcount > 0; i++) {
            if (IPCST_isActive(handle->cliConns[i].state) ) {

                DPRINT(DBG_IPC, 
                        "recv: adding client %d, sock fd %d to curSockSet",
                        IPC_indexToPA(i), handle->cliConns[i].sock);

                addToSSet(&sockMax, nextSSet, handle->cliConns[i].sock);

                fdcount--;
            }
        }
    }

    do {
        /* swap curSockSet and nextSockSet */
        freeSSet = curSSet;
        curSSet = nextSSet;
        nextSSet = freeSSet;


        fdcount = sel(sockMax, curSSet, selData) ; 

        if (fdcount < 0) {
            handle->errcode = IPCE_ERROR;
            handle->errnocode = errno;
            error = TRUE;                   /* bail now */
            break;
        }

        if (fdcount == 0) break;            /* nothing to do */

        /* further selects should be done with out internal select */
        sel = IPCSelect;
        selData = NULL;

        sockMax = -1;    
        FD_ZERO(nextSSet);

        /* Input modules and controllers differ heavily here */
        if (handle->mode == IPCM_SERVER) {

            /* check the main socket first */
            if (FD_ISSET(handle->mainState.sock, curSSet) ) {
                if (! acceptConn(handle, handle->mainState.sock, &i) ) {
                    error = TRUE;
                }
                else {
                    /* accept worked, add sock for next select */
                    addToSSet(&sockMax, nextSSet, handle->cliConns[i].sock);
                }
                fdcount--;
            }

            /* read data off of cliConns */
            for (i = 0; !error && i < IPC_MAX_IMPROCS && fdcount > 0; i++) {
                /**
                 * @bug This may lead to starvation if the same client is
                 * always erroring out then no clients with a higher sock ID
                 * will be able to get data through.
                 */
                if (IPCST_isActive(handle->cliConns[i].state) 
                    && FD_ISSET(handle->cliConns[i].sock, curSSet) ) 
                {
                    DPRINT(DBG_IPC, "recv: data from client %d",
                            IPC_indexToPA(i));

                    rdState = readSock(handle, i); 
                    if (rdState == IPCRS_ERROR) {
                        error = TRUE;
                    }
                    else if (rdState == IPCRS_AGAIN) {
                        addToSSet(&sockMax, nextSSet,handle->cliConns[i].sock);
                    }
                    else if (rdState == IPCRS_EOF) {
                        /* here we are gussing that it is dead, and can
                         * safely be cleaned up.
                         */
                        DPRINT(DBG_IPC, "recv: unexpected EOF", NO_VARG);
                        IPCCST_pipeClear(handle, i, FALSE);
                    }

                    fdcount--;
                }
            }
        }
        else {
            /* must be a client/input module, this FD_ISSET is almost a
             * forgone conculsion, but lets be safe */
            if (FD_ISSET(handle->mainState.sock, curSSet) ) {

                DPRINT(DBG_IPC,"recv: client has data from controller",NO_VARG);

                rdState = readSock(handle, IPC_MAINSOCK); 
                if (rdState == IPCRS_ERROR) {
                    error = TRUE;
                }
                else if (rdState == IPCRS_AGAIN) {
                    addToSSet(&sockMax, nextSSet, handle->mainState.sock);
                }
                else if (rdState == IPCRS_EOF) {
                    /* Connection closed */
                    IPCLL_handleBrokenPipe(handle);
                    error = TRUE;
                }
            }
        }
    } while (!error && sockMax > 0); 

    /* Handle any waiting signals.  This is the late check to deal with
     * anything that happened while we were reading sockets.
     */
    IPCSig_check(handle);

    DPRINT(DBG_IPC, "recv: done, returning %d", !error);

    return !error;
}


void
IPCLL_handleBrokenPipe(IPC *handle)
{
    struct IPC_event   *evt;

    /* we only deal with clients at the moment */
    assert(handle->mode == IPCM_CLIENT);

    /* was it expected? */
    if (handle->mainState.state == IPCST_SENTEXIT) {
        /* controller closed before we got the ack, no
         * problem, just move us into the right state.  
         */
        handle->mainState.state = IPCST_EXIT;
        DPRINT(DBG_IPC, "broken pipe: expected", NO_VARG);
    }
    else {
        /* not expected, notify the upper layers of our
         * problem, and close ourselves down.
         */
        evt = emalloc(sizeof(struct IPC_event));
        evt->type = IPCEVT_CTLRDEATH;
        evt->addr.proc = IPCAP_CTLR;
        evt->addr.channel = IPCAC_GENERAL;
        IPCLL_queueEvent(handle, evt);

        close(handle->mainState.sock);
        handle->mainState.sock = -1;

        handle->mainState.state = IPCST_BRKPIPE;
        DPRINT(DBG_IPC, "broken pipe: unexpected", NO_VARG);
    }
}

bool
acceptConn(IPC *handle, int sock, int *outpidx)
{
    int     newSock;
    int     i;

    assert(sock >= 0);
    assert(handle != NULL);

    DPRINT(DBG_IPC, "accept: start", NO_VARG);

    newSock = accept(sock, NULL, 0);
    if (newSock < 0) {
        /* acceptable errors, that we just ignore */
        if (errno == EAGAIN || errno == EWOULDBLOCK) 
            return TRUE;

        /* otherwise */
        handle->errcode = IPCE_ACCEPTERR;
        handle->errnocode = errno;
        return FALSE;
    }
    
    /* Find a free slot and add the connection */
    for (i = 0; i < IPC_MAX_IMPROCS; i++) {
        if (IPCST_isFree(handle->cliConns[i].state) ) {
            DPRINT(DBG_IPC, 
                "accept: Choosing client in sock id %d, for fd #%d", 
                i, newSock);

            IPCCST_init(handle, i, newSock);

            if (outpidx != NULL) *outpidx = i;
            break;
        }
    }
    if (i >= IPC_MAX_IMPROCS) {
        handle->errcode = IPCE_ACCEPTERR;
        return FALSE;
    }


    /* make sock non-blocking from here on out */
    if (fcntl(newSock, F_SETFL, O_NONBLOCK) < 0) {
        handle->errnocode = errno;
        handle->errcode = IPCE_BINDERR;
        return FALSE;
    }

    DPRINT(DBG_IPC, "accept: done (success)", NO_VARG);

    return TRUE;
}

IPC_readState
readSock(IPC *handle, int pidx)
{
    struct IPC_packet   *packet;
    struct IPC_buffer   *buf;
    struct IPC_buffer   headBuf;
    IPC_readState       rdState;
    struct IPC_cstate   *cst;

    assert(handle != NULL);

    DPRINT(DBG_IPC, "recv: reading data from client %d",
            IPC_indexToPA(pidx) );

    if (pidx == IPC_MAINSOCK) {
        cst = &(handle->mainState);
    }
    else {
        cst = &(handle->cliConns[pidx]);
    }

    rdState = IPCRS_SUCCESS;

    /* clear an error code, so we know if it is current or not if we
     * detect and error at the end
     */
    IPCLL_clearError(handle);

    /* NULL these to begin with, so that an error before allocation
     * will not cause problems
     */
    packet = NULL;
    buf = NULL;

    BUF_init(&headBuf);    

    /* make sure we don't already have the headers */
    if (cst->open_pkt == NULL) {
        /* Allocate space in the buffer */
        headBuf.size = IPCHEADER_LEN;
        BUF_alloc(&headBuf);

        rdState = BUF_read(&headBuf, cst->sock, TRUE); 
        if (rdState == IPCRS_EOF) {
            /*IPCLL_clientDeath(cst);*/
            handle->errcode = IPCE_UNEOF;
            handle->errnocode = errno;
            goto finish;    /* don't try to process what we don't have */
        }
        if (rdState == IPCRS_ERROR) {
            handle->errnocode = errno;
            goto finish;    /* don't try to process what we don't have */
        }
        if (rdState == IPCRS_AGAIN) {
            goto finish;    /* don't try to process what we don't have */
        }

        packet = IPCPacket_fromBuffer(handle, &headBuf);   
        if (packet == NULL) {
            rdState = IPCRS_ERROR;
            goto finish;    /* don't try to process what we don't have */
        }

        /* save the packet headers until we know that we have the data */
        cst->open_pkt = packet; 

        /* headBuf is freed in the finish block */
    }
    else {
        packet = cst->open_pkt;
    }

    DPRINT(DBG_IPC, "recv: packet = {%d, %d, %d, %d, %d, %d, %d, %d}",
            packet->majorVer, packet->minorVer, 
            packet->sp_addr, packet->sc_addr, 
            packet->dp_addr, packet->dc_addr,
            packet->type, packet->datalen);

    buf = packet->data;
    buf->size = packet->datalen;
    BUF_alloc(buf);

    DPRINT(DBG_IPC, "recv: trying to read %d bytes of body data", buf->size);

    rdState = BUF_read(buf, cst->sock, FALSE);
    if (rdState == IPCRS_EOF) {
        /*IPCLL_clientDeath(cst);*/
        handle->errcode = IPCE_UNEOF;
        handle->errnocode = errno;
        goto finish;    /* don't try to process what we don't have */
    }
    if (rdState == IPCRS_ERROR) {
        handle->errnocode = errno;
        goto finish;    /* don't try to process what we don't have */
    }
    if (rdState == IPCRS_AGAIN) {
        goto finish;    /* don't try to process what we don't have */
    }

    /* have data so clear out the partial packet */
    cst->open_pkt = NULL;

    DPRINT(DBG_IPC, "recv: body data read", NO_VARG);

    DPRINT(DBG_IPC, "recv: got packet type %d", packet->type);

    switch (packet->type) {
        case IPCT_CTRL:
        {
            rdState = IPCLL_processControl(handle, packet, pidx); 
            /* we are done with this */
            IPCPacket_free(packet);
            packet = NULL;
            break;
        }
        case IPCT_PATH: 
        {
            rdState = IPCLL_processPath(handle, packet, pidx);
            /* Not needed any longer */
            IPCPacket_free(packet);
            packet = NULL;
            break;
        }
        case IPCT_EVENT:
        {
            rdState = processEvent(handle, packet);
            /* Not needed any longer */
            IPCPacket_free(packet);
            packet = NULL;
            break;
        }
        default:
        {
            IPCLL_queuePacket(handle, packet);
            /* Just going to have to trust the consumer to free it */
            break;
        }
    }


    /* 
     * general cleanup and final error checks.  The last error check is
     * needed b/c not all failures cause an immediate goto.
     */
    finish:

        if (rdState == IPCRS_ERROR) {
            /** 
             * @todo check for a conn reset by peer error and notify the
             * upper layers the server is dead.
             */
            DPRINT(DBG_IPC, "recv: error reading/processing packet: (%d) %s", 
                    IPC_geterrno(handle), IPC_geterrstr(handle));

            if (packet != NULL) IPCPacket_free(packet);

            /* make sure some error code is set */
            if (handle->errcode == IPCE_SUCCESS) 
                handle->errcode = IPCE_IOERROR;
        }
        BUF_free(&headBuf);

        DPRINT(DBG_IPC, "recv: end (%d)", rdState);
    
        return rdState;
}

static IPC_readState
processEvent(IPC *handle, struct IPC_packet *pkt)
{
    struct IPC_event    *evt = NULL;
    uint32_t            type;

    assert(handle != NULL);
    assert(pkt != NULL);
    assert(pkt->type == IPCT_EVENT);


    if (! BUF_getInteger(pkt->data, &type) ) return IPCRS_ERROR;

    /* make sure the other side was alowed to send the packet; if it
     * wasn't just pretend that everything is ok, and drop the packet.
     */
    if (handle->mode == IPCM_SERVER) {
        if ( (type & IPCEVT_CLISEND) != type) return IPCRS_SUCCESS;
    }
    else if (handle->mode == IPCM_CLIENT) {
        if ( (type & IPCEVT_SRVSEND) != type) return IPCRS_SUCCESS;
    }
    else {
        throw_fatal("Unknown IPC mode");
    }

    /* now we know we will at least try to queue the event */
    evt = emalloc(sizeof(struct IPC_event));
    evt->addr.proc = pkt->sp_addr;
    evt->addr.channel = pkt->sc_addr;
    evt->type = type;

    if (! BUF_getInteger(pkt->data, (uint32_t*)&(evt->pid) ) ) goto error; 
    if (! BUF_getInteger(pkt->data, &(evt->data) ) ) goto error;

    IPCLL_queueEvent(handle, evt);

    return IPCRS_SUCCESS;
    
    /* this is only reached when something errored out above, cleanup
     * and return an error
     */
error:
    free(evt);

    return IPCRS_ERROR;
}

static int
IPCSelect(int sockMax, fd_set *readSet, void *data)
{
    struct timeval      readTimeout;
    int                 fdcount;

    DPRINT(DBG_IPC, "IPCSelect: start", NO_VARG);

    /* figure out which sockets have data waiting. */
    readTimeout.tv_sec = IPC_SELECT_TIMEOUT;
    readTimeout.tv_usec = 0;

    sockMax++;      /* select want's max FD + 1 */

    fdcount = select(sockMax, readSet, NULL, NULL, &readTimeout);

    if (fdcount < 0) {
        /* check errno and deal with the problem */
        if (errno == ENOMEM) {
            /* oops ... big problem */
            throw_fatal("Select Returned ENOMEM");
        }

        /* interruptions don't really bother us */
        if (errno == EINTR) fdcount = 0;
        else return -1;                      /* bail now */
    }

    return fdcount;
}

static void
addToSSet(int *sockMax, fd_set *sockSet, int sock)
{
    assert(sockSet != NULL);
    assert(sockMax != NULL);
    assert(sock >= 0);


    FD_SET(sock, sockSet);

    if (sock > *sockMax) 
        *sockMax = sock; 
}

