/**
 * @file ipcll.h
 *
 * @brief Low level functions for the IPC interface.
 * @par
 * @internal
 * This is to be used by the IPC module only.
 *
 * $Id: ipcll.h,v 1.10.2.5 2003/04/30 20:03:22 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Feb 26 12:37:33 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IPC_IPCLL_H
#define IPC_IPCLL_H

#include "ipc.h"
#include "ipc_types.h"
#include "ipc_pkt.h"

/**@defgroup ipc_ll Low Level IPC Functions
 * @ingroup ipc
 * @{
 */

/** 
 * The max back log on new incomming connections. I'm pulling this
 * number out of thin air right now, it should probably be tuned later.
 */
#define IPC_MAX_BACKLOG     100 

/** 
 * This is a proc index for the main socket in an IPC structure.  All other
 * proc indexes refer to particular client connections.
 */
#define IPC_MAINSOCK -1

/** 
 * The time (in seconds) to block waiting for file descriptors to be
 * ready.  On a timeout INW_getEvent() loops, so it will not return
 * until data is ready, however, this gives it some run time to deal
 * with outstanding things.
 */
#define IPC_SELECT_TIMEOUT 1

/** 
 * A structure to hold lists of path data buffers.   This whole sale
 * discards the packet headers, and only store the needed addressing
 * info and data buffers.
 * @internal
 * @ingroup ipc
 */
struct IPC_pathList
{
    List            bufList;    /**< The list of buffers with path frames */
    struct IPC_addr src;        /**< The source address */
};

/**
 * Send a packet.  
 * @par
 * This assumes that the packet has had all addressing type and data fields 
 * filled in.  Note that the \c data field should be filled in via 
 * IPC_setPacketData().
 * @param handle instance handle
 * @param packet The packet to send
 * @return TRUE on success, FALSE otherwise (and handle->errcode will be set)
 *
 * @pre handle != NULL
 * @pre packet != NULL
 */
bool IPCLL_send(IPC *handle, struct IPC_packet *packet);

/** 
 * Recieve a packet of a specified type.  
 * This will set \c packet to the address of the next waiting IPCT_* 
 * packet, if such a packet exists.  
 *
 * This is the default recv function, if an IPCT_ has a special recv
 * function associated with it, then this function cannot be used to
 * recieve thoes packets.
 *
 * @param handle The instance handle
 * @param type The \ref ipc_msgtypes to check for packets.
 * @return a pointer to the next waiting IPC_packet of the specified
 *      type, or NULL if none are present.
 *
 * @pre handle != NULL
 */
inline struct IPC_packet * IPCLL_recvPacket(IPC *handle, uint32_t type);

/**
 * Recieve an IPC_pathList.  This gets the next waiting path. 
 *
 * @param handle The instance handle
 * @return A pointer to the next waiting IPC_pathList in the queue, or
 *         NULL if none are present.
 *
 * @pre handle != NULL
 */
inline struct IPC_pathList * IPCLL_recvPath(IPC *handle);

/**
 * Recieve an IPC_event.  
 *
 * @param handle The instance handle
 * @return A pointer to the next waiting IPC_pathList in the queue, or
 *         NULL if none are present.
 *
 * @pre handle != NULL
 */
inline struct IPC_event * IPCLL_recvEvent(IPC *handle);

/** 
 * Recieve data from all the sockets and queue or process it as needed.
 *
 * The function for doing a select is the optional parameter, this
 * allows the InputWaiter code to check it's own FDs at the same time as
 * the IPC FDs.
 *
 * @param handle The handle to read from 
 * @param sel pointer to the function to use for select.
 * @param selData pointer to extra data needed for \c sel
 * @return TRUE if there were no problems, FALSE otherwise.
 *
 * @pre handle != NULL 
 */
bool IPCLL_readAllSocks(IPC *handle, IPC_select_t sel, void *selData);

/**
 * Append an packet to the correct queue.  This takes care of all the
 * state updates that are needed.  What queue to append to is determined
 * from the packet.
 *
 * @param handle The IPC handle to work on.
 * @param pkt The Packet to append.
 *
 * @pre handle != NULL
 * @pre pkt != NULL
 */
inline void IPCLL_queuePacket(IPC *handle, struct IPC_packet *pkt);

/**
 * Append a path packet list to the correct queue.
 *
 * @param handle The IPC handle to work on.
 * @param pth The pathList to add.
 *
 * @pre handle != NULL
 * @pre pth != NULL
 */
inline void IPCLL_queuePath(IPC *handle, struct IPC_pathList *pth);

/**
 * Append an event to the correct queue.
 *
 * @param handle The IPC handel to work on.
 * @param evt The imEvent to add
 *
 * @pre handle != NULL
 * @pre pth != NULL
 */
inline void IPCLL_queueEvent(IPC *handle, struct IPC_event *evt);

/** 
 * Check for queue packets.
 * The return value of this function is a bitmap of waiting 
 * \ref ipc_msgtypes.  This bit mask should be disassemble via the
 * IPC_tMapIsSet() function
 *
 * @param handle The IPC handle
 * @return 0 if no data is waiting, A bitmap of waiting \ref ipc_msgtypes 
 *  otherwise.
 *
 * @pre handle != NULL
 */
int IPCLL_hasQueued(IPC *handle);

/**
 * Clear the error state.
 *
 * @param handle The IPC handle to clear
 *
 * @pre handle != NULL
 */
void IPCLL_clearError(IPC *handle);

/**
 * Handle a broken pipe.  This is invoked when \c mode==IPCM_CLIENT and
 * either an EOF on the main socket was encountered, or a SIGPIPE was
 * recieved.
 *
 * @param handle The IPC to work with
 */
void IPCLL_handleBrokenPipe(IPC *handle);

/*@}*/


#endif /* IPC_IPCLL_H */

