/**
 * @file message.c
 *
 * @brief Implement utility functions for dealing with message and path
 * structures.
 * @ingroup msg
 *
 *
 * $Id: message.c,v 1.6.2.6 2003/10/22 04:09:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Jul 08 17:27:12 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <time.h>

#include "message.h"
#include "msg_const.h"
#include "util/netaddr.h"
#include "util/bitvec.h"

#include "debug.h"
#include "error.h"
#include "strcompat.h"

static char const rcsid[] 
    = "$Id: message.c,v 1.6.2.6 2003/10/22 04:09:03 devink Exp $";

void
PATH_free(struct path *head)
{
    struct path    *cur, *last;

    cur = head;
    while (cur != NULL) {
        last = cur;
        cur = cur->next;

        /* free the path structure */
		if(NULL!=last->fromFQDN)
        	free(last->fromFQDN);
        NA_destroy(last->fromIP);
		if(NULL!=last->toFQDN)
        	free(last->toFQDN);
        NA_destroy(last->toIP);
        free(last);
    }
}

bool
PATH_mkLinkProps(struct path *path, char const *str)
{
    char const  *p;
    bool        bad_prop;
    uint32_t    props; 

    bad_prop = FALSE;
    props = 0;
    p = str;
    while (*p != '\0') {
        switch (*p) {
            case 'o': BM_SET(props, WEAK_PRIV);   break;
            case 'O': BM_SET(props, STRONG_PRIV); break;
            case 'U': BM_SET(props, VALID_USER);  break;
            case 'A': BM_SET(props, AUTH_LAYER);  break;
            case 'R': BM_SET(props, REPLAY_PROT); break;
            case 'I': BM_SET(props, INTEGRETY);   break;
            case 'L': BM_SET(props, RELIABLE);    break;
            case 'D': BM_SET(props, FROM_DEVICE); break;

            default: bad_prop = TRUE; break;
        }
        p++;
    }

    path->props = props;

    return !bad_prop;
}

size_t
PATH_strLinkProps(const struct path *path, char *buf, size_t sz)
{
    uint32_t    props;
    char        *p;
    int         i;
    size_t      writ;

    props = path->props;
    p = buf;
    writ = 0;
    for (i = 0; i < PROP_COUNT; i++) {
        /* Skip properties which are not set */
        if (! BM_ISSET(props, i) ) continue;

        switch (i) {
            case WEAK_PRIV:     if (writ < sz) *p++ = 'o'; writ++; break;
            case STRONG_PRIV:   if (writ < sz) *p++ = 'O'; writ++; break;
            case VALID_USER:    if (writ < sz) *p++ = 'U'; writ++; break;
            case AUTH_LAYER:    if (writ < sz) *p++ = 'A'; writ++; break;
            case REPLAY_PROT:   if (writ < sz) *p++ = 'R'; writ++; break;
            case INTEGRETY:     if (writ < sz) *p++ = 'I'; writ++; break;
            case RELIABLE:      if (writ < sz) *p++ = 'L'; writ++; break;
            case FROM_DEVICE:   if (writ < sz) *p++ = 'D'; writ++; break;
            /* just ignore bad properties */
        }
    }

    *p = '\0';  /* We made sure to leave room for NULL */

    return writ;
}

void
MSG_free(struct message *msg)
{
    free(msg->msg);
    free(msg->host);
    free(msg->dfqdn);
    NA_destroy(msg->dip);

    memset(msg, 0, sizeof(struct message));
}

int
MSG_getPriority(const struct message *msg)
{
    return (msg->fac << 3) + msg->sev;
}

void
MSG_setTimeZone(struct message *msg)
{
    struct tm       tm;
    time_t          tt;

    if (msg->tz != MSG_TZ_UNKNOWN) 
        return;

    /* XXX The use of tm::tm_gmtoff may not be portable, but it is the
     * only method that works under linux
     */
    if (time(&tt) != (time_t) -1 && localtime_r(&tt, &tm) != NULL) {
        /* tm_gmtoff is in sections offset from GMT */
        msg->tz = tm.tm_gmtoff / 60;
    }

    DPRINT(DBG_MSG, "MSG_setTimeZone: tz set to %d", msg->tz);
}

struct iam*
MSG_iamCopy(const struct iam* src) 
{
    struct iam *dst;

    if (src == NULL)
        return NULL;

    dst = emalloc(sizeof(struct iam)); 
    
    dst->fqdnlen = src->fqdnlen;
    dst->type = src->type;
    dst->ip = NA_newNetaddr(src->ip);

    dst->fqdn = emalloc(dst->fqdnlen);
    strlcpy(dst->fqdn, src->fqdn, dst->fqdnlen);

    return dst;
}

void MSG_iamFree(struct iam *obj)
{
    free(obj->fqdn);
    obj->fqdn = NULL;
    NA_destroy(obj->ip);
    obj->ip = NULL;
}
