/**
 * @file message.h
 *
 * @brief Structures and operations for the internal representation of
 * syslog messages and %path messages.
 *
 * @design d3_1
 *
 * @ingroup msg
 *
 * $Id: message.h,v 1.10.2.8 2003/10/22 04:09:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Jan 18 14:17:28 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MSG_MESSAGE_H
#define MSG_MESSAGE_H

#include "util/netaddr.h"


/** @defgroup msg Syslog Messages
 *  @{
 */

#define MSGTAG_LEN  33 /**< Max Tag length + 1, defined by RFC3164 */

/** A structure that contains one syslog message. */
struct message {
    /** @name All formats use these */
    /**@{ */
    int         sev;            /**< Severity */ 
    int         fac;            /**< Facility */
    size_t      hostlen;        /**< length of HOSTNAME (inc null) */
    char        *host;          /**< HOSTNAME, must be free()ed */
    time_t      timestamp;      /**< TIMESTAMP */
	int			tz;				/**< TIMEZONE, in minutes E of Greenwich
								 	(use -1*tz_minutes_west ) */
    char        tag[MSGTAG_LEN];/**< TAG */
    size_t      msglen;         /**< length of CONTENT (inc null) */
    char        *msg;           /**< CONTENT, must be free()ed */
    /*@}*/

    /**@name TCP/COOKED specific */
    /**@{ */
    uint32_t    pathid;     /**< pathID Attribute */
    uint32_t    connid;     /**< Connection ID for the path */

    netaddr_t   dip;        /**< deviceIP attribute */ 
    size_t      dfqdnlen;   /**< length of deviceFQDN (inc null) */ 
    char        *dfqdn;     /**< deviceFQDN attribute */ 
    /*@}*/
};

/** 
 * A structure that contains one entry in a path block.  The next entry
 * (going from outter to innermost) will be pointed to by \c next and
 * the innermost path element will have <code>next == NULL</code>.
 */
struct path {
    uint32_t    id;         /**< path id */
    uint32_t    connid;     /**< connection id for this path */

    size_t      ffqdnlen;   /**< fromFQDN length (inc null) */
    char        *fromFQDN;  /**< fromFQDN attribute, must be free()ed */
    netaddr_t   fromIP;     /**< fromIP attribute */

    size_t      tfqdnlen;   /**< toFQDN length (inc null) */
    char        *toFQDN;    /**< toFQDN attribute, must be free()ed */
    netaddr_t   toIP;       /**< toIP attribute */
    uint32_t    props;      /**< linkprops attirbute, bit map */
    struct path *next;      /**< next path node */
};

/** Device types for IAM elements */
enum MSG_DTYPE 
{
    MSGDT_DEVICE,    /**< 'device' type */
    MSGDT_RELAY,     /**< 'relay' type */
    MSGDT_COLLECT    /**< 'collector' type */
};

/** A structure for storing IAM element information */
struct iam {
    char        *fqdn;          /**< 'fqdn' attribute, must be freed()ed */
    size_t      fqdnlen;        /**< FQDN length, inc null */

    netaddr_t   ip;             /**< 'ip' attribute */

    enum MSG_DTYPE type;        /**< 'type' attribute */
};


/** The types of data that can be held by a MSG_qelem */
typedef enum {
    MSGQE_MSG,      /**< struct message, MSG_qelem::data.msg */
    MSGQE_PATH,     /**< struct path, MSG_qelem::data.path */
    MSGQE_CLOSE     /**< connection close event, no data */
} MSGQE_TYPE;

/**
 * A structure used to queue up message and path objects.  Some things
 * (threaded input modules) need this.
 */
struct MSG_qelem {
    MSGQE_TYPE  type;           /**< type of object in this one */
    uint32_t    chan;           /**< Channel this came in on. */
    union {
        struct message  *msg;   /**< message object data */
        struct path     *path;  /**< path object data */
        uint32_t        connid; /**< connection id that just closed */
    } data;                     /**< data */
};

/** A value that signals an unset path id or connection id */
#define PATH_UNSET_ID   0

/** The smallest value for path ID */
#define PATH_ID_MIN     1

/** The largest value for a path ID */
#define PATH_ID_MAX     UINT_MAX

/** Path Link Properties, usable with BM_* functions */
typedef enum {
    WEAK_PRIV,      /**< Weak privacy */
    STRONG_PRIV,    /**< Strong privacy */
    VALID_USER ,    /**< A valid user has been negotiated */
    AUTH_LAYER,     /**< Authentication layer use on link */
    REPLAY_PROT,    /**< link protected against message replay */
    INTEGRETY,      /**< message integrety assure on link */
    RELIABLE,       /**< reliable link */
    FROM_DEVICE,    /**< The from side of the link is a device */
    PROP_COUNT      /**< Number of link properties */
} PATH_LINKPROPS;

/**
 * Make a Link Props bit mask from a RFC 3195 compliant string of link
 * props.  Note that this will return an error if any of the link
 * properties in the string were invalid, if none were valid then
 * \c path->props==0 on return.
 *
 * @param path The path to work on.
 * @param str The String of link props
 * @return TRUE if all link props were valid, FALSE otherwise.
 */
bool PATH_mkLinkProps(struct path *path, char const *str);

/**
 * Turn link props into a string.
 *
 * @param path The path object to look at the link properties of.
 * @param buf A buffer to place the link props string into.
 * @param sz The size of buf.
 * @param The numbeg of characets need to write the link props, not
 *        including NULL.
 */
size_t PATH_strLinkProps(const struct path *path, char *buf, size_t sz);

/**
 * Free a path structure.  This function recurses into the whole path
 * list and frees all the elements.
 *
 * @param pathHead The head of the path to free.
 */
void PATH_free(struct path *pathHead);

/**
 * Calculate the priority for a given message.
 *
 * @param msg The message to find the priority of.
 * @return the priority of \c msg 
 */
inline int MSG_getPriority(const struct message *msg);

/**
 * Free all the fields in a message structure.  This rather assumes that 
 * you will reuse the message structure so it clears some of the fields
 * out as well. \c msg is not free()ed.
 *
 * @param msg The message structure to free.
 */
void MSG_free(struct message *msg);

/** 
 * Set the timezone on a message if it is unset.
 *
 * @param msg The message structure to work with.
 */
void MSG_setTimeZone(struct message *msg);

/**
 * Copy a struct iam.  The resulting object must be free()ed.
 *
 * @param src The strict iam to copy
 * @return the new strict iam.
 */
struct iam *MSG_iamCopy(const struct iam *src);

/**
 * Free the fields of a struct iam.  This makes no attempt to actually
 * free the struct iam object its self.
 *
 * @param obj The object to free data from.
 */
void MSG_iamFree(struct iam *obj);
/*@}*/

#endif /* MSG_MESSAGE_H */
