/**
 * @file msg_frmt.c
 *
 * @brief The top-level generic message formatter support functions.
 *
 * $Id: msg_frmt.c,v 1.1.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Aug 28 19:13:56 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include "util/bitvec.h"

#include "error.h"
#include "log.h"

#include "msg_frmt.h"
#include "msg_frmtll.h"

static char const rcsid[] 
    = "$Id: msg_frmt.c,v 1.1.2.7 2003/10/22 04:09:03 devink Exp $";

/** Maximum length of a name */
#define MFLL_NAME_SZ     32

/** struct to store name -> info mappings of formatter types in. */
struct MFLL_typeMap
{
    char            name[MFLL_NAME_SZ]; /**< formatter name */
    MFRMT_TYPE      type;               /**< formatter type code */
};

/** struct to store name -> info mapptings of formatter options */
struct MFLL_optMap
{
    char            name[MFLL_NAME_SZ]; /**< option name */
    enum MFOPT      num;                /**< option number */
};

/** 
 * Store information used for checking options in formatter types and
 * matching them to output modules.
 */
struct MFLL_typeInfo
{
    MFRMT_TYPE      type;               /**< type */

    char            name[MFLL_NAME_SZ]; /**< primary name */

    MFLL_msg_frmt   msg_frmt;           /**< Message formatter */
    MFLL_pth_frmt   pth_frmt;           /**< Path formatter */

    uint32_t        valid_opts;         /**< Valid options */
};

/** Store extra information about an option. */
struct MFLL_optInfo
{
    enum MFOPT      num;                /**< option number */
    char            name[MFLL_NAME_SZ]; /**< option name */

    bool            depreciated;        /**< Is the option depreciated? */
};

/** Map names to formatter types */
static struct MFLL_typeMap type_map[] = {

    /**< RFC 3164 & variants 
     * @{
     */
    {"rfc3164", MFRMT_RFC3164},     
    {"udp", MFRMT_RFC3164},
    {"readable", MFRMT_RFC3164},  
    {"beep-raw", MFRMT_RFC3164}, 
    {"raw", MFRMT_RFC3164}, 
    /*@}*/

    /** BEEP COOKED format 
     * @{
     */
    {"beep-cooked", MFRMT_COOKED},  
    {"cooked", MFRMT_COOKED}, 
    /*@}*/

    {"", MFRMT_NOEXIST},            /**< error code really */
};

/** Map names to formatter options */
static struct MFLL_optMap opt_map[] = {
    /** RFC3339 timestamps */
    {"ts_rfc3339", MFOPT_RFC3339},

    /** RFC3164 + year, timezone */
    {"ts_extended", MFOPT_TSEXT},

    /** RFC3164 + timezone
     * @{
     */
    {"tz", MFOPT_TSTZ},
    {"timezone", MFOPT_TSTZ},
    /*@}*/

    /** RFC3164 + year */
    {"year", MFOPT_TSYEAR},

    /** Priority 
     * @{
     */
    {"pri", MFOPT_PRI},
    {"priority", MFOPT_PRI},
    /*@}*/

    /* use FQDN in RFC3164 */
    {"fqdn", MFOPT_FQDN},

    /** End marker */
    {"", MFOPT_NOEXIST},
};

static struct MFLL_optInfo opt_info[MFOPT_COUNT] = {
    {MFOPT_NOEXIST, "nonexistant", FALSE},
    {MFOPT_RFC3339, "ts_rfc3339", FALSE},
    {MFOPT_TSEXT, "ts_extended", FALSE},
    {MFOPT_TSTZ, "timezone", TRUE},
    {MFOPT_TSYEAR, "year", TRUE},
    {MFOPT_FQDN, "fqdn", FALSE},
    {MFOPT_PRI, "priority", FALSE},
    {MFOPT_MIMEHEAD, "mimeheaders", FALSE},
};

static struct MFLL_typeInfo type_info[MFRMT_COUNT] = {
    {MFRMT_NOEXIST, "no-existant", NULL, NULL,  0},

    {MFRMT_RFC3164, "rfc3164", MFLL_rfc3164, NULL, 
        BM_MBIT(MFOPT_RFC3339) | BM_MBIT(MFOPT_PRI) | BM_MBIT(MFOPT_TSEXT)
        | BM_MBIT(MFOPT_TSTZ) | BM_MBIT(MFOPT_TSYEAR) | BM_MBIT(MFOPT_FQDN)
        | BM_MBIT(MFOPT_MIMEHEAD)},

    {MFRMT_COOKED, "cooked", MFLL_cooked, MFLL_cookedPath, 
        BM_MBIT(MFOPT_RFC3339) | BM_MBIT(MFOPT_MIMEHEAD)},
};

/* Public Functions */
MFRMT_TYPE
MFRMT_getType(char const *name)
{
    struct MFLL_typeMap    *tm;

    for (tm = type_map; tm->type != MFRMT_NOEXIST; tm++) {
        if (strncmp(tm->name, name, MFLL_NAME_SZ) == 0) {
            break;
        }
    }

    return tm->type;
}

enum MFOPT
MFRMT_getOption(char const *name)
{
    struct MFLL_optMap      *om;

    for (om = opt_map; om->num != MFOPT_NOEXIST; om++) {
        if (strncmp(om->name, name, MFLL_NAME_SZ) == 0) {
            break;
        }
    }

    return om->num;
}

bool
MFRMT_checkOptions(struct MFRMT_opt *opts, char const *file, int line)
{
    struct MFLL_typeInfo    *ti;
    int                     i;
    uint32_t                invalid;

    assert(opts->type < MFRMT_COUNT);
    assert(opts->type >= 0);

    ti = &(type_info[opts->type]);

    invalid = BM_ISECT(~(ti->valid_opts), opts->options);

    DPRINT(DBG_MSG, "MFRMT_checkOptions(): %d: 0x%x", 
            opts->type, opts->options);

    for (i = MFOPT_NOEXIST+1; i < MFOPT_COUNT; i++) {
        DPRINT(DBG_MSG, "MFRMT_checkOptions(): checking %s, dep %d",
                opt_info[i].name, opt_info[i].depreciated);

        if (BM_ISSET(invalid, i) ) {
            LOG_print(LOG_INFO, 
                "%s:%d: Option '%s' is not allowed for formmater '%s'",
                file, line, opt_info[i].name, ti->name);
        }

        if (BM_ISSET(opts->options, i) && opt_info[i].depreciated) {
            LOG_print(LOG_INFO, 
                "%s:%d: Option '%s' is depreciated.", 
                file, line, opt_info[i].name); 
        }
    }

    /* now clear the invalid options */
    opts->options = BM_ISECT(~invalid, opts->options);

    return invalid == 0;
}

size_t 
MFRMT_format(char *retstr, size_t retlen, const struct message *msg,
             const struct MFRMT_opt *opts)
{
    MFLL_msg_frmt   func;
    uint32_t        rtn;

    assert(opts->type < MFRMT_COUNT);
    assert(opts->type >= 0);

    DPRINT(DBG_MSG, "MFRMT_format: options = 0x%x", opts->options);

    /* Print all options 
     * Handy at times ... but only just ... so leave it commented 
     */
#if 0
#define dopt(o) DPRINT(DBG_MSG, "format option: %s = 0x%x", #o, BM_MBIT(o)); 

        dopt(MFOPT_NOEXIST)  
        dopt(MFOPT_RFC3339) 
        dopt(MFOPT_TSEXT)    
        dopt(MFOPT_TSTZ)    
        dopt(MFOPT_TSYEAR)   
        dopt(MFOPT_PRI)     
        dopt(MFOPT_FQDN)     
        dopt(MFOPT_MIMEHEAD)
        dopt(MFOPT_COUNT)   
#endif
    func = type_info[opts->type].msg_frmt;
    rtn = 0;
    if (func != NULL) {
        rtn = func(retstr, retlen, msg, opts->options);
    }

    DPRINT(DBG_MSG, "MFRMT_format: return = %d", rtn);

    return rtn;
}

size_t 
MFRMT_formatPath(char *retstr, size_t retlen, const struct path *pth,
                 const struct MFRMT_opt *opts)
{
    MFLL_pth_frmt   func;
    size_t          rtn;

    assert(opts->type < MFRMT_COUNT);
    assert(opts->type >= 0);

    func = type_info[opts->type].pth_frmt;
    rtn = 0;
    if (func != NULL) {
        rtn = func(retstr, retlen, pth, opts->options);
    }

    return rtn;
}

size_t
MFRMT_formatIam(char *retstr, size_t retlen, const struct iam *iam)
{
    int     wlen;
    char    *typeName = "you-should-never-see-this";

    assert(iam != NULL);
    assert(retstr != NULL);

    switch (iam->type) {
        case MSGDT_DEVICE: typeName = "device"; break;
        case MSGDT_RELAY: typeName = "relay"; break;
        case MSGDT_COLLECT: typeName = "collector"; break;
    }

    wlen = snprintf(retstr, retlen, 
            "%s<iam fqdn=\"%s\" ip=\"%s\" type=\"%s\"/>",
            MFRMT_COOK_MimeHeaders, 
            iam->fqdn, NA_getCharAddr(iam->ip), typeName);
    if (wlen <= 0 || wlen >= retlen)
        return 0;

    return wlen + 1;    // +1 to include ending NUL
}

bool
MFRMT_hasPathFormatter(const struct MFRMT_opt *opts)
{
    assert(opts->type < MFRMT_COUNT);
    assert(opts->type >= 0);

    return type_info[opts->type].pth_frmt != NULL;
}
