/**
 * @file msg_frmt.h
 *
 * @brief Message Formatter Functions
 * @par
 * This provides functions to format messages, The formatter and the
 * formatting options are set in the MFRMT_opt struct and given to
 * MFRMT_format() which will select the correct formatter and do the
 * work.
 *
 * @ingroup msg_frmt
 *
 * $Id: msg_frmt.h,v 1.1.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Mon Mar 18 14:11:31 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef MSG_MSG_FRMT_H 
#define MSG_MSG_FRMT_H

#ifdef HAVE_CONFIG_H
# include "acconfig.h"
#endif

#include "message/message.h"

/**@defgroup msg_frmt Message Formatters
 * @ingroup msg
 * @{
 */

/** Types of formatters */
typedef enum 
{
    MFRMT_NOEXIST,  /**< Non-existant formatter (error code) */
    MFRMT_RFC3164,  /**< The RFC3164 formatter */
    MFRMT_COOKED,   /**< RFC3195 syslog COOKED formatter */
    MFRMT_COUNT     /**< A count of formatters */
} MFRMT_TYPE;

/** Options to the formatter */
enum MFOPT
{
    MFOPT_NOEXIST,  /**< No Such options */
    MFOPT_RFC3339,  /**< Use RFC3339 Time stamps */
    MFOPT_TSEXT,    /**< Use RFC3164 + year/timezone Time stamps */
    MFOPT_TSTZ,     /**< Add timezone to RFC3164 time stamps */
    MFOPT_TSYEAR,   /**< Add year to RFC3164 time stamps */
    MFOPT_PRI,      /**< Include the priority (<...>) */
    MFOPT_FQDN,     /**< Used the FQDN in RFC3164 messages if available */
    MFOPT_MIMEHEAD, /**< Add MIME headers to the message */
    MFOPT_COUNT     /**< Number of formatter options */
}; 

/** A Structure to contain formatting options */
struct MFRMT_opt
{
    MFRMT_TYPE  type;       /**< formatter type */
    uint32_t    options;    /**< formatetr options (bit map) */
};

/** MIME headers for SYSLOGD-COOKED */
#define MFRMT_COOK_MimeHeaders "Content-Type: application/beep+xml\r\n\r\n"
/** MIME headers for SYSLOG-RAW */
#define MFRMT_RAW_MimeHeaders "\r\n"

/** Okay message for SYSLOG-COOKED */ 
#define MFRMT_COOK_OK_MSG     MFRMT_COOK_MimeHeaders "<ok/>"

/** 
 * Lookup the formatter type and return it's code.  This uses the string
 * name of the formatters and will return MFRMT_NOEXIST if the given
 * name is not vaild.
 *
 * @param name The formatter name
 * @return The formatter ID
 */
MFRMT_TYPE MFRMT_getType(char const *name);

/**
 * Lookup a formatter option by name and return the bit which needs to
 * be set in \c MFRMT_opt::options
 * 
 * @param name The format option name
 * @return The bit to set in ::options member, or 0 if \c name doesn't exist
 */
enum MFOPT MFRMT_getOption(char const *name);

/**
 * Check options against the vaild options for the formatter.  If there
 * are invalid options in the list then they will be logged, and
 * cleared.  Additionally, this will check for depreciated options, and
 * log any that are found.
 *
 * @param opts The options struct
 * @param file The file name to show is log messages
 * @param line The line number to show in log messages.
 * @return True if no invalid options were found.
 */
bool MFRMT_checkOptions(struct MFRMT_opt *opts, char const *file, int line);

/**
 * Format a message according to the give options.
 *
 * @param retstr The string to put the forammted message in.
 * @param retlen The sizeof \c retstr
 * @param msg The message to format
 * @param opts The formatting options.
 * @return The number of bytes written to \c retstr (including NULL), 
 *          or 0 on error.
 */
size_t MFRMT_format(char *retstr, size_t retlen, const struct message *msg, 
                    const struct MFRMT_opt *opts);

/**
 * Format a path object.  
 *
 * @pre \c hasPathFormatter(opts)==TRUE
 *
 * @param retstr The string to put the forammted message in.
 * @param retlen The sizeof \c retstr
 * @param pth The head of the path chain to format.
 * @param opts The formatting options.
 * @return The number of bytes written to \c retstr (including NULL), 
 *          or 0 on error.
 */
size_t MFRMT_formatPath(char *retstr, size_t retlen, const struct path *pth,
                        const struct MFRMT_opt *opts);

/**
 * Format an IAM element.  Note that this exists independently of the
 * general formatter setup becasue it is a protocol specific element
 * which is handled at a much lower level (in the BEEP stack) than the
 * rest of the formatting.
 *
 * @pre \c hasIamFormatter(opts)==TRUE
 *
 * @param retstr The string to put the formatted message in
 * @param retlen The sizeof \c retstr
 * @param iam The strict iam with all the data.
 * @return The number of bytes written to \c retstr (including NULL). 
 *      or 0 on error
 */
size_t MFRMT_formatIam(char *retstr, size_t retlen, const struct iam *iam);

/**
 * Determine if there is a path formatter available.
 *
 * @param opts The formatter options
 * @return TRUE if there is a path formatter available.
 */
bool MFRMT_hasPathFormatter(const struct MFRMT_opt *opts);

/*@}*/


#endif /* MSG_MSG_FRMT_H */
