/**
 * @file msg_frmt_cooked.c
 *
 * @brief A formatter for the syslog COOKED (RFC 3195) format.
 *
 * $Id: msg_frmt_cooked.c,v 1.2.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Aug 27 17:35:02 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#  include "acconfig.h"
#endif

#include <stdio.h>
#include "strcompat.h"
#include "types.h"

#include "debug.h"

#include "util/bitvec.h"

#include "message.h"
#include "msg_frmtll.h"

/** Amount of space needed at the end of the message for the closing tag */
#define LEN_RESERVE     8

static char const rcsid[] 
    = "$Id: msg_frmt_cooked.c,v 1.2.2.7 2003/10/22 04:09:03 devink Exp $";

size_t
MFLL_cooked(char *retstr,size_t retlen,const struct message *in,uint32_t opts)
{
    char        *pm;
    size_t      wlen;
    size_t      totlen;
    int         i;
    char        *mime_headers = "";

	DPRINT(DBG_MSG, "MFLL_cooked(%X, %d, %s, %d)",retstr,retlen,in->msg,opts);

    totlen = retlen;    /* save for later */

    /* handle mime headers */
    if (BM_ISSET(opts, MFOPT_MIMEHEAD)) {
        mime_headers = MFRMT_COOK_MimeHeaders;
    }
    DPRINT(DBG_MSG, "MFLL_cooked: using mime headers: '%s'", 
            (mime_headers[0] == '\0') ? "no" : "yes");

    /* lets put all the required fields in */
    wlen = snprintf(retstr, retlen, 
        "%s<entry facility='%d' severity='%d' tag='%s' hostname='%s' timestamp='",
        mime_headers, in->fac, in->sev, in->tag, in->host);
    if (wlen <= 0 || wlen >= retlen) return 0;
    retstr += wlen;
    retlen -= wlen;

    /* now add in the timestamp */
    wlen = MFLL_timestamp(retstr, retlen, in->timestamp, in->tz, opts);
    if (wlen <= 0) return 0;
    retstr += wlen;
    retlen -= wlen;

    /* now try to copy the optional fields */
    if (in->dfqdnlen > 0) {
        if (retlen < 2) return 0;
        retstr[0] = '\''; retstr++;
        retstr[0] = ' '; retstr++;
        retlen -= 2;

        wlen = strlcpy(retstr, "deviceFQDN='", retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        wlen = strlcpy(retstr, in->dfqdn, retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;
    }

    if (! NA_isEmpty(in->dip) && ! NA_isError(in->dip)) {
        if (retlen < 2) return 0;
        retstr[0] = '\''; retstr++;
        retstr[0] = ' '; retstr++;
        retlen -= 2;

        wlen = strlcpy(retstr, "deviceIP='", retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        wlen = strlcpy(retstr, NA_getCharAddr(in->dip), retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

    }

    if (in->pathid != PATH_UNSET_ID) {
        if (retlen < 2) return 0;
        retstr[0] = '\''; retstr++;
        retstr[0] = ' '; retstr++;
        retlen -= 2;

        wlen = snprintf(retstr, retlen, "pathID='%d", in->pathid);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;
    }

    /* Close the last attribute with '\' ' and add the final '>' to the 
     * start tag, and makesure that we have enough space left for the 
     * closing tag.
     */
    if (retlen <= (2 + LEN_RESERVE) ) return 0;
    retstr[0] = '\''; retstr++;
    retstr[0] = '>'; retstr++;
    retlen -= 2;

    /* Now that we have a vaild open, and room for a vaild close, we try
     * to copy as much of the message as we can, and hope for the best 
     */

    /* now copy the body of the message with escapes */
    pm = in->msg;
    wlen = 0;
    i = 0;
    while (i < in->msglen && *pm != '\0' && retlen > LEN_RESERVE) {
        switch (*pm) {
            case '&':
                wlen = 5;
                if (retlen <= 5+LEN_RESERVE) break;
                retstr[0] = '&'; retstr++;
                retstr[0] = 'a'; retstr++;
                retstr[0] = 'm'; retstr++;
                retstr[0] = 'p'; retstr++;
                retstr[0] = ';'; retstr++;
                break;
            case '<':
                wlen = 4;
                if (retlen <= 4+LEN_RESERVE) break;
                retstr[0] = '&'; retstr++;
                retstr[0] = 'l'; retstr++;
                retstr[0] = 't'; retstr++;
                retstr[0] = ';'; retstr++;
                break;
            case '>':
                wlen = 4;
                if (retlen <= 4+LEN_RESERVE) break;
                retstr[0] = '&'; retstr++;
                retstr[0] = 'g'; retstr++;
                retstr[0] = 't'; retstr++;
                retstr[0] = ';'; retstr++;
                break;
            default:
                wlen = 1;
                retstr[0] = *pm;
                retstr++;
        }

        /* update */
        i++;
        retlen -= wlen; 
        pm++;
    }

    /* copy in the close, space should be reserved */
    wlen = strlcpy(retstr, "</entry>", retlen);
    retlen -= wlen;

    DPRINT(DBG_MSG, "totlen = %d; retlen = %d", totlen, retlen);

    return totlen - retlen + 1;
}

size_t 
MFLL_cookedPath(char *retstr, size_t retlen, const struct path *in, 
                uint32_t opts)
{
    const struct path   *curPath;
    int                 closeCount;
    size_t              wlen;
    size_t              totlen;
    int                 i;
    char                *mime_headers = "";

    totlen = retlen;    /* save for final chars written calc */

    DPRINT(DBG_MSG, "MFLL_cookedPath: start", NO_VARG);

    if (BM_ISSET(opts, MFOPT_MIMEHEAD)) {
        mime_headers = MFRMT_COOK_MimeHeaders;
    }
    DPRINT(DBG_MSG, "MFLL_path: using mime headers: '%s'", 
            (mime_headers[0] == '\0') ? "no" : "yes");

    /* no options apply to paths */
    curPath = in;
    closeCount = 0;
    while (curPath != NULL && retlen > 0) {

        wlen = snprintf(retstr, retlen, "%s<path pathID='%d'", mime_headers, curPath->id);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        /* Only want to put the headers on once */
        mime_headers = "";

        /* fromFQDN */
        if (curPath->fromFQDN != NULL && curPath->ffqdnlen > 0) {
            wlen = strlcpy(retstr, " fromFQDN='", retlen);
            if (wlen >= retlen) return 0;
            retstr += wlen;
            retlen -= wlen;

            wlen = strlcpy(retstr, curPath->fromFQDN, retlen);
            if (wlen >= retlen) return 0;
            retstr += wlen;
            retlen -= wlen;

            if (retlen < 1) return 0;
            retstr[0] = '\'';
            retstr++;
            retlen--;

        }

        /* fromIP */
        wlen = strlcpy(retstr, " fromIP='", retlen);
        if (wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        wlen = strlcpy(retstr, NA_getCharAddr(curPath->fromIP), retlen);
        if (wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        if (retlen < 1) return 0;
        retstr[0] = '\'';
        retstr++;
        retlen--;

        /* toFQDN */
        if (curPath->toFQDN != NULL && curPath->tfqdnlen > 0) {
            wlen = strlcpy(retstr, " toFQDN='", retlen);
            if (wlen >= retlen) return 0;
            retstr += wlen;
            retlen -= wlen;

            wlen = strlcpy(retstr, curPath->toFQDN, retlen);
            if (wlen >= retlen) return 0;
            retstr += wlen;
            retlen -= wlen;

            if (retlen < 1) return 0;
            retstr[0] = '\'';
            retstr++;
            retlen--;
        }

        /* toIP */
        wlen = strlcpy(retstr, " toIP='", retlen);
        if (wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        wlen = strlcpy(retstr, NA_getCharAddr(curPath->toIP), retlen);
        if (wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        if (retlen < 1) return 0;
        retstr[0] = '\'';
        retstr++;
        retlen--;

        /* linkprops */
        wlen = strlcpy(retstr, " linkprops='", retlen);
        if (wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        wlen = PATH_strLinkProps(curPath, retstr, retlen);
        if (wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;

        if (retlen < 2) return 0;
        retstr[0] = '\''; retstr++;
        retstr[0] = '>';  retstr++;
        retlen -= 2;

        closeCount++;

        curPath = curPath->next;
    }

    DPRINT(DBG_MSG, "MFLL_cookedPath: closeCount is %d", closeCount);

    /* print closing tags */
    for (i = 0; i < closeCount && retlen > 0; i++) {
        wlen = strlcpy(retstr, "</path>", retlen);
        if (wlen >= retlen) return 0;

        retstr += wlen;
        retlen -= wlen;
    }

    DPRINT(DBG_MSG, "MFLL_cookedPath: return %d", totlen - retlen + 1);

    /* number of characters written */
    return totlen - retlen + 1;
}

