/**
 * @file msg_frmt_read.c
 *
 * @brief Readable message formatter
 * @par
 * Implements a function to format a message from the internal message struct
 * to a readable message
 *
 * @design d4_2_1
 *
 * $Id: msg_frmt_read.c,v 1.5.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Mon Mar 18 14:11:31 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <strcompat.h>
#include <time.h>
#include <sys/time.h>

#include "types.h"
#include "error.h"

#include "util/bitvec.h"

#include "msg_frmt.h"
#include "msg_frmtll.h"
#include "message.h"

static char const rcsid[] 
    = "$Id: msg_frmt_read.c,v 1.5.2.7 2003/10/22 04:09:03 devink Exp $";

size_t
MFLL_rfc3164(char *retstr,size_t retlen,const struct message *in,uint32_t opts)
{
    size_t  wlen;
    size_t  totlen;
    int     pri;

    DPRINT(DBG_MSG, "MFLL_rfc3164: start(retlen %u; opts 0x%x", retlen, opts);

    totlen = retlen;    /* save for later */

    if (BM_ISSET(opts, MFOPT_MIMEHEAD)) {
        wlen = strlcpy(retstr, MFRMT_RAW_MimeHeaders, retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;
    }
    DPRINT(DBG_MSG, "Adding mime headers '%s'", 
            BM_ISSET(opts, MFOPT_MIMEHEAD) ? "yes" : "no" );

    if (BM_ISSET(opts, MFOPT_PRI) ) {
        DPRINT(DBG_MSG, "MFLL_rfc3164: printing priority", opts);
        /* put the priority on the front of the string */
        pri = MSG_getPriority(in);
        if (pri < 0 || pri > 999)
            return 0;

        wlen = snprintf(retstr, retlen, "<%d>", pri);
        if (wlen <= 0 || wlen >= retlen) return 0;

        retstr += wlen;
        retlen -= wlen;
    }

    DPRINT(DBG_MSG, "MFLL_rfc3164: about to print timestamp, retlen %u",retlen);


    wlen = MFLL_timestamp(retstr, retlen, in->timestamp, in->tz, opts); 
    if (wlen <= 0)
        return 0;
    
    

    retstr += wlen;
    retlen -= wlen;

    if (retlen < 1) return 0;
    retstr[0] = ' ';
    retstr++;
    retlen--;

    if (BM_ISSET(opts, MFOPT_FQDN) && in->dfqdn != NULL && in->dfqdnlen != 0) {
        wlen = strlcpy(retstr, in->dfqdn, retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;
    }
    else if(in->host!=NULL && in->hostlen>0)
	{
		wlen = strlcpy(retstr,in->host,retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;
        retstr += wlen;
        retlen -= wlen;
	}

	if(in->tag!=NULL && in->tag[0]!='\0')
	{
        if (retlen < 1) return 0;
        retstr[0] = ' ';
        retstr++; 
        retlen--;

		wlen = strlcpy(retstr,in->tag,retlen);
        if (wlen <= 0 || wlen >= retlen) return 0;

        retstr += wlen;
        retlen -= wlen;
	}

    DPRINT(DBG_MSG, "MFLL_rfc3164 about to print CONTENT, retlen %u", retlen);

    if (retlen <= 0) return 0;
	if(in->msg!=NULL && in->msglen>0)
	{
		wlen = strlcpy(retstr,in->msg,retlen);

        /* Only correct for an overflow don't error because of it. */
        if (wlen >= retlen) 
            wlen = retlen - 1;

        retstr += wlen;
        retlen -= wlen;
	}

	return totlen - retlen + 1;
}

