/**
 * @file msg_parse.c
 *
 * @brief Common parsing functions.
 *
 * $Id: msg_parse.c,v 1.2.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Aug 13 14:36:28 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <ctype.h>

#include "types.h"
#include "message.h"
#include "msg_parse.h"
#include "msg_const.h"
#include "debug.h"

static char const rcsid[] 
    = "$Id: msg_parse.c,v 1.2.2.7 2003/10/22 04:09:03 devink Exp $";

/** 
 * Parse an rfc3339 time stamp.
 * @param tmout The struct tm to fill in.
 * @param tzout The time zone to fill in.
 * @param in The input string
 * @param inlen The length of the input string.
 */
static int parse_rfc3339_ts(struct tm *tmout, int *tzout, 
                            char const *in, size_t inlen);


/** 
 * Parse an rfc3164 time stamp (old style).
 * @param tmout The struct tm to fill in.
 * @param tzout The time zone to fill in.
 * @param in The input string
 * @param inlen The length of the input string.
 */
static int parse_rfc3164_ts(struct tm *tmout, int *tzout, 
                            char const *in, size_t inlen);

int 
MPARSE_time(struct message *out, char const *in, int inlen)
{
    struct tm   tmptm;
    time_t      tmptt;
    int         len;
    int         tz;

    /* initilize to sane values. 
     * This has one small problem.  If the time stamp doesn't include
     * the year then the _hosts_ current year is assumed.  This will
     * cause dates to be radically wrong come the first of the year if
     * the sending host is in a different time zone in which the year
     * hasn't changed yet.
     */
    (void) time(&tmptt);
    (void) localtime_r(&tmptt, &tmptm);

    /* Setting this to -1 seems to cause mktime() to not make any
     * alterations to the time in the struct tm ... lets hope it works
     * on platforms other than linux 
     */
    tmptm.tm_isdst = -1; 

    /* First we need to figure out which style we are looking at */
    DPRINT(DBG_MSG, "MPARSE_time: parsing %.30s, len = %d", in, inlen);
    len = 0;
    if (inlen < 1) {
        /* It's not safe to check anything else ... but nothing needs to
         * be done to set an error either
         */
    }
    else if (isalpha(*in) ) {
        DPRINT(DBG_MSG, "MPARSE_time: parsing rfc3164 ts: %.30s", in);
        len = parse_rfc3164_ts(&tmptm, &tz, in, inlen);
    }
    else if (isdigit(*in) ) {
        DPRINT(DBG_MSG, "MPARSE_time: parsing rfc3339 ts: %.30s", in);
        len = parse_rfc3339_ts(&tmptm, &tz, in, inlen);
    }

    if (len > 0) {
        /* create our time_t */
        out->timestamp=mktime(&tmptm);
        out->tz = tz;

        /* check to make sure we're valid */
        if(-1==out->timestamp)
            len = 0;
    }

    if (len <= 0) {
        DPRINT(DBG_MSG,"MPARSE_time: invalid time, using current time",NO_VARG);

		(void)time(&(out->timestamp));
        out->tz = MSG_TZ_UNKNOWN;
        MSG_setTimeZone(out);

		return -1;
    }

    DPRINT(DBG_MSG, "MPARSE_time: end (%u, %-d)", out->timestamp, out->tz);
    return len;
}

bool 
MPARSE_strtoi(int *out, char const *str)
{
    char    *endptr;
    long    rtn;
    
    rtn = strtol(str, &endptr, 0);
    if (*endptr != '\0')
        return FALSE;
    if (rtn > INT_MAX || rtn < INT_MIN)
        return FALSE;

    *out = (int)rtn;

    return TRUE;
}

bool
MPARSE_strtoui(unsigned *out, char const *str)
{
    char    *endptr;
    long    rtn;

    rtn = strtol(str, &endptr, 0);
    if (*endptr != '\0')
        return FALSE;
    if (rtn < 0 || rtn > UINT_MAX)
        return FALSE;

    *out = (unsigned int)rtn;

    return TRUE;
}

bool
MPARSE_ip(netaddr_t *out, char const *str)
{
    char const  *p;
    bool        might_be_ip;

    *out = NULL;

    p = str;
    might_be_ip = TRUE;
    while (*p != '\0') {
        if (*p != '.' && *p != ':' && !isdigit(*p) ) {
            might_be_ip = FALSE;
            break;
        }
        p++;
    }

    if (! might_be_ip) 
        return FALSE;

    *out = NA_newINetLookup(str);

    if (NA_isError(*out) ) {
        NA_destroy(*out);
        return FALSE;
    }

    return TRUE;
}

static int
parse_rfc3339_ts(struct tm *tmout, int *tzout, char const *in, size_t inlen)
{
    int     num;
    int     inlen_sv;
    int     tzsign;

    inlen_sv = inlen;   /* save for later */

    /* Length check for date/time */
    if (inlen < 20) return 0;
    inlen -= 20;

    /* now convert the fields for time/date */
    DPRINT(DBG_MSG, "parse_rfc3339_ts: string = '%.40s'\n", in);

    /* year */
    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 1000;
    if (!isdigit(*in)) return 0; num += (*in++ - '0') * 100;
    if (!isdigit(*in)) return 0; num += (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    if (*in++ != '-')  return 0;
    tmout->tm_year = num - 1900;

    /* month */
    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    if (*in++ != '-')  return 0;
    tmout->tm_mon = num - 1;

    /* day */
    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    if (*in++ != 'T')  return 0;
    tmout->tm_mday = num;

    /* hour */
    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    if (*in++ != ':')  return 0;
    tmout->tm_hour = num;

    /* min */
    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    if (*in++ != ':')  return 0;
    tmout->tm_min = num;

    /* Sec */
    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    /* delimiter check is more involved ... so put it off */
    tmout->tm_sec = num;

    /* Check final time delimiter, and find tz start */
    if (*in == '.') {
        for (in++, inlen--; inlen > 0; in++, inlen--) {
            if (!isdigit(*in) ) {
                break;
            }
        }
    }
    if (inlen <= 0) return 0;

    tzsign = 0;
    if (*in == '-' )
        tzsign = -1;
    else if (*in == 'Z') {
        *tzout = 0;
        return TRUE;    /* Just parsed the timezone */
    }
    else if (*in != '+')
        return 0;   /* Invalid delimiter */
    
    /* Check timezone length */
    if (inlen < 6) return 0;
    inlen -= 6;
    in++;

    /* Timezone */
    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    if (*in++ != ':')  return 0;
    *tzout = num * 60;

    if (!isdigit(*in)) return 0; num  = (*in++ - '0') * 10;
    if (!isdigit(*in)) return 0; num += (*in++ - '0');
    *tzout += num;
    
    if (tzsign == -1) {
        /* -00:00 is an unknown time zone */
        if (*tzout == 0) *tzout = MSG_TZ_UNKNOWN;
        else *tzout = -(*tzout);
    }

    DPRINT(DBG_MSG, "parse_rfc3339_ts: end: %d-%.2d-%.2dT%.2d:%.2d:%.2d %-d", 
            tmout->tm_year+1900, tmout->tm_mon+1, tmout->tm_mday,
            tmout->tm_hour, tmout->tm_min, tmout->tm_sec, *tzout);

    return inlen_sv - inlen;
}

static int
parse_rfc3164_ts(struct tm *tmout, int *tzout, char const *in, size_t inlen)
{
	bool valid=TRUE;
	int read=0;
	int i=0;

    DPRINT(DBG_MSG, "MPARSE_time: start(\"%s\", %d)", in, inlen);

	if(in==NULL || inlen<16)
		valid=FALSE;

	if(valid)
	{
        /* Position 15 can be null */
		for(;read<15;read++)
		{
			if(in[read]=='\0')
			{
                DPRINT(DBG_MSG, "MPARSE_time: early null pos %d", read);
				valid=FALSE;
				break;
			}
		}

        /* Set read == 15 which is the last character of the normal
         * timestamp.  
         */
        read++;
	}
    read--; /* ignore the trailing space, NUL or whatever */

    DPRINT(DBG_MSG, "MPARSE_time: null check %d", valid);

	/* now we know that the length is ok, we do some advanced checking */
	if(valid)
	{
		for(i=0;i<16;i++)
		{
			/* ' ' seperators between fields */
			if( i==3 || i==6)
			{
				if(in[i]!=' ')
				{
					valid=FALSE;
					break;
				}
			}
            /* The last separator may be a space or a NIL */
            else if ( i==15 )
            {
                if (in[i] != ' ' && in[i] != '\0')  
                {
                    valid=FALSE;
                    break;
                }
            }
			/* : seperators in time */
			else if( i==9 || i==12 )
			{
				if(in[i]!=':')
				{
					valid=FALSE;
					break;
				}
			}
			/* stamps where the month is "08" are invalid; they
			 * must be " 8".
			 */
			else if( i==4 )
			{
				if( ! (
							in[i]==' ' || 
							(in[i]>='1' || in[i]<='9' )
					  )
				  )
				{
					valid=FALSE;
					break;
				}
			}
			/* month letters */
			else if( i<3 )
			{
				if( ! (
							(in[i]>='a' && in[i]<='z') ||
							(in[i]>='A' && in[i]<='Z')
					  )
				  )
				{
					valid=FALSE;
					break;
				}
			}
			/* finally, all the numbers */
			else 
			{
				if(in[i]<'0' || in[i]>'9')
				{
					valid=FALSE;
					break;
				}
			}
		}	/* end for */
	}

    DPRINT(DBG_MSG, "MPARSE_time: format check %d", valid);


	/* now, let's read in the fields.. ugh */
	if(valid)
	{
		/* month; messy */
		if(0==strncmp(in,"Jan",3))
			tmout->tm_mon=0;
		else if(0==strncmp(in,"Feb",3))
			tmout->tm_mon=1;
		else if(0==strncmp(in,"Mar",3))
			tmout->tm_mon=2;
		else if(0==strncmp(in,"Apr",3))
			tmout->tm_mon=3;
		else if(0==strncmp(in,"May",3))
			tmout->tm_mon=4;
		else if(0==strncmp(in,"Jun",3))
			tmout->tm_mon=5;
		else if(0==strncmp(in,"Jul",3))
			tmout->tm_mon=6;
		else if(0==strncmp(in,"Aug",3))
			tmout->tm_mon=7;
		else if(0==strncmp(in,"Sep",3))
			tmout->tm_mon=8;
		else if(0==strncmp(in,"Oct",3))
			tmout->tm_mon=9;
		else if(0==strncmp(in,"Nov",3))
			tmout->tm_mon=10;
		else if(0==strncmp(in,"Dec",3))
			tmout->tm_mon=11;
		else valid=FALSE;
    }
    /* The month is our best indicator to distinguish between something
     * that is a timestamp, and something that's not.
     */

    if (valid) {

		/* ranges for these were tested above */
		/* day of month */
		/* this entry, and only this entry, MUST be preceded by a ' '
		 * if it's a single digit number (i.e., " 7", not "07")
		 */
		if(in[4]!=' ')
			tmout->tm_mday=10*(in[4]-'0');
		else tmout->tm_mday=0;
		tmout->tm_mday+=in[5]-'0';

		/* hours */
		tmout->tm_hour=10*(in[7]-'0');
		tmout->tm_hour+=in[8]-'0';

		/* minutes */
		tmout->tm_min=10*(in[10]-'0');
		tmout->tm_min+=in[11]-'0';

		/* seconds */
		tmout->tm_sec=10*(in[13]-'0');
		tmout->tm_sec+=in[14]-'0';

        *tzout = MSG_TZ_UNKNOWN;
	}

    /* Make sure we return the error */
    if (!valid) {
        return -1;
    }

	return read;
}
