/**
 * @file msg_parse_cooked.c
 *
 * @brief Functions for parsing syslog COOKED messages.
 * @par
 * This implements two sets of functions, thoes called from the
 * syslog_cooked lexer and thoes defined in msg_parse_cooked.h
 *
 * @ingroup msg_parse_cooked
 *
 *
 * $Id: msg_parse_cooked.c,v 1.4.2.6 2003/04/30 20:03:27 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Aug 13 11:38:39 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include "strcompat.h"
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>

#include "types.h"

#if HAVE_LIBRR
/* Use glib Mutexs as long as glib is needed by librr */
/* When gthread.h is included before types.h there is an error, when
 * included below types.h there is no problem.  Can anyone tell JUST WHY
 * THE FUCK THAT IS?!?
 */
#include <glib/gthread.h>
#endif /* HAVE_LIBRR */

#include "syslog_cooked.h"
#include "msg_parse_cooked.h"
#include "msg_parse.h"

#include "log.h"
#include "debug.h"
#include "error.h"


static char const rcsid[]
    = "$Id: msg_parse_cooked.c,v 1.4.2.6 2003/04/30 20:03:27 devink Exp $";

/** Context for the current parsing job */
static struct MPCOOKED_ctx *curCtx = NULL;
#if HAVE_LIBRR
static GStaticMutex cooked_lock = G_STATIC_MUTEX_INIT;
#endif /* HAVE_LIBRR */

/* Public Functions from msg_parse_cooked.h */
bool
MPCOOKED_init(void)
{
    /* Nothing to do, but leaving it in incase */
    return TRUE;
}

void
MPCOOKED_free(struct MPCOOKED_ctx *ref)
{
	if(NULL==ref)
		return;

	switch(ref->type)
	{
		case MPC_PATH:
			if(NULL!=ref->data.path.fromFQDN)
			{
				free(ref->data.path.fromFQDN);
				ref->data.path.fromFQDN=NULL;
			}

			if(NULL!=ref->data.path.toFQDN)
			{
				free(ref->data.path.toFQDN);
				ref->data.path.toFQDN=NULL;
			}

			if(NULL!=ref->data.path.next)
			{
				PATH_free(ref->data.path.next);
				ref->data.path.next=NULL;
			}

			break;

		case MPC_ENTRY:
			MSG_free(&(ref->data.message));

			break;

		case MPC_IAM:
			if(NULL!=ref->data.iam.fqdn)
			{
				free(ref->data.iam.fqdn);
				ref->data.iam.fqdn=NULL;
			}

			break;

		case MPC_NONE:
			/* no type, no data */

			break;

		default:
			/* something's broken */

			break;
	}

	memset(ref,0,sizeof(struct MPCOOKED_ctx));
}

bool
MPCOOKED_parseString(struct MPCOOKED_ctx *out, char *buf, size_t sz)
{
    bool lexrtn;

    DPRINT(DBG_MSG, "MPCOOKED_parseString: start (sz = %d)", sz);

    /* setup the context object */
    memset(out, 0, sizeof(*out));
    out->type = MPC_NONE;
    out->error = FALSE;

    /* BEGIN critical section: 
     * This needs to be locked becasue lex is not re-entrant (and I
     * don't know how to make the functions that lex calls re-entrant.
     *
     * Locking is only needed to be done when librr forces threads.
     */
#if HAVE_LIBRR
    g_static_mutex_lock(&cooked_lock);
#endif /* HAVE_LIBRR */

    /* Setup the contex object now that it's safe */
    curCtx = out;

    lexrtn = MPCOOKED_lexString(buf, sz);

    /* Clear this just to be safe */
    curCtx = NULL;

#if HAVE_LIBRR
    g_static_mutex_unlock(&cooked_lock);
#endif /* HAVE_LIBRR */
    /* END critical section */

    if (out->type == MPC_ENTRY && out->data.message.msg == NULL) {
        size_t len = sz;
        DPRINT(DBG_MSG, "No Message Body", NO_VARG);
        if (buf[sz] != '\0') len++;

        out->data.message.msg = emalloc(len);

        strlcpy(out->data.message.msg, buf, len);
        if (len != sz) out->data.message.msg[sz] = '\0';

        out->data.message.msglen = len;
    }

    DPRINT(DBG_MSG, "MPCOOKED_parseString: end (yylex() == %d; error = %d)",
            lexrtn, out->error);

    return lexrtn == 0 && ! out->error;
}

void STag_iam(void)
{
    size_t  len;

    DPRINT(DBG_MSG, "STag_iam: start", NO_VARG);

    curCtx->type = MPC_IAM;

    len = strlen(A_iam_fqdn) + 1;
    curCtx->data.iam.fqdn = emalloc(len);
    strlcpy(curCtx->data.iam.fqdn, A_iam_fqdn, len);

    if (! MPARSE_ip(&(curCtx->data.iam.addr), A_iam_ip) ) {
        LOG_print(LOG_ERROR, "facility didn't contain a valid integer");
        curCtx->data.message.fac = RFC3195_DEFAULT_FAC;
        curCtx->error = TRUE;
    }

    curCtx->data.iam.type = A_iam_type;
}

void ETag_iam(void)
{
    /* Nothing to do */
}

void STag_entry(void)
{
    size_t  len;

    DPRINT(DBG_MSG, "STag_entry: start", NO_VARG);

    curCtx->type = MPC_ENTRY;

    if (! MPARSE_strtoi(&(curCtx->data.message.sev), A_entry_severity) ) {
        LOG_print(LOG_ERROR, "severity didn't contain a vaild integer");
        curCtx->data.message.sev = RFC3195_DEFAULT_SEV;
        curCtx->error = TRUE;
    }

    if (! MPARSE_strtoi(&(curCtx->data.message.fac), A_entry_facility) ) {
        LOG_print(LOG_ERROR, "facility didn't contain a vaild integer");
        curCtx->data.message.fac= RFC3195_DEFAULT_FAC;
        curCtx->error = TRUE;
    }

    len = strlen(A_entry_timestamp)+1;
    if (! MPARSE_time(&(curCtx->data.message), A_entry_timestamp, len)) {
        LOG_print(LOG_ERROR, "bad timestamp in <entry> tag.");
        curCtx->error = TRUE;
    }
    
	if (A_entry_hostname != AU_entry_hostname)
	{
		len = strlen(A_entry_hostname) + 1;
		curCtx->data.message.host = emalloc(len);
		strlcpy(curCtx->data.message.host, A_entry_hostname, len);
		curCtx->data.message.hostlen = len;
	}

    if (A_entry_deviceFQDN != AU_entry_deviceFQDN) {
        len = strlen(A_entry_deviceFQDN) + 1;
        curCtx->data.message.dfqdn = emalloc(len);
        strlcpy(curCtx->data.message.dfqdn, A_entry_deviceFQDN, len);
        curCtx->data.message.dfqdnlen = len;
    }

    if (A_entry_deviceIP != AU_entry_deviceIP) {
        if (! MPARSE_ip(&(curCtx->data.message.dip), A_entry_deviceIP) ) {
            LOG_print(LOG_ERROR, "bad IP in <entry> tag");
        }
    }

    if (A_entry_pathID != AU_entry_pathID) {
        if (! MPARSE_strtoui(&(curCtx->data.message.pathid), A_entry_pathID) ) {
            LOG_print(LOG_ERROR, 
                "pathID exists but doesn't contain a vaild integer");
            curCtx->error = TRUE;
        }
    }


    if (A_entry_tag != AU_entry_tag) {
        strlcpy(curCtx->data.message.tag, A_entry_tag, MSGTAG_LEN);
    }


} /* STag_entry */

void ETag_entry(void)
{
    size_t  len, i;
    char    *p, *pm;

    if (curCtx->error) {
        curCtx->error = FALSE;
        /* MPCOOKED_parseString() knows to put the entire buffer in the
         * ::msg field when type==MPC_ENTRY and ::msg==NULL. Thus we
         * turn off the error flag and let it ride.
         */
    }
    if (pcdata != NULL) { 
        len = strlen(pcdata) + 1;
        /* Note that the XML parser handles escaped characters. */
        curCtx->data.message.msg = emalloc(len);
        i = 0;
        pm = curCtx->data.message.msg;
        p = pcdata;
        while (*p != '\0' && i < len) {
            /* Turn all spaces into ' ' */
            if (isspace(*p) ) {
                *pm = ' ';
            }
            else {
                *pm = *p;
            }
            i++;
            pm++;
            p++;
        }
        *pm = '\0';
        curCtx->data.message.msglen = len;
    }
} /* ETag_entry */

void STag_path(void)
{
    struct path     *fillPath;
    int             id;

    DPRINT(DBG_MSG, "STag_path: start", NO_VARG);

    /* already hit the error, just give up now */
    if (curCtx->error) 
        return;

    /* We check to make sure that all fields are present, as all are
     * required.
     */

    if (A_path_pathID == AU_path_pathID) {
        curCtx->error = TRUE;
        return;
    }
    if (! MPARSE_strtoi(&id, A_path_pathID) ) {
        /* We are hosed, kill it and return an error */
        PATH_free(&(curCtx->data.path));
        curCtx->error = TRUE;
        return;
    }

    fillPath = &(curCtx->data.path);
    if (curCtx->type != MPC_NONE) {
        /* we have already found the first path so we now need to find
         * the next one to fill in.
         */
        while (fillPath->next != NULL) {
            fillPath = fillPath->next;
        }
        fillPath->next = emalloc(sizeof(struct path));
        memset(fillPath->next, 0, sizeof(struct path));

        fillPath = fillPath->next;
    }
    fillPath->id = id;


    curCtx->type = MPC_PATH;

    /* fromFQDN */
    if (A_path_fromFQDN != AU_path_fromFQDN) {
        fillPath->ffqdnlen = strlen(A_path_fromFQDN) + 1;
        fillPath->fromFQDN = emalloc(fillPath->ffqdnlen);
        strlcpy(fillPath->fromFQDN, A_path_fromFQDN, fillPath->ffqdnlen);
    }

    /* fromIP */
    if (A_path_fromIP == AU_path_fromIP) {
        curCtx->error = TRUE;
        return;
    }
    else if (! MPARSE_ip(&(fillPath->fromIP), A_path_fromIP)) {
        curCtx->error = TRUE;
        return;
    }

    /* toFQDN */
    if (A_path_toFQDN != AU_path_toFQDN) {
        fillPath->tfqdnlen = strlen(A_path_toFQDN) + 1;
        fillPath->toFQDN = emalloc(fillPath->tfqdnlen);
        strlcpy(fillPath->toFQDN, A_path_toFQDN, fillPath->tfqdnlen);
    }

    /* toIP */
    if (A_path_toIP == AU_path_toIP) {
        curCtx->error = TRUE;
        return;
    }
    else if (! MPARSE_ip(&(fillPath->toIP), A_path_toIP) ) {
        curCtx->error = TRUE;
        return;
    }

    /* linkprops */
    if (A_path_linkprops == AU_path_linkprops) {
        curCtx->error = TRUE;
        return;
    }
    else if (! PATH_mkLinkProps(fillPath, A_path_linkprops) ) {
        PATH_free(&(curCtx->data.path));
        curCtx->error = TRUE;
        return;
    }

} /* STag_path */

void ETag_path(void)
{
    /* Nothing to do */
} /* ETag_path */
