/**
 * @file msg_parse_cooked.h
 *
 * @brief Header for the syslog COOKED message format parser.
 * @ingroup msg_parse_cooked
 *
 * $Id: msg_parse_cooked.h,v 1.2.2.5 2003/04/30 20:03:27 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Aug 13 11:08:04 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MSG_MPCOOKED 
#define MSG_MPCOOKED

#include "util/netaddr.h"
#include "message.h"
#include "syslog_cooked.h"

/**@defgroup msg_parse_cooked Syslog Cooked Message Parser
 * @ingroup msg_parse
 * @{
 */

/** The possible types of syslog cooked messages */
typedef enum {
    MPC_PATH,   /**< path message */
    MPC_ENTRY,  /**< Syslog message */
    MPC_IAM,    /**< IAM (identifing) message */
    MPC_NONE    /**< No type yet (i.e. unknown) */
} MPCOOKED_TYPE;

/** Possible types for an IAM message */
typedef enum {
    IAM_DEVICE = A_iam_type_device,         /**< A source */
    IAM_RELAY = A_iam_type_relay,           /**< A relay */
    IAM_COLLECTOR = A_iam_type_collector    /**< A sink */
} MPC_IAM_TYPE;

/** Data for the IAM message type */
struct MPCOOKED_iam {
    char            *fqdn;  /**< FQDN */
    netaddr_t       addr;   /**< Address */
    MPC_IAM_TYPE    type;   /**< Type */
};

/** Context for the call to MPCOOKED_parseString(). */
struct MPCOOKED_ctx {
    MPCOOKED_TYPE   type;   /**< Type of message parsed */
    bool            error;  /**< Has a parsing error occured? */
    union {
        struct path         path;       /**< Data if a path was parsed */
        struct message      message;    /**< Data if an entry was parsed */
        struct MPCOOKED_iam iam;        /**< Data if an iam was parsed */
    } data;
};

/**
 * Initilize the syslog cooked parser.
 * @return TRUE on success.
 */
bool MPCOOKED_init(void);

/**
 * Parse a string containing a syslog-cooked message.  
 *
 * @param out The MPCOOKED_ctx object to fill in with data from a parse.
 * @param str The byte buffer to parse
 * @param sz The number of valid bytes in \c str (may include NULL)
 * @return TRUE on success, FALSE otherwise
 */
bool MPCOOKED_parseString(struct MPCOOKED_ctx *out, char *buf, size_t sz);

/**
 * Destroy an MPCOOKED_ctx structure; this only frees the (possible) subfields,
 * without attempting to free the actual MPCOOKED_ctx structure.
 * Because of this, passing in static variables by reference is just fine.
 *
 * @param ref The MPCOOKED_ctx structure to free the elements of
 */
void MPCOOKED_free(struct MPCOOKED_ctx *ref);

/**
 * Lex a string containing a syslog cooked message.  This does the
 * correct lex manipulation to make it take input from a byte buffer.
 *
 * @note This is defined in our modified flexml skeleton file.
 *
 * @param str The string to lex
 * @param len THe size of the string to lex.
 * @return 0 on success, non-zero otherwise.
 */
int MPCOOKED_lexString(char *buf, int len);

/*@}*/

#endif /* MSG_MPCOOKED */
