/**
 * @file msg_parse_rfc3164.c
 *
 * @brief This file implements message parsing for the old UDP/syslog classic
 * formatted strings
 * @ingroup msg_parse_rfc3164
 *
 * $Id: msg_parse_rfc3164.c,v 1.14.2.7 2003/10/22 04:09:03 devink Exp $
 *
 * @author	Matt Small <small@sdsc.edu>
 * @date	Mon Mar  4 14:38:30 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <ctype.h>

#include "types.h"
#include "error.h"
#include "const.h"

#include "msg_parse_rfc3164.h"
#include "msg_parse.h"

#include "message.h"
#include "strcompat.h"


static char const rcsid[] 
    = "$Id: msg_parse_rfc3164.c,v 1.14.2.7 2003/10/22 04:09:03 devink Exp $";

/** Parse the priority of a classic-syslog message
 * @internal
 * @param out Message pointer to modify
 * @param in The input string to be parsed
 * @param inlen The length of the input string
 * @return The number of characters read, or -1 on error
 */
static int 	MPOLD_parse_pri(struct message *out,char const *in,int inlen);

/** Parse the timestamp of a classic-syslog message
 * @internal
 * @param out Message pointer to modify
 * @param in The input string to be parsed
 * @param inlen The length of the input string
 * @return The number of characters read, or -1 on error
 * @todo Daylist savings (tm_isdst)?
 */
static int	MPOLD_parse_time(struct message *out,char const *in,int inlen);

/** Parse the hostname (or IP, in either IPv4 or IPv6) of a classic-syslog 
 * message
 * @internal
 * @param out Message pointer to modify
 * @param in The input string to be parsed
 * @param inlen The length of the input string
 * @return The number of characters read, or -1 on error
 * @todo Support IPv6 (see RFC 1884).
 */
static int	MPOLD_parse_host(struct message *out,char const *in,int inlen);

/** Parse the tag of a classic-syslog message
 * Unlike all other functions, this preserves whitespace if used as a seperator;
 * usually, we discard it, because it will be added back later.  Since the
 * TAG/MSG boundary has no specified seperator, we include the seperator in
 * the MSG, even if it's ' '.
 * @internal
 * @param out Message pointer to modify
 * @param in The input string to be parsed
 * @param inlen The length of the input string
 * @return The number of characters read, or -1 on error
 */
static int	MPOLD_parse_tag(struct message *out,char const *in,int inlen);

/** 
 * Parse the CONTENT part in an rfc3164 message.  This just copiens all
 * of \c in into \c out->msg, with the exception of CRLF terminators.
 * @param out The message strucutre to fill in
 * @param in The input message
 * @param inlen The length of the input string.
 * @param error TRUE if there was a parsing error earlier.
 * @return The number of characters read, or -1 on error.
 */
static int MPOLD_parse_msg(struct message *out, char const *in, size_t inlen, 
                            bool error);

/** 
 * find the first non space character in \c in.
 * @param in The input string.
 * @param inlen The length of in
 * @param read The number of characters that need to be skiped.
 */
#define SKIP_SPACE(in, inlen, read) \
    while (isspace(in[read]) && read < inlen) {read++;}

/** 
 * The tag to add to a message if it is not parsable.  This will be
 * added between the parsable and unparsable parts.
 */
#define MSG_UNPARSABLE_MARK  " UNPARSABLE"


int
MPOLD_parsein(struct message *out,char *in,int inlen, 
                bool ignoreHost, bool udpHacks)
{
	int     col;
    char    *str;
    int     left;

	/*
	 * Gotchas, not sure where they should go, but they're only useful for
	 * this function:
	 * <01> is invalid, and will generate a new priority by rfc 3164, sections
	 * 		4.1.1 and 4.3.3
	 */

	/* init our fields, in case we never write to them */
	out->host=NULL;		out->hostlen=0;
	out->msg=NULL;		out->msglen=0;
	out->tag[0]='\0';

	/* also, init the TCP cooked information, so that we don't accidentally
	 * run it */
	out->pathid=PATH_UNSET_ID;		
    out->dfqdnlen=0;
	out->dip=NULL;		
    out->dfqdn=NULL;

	/* we shouldn't transmit this packet any more */
	if(in==NULL || inlen<=1)
	{
		return FALSE;
	}

    str = in;
    left = inlen;

	/* priority */
	col=MPOLD_parse_pri(out,str,left);
	if(col<0) goto error_pri;
    str += col; left -= col;

    DPRINT(DBG_MSG, "MPOLD: priority: %d,%d", out->sev, out->fac);
    /* 
     * Note: udpHacks will treat invalid timestamps or hostnames as
     * simply a missig field, and not a hard error.
     *
     * Because of the way that certain other syslog client (linux, maybe
     * solaris) send their messages over the wire.  The host name will
     * be added by the UDP input module.
     */

	/* timestamp */
	col=MPOLD_parse_time(out,str,left);
    if(! udpHacks && col<0) goto error_time;
    else if (col > 0) {
        str += col; 
        left -= col;
    }

    DPRINT(DBG_MSG, "MPOLD: time: %lu", out->timestamp);

    /* host */
    if (!ignoreHost) {
        col=MPOLD_parse_host(out,str,left);
        if(!udpHacks && col<0) goto error_host;
        else if (col >= 0) {
            str += col; 
            left -= col;
        }
    }

    /* tag */
    col=MPOLD_parse_tag(out,str,left);
    if(col<0) goto error_tag;
    str += col; left -= col;

    DPRINT(DBG_MSG, "MPOLD: tag = '%s'", out->tag);

    /* msg */
    col = MPOLD_parse_msg(out, str, left, FALSE);
    left -= col;

    DPRINT(DBG_MSG, "MPOLD: message = (%d) '%s'", out->msglen, out->msg);

    /* and, i'm spent */
    return inlen - left;

    /* END NORMAL EXECUTION PATH */

    /* This is the unparsable hanlders.  When something becomes
     * unparsable it must jump to error_<pos>, where <pos> is the
     * current field.  These calls ignore returns and are just called 
     * to set any defaults, the last step is to copy all the message 
     * that was unparseable into the content (msg) field.
     *
     * It is expected that str will be begining of the unparsable
     * section; and left will be the unread charaters from in.
     *
     * This could be done with either an unparsable function and
     * positional constants, or chained functions, but I feel that gotos
     * are the most effecient and readable way.
     */
error_pri:
    MPOLD_parse_time(out, NULL, 0);
error_time:
    MPOLD_parse_host(out, NULL, 0);
error_host:
    MPOLD_parse_tag(out, NULL, 0);
error_tag:
    MPOLD_parse_msg(out, str, left, TRUE);

    DPRINT(DBG_MSG, "MPOLD: error, message = (%d) '%s'",out->msglen,out->msg);

    return -1;
}

static int
MPOLD_parse_pri(struct message *out,char const *in,int inlen)
{
	int read=0;
	bool valid=TRUE;
	int priority=0;

	if(inlen>0 && in[0]=='<')
	{
		/* inc first, so we don't include the '<' again */
		/* if this isn't true, we're going to fail the while loop below */
		/* very quickly, so we can let that take care of the 'else' */
		if(read+1<inlen)
			read++;

		while(read<inlen)
		{
			/* at this point, we should be done */
		 	if( in[read]=='>' )
				break;

			/* if we don't have a valid number, let's cause an error */
			if(! (in[read]>='0' && in[read]<='9') )
			{
				valid=FALSE;
				break;
			}

			/* calculate our priority */
			priority=priority*10+(in[read]-'0');

			/* let's move on to the next character */
			read++;
		}	/* end while */

		/* let's check our length */
		if( ! ( read>=2 && read<=4 ) )
			valid=FALSE;

		/* and, we still have to check the validity of the number */
		/* if there are leading 0's.. the only time a 0 may follow < is when */
		/* the priority is <0> */
		if( valid && in[1]=='0')
		{
			if( in[2]!='>' )
				valid=FALSE;
		}
	}
	/* if we don't start validly w/ a '<', we don't have a time stamp */
	else
	{
		valid=FALSE;
	}

	/* if our priority isn't readable, we generate a new one and copy
	 * the in[] into the MSG
	 */
	if(!valid)
	{
		priority=RFC3164_DEFAULT_PRI;

		/* calculate default facility, severity */
		out->sev=priority%8;
		out->fac=(priority - out->sev)/8;

		return -1;
	}
	else
	{
		out->sev=priority%8;
		out->fac=(priority - out->sev)/8;

		/* move to the start of msg */
		read++;

		return read;
	}
}

static int
MPOLD_parse_time(struct message *out,char const *in,int inlen)
{
    int     col;
    int     sps=0;

    if (in != NULL) { 
        SKIP_SPACE(in, inlen, sps);
    }

    col = MPARSE_time(out, in + sps, inlen - sps);
    if (col < 0) return -1;

	/* col will be < 0 if MPARSE_time failed */
	return col+sps;
}

static int
MPOLD_parse_host(struct message *out,char const *in,int inlen)
{
	bool valid=TRUE;
	int numdots=0;
	int numcolons=0;
	int read=0;
    int sps=0;
    int firstDot = -1;
	bool onlyNumbers=TRUE;
    bool startsWithNum=FALSE;

	if(in==NULL || inlen<=1)
		valid=FALSE;

	if(valid)
	{
        SKIP_SPACE(in, inlen, sps);
        in += sps;
        inlen -= sps;

        if (isdigit(in[read]) ) {
            startsWithNum = TRUE;
        }

		while( (in[read]!=' ') && (read<inlen) )
		{
			/* this still could be a dotted quad */
			if(in[read]=='.') {
                if (firstDot < 0) firstDot = read;
				numdots++;
            }
			/* and this could be an IPv6 addr */
			else if(in[read]==':')
				numcolons++;
			/* but if we're not doing address seperators, let's see if we're */
			/* doing numbers */
			else
			{
				if( onlyNumbers && isdigit(in[read]) )  
					onlyNumbers=TRUE;
				else 
					onlyNumbers=FALSE;
			}

			/* this could be valid, and just have no message, so trunc */
			/* the hostname here */
			if(in[read]=='\0')
				break;

			read++;
		}

		/* we've got a hostname (or FQDN) */
        if (!startsWithNum && numcolons == 0) 
            valid = TRUE;
		/* we've got IPv4 */
		else if(onlyNumbers && numdots==3)
			valid=TRUE;
        /* we've probably have an IPv6 */
        else if (onlyNumbers && numcolons > 0)
            valid = TRUE;
        /* We've got garbage */
		else valid=FALSE;
	}

	if(!valid)
	{
		out->hostlen=0;
		out->host=NULL;
		return -1;
	}
	else
	{
        /* Here we need to figure out if we have a hostname or FQDN */
        if (!startsWithNum && numdots > 0) {
            /* We have an FQDN, so put it in the dfqdn field, and put
             * the hostname (w/o the domain) in the host field
             */
            out->dfqdnlen = read + 1;
            out->dfqdn = emalloc(out->dfqdnlen);
            strlcpy(out->dfqdn, in, out->dfqdnlen);

            /* Okay, here the host len is only up to the first dot */
            out->hostlen = firstDot + 1;
        }
        else {
            /* hostlen is every thing we read. */
            out->hostlen=read+1;
        }
        out->host=emalloc(out->hostlen);
        strlcpy(out->host,in,out->hostlen);

		return read+sps;
	}

	/* should never be reached */
	return read+sps;
}

static int
MPOLD_parse_tag(struct message *out,char const *in,int inlen)
{
	int read=0;
    int i = 0;


    memset(out->tag,0,sizeof(out->tag));

	if(in==NULL || inlen<=1)
        return -1;

    SKIP_SPACE(in, inlen, read);

	while( (read<inlen) && (i < MSGTAG_LEN) )
	{
        /* We've reached the end */
		if(!isalnum(in[read]) ) 
            break;

        out->tag[i] = in[read];
        read++;
        i++;
	}

	return read; 
}

static int
MPOLD_parse_msg(struct message *out, char const *in, size_t inlen, bool error)
{
    char    *msg;
    size_t  wlen, msglen;

    DPRINT(DBG_MSG, "MPOLD_parse_msg: in = '%s', inlen = %d", in, inlen);

    /* strip trailing NULs */
    while (in[inlen-1] == '\0') inlen--;
    /* string trailing CR/LF */
    while (in[inlen-1] == '\n' || in[inlen-1] == '\r') inlen--;

    if (inlen < 1) {
        out->msglen = 0;
        out->msg = NULL;
        return 0;   /* Not really an error */
    }

    msglen = inlen + 1;

    /* Factor in the unparsable tag to the message length */
    if (ENABLE_UNPARSABLE_MARK) 
    {
        if (error) 
            msglen += strlen(MSG_UNPARSABLE_MARK);
    }

    msg = emalloc(msglen);

    /* set up for return */
    out->msglen = msglen;
    out->msg = msg;

    /* Add the unparsable tag into message if needed */
    if (ENABLE_UNPARSABLE_MARK)
    {
        if (error) {
            wlen = strlcpy(msg, MSG_UNPARSABLE_MARK, msglen);
            if (wlen <= 0 || wlen >= msglen) 
                goto error;
            msglen -= wlen;
            msg += wlen;
        }
    }
    wlen = strlcpy(msg, in, msglen);
    DPRINT(DBG_MSG, "MPOLD_parse_msg: after copy: wlen = %d, msglen = %d",
            wlen, msglen);
    /* No check for truncation b/c 'in' may not have been null
     * terminated, or may contain extra \n or \r characters.  In short
     * we trust inlen.
     */
    if (wlen < 0) 
        goto error;

    /* alls well that ends well */
    return inlen;

    /* Error handler */
error:
    free(out->msg);
    out->msglen = 0;
    out->msg = NULL;
    return -1;
}
