/**
 * @file omod_beep.c
 *
 * @brief Output module for sending messages over a BEEP link
 *
 * $Id: omod_beep.c,v 1.1.2.9 2003/10/22 04:09:04 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Aug 20 09:14:42 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#if HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "strcompat.h"
#include "error.h"
#include "types.h"
#include "privs.h"
#include "log.h"

#include "util/bitvec.h"

#include "omod_init.h"
#include "outputmod.h"

#include "message/message.h"
#include "message/msg_frmt.h"

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif

static char const rcsid[] 
    = "$Id: omod_beep.c,v 1.1.2.9 2003/10/22 04:09:04 devink Exp $";

/**@defgroup omod_beep BEEP Output
 * @ingroup omod
 * @{
 */

/** Initialize the BEEP Raw output module
 * @param options The options fields that we need to parse to initialize
 * @return a pointer to the state struct (must be free()'d)
 */
static struct OMOD_state* OMOD_BP_init(struct OMOD_opt *options);

/** Destroy the BEEP Raw output module indicated by state
 * @param state The state of the file output module to destroy
 */
static void OMOD_BP_destroy(struct OMOD_state *state);

/** Re-initialize the output module; useful for SIGHUP
 * @param state The state of the pre-SIGHUP output module
 * @todo Take down RR, and bring it back up?
 */
static void OMOD_BP_reinit(struct OMOD_state *state);


/** Sends a message (in this case, sends it across the network)
 * Obeys RFC 3195.
 * @note This function does not support sending multiple messages inside of a
 * 		single ANS reply; this is optional, and would require changes to the
 * 		formatting code.
 *
 * @param state The state of this output module
 * @param message The message to be send
 */
static void OMOD_BP_send_msg(struct OMOD_state *state,
		const struct message *msg);

/*@}*/

/* Private functions */

/**
 * Do a BEEP send with retries. 
 *
 * @param state The output module state.
 * @param buf The buffer containing the message to send.
 */
static bool doBEEPSend(struct OMOD_state *state, char *buf);

/**
 * Connect to a BEEP server.  This was pulled out so that it could be
 * shared in init() and reinit().
 */
static bool doBEEPConnect(struct OMOD_state *state);

/**
 * build an iam object
 */
static bool getIamData(int sock, struct iam *iamdata);

/** The Maximum number of times doBEEPSend() will try to reconnect */
#define MAX_RECONN_TRIES    1

void
OMOD_BP_iface_init(OMOD_interface *iface)
{
	iface->INIT=OMOD_BP_init;
	iface->REINIT=OMOD_BP_reinit;
	iface->DESTROY=OMOD_BP_destroy;
	iface->SEND_MSG=OMOD_BP_send_msg;
	iface->state=NULL;
}

bool
OMOD_BP_process_opts(struct OMOD_opt *opts, char const *f, int l)
{
    /* Set the default formatter */
    if (opts->format.type == MFRMT_NOEXIST) {
        opts->format.type = MFRMT_RFC3164;
    }
    /* always add the priority to the output */
    BM_SET(opts->format.options, MFOPT_PRI);

    /* always add mime headers to the output */
    BM_SET(opts->format.options, MFOPT_MIMEHEAD);

    if (BM_ISSET(opts->format.options, MFOPT_TSTZ)
        || BM_ISSET(opts->format.options, MFOPT_TSYEAR)
        || BM_ISSET(opts->format.options, MFOPT_TSEXT))
    {
        LOG_print(LOG_INFO, 
            "%s:%d: Format options create a time stamp which no known"
            " syslogd parses (this one included), using rfc3339 instead",
            f, l);
        BM_CLR(opts->format.options, MFOPT_TSTZ); 
        BM_CLR(opts->format.options, MFOPT_TSYEAR); 
        BM_CLR(opts->format.options, MFOPT_TSEXT); 
        BM_SET(opts->format.options, MFOPT_RFC3339);
    }

    if (opts->outPath == NULL) {
        LOG_print(LOG_ERROR, "%s:%d: Remote log host unspecified", f, l);
        return FALSE;   /* fatal */
    }

    opts->outAddr = NA_newINetLookup(opts->outPath);
    if (NA_isError(opts->outAddr) ) { 
        LOG_print(LOG_ERROR, 
            "%s:%d: Invalid address (%s) for remote log host", f, l);
        return FALSE;   /* fatal */
    }

    if (opts->outPort == -1) {
        opts->outPort = SYSLOG_CONN_PORT;
    }

    return TRUE;
}

static struct OMOD_state*
OMOD_BP_init(struct OMOD_opt *options)
{
	struct OMOD_state *state;
	
	state=emalloc(sizeof(struct OMOD_state));
	memset(state,0,sizeof(struct OMOD_state));

    /* Copy the foramtter options */
    state->format.type = options->format.type;
    state->format.options = options->format.options;

    /* Some assertions */
    assert(state->format.type != MFRMT_NOEXIST);
    assert(BM_ISSET(state->format.options, MFOPT_PRI));
    assert(BM_ISSET(state->format.options, MFOPT_MIMEHEAD));

	/* copy the netaddr */
	state->sendToAddr=NA_newINetLookup(options->outPath);

	if(! NA_mkSockaddr(state->sendToAddr, options->outPort, &(state->sendTo)) )
	{
		LOG_print(LOG_ERROR, "NA_mkSockaddr() failed");
		return NULL;
	}

    /* make the connection */
    if (! doBEEPConnect(state) ) {
        /* Error was already logged */
        free(state);
        return NULL;
    }

	/* copy the name */
    assert(options->outPath != NULL);
    state->namelen=strlen(options->outPath)+1;
    state->name=emalloc(state->namelen);
    strlcpy(state->name,options->outPath,state->namelen);
    
	/* set the error to false */
	state->error=FALSE;

	/* get the path cache ready for use */
	DPRINT(DBG_OMD, 
        "OMOD_BP_init(): create the pcache through PCACHE_regOutputMod(%X)",
        options->pcache);
	state->pcache=options->pcache;
	state->pcache_omID=PCACHE_regOutputMod(options->pcache);

	DPRINT(DBG_OMD, "OMOD_BP_init(): initialization complete", NO_VARG);


	return state;
}

static void
OMOD_BP_reinit(struct OMOD_state *state)
{
    /* Reconnect if needed */
    if (state->beep_conn == NULL || BEEP_NOTCON == state->beep_conn->state) { 
        doBEEPConnect(state);
        /* Clear the record of sent paths in the cache */
        PCACHE_omNewConn(state->pcache, state->pcache_omID);
    }

}

static void
OMOD_BP_destroy(struct OMOD_state *state)
{
	BEEP_IN_shutdown(state->beep_conn);

	close(state->filedes);

	PCACHE_unregOutputMod( state->pcache, state->pcache_omID );

	free(state->name);
    NA_destroy(state->sendToAddr);

	free(state);

	DPRINT(DBG_OMD, "OMOD_BP_destroy(): finished", NO_VARG);
}

static void
OMOD_BP_send_msg(struct OMOD_state *state,const struct message *msg)
{
	int		len, buflen;
	char	*msgbuf,*pathbuf;

	char 	*host;	/* leave malloc()d, because path sets should be rare */
	struct hostent *hostentry;

	uint32_t	pathID;
	struct path	*sendpath;

	DPRINT(DBG_OMD, "OMOD_BP_send_msg() called w/ message: '%s'",msg->msg);

    msgbuf = NULL;
    pathbuf = NULL;
    buflen = 0;

	/* see if we need to send the path */
	if(MFRMT_hasPathFormatter(&(state->format)))
	{
		DPRINT(DBG_OMD, "Trying to get path w/ connid: %d, pathid: %d, omID: %d",
				msg->connid, msg->pathid, state->pcache_omID);
		pathID=PCACHE_findOutgoing(state->pcache, msg->connid, msg->pathid,
				state->pcache_omID);
		if(PATH_UNSET_ID==pathID)
		{
			DPRINT(DBG_OMD, "Trying to create and send path id %d", pathID);
			sendpath=PCACHE_makeOutgoing(state->pcache, msg->connid, 
                    msg->pathid, state->pcache_omID);


			if(NULL!=sendpath)
			{
                /* fill in the sendpath */
                sendpath->toIP=NA_newNetaddr(state->sendToAddr);

                hostentry=gethostbyname(NA_getCharAddr(sendpath->toIP));
                sendpath->tfqdnlen=strlen(hostentry->h_name)+1;
                sendpath->toFQDN=emalloc(sizeof(char)*sendpath->tfqdnlen);
                strlcpy(sendpath->toFQDN, hostentry->h_name, 
                        sendpath->tfqdnlen);

                host=emalloc(sizeof(char)*HOST_NAME_MAX);
                if (gethostname(host, HOST_NAME_MAX) < 0) {
                    /* Don't want to add more logging, as it just fucks
                     * up in this output module.
                     * 
                    LOG_eprint("Failed to get our host name:", errno);
                     */
                    DPRINT(DBG_OMD, "gethostname() failed: %d", errno);
                    hostentry = NULL;
                }
                else {
                    hostentry = gethostbyname(host);
                }
                free(host);

                if (hostentry != NULL) {
                    DPRINT(DBG_DBG, "gethostbyname() said name: %s",
                            hostentry->h_name);

                    sendpath->ffqdnlen=strlen(hostentry->h_name)+1;
                    sendpath->fromFQDN=emalloc(sizeof(char)*sendpath->ffqdnlen);
                    strlcpy(sendpath->fromFQDN, hostentry->h_name, 
                            sendpath->ffqdnlen);

                    sendpath->fromIP=NA_newINetLookup(sendpath->fromFQDN);
                }
                else {
                    sendpath->ffqdnlen = 0;
                    sendpath->fromFQDN = NULL;
                    DPRINT(DBG_OMD, "(check) gethostbyname() failed: (%d) %s", 
                            h_errno, hstrerror(h_errno));
                }

                /* try to format it */
				DPRINT(DBG_OMD, "Creating path to send from %s to %s",
						sendpath->fromFQDN, sendpath->toFQDN);
				buflen=RFC3195_START_MSGLEN;

                /* Note that pathbuf will be freed by librr once it has
                 * been sent out.
                 */
				pathbuf=emalloc(sizeof(char)*buflen);
				len = MFRMT_formatPath(pathbuf, buflen, sendpath,
						&(state->format));
				while(len==buflen)
				{
					free(pathbuf);
					buflen*=2;
					if(buflen>RFC3195_MAX_MSGLEN)
					{
						pathbuf=NULL;
						len=0;
						break;
					}

					pathbuf=emalloc(sizeof(char)*buflen);

					len = MFRMT_formatPath(pathbuf, buflen, sendpath,
							&(state->format));
				}

				if(len==0)
				{
					LOG_print(LOG_INFO, "Failed to format BEEP Cooked Path");
				}
				else
				{
					DPRINT(DBG_OMD, 
                            "OMOD_BP_send_msg(): Sending path: %s (len: %d)",
							pathbuf, len);
                    /* Overload len, so that we can cleanup the path
                     * first 
                     */
                    len = doBEEPSend(state, pathbuf);
				}
                /* sendpath cleanup */
                free(sendpath->fromFQDN);
                free(sendpath->toFQDN);
                NA_destroy(sendpath->fromIP);
                NA_destroy(sendpath->toIP);
                free(sendpath);

                /* give up if the send above failed */
                if (!len) 
                    return;

				DPRINT(DBG_OMD, "Done sending and creating path id %d", pathID);
			}
		}
	}

	buflen=RFC3195_START_MSGLEN;
    /* note that msgbuf will be freed by librr once it has been sent out
     */
	msgbuf=emalloc(sizeof(char)*buflen);

	len = MFRMT_format(msgbuf, buflen, msg, &(state->format));
	DPRINT(DBG_OMD, "Formatted message, len: %d, max: %d", len, buflen);
	while(len==buflen)
	{
		free(msgbuf);
		buflen*=2;
		if(buflen>RFC3195_MAX_MSGLEN)
		{
			len=0;
			msgbuf=NULL;
			break;
		}

		msgbuf=emalloc(sizeof(char)*buflen);

		len = MFRMT_format(msgbuf, buflen, msg, &(state->format));
		DPRINT(DBG_OMD, "Formatted message (loop), len: %d, buflen: %d",
				len, buflen);
	}
	if(len<=0)
	{
		LOG_print(LOG_INFO, "Failed to format BEEP message (format type: %d)"
				" message: '%s'",
				state->format.type, msg->msg);
        return;
	}

    /* Ignore the return value, On error this will handle the error
     * messoge.
     */
	doBEEPSend(state, msgbuf);
}

/*-------------------------- Private Functions --------------------------*/

static bool
doBEEPSend(struct OMOD_state *state, char *buf) 
{
    bool        error;  /* set to true when a fatal error occurs */
    int         tries;
    BEEP_conn   *beep_conn;

    /* Warning, OMOD_BP_reinit() may change state->beep_conn, but I'm
     * sick of typing the whole expression ... just be careful.
     */
    beep_conn = state->beep_conn;

    error = FALSE;
    tries = 0;
    while (! error) {
        if((NULL != beep_conn && BEEP_NOTCON != beep_conn->state)
            && BEEP_sendLog(beep_conn, buf)) 
        {
            /* send was successful, so we are done */
            break;
        }
        error = TRUE;

        /* send failed */
        if (! state->beep_droped) {
            state->beep_droped = TRUE;
            LOG_print(LOG_INFO, "BEEP Connection to %s Dropped", state->name);
        }

        if ((NULL == beep_conn || BEEP_NOTCON == beep_conn->state) 
            && tries < MAX_RECONN_TRIES) 
        {
            DPRINT(DBG_OMD, "doBEEPSend: trying to reconnect", NO_VARG);
            /* Try to reconnect */
            OMOD_BP_reinit(state);

            /* state->beep_conn may have been changed */
            beep_conn = state->beep_conn;
            error = FALSE;
        }
        tries++;
    }

    if (state->beep_droped && !error) {
        state->beep_droped = FALSE;
        LOG_print(LOG_INFO, "BEEP Connection to %s re-established", 
                state->name);
    }
    /* This gets annoying, but I may want to bring it back later.
    else if (tries >= MAX_RECONN_TRIES && error) {
        LOG_print(LOG_ERROR, "Failed to reconnect to server %s",
                state->name);
    }
    */

    /* We have to free the message buffer now */
    free(buf);

    return !error;      /* TRUE on success */
}

static bool
doBEEPConnect(struct OMOD_state *state)
{
    int             fd;     /* needed due to recursion by LOG_* */
    struct iam      iamdata;

	/* open the socket */
	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(fd < 0)
	{
		LOG_eprint("Failed socket() call",errno);
        return FALSE;
	}

	DPRINT(DBG_OMD, "doBEEPConnect: attempting to connect to the socket",
			NO_VARG);

	/* connect the socket */
	if( connect(fd, (struct sockaddr *)&(state->sendTo), 
                sizeof(state->sendTo)) < 0 )
	{
		LOG_eprint("Failed connect() call",errno);
		close(fd);
		return FALSE;
	}

	DPRINT(DBG_OMD, "doBEEPConnect: starting the BEEP layer", NO_VARG);

    /* get data for iam element */
    getIamData(fd, &iamdata);

	/* init the RR layer */
	state->beep_conn = BEEP_IN_connect(fd, &(state->format), &iamdata);
    MSG_iamFree(&iamdata);  // cleanup
	if( NULL == state->beep_conn)
	{
        DPRINT(DBG_OMD, "doBEEPConnect: BEEP start unsuccessful", NO_VARG);
		return FALSE;
	}

	DPRINT(DBG_OMD, "doBEEPConnect: BEEP start successful",NO_VARG);

    state->filedes = fd;

    return TRUE;
}

static bool
getIamData(int sock, struct iam *iamData) {
    struct sockaddr_in  addr;
    socklen_t           alen;
    struct hostent      *hp;

    if (getsockname(sock, (struct sockaddr*)&addr, &alen) < 0) {
        return FALSE;
    }

    if (PF_INET != addr.sin_family) {
        return FALSE;
    }
    iamData->ip = NA_newINet4(addr.sin_addr);

    hp = gethostbyaddr((char*) &(addr.sin_addr), 
            sizeof(addr.sin_addr), AF_INET);
    if (NULL == hp) {
        NA_destroy(iamData->ip);
        return FALSE;
    }
    /* not going to include the header just for this inet_ntoa 
     * DPRINT(DBG_OMD, "Got iam ip = %s", inet_ntoa(addr.sin_addr));
     */
    DPRINT(DBG_OMD, "Got iam fqdn = %s", hp->h_name);

    iamData->fqdnlen = strlen(hp->h_name) + 1;
    iamData->fqdn = emalloc(iamData->fqdnlen);
    strlcpy(iamData->fqdn, hp->h_name, iamData->fqdnlen);

    /* we may be just blantently lying */
    iamData->type = MSGDT_DEVICE;

    return TRUE;
}
