/**
 * @file omod_fread.c
 *
 * @brief Implement all the output modules which work by writing to a
 * file.
 * @par
 * Because there are a lot of interesting things that can be accoplished
 * by writing to a file this file actually implements several output
 * modules which barely differ.
 * The current list of output modules is:
 * \li Normal File
 * \li Console
 *
 * @ingroup omod_file
 * @ingroup omod_cons
 *
 *
 * $Id: omod_fread.c,v 1.12.2.7 2003/04/30 20:03:31 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Aug 15 16:21:09 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>		/* for snprintf() */

#include "strcompat.h"
#include "error.h"
#include "types.h"
#include "privs.h"
#include "log.h"

#include "omod_init.h"
#include "outputmod.h"

#include "message/message.h"
#include "message/msg_frmt.h"

static char const rcsid[] 
    = "$Id: omod_fread.c,v 1.12.2.7 2003/04/30 20:03:31 devink Exp $";

/**@defgroup omod_file File Output Module
 * @ingroup omod
 * @{
 */

/** Initialize the readable file output module
 * @param options The options fields that we need to parse to initialize
 * @return a pointer to the state struct (must be free()'d)
 */
static struct OMOD_state * OMOD_FR_init(struct OMOD_opt *options);


/** Destroy the readable file output module indicated by state
 * @param state The state of the file output module to destroy
 */
static void OMOD_FR_destroy(struct OMOD_state *state);


/** Re-initialize the output module; useful for SIGHUP
 * @param state The state of the pre-SIGHUP output module
 * @param options The options fields that we need to parse to re-initialize
 */
static void OMOD_FR_reinit(struct OMOD_state *state);


/** Sends a message (in this case, writes it to disk)
 * Obeys RFC 3164 and truncates messages to 1024 characters (including the
 * '\0')
 * @param state The state of this output module
 * @param message The message to be send
 */
static void OMOD_FR_send_msg(struct OMOD_state *state,
		const struct message *msg);


/*@}*/

/**@defgroup omod_cons Console Output Module
 * @ingroup omod
 * @{
 */

/**
 * Initilize the Console output module.
 * Because the Console output module uses most of the same functions on
 * the file output module, all this does is set outPath to the correct
 * path for the console device and call OMOD_FR_init().
 *
 * @param options The output module options 
 * @return a pointer to the state struct 
 */
static struct OMOD_state* OMOD_CONS_init(struct OMOD_opt *options);

/*@}*/

/** Options to use on output log files */
#define OUTPUT_OPTS (O_WRONLY | O_CREAT | O_SYNC | O_APPEND | O_NOCTTY)

void
OMOD_FR_iface_init(OMOD_interface *iface)
{
	iface->INIT=OMOD_FR_init;
	iface->REINIT=OMOD_FR_reinit;
	iface->DESTROY=OMOD_FR_destroy;
	iface->SEND_MSG=OMOD_FR_send_msg;
	iface->state=NULL;
}

void
OMOD_CONS_iface_init(OMOD_interface *iface)
{
    iface->INIT=OMOD_CONS_init;

    /* All the other functions are the same as a file output module */
    iface->REINIT=OMOD_FR_reinit;
    iface->DESTROY=OMOD_FR_destroy;
    iface->SEND_MSG=OMOD_FR_send_msg;
    iface->state = NULL;
}

bool
OMOD_FR_process_opts(struct OMOD_opt *opts, char const *f, int l)
{
    /* Make sure that we have an output path */
    if (opts->outPath == NULL) {
        LOG_print(LOG_ERROR, "%s:%d: output-path left unspecified", f, l);
        return FALSE;   /* fatal error */
    }

    if (opts->format.type == MFRMT_NOEXIST) {
        opts->format.type = MFRMT_RFC3164;
    }

    return TRUE;
}

bool
OMOD_CONS_process_opts(struct OMOD_opt *opts, char const *f, int l) 
{
    if (opts->outPath != NULL || opts->outList != NULL) {
        LOG_print(LOG_ERROR, 
            "%s:%d: Console output module does not allow changing the path"
            " to the console, please remove the output-path options", f, l);
        opts->outPath = NULL;
        LL_destroy(&(opts->outList));
    }

    if (opts->format.type == MFRMT_NOEXIST) {
        opts->format.type = MFRMT_RFC3164;
    }

    return TRUE;
}

static struct OMOD_state*
OMOD_CONS_init(struct OMOD_opt *options)
{
    size_t  len;

    len = strlen(CONSOLE_PATH)+1;
    options->outPath = emalloc(len);
    strlcpy(options->outPath, CONSOLE_PATH, len);

    return OMOD_FR_init(options);
}


static struct OMOD_state*
OMOD_FR_init(struct OMOD_opt *options)
{
	struct OMOD_state *state;
	
	state=emalloc(sizeof(struct OMOD_state));
	
	/* open the file using the passed in options */
	state->filedes 
        = PRIV_openWriteFile(options->outPath, OUTPUT_OPTS);

	if(state->filedes<0)
	{
		free(state);
        return NULL;
	}

	/* copy the name */
    assert(options->outPath != NULL);
    state->namelen=strlen(options->outPath)+1;
    state->name=emalloc(state->namelen);
    strlcpy(state->name,options->outPath,state->namelen);

    /* Copy the formatter info, and set the default if not specified */
    state->format.type = options->format.type;
    state->format.options = options->format.options;

    DPRINT(DBG_OMD, "formatter = %d; options = 0x%x", 
            state->format.type, state->format.options);

	/* set the error to false */
	state->error=FALSE;

	return state;
}

static void
OMOD_FR_reinit(struct OMOD_state *state)
{
    close(state->filedes);

    state->filedes = PRIV_openWriteFile(state->name, OUTPUT_OPTS);
    if (state->filedes < 0) {
        state->error = TRUE;
    }
}

static void
OMOD_FR_destroy(struct OMOD_state *state)
{
	close( state->filedes );

	free( state->name );

	free(state);
}

static void
OMOD_FR_send_msg(struct OMOD_state *state,const struct message *msg)
{
	int nwritten,msgbuflen;
	char msgbuf[RFC3164_MAX_MSGLEN];

    DPRINT(DBG_OMD, "OMOD_FR_send_msg: start", NO_VARG);

	msgbuflen=MFRMT_format(msgbuf, RFC3164_MAX_MSGLEN, msg, &(state->format));

	/* if we don't have a real message to write, let's not write it */
	if(msgbuflen<=0)
		return;

	/* write everything except for the trailing '\0'; we know by above that 
	 * msgbuf will be at least 1.  Instead, we replace the '\0' with a '\n'
	 */
	msgbuf[msgbuflen-1]='\n';
	nwritten=write(state->filedes,msgbuf,msgbuflen);

	if(nwritten<0)
		state->error=TRUE;

    DPRINT(DBG_OMD, "OMOD_FR_send_msg: end", NO_VARG);
}
