/**
 * @file omod_udp.c
 *
 * @brief Output Module for sending messages over a UDP socket.
 * @par
 * This implements the compatibility output module that can send
 * messages to old style RFC3164 syslog daemons.
 *
 * $Id: omod_udp.c,v 1.3.2.7 2003/04/30 20:03:32 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jul 24 17:26:10 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
# include "acconfig.h"
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>


#include "const.h"
#include "log.h"
#include "error.h"
#include "privs.h"
#include "strcompat.h"

#include "util/bitvec.h"

#include "omod_init.h"
#include "outputmod.h"

#include "message/message.h"
#include "message/msg_frmt.h"

static char const rcsid[] 
    = "$Id: omod_udp.c,v 1.3.2.7 2003/04/30 20:03:32 devink Exp $";

/**@defgroup omod_udp UDP Output Module
 * @ingroup omod
 * @{
 */

/**
 * Intilize the UDP output module
 *
 * @param options The options for this output module.
 * @return A pointer to the output module state struct.
 */
static struct OMOD_state* OMOD_UDP_init(struct OMOD_opt *options);

/**
 * Destroy the UDP Output module state.
 * @param the OMOD_state object to destory.
 */
static void OMOD_UDP_destroy(struct OMOD_state *state);

/**
 * Send a message through the UDP output module.
 *
 * @param state The OMOD_state object for this OM.
 * @param msg The message object to send.
 */
static void OMOD_UDP_send_msg(struct OMOD_state *state,
		const struct message *msg);

/**
 * Re-initilize the ouput modules.
 * @note This is just a no-op at the moment, as there is nothing to
 * reinitilize.
 *
 * @param state The current state of this OM.
 */
static void OMOD_UDP_reinit(struct OMOD_state *state);

/*@}*/

/**
 * Create the shared UDP socket to send from.  
 * All instances of the UDP output module share the socket that the send
 * messages over.  This will create the socket if need be, return the
 * socket to use and update the reference count.
 *
 * @return The socket to send on or < 0 on error.
 */
static int openSocket(void);

/**
 * Close the shared UDP socket.  This will decrement the reference count
 * and close the socket when it reaches 0.
 */
static void closeSocket(void);

/** The shared UDP socket */
static int gSocket = -1;

/** The refrence count for gSocket */
static int gSockRefcount = 0;

void
OMOD_UDP_iface_init(OMOD_interface *iface)
{
    iface->INIT = OMOD_UDP_init;
    iface->REINIT = OMOD_UDP_reinit;
    iface->DESTROY = OMOD_UDP_destroy;
    iface->SEND_MSG = OMOD_UDP_send_msg;
    iface->state = NULL;
} 

bool
OMOD_UDP_process_opts(struct OMOD_opt *opts, char const *f, int l)
{
    if (opts->outPath == NULL) {
        LOG_print(LOG_ERROR, "%s:%d: remote log host not specified", f, l);
        return FALSE;   /* fatal */
    }

    opts->outAddr = NA_newINetLookup(opts->outPath);
    if (NA_isError(opts->outAddr) ) {
        LOG_print(LOG_ERROR, "%s:%d: Invalid address (%s) for remote log host",
                f, l, opts->outPath);
        return FALSE;   /* fatal */
    }

    /* Check the formatter type */
    if (opts->format.type == MFRMT_NOEXIST) {
        opts->format.type = MFRMT_RFC3164;
    }
    else if (opts->format.type != MFRMT_RFC3164) {
        LOG_print(LOG_ERROR, "%s:%d: udp output module can not be used with" 
                " any formatter other than rfc3164");
        return FALSE; /* fatal */
    }
    /* force the use of the priority option */
    BM_SET(opts->format.options, MFOPT_PRI);

    if (opts->outPort == -1) {
        opts->outPort = SYSLOG_UDP_PORT;
    }

    /* Consider checking outPath here with NA_newINetLookup. */
    return TRUE;
}


static struct OMOD_state *
OMOD_UDP_init(struct OMOD_opt *options)
{
    struct OMOD_state   *state;
    union NA_sockaddr   addr;

    DPRINT(DBG_OMD, "om-udp: init: start", NO_VARG);

    state = NULL;

    if (! NA_mkSockaddr(options->outAddr, options->outPort, &addr) ) {
        /* there is no _good_ reason this should fail */
        LOG_print(LOG_ERROR, "NA_mkSockaddr() failed");
        return NULL;
    }

    DPRINT(DBG_OMD, "om-udp: init: send to addr %s:%d, family %d",
            inet_ntoa(addr.sin.sin_addr), ntohs(addr.sin.sin_port),
            addr.sin.sin_family);

    state = emalloc(sizeof(struct OMOD_state));
    memset(state, 0, sizeof(struct OMOD_state));

    state->filedes = openSocket();
    if (state->filedes < 0) 
        return NULL;

    state->sendTo = addr;

    /* copy the formatter info, and check for sanity */
    state->format = options->format;

    /* Done, state will be NULL if there was an error */
    return state;
}

static void
OMOD_UDP_destroy(struct OMOD_state *state)
{
    closeSocket();
    free(state);
}

static void
OMOD_UDP_send_msg(struct OMOD_state *state, const struct message *msg)
{
    char    msgbuf[RFC3164_MAX_MSGLEN];
    int     len;

    DPRINT(DBG_OMD, "om-upd: send_msg: start", NO_VARG);
    DPRINT(DBG_OMD, "om-udp: send_msg: to %s:%d, family %d", 
            inet_ntoa(state->sendTo.sin.sin_addr), 
            ntohs(state->sendTo.sin.sin_port),
            state->sendTo.sin.sin_family);

    /* format the message */
    len = MFRMT_format(msgbuf, RFC3164_MAX_MSGLEN, msg, &(state->format));
    if (len <= 0) 
        return;     /* giveup */

    DPRINT(DBG_OMD, "om-udp: send_msg: message \"%s\"", msgbuf);

    /* send the message */
    len = sendto(state->filedes, msgbuf, len, 0, 
            &(state->sendTo.sa), sizeof(state->sendTo));

    DPRINT(DBG_OMD, "om-udp: send_msg: sendto result %d (%d); end", len, errno);
}

static void
OMOD_UDP_reinit(struct OMOD_state *state)
{
    /* NO-OP */
}

static int
openSocket(void)
{
    int         sock;
    int         val;        /* used later for setsockopt */
    bool        success;

    struct sockaddr_in  addr;

    sock = -1;
    success = FALSE;

    /* if the socket is already open we succeeded */
    if (gSocket >= 0) {
        success = TRUE;
        goto end;
    }

    /* not open, so we have to create it first */

    /**@todo This doesn't deal with IPv6 */
    sock = socket(PF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        LOG_eprint("upd om: socket()", errno);
        goto end;
    }
    DPRINT(DBG_OMD, "udp: socket = %d", sock);

    /* if we need to send from a specific port then bind to it now */
    addr.sin_family = PF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(SYSLOG_UDP_PORT);

    DPRINT(DBG_OMD, "udp: sin_port = %d, syslogd_port = %d", 
            addr.sin_port, SYSLOG_UDP_PORT);

    /* make sure that we can bind to the correct port.  This is needed
     * because it may be used by the UDP IM also
     */ 
    val = 1; /* need a real var for this one */
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) < 0) {
        LOG_eprint("setsockopt(SO_REUSEADDR)", errno);
        DPRINT(DBG_OMD, "setsockopt(SO_REUSEADDR): %s", strerror(errno));
        goto end;
    }

    if (! PRIV_bindSocket(sock, (struct sockaddr*)&addr, sizeof(addr))) {
        DPRINT(DBG_OMD, "udp-openSocket: bind failed", NO_VARG);
        goto end;
    }


    gSocket = sock;

    success = TRUE;

end:
    if (!success) {
        if (sock >= 0)
            close(sock);
    }
    else {
        /* any successful call means the refcount needs to be bumped */
        gSockRefcount++;
    }

    return gSocket;
}

static void
closeSocket(void)
{
    gSockRefcount--;

    if (gSockRefcount <= 0) {
        gSockRefcount = 0;

        close(gSocket);
    }
}
