/**
 * @file omod_user.c
 *
 * @brief Output module that sends messages to logged in users.
 *
 * $Id: omod_user.c,v 1.3.2.3 2003/04/30 20:03:32 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Aug 26 18:53:59 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include "strcompat.h"

#include "util/list.h"
#include "types.h"
#include "privs.h"
#include "error.h"
#include "log.h"

#include "outputmod.h"
#include "omod_init.h"

#include "message/message.h"
#include "message/msg_frmt.h"

static char const rcsid[] 
    = "$Id: omod_user.c,v 1.3.2.3 2003/04/30 20:03:32 devink Exp $";

/**@defgroup omod_user User Output Module
 * @ingroup omod
 * @{
 */

/** 
 * Initilize the output module
 * @param options The options for this output module.
 * @return A pointer to the output module state struct
 */
static struct OMOD_state * OMOD_USER_init(struct OMOD_opt *options);

/**
 * Destroy the USER output module state
 * @param the OMOD_state object to destroy
 */
static void OMOD_USER_destroy(struct OMOD_state *state);

/**
 * Send a message through the users output module.
 *
 * @param state The OMOD_state object for this OM.
 * @param msg The message object to send.
 */
static void OMOD_USER_send_msg(struct OMOD_state *state,
		const struct message *msg);

/**
 * Re-initilize the output modules
 * @note NO-OP right now.
 *
 * @param state The current state of this OM.
 */
static void OMOD_USER_reinit(struct OMOD_state *state);

/*@}*/

void
OMOD_USER_iface_init(OMOD_interface *iface)
{
    iface->INIT = OMOD_USER_init;
    iface->REINIT = OMOD_USER_reinit;
    iface->DESTROY = OMOD_USER_destroy;
    iface->SEND_MSG = OMOD_USER_send_msg;
    iface->state = NULL;
}

bool
OMOD_USER_process_opts(struct OMOD_opt *opts, char const *f, int l)
{
    if (opts->outPath == NULL && opts->outList == NULL) {
        LOG_print(LOG_ERROR, 
            "%s:%d: Required option output-path is not present", f, l);
        return FALSE;   /* fatal */
    }

    if (opts->format.type == MFRMT_NOEXIST) {
        opts->format.type = MFRMT_RFC3164;
    }

    return TRUE;
}

static struct OMOD_state *
OMOD_USER_init(struct OMOD_opt *options)
{
    struct OMOD_state   *state;
    int                 user_count, i;
    size_t              sz;
    char                *p, *cur_user;
    LLIterator          itr;

    state = emalloc(sizeof(struct OMOD_state));
    memset(state, 0, sizeof(struct OMOD_state));

    /* figure out how many users are in the list */
    if (options->outList == NULL) {
        user_count = 1;
    }
    else {
        user_count = LL_size(options->outList);
    }
    /* Allocate one big chunk of memory */
    sz = sizeof(char*)*user_count + UT_NAMESIZE*user_count;
    state->user_list = emalloc(sz);
    state->user_count = user_count;
    /* cur_user is the block of memory to put the next user in */
    cur_user = (char*) state->user_list + user_count;
    if (options->outList == NULL) {
        strlcpy(cur_user, options->outPath, UT_NAMESIZE);
        state->user_list[0] = cur_user;
        /* Done, there can only be one */
    }
    else {
        /* need to go through the entire list */
        itr = LL_iterator(options->outList);
        i = 0;
        while ( (p = LLI_next(itr)) != NULL) {
            strlcpy(cur_user, p, UT_NAMESIZE);
            state->user_list[i] = cur_user;

            cur_user += UT_NAMESIZE;
            i++;
        }
    }
    
    /* copy Format + check for default */
    if (options->format.type == MFRMT_NOEXIST) {
        state->format.type = MFRMT_RFC3164;
    }
    else {
        state->format.type = options->format.type;
    }
    state->format.options = options->format.options;

    return state;
}

static void
OMOD_USER_destroy(struct OMOD_state *state)
{
    free(state->user_list);
    free(state);
}

static void
OMOD_USER_reinit(struct OMOD_state *state) {}

static void
OMOD_USER_send_msg(struct OMOD_state *state, const struct message *msg)
{
    bool            allUsers;
    bool            doSend;
    struct utmp     *utent;
    char            **user;
    char            msg_str[2048];
    char            dev[PATH_MAX];
    int             msg_len;
    int             i;
    int             fd;

    allUsers = FALSE;
    doSend = FALSE;

    if (state->user_count < 1)
        return;     /* nothing to do */

    /* Make the message string */
    msg_len = strlcpy(msg_str, "Message from syslogd: ", sizeof(msg_str));
    msg_len += MFRMT_format(msg_str + msg_len, sizeof(msg_str) - msg_len, 
                            msg, &(state->format));
    if (msg_len + 1 < sizeof(msg_str)) {
        /* Add a new line if we can */
        msg_str[msg_len] = '\n';
        msg_len++;
        msg_str[msg_len] = '\0';
    }

    if (strncmp(state->user_list[0], "*", 2) == 0) 
        allUsers = TRUE;

    setutent();     /* start at the top of utmp */

    /* XXX This processing code was done while reading the linux man
     * page, it may not be too portable 
     */
    while ( (utent = getutent()) != NULL) { 
        /* Skip the stuff we really don't care about */
        if (utent->ut_type != USER_PROCESS) 
            continue;
        if (utent->ut_name[0] == '\0') 
            continue;
        
        /* If we aren't sending to all logged in users, then we need to
         * see if this user is on our list 
         */
        if (! allUsers) {
            /* XXX This would be better to do a binary search on a
             * sorted list.
             */
            user = state->user_list;
            for (i = 0; i < state->user_count; i++) {
                if (strncmp(*user, utent->ut_name, UT_NAMESIZE)) {
                    doSend = TRUE;
                    break;
                }
                user++;
            }
        }

        if (allUsers || doSend) {
            snprintf(dev, sizeof(dev), "%s/%s", _PATH_DEV, utent->ut_line);

            fd = PRIV_openWriteFile(dev, O_WRONLY | O_NOCTTY);
            if (fd > 0) {
                write(fd, msg_str, msg_len);
                close(fd);
            }
        }
    }
}


