/**
 * @file outputmod.c
 *
 * @brief Implement the output module management functions.
 *
 * $Id: outputmod.c,v 1.7.2.7 2003/04/30 20:03:32 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue May 28 14:39:39 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#if HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <stdlib.h>

#include "types.h"
#include "outputmod.h"
#include "omod_init.h"
#include "error.h"
#include "strcompat.h"

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif

static char const rcsid[] = 
    "$Id: outputmod.c,v 1.7.2.7 2003/04/30 20:03:32 devink Exp $";

/**@name Output Module Support Implementation
 * @ingroup omod
 * @{
 */

/** max size of an output module name in the omod_map */
#define OMODLL_NAME_SZ  32

/** Structure to map output module names to type ids */
struct OMODLL_map 
{
    char        name[OMODLL_NAME_SZ];   /**< Output module name */
    OMOD_TYPE   type;                   /**< Output module type */
};

static struct OMODLL_map omod_map[] = {
    {"file", OMOD_FILE},
    {"pipe", OMOD_PIPE},
    {"console", OMOD_CONS},
    {"beep", OMOD_BEEP},
    {"udp", OMOD_UDP},
    {"user", OMOD_USER},
    {"", OMOD_COUNT}        /**< Mark the end of the array */
};

/** List of output module initilization functions */
static struct OMOD_registration omod_reg[OMOD_COUNT] = {
    {OMOD_CONS_iface_init, OMOD_CONS_process_opts}, /**< Console OM */
    {OMOD_FR_iface_init, OMOD_FR_process_opts},     /**< File OM */
    {NULL, NULL},                                   /**< Pipe OM */
    {OMOD_UDP_iface_init, OMOD_UDP_process_opts},   /**< UDP OM */
    {OMOD_USER_iface_init, OMOD_USER_process_opts}, /**< USER OM */
#if HAVE_LIBRR
    {OMOD_BP_iface_init, OMOD_BP_process_opts},		/**< BEEP OM */
#else
    {NULL, NULL}
#endif /* HAVE_LIBRR */
};


OMOD_TYPE
OMOD_getTypeByName(char const *name)
{
    struct OMODLL_map   *om;

    for (om = omod_map; om->type != OMOD_COUNT; om++) {
        if (strncmp(name, om->name, OMODLL_NAME_SZ) == 0) 
            break;
    }

    return om->type;
}

bool
OMOD_processOptions(OMOD_TYPE id, struct OMOD_opt *opts, char const *f, int l)
{
    if (id < 0 || id >= OMOD_COUNT)
        return FALSE;

    if (omod_reg[id].process_opts == NULL) 
        return FALSE;

    return omod_reg[id].process_opts(opts, f, l);
}

OMOD_interface *
OMOD_getInterface(OMOD_TYPE id)
{
    OMOD_interface      *omi;

    if (id < 0 || id >= OMOD_COUNT) 
        return NULL;

    if (omod_reg[id].iface_init == NULL) 
        return NULL;

    omi = emalloc(sizeof(OMOD_interface));
    memset(omi, 0, sizeof(OMOD_interface));

    /* Call the type specific initilization function */
    omod_reg[id].iface_init(omi);

    return omi;
}

void 
OMOD_freeInterface(OMOD_interface *omi)
{
    if (omi->state != NULL) {
        omi->DESTROY(omi->state);
        omi->state = NULL;
    }
    free(omi);
}

/*@}*/
