/**
 * @file outputmod.h
 *
 * @brief This describes the common functions and data structures for
 * working with output modules.
 *
 * @design d4
 * @ingroup omod
 *
 * $Id: outputmod.h,v 1.12.2.7 2003/04/30 20:03:32 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Jan 28 14:26:12 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef OMOD_OUTPUTMOD_H
#define OMOD_OUTPUTMOD_H

#include "message/message.h"
#include "util/pathcache.h"
#include "message/msg_frmt.h"

/** @todo get rid of this? */
#if HAVE_LIBRR
# include <librr/rr-connection.h>
# include <librr/rr-channel.h>
# include <librr/rr-profileregistry.h>
# include "util/profile-beep.h"
#endif /* HAVE_LIBRR */

/** @defgroup omod Output Module
 *  @{
 */

/** Enumerate the possible output modules */
typedef enum
{
    OMOD_CONS,      /**< Console output */
    OMOD_FILE,      /**< File */
    OMOD_PIPE,      /**< Pipe output */
    OMOD_UDP,       /**< UDP output */
    OMOD_USER,      /**< Logged in users */
	OMOD_BEEP,		/**< BEEP (over TCP) */
    OMOD_COUNT      /**< number of defined output modules */
} OMOD_TYPE;

/** 
 * This contains the state information for an output module, including pointers
 * interface functions.
 */
struct OMOD_state
{
	int				filedes;		/**< The file descriptor */ 
	char			*name;			/**< The name, must be free()'ed */
	int				namelen;		/**< The length of name */
	bool			error;			/**< Do we have an unrecoverable error */
    PCache          *pcache;        /**< Path cache. */
	uint32_t		pcache_omID;	/**< Path cache output module ID */

#if HAVE_LIBRR
    BEEP_conn   *beep_conn;	        /**< Hold the BEEP connection information*/
    bool        beep_droped;        /**< True after droped conn. detected*/
#endif /* HAVE_LIBRR */ 


    union NA_sockaddr   sendTo;     /**< address to send messages to */
	netaddr_t			sendToAddr;	/**< the netaddr to send messages to */

    char            **user_list;    /**< User list for user output module */
    int             user_count;     /**< Number of users in \c user_list */

    struct MFRMT_opt    format; /**< Output Format */
};

/** 
 * This contains the options for all output modules; specific output modules
 * ignore options which don't apply to them (see the comments for the specific
 * module.)
 */
struct OMOD_opt
{
    struct MFRMT_opt    format; /**< Output Format */

    char        *outPath;       /**< Path write output on */ 
    List        outList;        /**< List of output Paths */

    netaddr_t   outAddr;        /**< Address of remote log host */
    int         outPort;        /**< Port to send output to */

    PCache      *pcache;        /**< Path Cache shared with controller */
};

/** 
 * This contains function pointers to the public interface of an output
 * module.
 */
typedef struct 
{
    /** Initilize the output module. This returns the output module
     *  instantce state.  It should be stored in the state member of
     *  this structure.
     *
     *  @param name The ouput module name option from the config file
     *  @return the output module state.
     */
    struct OMOD_state* (*INIT)(struct OMOD_opt *options);

    /** Re-initilize the output module.  This is used for things like
     *  SIGHUP.
     *  @param name The ouput module name option from the config file
     */
    void (*REINIT)(struct OMOD_state *state);

    /** Shutdown an output module. 
     *  @param state The output modules state
     */ 
    void (*DESTROY)(struct OMOD_state *state);

    /** Send a message to the output module 
     *  @param state The output modules state
     *  @param msg The message struct to send out.
     */
    void (*SEND_MSG)(struct OMOD_state *state, const struct message *msg);

    /** The output module's state variables */
    struct OMOD_state *state;
} OMOD_interface;

/** Get an ouput module by type. This returns the interface for an
 *  output module. It has not yet been initilized.
 *
 *  @param id The OMOD_ID of the desired output module.
 *  @return the OMOD_interface for the desired output module.
 */
OMOD_interface * OMOD_getInterface(OMOD_TYPE id);

/** 
 * Free an output module interface.  This will also cleanup the OM
 * state.
 *
 * @param omi The OMOD_interface to free.
 */
void OMOD_freeInterface(OMOD_interface *omi);

/**
 * Get an output module ID from the name string.
 *
 * @param name The name of the output module type.
 * @return The OMOD_TYPE or OMOD_COUNT if name isn't found.
 */
OMOD_TYPE OMOD_getTypeByName(char const *name);

/**
 * Validate and do final processing on output module options.  This
 * specifically makes sure that all required options are present, sets
 * defaults and other such tasks.  Each output module init function can
 * assume that this has been called previously.  Further, errors and
 * insanities can be logged from here.  Lastly, if this function returns
 * an error (FALSE) then there is assumed to be a fatal/serious problem
 * in the configuration and the output module will not be able to start.
 * 
 * @param id The output module type id.
 * @param opts The OMOD_opts structure.
 * @param f The file the output module is defined in.
 * @param l The line the output module def starts on.
 * @return FALSE if there is fatal error in the options, TRUE otherwise.
 */
bool OMOD_processOptions(OMOD_TYPE id, struct OMOD_opt *opts, 
                         char const *f, int l);

/*@}*/

#endif /* OMOD_OUTPUTMOD_H */
