/**
 * @file common.h
 *
 * @brief Testing tools common functions.
 *
 * $Id: common.h,v 1.1.2.2 2003/04/30 20:03:36 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Sep 26 16:12:01 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef TTOOLS_COMMON_H
#define TTOOLS_COMMON_H

#ifdef HAVE_CONFIG_H
# include "acconfig.h"
#endif

#include <sys/time.h>
#include <sys/times.h>

/** Types of timers */
enum TIME_mode {
    TIME_EXEC,  /**< execution time */
    TIME_WC     /**< wall clock time */
};

/** A struct for holding timing info */
struct TIME_ctx {
    struct tms      tms_start;
    struct timeval  tv_start;

    struct tms      tms_stop;
    struct timeval  tv_stop;

    enum TIME_mode  mode;
    int             done;
};


/**
 * Convert a strings conatining a severity into the correct number
 * @param sev The severity name
 * @return the severity number.
 */
inline int find_sev(char const *sev);

/**
 * convert a string containing a facility into a number.
 * @param fac The facility name
 * @return the facility number
 */
inline int find_fac(char const *fac);

/**
 * Start a new timer
 * @param ctx the time context to hold the start time in
 * @return 0 on error, > 0 otherwise
 */
inline int TIME_start(struct TIME_ctx *ctx, enum TIME_mode type);

/**
 * Stop a execution time timer
 * @param ctx the time context to hold the stop time in
 * @return 0 on error, > 0 otherwise
 *
 * @pre \c ctx->mode==TIME_EXEC
 */
inline int TIME_exec_stop(struct TIME_ctx *ctx);

/**
 * Stop a wall clock time timer
 * @param ctx The time context to hald the stop time in.
 * @param 0 on erorr, > 0 otherwise
 *
 * @pre \c ctx->mode=TIME_WC
 */
inline int TIME_wc_stop(struct TIME_ctx *cxt);

/**
 * Get the elapsed wall clock time.
 * @param ctx the time context to check
 * @return 0.0 if the timer is not done, elapsed time otherwise.
 *
 * @pre \c ctx->mode=TIME_WC
 */
double TIME_wcinterval(struct TIME_ctx *ctx);

/**
 * Get the elapsed user time from a time context.
 * @param ctx the time context to check
 * @return 0.0 if the timer is not done, elapsed time otherwise.
 * @pre \c ctx->mode==TIME_EXEC
 */
double TIME_uinterval(struct TIME_ctx *ctx);

/**
 * Get the elapsed system time from a time context.
 * @param ctx the time context to check
 * @return 0.0 if the timer is not done, elapsed time otherwise.
 * @pre \c ctx->mode==TIME_EXEC
 */
double TIME_sinterval(struct TIME_ctx *ctx);
#endif /* TTOOLS_COMMON_H */


