/**
 * @file prifac.c
 *
 * @brief Functions to handle string -> number severity/facility
 * conversions.
 * $Id: sevfac.c,v 1.1.2.2 2003/04/30 20:03:36 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Sep 26 16:07:49 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "strcompat.h"
#include "config/priority.h"
#include "common.h"

struct sevfac {
    char    *name;
    int     num;
};

static int find_sevfac(char const *name, struct sevfac *arr, int count);

static struct sevfac SEVS[] = {
    { "emerg", CONFFS_EMERG, },     /**< Emergency severity */
    { "panic", CONFFS_EMERG, },     /**< Deprecated; Emergency severity */
    { "alert", CONFFS_ALERT, },     /**< Alert severity */
    { "crit", CONFFS_CRIT, },       /**< Critical severity */
    { "error", CONFFS_ERR, },       /**< Deprecated; Error severity */
    { "err", CONFFS_ERR, },         /**< Error severity */
    { "warn", CONFFS_WARN, },       /**< Warning severity */
    { "warning", CONFFS_WARN, },    /**< Deprecated; Warning severity */
    { "notice", CONFFS_NOTICE, },   /**< Notice severity */
    { "info", CONFFS_INFO, },       /**< Informational severity */
    { "debug", CONFFS_DEBUG, },     /**< Debug severity */
};

static struct sevfac FACS[] = {
    { "kernel", CONFFS_KERN, },     /**< Map 'kernel' to kernel facility */
    { "user", CONFFS_USER, },       /**< Map 'user' to user facility */
    { "mail", CONFFS_MAIL, },       /**< Map 'mail' to mail facility */
    { "daemon", CONFFS_DAEMON, },   /**< Map 'daemon' to daemon facility */
    { "auth", CONFFS_AUTH0, },      /**< Map 'auth' to auth0 facility */
    { "auth0", CONFFS_AUTH0, },     /**< Map 'auth0' to auth0 facility */
    { "syslog", CONFFS_SYSLOG, },   /**< Map 'syslog' to syslog facility */
    { "lpr", CONFFS_LPR, },         /**< Map 'lpr' to LPR facility */
    { "news", CONFFS_NEWS, },       /**< Map 'news' to news facility */
    { "uucp", CONFFS_UUCP, },       /**< Map 'uucp' to UUCP facility */
    { "cron", CONFFS_CRON0, },      /**< Map 'cron' to cron0 facility */
    { "cron0", CONFFS_CRON0, },     /**< Map 'cron0' to cron0 facility */
    { "auth1", CONFFS_AUTH1, }  ,   /**< Map 'auth1' to auth1 facility */
    { "ftp", CONFFS_FTP, },         /**< Map 'ftp' to ftp facility */
    { "ntp", CONFFS_NTP, },         /**< Map 'ntp' to ntp facility */
    { "audit", CONFFS_AUDIT0, },    /**< Map 'audit' to audit0 facility */
    { "audit0", CONFFS_AUDIT0, },   /**< Map 'audit0' to audit0 facility */
    { "audit1", CONFFS_AUDIT1, },   /**< Map 'audit1' to audit1 facility */
    { "cron1", CONFFS_CRON1, },     /**< Map 'cron1' to cron1 facility */
    { "local0", CONFFS_LOCAL0, },   /**< Map 'local0' to local0 facility */
    { "local1", CONFFS_LOCAL1, },   /**< Map 'local1' to local1 facility */
    { "local2", CONFFS_LOCAL2, },   /**< Map 'local2' to local2 facility */
    { "local3", CONFFS_LOCAL3, },   /**< Map 'local3' to local3 facility */
    { "local4", CONFFS_LOCAL4, },   /**< Map 'local4' to local4 facility */
    { "local5", CONFFS_LOCAL5, },   /**< Map 'local5' to local5 facility */
    { "local6", CONFFS_LOCAL6, },   /**< Map 'local6' to local6 facility */
    { "local7", CONFFS_LOCAL7, },   /**< Map 'local7' to local7 facility */
}; 

#define FAC_COUNT   (sizeof(FACS)/sizeof(struct sevfac))
#define SEV_COUNT   (sizeof(SEVS)/sizeof(struct sevfac))

int
find_fac(char const *fac)
{
    return find_sevfac(fac, FACS, FAC_COUNT);
}

int 
find_sev(char const *sev)
{
    return find_sevfac(sev, SEVS, SEV_COUNT);
}

int
find_sevfac(char const *name, struct sevfac *arr, int count)
{
    int     i;
    for (i = 0; i < count; i++) {
        if (strcmp(arr[i].name, name) == 0) {
            return arr[i].num;
        }
    }

    return 0;
}
