/**
 * @file timing.c
 *
 * @brief Functions for timing things.
 * @par
 * The portability stuff in here is because I'm copying portions of it
 * from another project of mine.  It should compile on linux, solaris,
 * and AIX.
 *
 * $Id: timing.c,v 1.1.2.2 2003/04/30 20:03:36 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Sep 26 16:19:02 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/times.h>
#include <sys/time.h>

#ifdef SUN
# include <limits.h>
#else
# include <unistd.h>
#endif
#include "strcompat.h"

#include "common.h"

#if !defined(CLK_TCK) && defined(_SC_CLK_TCK)
# define CLK_TCK sysconf(_SC_CLK_TCK)
#endif

#ifndef CLK_TCK
# error "Don't know how long a clock tick is"
#endif

#define CLOCK_HZ    (double)(CLK_TCK)

int TIME_start(struct TIME_ctx *ctx, enum TIME_mode type)
{
    memset(ctx, 0, sizeof(struct TIME_ctx));
    ctx->mode = type;
    if (type == TIME_WC) {
        return gettimeofday(&(ctx->tv_start), NULL) == 0;
    }
    else if (type == TIME_EXEC) {
        return times(&(ctx->tms_start)) != (clock_t)-1;
    }
    return 0;
}

int TIME_wc_stop(struct TIME_ctx *ctx)
{
    int rtn;
    rtn = gettimeofday(&(ctx->tv_stop), NULL); 

    /* after the fact checking so the actual timing is affected as
     * little as possible.
     */
    if (ctx->mode != TIME_WC) return 0;
    ctx->done = 1;
    return rtn == 0;
}

int TIME_exec_stop(struct TIME_ctx *ctx)
{
    clock_t rtn;
    rtn = times(&(ctx->tms_stop)); 

    /* after the fact checking so the actual timing is affected as
     * little as possible.
     */
    if (ctx->mode != TIME_EXEC) return 0;
    ctx->done = 1;
    return rtn != (clock_t)-1;
}

double TIME_wcinterval(struct TIME_ctx *ctx)
{
    if (! ctx->done) return 0.0;
    if (ctx->mode != TIME_WC) return 0.0;

    return ((double)(ctx->tv_stop.tv_sec) + 1.0e-6*ctx->tv_stop.tv_usec)
        - ((double)(ctx->tv_start.tv_sec) + 1.0e-6*ctx->tv_start.tv_usec);
}

double TIME_uinterval(struct TIME_ctx *ctx)
{
    if (! ctx->done) return 0.0;
    if (ctx->mode != TIME_EXEC) return 0.0;

    return (double)(ctx->tms_stop.tms_utime - ctx->tms_start.tms_utime)
                / CLOCK_HZ;
}

double TIME_sinterval(struct TIME_ctx *ctx)
{
    if (! ctx->done) return 0.0;
    if (ctx->mode != TIME_EXEC) return 0.0;

    return (double)(ctx->tms_stop.tms_stime - ctx->tms_start.tms_stime)
                / CLOCK_HZ;
}
