/**
 * @file udpsend.c
 *
 * @brief Send a syslog message via UDP
 *
 * $Id: udpsend.c,v 1.8.2.8 2003/10/22 04:09:04 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Aug 22 16:40:23 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "strcompat.h"
#include "config/priority.h"

#define OPTIONS     "f:s:p:h:c"
#define OPT_FAC     'f'
#define OPT_SEV     's'
#define OPT_PORT    'p'
#define OPT_HOST    'h'
#define OPT_COMPAT  'c'


static int connect_socket(int port, char const *ip);

static void send_message(int sock, int fac, int sev, char const *msg);

static const char   old_fmt[] = "%b %d %H:%M:%S";
static const char   new_fmt[] = "%Y-%m-%dT%H:%M:%S%z";
static const char   *fmt = new_fmt;

int main(int argc, char **argv)
{
    int     arg;
    int     port; 
    int     fac, sev;
    int     sock;
    char    ip[128];

    /* defaults */
    port = 5000;
    ip[0] = '\0';
    sev = CONFFS_NOTICE;
    fac = CONFFS_LOCAL0;

    while ( (arg = getopt(argc, argv, OPTIONS)) >= 0) {
        switch (arg) {
            case OPT_FAC:
            {
                if (optarg == 0) {
                    exit(1);
                }
                fac = find_fac(optarg);
                break;
            }
            case OPT_SEV:
            {
                if (optarg == 0) {
                    exit(1);
                }
                sev = find_sev(optarg);
                break;
            }
            case OPT_PORT:
            {
                if (optarg == 0) {
                    exit(1);
                }
                port = atoi(optarg);
                break;
            }
            case OPT_HOST:
            {
                if (optarg == 0) {
                    exit(1);
                }
                strlcpy(ip, optarg, sizeof(ip));
                break;
            }
            case OPT_COMPAT:
            {
                fmt = old_fmt;
                break;
            }
            case '?': 
            {
                /* unknown option */
                exit(1);
                break;
            }
            default:
            {
                fprintf(stderr, "Unknown option \"%c\"\n", arg);
                exit(1);
                break;
            }
        }
    }

    sock = connect_socket(port, ip);
    while (optind < argc) {
        send_message(sock, fac, sev, argv[optind]);
        optind++;
    }

    close(sock);
    exit(0);
}

static int 
connect_socket(int port, char const *ip)
{
    int                 sock;
    struct sockaddr_in  addr;

	sock=socket(AF_INET,SOCK_DGRAM,0);
	if( sock < 0 ) {
        perror("socket()");
        exit(1);
    }

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);

    if (ip == NULL || *ip == '\0') {
	    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    }
    else if (! inet_aton(ip, &(addr.sin_addr))) {
        fprintf(stderr, "ip address is bad\n");
        exit(1);
    }

	/* connect and send the first packet */
	if(connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		perror("connect()");
        exit(1);
    }

    return sock;
}

static void
send_message(int sock, int fac, int sev, char const *msg)
{
    uint32_t    pri;
    char        buf[1024];
    char        host[255];
    char        datetime[128];
    time_t      ts;


    ts = time(NULL);
    if (fmt == NULL) {
        fprintf(stderr, "No timestamp format\n");
    }
    pri = strftime(datetime, sizeof(datetime), 
        fmt, localtime(&ts));
    if (pri <= 0) {
        fprintf(stderr, "strftime() failed\n");
        exit(1);
    }
    if (gethostname(host, sizeof(host)) < 0) {
        perror("gethostname():");
        exit(1);
    }
    pri = (fac << 3) + sev;


    snprintf(buf, sizeof(buf), 
        "<%d>%s %s udpsendtest: %s", pri, datetime, host, msg);

    printf("Sending: %s\n", buf);
    if (send(sock, buf, strlen(buf), 0) < 0) {
        perror("send()");
        exit(1);
    }
}
