/**
 * @file ut_acl.c
 *
 * @brief Define the Base Test suite, and implements it
 * @ingroup ut_acl
 *
 *
 * $Id: ut_acl.c,v 1.8.2.1 2003/04/30 20:03:42 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Apr  2 16:16:22 PST 2002
 *
 * @todo Path checking checking needs to be implemented; the start of an
 * 	implementation is in ut_acl.c
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>      /* for NULL */

#include "ut_driver.h"

/* Include all the needed test suite headers */
#include "ut_acl.h"

/* and, include the ACL headers */
#include "util/acl.h"
#include "error.h"

#include "strcompat.h"
#include <string.h>


/**@defgroup ut_acl ACL Base Tests 
 * @ingroup unit_test
 */


/**@name Globals needed by generic test driver 
 * @ingroup ut_acl
 * @{
 */
/* our suite setup functions */
SuiteInfo suites[] = {
	ADD_SUITE(acl_suite, "ACL"),

	/* Don't change the ordering; ACL_crunch DEPENDS ON ACL_buf */
	ADD_SUITE(acl_suite_crunch, "ACL_crunch"),
	ADD_SUITE(acl_suite_buf, "ACL_buf")
};

int suite_count = SUITE_COUNT;

char const *log_name = "acl";

bool disablePrivs = TRUE;
/*@}*/


/**@name ACL Test Cases
 * @ingroup ut_acl
 * @{
 */

/** Test basic ACL functionality through denies.
 * Also, test ACL_addACL() */
static void TESTacl_deny(void);

/** Test basic ACL functionality through allows */
static void TESTacl_allow(void);

/* static void TESTacl_path_allow(void); we dont' do this yet */

/*@}*/

static void debug_init(void)
{
    DBG_init("an", "acl.debug");
}

static void debug_done(void)
{
    DPRINT(DBG_DBG, "Test Case Done", NO_VARG);
}

void acl_suite(Suite *suite)
{
	TCase *tc_acl;

	tc_acl=tcase_create("ACL");
	suite_add_tcase(suite,tc_acl);

    tcase_add_unchecked_fixture(tc_acl, debug_init, debug_done);

	tcase_add_test(tc_acl,TESTacl_deny);
	tcase_add_test(tc_acl,TESTacl_allow);
	/* tcase_add_test(tc_io,TESTacl_path_allow); not impelemented yet */
}

START_TEST(TESTacl_deny)
{
	ACLChecker *aclchkr;
	ACL *acl,*acl_add_difftype,*acl_add_diffmode,*acl_add;
	netaddr_t ip_acl,ip_eacl,ip_dacl,ip_acl_add;

	/* init ACL */
	aclchkr=ACL_init();
	acl=ACL_create(ACLT_GENHOST,ACLM_DENY);
	acl_add=ACL_create(ACLT_GENHOST,ACLM_DENY);

	/* check add permissions */
	acl_add_difftype=ACL_create(ACLT_PATH,ACLM_DENY);
	acl_add_diffmode=ACL_create(ACLT_GENHOST,ACLM_ALLOW);
	fail_if(ACL_addACL(acl,acl_add_difftype),"Unexpected success in adding ACLs of different types");
	fail_if(ACL_addACL(acl,acl_add_diffmode),"Unexpected success in adding ACLs of different modes");
	ACL_free(acl_add_difftype);
	ACL_free(acl_add_diffmode);

	/* add hosts to be denied */
	fail_unless(ACL_addHost(acl,"www.sdsc.edu"),
            "Failed to add host to ACL list." HOST_LOOKUP_FAIL_MSG);
	fail_unless(ACL_addHost(acl,"mail.sdsc.edu"),
            "Failed to add host to ACL list." HOST_LOOKUP_FAIL_MSG);

	/* add ips to be denied */
	ip_acl=NA_newINetLookup("192.168.1.1");
	fail_if(NA_isError(ip_acl),"netaddr_t ip_acl was error");
	ip_eacl=NA_newINetLookup("192.168.1.1");
	fail_if(NA_isError(ip_eacl),"netaddr_t ip_eacl was error");
	ip_dacl=NA_newINetLookup("192.168.12.1");
	fail_if(NA_isError(ip_dacl),"netaddr_t ip_dacl was error");
	fail_unless(ACL_addIP(acl,ip_acl),"Failed to add ip_acl");

	/* add through ACL_addACL() */
	fail_unless(ACL_addHost(acl_add,"security.sdsc.edu"),
            "Failed to add host to ACL list" HOST_LOOKUP_FAIL_MSG);
	ip_acl_add=NA_newINetLookup("192.168.1.2");
	fail_if(NA_isError(ip_acl_add),"netaddr_t ip_acl_add was error");
	fail_unless(ACL_addIP(acl_add,ip_acl_add),"Failed to add ip_acl_add");

	fail_unless(ACL_addACL(acl,acl_add),"Failed to add ACL's with the same type and mode");

	fail_unless(ACL_setAcl(aclchkr,acl),"Failed to add ACL to ACLChecker");

	/* check to make sure the hosts were dealt with */
	fail_if(ACL_checkHost(aclchkr,"www.sdsc.edu",ACLT_GENHOST),
            "Host not denied." HOST_LOOKUP_FAIL_MSG);
	fail_if(ACL_checkHost(aclchkr,"mail.sdsc.edu",ACLT_GENHOST),
            "Host not denied." HOST_LOOKUP_FAIL_MSG);
	fail_if(ACL_checkHost(aclchkr,"security.sdsc.edu",ACLT_GENHOST),
            "Host not denied." HOST_LOOKUP_FAIL_MSG);
	fail_unless(ACL_checkHost(aclchkr,"ftp.sdsc.edu",ACLT_GENHOST),
            "Unknown host not allowed." HOST_LOOKUP_FAIL_MSG);

	/* check to make sure the IPs are dealt with */
	fail_if(ACL_checkIP(aclchkr,ip_acl,ACLT_GENHOST),"IP not denied");
	fail_if(ACL_checkIP(aclchkr,ip_eacl,ACLT_GENHOST),"IP not denied");
	fail_if(ACL_checkIP(aclchkr,ip_acl_add,ACLT_GENHOST),"IP not denied");
	fail_unless(ACL_checkIP(aclchkr,ip_dacl,ACLT_GENHOST),"Uknown IP not allowed");

	ACL_free(acl);
	ACL_free(acl_add);
	ACL_destroy(aclchkr);
}
END_TEST

START_TEST(TESTacl_allow)
{
	ACLChecker *aclchkr;
	ACL *acl;
	netaddr_t ip_acl,ip_eacl,ip_dacl;

	/* init ACL */
	aclchkr=ACL_init();
	acl=ACL_create(ACLT_GENHOST,ACLM_ALLOW);

	/* add hosts to be allowed */
	fail_unless(ACL_addHost(acl,"www.sdsc.edu"),
            "Failed to add host to ACL list." HOST_LOOKUP_FAIL_MSG);
	fail_unless(ACL_addHost(acl,"mail.sdsc.edu"),
            "Failed to add host to ACL list." HOST_LOOKUP_FAIL_MSG);

	/* add ips to be allowed */
	ip_acl=NA_newINetLookup("192.168.1.1");
	fail_if(NA_isError(ip_acl),"netaddr_t ip_acl was error");
	fail_unless(ACL_addIP(acl,ip_acl),"Failed to add ip_acl");

	/* create misc ips */
	ip_eacl=NA_newINetLookup("192.168.1.1");
	fail_if(NA_isError(ip_eacl),"netaddr_t ip_eacl was error");

	ip_dacl=NA_newINetLookup("192.168.12.1");
	fail_if(NA_isError(ip_dacl),"netaddr_t ip_dacl was error");

	/* add ACL to aclchkr */
	fail_unless(ACL_setAcl(aclchkr,acl),"Failed to add ACL to ACLChecker");

	/* check to make sure the hosts were dealt with */
	fail_unless(ACL_checkHost(aclchkr,"www.sdsc.edu",ACLT_GENHOST),
            "Host not allowed." HOST_LOOKUP_FAIL_MSG);
	fail_unless(ACL_checkHost(aclchkr,"mail.sdsc.edu",ACLT_GENHOST),
            "Host not allowed." HOST_LOOKUP_FAIL_MSG);
	fail_if(ACL_checkHost(aclchkr,"ftp.sdsc.edu",ACLT_GENHOST),
            "Unknown host not denied." HOST_LOOKUP_FAIL_MSG);

	/* check to make sure the ips were dealt with */
	fail_unless(ACL_checkIP(aclchkr,ip_acl,ACLT_GENHOST),"IP not allowed");
	fail_unless(ACL_checkIP(aclchkr,ip_eacl,ACLT_GENHOST),"IP not allowed");
	fail_if(ACL_checkIP(aclchkr,ip_dacl,ACLT_GENHOST),"Unknown IP not denied");

	/* cleanup */
	NA_destroy(ip_acl);
	NA_destroy(ip_eacl);
	NA_destroy(ip_dacl);
	ACL_destroy(aclchkr);
}
END_TEST

/*************************************************************
 * *********** PATH CHECKING IS NOT IMPELEMTED ***************
 * **************** but the start of it is below *************
 */
#if 0
int pathcnt;

void addpathelem(struct path *pth,char *from,char *to)
{
	struct path *wpth=pth;

	while(wpth!=NULL)
		wpth=wpth->next;

	wpth=emalloc(sizeof(struct path));
	wpth->id=pathcnt++;

	wpth->ffqdnlen=strlen(from)+1;
	wpth->fromFQDN=emalloc(wpth->ffqdnlen);
	strlcpy(wpth->fromFQDN,from,wpth->ffqdnlen);

	wpth->tfqdnlen=strlen(to)+1;
	wpth->toFQDN=emalloc(wpth->tfqdnlen);
	strlcpy(wpth->toFQDN,to,wpth->tfqdnlen);
}

void freepath(struct path *pth)
{
	if(pth->next!=NULL)
		freepath(pth->next);

	free(pth->toFQDN);
	free(pth->fromFQDN);
	free(pth);
}

#define host1	"buckaroo.sdsc.edu"
#define host2	"mail.sdsc.edu"
#define host3	"192.168.1.1"
#define extrahost	"www.sdsc.edu"
START_TEST(TESTacl_path_allow)
{
	ACLChecker *aclchkr;
	ACL *acl;
	struct path *pth;

	/* create the path */
	pathcnt=0;
	addpathelem(pth,host1,host2);
	addpathelem(pth,host2,host3);




	/* init ACL */
	aclchkr=ACL_init();
	acl=ACL_create(ACLT_PATH,ACLM_ALLOW);

	/* add hosts to be allowed */
	fail_unless(ACL_addHost(acl,"buckaroo.sdsc.edu"),"Failed to add host to ACL list");
	fail_unless(ACL_addHost(acl,"mail.sdsc.edu"),"Failed to add host to ACL list");

}
END_TEST
#undef host1
#undef host2
#undef host3
#undef extrahost
#endif /* 0 */

