/**
 * @file ut_acl_buf.c
 *
 * @brief Define the Base Test suite, and implements it
 * @ingroup ut_acl
 *
 * $Id: ut_acl_buf.c,v 1.7.2.1 2003/04/30 20:03:42 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Apr  2 16:16:22 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>      /* for NULL */
#include <stdlib.h>

#include "ut_driver.h"

/* Include all the needed test suite headers */
#include "ut_acl.h"

/* and, include the ACL headers */
#include "util/acl.h"
#include "error.h"

/**@name ACL Test Cases
 * @ingroup ut_acl
 * @{
 */
static void TESTacl_buffer(void);
/*@}*/

/** Buffer tests */

void acl_suite_buf(Suite *suite)
{
	TCase *tc_io;

	tc_io=tcase_create("ACL_buf");
	suite_add_tcase(suite,tc_io);

	tcase_add_test(tc_io,TESTacl_buffer);

    if (! DBG_init("ab", "acl-buf.debug") ) {
        perror("debug init failed");
        exit(1);
    }
}

START_TEST(TESTacl_buffer)
{
	ACLChecker *aclchk,*aclchk_new;
	ACL *acl,*acl_new;
	struct IPC_buffer *buf;
	netaddr_t ip_acl,ip_dacl;
    IPC_bufMark  startPos;

	aclchk=ACL_init();
	aclchk_new=ACL_init();
	acl=ACL_create(ACLT_GENHOST,ACLM_ALLOW);

	/* add hosts to be allowed */
	fail_unless(ACL_addHost(acl,"www.sdsc.edu"),
            "Failed to add host to ACL list." HOST_LOOKUP_FAIL_MSG);
	ip_acl=NA_newINetLookup("192.168.1.1");
	fail_unless(ACL_addIP(acl,ip_acl),"Failed to add ip_acl");

	/* create misc ip */
	ip_dacl=NA_newINetLookup("192.168.12.1");
	fail_if(NA_isError(ip_dacl),"netaddr_t ip_dacl was error");

	/* add ACL to aclchkr */
	fail_unless(ACL_setAcl(aclchk,acl),"Failed to add ACL to ACLChecker");

	/* test this new acl */
	fail_unless(ACL_checkHost(aclchk,"www.sdsc.edu",ACLT_GENHOST),
            "Host not allowed." HOST_LOOKUP_FAIL_MSG);
	fail_if(ACL_checkHost(aclchk,"ftp.sdsc.edu",ACLT_GENHOST),
            "Unknown host not denied." HOST_LOOKUP_FAIL_MSG);
	fail_unless(ACL_checkIP(aclchk,ip_acl,ACLT_GENHOST),"IP not allowed");
	fail_if(ACL_checkIP(aclchk,ip_dacl,ACLT_GENHOST),"Unknown IP not denied");

	/* init the buffer */
	buf=emalloc(sizeof(struct IPC_buffer));
	BUF_init(buf);
	buf->size=ACL_getBufferSize(acl);
	BUF_alloc(buf);

    startPos = BUF_markPos(buf);

	fail_if(ACL_getBufferSize(acl)<ACL_toBuffer(acl,buf),"Amount written to buffer greater than amount we allocated for it");

    BUF_resetToPos(buf, startPos);

	/* get our ACL from the buffer, put it in a new ACLChecker */
	acl_new=ACL_fromBuffer(buf);
	fail_if(acl_new == NULL,"ACL_fromBuffer() returned NULL");
	fail_unless(ACL_setAcl(aclchk_new,acl_new),"Failed to add ACL to ACLChecker");

	/* test the acl from the buffer */
	fail_unless(ACL_checkHost(aclchk_new,"www.sdsc.edu",ACLT_GENHOST),
            "Host not allowed." HOST_LOOKUP_FAIL_MSG);
	fail_if(ACL_checkHost(aclchk_new,"ftp.sdsc.edu",ACLT_GENHOST),
            "Unknown host not denied." HOST_LOOKUP_FAIL_MSG);
	fail_unless(ACL_checkIP(aclchk_new,ip_acl,ACLT_GENHOST),"IP not allowed");
	fail_if(ACL_checkIP(aclchk_new,ip_dacl,ACLT_GENHOST),"Unknown IP not denied");

	/* cleanup */
	NA_destroy(ip_acl);
	NA_destroy(ip_dacl);
	ACL_destroy(aclchk);
	ACL_destroy(aclchk_new);
	BUF_free(buf);
	free(buf);
}
END_TEST

