/**
 * @file ut_common_func.c
 *
 * @brief Test the common functions
 * @par
 *
 *
 * $Id: ut_common_func.c,v 1.1.2.3 2003/04/30 20:03:44 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Sep 10 12:10:03 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>

#include "check.h"
#include "ut_driver.h"
#include "ut_base.h"

#include "error.h"

#include "common_funcs.h"


/**@name Common Function Test Cases
 * @ingroup utb_common_funcs
 * @{
 */

/** Test integer mismatch failure */
static void cf_diffint(void);

/** Integer mismatch pass */
static void cf_diffint_pass(void);

/** Integer mismatch failure */
static void cf_diffint_fail(void);

/** Test string mismatch failure */
static void cf_diffstr(void);

/** String mismatch pass */
static void cf_diffstr_pass(void);

/** String mismatch failure */
static void cf_diffstr_fail(void);

/** Test NA mismatch failure */
static void cf_diffna(void);

/** NA mismatch pass */
static void cf_diffna_pass(void);

/** NA mismatch failure */
static void cf_diffna_fail(void);

/*@}*/

void
cf_suite(Suite *suite)
{
	TCase *tc;

	tc=tcase_create("Common Functions");
	suite_add_tcase(suite,tc);

	tcase_add_test(tc, cf_diffint);
	tcase_add_test(tc, cf_diffstr);
	tcase_add_test(tc, cf_diffna);
}

START_TEST(cf_diffint_pass)
{
	struct message *a,*b;

	a=emalloc(sizeof(struct message));
	b=emalloc(sizeof(struct message));

	memset(a,0,sizeof(struct message));
	memset(b,0,sizeof(struct message));

	a->msglen=2;
	b->msglen=2;

	a->host="hello";
	b->host="hello";

	a->hostlen=2;
	b->hostlen=2;

	CMP_msg(a,b);
}
END_TEST

START_TEST(cf_diffint_fail)
{
	struct message *a,*b;

	a=emalloc(sizeof(struct message));
	b=emalloc(sizeof(struct message));

	memset(a,0,sizeof(struct message));
	memset(b,0,sizeof(struct message));

	a->msglen=1;
	b->msglen=2;

	a->host="hello";
	b->host="hello";

	a->hostlen=2;
	b->hostlen=2;

	CMP_msg(a,b);
}
END_TEST

START_TEST(cf_diffstr_pass)
{
	struct message *a,*b;

	a=emalloc(sizeof(struct message));
	b=emalloc(sizeof(struct message));

	memset(a,0,sizeof(struct message));
	memset(b,0,sizeof(struct message));

	a->msglen=2;
	b->msglen=2;

	a->host="hello";
	b->host="hello";

	a->hostlen=2;
	b->hostlen=2;

	CMP_msg(a,b);
}
END_TEST

START_TEST(cf_diffstr_fail)
{
	struct message *a,*b;

	a=emalloc(sizeof(struct message));
	b=emalloc(sizeof(struct message));

	memset(a,0,sizeof(struct message));
	memset(b,0,sizeof(struct message));

	a->msg="fourteen";
	b->msg="fourteen";

	a->host="hello";
	b->host="goodbye";

	a->hostlen=2;
	b->hostlen=2;

	CMP_msg(a,b);
}
END_TEST

START_TEST(cf_diffna_pass)
{
	struct path *a,*b;

	a=emalloc(sizeof(struct path));
	b=emalloc(sizeof(struct path));

	memset(a,0,sizeof(struct path));
	memset(b,0,sizeof(struct path));

	b->id=a->id=1;

	b->fromFQDN="hello";
	a->fromFQDN="hello";

	a->toIP=NA_newINetLookup("10.0.0.50");
	b->toIP=NA_newINetLookup("10.0.0.50");

	CMP_path(a,b);
}
END_TEST

START_TEST(cf_diffna_fail)
{
	struct path *a,*b;

	a=emalloc(sizeof(struct path));
	b=emalloc(sizeof(struct path));

	memset(a,0,sizeof(struct path));
	memset(b,0,sizeof(struct path));

	b->id=a->id=1;

	b->fromFQDN="hello";
	a->fromFQDN="hello";

	a->toIP=NA_newINetLookup("10.0.0.50");
	b->toIP=NA_newINetLookup("10.0.0.51");

	CMP_path(a,b);
}
END_TEST

START_TEST(cf_diffint)
{
	TCase *tc;
	Suite *s;
	SRunner *sr;
	char errstr[1024];

	s=suite_create("DiffInt Fail");
	tc=tcase_create("Core");
	suite_add_tcase(s,tc);
	tcase_add_test(tc,cf_diffint_pass);
	tcase_add_test(tc,cf_diffint_fail);
	sr=srunner_create(s);
	srunner_run_all(sr,CK_SILENT);		/* CK_SILENT or CK_VERBOSE */

	snprintf(errstr,1024,"Failure counts not correct (expected 1, got %d",
			srunner_ntests_failed(sr));
	fail_unless(srunner_ntests_failed(sr)==1,
			errstr);

	snprintf(errstr,1024,"Bad failed check message: '%s'",
			tr_msg(srunner_failures(sr)[0]));
	fail_unless(0==strcmp(tr_msg(srunner_failures(sr)[0]),
			"msglen doesn't match ('1' != '2')"),
			errstr);
}
END_TEST

START_TEST(cf_diffstr)
{
	TCase *tc;
	Suite *s;
	SRunner *sr;
	char errstr[1024];

	s=suite_create("DiffStr Fail");
	tc=tcase_create("Core");
	suite_add_tcase(s,tc);
	tcase_add_test(tc,cf_diffstr_pass);
	tcase_add_test(tc,cf_diffstr_fail);
	sr=srunner_create(s);
	srunner_run_all(sr,CK_SILENT);		/* CK_SILENT or CK_VERBOSE */

	snprintf(errstr,1024,"Failure counts not correct (expected 1, got %d",
			srunner_ntests_failed(sr));
	fail_unless(srunner_ntests_failed(sr)==1,
			errstr);

	snprintf(errstr,1024,"Bad failed check message: '%s'",
			tr_msg(srunner_failures(sr)[0]));
	fail_unless(0==strcmp(tr_msg(srunner_failures(sr)[0]),
			"host doesn't match ('hello' != 'goodbye')"),
			errstr);
}
END_TEST

START_TEST(cf_diffna)
{
	TCase *tc;
	Suite *s;
	SRunner *sr;
	char errstr[1024];

	s=suite_create("DiffNA Fail");
	tc=tcase_create("Core");
	suite_add_tcase(s,tc);
	tcase_add_test(tc,cf_diffna_pass);
	tcase_add_test(tc,cf_diffna_fail);
	sr=srunner_create(s);
	srunner_run_all(sr,CK_SILENT);		/* CK_SILENT or CK_VERBOSE */

	snprintf(errstr,1024,"Failure counts not correct (expected 1, got %d",
			srunner_ntests_failed(sr));
	fail_unless(srunner_ntests_failed(sr)==1,
			errstr);

	snprintf(errstr,1024,"Bad failed check message: '%s'",
			tr_msg(srunner_failures(sr)[0]));
	fail_unless(0==strcmp(tr_msg(srunner_failures(sr)[0]),
			"toIP doesn't match (names can't be provided)"),
			errstr);
}
END_TEST
