/**
 * @file ut_list.c
 *
 * @brief List Unit Test driver.
 * @ingroup utb_llist
 *
 *
 * $Id: ut_list.c,v 1.6.2.1 2003/04/30 20:03:44 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Feb 04 14:10:02 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>  /* for NULL */

#include "check.h"
#include "util/list.h"
#include "types.h"

#include "ut_driver.h"
#include "ut_base.h"

/**@name Linked List Test Cases
 * @ingroup utb_llist
 * @{
 */

/** Test list creation, and initilization */
static void list_create(void);

/** Test appending to a list */
static void list_append(void);

/** Test prepending to a list */ 
static void list_prepend(void);

/** Test removal of the first element of a list */
static void list_rmfirst(void);

/** Test removal of the last element of a list */
static void list_rmlast(void);

/** Test removal at both ends of the list */
static void list_remove(void);

/*@}*/

/**@name Linked List Iterator Test Cases
 * @ingroup utb_list
 * @{
 */

/** Test LL iterator creations */
static void lliterator_create(void);

/** Test LL iteration via next */
static void lliterator_next(void);

/** Test LL iteration via prev */
static void lliterator_prev(void);

/** Test that LLI_elem() matches the last returned element from next() */
static void lliterator_elem(void);

/** Test bi-directional iteration. */
static void lliterator_bidir(void);

/** Test insertions through an iterator */
static void lliterator_insert(void);

/** Test removal of elements through an iterator */
static void lliterator_remove(void);

/*@}*/

/** The number of test elements we will use */
#define TEST_ELEM_COUNT 10

/** The list we work all our tests on */
static List     list;
/** A List Iterator to run our tests on */
static LLIterator itr;
/** The master values for our test data */
static int      master[TEST_ELEM_COUNT];
/** The data that is actually inserted in our list */
static int      data[TEST_ELEM_COUNT];

/** Create a new list and initilize \c master and \c data */
static void list_setup(void);
/** Destroy our test list and cleanup \c master and \c data if needed */
static void list_finish(void);

/** Create a new list and load data into it. Data is loaded from \c
 * data, and is the same as from list_setup().
 */ 
static void lliterator_setup(void);

/** Cleanup the iterator, and whatever needs cleaning from list_finish()
 */
static void lliterator_finish(void);


/* test suit setup */
void list_suite(Suite *suite)
{
    TCase   *tc_list, *tc_iterator;

    /* Setup test case for list functions */
    tc_list = tcase_create("List Functions");
    suite_add_tcase(suite, tc_list);
    tcase_add_checked_fixture(tc_list, list_setup, list_finish);

    tcase_add_test(tc_list, list_create);
    tcase_add_test(tc_list, list_append);
    tcase_add_test(tc_list, list_prepend);
    tcase_add_test(tc_list, list_rmfirst);
    tcase_add_test(tc_list, list_rmlast);
    tcase_add_test(tc_list, list_remove);

    /* Setup test case for List Iterator Functions */
    tc_iterator = tcase_create("List Iterator Functions");
    suite_add_tcase(suite, tc_iterator);
    tcase_add_checked_fixture(tc_iterator, lliterator_setup, lliterator_finish);

    tcase_add_test(tc_iterator, lliterator_create);
    tcase_add_test(tc_iterator, lliterator_next);
    tcase_add_test(tc_iterator, lliterator_prev);
    tcase_add_test(tc_iterator, lliterator_elem);
    tcase_add_test(tc_iterator, lliterator_bidir);
    tcase_add_test(tc_iterator, lliterator_insert);
    tcase_add_test(tc_iterator, lliterator_remove);
}

/* fixtures */
static void list_setup(void)
{
    int     i;

    /* initilize test data */
    for (i = 0; i < TEST_ELEM_COUNT; i++) {
        master[i] = i * 2;
        data[i] = master[i];
    }

    /* create a list */
    list = LL_create();
}

static void list_finish(void)
{
    LL_destroy(&list);
}

static void lliterator_setup(void)
{
    int     i;

    list_setup();

    for (i = 0; i < TEST_ELEM_COUNT; i++) {
        LL_append(list, &(data[i]));
    }

    fail_unless(LL_size(list) == TEST_ELEM_COUNT, 
        "LL_size() returned incorrect element count after setup");

    itr = LL_iterator(list);
}

static void lliterator_finish(void)
{
    list_finish();
    LLI_destroy(&itr);
}

/************************** List Test Cases *****************************/
START_TEST(list_create)
{
    fail_unless(list != NULL, "list creation returned NULL");
    fail_unless(LL_isEmpty(list) == TRUE, "New List is not empty");
    fail_unless(LL_first(list) == NULL, "New List has a first node");
    fail_unless(LL_last(list) == NULL,  "New List has a last node");
}
END_TEST

START_TEST(list_append)
{
    int     *elem;
    char    msg[1024];
    int     i;
    int     sz;

    LL_append(list, &(data[0]) );
    
    elem = LL_first(list);
    fail_unless(*elem == master[0], 
        "first() didn't return correct element, on first append");

    elem = LL_last(list);
    fail_unless(*elem == master[0], 
        "last() didn't return correct element, on first append");
    
    sz = LL_size(list);
    if (sz != 1) {
        snprintf(msg, sizeof(msg),
                "size() returned incorrect element count (%d, expecting %d)",
                sz, 1);
        fail(msg);
    }

    /* insert all elements checking as we go */
    for (i = 1; i < TEST_ELEM_COUNT; i++) {
        LL_append(list, &(data[i]));

        fail_unless(! LL_isEmpty(list), 
            "List with data still claims to be empty");

        sz = LL_size(list);
        if (sz != i+1) {
            snprintf(msg, sizeof(msg),
                "size() returned incorrect element count (%d, expecting %d)",
                sz, i+1);
            fail(msg);
        }

        /* first should never change, we are _appending_ */ 
        elem = LL_first(list);
        if (*elem != master[0]) {
            snprintf(msg, sizeof(msg), 
                "first() didn't return correct element, on append #%d", i);

            fail(msg);
        }

        /* last should be the element we just inserted */
        elem = LL_last(list);
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg), 
                "last() didn't return correct element, on append #%d", i);

            fail(msg);
        }
    }
}
END_TEST

START_TEST(list_prepend)
{
    int     *elem;
    char    msg[1024];
    int     i, sz;

    LL_prepend(list, &(data[0]));

    elem = LL_first(list);
    fail_unless(*elem == master[0],
        "first() didn't return correct element, on first prepend");

    elem = LL_last(list);
    fail_unless(*elem == master[0],
        "last() didn't return correct element, on first prepend");

    sz = LL_size(list);
    if (sz != 1) {
        snprintf(msg, sizeof(msg),
                "size() returned incorrect element count (%d, expecting %d)",
                sz, 1);
        fail(msg);
    }

    for (i = 1; i < TEST_ELEM_COUNT; i++) {
        LL_prepend(list, &(data[i]));

        fail_unless(! LL_isEmpty(list), 
            "List with data still claims to be empty");

        sz = LL_size(list);
        if (sz != i+1) {
            snprintf(msg, sizeof(msg),
                "size() returned incorrect element count (%d, expecting %d)",
                sz, i+1);
            fail(msg);
        }

        /* first should return the just inserted element */
        elem = LL_first(list);
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "first() didn't return correct element, on prepend #%d", i);

            fail(msg);
        }

        /* last should always return the same element, we are _prepending_ */
        elem = LL_last(list);
        if (*elem != master[0]) {
            snprintf(msg, sizeof(msg),
                "last() didn't return correct element, on prepend #%d", i);

            fail(msg);
        }
    }
}
END_TEST

START_TEST(list_rmfirst)
{
    int     *elem;
    char    msg[1024];
    int     i, sz;

    for (i = 0; i < TEST_ELEM_COUNT; i++) {
        LL_append(list, &(data[i]));
    }

    sz = LL_size(list);
    fail_unless(sz == TEST_ELEM_COUNT, 
        "size() returned incorrect element count after insertion");
    
    i = 0;
    elem = LL_rmFirst(list);
    while (elem != NULL) {
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "rmFirst() didn't return the correct element, position %d", i);
            fail(msg);
        }

        sz = LL_size(list);
        if (sz != TEST_ELEM_COUNT - i - 1) {
            snprintf(msg, sizeof(msg),
                "size() didn't return correct element count (%d, expected %d)",
                sz, TEST_ELEM_COUNT - i - 1);
            fail(msg);
        }

        elem = LL_rmFirst(list);
        i++;
    }
}
END_TEST

START_TEST(list_rmlast)
{
    int     *elem;
    char    msg[1024];
    int     i, sz;

    mark_point();

    for (i = 0; i < TEST_ELEM_COUNT; i++) {
        LL_append(list, &(data[i]));
    }

    sz = LL_size(list);
    fail_unless(sz == TEST_ELEM_COUNT, 
        "size() returned incorrect element count after insertion");
    
    i = TEST_ELEM_COUNT - 1;
    elem = LL_rmLast(list);
    while (elem != NULL) {
        mark_point();

        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "rmLast() didn't return the correct element, position %d", i);
            fail(msg);
        }

        mark_point();
        sz = LL_size(list);
        if (sz != i) {
            snprintf(msg, sizeof(msg),
                "size() didn't return correct element count (%d, expected %d)",
                sz, i);
            fail(msg);
        }

        mark_point();
        elem = LL_rmLast(list);
        i--;

        mark_point();
    }
}
END_TEST

START_TEST(list_remove)
{
    int     *elem;
    char    msg[4096];
    int     tail, head; 
    int     sz, expectedSz;
    int     i;

    for (i = 0; i < TEST_ELEM_COUNT; i++) {
        LL_append(list, &(data[i]));
    }

    sz = LL_size(list);
    fail_unless(sz == TEST_ELEM_COUNT, 
        "size() returned incorrect element count after insertion");
    
    head = 0;
    tail = TEST_ELEM_COUNT - 1;
    i = 1;
    expectedSz = TEST_ELEM_COUNT;
    do {
        elem = LL_rmLast(list);
        expectedSz--;

        if (elem == NULL) break; 

        if (*elem != master[tail]) {
            snprintf(msg, sizeof(msg),
                "rmLast() didn't return the correct element, iteration %d", i);
            fail(msg);
        }

        sz = LL_size(list);
        if (sz != expectedSz) {
            snprintf(msg, sizeof(msg),
                "size() didn't return correct element count (%d, expected %d)",
                sz, expectedSz);
            fail(msg);
        }
        tail--;


        elem = LL_rmFirst(list);
        expectedSz--;

        if (elem == NULL) break;

        if (*elem != master[head]) {
            snprintf(msg, sizeof(msg),
                "rmFirst() didn't return the correct element, iteration %d",i);
            fail(msg);
        }

        sz = LL_size(list);
        if (sz != expectedSz) {
            snprintf(msg, sizeof(msg),
                "size() didn't return correct element count (%d, expected %d)",
                sz, expectedSz);
            fail(msg);
        }
        head++;

        /* condition is redundant */
        i++;
    } while (elem != NULL);

    fail_unless(LL_size(list) == 0, 
            "Not all elements were removed by size()");
    fail_unless(LL_first(list) == NULL, 
            "Not all elements were removed by first()");
    fail_unless(LL_last(list) == NULL,
            "Not all elements were removed by first()");
}
END_TEST

/*********************** List Iterator Test Cases **************************/

START_TEST(lliterator_create)
{
    fail_unless(itr != NULL, "Iterator is NULL after creation");
    fail_unless(LLI_elem(itr) == NULL, 
        "Iterator returns element just after creation");
}
END_TEST

START_TEST(lliterator_next)
{
    int         *elem;
    int         i;
    char        msg[1024];

    i = 0;
    elem = LLI_next(itr);
    while (elem != NULL) {
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "LLI_next() didn't return correct value (element %d)", i);

            fail(msg);
        }
        i++;
        elem = LLI_next(itr);
    }

    if (i != TEST_ELEM_COUNT) {
        snprintf(msg, sizeof(msg),
            "found %d elements, was expecting %d", i, TEST_ELEM_COUNT);

        fail(msg);
    }

    /* now try to advance past the end a few times */
    for (i = 0; i < 3; i++) {
        elem = LLI_next(itr);

        fail_unless(elem == NULL,
            "iterator resets after the end of the list.");

    }
}
END_TEST

START_TEST(lliterator_prev)
{
    int         *elem, i;
    char        msg[1024];

    i = TEST_ELEM_COUNT - 1;
    elem = LLI_prev(itr);
    while (elem != NULL) {
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "LLI_prev() didn't return correct value (element #%d)", i);

            fail(msg);
        }
        i--;
        elem = LLI_prev(itr);
    }

    if (i != -1) {
        snprintf(msg, sizeof(msg),
            "found %d elements, was expecting %d", 
            TEST_ELEM_COUNT - 1 - i, TEST_ELEM_COUNT);

        fail(msg);
    }

    /* now try to advance past the end a few times */
    for (i = 0; i < 3; i++) {
        elem = LLI_prev(itr);

        fail_unless(elem == NULL,
            "iterator resets after the end of the list.");
    }
}
END_TEST

START_TEST(lliterator_elem)
{
    int             *elem, *melem;
    char            msg[1024];

    melem = LLI_next(itr);
    while (melem != NULL) {
        elem = LLI_elem(itr);
        if (*elem != *melem) {
            snprintf(msg, sizeof(msg),
                "elem (value %d) does not match current element (value %d)", 
                *elem, *melem);

            fail(msg);
        }
        melem = LLI_next(itr);
    }
}
END_TEST

START_TEST(lliterator_bidir)
{
    int     i, *elem;
    char    msg[1024];

    /* first go forward */
    i = 0;
    elem = LLI_next(itr);
    while (elem != NULL) {
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "LLI_next() didn't return correct value (element %d)", i);

            fail(msg);
        }
        i++;
        elem = LLI_next(itr);
    }

    if (i != TEST_ELEM_COUNT) {
        snprintf(msg, sizeof(msg),
            "found %d elements, was expecting %d", i, TEST_ELEM_COUNT);

        fail(msg);
    }

    /* then go back */
    i--;
    elem = LLI_prev(itr);
    while (elem != NULL) {
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "LLI_prev() didn't return correct value (element #%d)", i);

            fail(msg);
        }
        i--;
        elem = LLI_prev(itr);
    }

    if (i != -1) {
        snprintf(msg, sizeof(msg),
            "found %d elements, was expecting %d", 
            TEST_ELEM_COUNT - 1 - i, TEST_ELEM_COUNT);

        fail(msg);
    }

    /* and forward again */
    i++;
    elem = LLI_next(itr);
    while (elem != NULL) {
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "LLI_next() didn't return correct value (element %d)", i);

            fail(msg);
        }
        i++;
        elem = LLI_next(itr);
    }

    if (i != TEST_ELEM_COUNT) {
        snprintf(msg, sizeof(msg),
            "found %d elements, was expecting %d", i, TEST_ELEM_COUNT);

        fail(msg);
    }

    /* and back once more */
    i--;
    elem = LLI_prev(itr);
    while (elem != NULL) {
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "LLI_prev() didn't return correct value (element #%d)", i);

            fail(msg);
        }
        i--;
        elem = LLI_prev(itr);
    }

    if (i != -1) {
        snprintf(msg, sizeof(msg),
            "found %d elements, was expecting %d", 
            TEST_ELEM_COUNT - 1 - i, TEST_ELEM_COUNT);

        fail(msg);
    }
}
END_TEST

START_TEST(lliterator_insert)
{
    int     data2[TEST_ELEM_COUNT + 1];
    int     master2[TEST_ELEM_COUNT + 1];
    int     i, *elem, sz;
    char    msg[1024];

    for (i = 0; i < TEST_ELEM_COUNT + 1; i++) {
        data2[i] = i * 20;
        master2[i] = data2[i];
    }

    for (i = 0; i < TEST_ELEM_COUNT + 1; i++) {
        LLI_insert(itr, &(data2[i]));
        elem = LLI_next(itr);
        if (*elem != master2[i]) {
            snprintf(msg, sizeof(msg),
                "LLI_insert() failed at position %d", i);
            fail(msg);
        }

        sz = LL_size(list);
        if (sz != i+1+TEST_ELEM_COUNT) {
            snprintf(msg, sizeof(msg),
                "LL_size() returned incorrect element count (%d, expected %d)",
                sz, i+1+TEST_ELEM_COUNT);
            fail(msg);
        }

        if (i >= TEST_ELEM_COUNT) continue;

        elem = LLI_next(itr);
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "original data not properly preserved, position %d", i);
            fail(msg);
        }
    }
}
END_TEST

START_TEST(lliterator_remove)
{
    int     i, *elem, *relem;
    int     sz, expectedSz;
    char    msg[1024];

    expectedSz = TEST_ELEM_COUNT;

    for (i = 0; i < TEST_ELEM_COUNT; i += 2) {
        elem = LLI_next(itr);
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "next didn't return correct elemnt, position %d", i);
            fail(msg);
        }

        relem = LLI_remove(itr);
        expectedSz--;
        if (*relem != *elem) {
            snprintf(msg, sizeof(msg),
                "remove didn't return correct element, position %d", i);
            fail(msg);
        }

        sz = LL_size(list);
        if (sz != expectedSz) {
            snprintf(msg, sizeof(msg),
                "LL_size() returned incorrect element count (%d, expected %d)",
                sz, expectedSz);
            fail(msg);
        }

        elem = LLI_next(itr);
        if (*elem != master[i+1]) {
            snprintf(msg, sizeof(msg),
                "next didn't return correct elemnt, position %d", i);
            fail(msg);
        }
    }

    elem = LLI_next(itr);
    fail_unless(elem == NULL, "not on last element after first remove round");

    for (i = TEST_ELEM_COUNT - 1; i > 0; i -= 2) {
        elem = LLI_prev(itr);
        if (*elem != master[i]) {
            snprintf(msg, sizeof(msg),
                "prev didn't return correct elemnt, position %d", i);
            fail(msg);
        }

        relem = LLI_remove(itr);
        expectedSz--;
        if (*relem != *elem) {
            snprintf(msg, sizeof(msg),
                "remove didn't return correct element, position %d", i);
            fail(msg);
        }

        sz = LL_size(list);
        if (sz != expectedSz) {
            snprintf(msg, sizeof(msg),
                "LL_size() returned incorrect element count (%d, expected %d)",
                sz, expectedSz);
            fail(msg);
        }
    }
}
END_TEST

