/**
 * @file ut_netaddr.c
 *
 * @brief Test cases for network address abstraction.
 * @ingroup utb_netaddr
 *
 * $Id: ut_netaddr.c,v 1.10.2.1 2003/04/30 20:03:44 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Feb 14 11:16:26 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>

#include "check.h"

#include "util/netaddr.h"
#include "ipc/buffer.h"
#include "debug.h"

#include "ut_driver.h"
#include "ut_base.h"

/**@name Network Address Test Cases
 * @ingroup utb_netaddr
 * @{
 */

/** Test NetAddr creation. */
static void netaddr_create(void);

/** Test NetAddr Comparison */
static void netaddr_compare(void);

/** Test Creation of struct sockaddrs */
static void netaddr_mksockaddr(void);

/** Test NetAddr serialization */
static void netaddr_buffer(void);

/** Test NetAddr printing */
static void netaddr_print(void);

/*@}*/

void debug_init(void)
{
    if (! DBG_init("n", "netaddr.debug") ) {
        perror("Error setting up debugging:");
        exit(1);
    }
}

void debug_finish(void)
{
    DPRINT(DBG_DBG, "Test Case Done", NO_VARG);
}

/* test suite setup */
void netaddr_suite(Suite *suite)
{
    TCase   *tc_netaddr;

    tc_netaddr = tcase_create("Network Address");
    suite_add_tcase(suite, tc_netaddr);

    tcase_add_unchecked_fixture(tc_netaddr, debug_init, debug_finish);


    tcase_add_test(tc_netaddr, netaddr_create);
    tcase_add_test(tc_netaddr, netaddr_compare);
    tcase_add_test(tc_netaddr, netaddr_mksockaddr);
    tcase_add_test(tc_netaddr, netaddr_buffer);
	tcase_add_test(tc_netaddr, netaddr_print);
}


/**@fn netaddr_create
 * @bug Doesn't test INET6 addresses 
 */ 
START_TEST(netaddr_create)
{
    netaddr_t       na1, na2;
	netaddr_t		localhost,lookupLocal;
    struct in_addr  addr;


    inet_aton("192.168.5.1", &addr);

    na1 = NA_newINet4(addr);

    fail_unless(na1 != NULL, 
        "create returned NULL");

    fail_unless(na1->type == NAT_IP4, 
        "wrong address type in first netaddr.");

    fail_unless(na1->addr.ip4 == addr.s_addr,
        "wrong IP address in first netaddr");


    na2 = NA_newNetaddr(na1);

    fail_unless(na2 != NULL, 
        "copy ctor returned NULL");

    fail_unless(na1->type == na2->type, 
        "copied netaddr has different type than original");

    fail_unless(na1->addr.ip4 == na2->addr.ip4,
        "Copied netaddr has different IP than original");

	inet_aton("127.0.0.1", &addr);
	localhost=NA_newINet4(addr);
	lookupLocal=NA_newINetLookup("localhost");
	fail_if(NA_isError(localhost),"Failed to create netaddr from 127.0.0.1");
	fail_if(NA_isError(lookupLocal),
			"Failed to create netaddr from localhost; perhaps you need to add it to /etc/hosts");
	fail_unless(localhost->addr.ip4 == lookupLocal->addr.ip4,
			"localhost seems to not point to 127.0.0.1");

    NA_destroy(na1);
    NA_destroy(na2);
	NA_destroy(localhost);
	NA_destroy(lookupLocal);
}
END_TEST

START_TEST(netaddr_compare)
{
    netaddr_t       rhs, lhs;
    struct in_addr  addr;
    int             rslt;

    inet_aton("192.168.0.0", &addr);
    rhs = NA_newINet4(addr);

    inet_aton("192.168.0.1", &addr);
    lhs = NA_newINet4(addr);

    rslt = NA_compare(lhs, rhs);
    fail_unless(rslt > 0, "lhs > rhs, but compare did not return > 0");

    rslt = NA_compare(rhs, lhs);
    fail_unless(rslt < 0, "lhs < rhs, but compare did not return < 0");

    rhs = NA_newINet4(addr);

    rslt = NA_compare(lhs, rhs);
    fail_unless(rslt == 0, "lhs == rhs, but compare did not return == 0");
}
END_TEST

#define TEST_ADDR   "127.0.0.1"
#define TEST_PORT   514
#define TEST_PATH   "/tmp/test-path"
START_TEST(netaddr_mksockaddr)
{
    netaddr_t           na_inet;
    netaddr_t           na_unix;
    struct in_addr      addr;
    union NA_sockaddr   sa_inet;
    union NA_sockaddr   sa_unix;

    fail_unless(inet_aton(TEST_ADDR, &addr), "inet_aton()");

    na_inet = NA_newINet4(addr);
    na_unix = NA_newUnix(TEST_PATH);

    fail_if(NA_isError(na_inet), "NA_newINet4()");
    fail_if(NA_isError(na_unix), "NA_newUnix()");

    fail_unless(NA_mkSockaddr(na_inet, TEST_PORT, &sa_inet), 
            "NA_mkSockaddr(INET)");
    DPRINT(DBG_NA, "test: sa_inet = %s:%d, family %d (%d)",
            inet_ntoa(sa_inet.sin.sin_addr), ntohs(sa_inet.sin.sin_port),
            sa_inet.sin.sin_family, sa_inet.sa.sa_family);

    fail_unless(NA_mkSockaddr(na_unix, 0, &sa_unix),
            "NA_mkSockaddr(UNIX)");

    DPRINT(DBG_NA, "test: sa_inet = %s:%d, family %d (%d)",
            inet_ntoa(sa_inet.sin.sin_addr), ntohs(sa_inet.sin.sin_port),
            sa_inet.sin.sin_family, sa_inet.sa.sa_family);

    /* check the inet sockaddr */
    fail_unless(sa_inet.sa.sa_family == PF_INET, 
            "INET: family is wrong");
    fail_unless(memcmp(&(sa_inet.sin.sin_addr), &addr, sizeof(addr)) == 0,
            "INET: addr is wrong");
    fail_unless(ntohs(sa_inet.sin.sin_port) == TEST_PORT, 
            "INET: port is wrong");

    /* check the unix sockaddr */
    fail_unless(sa_unix.sa.sa_family == PF_UNIX,
            "UNIX: family is wrong");
    fail_unless(strcmp(sa_unix.sun.sun_path, TEST_PATH) == 0,
            "UNIX: path is wrong");

    NA_destroy(na_inet);
    NA_destroy(na_unix);
}
END_TEST
#undef TEST_PATH
#undef TEST_PORT
#undef TEST_ADDR

START_TEST(netaddr_buffer)
{
    netaddr_t           na;
    netaddr_t           nab;
    struct in_addr      addr;
    struct IPC_buffer   buf;
    IPC_bufMark         bStart;
    int                 rslt;

    fail_unless(inet_aton("192.168.5.2", &addr), "inet_aton() failed");

    na = NA_newINet4(addr);
    fail_unless(!NA_isError(na), "NA_newINet4() failed");

    BUF_init(&buf);
    buf.size = NA_MAX_BUFLEN;
    BUF_alloc(&buf);

    bStart = BUF_markPos(&buf);

    rslt = NA_toBuffer(na, &buf);
    fail_unless(rslt >= 0, "NA_toBuffer() had error");
    fail_unless(rslt == sizeof(addr)+1, "NA_toBuffer() didn't write enough data");

    fail_unless(BUF_resetToPos(&buf, bStart), "resting buf position failed");

    nab = NA_newBuffer(&buf);
    fail_unless(!NA_isError(nab), "NA_newBuffer() failed");

    rslt = NA_compare(na, nab);
    fail_unless(rslt == 0, "the Netaddr got changed in the IPCBuffer");
}
END_TEST

START_TEST(netaddr_print)
{
	netaddr_t			na;
	struct in_addr		addr;

	fail_unless(inet_aton("192.168.12.1",&addr),"inet_aton() failed");

	na=NA_newINet4(addr);
	fail_if(NA_isError(na),"NA_newINet4() failed");
	fail_unless(0==strncmp(NA_getCharAddr(na),"192.168.12.1",
				strlen("192.168.12.1")+1),"NA_getCharAddr() failed");

	NA_destroy(na);

	na=NA_newUnix("/dev/null");
	fail_if(NA_isError(na),"NA_newUnix() failed");
	fail_unless(0==strncmp(NA_getCharAddr(na),"/dev/null",
				strlen("/dev/null")+1),"NA_getCharAddr() failed");

	NA_destroy(na);
}
END_TEST


