/**
 * @file ut_pathcache.c
 *
 * @brief Path Cache Unit Test driver.
 * @ingroup utb_pathcache
 *
 * $Id: ut_pathcache.c,v 1.3.2.5 2003/04/30 20:03:44 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Jul 08 19:37:01 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>

#include "check.h"
#include "util/pathcache.h"
#include "types.h"
#include "error.h"

#include "ut_driver.h"
#include "ut_base.h"

/**@name Path Cache Test Cases
 * @ingroup utb_pathcache
 * @{
 */

/** Test the path cache as the controller will use it. */
static void pc_ctlr(void);

/** Test the path cache as the input modules will use it. */
static void pc_im(void);

/** Test the path cache as the output modules will use it. */
static void pc_om(void);

/*@}*/

/** Number of path objects to test with */
#define PATH_COUNT  10

/** Number of connection id's used (sequential) */
#define CONN_COUNT  4

/** Test paths to insert malloc()ed */
static struct path  *ipaths[PATH_COUNT];

/** Copies of ins_paths that will always be valid */
static struct path  mpaths[PATH_COUNT];

/** Initilize the ipaths and mpaths structures */
static void pc_setup(void)
{
    int         i;
    uint32_t    connid;

    connid = 1;
    for (i = 0; i < PATH_COUNT; i++) {
        if (i > 0 && (i % 2) == 0 && i < CONN_COUNT) {
            connid++;
        }

        ipaths[i] = emalloc(sizeof(struct path));
        memset(ipaths[i], 0, sizeof(struct path));
        ipaths[i]->id = i+1;
        ipaths[i]->connid = connid;
        /* only fields we care about */

        mpaths[i] = *(ipaths[i]);
    }
    DPRINT(DBG_PC, "========== Start New Test ===========", NO_VARG);
}

static void pc_teardown(void)
{
    /* do nothing. The Path cache destroy function should delete the
     * paths.  
     */
    DPRINT(DBG_PC, "============ End Test =============", NO_VARG);
}

static void debug_init(void)
{
    DBG_init("te", "pathcache.debug");
}

static void noop(void)
{
}


void pathcache_suite(Suite *suite)
{
    TCase   *tc_pc;

    tc_pc = tcase_create("Path Cache");
    suite_add_tcase(suite, tc_pc);

    tcase_add_checked_fixture(tc_pc, pc_setup, pc_teardown);
    tcase_add_unchecked_fixture(tc_pc, debug_init, noop);

    tcase_add_test(tc_pc, pc_ctlr);
    tcase_add_test(tc_pc, pc_im);
    tcase_add_test(tc_pc, pc_om);
}

START_TEST(pc_ctlr)
{
    PCache      *cache;
    int         conns[CONN_COUNT];
    bool        found;
    int         i;

    memset(conns, 0, sizeof(conns));
    srand(time(NULL));  /* not quite deterministic test cases can be handy */

    cache = PCACHE_ctlrCreate();
    fail_if(cache == NULL, "PCACHE_ctlrCreate() failed");

    for (i = 0; i < PATH_COUNT; i++) {
        /* make sure that it is not present */
        fail_if(PCACHE_findPath(cache,mpaths[i].connid,mpaths[i].id),
            "Found Path that was not inserted");
        
        fail_unless(PCACHE_addPath(cache, ipaths[i], FALSE),
            "Failed to add path");

        fail_if(NULL ==
            PCACHE_findPath(cache, mpaths[i].connid, mpaths[i].id),
            "Couldn't find path that was just inserted");

        conns[mpaths[i].connid]++;
    }

    for (i = 0; i < PATH_COUNT; i++) {
        if (conns[mpaths[i].connid] ) {
            fail_if(NULL == 
                PCACHE_findPath(cache, mpaths[i].connid, mpaths[i].id),
                "Failed to find path after all inserts");
        }

        /* choose about 1/2 of the paths and clear all paths with
         * matching connection ids.
         */
        if ( (rand() % 100) < 50) {
            if (conns[mpaths[i].connid] ) {

                fail_unless(PCACHE_clearConn(cache, mpaths[i].connid),
                    "PCACHE_clearConn() failed");

                conns[mpaths[i].connid] = 0;
            }
        }
    }

    for (i = 0; i < PATH_COUNT; i++) {
        found = (NULL != PCACHE_findPath(cache,mpaths[i].connid,mpaths[i].id));

        if (conns[mpaths[i].connid]) {
            fail_unless(found, "Can't find path that should still be present");
        }
        else {
            fail_if(found, "Found Path that should have been deleted");

        }
    }

    PCACHE_destroy(cache);
}
END_TEST

START_TEST(pc_im)
{
    PCache      *cache;
    int         conns[CONN_COUNT];
    int         rtn;
    int         i;


    memset(conns, 0, sizeof(conns));
    srand(time(NULL));  /* not quite deterministic test cases can be handy */

    cache = PCACHE_imCreate();
    fail_if(cache == NULL, "PCACHE_ctlrCreate() failed");

    for (i = 0; i < PATH_COUNT; i++) {
        /* make sure that it is not present */
        fail_unless(
            PCACHE_aclCheck(cache,mpaths[i].connid,mpaths[i].id) < 0,
            "Found Path that was not inserted");
        
        /* alternate between true and false */
        fail_unless(PCACHE_addPath(cache, ipaths[i], (i % 2) == 0),
            "Failed to add path");

        rtn = PCACHE_aclCheck(cache,mpaths[i].connid,mpaths[i].id);
        DPRINT(DBG_PC, "test: aclCheck = %d, i = %d", rtn, i);
        fail_if(rtn < 0, "Couldn't find path that was just inserted");
        fail_if((i % 2) == 0 && rtn == 0, "Acl Check was wrong");
        fail_if((i % 2) != 0 && rtn != 0, "Acl Check was wrong");

        conns[mpaths[i].connid]++;
    }

    for (i = 0; i < PATH_COUNT; i++) {
        /* only look if we haven't deleted it yet */
        if (conns[mpaths[i].connid]) {
            rtn = PCACHE_aclCheck(cache,mpaths[i].connid,mpaths[i].id);
            fail_if(rtn < 0, "Failed to find path after all inserts");
            fail_if((i % 2) == 0 && rtn == 0, "Acl Check was wrong");
            fail_if((i % 2) != 0 && rtn != 0, "Acl Check was wrong");
        }

        /* choose about 1/2 of the paths and clear all paths with
         * matching connection ids.
         */
        if ( (rand() % 100) < 50) {
            if (conns[mpaths[i].connid] ) {
                fail_unless(PCACHE_clearConn(cache, mpaths[i].connid),
                    "PCACHE_clearConn() failed");

                conns[mpaths[i].connid] = 0;
            }
        }
    }

    for (i = 0; i < PATH_COUNT; i++) {
        rtn = PCACHE_aclCheck(cache,mpaths[i].connid,mpaths[i].id);

        if (conns[mpaths[i].connid]) {
            /* Should be found and aclPass should be correct */
            fail_if(rtn < 0, "Couldn't find path that should be there");
            fail_if((i % 2) == 0 && rtn == 0, "Acl Check was wrong");
            fail_if((i % 2) != 0 && rtn != 0, "Acl Check was wrong");
        }
        else {
            fail_unless(rtn < 0, "Found path that shouldn't be there");
        }
    }
    PCACHE_destroy(cache);
}
END_TEST

START_TEST(pc_om)
{
    PCache      *cache;
    int         conns[CONN_COUNT];
    int         i;
    uint32_t    out_ids[PATH_COUNT];
    uint32_t    rtn;
    uint32_t    om[2];
    struct path *path;


    memset(conns, 0, sizeof(conns));
    memset(out_ids, 0, sizeof(out_ids));
    srand(time(NULL));  /* not quite deterministic test cases can be handy */

    cache = PCACHE_ctlrCreate();
    fail_if(cache == NULL, "PCACHE_ctlrCreate() failed");

    om[0] = PCACHE_regOutputMod(cache);
    om[1] = PCACHE_regOutputMod(cache);

    for (i = 0; i < PATH_COUNT; i++) {
        /* make sure that it is not present */
        fail_if(PCACHE_findPath(cache,mpaths[i].connid,mpaths[i].id),
            "Found Path that was not inserted");
        
        fail_unless(PCACHE_addPath(cache, ipaths[i], FALSE),
            "Failed to add path");

        rtn = PCACHE_findOutgoing(cache,
                mpaths[i].connid,mpaths[i].id, om[0]);

        fail_unless(rtn == PATH_UNSET_ID, 
            "PCACHE_findOutgoing inccorectly succededed");

        path = PCACHE_makeOutgoing(cache, 
                mpaths[i].connid, mpaths[i].id, om[0]);
        fail_if(path == NULL, "PCACHE_makeOutgoing() failed");
        out_ids[i] = path->id;

        conns[mpaths[i].connid]++;
    }

    for (i = 0; i < PATH_COUNT; i++) {
        if (conns[mpaths[i].connid]) {
            rtn = PCACHE_findOutgoing(cache, 
                    mpaths[i].connid, mpaths[i].id, om[i % 2]);
            if (i % 2 == 1) {
                /* we should have _not_ found the path */
                fail_unless(rtn == PATH_UNSET_ID, 
                    "found outgoing pathID for OM that never sent it");
                path = PCACHE_makeOutgoing(cache,
                        mpaths[i].connid, mpaths[i].id, om[1]);
                fail_if(path == NULL, "PCACHE_makeOutgoing() failed");
                rtn = path->id;
            }
            DPRINT(DBG_PC, "rtn == %u, out_ids[i == %d] == %u",
                    rtn, i, out_ids[i]);
            fail_unless(rtn == out_ids[i],
                "Failed to find correct outgoing pathID after all inserts");
        }

        /* choose about 1/2 of the paths and clear all paths with
         * matching connection ids.
         */
        if ( (rand() % 100) < 50) {
            if (conns[mpaths[i].connid] ) {
                fail_unless(PCACHE_clearConn(cache, mpaths[i].connid),
                    "PCACHE_clearConn() failed");

                conns[mpaths[i].connid] = 0;
            }
        }
    }

    for (i = 0; i < PATH_COUNT; i++) {
        rtn = PCACHE_findOutgoing(cache,
                mpaths[i].connid,mpaths[i].id, om[i % 2]);

        if (conns[mpaths[i].connid]) {
            DPRINT(DBG_PC, "rtn == %u, i == %d, connid = %u",
                    rtn, i, mpaths[i].connid);

            fail_if(rtn == PATH_UNSET_ID, 
                "Can't find path that should still be present");
            fail_unless(rtn == out_ids[i], "Found wrong outgoing path id");
        }
        else {
            fail_unless(rtn == PATH_UNSET_ID, 
                "Found Path that should have been deleted");
        }
    }

    /* Now we pretend that we got a new connection, and need to make
     * sure that no paths have been set.
     */
    PCACHE_omNewConn(cache, om[0]);
    for (i = 0; i < PATH_COUNT; i++) {
        rtn = PCACHE_findOutgoing(cache, mpaths[i].connid, mpaths[i].id, om[0]);
        fail_unless(rtn == PATH_UNSET_ID, "Found sent path after new conn");
    }

    /* Last check, make sure that we can create a new local path
     * correctly.
     */
    rtn = PCACHE_findOutgoing(cache, PATH_UNSET_ID, PATH_UNSET_ID, om[0]);
    fail_unless(rtn == PATH_UNSET_ID, "Got local path unexpectedly");

    path = PCACHE_makeOutgoing(cache, PATH_UNSET_ID, PATH_UNSET_ID, om[0]);
    fail_if(path == NULL, "Failed to make new local path");

    rtn = PCACHE_findOutgoing(cache, PATH_UNSET_ID, PATH_UNSET_ID, om[0]);
    fail_unless(path->id == rtn, "Wrong PID, in second local findOutgoing");

    PCACHE_unregOutputMod(cache, om[0]);
    PCACHE_unregOutputMod(cache, om[1]);
    PCACHE_destroy(cache);
}
END_TEST
