/**
 * @file common_funcs.c
 *
 * @brief Common functions to several unit tests
 * @par
 *
 * $Id: common_funcs.c,v 1.1.2.6 2003/04/30 20:03:40 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Sep  5 14:44:01 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <sys/time.h>
#include <errno.h>

#include "debug.h"
#include "message/message.h"

#include "check.h"

#include "common_funcs.h"

#define CMP_int(a,b,member,buf,file,line)									\
	if(a->member!=b->member) {												\
		snprintf(buf,sizeof(buf), "%s doesn't match ('%d' != '%d')",		\
				#member, (int)a->member, (int)b->member);					\
		_fail_unless(0,file,line,buf);										\
	}

#define CMP_str(a,b,member,buf,file,line)									\
	if( NULL==a->member && NULL==b->member) {								\
		/* do nothing */													\
	}																		\
	else if ( NULL==a->member || NULL==b->member) {							\
		snprintf(buf,sizeof(buf), "%s doesn't match ('%s' != '%s')",		\
				#member, a->member, b->member);								\
		_fail_unless(0,file,line,buf);										\
	}																		\
	else if ( 0!=strcmp(a->member,b->member) ) {							\
		snprintf(buf,sizeof(buf), "%s doesn't match ('%s' != '%s')",		\
				#member, a->member, b->member);								\
		_fail_unless(0,file,line,buf);										\
	}

#define CMP_na(a,b,member,buf,file,line)									\
	if( NA_compare(a->member,b->member) != 0 ) {							\
		snprintf(buf,sizeof(buf),"%s doesn't match (names can't be provided)",			\
				#member);													\
		_fail_unless(0,file,line,buf);										\
	}

void
internal_msgCompare(struct message *a,struct message *b,char *file,int line)
{
	char errstr[1024];

	if(a==NULL && b==NULL)
		return;

	if(a==NULL)
		fail("a is null, b is not");
	if(b==NULL)
		fail("b is null, a is not");

	CMP_int(a,b,sev,errstr,file,line);
	CMP_int(a,b,fac,errstr,file,line);

	CMP_int(a,b,hostlen,errstr,file,line);
	CMP_str(a,b,host,errstr,file,line);

	/* if we set one of these to -1, it means we're expecting for them to be
	 * different, so we ignore the error */
	if( a->timestamp!=-1 && b->timestamp!=-1)
	{
		CMP_int(a,b,timestamp,errstr,file,line);
		CMP_int(a,b,tz,errstr,file,line);
	}

	CMP_str(a,b,tag,errstr,file,line);

	CMP_int(a,b,msglen,errstr,file,line);
	CMP_str(a,b,msg,errstr,file,line);
}

void
internal_CMP_path(struct path *a,struct path *b,char *file,int line)
{
	char errstr[1024];

	CMP_int(a,b,id,errstr,file,line);
	CMP_int(a,b,connid,errstr,file,line);

	CMP_int(a,b,ffqdnlen,errstr,file,line);
	CMP_str(a,b,fromFQDN,errstr,file,line);
	CMP_na(a,b,fromIP,errstr,file,line);

	CMP_int(a,b,tfqdnlen,errstr,file,line);
	CMP_str(a,b,toFQDN,errstr,file,line);
	CMP_na(a,b,toIP,errstr,file,line);

	CMP_int(a,b,props,errstr,file,line);
}

int
getTZ(void)
{
	char errstr[1024];
	struct timezone tz;

	if(-1==gettimeofday(NULL,&tz))
	{
		snprintf(errstr,1024,"gettimeofday() failed: %s",strerror(errno));
		fail(errstr);
	}

	return -1*tz.tz_minuteswest;
}
