/**
 * @file ut_config.h
 *
 * @brief Declare the suite setup functions for the config  unit tests.
 * @ingroup ut_config
 *
 * $Id: ut_config.h,v 1.16.2.6 2003/04/30 20:03:46 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Feb 05 16:53:32 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef TEST_UNIT_CONFIG_BASE_H
#define TEST_UNIT_CONFIG_BASE_H

#include "chaninfo.h"

/**@name Config Reader Tests
 * @ingroup ut_config
 * @{
 */

/** Setup the file test suites
 * @param suite The test suite to setup.
 */
void config_read_suite(Suite *suite);

/*@}*/

/**@name Config Symbol Table Tests
 * @ingroup ut_config
 * @{
 */

/** Setup the symbol table test suites
 * @param suite The test suite to setup
 */
void config_symtab_suite(Suite *suite);

/*@}*/

/**@name Config Object Tests
 * @ingroup ut_config
 * @{
 */

/** Setup the object creation/deletion test suites
 * @param suite THe test suite to setup
 */
void config_object_suite(Suite *suite);

/*@}*/

/**@name Config Test Helper Functions
 * @ingroup ut_config
 * @{
 */

/** Check two inchan_info together
 *
 * @param a The first input to check
 * @param b The second input to check
 */
#define check_inchan_info(a,b) check_inchan_info_internal(a,b,__LINE__,__FILE__)

/** Check two inchan_info together; private function, use check_inchan_info() instead
 *
 * @param a The first input to check
 * @param b The second input to check
 * @param line The line number we were called from
 * @param file The file name we were called from
 */
void check_inchan_info_internal(struct inchan_info *a,struct inchan_info *b,int line,char *file);

/** Check two outchan_infos against each other
 *
 * @param a The first output to check
 * @param b The second output to check
 */
#define check_outchan_info(a,b) check_outchan_info_internal(a,b,__LINE__,__FILE__)

/** Check two outchan_infos together; private function, use check_outchan_info() 
 *  instead
 *
 * @param a The first output to check
 * @param b The second output to check
 * @param line The line number we were called from
 * @param file The file name we were called from
 */
void check_outchan_info_internal(struct outchan_info *a,struct outchan_info *b,
		int line,char *file);

/** Check two netaddr_t objects
 *
 * @param a The first netaddr_t to check
 * @param b The second netaddr_t to check
 * @param errstr The error string to print
 */
void check_netaddr_t(netaddr_t a,netaddr_t b,char *errstr);

/** Create a inchan_info structure with the specified values
 *
 * @param name The name member (the function copies this string, if it's not NULL)
 * @param type The type member
 * @param inAddr The listen member (the function copies this, if it's not NULL)
 * @param inPath The path (as a char *) to listen on
 * @param path_acl The path ACL (the function copies this pointer directly)
 * @param psec_acl The path security ACL (again, this pointer is copied)
 * @param hsrc_acl The host source ACL (the pointer is copied)
 * @param hcon_acl THe connecting host ACL (the pointer is copied)
 *
 * @return The new struct inchan_info; this must be destroyed with the proper 
 * 				conf_sdsc_destroy* function
 */
struct inchan_info * create_inchan_info(char *name,IMOD_TYPE type, 
        netaddr_t inAddr, char *inPath,int listen_port,ACL *path_acl,
        ACL *psec_acl,ACL *hsrc_acl,ACL *hcon_acl);

/** Create a outchan_info structure with the specified values
 *
 * @param type The type member
 * @param formattype The format type member
 * @param fopts The formatter options
 * @param name The name (the function copies this string, if it's not NULL)
 * @param outPath The output path (this string is copied, if it's not NULL)
 * @param outAddr The output address (this is copied, if it's not NULL)
 * @param outPort The output port
 * @param facmask The facility mask
 * @param sevmask The severity mask
 *
 * @return The new struct outchan_info; this must be destroyed with 
 * 				conf_sdsc_destroy_outchan_info()
 */
struct outchan_info *create_outchan_info(OMOD_TYPE type,
        MFRMT_TYPE formattype, uint32_t fopts, char *name,
        char *outPath,netaddr_t outAddr,int outPort,uint8_t *priority);

/*@}*/


#endif /* TEST_UNIT_CONFIG_BASE_H */
