/**
 * @file ut_config_helper.c
 *
 * @brief Define the helper functions
 * @par
 *
 * $Id: ut_config_helper.c,v 1.14.2.6 2003/04/30 20:03:46 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Fri Apr 26 15:35:51 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "check.h"
#include "error.h"

#include "strcompat.h"
#include "util/list.h"
#include "chaninfo.h"

#include "config/config.h"
#include "config/conf_private.h"
#include "config/conf_obj.h"

#include "ut_config.h"

void
check_inchan_info_internal(struct inchan_info *a,struct inchan_info *b,int line,char *file)
{
	char *errstr=emalloc(sizeof(char)*200);

	if( (NULL==a) && (NULL==b) )
		return;

	/* check to make sure that one isn't null; this is cause for not continuing with the
	 * function, or we'll segfault */
	snprintf(errstr,200,"check_inchan_info() called from %s, %d: %s",file,line,
			"struct inchan_info *a is NULL, could not compare");
	fail_if(NULL==a,errstr);

	snprintf(errstr,200,"check_inchan_info() called from %s, %d: %s",file,line,
			"struct inchan_info *b is NULL, could not compare");
	fail_if(NULL==b,errstr);

	/* check all our other variables
	 */
	snprintf(errstr,200,"check_inchan_info() called from %s, %d: %s",file,line,
			"Names of a and b don't match");
	fail_unless(0==strncmp(a->name,b->name,strlen(a->name)),errstr);

	snprintf(errstr,200,"check_inchan_info() called from %s, %d: %s",file,line,
			"Types of a and b don't match");
	fail_unless(a->type==b->type,errstr);

	snprintf(errstr,200,"check_inchan_info() called from %s, %d: %s",file,line,
			"inAddr statements of a and b don't match");
	check_netaddr_t(a->inAddr,b->inAddr,errstr);

	snprintf(errstr,200,"check_inchan_info() called from %s, %d: %s",file,line,
			"inPath statements of a and b don't match");
	fail_if( (a->inPath==NULL) && (b->inPath!=a->inPath),errstr);
	if(a->inPath != b->inPath)
		fail_unless(0==strncmp(a->inPath,b->inPath,strlen(a->inPath)+1),
				errstr);

	snprintf(errstr,200,"check_inchan_info() called from %s, %d: %s",file,line,
			"inPort statements don't match");
	fail_unless((a->inPort)==(b->inPort),errstr);

	free(errstr);
}

void
check_outchan_info_internal(struct outchan_info *a,struct outchan_info *b,int line,char *file)
{
	int i;
	char *errstr=emalloc(sizeof(char)*200);

	if( (NULL==a) && (NULL==b) )
		return;

	/* check to make sure that one isn't null; this is cause for not continuing with the
	 * function, or we'll segfault */
	snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s",file,line,
			"struct outchan_info *a is NULL, could not compare");
	fail_if(NULL==a,errstr);

	snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s",file,line,
			"struct outchan_info *b is NULL, could not compare");
	fail_if(NULL==b,errstr);

	/* check all our other variables
	 */
	snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s",file,line,
			"Names of a and b don't match");
	fail_unless(0==strncmp(a->name,b->name,strlen(a->name)),errstr);

	snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s",file,line,
			"Types of a and b don't match");
	fail_unless(a->type==b->type,errstr);

	snprintf(errstr,200,"check_outchan_info() called from %s, %d: "
			"Format type of a and b don't match (%d != %d)", 
            file, line, a->opts->format.type, b->opts->format.type);
	fail_unless(a->opts->format.type==b->opts->format.type,errstr);
	snprintf(errstr,200,"check_outchan_info() called from %s, %d: "
			"Format options of a and b don't match (0x%x != 0x%x)", 
            file, line, a->opts->format.options, b->opts->format.options);
    fail_unless(a->opts->format.options == b->opts->format.options,errstr);

	snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s",file,line,
			"Output addr statements of a and b don't match");
	check_netaddr_t(a->opts->outAddr,b->opts->outAddr,errstr);

	snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s",file,line,
			"Output path statements of a and b don't match");
	fail_if( (a->opts->outPath==NULL) && (a->opts->outPath!=b->opts->outPath),errstr);
	if( a->opts->outPath != b->opts->outPath)
		fail_unless(0==strncmp(a->opts->outPath,b->opts->outPath,strlen(a->opts->outPath)),errstr);

	snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s",file,line,
			"Output port statements of a and b don't match");
	fail_unless(a->opts->outPort==b->opts->outPort,errstr);

	for(i=0;i<256;i++)
	{
		if(a->pri[i]!=b->pri[i])
		{
			snprintf(errstr,200,"check_outchan_info() called from %s, %d: %s%d%s",file,line,
					"Priority ",i," of a and b didn't match");
			fail(errstr);
		}
	}

	free(errstr);
}

void
check_netaddr_t(netaddr_t a,netaddr_t b,char *errstr)
{
	/* if they're both NULL, we're OK */
	if( (NULL==a) && (NULL==b) )
		return;

	/* if one of them is NULL, we've got a problem */
	if(b==NULL)
	{
		fail(errstr);
		return;			/* needed in check_nofork */
	}
	if(a==NULL)
	{
		fail(errstr);
		return;			/* needed in check_nofork */
	}


	if( (a->type)!=(b->type) )
		fail(errstr);

	if(NAT_IP4==a->type)
	{
		if(a->addr.ip4!=b->addr.ip4)
			fail(errstr);
	}
	else if(NAT_IP6==a->type)
	{
		int i;

		for(i=0;i<4;i++)
		{
			if(a->addr.ip6[i]!=b->addr.ip6[i])
				fail(errstr);
		}
	}
	else if(NAT_UNIX==a->type)
	{
		if(0!=strncmp(a->addr.sun_path,b->addr.sun_path,
					strlen(a->addr.sun_path)))
			fail(errstr);
	}
	else
	{
		fail(errstr);
	}
}

struct inchan_info *
create_inchan_info(char *name,IMOD_TYPE type, netaddr_t inAddr, char *inPath,
		int listen_port,ACL *path_acl,ACL *psec_acl,ACL *hsrc_acl,ACL *hcon_acl)
{
	struct inchan_info *ret;

	ret=CONFOBJ_createInput();

	if(NULL==name)
		ret->name=NULL;
	else
	{
		ret->name=emalloc(strlen(name)+1);
		strlcpy(ret->name,name,strlen(name)+1);
	}

	ret->type=type;

	if(NULL==inAddr)
		ret->inAddr=NULL;
	else
	{
		ret->inAddr=emalloc(sizeof(*inAddr));
		memcpy(ret->inAddr,inAddr,sizeof(*inAddr));
	}

	if(NULL==inPath)
		ret->inPath=NULL;
	else
	{
		ret->inPath=emalloc( (strlen(inPath)+1)*sizeof(char));
		strlcpy(ret->inPath,inPath,strlen(inPath)+1);
	}

	ret->inPort=listen_port;

	ret->path_acl=path_acl;
	ret->psec_acl=psec_acl;
	ret->hsrc_acl=hsrc_acl;
	ret->hcon_acl=hcon_acl;

	return ret;
}

struct outchan_info *
create_outchan_info(OMOD_TYPE type,MFRMT_TYPE formattype, uint32_t fopts,
		char *name,char *outPath,netaddr_t outAddr,int outPort,uint8_t * priority)
{
	int i;
	struct outchan_info *ret;

	ret=CONFOBJ_createOutput();

	ret->type=type;
	ret->opts->format.type=formattype;
    ret->opts->format.options = fopts;

	if(name==NULL)
		ret->name=NULL;
	else
	{
		ret->name=emalloc(strlen(name)+1);
		strlcpy(ret->name,name,strlen(name)+1);
	}

	if(outAddr==NULL)
		ret->opts->outAddr=NULL;
	else
	{
		ret->opts->outAddr=emalloc(sizeof(*outAddr));
		memcpy(ret->opts->outAddr,outAddr,sizeof(*outAddr));
	}

	if(outPath==NULL)
		ret->opts->outPath=NULL;
	else
	{
		ret->opts->outPath=emalloc(strlen(outPath)+1);
		memcpy(ret->opts->outPath,outPath,strlen(outPath)+1);
	}

	ret->opts->outPort=outPort;

	if(priority==NULL)
		memset(ret->pri,0,sizeof(ret->pri));
	else
	{
		for(i=0;i<256;i++)
			ret->pri[i]=priority[i];
	}

	return ret;
}
