/**
 * @file ut_config_objects.c
 *
 * @brief Define the symbol table tests
 * @ingroup ut_config
 *
 * @todo Need to check ACL, connects stuff
 *
 * $Id: ut_config_objects.c,v 1.12.2.6 2003/04/30 20:03:46 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Apr 30 14:16:33 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "error.h"
#include "check.h"
#include "strcompat.h"
#include "chaninfo.h"

#include "config/config.h"
#include "config/conf_private.h"
#include "config/conf_obj.h"

/**@name Config File Object Tests 
 * @note Objects of type config_info are tested in the Symbol Table tests
 * @ingroup ut_config
 * @{
 */

/** Test creation of outchan_info */
static void TESTconfig_object_output(void);

/** Test creation of inchan_info */
static void TESTconfig_object_input(void);

/** Test creation of CONF_acl */
static void TESTconfig_object_acl(void);

/** Test creation of CONF_con */
static void TESTconfig_object_con(void);

/*@}*/

/* test suite setup */
void config_object_suite(Suite *suite)
{
	TCase *tc_io;

	tc_io=tcase_create("Config_Obj");
	suite_add_tcase(suite,tc_io);

	tcase_add_test(tc_io,TESTconfig_object_output);
	tcase_add_test(tc_io,TESTconfig_object_input);
	tcase_add_test(tc_io,TESTconfig_object_acl);
	tcase_add_test(tc_io,TESTconfig_object_con);
}

START_TEST(TESTconfig_object_output)
{
	struct outchan_info *out=NULL;

	out=CONFOBJ_createOutput();

	fail_if(out==NULL,"conf_sdsc_malloc_outchan_info() outurned NULL");
	fail_unless(out->name==NULL,"conf_sdsc_malloc_outchan_info() didn't init name correctly");
	fail_unless(out->opts->outPath==NULL,"conf_sdsc_malloc_outchan_info() didn't init outPath correctly");
	fail_unless(out->opts->outAddr==NULL,"conf_sdsc_malloc_outchan_info() didn't init outAddr correctly");
	fail_unless(out->type==-1,"conf_sdsc_malloc_outchan_info() didn't init type correctly");
	fail_unless(out->opts->format.type==MFRMT_NOEXIST,
        "conf_sdsc_malloc_outchan_info() didn't init format.type correctly");
    fail_unless(out->opts->format.options == 0,
        "conf_sdsc_mailloc_outchan_info didn't init format.options correctly");
	fail_unless(out->opts->outPort==-1,"conf_sdsc_malloc_outchan_info() didn't init outPort correctly");

	out->name=emalloc(30);

	CONFOBJ_destroyOutput(&out);
	fail_if(out!=NULL,"conf_sdsc_destroy_outchan_info() didn't destroy parameter");
}
END_TEST

START_TEST(TESTconfig_object_input)
{
	struct inchan_info *in=NULL;

	in=CONFOBJ_createInput();

	fail_if(in==NULL,"conf_sdsc_malloc_inchan_info() inurned NULL");
	fail_unless(in->name==NULL,"conf_sdsc_malloc_inchan_info() didn't init name correctly");
	fail_unless(in->inAddr==NULL,"conf_sdsc_malloc_inchan_info() didn't init inAddr correctly");
	fail_unless(in->inPath==NULL,"conf_sdsc_malloc_inchan_info() didn't init inPath correctly");
	fail_unless(in->type==-1,"conf_sdsc_malloc_inchan_info() didn't init type correctly");
	fail_unless(in->path_acl==NULL,"conf_sdsc_malloc_inchan_info() didn't init path_acl correctly");
	fail_unless(in->psec_acl==NULL,"conf_sdsc_malloc_inchan_info() didn't init psec_acl correctly");
	fail_unless(in->hsrc_acl==NULL,"conf_sdsc_malloc_inchan_info() didn't init hsrc_acl correctly");
	fail_unless(in->hcon_acl==NULL,"conf_sdsc_malloc_inchan_info() didn't init hcon_acl correctly");

	in->name=emalloc(30);

	CONFOBJ_destroyInput(&in);
	fail_unless(in==NULL,"conf_sdsc_destroy_inchan_info() didn't destroy the object");
}
END_TEST

START_TEST(TESTconfig_object_acl)
{
}
END_TEST

START_TEST(TESTconfig_object_con)
{
	struct CONF_con *con=NULL;
	struct inchan_info *in=NULL;

	/* create an input object */
	in=CONFOBJ_createInput();
	in->name=emalloc(30);
	strlcpy(in->name,"Hello",30);

	/* create a CONF_con, and point it there */
	con=CONFOBJ_createCon();
	fail_if(NULL==con,"conf_sdsc_malloc_CONF_con() returned NULL");
	fail_unless(NULL==con->input,"conf_sdsc_malloc_CONF_con() failed to init input");
	fail_unless(NULL==con->output,"conf_sdsc_malloc_CONF_con() failed to init output");

	con->input=in->name;
	fail_unless(0==strncmp(con->input,in->name,strlen(in->name)),"CONF_con input field broken");

	CONFOBJ_destroyCon(&con);
	fail_unless(NULL==con,"conf_sdsc_destroy_CONF_con() failed to destroy an object");
	fail_unless(0==strncmp(in->name,"Hello",strlen("Hello")),"conf_sdsc_destroy_CONF_con() destroyed too much");

	CONFOBJ_destroyInput(&in);
}
END_TEST

