/**
 * @file ut_config_read.c
 *
 * @brief Define the config read tests
 * @ingroup ut_config
 *
 * @todo Test multiple calls to getConfigInfo() in one process.
 *
 * $Id: ut_config_read.c,v 1.30.2.7 2003/10/22 04:09:04 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Fri Apr 26 15:35:51 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "error.h"
#include "check.h"
#include "util/list.h"
#include "chaninfo.h"
#include "util/bitvec.h"

#include "message/message.h"
#include "message/msg_parse_rfc3164.h"
#include "message/msg_frmt.h"
#include "config/config.h"
#include "config/conf_private.h"
#include "config/conf_obj.h"
#include "outputmod/outputmod.h"

#include "ut_config.h"

/**@name Config File Reading Test Cases
 * @ingroup ut_config
 * @{
 */

/** Test config file reading pass/fail */
static void TESTconfig_read(void);

/** Test getConfigInfo() */
static void TESTconfig_multiple(void);

/** Test getConfigInfo(), called once */
static void TESTconfig_info(void);

/** Test the defaults created by the parsing algorithm */
static void TESTconfig_default(void);

/** Test the addressing */
static void TESTconfig_address(void);

/*@}*/

/* test suite setup */
void config_read_suite(Suite *suite)
{
	TCase *tc_io;

	tc_io=tcase_create("Config_Read");
	suite_add_tcase(suite,tc_io);

	tcase_add_test(tc_io,TESTconfig_read);
	/* tcase_add_test(tc_io,TESTconfig_multiple); */
	tcase_add_test(tc_io,TESTconfig_info);
	tcase_add_test(tc_io,TESTconfig_default);
	tcase_add_test(tc_io,TESTconfig_address);
}

#define	TESTconfig_read_pass		"config_tests/TESTconfig_read_pass"
#define	TESTconfig_read_fail		"config_tests/TESTconfig_read_fail"
extern int conf_debug;
START_TEST(TESTconfig_read)
{
	FILE *fp_pass,*fp_fail,*fp_temp;
	char repeat_fname[]="config_tests/TESTconfig_read_fail0";
	int i;

	/* enable for extra debugging output */
	/* conf_debug=1; */

	/* stop stdout */
	close(1);

	fp_pass=fopen(TESTconfig_read_pass,"r");
	fp_fail=fopen(TESTconfig_read_fail,"r");

	conf_prep(fp_pass,TESTconfig_read_pass);
	mark_point();
	fail_if(0!=conf_parse(),"Parsing unexpectedly failed");
	mark_point();

	conf_reset(fp_fail,TESTconfig_read_fail);
	fail_if(0==conf_parse(),"Parsing unexpectedly succeeded");

	/*
	 * TESTconfig_read_fail0: Improper input name
	 * TESTconfig_read_fail1: Invalid input type
	 * TESTconfig_read_fail2: No ending '};'
	 * TESTconfig_read_fail3: Invalid listen name
	 * TESTconfig_read_fail4: Incorrect version number
	 * TESTconfig_read_fail5: Incorrect type name
	 * TESTconfig_read_fail6: Incorrect format type name
	 * TESTconfig_read_fail7: Undefined input type
	 * TESTconfig_read_fail8: Incorrect facility type
	 */
	for(i=0;i<=8;i++)
	{
		repeat_fname[strlen(repeat_fname)-1]=i+'0';
		fp_temp=fopen(repeat_fname,"r");
		conf_reset(fp_temp,repeat_fname);
		fail_if(0==conf_parse(),"Parsing unexpectedly succeeded");
	}
	
	fp_pass=fopen(TESTconfig_read_pass,"r");
	conf_reset(fp_pass,TESTconfig_read_pass);
	fail_if(0!=conf_parse(),"Parsing unexpectedly failed");
	conf_stop();
	CONFSYM_destroy();
}
END_TEST

START_TEST(TESTconfig_multiple)
{
	/* enable for extra debugging output */
	/* conf_debug=1; */
	struct config_info *t;

	t=getConfigInfo(TESTconfig_read_fail);
	fail_if(NULL!=t,"Parsing unexpectedly succeeded");
	if(NULL!=t)
		destroyConfigInfo(&t);

	mark_point();

	t=getConfigInfo(TESTconfig_read_pass);
	fail_if(NULL==t,"Parsing unexpectedly failed");
	destroyConfigInfo(&t);

	mark_point();
}
END_TEST
#undef TESTconfig_read_pass
#undef TESTconfig_read_fail

#define TESTconfig_read_pass_small "config_tests/TESTconfig_read_pass_small"
START_TEST(TESTconfig_info)
{
	struct config_info *t;

	struct inchan_info *F_09,*sys2,*in_temp;
	struct outchan_info *out1,*out_temp;
	struct CONF_con *con_temp;

	LLIterator it;

	struct in_addr addr;
	netaddr_t ip192,ulog;

	/* for the priority */
	uint8_t priarr[256];
	int j;
	memset(priarr,0,sizeof(uint8_t)*256);

    /* kernel,mail,local0,local7.alert */
    priarr[0] = priarr[1] = priarr[16] = priarr[17] = 
        priarr[128] = priarr[129] = priarr[184] = priarr[185] = 1;
    /* '*.=debug' */
	for(j=0;j<CONFFS_FACCOUNT;j++)
		priarr[j*8+7]=1;
    /* cron.error-notice */
    priarr[75] = priarr[76] = priarr[77] = 1;
    /* user.none */
    priarr[15] = 0;
	/* ftp.* */
	for(j=0;j<=CONFFS_DEBUG;j++)
		priarr[11*8+j]=1;

	/* enable for extra debugging output */
	/* conf_debug=1; */

	/* stop stdout */
	close(1);

	/* create the netaddr_t */
	inet_pton(AF_INET,"192.168.12.1",&addr);
	ip192=NA_newINet4(addr);
	ulog=NA_newUnix("/dev/log");

	/* init the different structs, for comparison */
	F_09=create_inchan_info("F_09_as1_232DSF", IMOD_BEEP, ip192, "192.168.12.1",514, NULL, NULL, NULL, NULL);
	sys2=create_inchan_info("sys2", IMOD_SYSTEM, ulog, "/dev/log", 999, NULL, NULL, NULL, NULL);
	out1=create_outchan_info(OMOD_UDP, MFRMT_RFC3164, BM_MBIT(MFOPT_PRI), "out1", "192.168.12.1",
			ip192, 12, priarr);

	/* destroy the netaddr_t's we just made */
	NA_destroy(ip192);
	NA_destroy(ulog);

	t=getConfigInfo(TESTconfig_read_pass_small);
	fail_if(NULL==t,"Parsing unexpectedly failed");

	/* we need to get everything out of the config_info object, and check them */
	if(t->input_count>0)
	{
		it=LL_iterator(t->inputs);
		while(NULL!=LLI_next(it))
		{
			in_temp=LLI_elem(it);

			if(0==strncmp(in_temp->name,F_09->name,strlen(F_09->name)+1))
				check_inchan_info(in_temp,F_09);
			else if(0==strncmp(in_temp->name,sys2->name,strlen(sys2->name)+1))
				check_inchan_info(in_temp,sys2);
			else
				fail("Unrecognized input type in file; please add to test");
		}
		LLI_destroy(&it);
	}
	else
		fail("Unexpected number of inputs");

	if(t->output_count>0)
	{
		it=LL_iterator(t->outputs);
		while(NULL!=LLI_next(it))
		{
			out_temp=LLI_elem(it);

			if(0==strncmp(out_temp->name,out1->name,strlen(out1->name)+1))
				check_outchan_info(out_temp,out1);
			else
				fail("Unrecognized output type in file; please add to test");
		}
		LLI_destroy(&it);
	}
	else
		fail("Unexpected number of outputs");

	/* make sure the connection is good, too */
	if(t->connect_count>0)
	{
		it=LL_iterator(t->connects);
		while(NULL!=LLI_next(it))
		{
			con_temp=LLI_elem(it);

			/* since we know we only have one connection, we'll compare it
			 * manually, even though it's in a while loop.
			 */
			fail_unless(0==strncmp(con_temp->input,sys2->name,strlen(sys2->name)+1),
					"Input field of connection has unexpected value");
			fail_unless(0==strncmp(con_temp->output,out1->name,strlen(out1->name)+1),
					"Input field of connection has unexpected value");
		}
		LLI_destroy(&it);
	}
	else
		fail("Unexpected number of connections");

	/* check the user and group */
	fail_if(NULL==t->global,"Globals not correctly read");
	fail_if(NULL==t->global->user,"Globals not correctly read");
	fail_if(NULL==t->global->group,"Globals not correctly read");
	fail_unless(NULL==t->global->stdredir,"Globals not correctly read");

	fail_unless(0==strncmp("small",t->global->user,6),"Incorrect user name");
	fail_unless(0==strncmp("users",t->global->group,6),"Incorrect group name");

	CONFOBJ_destroyInput(&F_09);
	fail_unless(NULL==F_09,"CONFOBJ_destroyInput() failed to set input to NULL");
	CONFOBJ_destroyInput(&sys2);
	fail_unless(NULL==sys2,"CONFOBJ_destroyInput() failed to set input to NULL");
	CONFOBJ_destroyOutput(&out1);
	fail_unless(NULL==out1,"CONFOBJ_destroyInput() failed to set input to NULL");
}
END_TEST
#undef TESTconfig_read_pass_small

#define TESTconfig_read_pass_default "config_tests/TESTconfig_read_pass_default"
START_TEST(TESTconfig_default)
{
	struct config_info *t;

	struct inchan_info *sys,*udp;
	struct outchan_info *tcp;

	struct in_addr tempaddr;

	LLIterator it;

	netaddr_t ulog,slog;

	/* stop stdout */
	close(1);

	/* create the netaddr_t */
	ulog=NA_newUnix("/dev/log");
	tempaddr.s_addr=INADDR_ANY;
	slog=NA_newINet4(tempaddr);

	/* init the different structs, for comparison */
	sys=create_inchan_info("sys", IMOD_SYSTEM, ulog, "/dev/log", -1, 
            NULL, NULL, NULL, NULL );
	udp=create_inchan_info("udp", IMOD_UDP, slog, "0.0.0.0", 
            SYSLOG_UDP_PORT, NULL, NULL, NULL, NULL);
	slog=NA_newINetLookup("192.168.12.1");
	tcp=create_outchan_info(OMOD_BEEP, MFRMT_RFC3164, 
            BM_MBIT(MFOPT_MIMEHEAD) | BM_MBIT(MFOPT_PRI), "tcp", 
            "192.168.12.1", slog, SYSLOG_CONN_PORT, NULL);

	/* destroy the netaddr_t's we just made */
	NA_destroy(ulog);
	NA_destroy(slog);

	t=getConfigInfo(TESTconfig_read_pass_default);
	fail_if(NULL==t,"Parsing unexpectedly failed");

	/* we need to get everything out of the config_info object, and check them */
	if(t->input_count>0)
	{
		struct inchan_info *in_temp;

		it=LL_iterator(t->inputs);
		while(NULL!=LLI_next(it))
		{
			in_temp=LLI_elem(it);

			if(0==strncmp(in_temp->name,sys->name,strlen(sys->name)+1))
				check_inchan_info(in_temp,sys);
			else if(0==strncmp(in_temp->name,udp->name,strlen(udp->name)+1))
				check_inchan_info(in_temp,udp);
			else
				fail("Unrecognized input type in file; please add to test");
		}
		LLI_destroy(&it);
	}
	else
		fail("Unexpected number of inputs");

	if(t->output_count>0)
	{
		struct outchan_info *out_temp;

		it=LL_iterator(t->outputs);
		while(NULL!=LLI_next(it))
		{
			out_temp=LLI_elem(it);

			if(0==strncmp(out_temp->name,tcp->name,strlen(tcp->name)+1))
				check_outchan_info(out_temp,tcp);
			else
				fail("Unrecognized output type in file; please add to test");
		}
		LLI_destroy(&it);
	}
	else
		fail("Unexpected number of outputs");

	/* check the user and group */

	CONFOBJ_destroyInput(&sys);
	fail_unless(NULL==sys,"CONFOBJ_destroyInput() failed to set input to NULL");
	CONFOBJ_destroyInput(&udp);
	fail_unless(NULL==udp,"CONFOBJ_destroyInput() failed to set input to NULL");
}
END_TEST
#undef TESTconfig_read_pass_default

#define TESTconfig_read_pass_addr "config_tests/TESTconfig_read_pass_addr"
START_TEST(TESTconfig_address)
{
	struct config_info *t;

	struct outchan_info *out_temp;

	/* stop stdout */
	close(1);

	t=getConfigInfo(TESTconfig_read_pass_addr);
	fail_if(NULL==t,"Parsing unexpectedly failed");

	/* we need to get everything out of the config_info object, and check them */
	if(t->output_count>0)
	{
		LLIterator it=LL_iterator(t->outputs);
		while(NULL!=LLI_next(it))
		{
			out_temp=LLI_elem(it);

			if(0==strncmp(out_temp->name,"out1",strlen("out1")+1))
				fail_unless(out_temp->opts->outAddr->type==NAT_IP4,
						"out1 failed to have an ip4 address");
			else if(0==strncmp(out_temp->name,"out2",strlen("out2")+1))
				fail_unless(out_temp->opts->outAddr->type==NAT_IP4,
						"out2 failed to have an ip4 address");
			else
				fail("Unrecognized input type in file; please add to test");
		}
		LLI_destroy(&it);
	}
	else
		fail("Unexpected number of inputs");
}
END_TEST
#undef TESTconfig_read_pass_addr
