/**
 * @file ut_config_symtab.c
 *
 * @brief Define the symbol table tests
 * @ingroup ut_config
 *
 * @todo Need to check ACL, connects stuff
 *
 * $Id: ut_config_symtab.c,v 1.18.2.6 2003/04/30 20:03:47 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Fri Apr 26 15:35:51 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "strcompat.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "error.h"
#include "check.h"
#include "chaninfo.h"

#include "config/config.h"
#include "config/conf_private.h"
#include "config/conf_obj.h"

#include "ut_config.h"

/**@name Config File Symbol Table
 * @ingroup ut_config
 * @{
 */

/** Test adding elements to the symbol table */
static void TESTconfig_symtab_add(void);

/** Test retrieval of elements by name */
static void TESTconfig_symtab_get(void);

/** Test conversion of the symbol table to a struct config_info */
static void TESTconfig_symtab_info(void);

/** Test attemped double calls of ...get_info() */
static void TESTconfig_symtab_double(void);

/** Initialize and check the symbol table */
static void init_check(void);

/** Destroy and check the symbol table */
static void destroy_check(void);

/*@}*/

/* test suite setup */
void config_symtab_suite(Suite *suite)
{
	TCase *tc_io;

	tc_io=tcase_create("Config_SymTab");
	suite_add_tcase(suite,tc_io);

	tcase_add_test(tc_io,TESTconfig_symtab_add);
	tcase_add_test(tc_io,TESTconfig_symtab_get);
	tcase_add_test(tc_io,TESTconfig_symtab_info);
	tcase_add_test(tc_io,TESTconfig_symtab_double);
}

/* externs for our symbol table */
extern struct CONFSYM_str *CONFSYM_table;

START_TEST(TESTconfig_symtab_add)
{
	struct inchan_info *in;
	struct outchan_info *out;
	struct CONF_con *con;

	struct inchan_info *in_returned;
	struct outchan_info *out_returned;
	struct CONF_con *con_returned;

	struct in_addr addr;
	netaddr_t ip192,unull;

	/* for the priority */
	uint8_t priarr[256];
	memset(priarr,0,sizeof(uint8_t)*256);
	priarr[0]=priarr[1]=priarr[16]=priarr[17]=priarr[128]=priarr[129]=priarr[185]=priarr[184]=1;

	/* create the netaddr_t */
	inet_pton(AF_INET,"192.168.12.1",&addr);
	ip192=NA_newINet4(addr);
	unull=NA_newUnix("/dev/null");

	in=create_inchan_info("type in", IMOD_BEEP, ip192, "192.168.12.1", 514,  NULL, NULL, NULL, NULL);
	out=create_outchan_info(OMOD_FILE, MFRMT_RFC3164, 0, "type out", 
            "/dev/null", unull, -1, priarr);
	con=CONFOBJ_createCon();
		con->input=in->name;
		con->output=out->name;

	/* destroy the netaddr_t's we just made */
	NA_destroy(ip192);
	NA_destroy(unull);

	init_check();

	CONFSYM_addInput(in);
	fail_unless(CONFSYM_table->input_count==1,"Input count not 1 after added an input");

	CONFSYM_addOutput(out);
	fail_unless(CONFSYM_table->output_count==1,"Output count not 1 after added an output");

	CONFSYM_addConnect(con);
	fail_unless(CONFSYM_table->connect_count==1,"Connection count not 1 after added a connect");

	in_returned=LL_first(CONFSYM_table->input_l);
	fail_if(NULL==in_returned,"Tried to get an input from the list, got NULL instead");
	check_inchan_info(in_returned,in);

	out_returned=LL_first(CONFSYM_table->output_l);
	fail_if(NULL==out_returned,"Tried to get an output from the list, got NULL instead");
	check_outchan_info(out_returned,out);

	con_returned=LL_first(CONFSYM_table->connect_l);
	fail_if(NULL==con_returned,"Tried to get a connection from the list, got NULL instead");
	fail_unless(con_returned->input==in->name,"Retrieved connect input pointer incorrect");
	fail_unless(con_returned->input==in_returned->name,"Retrieved connect input pointer incorrect");
	fail_unless(con_returned->output==out->name,"Retrieved connect output pointer incorrect");
	fail_unless(con_returned->output==out_returned->name,"Retrieved connect output pointer incorrect");

	destroy_check();
}
END_TEST

START_TEST(TESTconfig_symtab_get)
{
	struct inchan_info *in;
	struct inchan_info *in2;
	struct outchan_info *out;
	struct outchan_info *out2;
	struct CONF_con *con;
	struct CONF_con *con2;

	struct inchan_info *in_returned;
	struct outchan_info *out_returned;
	struct CONF_con *con_returned;

	struct in_addr addr;
	netaddr_t ip1,ip14,unull,uexists;

	/* for the priority */
	uint8_t priarr0[256];
	uint8_t priarr1[256];
	memset(priarr0,0,sizeof(uint8_t)*256);
	memset(priarr1,0,sizeof(uint8_t)*256);
	priarr0[0]=priarr0[1]=priarr0[16]=priarr0[17]=priarr0[128]=priarr0[129]=priarr0[185]=priarr0[184]=1;
	priarr1[2]=priarr1[5]=priarr1[17]=priarr1[27]=priarr1[158]=priarr1[139]=priarr1[186]=priarr1[256]=1;

	/* create the netaddr_t */
	inet_pton(AF_INET,"192.168.12.1",&addr);
	ip1=NA_newINet4(addr);
	inet_pton(AF_INET,"192.168.12.14",&addr);
	ip14=NA_newINet4(addr);
	unull=NA_newUnix("/dev/null");
	uexists=NA_newUnix("/dev/exists");

	/* initialize our test inputs */
	in=create_inchan_info("type in", IMOD_BEEP, ip1, "192.168.12.1", 514, NULL, NULL, NULL, NULL );
	in2=create_inchan_info("type2 in", IMOD_UDP, ip14, "192.168.12.14", 990, NULL, NULL, NULL, NULL);
	out=create_outchan_info(OMOD_FILE, MFRMT_RFC3164, 0, "type out", 
            "/dev/null", unull, -1, priarr0);
	out2=create_outchan_info(OMOD_FILE, MFRMT_RFC3164, 0, "type2 out", 
            "/dev/exists", uexists, -1, priarr1);
	con=CONFOBJ_createCon();
		con->input=in->name;
		con->output=out2->name;
	con2=CONFOBJ_createCon();
		con2->input=in2->name;
		con2->output=out->name;

	/* destroy the NA's */
	NA_destroy(ip1);
	NA_destroy(ip14);
	NA_destroy(unull);
	NA_destroy(uexists);

	init_check();

	/* add in the inputs */
	CONFSYM_addInput(in);
	fail_unless(CONFSYM_table->input_count==1,"Input count not 1 after added an input");
	CONFSYM_addInput(in2);
	fail_unless(CONFSYM_table->input_count==2,"Input count not 2 after added an input");

	/* add in the outputs */
	CONFSYM_addOutput(out);
	fail_unless(CONFSYM_table->output_count==1,"Output count not 1 after added an output");
	CONFSYM_addOutput(out2);
	fail_unless(CONFSYM_table->output_count==2,"Output count not 2 after added an output");

	/* add in the connects */
	CONFSYM_addConnect(con);
	fail_unless(CONFSYM_table->connect_count==1,"Connect count not 1 after added a connect");
	CONFSYM_addConnect(con2);
	fail_unless(CONFSYM_table->connect_count==2,"Connect count not 2 after added a connect");

	/* check to make sure we can get inputs out */
	in_returned=CONFSYM_getInput("type in");
	fail_if(NULL==in_returned,"conf_symtab_get_input() unexpectedly returned NULL");
	check_inchan_info(in,in_returned);

	in_returned=CONFSYM_getInput("type2 in");
	fail_if(NULL==in_returned,"conf_symtab_get_input() unexpectedly returned NULL");
	check_inchan_info(in2,in_returned);

	in_returned=CONFSYM_getInput("not real");
	fail_unless(NULL==in_returned,"conf_symtab_get_input() failed to return NULL");

	/* and, can we get outputs out */
	out_returned=CONFSYM_getOutput("type out");
	fail_if(NULL==out_returned,"conf_symtab_get_output() unexpectedly returned NULL");
	check_outchan_info(out,out_returned);

	out_returned=CONFSYM_getOutput("type2 out");
	fail_if(NULL==out_returned,"conf_symtab_get_output() unexpectedly returned NULL");
	check_outchan_info(out2,out_returned);

	out_returned=CONFSYM_getOutput("doesn't exist");
	fail_unless(NULL==out_returned,"conf_symtab_get_input() failed to return NULL");

	/* finally, can we get connects out */
	con_returned=CONFSYM_getConnect(in->name,NULL);
	fail_if(con_returned==NULL,"conf_symtab_get_connect() returned NULL");
	fail_unless(0==strncmp(con_returned->input,in->name,strlen(in->name)),
			"Symbol table didn't return the proper connection");
	fail_unless(0==strncmp(con_returned->output,out2->name,strlen(out->name)),
			"Symbol table didn't return the proper connection");

	con_returned=CONFSYM_getConnect(NULL,out2->name);
	fail_if(con_returned==NULL,"conf_symtab_get_connect() returned NULL");
	fail_unless(0==strncmp(con_returned->input,in->name,strlen(in->name)),
			"Symbol table didn't return the proper connection");
	fail_unless(0==strncmp(con_returned->output,out2->name,strlen(out->name)),
			"Symbol table didn't return the proper connection");

	con_returned=CONFSYM_getConnect(NULL,NULL);
	fail_unless(con_returned==NULL,"conf_symtab_get_connect() didn't return NULL");

	con_returned=CONFSYM_getConnect(in->name,out->name);
	fail_unless(con_returned==NULL,"conf_symtab_get_connect() didn't return NULL");

	con_returned=CONFSYM_getConnect(in2->name,NULL);
	fail_if(con_returned==NULL,"conf_symtab_get_connect() returned NULL");
	fail_unless(0==strncmp(con_returned->output,out->name,strlen(out->name)),
			"Symbol table didn't return the proper connection");
	fail_unless(0==strncmp(con_returned->input,in2->name,strlen(out->name)),
			"Symbol table didn't return the proper connection");

	con_returned=CONFSYM_getConnect(NULL,out->name);
	fail_if(con_returned==NULL,"conf_symtab_get_connect() returned NULL");
	fail_unless(0==strncmp(con_returned->output,out->name,strlen(out->name)),
			"Symbol table didn't return the proper connection");
	fail_unless(0==strncmp(con_returned->input,in2->name,strlen(out->name)),
			"Symbol table didn't return the proper connection");

	destroy_check();
}
END_TEST

START_TEST(TESTconfig_symtab_info)
{
	struct inchan_info *in;
	struct inchan_info *in2;
	struct outchan_info *out;
	struct outchan_info *out2;
	struct CONF_con *con;
	struct CONF_con *con2;

	struct inchan_info *in_returned;
	struct outchan_info *out_returned;
	struct CONF_con *con_returned;

	struct config_info *inf_ret;

	LLIterator it;

	struct in_addr addr;
	netaddr_t ip1,ip14,unull,uexists;

	/* for the priority */
	uint8_t priarr0[256];
	uint8_t priarr1[256];
	memset(priarr0,0,sizeof(uint8_t)*256);
	memset(priarr1,0,sizeof(uint8_t)*256);
	priarr0[0]=priarr0[1]=priarr0[16]=priarr0[17]=priarr0[128]=priarr0[129]=priarr0[185]=priarr0[184]=1;
	priarr1[2]=priarr1[5]=priarr1[17]=priarr1[27]=priarr1[158]=priarr1[139]=priarr1[186]=priarr1[256]=1;

	/* create the netaddr_t */
	inet_pton(AF_INET,"192.168.12.1",&addr);
	ip1=NA_newINet4(addr);
	inet_pton(AF_INET,"192.168.12.14",&addr);
	ip14=NA_newINet4(addr);
	unull=NA_newUnix("/dev/null");
	uexists=NA_newUnix("/dev/exists");

	/* finish constructing our test struct CONF_*'s */
	in=create_inchan_info("type in", IMOD_BEEP, ip1, "192.168.12.1", 514, NULL, NULL, NULL, NULL);
	in2=create_inchan_info("type2 in", IMOD_UDP, ip14, "192.168.12.14", 514, NULL, NULL, NULL, NULL);
	out=create_outchan_info(OMOD_FILE, MFRMT_RFC3164,0, "type out", "/dev/nulL",
			unull, -1, priarr0);
	out2=create_outchan_info(OMOD_FILE, MFRMT_RFC3164,0, "type2 out", 
            "/dev/exists", uexists, -1, priarr1);
	con=CONFOBJ_createCon();
		con->input=in->name;
		con->output=out2->name;
	con2=CONFOBJ_createCon();
		con2->input=in2->name;
		con2->output=out->name;

	/* destroy the NA's */
	NA_destroy(ip1);
	NA_destroy(ip14);
	NA_destroy(unull);
	NA_destroy(uexists);

	init_check();

	/* add in the inputs */
	CONFSYM_addInput(in);
	fail_unless(CONFSYM_table->input_count==1,"Input count not 1 after added an input");
	CONFSYM_addInput(in2);
	fail_unless(CONFSYM_table->input_count==2,"Input count not 2 after added an input");

	/* add in the outputs */
	CONFSYM_addOutput(out);
	fail_unless(CONFSYM_table->output_count==1,"Output count not 1 after added an output");
	CONFSYM_addOutput(out2);
	fail_unless(CONFSYM_table->output_count==2,"Output count not 2 after added an output");
	
	/* add in the connects */
	CONFSYM_addConnect(con);
	fail_unless(CONFSYM_table->connect_count==1,"Connect count not 1 after added a connect");
	CONFSYM_addConnect(con2);
	fail_unless(CONFSYM_table->connect_count==2,"Connect count not 2 after added a connect");

	/* see if we can get a valid config_info structure out */
	inf_ret=CONFSYM_getInfo();

	fail_unless(inf_ret->input_count==2,"struct config_info doesn't contain the proper number of inputs");
	fail_unless(inf_ret->output_count==2,"struct config_info doesn't contain the proper number of outputs");

	/* check the inputs */
	it=LL_iterator(inf_ret->inputs);
	while(NULL!=LLI_next(it))
	{
		in_returned=LLI_elem(it);

		if(0==strncmp(in_returned->name,in->name,strlen(in->name)+1))
			check_inchan_info(in_returned,in);
		else check_inchan_info(in_returned,in2);
	}
	LLI_destroy(&it);

	/* check the outputs */
	it=LL_iterator(inf_ret->outputs);
	while(NULL!=LLI_next(it))
	{
		out_returned=LLI_elem(it);

		if(0==strncmp(out_returned->name,out->name,strlen(out->name)+1))
			check_outchan_info(out_returned,out);
		else check_outchan_info(out_returned,out2);
	}
	LLI_destroy(&it);

	/* check the connects */
	it=LL_iterator(inf_ret->connects);
	while(NULL!=LLI_next(it))
	{
		con_returned=LLI_elem(it);

		if(0==strncmp(con_returned->input,in->name,strlen(in->name)+1))
		{
			fail_unless(0==strncmp(con_returned->input,in->name,strlen(in->name)),
					"Symbol table didn't return the proper connection");
			fail_unless(0==strncmp(con_returned->output,out2->name,strlen(out->name)),
					"Symbol table didn't return the proper connection");
		}
		else
		{
			fail_unless(0==strncmp(con_returned->output,out->name,strlen(out->name)),
					"Symbol table didn't return the proper connection");
			fail_unless(0==strncmp(con_returned->input,in2->name,strlen(out->name)),
					"Symbol table didn't return the proper connection");
		}
	}
	LLI_destroy(&it);

	CONFOBJ_destroyInfo(&inf_ret);
	destroy_check();
}
END_TEST

START_TEST(TESTconfig_symtab_double)
{
	CONFSYM_table=NULL;

	fail_if(NULL!=CONFSYM_getInfo(),"CONFSYM_getInfo() on NULL symbol table failed incorrectly");
}
END_TEST

static void
init_check(void)
{
	CONFSYM_init();

	/* check our initialization */
	fail_if(CONFSYM_table->acl_l==NULL,"ACL list init failed");
	fail_if(CONFSYM_table->input_l==NULL,"Input list init failed");
	fail_if(CONFSYM_table->output_l==NULL,"Output list init failed");
	fail_if(CONFSYM_table->connect_l==NULL,"Connect list init failed");

	fail_unless(CONFSYM_table->acl_count==0,"ACL count not 0 at init");
	fail_unless(CONFSYM_table->input_count==0,"Input count not 0 at init");
	fail_unless(CONFSYM_table->output_count==0,"Output count not 0 at init");
	fail_unless(CONFSYM_table->connect_count==0,"Connect count not 0 at init");
}

static void
destroy_check(void)
{
	CONFSYM_destroy();

	fail_unless(NULL==CONFSYM_table,"Symtab failed to be destroyed");
}

