/**
 * @file ut_imod_beep.c
 *
 * @brief BEEP Input Module Tests
 * @par
 *
 * $Id: ut_imod_beep.c,v 1.1.2.8 2003/10/22 04:09:04 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Aug 29 17:16:39 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/types.h>
#include <signal.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "check.h"

#include "error.h"
#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/inputwait.h"
#include "outputmod/outputmod.h"
#include "strcompat.h"
#include "ut_imod.h"
#include "debug.h"

#include "common_funcs.h"

/**@defgroup ut_imod_beep BEEP Input Module Test Cases
 * @ingroup ut_omod
 * @{
 */

/** Receive messages; send them to the IPC layer */
static void TESTimod_beep_module(void);

/** Send test messages to the server */
static void TESTimod_beep_send(void);

/** Receive IPC messages from the module */
static void TESTimod_beep_controller(void);

/*@}*/

/**@defgroup ut_imod_beep_impl BEEP Input Module Test Cases Implementation
 * @ingroup ut_imod_beep
 * @{
 */

/** Start debugging */
static void
debug_init(void)
{
	DBG_init("IBm","imod-beep.debug");
}

/** Mark that we're done debugging */
static void
debug_done(void)
{
	DPRINT(DBG_DBG, "Test Case Done", NO_VARG);
}

/** Path to the IPC socket */
#define IPCSOCK "imod-beep.ipcsock"

/** BEEP port */
#define BEEPPORT 5000

/** The number of senders */
#define NUMSENDER	4

/** The number of messages */
#define NUMMSG	4

/** The raw messages to send */
static char sendmsgs[NUMMSG][1024]= {
		"<72>1980-11-11T22:14:15-06:00 192.168.12.12 su[12]: 'su root' failed for small",
		"<17>1980-11-11T22:14:15-06:00 192.168.12.13 su[13]: 'su nobody' failed for small",
		"<26>1980-11-11T22:14:15-06:00 192.168.12.14 su[14]: 'su mysql' failed for small",
		"<35>Nov 11 22:14:15 buckaroo su: 'su root' failed for lonvick",
};

/** The messages we should receive in the controller (already formatted to the
 * internal struct message format
 * 
 * @note This array is indexed by the severity field; this must increment as
 * below
 */
static struct message recvmsgs[NUMMSG]= {
		{ 0, 9, 14, "192.168.12.12", 342857655, -360, "su", 33, "[12]: 'su root' failed for small", },
		{ 1, 2, 14, "192.168.12.13", 342857655, -360, "su", 35, "[13]: 'su nobody' failed for small", },
		{ 2, 3, 14, "192.168.12.14", 342857655, -360, "su", 34, "[14]: 'su mysql' failed for small", },
		{ 3, 4, 9, "buckaroo", -1, -1, "su", 31, ": 'su root' failed for lonvick", },
};

/*@}*/

void
imod_beep_suite(Suite *suite)
{
	int i;

	TCase *tc;

	tc = tcase_create("IMOD_BEEP");
	suite_add_tcase(suite,tc);

	tcase_add_unchecked_fixture(tc,debug_init,debug_done);

	tcase_set_run_mode(tc, CK_PARALLEL);

	tcase_add_test(tc,TESTimod_beep_controller);
	tcase_add_test(tc,TESTimod_beep_module);

	/* run this twice, to send even more */
	for(i=0;i<NUMSENDER;i++)
		tcase_add_test(tc,TESTimod_beep_send);
}

START_TEST(TESTimod_beep_module)
{
	struct IMOD_info *info;
	int sock;
	struct sockaddr_in addr;
	char errstr[1024];

	DBG_newproc("Module");

	DPRINT(DBG_DBG, "Starting module", NO_VARG);

	EXIT_initIM();

	info=emalloc(sizeof(struct IMOD_info));
	memset(info,0,sizeof(struct IMOD_info));
	info->ipc_path=emalloc(sizeof(char)*(1+strlen(IPCSOCK)));
	strlcpy(info->ipc_path,IPCSOCK, strlen(IPCSOCK)+1);

	sleep(1);		/* let the IPC layer init */

	/* set up the listener */
	sock=socket(AF_INET, SOCK_STREAM, 0);
	if(sock<0)
		fail("Failed to create socket");

	memset(&(addr),0,sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_port=htons(BEEPPORT);
	addr.sin_addr.s_addr=htonl(INADDR_ANY);

	if(bind(sock,(struct sockaddr *)&addr,sizeof(addr))<0)
	{
		snprintf(errstr,1024,"Could not bind to socket: port %d", BEEPPORT);
		fail(errstr);
	}

	if(listen(sock, 5)<0)
	{
		snprintf(errstr,1024,"Could not listen(): %s",strerror(errno));
		fail(errstr);
	}

	info->fd_len=1;
	info->fd=emalloc(sizeof(int)*info->fd_len*2);
	info->chans=info->fd+info->fd_len;
	info->fd[0]=sock;
	info->chans[0]=2;

	DPRINT(DBG_DBG, "calling IMOD_BEEP_start()", NO_VARG);

	IMOD_BEEP_start(info, IMOD_BEEP);

	DPRINT( DBG_DBG, "Test case done", NO_VARG);
}
END_TEST

START_TEST(TESTimod_beep_send)
{
	int sock;
	char errstr[1024];
	char *sendmsg;

	BEEP_conn *connection;

	struct MFRMT_opt frmt_opt = { MFRMT_RFC3164, 0};

	int i;

	DBG_newproc("Sender");

	DPRINT(DBG_DBG, "Starting sender", NO_VARG);

	sleep(2);			/* let everything else start */

	/* could be in its own function */
	{
		struct sockaddr_in addr;

		sock=socket(AF_INET,SOCK_STREAM,0);
		if(sock<0)
		{
			snprintf(errstr,1024,"socket() failed: %s",strerror(errno));
			fail(errstr);
		}

		addr.sin_family=AF_INET;
		addr.sin_port=htons(BEEPPORT);
		addr.sin_addr.s_addr=inet_addr("127.0.0.1");

		if(connect(sock,(struct sockaddr *)&addr, sizeof(struct sockaddr_in))<0)
		{
			snprintf(errstr,1024,"connect() failed: %s",strerror(errno));
			fail(errstr);
		}
	}

	DPRINT(DBG_DBG, "Try to start BEEP", NO_VARG);
	connection=BEEP_IN_connect(sock, &frmt_opt, NULL);
	if(NULL==connection)
	{
		fail("BEEP_IN_connect() failed, see debug output");
	}

	for(i=0;i<NUMMSG;i++)
	{
		DPRINT(DBG_DBG, "Try to send message: '%s'", sendmsgs[i]);

		/* we can't send the messages directly from static memory; it has to be
		 * copied into a malloc()d area so that BRI_sendLog() can free() it */
		sendmsg=emalloc(sizeof(char)*(1+strlen(sendmsgs[i])));
		strcpy(sendmsg,sendmsgs[i]);

		// XXX there shouldn't be a NULL below
		if(FALSE==BEEP_sendLog(connection,sendmsg))
		{
			fail("BRI_sendLog() failed, see debug output");
		}
	}

	BEEP_IN_shutdown(connection);

	DPRINT(DBG_DBG, "Test case done", NO_VARG);
}
END_TEST

START_TEST(TESTimod_beep_controller)
{
	IPC *ipc_socket;
	INW *waiter;
	struct INW_eInfo *event;

	struct message *msg;
	struct IPC_event *evt;
	struct IPC_addr *src;
	struct message *check_msg;

	pid_t module_pid;
	int i,j;

	bool recvd_msg_list[NUMSENDER][NUMMSG];

	char errstr[1024];

	DBG_newproc("Controller");

	DPRINT(DBG_DBG, "Starting controller", NO_VARG);

	msg=emalloc(sizeof(struct message));
	src=emalloc(sizeof(struct IPC_addr));
	evt=emalloc(sizeof(struct IPC_event));
	check_msg=emalloc(sizeof(struct message));

	for(i=0;i<NUMSENDER;i++)
		for(j=0;j<NUMMSG;j++)
			recvd_msg_list[i][j]=FALSE;

	/* init the IPC stuffs */
	DPRINT(DBG_DBG, "Starting IPC controller", NO_VARG);
	ipc_socket=IPC_initController(IPCSOCK);
	if(IPCE_SUCCESS!=IPC_geterrno(ipc_socket))
	{
		IPC_destroy(ipc_socket);

		snprintf(errstr,1024,
				"Failed to init IPC module: %s (%d)",
				IPC_geterrstr(ipc_socket),IPC_geterrno(ipc_socket));
		fail(errstr);
	}

	/* setup the waiter */
	DPRINT(DBG_DBG, "Creating the input waiter", NO_VARG);
	INW_initialize();
	waiter=INW_create();
	INW_addIPC(waiter,ipc_socket);

	/* get the first event */
	DPRINT(DBG_DBG, "Getting the first IPC event", NO_VARG);
	event=INW_getEvent(waiter);
	fail_unless(INWE_IPC==event->event_id, "non-IPC event received");
	fail_unless(event->info == IPCT_EVENT, "IPC event not of expected type");
	fail_unless(IPC_recvEvent(ipc_socket,evt), "Failed to receive IPC event");
	fail_unless(IPCEVT_IMSTART==evt->type, "event received was not Input Module Start");
	module_pid = evt->pid;

	for(i=0; i<(NUMMSG*NUMSENDER); /* inc i below */)
	{
		event=INW_getEvent(waiter);
	
		fail_unless(INWE_IPC==event->event_id,
				"non-IPC event received");

		if (event->info == IPCT_MSG)
		{
			fail_unless(IPC_cRecvMessage(ipc_socket,src,msg),
					"Failed to receive IPC MSG message");

			DPRINT(DBG_DBG, "Received message, comparing '%s' w/ '%s'",
					msg->msg,recvmsgs[msg->sev].msg);

			CMP_msg(msg,&(recvmsgs[msg->sev]));

			/* mark this message as having been received */
			j=0;
			for(j=0;j<NUMSENDER;j++)
			{
				if(recvd_msg_list[j][msg->sev]==FALSE)
				{
					DPRINT(DBG_DBG,
							"Guessing that we've received message %d from initiator %d",
							msg->sev,j);
					recvd_msg_list[j][msg->sev]=TRUE;
					j=-1;
					break;
				}
			}
			if(j!=-1)
			{
				snprintf(errstr,1024,"Received message (%s) more than once",
						msg->msg);
				fail(errstr);
			}

			i++;
		}
		else if (event->info == IPCT_EVENT)
		{
			struct IPC_event   ipcevt;
			IPC_recvEvent(ipc_socket, &ipcevt);
			/* ignore the value, we don't really care */
		}
	}

	/* check to make sure that we've got all our messages */
	for(j=0;j<NUMSENDER;j++)
	{
		for(i=0;i<NUMMSG;i++)
		{
			snprintf(errstr,1024,
					"Failed to receive message no. %d, from sender %d",
					i,j);
			fail_if(recvd_msg_list[j][i]==FALSE,errstr);
		}
	}

	DPRINT(DBG_DBG, "Got all messages, killing the input module (pid %u)",
			module_pid);
	kill(module_pid,SIGTERM);

	sleep(2);

	/* clean up */
	DPRINT(DBG_OMD, "Cleaning up IPC layer", NO_VARG);
	IPC_destroy(ipc_socket);
	free(msg);
	free(src);
	free(evt);
	free(check_msg);
	INW_shutdown();

	DPRINT( DBG_DBG, "Test case done", NO_VARG);
}
END_TEST
