/**
 * @file ut_imod_mark.c
 *
 * @brief Test cases for MARK input module.
 *
 * $Id: ut_imod_mark.c,v 1.1.2.7 2003/04/30 20:03:49 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Sep 04 11:35:11 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "check.h"
#include "ut_imod.h"

#include "error.h"
#include "debug.h"
#include "strcompat.h"

#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/inputwait.h"

#include "config/priority.h"

#include "inputmod/inputmod.h"

/**@defgroup ut_imod_mark Mark Input Module Test Cases
 * @ingroup ut_imod
 * @{
 */

/** Function to run the module */
static void imod_mark_module(void);

/** Function to act as the controller */
static void imod_mark_ctlr(void);

/** Mark Interval in seconds */
#define MARK_INTERVAL   5

/** Allowable variance in timestamps */
#define MARK_FUDGE_FACTOR    2

/*@}*/

static void debug_init(void);
static void debug_finish(void);

void
imod_mark_suite(Suite *suite)
{
    TCase   *tc;

    tc = tcase_create("IMOD_MARK");
    suite_add_tcase(suite,tc);

    tcase_set_run_mode(tc, CK_PARALLEL);

    tcase_add_unchecked_fixture(tc, debug_init, debug_finish);

    tcase_add_test(tc, imod_mark_module);
    tcase_add_test(tc, imod_mark_ctlr);
}

static void
debug_init(void)
{
    DBG_init("Ii", "imod-mark.debug");
}

static void
debug_finish(void)
{
    DPRINT(DBG_DBG, "---Test Case Done----", NO_VARG);
}

#define TEST_IPC_PATH "ut_imod_mark_ipc.sock"
#define TEST_MSG_COUNT 3

START_TEST(imod_mark_module)
{
    struct IMOD_info    *info;

    DBG_newproc("MARK-IM");

    EXIT_initIM();

    info = emalloc(sizeof(struct IMOD_info));
    memset(info, 0, sizeof(struct IMOD_info));

    info->fd_len = 1;
    info->fd = emalloc(sizeof(int) * 2);
    info->chans = info->fd + 1;

    info->fd[0] = -1;
    info->chans[0] = 2;
    info->mark_interval = MARK_INTERVAL;

    info->ipc_path = emalloc(strlen(TEST_IPC_PATH) + 1);
    strlcpy(info->ipc_path, TEST_IPC_PATH, strlen(TEST_IPC_PATH) + 1);

    info->sysChan = 1;

    sleep(5);   /* give the controller a chance to start up */

    IMOD_MARK_start(info, IMOD_MARK);
}
END_TEST

START_TEST(imod_mark_ctlr)
{
    IPC                 *ipc;
    INW                 *inw;
    struct INW_eInfo    *event;
    struct IPC_event    ipcevent;
    struct message      msg;
    char                buf[1024];
    int                 count;
    time_t              last_ts;
    time_t              cur_ts;
    int                 drift;
    struct IPC_addr     mod_ipcaddr;
    struct IPC_addr     src;
    pid_t               mod_pid = 0;
    char                exp_content[1024];
    bool                shutdown;

    DBG_newproc("CTLR");

    unlink(TEST_IPC_PATH);
    ipc = IPC_initController(TEST_IPC_PATH);
    if (IPC_geterrno(ipc) != IPCE_SUCCESS) {
        IPC_destroy(ipc);
        snprintf(buf, sizeof(buf), "IPC_initController() failed (%d) %s",
                IPC_geterrno(ipc), IPC_geterrstr(ipc));
        fail(buf);
    }

    inw = INW_create();
    fail_if(inw == NULL, "INW_create failed");
    fail_unless(INW_addIPC(inw, ipc), "INW_addIPC() failed");


    count = 0;
    shutdown = FALSE;
    last_ts = 0;
    cur_ts = 0;
    while (!shutdown)
    {
        event = INW_getEvent(inw);
        if (event->event_id != INWE_IPC) {
            fail("Unexpected event (non-IPC)");
        }

        if (event->info == IPCT_MSG) {
            fail_unless(IPC_cRecvMessage(ipc, &src, &msg), 
                    "IPC_cRecvMessage() failed");

            if (src.channel == 1) {
                /* log message */
                snprintf(buf, sizeof(buf), "Log message from IM: %s", msg.msg);
                fail(buf);
            }

            fail_unless(src.proc == mod_ipcaddr.proc, 
                    "message proc addr and MARK IM proc addr don't match");
            fail_unless(src.channel == 2, 
                    "message chan doesn't match assigned chan");

            fail_unless(msg.fac == CONFFS_MARK, "bad facility in message");
            fail_unless(msg.sev == CONFFS_INFO, "bad severity in message");
            fail_unless(strcmp(msg.tag, "syslog-im")==0, "bad tag in message");
            fail_unless(strcmp(msg.msg, exp_content)==0, "bad msg in message");

            /* now check the time stamp */

            /* figure out how far off it is from the expected mark
             * time
             */
            drift = msg.timestamp % MARK_INTERVAL;
            if (drift >= (MARK_INTERVAL - MARK_FUDGE_FACTOR)) {
                /* drift < MARK_INTERVAL, by def of MOD */
                cur_ts = msg.timestamp + (MARK_INTERVAL - drift);
            }
            else if (drift <= MARK_FUDGE_FACTOR && drift > 0) {
                cur_ts = msg.timestamp - drift;
            }
            else if (drift != 0) {
                /* Error, drift too far */
                snprintf(buf, sizeof(buf), 
                    "Got mark at %lu, for interval %d, allowable fuge is %d, "
                    "but drift is %d",
                    msg.timestamp, MARK_INTERVAL, MARK_FUDGE_FACTOR, drift);
                fail(buf);
            }
            else {
                cur_ts = msg.timestamp;
            }

            if (last_ts > 0 && cur_ts != (last_ts + MARK_INTERVAL) ) {
                snprintf(buf, sizeof(buf), 
                    "Too long between marks: %d secs, expected %d",
                    (int)(cur_ts - last_ts), MARK_INTERVAL);
                fail(buf);
            }

            last_ts = cur_ts;

            count++;
            if (count >= TEST_MSG_COUNT) {
                DPRINT(DBG_DBG, "Got all %d MARK messages", TEST_MSG_COUNT);

                /* now we need to kill off the input module */
                kill(mod_pid, SIGTERM);
            }

        }
        else if (event->info == IPCT_EVENT) {
            fail_unless(IPC_recvEvent(ipc, &ipcevent), 
                "IPC_recvEvent() failed");

            if (ipcevent.type == IPCEVT_IMSTART) {
                mod_ipcaddr = ipcevent.addr;
                mod_pid = ipcevent.pid;
                snprintf(exp_content, sizeof(exp_content), 
                    "[%d]: -- MARK --", mod_pid);
            }
            /* Note the clean shutdown so we can exit. */
            else if (ipcevent.type == IPCEVT_IMSTOP) {
                fail_unless(count >= TEST_MSG_COUNT, 
                    "Not Enough messages sent before clean shutdown");
                shutdown = TRUE;

            }
            /* fail on everything else */
            else {
                snprintf(buf, sizeof(buf), "Unexpected IPC event %d", 
                        ipcevent.type);
                fail(buf);
            }
        }
        else {
            snprintf(buf, sizeof(buf), "Unexpected IPC message %d", 
                    event->info);
            fail(buf);
        }
    }

    /* cleanup */
    /* This causes problems b/c the IPC queues are trashed ... but I'm
     * not sure why (no other problems known).
     * IPC_destroy(&ipc); */
    INW_destroy(&inw);
    INW_shutdown();
}
END_TEST
