/**
 * @file ut_imod_udp.c
 *
 * @brief UDP Input Module Tests
 * @par
 *
 * $Id: ut_imod_udp.c,v 1.17.2.7 2003/04/30 20:03:49 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu May  9 16:21:54 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#undef WITH_DMALLOC
#define DMALLOC_DISABLE

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "check.h"

#include "error.h"
#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/inputwait.h"
#include "inputmod/inputmod.h"
#include "strcompat.h"
#include "ut_imod.h"
#include "debug.h"

#include "common_funcs.h"

/**@defgroup ut_imod_udp UDP Input Module Test Cases
 * @ingroup ut_imod
 * @{
 */

/** The module portion of the test */
static void TESTimod_udp_module(void);

/** The main loop portion of the test */
static void TESTimod_udp_main(void);

/** The transmitters */
static void TESTimod_udp_send(void);

/*@}*/


/**@defgroup ut_imod_udp_impl UDP Input Module Test Cases Implementation
 * @ingroup ut_imod_udp
 * @{
 */

/** Start log for debugging messages */
static void debug_init(void);

/** End log for debugging messages */
static void debug_done(void);

/** Send several log messages to addresses that are passed in.
 * The addresses are opened with connect(2), then data sent with send(2),
 * and the connection closed with close(2).
 *
 * @param num The number of connections to send
 * @param size The size of each sockaddr
 * @param socks An array of sockets
 * @param addrs An array of sockaddrs
 * @param msgs An array of messages (char *)
 */
#define msgSend(n,s,so,a,m) internal_msgSend(n,s,so,a,m,__FILE__,__LINE__)

/** Send several log messages to addresses that are passed in.
 * Do not use this; instead, use msgSend()
 *
 * @param num The number of connections to send
 * @param size The size of each sockaddr
 * @param socks An array of sockets
 * @param addrs An array of sockaddrs
 * @param msgs An array of messages (char *)
 * @param file The file it was called from
 * @param line The line it was called from
 */
static void internal_msgSend(int num,int size,int *socks,struct sockaddr *addrs,
		char **msgs,char *file,int line);


/** Receive a number of messages across the IPC channel; verify them against
 * a set of pre-defined message formats (using msgCompare())
 *
 * @param num The number of receives expected
 * @param msgs The messages for comparison
 * @param ipc_path The path of the IPC socket to use
 */
#define msgRecv(n,m,i) internal_msgRecv(n,m,i,__FILE__,__LINE__)

/** Receive a number of messages across the IPC channel; verify them against
 * a set of pre-defined message formats (using msgCompare())
 * Do not use this; instead, use msgRecv()
 *
 * @param num The number of receives expected
 * @param msgs The messages for comparison
 * @param ipc_path The path of the IPC socket to use
 * @param file The file name of the calling function
 * @param line The line of the caller
 */
static void internal_msgRecv(int num,struct message *msgs,char *ipc_path,
		char *file,int line);

/** The path to use for the socket */
#define TEST_IMOD_UDP_SOCKET_PATH_IPC "ut_imod_udp_socket_ipc"

/** The communcations port */
#define TEST_IMOD_UDP_PORT	1500

/** The communications host */
#define TEST_IMOD_UDP_HOST	"127.0.0.1"

/*@}*/

static void 
debug_init(void)
{
	if(! DBG_init("Ii", "imod-udp.debug"))
	{
		perror("Error setting up debugging");
		exit(1);
	}
}

static void 
debug_done(void)
{
	DPRINT(DBG_DBG, "Test Case Done", NO_VARG);
}

void 
imod_udp_suite(Suite *suite)
{
	TCase *tc_io;

	tc_io=tcase_create("IMOD_UDP");
	suite_add_tcase(suite,tc_io);

	tcase_add_unchecked_fixture(tc_io, debug_init, debug_done);

	tcase_set_run_mode(tc_io,CK_PARALLEL);

	tcase_add_test(tc_io,TESTimod_udp_main);
	tcase_add_test(tc_io,TESTimod_udp_module);
	tcase_add_test(tc_io,TESTimod_udp_send);
}

START_TEST(TESTimod_udp_module)
{
	struct IMOD_info *info;
	int sock;
	struct sockaddr_in addr;

	DBG_newproc("Module");

    EXIT_initIM();

	info=emalloc(sizeof(struct IMOD_info));

	sleep(1);	/* make sure we construct main first */

	/* init the IPC stuffs */
	info->ipc_path=emalloc(sizeof(char)*(strlen(TEST_IMOD_UDP_SOCKET_PATH_IPC)+1));
	strlcpy(info->ipc_path,TEST_IMOD_UDP_SOCKET_PATH_IPC,
				strlen(TEST_IMOD_UDP_SOCKET_PATH_IPC)+1);

	/* set up our listener */
	sock=socket(AF_INET,SOCK_DGRAM,0);
	if(sock<0)
		fail("Failed to create socket");

	memset(&(addr),0,sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_port=htons(TEST_IMOD_UDP_PORT);
	addr.sin_addr.s_addr=htonl(INADDR_ANY);

	if(bind(sock,(struct sockaddr*)&addr,sizeof(addr))<0)
	{
		char *errmsg=emalloc(40*sizeof(char));
		snprintf(errmsg,40,"Could not bind to socket (port %d)",TEST_IMOD_UDP_PORT);
		fail(errmsg);
		free(errmsg);
	}

	info->fd_len=1;
	info->fd=emalloc(sizeof(int)*info->fd_len * 2);
    info->chans = info->fd + info->fd_len;
	info->fd[0]=sock;
    info->chans[0] = 2;

    IMOD_UDP_start(info, IMOD_UDP);
	/* clean up */
	/* we don't free anything; let IMOD_UDP_start() free what it will */
}
END_TEST

START_TEST(TESTimod_udp_main)
{
	struct message checkmsgs[3]={
		{ 3, 4, 9, "buckaroo", 342857655, -360, "su", 31, ": 'su root' failed for lonvick", },
		{ 0, 9, 14, "192.168.12.12", 342857655, -360, "su", 33, "[12]: 'su root' failed for small", },
		{ 5, 1, 10, "localhost", -1, -1, "", 
#if ENABLE_UNPARSABLE_MARK
            45, " UNPARSABLE<00>this should be in the message",
#else
            34, "<00>this should be in the message", 
#endif
        },
	};

	DBG_newproc("Controller");

	msgRecv(3,checkmsgs,TEST_IMOD_UDP_SOCKET_PATH_IPC);
}
END_TEST

START_TEST(TESTimod_udp_send)
{
	int sock[3];
	struct sockaddr_in addrs[3];
	char *messages[]={
		"<35>1980-11-11T22:14:15-06:00 buckaroo su: 'su root' failed for lonvick",
		"<72>1980-11-11T22:14:15-06:00 192.168.12.12 su[12]: 'su root' failed for small",
		"<00>this should be in the message",
	};

	DBG_newproc("Sender");

	sleep(5);		/* hopefully make sure we do this last */

	/* set up the transmitters */
	sock[0]=socket(AF_INET,SOCK_DGRAM,0);
	sock[1]=socket(AF_INET,SOCK_DGRAM,0);
	sock[2]=socket(AF_INET,SOCK_DGRAM,0);
	if( (sock[0]<0) || (sock[1]<0) || (sock[2]<0) )
		fail("Failed to create socket");

	addrs[0].sin_family=addrs[1].sin_family=addrs[2].sin_family=AF_INET;
	addrs[0].sin_port=addrs[1].sin_port=addrs[2].sin_port=
			htons(TEST_IMOD_UDP_PORT);
	addrs[0].sin_addr.s_addr=addrs[1].sin_addr.s_addr=
			addrs[2].sin_addr.s_addr=inet_addr(TEST_IMOD_UDP_HOST);

	msgSend(3,sizeof(struct sockaddr_in),sock,(struct sockaddr *)addrs,messages);

}
END_TEST

static void internal_msgSend(int num,int size,int *socks,struct sockaddr *addrs,
		char **msgs,char *file,int line)
{
	int i;
	char *errmsg;

	for(i=0;i<num;i++)
	{
		if(connect(socks[i],(struct sockaddr *)&(addrs[i]),size)<0)
		{
			errmsg=emalloc(sizeof(char)*1024);
			snprintf(errmsg,1024,"msgSend() from %s:%d: could not establish connection on socket %d",
					file,line,i);
			fail(errmsg);
			free(errmsg);
		}

		if(-1==send(socks[i],msgs[i],strlen(msgs[i])+1,0))
		{
			errmsg=emalloc(sizeof(char)*1024);
			snprintf(errmsg,1024,"msgSend() from %s:%d: could not send message %d",
					file,line,i);
			fail(errmsg);
			free(errmsg);
		}

        DPRINT(DBG_IM, "Sent message #%d (%s)", i, msgs[i]);

		close(socks[i]);
	}
}

static void internal_msgRecv(int num,struct message *msgs,char *ipc_path,
		char *file,int line)
{
	IPC *ipc_socket;
	INW *waiter;
	struct INW_eInfo *event;
	struct message *msg;
	struct IPC_event ipcevent;
	struct IPC_addr *src;

	char *errmsg=emalloc(sizeof(char)*200);
	int count;
    bool shutdown;
    pid_t mod_pid;

	msg=emalloc(sizeof(struct message));
	src=emalloc(sizeof(struct IPC_addr));

	/* init the IPC stuffs */
	ipc_socket=IPC_initController(ipc_path);
	if(IPCE_SUCCESS!=IPC_geterrno(ipc_socket))
	{
		IPC_destroy(ipc_socket);
		snprintf(errmsg,200,"msgRecv() called from %s:%d: \
				Failed to init IPC module, (%d): %s\n",
				file,line,
				IPC_geterrno(ipc_socket),IPC_geterrstr(ipc_socket));
		fail(errmsg);
	}

	/* setup the waiter */
    INW_initialize();
	waiter=INW_create();
	INW_addIPC(waiter,ipc_socket);

	mark_point();

	/* get our messages */
    shutdown = FALSE;
    count = 0;
    mod_pid = 0;
    while (!shutdown)
	{
		event=INW_getEvent(waiter);

		snprintf(errmsg,200,"msgRecv() called from %s:%d: \
				non-IPC event received",
				file,line);
		fail_unless(INWE_IPC==event->event_id,errmsg);

        if (event->info == IPCT_EVENT) {
            fail_unless(IPC_recvEvent(ipc_socket, &ipcevent),
                "IPC_recvEvent failed");

            if (ipcevent.type == IPCEVT_IMSTART) {
                mod_pid = ipcevent.pid;
            }
            else if (ipcevent.type == IPCEVT_IMSTOP) {
                shutdown = TRUE;
            }
            else {
                snprintf(errmsg, 200, "Unexpected IPC message %d",
                        ipcevent.type);
                fail(errmsg);
            }
        }
        else if (event->info == IPCT_MSG) {
            snprintf(errmsg,200,"msgRecv() called from %s:%d: \
                    Failed to receive IPC MSG message",
                    file,line);
            fail_unless(IPC_cRecvMessage(ipc_socket,src,msg),errmsg);

            internal_msgCompare(msg,&(msgs[count]),file,line);


            count++;
            DPRINT(DBG_IM, "Got Message #%d", count);
            if (count >= num) {
                DPRINT(DBG_IM, "Got all the messages, killig %u", mod_pid);
                kill(mod_pid, SIGTERM);
            }
        }
        else {
            snprintf(errmsg,200,"msgRecv() called from %s:%d:" 
                    " IPC event not of expected type", file,line);
            fail(errmsg);
        }
	}

	mark_point();

	sleep(2);

	/* clean up */
	IPC_destroy(ipc_socket);
	free(msg);
	free(src);
	free(errmsg);

    INW_shutdown();
}

/*@}*/
