/**
 * @file ut_buffer.c
 *
 * @brief Test cases for IPC buffers.
 * @ingroup ut_ipc
 *
 * $Id: ut_buffer.c,v 1.6.2.5 2003/04/30 20:03:50 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Mon Apr  8 17:01:28 PDT 2002
 *
 * @todo Add IPC_buffer tests across processes; probably needs IPC functions.
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "check.h"
#include "error.h"
#include "debug.h"
#include "ipc/buffer.h"

#include "ut_driver.h"
#include "ut_ipc.h"

/**@name Buffer Test Cases
 * @ingroup ut_ipc
 * @{
 */

/** Test buffer creation and destruction */
static void buffer_create(void);

/** Test buffer fixed-length string use */
static void buffer_fixstring(void);

/** Test buffer variable length string use */
static void buffer_varstring(void);

/** Test massive buffer writes, and writes of different types */
static void buffer_crush(void);

/** Test what happens when we overwrite values in the buffer */
static void buffer_overwrite(void);

/** Test getting a variable length string from an empty string */
static void buffer_varfromempty(void);
/*@}*/

/** Debug Log name to use */
#define debug_log_name "ipc-buffer.debug"

/** Initilize debugging */
static void debug_init(void);

/* test suite setup */
void buffer_suite(Suite *suite)
{
    TCase   *tc;

    tc = tcase_create("IPC_buffer");
    suite_add_tcase(suite, tc);
    
    tcase_add_unchecked_fixture(tc, debug_init, debug_done);

    tcase_add_test(tc, buffer_create);
	tcase_add_test(tc, buffer_fixstring);
	tcase_add_test(tc, buffer_varstring);
	tcase_add_test(tc, buffer_varfromempty);
	tcase_add_test(tc, buffer_overwrite);
	tcase_add_test(tc, buffer_crush);
}

static void debug_init(void)
{
    if (! DBG_init("i", debug_log_name) ) {
        perror("Error setting up debugging");
        exit(1);
    }
}

START_TEST(buffer_create)
{
	struct IPC_buffer *buf;
	IPC_bufMark mark;
	unsigned int ret;

	/* first, let's init & allocate */
    buf = emalloc(sizeof(struct IPC_buffer));
	BUF_init(buf);

	buf->size=BUF_INTSZ;
	BUF_alloc(buf);

	mark = BUF_markPos(buf);

	/* let's insert and retrieve and integer */
	fail_unless(BUF_setInteger(buf,0x11223344),
            "Failed to insert '0x11223344' into IPC_buffer");
	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to the initial position");
	fail_unless(BUF_getInteger(buf,&ret),
            "Failed to retrieve an integer from an IPC_buffer");

	fail_unless(ret==0x11223344,
            "Didn't get a correct value back from IPC_buffer");

	/* and let's clean up */
	BUF_free(buf);
	free(buf);
}
END_TEST

START_TEST(buffer_fixstring)
{
	struct IPC_buffer *buf;
	IPC_bufMark mark;
	char *retstr;
	char *instr="Hello World.";

	/* first, let's init & allocate */
	buf=emalloc(sizeof(struct IPC_buffer));
	retstr=emalloc(strlen(instr)+1);
	BUF_init(buf);
	buf->size=BUF_BYTESZ*(strlen(instr)+1);
	BUF_alloc(buf);
	mark=BUF_markPos(buf);

	/* let's insert and retrieve a string */
	fail_unless(BUF_setFixString(buf,instr,strlen(instr)+1),
            "Failed to insert string into IPC_buffer");
	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to the initial position");
	fail_unless(BUF_getString(buf,retstr,strlen(instr)+1),
            "Failed to retrieve a string from IPC_buffer");
	fail_unless(0==strncmp(instr,retstr,strlen(instr)+1),
            "String from buffer doesn't match input string");
    
	/* and let's clean up */
	BUF_free(buf);
	free(buf);
	free(retstr);
}
END_TEST

START_TEST(buffer_varstring)
{
	struct IPC_buffer *buf;
	IPC_bufMark mark;
	char *retstr;
	size_t retlen;
    size_t rsLen;
	char *instr="Hello World.";

	/* first, let's init & allocate */
	buf=emalloc(sizeof(struct IPC_buffer));
    rsLen = strlen(instr)+1;
	retstr=emalloc(rsLen);
	BUF_init(buf);
	buf->size=BUF_BYTESZ*(rsLen)+BUF_INTSZ;
	BUF_alloc(buf);
	mark=BUF_markPos(buf);

	/* let's insert and retrieve a string */
	fail_unless(strlen(instr)+1 == BUF_setVarString(buf,instr,0),
            "Failed to insert string into IPC_buffer");

	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to the initial position");
	fail_unless(BUF_getVarString(buf,&retlen, &retstr, rsLen),
            "Failed to retrieve string length");
	fail_unless(retlen==strlen(instr)+1,
            "String length as reported by buffer not equal to real string length");
	fail_unless(0==strcmp(instr,retstr),
            "String from buffer doesn't match input string");

	/* and let's clean up */
	BUF_free(buf);
	free(buf);
	free(retstr);
}
END_TEST

START_TEST(buffer_varfromempty)
{
	struct IPC_buffer *buf;
	IPC_bufMark mark;
	size_t retlen;
	char *retstr;

	/* first, let's init & allocate */
	buf=emalloc(sizeof(struct IPC_buffer));
	BUF_init(buf);
	buf->size=BUF_EMPTYSTRSZ;
	BUF_alloc(buf);
	mark=BUF_markPos(buf);

	/* let's insert and retrieve the string */
	fail_unless(BUF_setEmptyString(buf),
			"Failed to insert an empty string into the buffer");

	fail_unless(BUF_resetToPos(buf,mark),
			"Failed to reset to the initial position");
	fail_unless(BUF_getVarString(buf,&retlen, &retstr, 0),
            "Failed to retrieve string");

	fail_unless(0==retlen,
			"The empty string has non-zero length");

	fail_unless(NULL==retstr,
			"Incorrect string returned");

    fail_unless(BUF_resetToPos(buf, mark),
            "Failed to reset to the initial position");
    retstr = malloc(10);
    fail_unless(BUF_getVarString(buf, &retlen, &retstr, 10),
            "Failed to retrieve string");
    fail_unless(retlen == 1, 
            "Empty string to a buffer has length > 1 (1 for '\\0'");
    fail_unless(retstr[0]=='\0', 
            "Empty string has non-zero first character");

	/* clean up */
	BUF_free(buf);
	free(buf);
}
END_TEST

START_TEST(buffer_overwrite)
{
	struct IPC_buffer *buf;
	IPC_bufMark mark;
	unsigned int ret;
    uint8_t bret;

	/* first, let's init & allocate */
	buf=emalloc(sizeof(struct IPC_buffer));
	BUF_init(buf);
	buf->size=BUF_INTSZ;
	BUF_alloc(buf);
	mark=BUF_markPos(buf);

	/* let's insert and retrieve and integer */
	fail_unless(BUF_setInteger(buf,0xAABBCCDD),
            "Failed to insert '0xAABBCCDD' into IPC_buffer");
	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to the initial position");
	fail_unless(BUF_getInteger(buf,&ret),
            "Failed to retrieve an integer from an IPC_buffer");
	fail_unless(ret == 0xAABBCCDD,
            "Didn't get a correct value back from IPC_buffer");

	/* now, insert the byte after the integer, and make sure we get an error */
	fail_unless(!BUF_setByte(buf,0xFF),
            "We added a byte, but there should have been no more space");
	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to the initial position");
	fail_unless(BUF_getInteger(buf,&ret),
            "Failed to retrieve an integer from an IPC_buffer");
	fail_unless(ret==0xAABBCCDD,
            "Didn't get a correct value back from IPC_buffer");

	/* and insert the byte over the integer, simulate a reset and
     * further inserts
     */ 
	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to the initial position");
	fail_unless(BUF_setByte(buf,0x11),
            "Failed to add a byte");
	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to the initial position");
	fail_unless(BUF_getByte(buf,&bret),
            "Failed to retrieve an integer from an IPC_buffer");
	fail_unless(bret == 0x11,
            "Didn't get a correct value back from IPC_buffer");

	/* and let's clean up */
	BUF_free(buf);
	free(buf);
}
END_TEST

/** the maximum length of a string in buffer_crush() */
#define	BUF_CRUSH_MAXLEN	1024	

START_TEST(buffer_crush)
{
	struct IPC_buffer *buf;
	IPC_bufMark mark;

	/* return values */
	char        *retstr;
	int8_t      retbyte;
	int32_t     retint;
	time_t      rettime;

	/* test arrays */
	char varstr[4][BUF_CRUSH_MAXLEN]={ 
        "test1", 
        "This is test two, and it's longer",
		"This is test 3, and it's even longer.. these just keep increasing "
            "in size, don't they",
		"This is test 4, and it's not so long"
    };
	int ints[4]={ 0x11223344, 0xFF00FFDD, 0xCCDDEE11, 0x000FF001};
	char bytes[4]={12,123,-127,127,};
	char fixstr[] = "Hello.";

	int i;
	int size;

	time_t temptt;

	(void)time(&temptt);

	/* calculate the size */
	size=0;
	for(i=0;i<4;i++)	/* string length */
		size+=(strlen(varstr[i])+1)*BUF_BYTESZ + BUF_INTSZ;
	for(i=0;i<4;i++)	/* integers */
		size+=BUF_INTSZ;
	for(i=0;i<4;i++)	/* and for bytes */
		size+=BUF_BYTESZ;

	size+=(strlen(fixstr)+1)*BUF_BYTESZ;	/* fixstr */
	size+=BUF_TIMESZ;


	/* init the buffer */
	buf=emalloc(sizeof(struct IPC_buffer));
	BUF_init(buf);
	buf->size=size;
	BUF_alloc(buf);
    memset(buf->data, 0, buf->size);
	mark=BUF_markPos(buf);

	/* add everything */
	for(i=0;i<4;i++)
	{
        DPRINT(DBG_IPC, "inserting %d\n", i);

		fail_unless(BUF_setInteger(buf,ints[i]),
                "Failed to add integer");
		fail_unless(BUF_setByte(buf,bytes[i]),
                "Failed to add byte");
		fail_unless( (strlen(varstr[i])+1) == BUF_setVarString(buf,varstr[i],0),
                "Failed to add variable length string");
	}
	fail_unless(BUF_setFixString(buf,fixstr,strlen(fixstr)+1),
            "Failed to add fixed string");
    DPRINT_BUF(DBG_IPC, "before ts insert, buf = ", buf, -1);
	fail_unless(BUF_setTimestamp(buf,temptt),
            "Failed to add timestamp");

    DPRINT_BUF(DBG_IPC, "done inserting, buf = ", buf, -1);

	/* test to make sure we actually do have everything */
	fail_unless(BUF_resetToPos(buf,mark),
            "Failed to reset to original buffer position");
	for(i=0;i<4;i++)
	{
        DPRINT(DBG_IPC, "retrieving %d\n", i);
		fail_unless(BUF_getInteger(buf, (uint32_t*)&retint),
                "Failed to retrive integer");
        DPRINT(DBG_IPC, "ints[%d] =? 0x%08x", i, retint);
		fail_unless(retint == ints[i],
                "Retrieved integer not equal to inserted integer");

		fail_unless(BUF_getByte(buf, (uint8_t*)&retbyte),
                "Failed to retrieve byte");
        DPRINT(DBG_IPC, "bytes[%d](%02x) =? 0x%02x", 
                i, bytes[i], retbyte);
        DPRINT(DBG_IPC, "is decimal %d =? %d", bytes[i], retbyte);
		fail_unless(retbyte == bytes[i],
                "Retrieved byte not equal to inserted byte");

		fail_unless(BUF_getVarString(buf, (uint32_t*)&retint, &retstr, 0),
                "Failed to get the length of the variable length string");
		fail_unless(retint == strlen(varstr[i])+1,
                "Length of the variable string doesn't match expectations");
        fail_if(retstr == NULL, "Returned string is null");
		fail_unless(strcmp(varstr[i],retstr) == 0,
                "String from buffer doesn't match input string");
        free(retstr);
	}

    retstr = emalloc(strlen(fixstr)+1);
    fail_unless(BUF_getString(buf, retstr, strlen(fixstr)+1),
            "Failed to get fixed string");
    fail_unless(strcmp(retstr, fixstr) == 0,
            "Fixed string doesn't match input string");
    fail_unless(BUF_getTimestamp(buf, &rettime),
            "Failed to get timestamp");
    fail_unless(rettime == temptt,
            "Timestamp doesn't match input value");

    free(retstr);
	BUF_free(buf);
	free(buf);
}
END_TEST
#undef BUF_CRUSH_MAXLEN

