/**
 * @file ut_ipc_init.c
 *
 * @brief IPC Layer Initilization tests
 * @ingroup ut_ipc
 *
 *
 * $Id: ut_ipc_init.c,v 1.12.2.5 2003/04/30 20:03:51 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Apr 02 13:46:42 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include "check.h"

#include "util/tsqueue.h"

#include "ipc/ipc_private.h"
#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/ipc_ctl.h"
#include "ipc/ipcll.h"
#include "ipc/inputwait.h"
#include "ipc/ipc_cstate.h"

#include "ut_ipc.h"
#include "strcompat.h"
#include "debug.h"

/**@name IPC Initilization Test Cases
 * @ingroup ut_ipc
 * @{
 */

/** Test IPC controller connection creation/destruction */
static void ipc_ctlr_create(void);

/** Test IPC client connection creation/destruction */
static void ipc_im_create(void);

/*@}*/

/** Log for debugging messages */
#define debug_log_name      "ipc-init.debug"

/** Initilize Debugging */
static void debug_init(void);

void ipc_init_suite(Suite *suite)
{
    TCase   *tc;
    int i;

    /* setup debugging */

    tc = tcase_create("Creation/Start/Stop/Destruction");
    suite_add_tcase(suite, tc);

    tcase_add_unchecked_fixture(tc, debug_init, debug_done);


    tcase_set_run_mode(tc, CK_PARALLEL);
    tcase_add_test(tc, ipc_ctlr_create);
    for (i = 0 ; i < IPC_TEST_CLI_COUNT; i++) {
        tcase_add_test(tc, ipc_im_create); 
    }
}

static void debug_init(void)
{
    if (! DBG_init("ip", debug_log_name) ) {
        perror("Error setting up debugging:");
        exit(1);
    }
}

/******** Test Cases **********/

START_TEST(ipc_ctlr_create)
{
    int                 pidx;
    char                msg[1024];
    int                 waitType;
    struct IPC_event    event;
    time_t              giveup_time;
    int                 cliStatus[IPC_TEST_CLI_COUNT];
    int                 cliStartCount;
    int                 cliStopCount;
    int                 i;
    struct INW_eInfo    *inwevt;

    memset(cliStatus, 0, sizeof(cliStatus));
    cliStartCount = 0;
    cliStopCount = 0;

    DBG_newproc("CTLR");

    ipc_init_ctlr();

    fail_unless(strcmp(ctlrHandle->sock_path, SOCK_PATH) == 0, 
            "ctlr::sock_path not copied correctly");

    fail_unless(ctlrHandle->paddr == IPCAP_CTLR,
            "ctlr::paddr not set correctly");

    fail_unless(ctlrHandle->mode == IPCM_SERVER,
            "ctlr::mode not set correctly");

    fail_unless(ctlrHandle->cliCount == 0,
            "ctlr::cliCount not zero");

    fail_unless(ctlrHandle->cliConns != NULL,
            "ctlr::cliConns not allocated");

    fail_unless(ctlrHandle->mainState.sock >= 0,
            "ctlr::sock not set correctly");

    fail_unless(ctlrHandle->mainState.state == IPCST_READY,
            "ctlr::state not ready");

    for (i = 0; i < IPCQ_COUNT; i++) {
        fail_if(ctlrHandle->queues[i] == NULL,
                "ctlr::queue not fully initilized");
    }

    giveup_time = time(NULL) + 300;     /* try for 5 min */
    while (time(NULL) < giveup_time && cliStopCount < IPC_TEST_CLI_COUNT) { 
        /* make the ipc layer check the sockt */
        inwevt = INW_getEvent(waiter);
        if (inwevt == NULL) {
            snprintf(msg, sizeof(msg), "Error in INW_getEvent(): (%d) %s",
                    IPC_geterrno(ctlrHandle),
                    IPC_geterrstr(ctlrHandle));

            DPRINT(DBG_IPC, "FAIL: %s", msg);
            fail(msg);
        }
        fail_unless(inwevt->event_id == INWE_IPC, "Unexpected event from INW");

        DPRINT(DBG_IPC, "Got event: id = %d; info = %X",
                inwevt->event_id, inwevt->info);

        waitType = inwevt->info;


        if (waitType == IPCT_EVENT) {
            if (IPC_recvEvent(ctlrHandle, &event) < 0) {
                snprintf(msg, sizeof(msg),
                    "Error recieving input module event: (%d) %s",
                    IPC_geterrno(ctlrHandle),
                    IPC_geterrstr(ctlrHandle));

                fail(msg);
            }
            /* Process the event */
            pidx = IPC_PAToIndex(event.addr.proc);
            if (event.addr.channel != IPCAC_GENERAL) {
                snprintf(msg, sizeof(msg),
                    "Inputmodule (%d) event (%d) not on general channel",
                    pidx, event.type);
            }

            if (event.type == IPCEVT_IMSTART && cliStatus[pidx] == 0) {
                DPRINT(DBG_IPC, "got start from: %d", 
                        event.addr.proc);
                cliStatus[pidx] = 1;
                cliStartCount++;
            }
            else if (event.type == IPCEVT_IMSTOP && cliStatus[pidx] == 1) {
                DPRINT(DBG_IPC, "got stop from: %d", 
                        event.addr.proc);
                /* Address my be reused */
                cliStatus[pidx] = 0;
                cliStopCount++;
            }
            else {
                snprintf(msg, sizeof(msg),
                    "client %d sent unexpected event %d, in state %d", 
                    event.addr.proc, event.type, cliStatus[pidx]);

                DPRINT(DBG_IPC, "FAIL: %s", msg);

                fail(msg);
            }
        }
        else if (waitType != 0) {
            snprintf(msg, sizeof(msg),
                "Got unexpected packets (waiting map = 0x%X)", waitType);

            fail(msg);
        }
    }

    /* okay, final checks that everything completed */
    if (cliStartCount < IPC_TEST_CLI_COUNT || cliStopCount != cliStartCount) {
        snprintf(msg, sizeof(msg), 
            "%d clients did not even start, %d clients started but did not"
            " finish", 
            IPC_TEST_CLI_COUNT - cliStartCount, cliStartCount - cliStopCount);
        fail(msg);
    }

    ipc_destroy_ctlr();

    DPRINT(DBG_IPC, "Controller Stop", NO_VARG);
}
END_TEST

START_TEST(ipc_im_create)
{
    char    msg[1024];
    int     i;

    DBG_newproc("CLI");

    ipc_init_client();

    fail_unless(strcmp(cliHandle->sock_path, SOCK_PATH) == 0, 
            "cli::sock_path not copied correctly");

    fail_unless(cliHandle->paddr == IPCAP_UNSET,
            "cli::paddr not unset after create");

    fail_unless(cliHandle->mode == IPCM_CLIENT,
            "cli::mode not set correctly");

    fail_unless(cliHandle->cliCount == 0,
            "cli::cliCount not zero after create");

    fail_unless(cliHandle->cliConns == NULL,
            "cli::cliConns allocated after create");

    fail_unless(cliHandle->mainState.sock == -1,
            "cli::sock not set correctly after create");

    fail_unless(cliHandle->mainState.state == IPCST_FREE,
            "cli::state not inactive after create");

    for (i = 0; i < IPCQ_COUNT; i++) {
        fail_if(cliHandle->queues[i] == NULL,
                "cli::queue not fully initilized");
    }

    /* mark a point just before and just after start up */
    mark_point();

    if (! IPC_inputmoduleStart(cliHandle) ) {
        snprintf(msg, sizeof(msg),
            "Starting Input module failed (%d), %s",
            IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
        fail(msg);
    }
    mark_point();

    DPRINT(DBG_IPC, "inputmodule is started", NO_VARG);

    fail_unless(cliHandle->mainState.state == IPCST_READY,
            "cli::state not ready after imstart");

    fail_unless(cliHandle->paddr != IPCAP_UNSET,
            "cli::paddr not ready after imstart");

    fail_unless(cliHandle->cliCount == 0,
            "cli::cliCount not zero after imstart");

    fail_unless(cliHandle->cliConns == NULL,
            "cli::cliConns allocated after imstart");

    fail_unless(cliHandle->mainState.sock >= 0,
            "cli::sock not set correctly after imstart");

    /* Do a proper shutdown */
    mark_point();   /* mark point just before shutdown */
    sleep(10);
    ipc_destroy_client();

    DPRINT(DBG_IPC, "inputmodule is exiting", NO_VARG);
}
END_TEST
