/**
 * @file ut_ipc_lock.c
 *
 * @brief Test cases for IPC lock passing functions.
 * @ingroup ut_ipc
 *
 * $Id: ut_ipc_lock.c,v 1.3.2.5 2003/04/30 20:03:51 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Sat Jul 13 15:59:53 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>

#include "ipc/ipc_private.h"
#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/ipcll.h"
#include "util/list.h"
#include "util/tsqueue.h"

#include "check.h"
#include "ut_ipc.h"
#include "ut_driver.h"

/**@name IPC Lock Passing Test Cases
 * @ingroup ut_ipc
 * @{
 */

/** Test IPC controller part */
static void ipc_lock_ctlr(void);

/** Test IPC Input Module */
static void ipc_lock_im(void);

/*@}*/

/** Debug log name */
#define debug_log_name  "ipc-lock.debug"

/** Initilize debugging */
static void debug_init(void);

/** 
 * Search a list of IPC_TEST_CLI_COUNT elements for a maching process
 * id, and return the index.
 *
 * @param proc The process address to look for
 * @paddrs The list of process addresses
 * @return the index into \c paddrs that contains \c proc; < 0 if not found.
 */
static int findCliNum(uint32_t proc, uint32_t *paddrs);

/**@name Client States.
 * These make a natural progression, so each client should be moved
 * through each state.  And should be in next_state-1 before entering
 * next_state.
 *
 * They are all bit masks, so setting a state is just an |=, the only
 * state that can be set in a range of time is CLI_GOTMSG, which might
 * come in anytime after CLI_GOTLOCK is set.
 * @{
 */
#define CLI_NOSTART     0x00
#define CLI_START       0x01
#define CLI_GOTLOCK     0x02
#define CLI_DONE        0x04
#define CLI_EXIT        0x08
#define CLI_GOTMSG      0x10
/*@}*/

/** Value that the client state must have to be considered complete. */
#define CLI_COMPLETE    0x1F

void
ipc_lock_suite(Suite *suite)
{
    TCase   *tc;
    int     i;

    tc = tcase_create("Lock Passing");
    suite_add_tcase(suite, tc); 

    tcase_add_unchecked_fixture(tc, debug_init, debug_done);

    tcase_set_run_mode(tc, CK_PARALLEL);

    tcase_add_test(tc, ipc_lock_ctlr);
    for (i = 0; i < IPC_TEST_CLI_COUNT; i++) {
        tcase_add_test(tc, ipc_lock_im);
    }
}

static void
debug_init(void)
{
    if (! DBG_init("i", debug_log_name) ) {
        perror("Error setting up debugging");
        exit(1);
    }
}

static int
findCliNum(uint32_t proc, uint32_t *paddrs)
{
    int         i;
    for (i = 0; i < IPC_TEST_CLI_COUNT; i++) {
        if (paddrs[i] == proc) 
            return i;
    }

    return -1;
}


/* ************** Test Cases ****************** */


START_TEST(ipc_lock_ctlr)
{
    int                 cliState[IPC_TEST_CLI_COUNT];
    uint32_t            cliPaddr[IPC_TEST_CLI_COUNT];
    int                 cliIndex;
    int                 clisDone;
    struct INW_eInfo    *inwevt;
    struct IPC_event    ipcevt;
    struct IPC_addr     src;
    char                msg[1024];
    int                 cli;
    List                lkreqs;
    IPC_lockType        lktype;
    IPC_lock            lock;
    bool                have_lock;
    int                 i;

    DBG_newproc("CTLR");

    memset(cliState, 0, sizeof(cliState));
    memset(cliPaddr, 0, sizeof(cliPaddr));
    cliIndex = 0;
    clisDone = 0;
    lkreqs = LL_create();

    have_lock = TRUE;   /* we do start with the lock */

    ipc_init_ctlr();

    while (clisDone < IPC_TEST_CLI_COUNT) {
        inwevt = INW_getEvent(waiter);
        if (inwevt == NULL) fail("INW_getEVent() failed");
        if (inwevt->event_id != INWE_IPC) fail("Unexpected INW Event");

        DPRINT(DBG_IPC, "test: got event type %d", inwevt->info);

        switch (inwevt->info) {
            case IPCT_EVENT:
            {
                if (! IPC_recvEvent(ctlrHandle, &ipcevt) ) {
                    snprintf(msg, sizeof(msg),
                        "IPC_recvEvent(): (%d) %s",
                        IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));

                    fail(msg);
                }

                cli = findCliNum(ipcevt.addr.proc, cliPaddr);


                if (ipcevt.type == IPCEVT_IMSTART) {
                    if (cli >= 0) {
                        snprintf(msg, sizeof(msg),
                            "Got IMSTART for paddr that exists (%d)",
                            ipcevt.addr.proc);

                        fail(msg);
                    }
                    if (cliIndex >= IPC_TEST_CLI_COUNT) {
                        fail("too many clients have started");
                    }

                    cli = cliIndex;
                    cliIndex++;

                    cliPaddr[cli] = ipcevt.addr.proc;
                    cliState[cli] |= CLI_START;
                }
                else if (ipcevt.type == IPCEVT_CONEND) {
                    /*
                     * The test IM sends this event after it gets the
                     * lock.  IPCEVT_CONEND is used b/c input modules
                     * can explicitly send them, and they are easier to
                     * send than say paths or messages.
                     */
                    if ( (cliState[cli] & CLI_GOTLOCK) == 0) {
                        snprintf(msg, sizeof(msg),
                            "Got CONEND in wrong state (%x) from %d",
                            cliState[cli], ipcevt.addr.proc);
                        fail(msg);
                    }
                    cliState[cli] |= CLI_GOTMSG;
                }
                else if (ipcevt.type == IPCEVT_IMSTOP) {
                    DPRINT(DBG_IPC, "test: Got IMSTOP from %d", 
                            ipcevt.addr.proc);

                    if ( (cliState[cli] & CLI_DONE) == 0) {
                        snprintf(msg, sizeof(msg),
                            "Got IMSTOP in wrong state (%x) from %d",
                            cliState[cli], ipcevt.addr.proc);
                        fail(msg);
                    }
                    cliState[cli] |= CLI_EXIT;

                    /* Marke this paddr as inactive, but allow its
                     * recovery if needed.
                     */
                    cliPaddr[cli] = -cliPaddr[cli];

                    /* One more client finished */
                    clisDone++;
                }
                else {
                    snprintf(msg, sizeof(msg),
                        "Got Unexpected event (%d) from %d in state %x",
                        ipcevt.type, ipcevt.addr.proc, cliState[cli]);
                    fail(msg);
                }

                break;
            }
            case IPCT_LOCK:
            {
                if (! IPC_cRecvLock(ctlrHandle, &src, &lktype, &lock) ) {
                    snprintf(msg, sizeof(msg), "IPC_cRecvLock(): (%d) %s",
                        IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));
                    DPRINT(DBG_IPC, "FAIL: %s", msg);
                    fail(msg);
                }
                cli = findCliNum(src.proc, cliPaddr);

                if (lktype == IPCLT_REQ) {
                    if ( (cliState[cli] & CLI_START) == 0) {
                        snprintf(msg, sizeof(msg),
                            "Got lock request before IMSTART from %d",
                            src.proc);
                        fail(msg);
                    }
                    if (lock != IPCL_ACCEPT) {
                        fail("Client requested the wrong lock");
                    }

                    LL_append(lkreqs, &(cliPaddr[cli]));
                }
                else if (lktype == IPCLT_LOCK) {
                    /* All we check is that the client has the lock, the
                     * message is lower priority, so it my not have come
                     * in yet.
                     */
                    if ( (cliState[cli] & CLI_GOTLOCK) == 0) {
                        snprintf(msg, sizeof(msg),
                            "Client %d released lock before getting it"
                            " (state = %x)", src.proc, cliState[cli]);
                        fail(msg);
                    }
                    if (lock != IPCL_ACCEPT) {
                        fail("Client requested the wrong lock");
                    }
                    cliState[cli] |= CLI_DONE;
                    have_lock = TRUE;
                }
                else {
                    snprintf(msg, sizeof(msg),
                        "Got unexpected lock subtype %d from %d",
                        lktype, src.proc);
                    fail(msg);
                }
                break;
            }
            default:
            {
                snprintf(msg, sizeof(msg),
                        "Got unexpected IPC event %d", inwevt->info);
                fail(msg);
                break;
            }
        }

        if (have_lock && ! LL_isEmpty(lkreqs) ) {
            /* abusing the src IPC_addr structure here */
            src.proc = *(int*)LL_first(lkreqs);
            src.channel = IPCAC_GENERAL;
            DPRINT(DBG_IPC, "test: giving lock to %d", src.proc);
            if (! IPC_cSendLock(ctlrHandle, &src, IPCL_ACCEPT) ) {
                snprintf(msg, sizeof(msg),
                    "IPC_cSendLock(): (%d) %s", 
                    IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));
                fail(msg);
            }
            fail_unless(NULL!=LL_rmFirst(lkreqs), "LL_rmFirst() failed");
            have_lock = FALSE;

            cli = findCliNum(src.proc, cliPaddr);
            cliState[cli] |= CLI_GOTLOCK;
        }
    }

    /* final check */
    for (i = 0; i < IPC_TEST_CLI_COUNT; i++) {
        if (cliState[i] != CLI_COMPLETE) {
            snprintf(msg, sizeof(msg), "Client %d (%d) is not finished yet", 
                    i, cliPaddr[i]);
            fail(msg);
        }
    }


    ipc_destroy_ctlr();
}
END_TEST

START_TEST(ipc_lock_im)
{
    char                msg[1024];
    struct INW_eInfo    *evt;
    struct IPC_addr     dst;
    IPC_lock            lock;

    DBG_newproc("CLI");

    ipc_init_client();

    if (! IPC_inputmoduleStart(cliHandle) ) {
        snprintf(msg, sizeof(msg),
            "IPC_inputmoduleStart(): (%d) %s",
            IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
        fail(msg);
    }

    /* request the lock */
    if (! IPC_lockRequest(cliHandle, IPCL_ACCEPT) ) {
        snprintf(msg, sizeof(msg),
            "IPC_lockRequest(): (%d) %s",
            IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
        fail(msg);
    }

    /* wait for the lock */
    evt = INW_getEvent(waiter);
    fail_if(evt == NULL, "INW_getEvent() failed");
    fail_unless(evt->event_id == INWE_IPC, "Unexpected INW event type");
    fail_unless(evt->info == IPCT_LOCK, "Unexpected IPC event type");

    if (! IPC_imRecvLock(cliHandle, &lock) ) {
        snprintf(msg, sizeof(msg),
            "IPC_imRecvLock(): (%d) %s",
            IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
        fail(msg);
    }
    fail_unless(lock == IPCL_ACCEPT, "Got Wrong Lock Type");

    /* Send our event that signals that we got the lock */
    dst.proc = IPCAP_CTLR;
    dst.channel = IPCAC_GENERAL;
    if (! IPC_sendEvent(cliHandle, &dst, IPCAC_GENERAL, IPCEVT_CONEND, 0) ) {
        snprintf(msg, sizeof(msg),
            "IPC_imRecvLock(): (%d) %s",
            IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
        fail(msg);
    }

    /* release the lock */
    IPC_lockRelease(cliHandle, IPCL_ACCEPT);

    ipc_destroy_client();
}
END_TEST
