/**
 * @file ut_ipc_msg.c
 *
 * @brief Test message sending and recieving over the IPC channle.
 * @ingroup ut_ipc
 *
 *
 * $Id: ut_ipc_msg.c,v 1.11.2.5 2003/04/30 20:03:51 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Apr 17 16:32:11 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "check.h"

#include "util/tsqueue.h"

#include "ut_ipc.h"

#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/ipc_ctl.h"
#include "ipc/ipcll.h"
#include "ipc/inputwait.h"
#include "message/message.h"

#include "strcompat.h"
#include "error.h"
#include "debug.h"

#include "util/bitvec.h"

#include "common_funcs.h"

/**@name IPC Syslog Message Test Cases
 * @ingroup ut_ipc
 * @{
 */

/** Test IPC Controller recieving of messages */
static void ipc_ctlr_msg(void);

/** Test IPC input module sending of messages */
static void ipc_im_msg(void);

/*@}*/


/**@name Test Messages
 * @ingroup ut_ipc
 * Arrays that defines the list of test messages to send (or to expect).
 * Some of this info is meant to be filled in by ipc_msg_setup(). 
 * @{
 */ 

/** Number of test messages each input process must send */
#define TEST_MSG_COUNT 5

/** Message structures to hold the messages */
static struct message   testMsgs[TEST_MSG_COUNT]; 

/* The message part */
static char const   *msgStr[TEST_MSG_COUNT] = {
    "The Quick Brown Fox",
    "Testing Message Sending",
    "A Short Message",
    "A Very very very very long test message",
    "Anothe test message",
};

/** Host names */
static char const   *hostStr[TEST_MSG_COUNT] = {
    "scott", "empire", "snowball", "bird", "pitofdespair",
};

/** IP addrs */
static char const   *ipStr[TEST_MSG_COUNT] = {
    "129.12.23.21", "31.32.39.124", "249.139.23.29", "192.31.21.25", "192.168.123.3",
};

/** Tag names */
static char const   *tagStr[TEST_MSG_COUNT] = {
    "sendmail", "rsh", "fingerd", "ssh", "in.ftpd",
};

/** The domain part of the host name */
#define TEST_DOMAIN ".sdsc.edu"

/** Number of test paths each input process should send */
#define TEST_PATH_COUNT 3

/** Path structures to hold the paths */
static struct path		testPaths[TEST_PATH_COUNT];

static char const	*path_fromFQDN[TEST_PATH_COUNT] = {
	"a.root-servers.net", "www.sdsc.edu", "mail.sdsc.edu", 
};

static char const	*path_fromIP[TEST_PATH_COUNT] = {
	"10.0.0.1", "10.1.2.3", "10.2.3.4",
};

static char const	*path_toFQDN[TEST_PATH_COUNT] = {
	"b.root-servers.net", "ftp.sdsc.edu", "mail.ucsd.edu",
};

static char const	*path_toIP[TEST_PATH_COUNT] = {
	"192.168.1.2", "10.230.12.45", "10.34.54.59",
};

/*@}*/

#define debug_log_name "ipc-msg.debug"
static void debug_init(void);

/** Setup the list of messages to send from im -> ctlr */
static void ipc_msg_setup(void);

/** cleanup after the test case */
static void ipc_msg_teardown(void);

void 
ipc_msg_suite(Suite *suite)
{
    TCase   *tc;
    int     i;


    tc = tcase_create("Syslog Messages");
    suite_add_tcase(suite, tc);

    tcase_add_unchecked_fixture(tc, debug_init, debug_done);

    tcase_set_run_mode(tc, CK_PARALLEL);
    tcase_add_unchecked_fixture(tc, ipc_msg_setup, ipc_msg_teardown);

    tcase_add_test(tc, ipc_ctlr_msg);
    for (i = 0; i < IPC_TEST_CLI_COUNT; i++) {
        tcase_add_test(tc, ipc_im_msg);
    }
}

/************* Supporting Functions ****************/

static void
debug_init(void)
{
    /* setup debugging */
    if (! DBG_init("i", debug_log_name) ) {
        perror("Error setting up debugging:");
        exit(1);
    }
}
        
static void
ipc_msg_setup(void)
{
    int             i;
    struct message  *msg;
    struct in_addr  addr;
	struct path		*pth;

    for (i = 0; i < TEST_MSG_COUNT; i++) {
        msg = &(testMsgs[i]);
        msg->sev = i; 
        msg->fac = (i * 2) % 9;

        msg->hostlen = strlen(hostStr[i])+1;
        msg->host = emalloc(msg->hostlen);
        strlcpy(msg->host, hostStr[i], msg->hostlen);

        msg->timestamp = time(NULL);
        msg->tz = 480;

        strlcpy(msg->tag, tagStr[i], MSGTAG_LEN);

        msg->msglen = strlen(msgStr[i])+1;
        msg->msg = emalloc(msg->msglen);
        strlcpy(msg->msg, msgStr[i], msg->msglen);

        msg->pathid = (i * 3) % 7;

        inet_aton(ipStr[i],  &addr);
        msg->dip = NA_newINet4(addr);

        msg->dfqdnlen = strlen(hostStr[i]) + strlen(TEST_DOMAIN) + 1; 
        msg->dfqdn = emalloc(msg->dfqdnlen);
        strlcpy(msg->dfqdn, hostStr[i], msg->dfqdnlen);
        strlcat(msg->dfqdn, TEST_DOMAIN, msg->dfqdnlen);

    }

	for(i=0;i<TEST_PATH_COUNT;i++)
	{
		pth = &(testPaths[i]);

		pth->id=i;
		pth->connid=i*42;

		pth->ffqdnlen=strlen(path_fromFQDN[i])+1;
		pth->fromFQDN=emalloc(pth->ffqdnlen);
		strlcpy(pth->fromFQDN, path_fromFQDN[i], pth->ffqdnlen);
		pth->fromIP=NA_newINetLookup(path_fromIP[i]);

		pth->tfqdnlen=strlen(path_toFQDN[i])+1;
		pth->toFQDN=emalloc(pth->tfqdnlen);
		strlcpy(pth->toFQDN, path_toFQDN[i], pth->tfqdnlen);
		pth->toIP=NA_newINetLookup(path_toIP[i]);

		pth->props=BM_MBIT(VALID_USER);

		pth->next=NULL;
	}

	/* also, set path 2 to point to path 1 as the next field */
	testPaths[2].next=&(testPaths[1]);
}

static void
ipc_msg_teardown(void)
{
    int             i;
    struct message  *msg;

    for (i = 0; i < TEST_MSG_COUNT; i++) {
        msg = &(testMsgs[i]);
        free(msg->host);
        free(msg->msg);
        free(msg->dfqdn);
    }
}



/***************** Test Cases **********************/
START_TEST(ipc_ctlr_msg)
{
    int                 clisDone;
    int                 clisStarted;
    int                 cliState[IPC_TEST_CLI_COUNT];
	int					cliPathState[IPC_TEST_CLI_COUNT];
    int                 waitType;
    int                 cli;
    struct IPC_event    evt;
    struct message      msg;
	struct path			pth;
    struct IPC_addr     src;
    char                failmsg[1024];
    int                 i;

    struct INW_eInfo    *inwevt;

    DBG_newproc("CTLR");

    ipc_init_ctlr();
    
    cli = -1;
    clisStarted = 0;
    clisDone = 0;
    for (i = 0; i < IPC_TEST_CLI_COUNT; i++) {
        cliState[i] = -1;
		cliPathState[i] = -1;
    }

    while (clisDone < IPC_TEST_CLI_COUNT) {
        inwevt = INW_getEvent(waiter);

        /* if statements are used to reduce the number of times
         * _fail_unless() is called.  At the moment it is called enough
         * times to fill up the test info pipe, causing the test to
         * hang.
         */
        if (inwevt == NULL) 
            fail("Error waiting for an event");
        if (inwevt->event_id != INWE_IPC)
            fail("Unexpected Event ID from INW");

        waitType = inwevt->info;

        DPRINT(DBG_IPC, "waitType = 0x%x", waitType);

        if (waitType == IPCT_EVENT) {
            if (! IPC_recvEvent(ctlrHandle, &evt) ) {
                snprintf(failmsg, sizeof(failmsg),
                    "Error recieving IM event: (%d) %s",
                    IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));
                fail(failmsg);
            }

            cli = IPC_PAToIndex(evt.addr.proc);

            if (evt.type == IPCEVT_IMSTART) {
                if (cliState[cli] != -1) {
                    snprintf(failmsg, sizeof(failmsg), 
                            "Got 2 start events from client %d", 
                            evt.addr.proc);
                    fail(failmsg);
                }
                else {
					cliPathState[cli]++;
                    cliState[cli]++;    /* == 0 now */
                    clisStarted++;
                }
            }
            else if (evt.type == IPCEVT_IMSTOP) {
                if (cliState[cli] < TEST_MSG_COUNT) {
                    snprintf(failmsg, sizeof(failmsg), 
                        "client %d stoped with only %d messages, "
                        "expecting %d messages",
                        evt.addr.proc, cliState[cli], TEST_MSG_COUNT);
                    fail(failmsg);
                }
				else if (cliPathState[cli] < TEST_PATH_COUNT)
				{
					snprintf(failmsg, sizeof(failmsg),
							"client %d stopped with only %d paths, "
							"expected %d paths",
							evt.addr.proc, cliPathState[cli], TEST_PATH_COUNT);
					fail(failmsg);
				}
                else {
                    cliState[cli] = -1; /* prep for reuse */
					cliPathState[cli] = -1;
                    clisDone++;
                }
            }
            else if (evt.type == IPCEVT_IMDEATH) {
                snprintf(failmsg, sizeof(failmsg),
                    "client %d died early", evt.addr.proc);
                fail(failmsg);
            }
            else {
                snprintf(failmsg, sizeof(failmsg),
                    "client %d send unknow event (%d)", 
                    evt.addr.proc, evt.type);
                fail(failmsg);
            }
        }
        else if (waitType == IPCT_MSG) {
            if (! IPC_cRecvMessage(ctlrHandle, &src, &msg)) {
                snprintf(failmsg, sizeof(failmsg),
                    "Error recieving message (%d) %s",  
                    IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));
                fail(failmsg);
            }

            cli = IPC_PAToIndex(src.proc);
            if (cliState[cli] < 0) {
                snprintf(failmsg, sizeof(failmsg), 
                    "Message from unstarted client %d", evt.addr.proc);
                fail(failmsg);
            }

            if (cliState[cli] >= TEST_MSG_COUNT) {
                snprintf(failmsg, sizeof(failmsg), 
                    "Too Many messages from client %d", evt.addr.proc);
                fail(failmsg);
            }

            DPRINT(DBG_IPC, "dfqdnlen = %d; dfqdn = '%s'", 
                    msg.dfqdnlen, msg.dfqdn);
            CMP_msg(&msg, &(testMsgs[cliState[cli]]));
                    
            cliState[cli]++;
            DPRINT(DBG_IPC, "Got good message (%d) from %d", 
                    cliState[cli], evt.addr.proc);
        }
		else if(waitType==IPCT_PATH)
		{
			if(! IPC_cRecvPath(ctlrHandle, &src, &pth))
			{
				snprintf(failmsg,sizeof(failmsg),
						"Error receiving path: (%d) %s",
						IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));
			}

			cli = IPC_PAToIndex(src.proc);
			if(cliPathState[cli] < 0)
			{
				snprintf(failmsg, sizeof(failmsg),
						"Message from unstarted client %d", evt.addr.proc);
				fail(failmsg);
			}

			if(cliPathState[cli] >= TEST_PATH_COUNT)
			{
				snprintf(failmsg,sizeof(failmsg),
						"Too many messages from client %d", evt.addr.proc);
				fail(failmsg);
			}

			CMP_path(&pth, &(testPaths[cliPathState[cli]]));

			cliPathState[cli]++;

			DPRINT(DBG_IPC, "Got good path (%d) from %d",
					cliPathState[cli], evt.addr.proc);
		}

    }

    ipc_destroy_ctlr();
}
END_TEST

START_TEST(ipc_im_msg)
{
    char        failmsg[1024];
    int         sleeptm;
    int         i;

    DBG_newproc("CLI");

    ipc_init_client();

    if (! IPC_inputmoduleStart(cliHandle) ) {
        snprintf(failmsg, sizeof(failmsg),
                "Starting input module: (%d) %s", 
                IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
        fail(failmsg);
    }

    /* bad way to seed this, but it should do */
    srand((int)time(NULL));

    mark_point();

    for (i = 0; i < TEST_MSG_COUNT; i++) {
        sleeptm = 1+(rand() % 4);   /* again I'm being lazy */
        sleep(sleeptm);
        if (! IPC_imSendMessage(cliHandle, 10, &(testMsgs[i]))) {
            snprintf(failmsg, sizeof(failmsg),
                    "Sending message %d: (%d) %s", i,
                    IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
            fail(failmsg);
        }
    }

	mark_point();

	for(i=0; i<TEST_PATH_COUNT; i++)
	{
		sleeptm = 1+(rand() % 4);   /* again I'm being lazy */
		sleep(sleeptm);

		if (! IPC_imSendPath(cliHandle, 10, &(testPaths[i])))
		{
			snprintf(failmsg, sizeof(failmsg),
					"Sending path %d: (%d) %s", i,
					IPC_geterrno(cliHandle), IPC_geterrstr(cliHandle));
			fail(failmsg);
		}
	}

    mark_point();
    ipc_destroy_client();
}
END_TEST
