/**
 * @file ut_ipc_sig.c
 *
 * @brief Tests for signal handling in both the input waiter and IPC
 * layer.
 * @ingroup ut_ipc
 *
 * $Id: ut_ipc_sig.c,v 1.5.2.1 2003/04/30 20:03:51 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Jun 25 14:10:02 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include "check.h"
#include "test.h"
#include "strcompat.h"
#include "error.h"

#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/ipcll.h"

#include "ut_ipc.h"

/**@name Signal Handling IPC Layer Test Cases
 * @ingroup ut_ipc
 * @{
 */

/** Test pre and post initilization child death. */
static void ipc_sig_chld(void);

/*@}*/


/**@name Child States
 * @{
 */
#define CSTART  0x01
#define CSTOP   0x02
#define CMSG    0x04
#define CQUEUE  0x08
#define CDEATH  0x10
/*@}*/

/** Number of children that we are running with */
#define CHILD_COUNT     5

/** Function to run the child's code */
static void child(int exitpt);

/** debug init */
static void debug_init(void);

#define debug_log_name  "ipc-sig.debug"

void
ipc_sig_suite(Suite *suite)
{
    TCase   *tc;

    tc = tcase_create("IPC Signal Handling");
    suite_add_tcase(suite, tc);

    tcase_add_unchecked_fixture(tc, debug_init, debug_done);

    tcase_add_test(tc, ipc_sig_chld);
}

static void debug_init(void)
{
    if (! DBG_init("i", debug_log_name) ) {
        perror("Error setting up debugging:");
        exit(1);
    }
}

START_TEST(ipc_sig_chld)
{
    pid_t               pids[CHILD_COUNT];
    int                 cMap[CHILD_COUNT];
    int                 childState[CHILD_COUNT];
    int                 i, cidx;
    int                 exitCount;
    struct INW_eInfo    *evt;
    struct message      msg;
    struct IPC_event    ipcevt;
    struct IPC_addr     addr;
    char                failmsg[1024];


    unit_test = FALSE;  /* we can test the real system here */

    DBG_newproc("CTLR");

    ipc_init_ctlr();

    memset(childState, 0, sizeof(childState));
    exitCount = 0;

    for (i = 0; i < CHILD_COUNT; i++) {
        pids[i] = fork();
        if (pids[i] < 0) {
            /* can only occure in the parent */
            fail("fork() failed");
        }

        if (pids[i] == 0) {
            /* child */
            child(i);
        }
        DPRINT(DBG_IPC, "sig_test: forking child %d", pids[i]);
    }

    while (exitCount < CHILD_COUNT) {
        evt = INW_getEvent(waiter);

        if (evt == NULL)
            fail("INW_getEvent() failed");
        if (evt->event_id != INWE_IPC) 
            fail("unexpecetd event");

        if (evt->info == IPCT_EVENT) {
            if (IPC_recvEvent(ctlrHandle, &ipcevt) < 0) {
                snprintf(failmsg, sizeof(failmsg),
                    "IPC_recvEvent(): (%d) %s", 
                    IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));

                DPRINT(DBG_IPC, "FAIL: %s", failmsg);
                fail(failmsg);
            }
            
            cidx = cMap[IPC_PAToIndex(ipcevt.addr.proc)];
            if (ipcevt.type == IPCEVT_IMSTART) {
                for (cidx = 0; cidx < CHILD_COUNT; cidx++) {
                    if (pids[cidx] == ipcevt.pid) {
                        cMap[IPC_PAToIndex(ipcevt.addr.proc)] = cidx;
                        break;
                    }
                }
                childState[cidx] |= CSTART;
            }
            else if (ipcevt.type == IPCEVT_IMSTOP) {
                childState[cidx] |= CSTOP;
                if (cidx >= 2 && (childState[cidx] & CMSG) == 0) {
                    fail("Stop before the message was recieved");
                }
                exitCount++;
            }
            else if (ipcevt.type == IPCEVT_IMDEATH) {
                childState[cidx] |= CDEATH;
                if (cidx >= 2 && (childState[cidx] & CMSG) == 0) {
                    fail("Death before the messages was recieved");
                }
                exitCount++;
            }
            else if (ipcevt.type == IPCEVT_CHLDDEATH) {
                for (cidx = 0; cidx < CHILD_COUNT; cidx++) {
                    if (pids[cidx] == ipcevt.pid) 
                        break;
                }
                childState[cidx] |= CDEATH;
                exitCount++;
            }

        }
        else if (evt->info == IPCT_MSG) {
            if (IPC_cRecvMessage(ctlrHandle, &addr, &msg) < 0) {
                snprintf(failmsg, sizeof(failmsg),
                    "IPC_cRecvMessage(): (%d) %s", 
                    IPC_geterrno(ctlrHandle), IPC_geterrstr(ctlrHandle));
                DPRINT(DBG_IPC, "FAIL: %s", failmsg);
                fail(failmsg);
            }

            cidx = cMap[IPC_PAToIndex(addr.proc)];

            childState[cidx] |= CMSG;
        }
    }

    /* now we need to check all the children */

    fail_unless(childState[0] == CDEATH, 
            "child 1: Wrong state");

    fail_unless(childState[1] == (CSTART | CDEATH), 
            "child 2: Wrong state");

    fail_unless(childState[2] == (CSTART | CDEATH | CMSG), 
            "child 3: Wrong state");

    fail_unless(childState[3] == (CSTART | CDEATH | CMSG), 
            "child 4: Wrong state");

    fail_unless(childState[4] == (CSTART | CSTOP | CMSG),
            "child 5: Wrong state");
}
END_TEST

static void
child(int exitpt)
{
    struct message  msg;

    DBG_newproc("CHLD");

    if (exitpt == 0) exit(0);

    ipc_init_client();

    if (! IPC_inputmoduleStart(cliHandle)  ) {
        DPRINT(DBG_IPC, "input module start failed", NO_VARG);
        exit(1);
    }

    if (exitpt == 1) exit(0);

    memset(&msg, 0, sizeof(msg));

    msg.sev = 1;
    msg.fac = 2;

    msg.hostlen = strlen("empire")+1;
    msg.host = emalloc(msg.hostlen);
    strlcpy(msg.host, "empire", msg.hostlen);

    msg.timestamp = time(NULL);
    msg.tz = 480;

    strlcpy(msg.tag, "test", MSGTAG_LEN);

    msg.msglen = strlen("blah blah test") + 1;
    msg.msg = emalloc(msg.msglen);
    strlcpy(msg.msg, "blah blah test", msg.msglen);

    IPC_imSendMessage(cliHandle, 0, &msg);

    if (exitpt == 2) exit(0);

    sleep(5);   /* give the controller a chance to recieve the message */

    if (exitpt == 3) exit(0);

    ipc_destroy_client();
    
    exit(0);

    /* NOT REACHED */
}
