/**
 * @file ut_waiter.c
 *
 * @brief Input Waiter Tests
 * @ingroup ut_ipc
 *
 * $Id: ut_waiter.c,v 1.9.2.7 2003/04/30 20:03:52 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue May  7 15:42:12 PDT 
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>  /* for NULL */
#include <stdlib.h>
#include <signal.h>

#include "check.h"
#include "error.h"
#include "types.h"

#include "ipc/ipc_private.h"
#include "ipc/ipc.h"
#include "util/list.h"
#include "util/tsqueue.h"

#include "ipc/inputwait.h"

#include "message/message.h"

#include "ut_driver.h"
#include "ut_ipc.h"

#if HAVE_LIBRR
#include <glib/gthread.h>
#endif /* HAVE_LIBRR */

/**@name Input Waiter Initilization Test Cases 
 * @ingroup ut_ipc
 * @{
 */

/** Test creation and deletion */
static void TESTiw_create(void);

/** Test FD addition */
static void TESTiw_add_getAllFD(void);

/** Test IPC addition */
static void iw_add_ipc(void);

/** Test Signal addition */
static void iw_add_signal(void);

/** Test TSQueue addition, and a simple event */
static void iw_add_tsq(void);
/*@}*/

void waiter_suite(Suite *suite)
{
	TCase *tc_io;

	tc_io=tcase_create("IW");
	suite_add_tcase(suite,tc_io);

	tcase_add_test(tc_io,TESTiw_create);
	tcase_add_test(tc_io,TESTiw_add_getAllFD);
    tcase_add_test(tc_io,iw_add_ipc);
    tcase_add_test(tc_io,iw_add_signal);
	tcase_add_test(tc_io,iw_add_tsq);
}

START_TEST(TESTiw_create)
{
	INW *in;
	struct INW_eInfo *add;

	/* create the INW */
    INW_initialize();
	in=INW_create();
	fail_if(NULL==in,"INW_init() returned NULL");
	fail_if(NULL==in->fds,"INW_init() returned object with null list");

	/* create the INW_eCOntain */
	add=emalloc(sizeof(struct INW_eInfo));
	add->info=2;
	add->event_id=2;

	/* add a new element */
	LL_append(in->fds,add);

	/* try clear */
	INW_clearFDs(in);
	fail_unless(0==LL_size(in->fds),"Failed to remove items from list");

	/* finally, delete */
	INW_destroy(&in);
	fail_unless(NULL==in,"Failed to delete");

    INW_shutdown();
}
END_TEST

START_TEST(TESTiw_add_getAllFD)
{
	INW *in;
	struct INW_eInfo *elem;
	int id;

    INW_initialize();
	in=INW_create();

	/* check some good adds */
	id=INW_getNextEventID(in);
	fail_unless(id==INWE_FDSTART,"getEventID() returned incorrect value for first run");
	fail_unless(INW_addFD(in,2,id),"Failed to add event");

	id=INW_getNextEventID(in);
	fail_unless(id==INWE_FDSTART+1,"getEventID() returned incorrect value");
	fail_unless(INW_addFD(in,3,id),"Failed to add event");

	/* check some bad adds */
	fail_if(INW_addFD(in,4,INWE_FDSTART-1),"Added event with bad eventid");

	/* check to make sure our adds work */
	elem=LL_first(in->fds);
	fail_unless(2==elem->info,"Stored wrong INW_eContain");
	fail_unless(id-1==elem->event_id,"Stored wrong INW_eContain");

	elem=LL_last(in->fds);
	fail_unless(3==elem->info,"Stored wrong INW_eContain");
	fail_unless(id==elem->event_id,"Stored wrong INW_eContain");

	INW_destroy(&in);

    INW_shutdown();
}
END_TEST

START_TEST(iw_add_ipc)
{
    INW                 *inw;
    IPC                 ipc;

    INW_initialize();

    inw = INW_create();

    /* I don't care if ipc is not a real ipc handle, just that is not
     * null
     */
    fail_unless(INW_addIPC(inw, &ipc), "INW_addIPC() failed");

    fail_unless(inw->ipc == &ipc, "IPC Handle set wrong");
    fail_unless(inw->ipcEInfo.event_id == INWE_IPC, 
        "ipcEInfo.event_id set wrong");

}
END_TEST

START_TEST(iw_add_signal)
{
    INW                 *inw;
    struct INW_eInfo    *einfo;
    struct sigaction    act;

    INW_initialize();

    inw = INW_create();

    fail_unless(INW_addSignal(inw, SIGTERM), "INW_addSignal(SIGTERM) failed");
    fail_unless(INW_addSignal(inw, SIGHUP), "INW_addSignal(SIGHUP) failed");

    /* now check the list */
    einfo = LL_first(inw->sigs);
    fail_unless(einfo->event_id == INWE_SIG, "first elem event_id wrong");
    fail_unless(einfo->info == SIGTERM, "first elem info wrong");

    einfo = LL_last(inw->sigs);
    fail_unless(einfo->event_id == INWE_SIG, "second elem event_id wrong");
    fail_unless(einfo->info == SIGHUP, "second elem info wrong");

    /* now check the signal handlers */
    fail_unless( sigaction(SIGTERM, NULL, &act) == 0, "sigaction failed");
    fail_unless(act.sa_handler != SIG_DFL, "SIGTERM handler not installed");
    fail_unless(act.sa_handler != SIG_IGN, "SIGTERM handler not installed");
    fail_unless(act.sa_handler != NULL, "SIGTERM handler not installed");

    fail_unless( sigaction(SIGHUP, NULL, &act) == 0, "sigaction failed");
    fail_unless(act.sa_handler != SIG_DFL, "SIGHUP handler not installed");
    fail_unless(act.sa_handler != SIG_IGN, "SIGHUP handler not installed");
    fail_unless(act.sa_handler != NULL, "SIGHUP handler not installed");

}
END_TEST

START_TEST(iw_add_tsq)
{
	INW					*inw;
	struct INW_eInfo	*einfo;

	TSQueue				*queue=NULL;
	struct MSG_qelem	*queue_elem=NULL;

#if HAVE_LIBRR
	g_thread_init(NULL);
#endif /* HAVE_LIBRR */

	INW_initialize();

	inw=INW_create();
	fail_if(NULL==inw,"INW_create() failed");

	/* setup the queue and the msg element */
	queue=TSQ_create();
	fail_if(NULL==queue,"TSQ_create() failed");

	/* try to add the queue to the input waiter */
	fail_unless(INW_addQueue(inw, queue),"INW_addQueue() failed");
	fail_if(NULL==inw->queue,"INW_addQueue() failed to add the TSQueue object");
	fail_unless(INWE_TSQ == inw->qEInfo.event_id, "TSQueue event_id wrong");
	fail_unless(inw->qEInfo.info == (int)queue, "TSQueue elem info wrong");

	/* add an element to the queue */
	queue_elem=emalloc(sizeof(struct MSG_qelem));
	queue_elem->type=MSGQE_MSG;
	queue_elem->chan=12;
	TSQ_enqueue(queue,queue_elem);

	/* wait on the input waiter; it should return immediately, since there's
	 * a queue event on it */
	einfo=INW_getEvent(inw);
	fail_if(einfo==NULL,"INW_getEvent() returned NULL");
	fail_if(einfo->event_id!=INWE_TSQ,
			"event_id of returned event was not INWE_TSQ");
	fail_if(einfo->info!=(int)queue,
			"info of returned event was not the same queue");

	/* cleanup */
	TSQ_destroy(&queue);
	fail_if(NULL!=queue,"TSQ_destroy() failed to cleanup properly");
}
END_TEST
