/**
 * @file ut_waiter_fork.c
 *
 * @brief Input Waiter Forking Tests
 * @ingroup ut_ipc
 *
 * $Id: ut_waiter_fork.c,v 1.6.2.6 2003/04/30 20:03:52 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Mon May 10 15:42:12 PDT 
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>  /* for NULL */
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>

#include "check.h"
#include "error.h"
#include "types.h"
#include "strcompat.h"

#include "ipc/ipc.h"
#include "util/list.h"

#include "ipc/inputwait.h"

#include "ut_driver.h"
#include "ut_ipc.h"

/**@name Input Waiter Test Cases 
 * @ingroup ut_ipc
 * @{
 */

/** Test simple communication (receiver) */
static void TESTiw_comm_test(void);

/** Test simple communication (try to connect) */
static void TESTiw_comm_connect(void);

/** Test Signal Events (receiver) */
static void TESTiw_sig_test(void);

/** Test Signal Events (sender) */
static void TESTiw_sig_sender(void);

/*@}*/

/** A nonce so that TESTiw_sig_sender doesn't treat stale data in the
 * pid file as correct.
 */
static int sig_nonce;

void waiter_fork_suite(Suite *suite)
{
	TCase *tc_io, *tc_sig;

	tc_io=tcase_create("IW_Fork");
	suite_add_tcase(suite,tc_io);

	tcase_set_run_mode(tc_io,CK_PARALLEL);

	tcase_add_test(tc_io,TESTiw_comm_test);
	tcase_add_test(tc_io,TESTiw_comm_connect);

    tc_sig = tcase_create("IW_Signal");
    suite_add_tcase(suite, tc_sig);

    tcase_set_run_mode(tc_sig, CK_PARALLEL);

    tcase_add_test(tc_sig, TESTiw_sig_test);
    tcase_add_test(tc_sig, TESTiw_sig_sender);

    /* comeup with a nonce to use */
    srand( time(NULL));
    sig_nonce = rand();
}


/** The path for our socket */
#define UT_IPC_IW_COMMSOCK_PATH		"ut_ipc_iw_comm.path"

/** How large of a backlog to keep */
#define UT_IPC_IW_BACKLOG			20

/** Test message to receive */
#define	UT_IPC_IW_COMM_MSG			"msg"

/** Length of test message */
#define UT_IPC_IW_COMM_MSG_LEN		4


START_TEST(TESTiw_comm_connect)
{
	struct sockaddr_un addr;
	int sock;

	/* sleep for just a bit so we context switch to the other thread */
	sleep(1);

	/* setup the connection socket */
	sock=socket(PF_UNIX,SOCK_STREAM,0);
	if(sock<0)
	{
		fail("Failed to create a socket");
	}

	addr.sun_family=PF_UNIX;
	strlcpy(addr.sun_path,UT_IPC_IW_COMMSOCK_PATH,UNIX_PATH_MAX);

	if(connect(sock,(struct sockaddr*)&addr,sizeof(addr))<0)
	{
		fail("Could not establish connection on socket");
	}

	/* try to send something over the socket */
	if(-1==send(sock,UT_IPC_IW_COMM_MSG,UT_IPC_IW_COMM_MSG_LEN,0))
	{
		fail("Couldn't send data over the socket");
	}

	/* clean up the socket */
	close(sock);
}
END_TEST

START_TEST(TESTiw_comm_test)
{
	struct sockaddr_un addr;
	int sock;
	int accept_sock;
	int sock_eventid;
	INW *waiter;

	struct INW_eInfo * recvd_event;
	char recvd_msg[UT_IPC_IW_COMM_MSG_LEN];

	/* setup our socket, see if we can get ready to send/recv */
	sock=socket(PF_UNIX,SOCK_STREAM,0);
	if(sock<0)
	{
		fail("Failed to create a socket");
	}

	addr.sun_family=PF_UNIX;
	strlcpy(addr.sun_path, UT_IPC_IW_COMMSOCK_PATH, UNIX_PATH_MAX);

	if( bind(sock,(struct sockaddr*)&addr,sizeof(addr))<0 )
	{
		/* try removing the (presumably stale) socket */
		unlink(UT_IPC_IW_COMMSOCK_PATH);

		/* try binding again */
		if( bind(sock,(struct sockaddr*)&addr,sizeof(addr))<0 )
		{
			fail("Could not bind to socket");
		}
	}

	if( listen(sock,UT_IPC_IW_BACKLOG) )
	{
		fail("Could not listen on socket");
	}

	/* create the input waiter */
    INW_initialize();
	waiter=INW_create();
	sock_eventid=INW_getNextEventID(waiter);
	INW_addFD(waiter,sock,sock_eventid);

	/* wait on our socket until we've got stuff to read */
	recvd_event=INW_getEvent(waiter);

	/* check to make sure we get the right stuff back */
	fail_if(NULL==recvd_event,"INW_getEvent() returned an unexpected failure");
	fail_if(recvd_event->event_id!=sock_eventid,"INW_getEvent() didn't return the expected event ID");

	/* accept this connection */
	if(-1==(accept_sock=accept(sock,NULL,NULL)))
	{
		fail("Could not accept connection");
	}

	/* try to receive over the socket */
	if(-1==recv(accept_sock,recvd_msg,UT_IPC_IW_COMM_MSG_LEN,0))
	{
		fail("Could not receive data over socket");
	}

	fail_unless(0==strncmp(recvd_msg,UT_IPC_IW_COMM_MSG,UT_IPC_IW_COMM_MSG_LEN),
			"Received message not the same as the sent message");

	/* clean up the socket */
	close(sock);
	unlink(UT_IPC_IW_COMMSOCK_PATH);

	INW_destroy(&waiter);

    INW_shutdown();
}
END_TEST

#undef UT_IPC_IW_BACKLOG
#undef UT_IPC_IW_COMMSOCK_PATH

/** PID file for TESTiw_sig_test */
#define SIG_PID_PATH        "test-inw-sig.pid"

#define SIG_EVENTS          15
#define SIG_TERMEVENTS      7
#define SIG_HUPEVENTS       3
#define SIG_USREVENTS       1

START_TEST(TESTiw_sig_test)
{
    pid_t               mypid;
    FILE                *fout;
    INW                 *inw;
    int                 eventCount, termCount, hupCount, usrCount;
    struct INW_eInfo    *evt;
    char                msg[1024];

    /* first setup the signal handlers */
    INW_initialize();
    inw = INW_create();

    fail_unless(INW_addSignal(inw, SIGTERM), "addSignal(SIGTERM) failed");
    fail_unless(INW_addSignal(inw, SIGHUP), "addSignal(SIGHUP) failed");
    fail_unless(INW_addSignal(inw, SIGUSR1), "addSignal(SIGUSR1) failed");

    /* now we need to put our PID in a file */
    fout = fopen(SIG_PID_PATH, "w");
    fail_unless(fout != NULL, "Opening pid file failed");

    mypid = getpid();
    fprintf(fout, "%d %d", sig_nonce, mypid);

    fclose(fout);

    eventCount = termCount = hupCount = usrCount = 0;
    while (termCount < 1 || hupCount < 1 || usrCount < 1) {
        evt = INW_getEvent(inw);

        fail_unless(evt != NULL, "INW_getEvent() failed");
        fail_unless(evt->event_id == INWE_SIG, 
            "Unexpected event from INW_getEvent");

        if (evt->info == SIGTERM)
            termCount++;
        else if (evt->info == SIGHUP)
            hupCount++;
        else if (evt->info == SIGUSR1)
            usrCount++;
        else {
            snprintf(msg, sizeof(msg),
                "Unexpected signal %d", evt->info);
            fail(msg);
        }

        eventCount++;
    }

    /*
    printf("Total events recieved: %d SIGTERM; %d SIGHUP; %d SIGUSR1\n",
            termCount, hupCount, usrCount);
    */

    fail_unless(termCount <= SIG_TERMEVENTS, "Too many SIGTERM events");
    fail_unless(hupCount <= SIG_HUPEVENTS, "Too many SIGHUP events");
    fail_unless(usrCount <= SIG_USREVENTS, "Too many SIGUSR1 events");

    fail_unless(termCount > 0, "Too few SIGTERM events");
    fail_unless(hupCount > 0, "Too few SIGHUP events");
    fail_unless(usrCount > 0, "Too few SIGUSR1 events");


}
END_TEST

START_TEST(TESTiw_sig_sender)
{
    FILE    *fin;
    pid_t   recv;
    bool    gotPID;
    int     sig;
    int     usrCount, hupCount, termCount, sentCount;
    int     nonce;


    gotPID = FALSE;
    while (!gotPID) {
        sleep(1);   /* allow a context switch */

        fin = fopen(SIG_PID_PATH, "r");

        /* fail_unless will just backup the pipe */
        if (fin == NULL) 
            fail("fopen() failed");

        fscanf(fin, "%d %d", &nonce, &recv);
        if (nonce == sig_nonce) 
            gotPID = TRUE;

        fclose(fin);
    }

    sentCount = termCount = hupCount = usrCount = 0;
    while (sentCount < SIG_EVENTS) {

        sig = rand() % 3;

        switch (sig) {
            case 0: 
            {
                if (termCount < SIG_TERMEVENTS) {
                    kill(recv, SIGTERM);
                    termCount++;
                    break;
                }
            }
            case 1: 
            {
                if (hupCount < SIG_HUPEVENTS) {
                    kill(recv, SIGHUP);
                    hupCount++;
                    break;
                }
            }
            case 2:
            {
                if (usrCount < SIG_USREVENTS) {
                    kill(recv, SIGUSR1);
                    usrCount++;
                    break;
                }
            }
        }
        if ( (rand() % 3) == 0) {
            sleep(1);
        }
        sentCount++;
    }
}
END_TEST







