/**
 * @file ut_msg_frmt_cooked.c
 *
 * @brief Test cases for the COOKED formatter.
 *
 * $Id: ut_msg_frmt_cooked.c,v 1.2.2.6 2003/04/30 20:03:54 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Aug 28 13:56:08 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>

#include "check.h"

#include "debug.h"
#include "util/netaddr.h"
#include "util/bitvec.h"

#include "message/message.h"
#include "message/msg_frmt.h"

/**@name Message Formatter COOKED Test Cases
 * @ingroup ut_msg_frmt
 * @{
 */

/** Test COOKED message formattinrg (<entry>) */
static void msg_frmt_cooked_msg(void);

/** Test COOKED path formatting (<path>) */
static void msg_frmt_cooked_path(void);

/*@}*/

static void debug_setup(void);
static void debug_teardown(void);
static void init_msgs(void);
static void destroy_msgs(void);

void msg_frmt_cooked_suite(Suite *suite)
{
    TCase   *tc;

    tc = tcase_create("MFormatting_COOKED");
    suite_add_tcase(suite, tc);

    tcase_add_unchecked_fixture(tc, debug_setup, debug_teardown);
    tcase_add_unchecked_fixture(tc, init_msgs, destroy_msgs);

    tcase_add_test(tc, msg_frmt_cooked_msg);
    tcase_add_test(tc, msg_frmt_cooked_path);
}

#define TEST_MSG_COUNT  3
#define TEST_PATH_COUNT 1
static struct message msgObjs[] = {
    {5, 24, 5, "jack", 1010006535, -420, "imapd", 37,
        "Replacement device found in nostril.", 0, 0, NULL, 17,
        "jack.example.net"},
    {5, 24, 5, "jack", 946848135, -180, "imapd", 37,
        "Replacement device found in nostril.", 0, 0, NULL, 17,
        "jack.example.net"},
    {5, 24, 5, "jack", 631315335, -420, "imapd", 38,
        "\nReplacement device found in nostril.", 173, 0, NULL, 17,
        "jack.example.net"}
};

static struct path pathObjs[] = {
    {173, 0, 26, "lowry.records.example.com", NULL,
        29, "kurtzman.records.example.com", NULL,
        (BM_MBIT(VALID_USER) | BM_MBIT(INTEGRETY) 
         | BM_MBIT(REPLAY_PROT) | BM_MBIT(RELIABLE)), NULL},
    /* Start Sub Path objects */
    {24, 0, 33, "screen.lowry.records.example.com", NULL,
        26, "lowry.records.example.com", NULL,
        (BM_MBIT(FROM_DEVICE) | BM_MBIT(INTEGRETY) | BM_MBIT(RELIABLE)), NULL}
};

char *msgStrs[] = {
    "<entry facility='24' severity='5' tag='imapd' hostname='jack' "
        "timestamp='Jan  2 13:22:15' deviceFQDN='jack.example.net' "
        "deviceIP='10.0.0.1'>Replacement device found in nostril.</entry>",
    "<entry facility='24' severity='5' tag='imapd' hostname='jack' "
        "timestamp='Jan  2 13:22:15' "
        "deviceFQDN='jack.example.net' deviceIP='10.0.0.4'>"
        "Replacement device found in nostril.</entry>",
    "<entry facility='24' severity='5' tag='imapd' hostname='jack' "
        "timestamp='Jan  2 13:22:15' "
        "deviceFQDN='jack.example.net' deviceIP='10.0.0.1' "
        "pathID='173'>\n"
        "Replacement device found in nostril.</entry>",
};

char *pathStrs[] = {
    "<path pathID='173' fromFQDN='lowry.records.example.com'"
        " fromIP='10.0.0.50' toFQDN='kurtzman.records.example.com'"
        " toIP='10.0.0.51' linkprops='URIL'>" 
        "<path pathID='24'"
        " fromFQDN='screen.lowry.records.example.com' fromIP='10.0.0.47'"
        " toFQDN='lowry.records.example.com' toIP='10.0.0.50'"
        " linkprops='ILD'></path></path>"
};

static struct MFRMT_opt fopt = 
    {MFRMT_COOKED, 0};

static void init_msgs(void)
{

    pathObjs[0].fromIP = NA_newINetLookup("10.0.0.50");
    pathObjs[0].toIP = NA_newINetLookup("10.0.0.51");
    pathObjs[0].next = &(pathObjs[1]);

    pathObjs[1].fromIP = NA_newINetLookup("10.0.0.47");
    pathObjs[1].toIP = NA_newINetLookup("10.0.0.50");

    msgObjs[0].dip = NA_newINetLookup("10.0.0.1");
    msgObjs[1].dip = NA_newINetLookup("10.0.0.4");
    msgObjs[2].dip = NA_newINetLookup("10.0.0.1");
}

static void destroy_msgs(void) {}

static void debug_setup(void)
{
    DBG_init("m", "msg-frmt-cooked.debug");
}
static void debug_teardown(void)
{
    DPRINT(DBG_DBG, "TEST CASE DONE", NO_VARG);
}

START_TEST(msg_frmt_cooked_msg)
{
    char    buf[4096];
    int     rtn;
    int     i;


    for (i = 0; i < TEST_MSG_COUNT; i++) {
        /* One big string of 'A' */
        memset(buf, 'A', sizeof(buf));
        buf[sizeof(buf) - 1] = '\0';

        rtn = MFRMT_format(buf, sizeof(buf), &(msgObjs[i]), &fopt);
        if (rtn <= 0) {
            fail("Error Formatting Message");
        }

        if (rtn != (strlen(buf) + 1) ) {
            DPRINT(DBG_MSG, "rtn = %d; strlen(buf) = %d", rtn, strlen(buf));
            DPRINT(DBG_MSG, "buf = '%s'", buf);
            fail("Formatter return doesn't match buffer space used");
        }

        if (rtn != (strlen(msgStrs[i]) + 1) ) {
            DPRINT(DBG_MSG, "formatted message length = %d", rtn);
            DPRINT(DBG_MSG, "formatted message '%s'", buf);
            fail("Expected Message Length doesn't match formatted length");
        }

        if (strcmp(msgStrs[i], buf) != 0) {
            DPRINT(DBG_MSG, "formatted message '%s'", buf);
            fail("Expecetd message doesn't match formatted message");
        }

        DPRINT(DBG_MSG, "strlen(buf) = %d", strlen(buf));
    }

}
END_TEST

START_TEST(msg_frmt_cooked_path)
{
    char    buf[4096];
    int     rtn;
    int     i;

    for (i = 0; i < TEST_PATH_COUNT; i++) {
        memset(buf, 'A', sizeof(buf));
        buf[sizeof(buf) -1] = '\0';

        rtn = MFRMT_formatPath(buf, sizeof(buf), &(pathObjs[i]), &fopt);
        if (rtn <= 0) {
            fail("Error formatting path");
        }

        if (rtn != (strlen(buf) + 1) ) {
            DPRINT(DBG_MSG, "return %d, buf len %d", rtn, strlen(buf) + 1);
            DPRINT(DBG_MSG, "formatted path '%s'", buf);
            fail("Formatter return doesn't match buffer space used");
        }

        if (rtn != (strlen(pathStrs[i]) + 1) ) {
            DPRINT(DBG_MSG, "expected len %d, return %d", 
                    strlen(pathStrs[i]) + 1, rtn);
            DPRINT(DBG_MSG, "formatted path '%s'", buf);
            fail("Expected path length doesn't match formatted length");
        }

        if (strcmp(pathStrs[i], buf) != 0) {
            DPRINT(DBG_MSG, "formatted path '%s'", buf);
            fail("Expected path doesn't match formatted path");
        }

    }
}
END_TEST
