/**
 * @file ut_msg_frmt_read.c
 *
 * @brief 
 * @par
 *
 *
 * $Id: ut_msg_frmt_read.c,v 1.4.2.6 2003/04/30 20:03:54 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Aug 22 16:33:10 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>

#include "error.h"
#include "util/bitvec.h"

#include "check.h"
#include "message/message.h"
#include "message/msg_parse_rfc3164.h"
#include "message/msg_frmt.h"

/** The number and size of test elements */
#define MAXLEN		2048
#define	NUMTESTS	3
#define NUMOVERFLOW 4

/**@name Message Formatter Readable Test Cases
 * @ingroup ut_msg_frmt
 * @{
 */

/** Test readable message formatting */
static void TESTmsg_frmt_read(void);

/** Test readable message formatting overflow */
static void TESTmsg_frmt_read_ovrflw(void);

/*@}*/

/* test suite setup */
void msg_frmt_read_suite(Suite *suite)
{
	TCase *tc_io;

	tc_io = tcase_create("MFormatting_Readable");
	suite_add_tcase(suite,tc_io);

	tcase_add_test(tc_io,TESTmsg_frmt_read);
	tcase_add_test(tc_io,TESTmsg_frmt_read_ovrflw);
}

/* Use a full and complete ts */
static struct MFRMT_opt fopt_fts = 
    {MFRMT_RFC3164, BM_MBIT(MFOPT_RFC3339)}; 

/* Use the incomplete rfc3164 ts */
static struct MFRMT_opt fopt_its = 
    {MFRMT_RFC3164, 0};

START_TEST(TESTmsg_frmt_read)
{
	char ins_fts[NUMTESTS][MAXLEN]={
		{ "1980-10-11T23:14:15-03:50 asdf", },
		{ "1980-10-11T23:14:15-03:50 asdf su", },
		{ "1980-10-11T23:14:15+04:00 asdf inetd[76]: inetd service failed to read inetd.conf", },
	};
    char ins_its[NUMTESTS][MAXLEN]={
        { "Oct 11 23:14:15 asdf"},
        { "Oct 11 23:14:15 asdf su"},
        { "Oct 11 23:14:15 asdf inetd[76]: inetd service failed to read inetd.conf"},
    };
	struct message checkmsgs[NUMTESTS]={
		/*S, F, HL, H, T, TZ, TA, ML, M */
		{ 0, 9, 5, "asdf", 340179255, -230, "", 0, NULL, },
		{ 0, 9, 5, "asdf", 340179255, -230, "su", 0, NULL, },
		{ 0, 9, 5, "asdf", 340179255, 240, "inetd", -1, "[76]: inetd service failed to read inetd.conf"},
	};
	char failmsg[MAXLEN];
    char retmsg[MAXLEN];
	int i=0;

	/* add some strlen things that's easier to do here */
	checkmsgs[2].msglen=strlen(checkmsgs[2].msg)+1;

	/* first, check string comparisons */
	for(i=0;i<NUMTESTS;i++)
	{
		MFRMT_format(retmsg,MAXLEN,&(checkmsgs[i]),&fopt_fts);
		snprintf(failmsg,MAXLEN,"ins_fts[%d] != retmsg (%s)",i,retmsg);
		fail_unless(0==strncmp(retmsg,ins_fts[i],MAXLEN),failmsg);

        MFRMT_format(retmsg, MAXLEN, &(checkmsgs[i]), &fopt_its);
        snprintf(failmsg,MAXLEN, "ins_its[%d] != retmsg (%s)",i, retmsg);
        fail_unless(strncmp(retmsg, ins_its[i], MAXLEN) == 0, failmsg);
	}
}
END_TEST

START_TEST(TESTmsg_frmt_read_ovrflw)
{
	struct message overflowmsg = 
		{ 0, 9, 5, "asdf", 340179255, 240, "ircd", -1, ": clock is running backwards at an incredible rate", };

	char overflowed_fts[NUMOVERFLOW][MAXLEN]={
		{"1980-10-11T23:14:15"},
		{"1980-10-11T23:14:15+04:00 asdf ircd: ",},         /* 38 */
		{"1980-10-11T23:14:15+04:00 asdf ircd: clock",},	/* 43 */
		{"1980-10-11T23:14:15+04:00 asdf ircd: clock is ",},/* 47 */
	};
    char overflowed_its[NUMOVERFLOW][MAXLEN] = {
        {"Oct 11 23:14:15 asd"},                            
        {"Oct 11 23:14:15 asdf ircd: clock is r"},          /* 38 */
        {"Oct 11 23:14:15 asdf ircd: clock is runnin"},     /* 43 */
        {"Oct 11 23:14:15 asdf ircd: clock is running ba"}, /* 47 */
    };
	char retmsg[MAXLEN];
	int retlen;
	int curtest=0;
	int compared;
    int msglen;
    char buf[1024];

	/* add some strlen that's easier to have it do */
	overflowmsg.msglen=strlen(overflowmsg.msg)+1;

	/* the first overflow test */
	curtest=0;
    msglen = 20;
	retlen=MFRMT_format(retmsg,msglen,&overflowmsg,&fopt_its);
    /* check our results */
	if (0 != retlen) {
        snprintf(buf, sizeof(buf),
            "MFRMT_format() returned an incorrect # of chars (%d)", retlen);
        fail(buf);
    }
    retlen = MFRMT_format(retmsg,msglen, &overflowmsg, &fopt_fts);
    if (0 != retlen) {
        snprintf(buf, sizeof(buf),
            "MFRMT_format() returned an incorrect # of chars (%d)", retlen);
        fail(buf);
    }

	/* and the second, with extra tests for OOB writing */
    for (curtest = 1; curtest < NUMOVERFLOW; curtest++) {
        msglen = strlen(overflowed_fts[curtest])+1;

        retmsg[msglen]='a';
        retlen=MFRMT_format(retmsg,msglen,&overflowmsg,&fopt_its);
        if (msglen != retlen) {
            snprintf(buf, sizeof(buf),
                "MFRMT_format() (its) returned an incorrect # of chars (%d)", 
                retlen);
            fail(buf);
        }
        compared=strncmp(retmsg,overflowed_its[curtest],msglen);
        fail_unless(0==compared,
            "MFRMT_format() (its) formatted something unexpected");
        fail_unless('a'==retmsg[msglen],
            "MFRMT_format() (its) wrote out of bounds");

        retmsg[msglen]='a';
        retlen = MFRMT_format(retmsg, msglen, &overflowmsg, &fopt_fts);
        if (msglen != retlen) {
            snprintf(buf, sizeof(buf),
                "MFRMT_format() (fts) returned an incorrect # of chars (%d)", 
                retlen);
            fail(buf);
        }
        compared=strncmp(retmsg,overflowed_fts[curtest],msglen);
        fail_unless(0==compared,
            "MFRMT_format() (fts) formatted something unexpected");
        fail_unless('a'==retmsg[msglen],
                "MFRMT_format() (its) wrote out of bounds");
            
    }
}
END_TEST
