/**
 * @file ut_msg_parser_cooked.c
 *
 * @brief Test cases for the SYSLOG-COOKED message parser.
 *
 * $Id: ut_msg_parser_cooked.c,v 1.1.2.7 2003/04/30 20:03:54 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Aug 14 15:44:18 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "check.h"


#include "message/msg_parse_cooked.h"
#include "message/message.h"
#include "message/msg_const.h"

#include "debug.h"
#include "error.h"

#include "common_funcs.h"

#include "util/bitvec.h"

/**@name SYSLOG-COOKED Message Parser Test Cases
 * @ingroup ut_msg_parse
 * @{
 */

/** Test cooked parser for <entry> elements */
static void msg_parse_cooked_entry(void);

/** Test cooked parser for <iam> elements */
static void msg_parse_cooked_iam(void);

/** Test cooked parser for <path> elements */
static void msg_parse_cooked_path(void);

/** Test cooked parser free routine for segfaults */
static void msg_parse_cooked_free(void);

/*@}*/

static void debug_init(void)
{
    /* setup debugging */
    if (! DBG_init("m", "msg-parse-cooked.debug") ) {
        perror("Error setting up debugging:");
        exit(1);
    }
    DPRINT(DBG_DBG, "======= Test Case Start =======", NO_VARG);
}

static void debug_finish(void)
{
    DPRINT(DBG_DBG, "======= Test Case Done =======", NO_VARG);
}


void msg_parse_cooked_suite(Suite *suite)
{
    TCase   *tc;

    tc = tcase_create("MParse_Cooked");
    suite_add_tcase(suite, tc);

    tcase_add_unchecked_fixture(tc, debug_init, debug_finish);

    tcase_add_test(tc, msg_parse_cooked_entry);
    tcase_add_test(tc, msg_parse_cooked_iam);
    tcase_add_test(tc, msg_parse_cooked_path);
	tcase_add_test(tc, msg_parse_cooked_free);
}

START_TEST(msg_parse_cooked_entry)
{
#define NUMTESTS    3
    struct MPCOOKED_ctx ctx;
    int                 i;
    struct tm           *ltm;
    time_t              tt;

    char *xmlStrs[NUMTESTS] = {
        "<entry facility='24' severity='5' timestamp='Jan  2 13:22:15'\n"
            "deviceFQDN='jack.example.net' deviceIP='10.0.0.1'\n"
			"hostname='localhost'\n"
            "tag='imapd'>Replacement device found in nostril.</entry>",
        "<entry facility='24' severity='5'\n"
            "timestamp='2000-01-02T13:22:15-03:00'\n"
            "deviceFQDN='jack.example.net' deviceIP='10.0.0.4'\n"
			"hostname='buckaroo'\n"
            "tag='imapd'>Replacement device found in nostril.</entry>",
        "<entry facility='24' severity='5'\n"
            "timestamp='1990-01-02T13:22:15-00:00'\n"
            "deviceFQDN='jack.example.net' deviceIP='10.0.0.1'\n"
			"hostname='www'\n"
            "tag='imapd' pathID='173'>"
            "Replacement device found in nostril.</entry>",
    };
    struct message msgObjs[NUMTESTS] = {
        {5, 24, 10, "localhost", 1010006535, MSG_TZ_UNKNOWN, "imapd", 37,
            "Replacement device found in nostril.", 0, 0, NULL, 17,
            "jack.example.net"},
        {5, 24, 9, "buckaroo", 946848135, -180, "imapd", 37,
            "Replacement device found in nostril.", 0, 0, NULL, 17,
            "jack.example.net"},
        {5, 24, 4, "www", 631315335, MSG_TZ_UNKNOWN, "imapd", 37,
            "Replacement device found in nostril.", 173, 0, NULL, 17,
            "jack.example.net"},
    };

    /* finish filling in msgObjs */
    msgObjs[0].dip = NA_newINetLookup("10.0.0.1");
    msgObjs[1].dip = NA_newINetLookup("10.0.0.4");
    msgObjs[2].dip = NA_newINetLookup("10.0.0.1");

    /* Fix up the time stamp in the yearless entry */
    tt = time(NULL);
    ltm = localtime(&tt);
    ltm->tm_mon = 0;
    ltm->tm_mday = 2;
    ltm->tm_hour = 13;
    ltm->tm_min = 22;
    ltm->tm_sec = 15;
    ltm->tm_isdst = -1;  /* Hopefully portable */
    msgObjs[0].timestamp = mktime(ltm);

    fail_unless(MPCOOKED_init(), "init failed");

    for (i = 0; i < NUMTESTS; i++) {
		DPRINT(DBG_MSG, "test: Checking message: %s", xmlStrs[i]);

        if (! MPCOOKED_parseString(&ctx, xmlStrs[i], strlen(xmlStrs[i]))) {
            fail("MPCOOKED_parseString() failed");
        }
        DPRINT(DBG_MSG, "test: got context type: %d", ctx.type);
        fail_unless(ctx.type == MPC_ENTRY, "bad message type");

        DPRINT(DBG_MSG, "test: tz check (%d) %-d == %-d", 
                i, ctx.data.message.tz, msgObjs[i].tz);
    
        /* This will fail() on error */
        CMP_msg(&(ctx.data.message), &(msgObjs[i]));

		/* check MPCOOKED_free() */
		MPCOOKED_free(&ctx);
		fail_unless(NULL==ctx.data.message.msg,
				"MPCOOKED_free() failed to set ctx.data.message to NULL");
    }
#undef NUMTESTS
}
END_TEST

START_TEST(msg_parse_cooked_iam)
{
#define NUMTESTS    1
    struct MPCOOKED_ctx     ctx;
    int                     num;
    struct MPCOOKED_iam     *l;
    struct MPCOOKED_iam     *r;
    char *xmlStrs[NUMTESTS] = {
        "<iam fqdn='lowry.example.com' ip='10.0.0.27' type='device'/>"
    };
    struct MPCOOKED_iam     iamObjs[NUMTESTS] = {
        {"lowry.example.com", NULL, IAM_DEVICE}
    };
	char errstr[1024];

    iamObjs[0].addr = NA_newINetLookup("10.0.0.27");

    fail_unless(MPCOOKED_init(), "init failed");

    for (num = 0; num < NUMTESTS; num++) {
        if (! MPCOOKED_parseString(&ctx, xmlStrs[num], strlen(xmlStrs[num]))) {
            fail("MPCOOKED_parseString() failed");
        }

        fail_unless(ctx.type == MPC_IAM, "Bad message type");
        l = &(ctx.data.iam);
        r = &(iamObjs[num]);

		if(0!=strcmp(l->fqdn,r->fqdn))
		{
			snprintf(errstr,1024,"fqdn not equal from iam statement ('%s' != '%s')",
					l->fqdn,r->fqdn);
			fail(errstr);
		}

		/* check MPCOOKED_ctx() free */
		MPCOOKED_free(&ctx);
		fail_unless(NULL==ctx.data.iam.fqdn,
				"MPCOOKED_fail() failed to set ctx.data.iam.fqdn to NULL");
    }
#undef NUMTESTS
}
END_TEST

START_TEST(msg_parse_cooked_path)
{
#define NUMTESTS 1
    struct MPCOOKED_ctx ctx;
    int                 num;
    struct path         *l;
    struct path         *r;
    char *xmlStrs[NUMTESTS] = {
        "<path fromFQDN='lowry.records.example.com'\n"
            "fromIP='10.0.0.50' toFQDN='kurtzman.records.example.com'\n"
            "toIP='10.0.0.51' linkprops='ULRI' pathId='173'>\n"
            "<path \n"
            "fromFQDN='screen.lowry.records.example.com' fromIP='10.0.0.47'\n"
            "toFQDN='lowry.records.example.com' toIP='10.0.0.50'\n"
            "linkprops='DLI' pathID='24'></path></path>"
    };
    struct path pathObjs[] = {
        {173, 0, 26, "lowry.records.example.com", NULL, 
            29, "kurtzman.records.example.com", NULL,
            (BM_MBIT(VALID_USER) | BM_MBIT(INTEGRETY) 
             | BM_MBIT(REPLAY_PROT) | BM_MBIT(RELIABLE)), NULL},
        /* Start Sub Path objects */
        {24, 0, 33, "screen.lowry.records.example.com", NULL,
            26, "lowry.records.example.com", NULL,
            (BM_MBIT(FROM_DEVICE) | BM_MBIT(INTEGRETY) | BM_MBIT(RELIABLE)), 
            NULL}
    };

    pathObjs[0].fromIP = NA_newINetLookup("10.0.0.50");
    pathObjs[0].toIP = NA_newINetLookup("10.0.0.51");
    pathObjs[0].next = &(pathObjs[1]);

    pathObjs[1].fromIP = NA_newINetLookup("10.0.0.47");
    pathObjs[1].toIP = NA_newINetLookup("10.0.0.50");

    fail_unless(MPCOOKED_init(), "init failed");

    for (num = 0; num < NUMTESTS; num++) {
        if (! MPCOOKED_parseString(&ctx, xmlStrs[num], strlen(xmlStrs[num]))) {
            fail("MPCOOKED_parseString() failed");
        }

        fail_unless(ctx.type == MPC_PATH, "Bad message type");

        l = &(ctx.data.path);
        r = &(pathObjs[num]);

        while (l != NULL) {
            if (r == NULL) 
                fail("Parsed path is too short");

			CMP_path(l,r);

            r = r->next;
            l = l->next;
        }
        fail_unless(r == NULL, "Parsed path is too long");

		/* check MPCOOKED_free() */
		MPCOOKED_free(&ctx);
		fail_unless(ctx.data.path.fromFQDN==NULL,
				"MPCOOKED_free() failed to set ctx.data.path.fromFQDN to NULL");
		fail_unless(ctx.data.path.toFQDN==NULL,
				"MPCOOKED_free() failed to set ctx.data.path.toFQDN to NULL");
		fail_unless(ctx.data.path.next==NULL,
				"MPCOOKED_free() failed to set ctx.data.path.next to NULL");
    }
#undef NUMTESTS
}
END_TEST

START_TEST(msg_parse_cooked_free)
{
	struct MPCOOKED_ctx ctx;

	memset(&ctx,0,sizeof(struct MPCOOKED_ctx));

	/* test path */
	ctx.type=MPC_PATH;
	ctx.data.path.fromFQDN=emalloc(strlen("hello")+1);
	strcpy(ctx.data.path.fromFQDN,"hello");

	mark_point();
	MPCOOKED_free(&ctx);

	/* test entry */
	ctx.type=MPC_ENTRY;

	mark_point();
	MPCOOKED_free(&ctx);

	/* test iam */
	ctx.type=MPC_IAM;

	mark_point();
	MPCOOKED_free(&ctx);

	/* test weird entry */
	ctx.type=MPC_NONE;

	mark_point();
	MPCOOKED_free(&ctx);
}
END_TEST
