/**
 * @file ut_omod_beep.c
 *
 * @brief BEEP output test suite.
 *
 * $Id: ut_omod_beep.c,v 1.6.2.7 2003/10/22 04:09:04 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Tue Aug 27 18:35:02 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "check.h"
#include "ut_omod_base.h"

#include "error.h"
#include "debug.h"
#include "log.h"
#include "strcompat.h"

#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "ipc/inputwait.h"

#include "util/bitvec.h"

#include "message/message.h"
#include "message/msg_parse_rfc3164.h"

#include "common_funcs.h"

/**@defgroup ut_omod_beep BEEP Output Module Test Cases
 * @ingroup ut_omod
 * @{
 */

/** Listener used in the test */
static void omod_beep_listener(void);

/** Initiater test (uses the OMOD interface); sends raw packets */
static void omod_beepraw_initiator(void);

/** Initiator test (uses the OMOD interface); sends cooked packets */
static void omod_beepcook_initiator(void);

/** Controller (receives and verifies messages from the listener */
static void omod_beep_controller(void);

/*@}*/

/**@defgroup ut_omod_beep_impl BEEP Output Module Test Cases Implementation
 * @ingroup ut_omod_beep
 * @{
 */

static void
debug_init(void)
{
	DBG_init("PoBm", "omod-beep.debug");
}

static void
debug_done(void)
{
	DPRINT(DBG_DBG, "Test Case Done", NO_VARG);
}

void
omod_beep_suite(Suite *suite)
{
	TCase *tc;

	tc = tcase_create("OMOD_BEEP");
	suite_add_tcase(suite,tc);

	tcase_add_unchecked_fixture(tc, debug_init, debug_done);

	tcase_set_run_mode(tc, CK_PARALLEL);
	tcase_add_test(tc, omod_beep_controller);
	tcase_add_test(tc, omod_beep_listener);

	tcase_add_test(tc, omod_beepraw_initiator);
	tcase_add_test(tc, omod_beepcook_initiator);
}

/** The port for the listener to listen on */
#define	BEEP_PORT		5000

/** The number of messages to send */
#define BEEP_NUMMSG		5

/** The number of raw senders to create */
#define BEEPRAW_NUM_INITIATOR	1

/** The number of cooked senders to create */
#define	BEEPCOOK_NUM_INITIATOR	1

/** The maximum length of our messages */
#define MSG_LEN			512

/** The path to the controller socket */
#define BEEP_IPCPATH	"omod-beep.ipcsock"

/** Initialize the message with some sane values
 *
 * @param msg The message to initialize
 */
static void
initMsg(struct message *msg)
{
	memset(msg,0,sizeof(struct message));
	msg->sev = 4;

	msg->hostlen = strlen("localhost")+1;
	msg->host = emalloc(msg->hostlen);
	strlcpy(msg->host, "localhost", msg->hostlen);

	msg->timestamp = time(NULL);
	msg->tz = -700;
	strlcpy(msg->tag, "udptest", MSGTAG_LEN);
	msg->msglen = strlen(": Test message 0 from initiator 0")+1;
	msg->msg = emalloc(msg->msglen);
}

/** Keep track of the paths that we are expecting */
struct path     *pathList[BEEP_NUMMSG][BEEPCOOK_NUM_INITIATOR];

/** Add a few starting values to the output module's "controller" path cache
 *
 * @param pcache The path cache to add to
 * @param omID The output module ID to use
 * @param numInitiator The number of initiators
 */
static void
initPCache(PCache *pcache, int omID,int numInitiator)
{
	int i,j;
	struct path *tpath;
    struct path *opath; /* outter path */
    struct hostent  *phe;
    char            host[HOST_NAME_MAX];

    opath = emalloc(sizeof(struct path));
    opath->id = 4;
    opath->connid = 0;

    opath->tfqdnlen = strlen("127.0.0.1")+1;
    opath->toFQDN = emalloc(opath->tfqdnlen);
    strlcpy(opath->toFQDN, "127.0.0.1", opath->tfqdnlen);
    opath->toIP = NA_newINetLookup(opath->toFQDN);

    gethostname(host, HOST_NAME_MAX);
    phe = gethostbyname(host);
    DPRINT(DBG_OMD, "gethostname() = %s", host);
    DPRINT(DBG_OMD, "gethostbyname(%s)->h_name = %s", host, phe->h_name);

    opath->ffqdnlen = strlen(phe->h_name)+1;
    opath->fromFQDN = emalloc(opath->ffqdnlen);
    strlcpy(opath->fromFQDN, phe->h_name, opath->ffqdnlen);
    opath->fromIP = NA_newINetLookup(opath->fromFQDN);

	for(j=0;j<numInitiator;j++)
	{
		for(i=0;i<BEEP_NUMMSG;i++)
		{
			tpath=emalloc(sizeof(struct path));
			memset(tpath,0,sizeof(struct path));

			tpath->ffqdnlen=strlen("host##,##")+1;
			tpath->fromFQDN=emalloc(tpath->ffqdnlen);
			snprintf(tpath->fromFQDN, tpath->ffqdnlen, "host%2d,%2d",
					j,i+1);
			tpath->fromIP=NA_newINetLookup("10.0.1.1");

			tpath->tfqdnlen=strlen("tohost##,##")+1;
			tpath->toFQDN=emalloc(tpath->tfqdnlen);
			snprintf(tpath->toFQDN, tpath->tfqdnlen, "tohost%2d,%2d",
					j,i+1);
			tpath->toIP=NA_newINetLookup("10.0.2.1");

			tpath->props=BM_MBIT(VALID_USER);

			tpath->id=i+1;
			tpath->connid=j;

			DPRINT(DBG_DBG, "Adding path w/ connid: %d, pathid: %d, omID: %d",
					tpath->connid, tpath->id, omID);

            if (pcache != NULL) {
			    PCACHE_addPath(pcache, tpath, TRUE);
            }
            else {
                opath->next = tpath;
                tpath = emalloc(sizeof(struct path));
                *tpath = *opath;
                tpath->id = i+1;
                tpath->connid = j;
                pathList[i][j] = tpath;

                DPRINT(DBG_OMD, "pathList[%d][%d].fromFQDN='%s'", 
                        i, j, tpath->fromFQDN);
            }
		}
	}
}

/*@}*/

START_TEST(omod_beepraw_initiator)
{
	struct OMOD_opt		opts;
	struct in_addr		localhost;
	OMOD_interface		*oms[BEEPRAW_NUM_INITIATOR];
	struct message		sendMsg;
	int					i,j;

	DBG_newproc("RawInitiator");

	/* make sure that the listener is set up before we try to connect to it */
	sleep(3);

	/* fill in omod options */
	opts.format.type = MFRMT_RFC3164;
    opts.format.options = 
        BM_MBIT(MFOPT_PRI) | BM_MBIT(MFOPT_RFC3339) | BM_MBIT(MFOPT_MIMEHEAD); 
	opts.outPath = "localhost";
	opts.pcache = NULL;
	fail_unless(inet_aton("127.0.0.1", &localhost), "inet_aton()");
	opts.outAddr = NA_newINet4(localhost);
	opts.outPort = BEEP_PORT;
	opts.pcache=PCACHE_ctlrCreate();

    fail_unless(OMOD_processOptions(OMOD_BEEP, &opts, "(none)", 0),
            "Invalid Options");
    DPRINT(DBG_OMD, "Format options = 0x%x", opts.format.options);

	/* create the message */
	initMsg(&sendMsg);

	DPRINT(DBG_OMD, "Starting initiator initialization", NO_VARG);

	for(j =0; j<BEEPRAW_NUM_INITIATOR; j++)
	{
		oms[j] = OMOD_getInterface(OMOD_BEEP);
		oms[j]->state = oms[j]->INIT(&opts);
		if(NULL==oms[j]->state)
			fail("OM INIT() failed");

		initPCache(oms[j]->state->pcache, oms[j]->state->pcache_omID,
				BEEPRAW_NUM_INITIATOR);
	}

	DPRINT(DBG_OMD, "Done w/ initiator initialization", NO_VARG);

	for ( i=0; i<BEEP_NUMMSG; i++)
	{
		for( j=0; j<BEEPRAW_NUM_INITIATOR; j++)
		{
			/* send the message */
			snprintf(sendMsg.msg, sendMsg.msglen,
					": Test message %d from initiator %d", i, j);
			sendMsg.pathid=i;
			sendMsg.connid=j;

			oms[j]->SEND_MSG(oms[j]->state, &sendMsg);
		}
	}

	DPRINT(DBG_OMD, "Done sending messages, close everything", NO_VARG);

	for (j=0; j<BEEPRAW_NUM_INITIATOR; j++)
		oms[j]->DESTROY(oms[j]->state);

	DPRINT(DBG_DBG, "BEEP OM Initiator Done",NO_VARG);
}
END_TEST

START_TEST(omod_beepcook_initiator)
{
	struct OMOD_opt		opts;
	struct in_addr		localhost;
	OMOD_interface		*oms[BEEPCOOK_NUM_INITIATOR];
	struct message		sendMsg;
	int					i,j;

	DBG_newproc("CookInitiator");

	/* make sure that the listener is set up before we try to connect to it */
	sleep(3);

	/* fill in omod options */
	opts.format.type = MFRMT_COOKED;
    opts.format.options = 
        BM_MBIT(MFOPT_PRI) | BM_MBIT(MFOPT_RFC3339); 
	opts.outPath = "localhost";
	opts.pcache = NULL;
	fail_unless(inet_aton("127.0.0.1", &localhost), "inet_aton()");
	opts.outAddr = NA_newINet4(localhost);
	opts.outPort = BEEP_PORT;
	opts.pcache=PCACHE_ctlrCreate();

    /* finish setting up options */
    fail_unless(OMOD_processOptions(OMOD_BEEP, &opts, "(none)", 0), 
            "Invalid options");

	/* create the message */
	initMsg(&sendMsg);

	DPRINT(DBG_OMD, "Starting initiator initialization", NO_VARG);

	for(j =0; j<BEEPCOOK_NUM_INITIATOR; j++)
	{
		oms[j] = OMOD_getInterface(OMOD_BEEP);
		oms[j]->state = oms[j]->INIT(&opts);
		if(NULL==oms[j]->state)
			fail("OM INIT() failed");

		initPCache(oms[j]->state->pcache, oms[j]->state->pcache_omID,
				BEEPCOOK_NUM_INITIATOR);
	}

	DPRINT(DBG_OMD, "Done w/ initiator initialization", NO_VARG);

	for ( i=0; i<BEEP_NUMMSG; i++)
	{
		for( j=0; j<BEEPCOOK_NUM_INITIATOR; j++)
		{
			/* send the message */
			snprintf(sendMsg.msg, sendMsg.msglen,
					": Test message %d from initiator %d", i,
					j+BEEPRAW_NUM_INITIATOR);
			sendMsg.pathid=i+1;
			sendMsg.connid=j;

			/* for the last message, use a pathid from earlier, so we get
			 * BEEPCOOK_NUM_INITIATOR-1 paths sent */
			if(i==BEEPCOOK_NUM_INITIATOR-1)
				sendMsg.pathid=1;

			oms[j]->SEND_MSG(oms[j]->state, &sendMsg);
		}
	}

	DPRINT(DBG_OMD, "Done sending messages, close everything", NO_VARG);

	for (j=0; j<BEEPCOOK_NUM_INITIATOR; j++)
		oms[j]->DESTROY(oms[j]->state);

	DPRINT(DBG_DBG, "BEEP OM Initiator Done",NO_VARG);
}
END_TEST

/* we include this late in the game because we want to make sure none of the
 * functions used here are accidentally used above */
#include <librr/rr.h>
#include "util/profile-beep.h"

START_TEST(omod_beep_listener)
{
	struct IMOD_info *info;
	int sock;
	struct sockaddr_in addr;

	char errstr[1024];

	DBG_newproc("Listener");

	info=emalloc(sizeof(struct IMOD_info));
	memset(info,0,sizeof(struct IMOD_info));
	info->ipc_path=emalloc(strlen(BEEP_IPCPATH)+1);
	strlcpy(info->ipc_path, BEEP_IPCPATH, strlen(BEEP_IPCPATH)+1);

	sock=socket(AF_INET, SOCK_STREAM, 0);
	fail_if(sock<0, "socket() failed");

	memset(&(addr),0,sizeof(addr));
	addr.sin_family=AF_INET;
	addr.sin_port=htons(BEEP_PORT);
	addr.sin_addr.s_addr=htonl(INADDR_ANY);

	if(bind(sock,(struct sockaddr *)&addr,sizeof(addr))<0)
	{
		snprintf(errstr,1024,"Could not bind to socket; port: %d",BEEP_PORT);
		fail(errstr);
	}

	if(listen(sock, 5)<0)
	{
		snprintf(errstr,1024,"Could not listen(): %s",strerror(errno));
		fail(errstr);
	}

	info->fd_len=1;
	info->fd=emalloc(sizeof(int)*info->fd_len*2);
	info->chans=info->fd+info->fd_len;
	info->fd[0]=sock;
	info->chans[0]=2;
	
	/* try to sleep until the controller has come up */
	sleep(1);

	DPRINT(DBG_DBG, "Calling IMOD_BEEP_start()", NO_VARG);

	IMOD_BEEP_start(info, IMOD_BEEP);

	DPRINT( DBG_DBG, "Test case done", NO_VARG);
}
END_TEST

START_TEST(omod_beep_controller)
{
	IPC *ipc_socket;
	INW *waiter;
	struct INW_eInfo *event;
	struct message *msg;
	struct IPC_event *evt;
	struct IPC_addr *src;
	struct message *check_msg;
	struct path *path, *mpath;
    bool pathSeen[BEEP_NUMMSG][BEEPCOOK_NUM_INITIATOR];

	char errmsg[1024];

	int i,j;

	bool recvd_msg_no[BEEPRAW_NUM_INITIATOR+BEEPCOOK_NUM_INITIATOR][BEEP_NUMMSG];
	int initiator,msgno;

	pid_t module_pid;

	DBG_newproc("Controller");

	msg=emalloc(sizeof(struct message));
	path=emalloc(sizeof(struct path));
	src=emalloc(sizeof(struct IPC_addr));
	evt=emalloc(sizeof(struct IPC_event));
	check_msg=emalloc(sizeof(struct message));
	memset(recvd_msg_no,0,sizeof(recvd_msg_no));
    
    memset(pathSeen, 0, sizeof(pathSeen));

    /* initilize pathList  */
    initPCache(NULL, 0, BEEPCOOK_NUM_INITIATOR);

	/* init check_msg just like we initialized the other message, above */
	initMsg(check_msg);
	check_msg->timestamp=-1;	/* so we don't accidentally check this */

	/* init the IPC stuffs */
	ipc_socket=IPC_initController(BEEP_IPCPATH);
	if(IPCE_SUCCESS!=IPC_geterrno(ipc_socket))
	{
		IPC_destroy(ipc_socket);

		snprintf(errmsg,1024,
				"Failed to init IPC module: %s (%d)",
				IPC_geterrstr(ipc_socket),IPC_geterrno(ipc_socket));
		fail(errmsg);
	}

	/* setup the waiter */
	INW_initialize();
	waiter=INW_create();
	INW_addIPC(waiter,ipc_socket);

	/* get the first event */
	DPRINT(DBG_OMD, "Getting the first IPC event", NO_VARG);
	event=INW_getEvent(waiter);
	fail_unless(INWE_IPC==event->event_id, "non-IPC event received");
	fail_unless(event->info == IPCT_EVENT, "IPC event not of expected type");
	fail_unless(IPC_recvEvent(ipc_socket,evt), "Failed to receive IPC event");
	fail_unless(IPCEVT_IMSTART==evt->type, "event received was not Input Module Start");
	module_pid = evt->pid;

	DPRINT(DBG_DBG, "Input Module start event received", NO_VARG);

	/* get the messages */
	for ( i=0;
			i< (BEEP_NUMMSG * (BEEPRAW_NUM_INITIATOR+BEEPCOOK_NUM_INITIATOR) );
			/* incremented below */)
	{
		DPRINT(DBG_DBG, "Looping (i %d)",
				i);
		event=INW_getEvent(waiter);

		DPRINT(DBG_DBG, "Got event", NO_VARG);

		fail_unless(INWE_IPC==event->event_id,
				"non-IPC event received");

		if (event->info == IPCT_MSG)
		{
            memset(msg, 0, sizeof(*msg));
			fail_unless(IPC_cRecvMessage(ipc_socket,src,msg),
					"Failed to receive IPC MSG message");

			i++;	/* increment i */

			initiator=msg->msg[32]-'0';
			msgno=msg->msg[15]-'0';

			DPRINT(DBG_OMD, "received message: \'%s\' (%d,%d)", msg->msg,
					initiator,msgno);

			/* we do this to ensure that we don't care about order; check
			 * for duplicates, though */
			snprintf(check_msg->msg, check_msg->msglen,
					": Test message %d from initiator %d",
					msgno, initiator);

			if(initiator>=(BEEPRAW_NUM_INITIATOR+BEEPCOOK_NUM_INITIATOR)
					|| msgno>=BEEP_NUMMSG)
				fail("Initiator or msg number out of bounds");

			if(recvd_msg_no[initiator][msgno]!=FALSE)
			{
				DPRINT(DBG_DBG, "failing because a previoud message from \
						initiator %d, msgno %d was received",
						initiator,msgno);
				fail("Received duplicate message");
			}
			else
			{
				recvd_msg_no[initiator][msgno]=TRUE;
			}

			CMP_msg(check_msg,msg);

            if (msg->pathid != PATH_UNSET_ID) {
                if (! pathSeen[msg->pathid - 1][msg->connid] ) {
                    snprintf(errmsg, sizeof(errmsg),
                        "Haven't seen path ID %d, on connection %d",
                        msg->pathid, msg->connid);
                    fail(errmsg);
                }
            }
		}
		else if(event->info == IPCT_PATH)
		{
			DPRINT(DBG_DBG, "Path event", NO_VARG);

			fail_unless(IPC_cRecvPath(ipc_socket, src, path),
					"IPC_cRecvPath failed to receive a path");

            mpath = pathList[path->id - 1][path->connid];

            DPRINT(DBG_OMD, 
                "Got Path: id='%d' con='%d' to='%s' from='%s' props='%x'", 
                path->id, path->connid, path->toFQDN, path->fromFQDN, 
                path->props);
            DPRINT(DBG_OMD, 
                "Expected Path: id='%d' con='%d' to='%s' from='%s' props='%x'",
                mpath->id, mpath->connid, mpath->toFQDN, mpath->fromFQDN,
                mpath->props);

            /* make sure it was expected */
            CMP_path(mpath, path);

            /* mark that we have seen this path */
            pathSeen[path->id - 1][path->connid] = TRUE;

		}
		else if (event->info == IPCT_EVENT)
		{
			struct IPC_event   ipcevt;
			IPC_recvEvent(ipc_socket, &ipcevt);

			if( IPCEVT_IMSTART==evt->type )
				DPRINT(DBG_DBG, "Input Module start event received", NO_VARG);
		}
	}

	/* check to make sure we received all the proper messages */
	for(i=0;i<BEEP_NUMMSG;i++)
		for(j=0;j<(BEEPRAW_NUM_INITIATOR+BEEPCOOK_NUM_INITIATOR);j++)
		{
			if(recvd_msg_no[j][i]!=TRUE)
			{
				DPRINT(DBG_DBG, "Failing because we missed message initiator %d, no. %d",
						j,i);
				fail_if(recvd_msg_no[j][i]!=TRUE,
						"Not all sent messages were received");
			}
		}

	/* kill the module */
	DPRINT(DBG_OMD, "Kill the input module", NO_VARG);
	kill(module_pid, SIGTERM);

	sleep(2);

	/* clean up */
	DPRINT(DBG_OMD, "Cleaning up IPC layer", NO_VARG);
	IPC_destroy(ipc_socket);
	free(msg);
	free(src);
	free(evt);
	free(check_msg);
	INW_shutdown();

	DPRINT( DBG_DBG, "Test case done", NO_VARG);
}
END_TEST
