/**
 * @file ut_omod_udp.c
 *
 * @brief UDP output test suite.
 *
 * $Id: ut_omod_udp.c,v 1.2.2.6 2003/04/30 20:03:56 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Jul 25 11:25:58 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "check.h"
#include "ut_omod_base.h"

#include "const.h"
#include "error.h"
#include "debug.h"
#include "strcompat.h"
#include "util/bitvec.h"

#include "message/message.h"
#include "message/msg_frmt.h"
#include "outputmod/outputmod.h"

/**@name UDP Output Module Test Cases
 * @ingroup ut_omod
 * @{
 */

/** Echo server used in the test */
static void omod_udp_echo(void);

/** Actual Test OMOD test */
static void omod_udp_sender(void);

/*@}*/

static void debug_init(void)
{
    DBG_init("po", "omod-udp.debug");
}

static void debug_done(void)
{
    DPRINT(DBG_DBG, "Test Case Done", NO_VARG);
}

void omod_udp_suite(Suite *suite)
{
    TCase   *tc;

    tc = tcase_create("OMOD_UDP");
    suite_add_tcase(suite, tc);

    tcase_add_unchecked_fixture(tc, debug_init, debug_done);

    tcase_set_run_mode(tc, CK_PARALLEL);
    tcase_add_test(tc, omod_udp_echo);
    tcase_add_test(tc, omod_udp_sender);
}

/** The port for the echo server to listen on */
#define ECHO_SRV_PORT   5001

/** The number of senders */
#define SENDER_COUNT    2

/** Number of messages to send */
#define MESSAGE_COUNT   5

/** Piority tag signaling doneness (0, 0) */
#define PRI_DONE        "<0>"

START_TEST(omod_udp_sender)
{
    OMOD_interface      *oms[SENDER_COUNT];
    struct OMOD_opt     opts;
    struct in_addr      localhost;
    struct message      sendMsg;
    struct sockaddr_in  addr;
    int                 sock;
    int                 i, j;
    int                 rlen;
    char                rbuf[RFC3164_MAX_MSGLEN];
    char                mbuf[RFC3164_MAX_MSGLEN];


    /* fillin omod options */
    opts.format.type = MFRMT_RFC3164;
    opts.format.options = BM_MBIT(MFOPT_PRI);

    opts.outPath = "localhost";
    opts.pcache = NULL;
    fail_unless(inet_aton("127.0.0.1", &localhost), "inet_aton()");
    opts.outAddr = NA_newINet4(localhost);
    opts.outPort = ECHO_SRV_PORT;

    /* create omods */
    for (j = 0; j < SENDER_COUNT; j++) {
        oms[j] = OMOD_getInterface(OMOD_UDP);
        oms[j]->state = oms[j]->INIT(&opts);
        if (oms[j]->state == NULL) {
            fail("OM INIT() failed");
        }
    }

    /* bring up the receiving socket */
    addr.sin_family = PF_INET;
    addr.sin_port = htons(SYSLOG_UDP_PORT);
    inet_aton("127.0.0.1", &(addr.sin_addr));

    sock = socket(PF_INET, SOCK_DGRAM, 0);
    fail_if(sock < 0, "socket()");
    DPRINT(DBG_OMD, "test: sender, socket = %d", sock);

    i = 1;  /* need a real variable for setsockopt() */
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0) {
        fail("setsockopt(SO_REUSEADDR)");
    }

    if (bind(sock, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
        char    buf[1024];
        snprintf(buf, sizeof(buf), "bind(): (%d) %s", 
                SYSLOG_UDP_PORT, strerror(errno));
        fail(buf);
    }

    sendMsg.sev = 4;
    sendMsg.hostlen = strlen("localhost")+1;
    sendMsg.host = emalloc(sendMsg.hostlen);
    strlcpy(sendMsg.host, "localhost", sendMsg.hostlen);
    sendMsg.timestamp = time(NULL);
    sendMsg.tz = -700;
    strlcpy(sendMsg.tag, "udptest", MSGTAG_LEN);
    sendMsg.msglen = 512;
    sendMsg.msg = emalloc(sendMsg.msglen);

    for (i = 0; i < MESSAGE_COUNT; i++) {
        for (j = 0; j < SENDER_COUNT; j++) {
            if (i == MESSAGE_COUNT - 1) {
                /* done message */
                sendMsg.sev = 0;
                sendMsg.fac = 0;
            }
            else {
                sendMsg.fac = j + 1;
            }

            snprintf(sendMsg.msg, sendMsg.msglen, 
                    ": Test message %d from %d", i, j);

            DPRINT(DBG_OMD, "test: Sending message \"%s\"", sendMsg.msg);
            oms[j]->SEND_MSG(oms[j]->state, &sendMsg);

            rlen = recv(sock, rbuf, sizeof(rbuf), 0);
            if (rlen < 0) {
                fail("recv()");
            }
            DPRINT(DBG_OMD, "test: Received message \"%s\"", rbuf);

            if (MFRMT_format(mbuf,sizeof(mbuf),&sendMsg,&(opts.format)) <= 0) {
                fail("MFRMT_format()");
            }

            if (strcmp(mbuf, rbuf) != 0) {
                DPRINT(DBG_OMD, "test: nomatch-msg \"%s\"", mbuf);
                DPRINT(DBG_OMD, "test: nomatch-rcv \"%s\"", rbuf);
                fail("received and send messages don't match");
            }
        }
    }

    for (j = 0; j < SENDER_COUNT; j++) {
        oms[j]->DESTROY(oms[j]->state);
    }

    close(sock);
    DPRINT(DBG_DBG, "UDP OM Sender Done", NO_VARG);
} 
END_TEST


START_TEST(omod_udp_echo)
{
    int                 sock;
    struct sockaddr_in  addr;
    socklen_t           addr_len;
    int                 len;
    char                buf[RFC3164_MAX_MSGLEN];
    int                 done_msgs;

    DBG_newproc("ECHO");

    sock = socket(PF_INET, SOCK_DGRAM, 0);
    fail_if(sock < 0, "socket()");
    DPRINT(DBG_OMD, "test: echo: socket = %d", sock);

    addr.sin_family = PF_INET;
    addr.sin_port = htons(ECHO_SRV_PORT);
    addr.sin_addr.s_addr = INADDR_ANY;
    if (bind(sock, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
        fail("bind()");
    }

    done_msgs = 0;
    while (done_msgs < SENDER_COUNT) {

        addr_len = sizeof(addr);
        len = recvfrom(sock, buf, RFC3164_MAX_MSGLEN, 0, 
                       (struct sockaddr*)&addr, &addr_len);
        if (len < 0) {
            fail("recvfrom()");
        }
        DPRINT(DBG_OMD, "test: Got message \"%s\"", buf);

        DPRINT(DBG_OMD, "message from %s:%d", 
                inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

        if (strncmp(buf, PRI_DONE, strlen(PRI_DONE)) == 0) {
            done_msgs++;
        }

        len = sendto(sock, buf, len, 0, (struct sockaddr*)&addr, addr_len);
        if (len < 0) {
            fail("sendto()");
        }
    }

    close(sock);

    DPRINT(DBG_DBG, "Echo Server Exiting", NO_VARG);
}
END_TEST
