/**
 * @file ut_privs.c
 *
 * @brief Unit test for functions in controller/privs.c.  Must be run as
 * root.
 *
 * $Id: ut_privs.c,v 1.3.2.1 2003/04/30 20:03:57 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Jun 18 12:37:44 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "check.h"
#include "types.h"
#include "strcompat.h"
#include "privs.h"

#include "ut_driver.h"

/**@defgroup ut_privs PRIV function tests
 * @{
 */

/** Test temporary priv dropping */
static void priv_droptemp(void);

/** Test permanent priv dropping */
static void priv_dropperm(void);

/** Test open for read */
static void priv_openread(void);

/** Test open for write */
static void priv_openwrite(void);

/** Suite setup function for privs test */
void priv_suite(Suite *suite);

/*@}*/

/**@defgroup ut_privs_impl PRIV implementation
 * @{
 */

/* the suites that we run */
SuiteInfo suites[] = {
    ADD_SUITE(priv_suite, "PRIV_functions")
};

/* suite count */
int suite_count = SUITE_COUNT;

/* name of the log file */
char const *log_name = "privs";

bool disablePrivs = FALSE;

/** setup fixture */
static void priv_setup(void);

/** teardown fixture */
static void priv_finish(void);

/** user without privs to run as */
#define UNPRIV_USER "nobody"

/** group without privs to run as */
#define UNPRIV_GROUP "nogroup"

/** UID of root (for clarity, not b/c I expect it to change) */
#define ROOT_UID    0

/** Temp file template */
#define TMP_TMPL    "/tmp/ut_priv.XXXXXX"

/* test suite setup */
void priv_suite(Suite *suite)
{
    TCase   *tc_priv;

    tc_priv = tcase_create("PRIV Functions");
    suite_add_tcase(suite, tc_priv);
    tcase_add_checked_fixture(tc_priv, priv_setup, priv_finish);

    tcase_add_test(tc_priv, priv_droptemp);
    tcase_add_test(tc_priv, priv_dropperm);
    tcase_add_test(tc_priv, priv_openread);
    tcase_add_test(tc_priv, priv_openwrite);
}

static void
priv_setup(void)
{
    char    buf[1024];

    fail_unless(getuid() == 0, "Test cases must be run as root");

    if (! PRIV_setUnprivUser(UNPRIV_USER, UNPRIV_GROUP) ) {
        snprintf(buf, sizeof(buf), "Can't set %s:%s as unprived user:group", 
                UNPRIV_USER, UNPRIV_GROUP);
        fail(buf);
    }

}

static void
priv_finish(void)
{
    /* nothing to do */
}

START_TEST(priv_droptemp)
{
    fail_unless(PRIV_dropTemp(), "Can't do temp. drop of perms");

    fail_unless(getuid() != 0, "UID still 0");
    fail_unless(getgid() != 0, "GID still 0");

    /* now try to regain them */
    fail_unless(setreuid(-1, ROOT_UID) == 0, "Can't regain perms");

}
END_TEST

START_TEST(priv_dropperm)
{
    fail_unless(PRIV_dropPerm(), "Can't do permenant drop of perms");

    fail_unless(getuid() != 0, "UID still 0");
    fail_unless(getgid() != 0, "GID still 0");

    fail_unless(setreuid(-1, ROOT_UID) != 0, "Can regain perms");
    fail_unless(errno == EPERM, "setreuid() didn't get a permission error");
}
END_TEST

START_TEST(priv_openread)
{
    char    insecfn[256], secfn[256];
    char    isfd, sfd;


    /* need a couple of temp files */
    strlcpy(insecfn, TMP_TMPL, sizeof(insecfn));
    strlcpy(secfn, TMP_TMPL, sizeof(secfn));


    isfd = mkstemp(insecfn);
    fail_unless(isfd >= 0, "mkstemp(insecfn) failed");
    fail_unless(
        fchmod(isfd, S_IWUSR | S_IRUSR | S_IWGRP | S_IRGRP | S_IWOTH) == 0,
        "fchmod(isfd) failed");
    close(isfd);

    sfd = mkstemp(secfn);
    fail_unless(sfd >= 0, "mkstemp(secfn) failed");
    fail_unless(
        fchmod(sfd, S_IWUSR | S_IRUSR | S_IWGRP | S_IRGRP) == 0,
        "fchmod(sfd) failed");

    printf("insecfn = %s\n", insecfn);
    printf("secfn = %s\n", secfn);

    fail_unless(PRIV_dropTemp(), "dropTemp() failed");

    isfd = PRIV_openReadFile(insecfn, 0);
    fail_unless(isfd < 0, "Insecure file opened successfully");

    sfd = PRIV_openReadFile(secfn, 0);
    fail_unless(sfd >= 0, "Secure faile not opened successfully");
    close(sfd);

    if (setuid(0) == 0) {
        unlink(secfn);
        unlink(insecfn);
    }
}
END_TEST

START_TEST(priv_openwrite)
{
    char    symfn[256], secfn[256];
    char    symfd, sfd;

    /* need a couple of temp files */
    strlcpy(symfn, TMP_TMPL, sizeof(symfn));
    strlcpy(secfn, TMP_TMPL, sizeof(secfn));


    sfd = mkstemp(secfn);
    fail_unless(sfd >= 0, "mkstmp(sfd) failed");
    fail_unless(fchmod(sfd, S_IWUSR | S_IRUSR | S_IWGRP | S_IRGRP) == 0, 
            "fchmod(sfd) failed)");
    close(sfd);

    symfd = mkstemp(symfn);
    fail_unless(symfd >= 0, "mkstmp(sfd) failed");
    close(symfd);
    fail_unless( unlink(symfn) == 0, "unlink(symfn) failed");
    fail_unless( symlink(secfn, symfn) == 0, "symlink() failed");

    printf("symfn = %s\n", symfn);
    printf("secfn = %s\n", secfn);

    fail_unless(PRIV_dropTemp(), "PRIV_dropTemp() failed");

    symfd = PRIV_openWriteFile(symfn, 0);
    fail_unless(symfd < 0, "Symlink incorrectly opened for write"); 

    sfd = PRIV_openWriteFile(secfn, 0);
    fail_unless(sfd >= 0, "Open failed on Safe file");
    close(sfd);

    if (setuid(0) == 0) {
        unlink(symfn);
        unlink(secfn);
    }
}
END_TEST

/*@}*/

