/**
 * @file ut_driver.c
 *
 * @brief This defines a common setup function that various tests can
 *  use to run tests.
 * @par
 *
 *
 * $Id: ut_driver.c,v 1.4.2.1 2003/04/30 20:03:41 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Feb 04 14:17:03 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <linux/limits.h>   /* for PATH_MAX */
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
#include "acconfig.h"
#endif

#include "check.h"
#include "ut_driver.h"

#include "debug.h"
#include "test.h"

/**@defgroup ut_driver_impl Unit Test Driver Implementation
 * @ingroup ut_driver
 * @{
 */

/** Command like arg for no fork mode. */
#define NOFORK_ARG "-n"

/** Setup the master suite */
static void setup_master(Suite *msuite);

int main(int argc, char *argv[])
{
    int         failed = 0;
    int         i;
    SRunner     *sr;
    Suite       *master_suite;
    char        logfile[PATH_MAX];
    int         nofork = 0;

    /* Tell everyone that we are unit testing */
    unit_test = TRUE;

    /* disable privs unless otherwise requested */
    if (disablePrivs) 
        PRIV_disableChanges();

    if (argc > 1 && strcmp(argv[1], NOFORK_ARG) == 0) {
        nofork = 1;
    }

    /* build the logfile name */
    snprintf(logfile, sizeof(logfile), "%s.log", log_name);
    unlink(logfile);

    /* setup the master suite */
    master_suite = suite_create("Master");
    setup_master(master_suite);

    /* create the driver */
    sr = srunner_create(master_suite);
    srunner_set_log(sr, logfile);
    if (nofork > 0) {
        srunner_set_fork_status(sr, CK_NOFORK);
    }

    /* add all the suites */
    for (i = 0; i < suite_count; i++) {
        suites[i].suite = suite_create(suites[i].name);
        suites[i].setup(suites[i].suite);
        srunner_add_suite(sr, suites[i].suite);
    }

    /* run tests */
    srunner_run_all (sr, CK_NORMAL); 
    failed = srunner_ntests_failed(sr); 

    /* Clean up */
    srunner_free(sr); 
    suite_free(master_suite);
    for (i = 0; i < suite_count; i++) {
        suite_free(suites[i].suite); 
    }

    /* in unix 0 == success */
    return failed;
}

static void
setup_master(Suite *master)
{
    /* run no tests */
}

/*@}*/
