/**
 * @file ut_driver.h
 *
 * @brief The header for common test suite stuff.
 *
 *
 * $Id: ut_driver.h,v 1.3.2.1 2003/04/30 20:03:41 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Feb 04 14:21:09 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef TEST_UNIT_MAIN_H
#define TEST_UNIT_MAIN_H

/**@defgroup unit_test Unit Tests */
/**@defgroup ut_driver Unit Test Driver 
 * @ingroup unit_test
 */

#include "check.h"

#include "types.h"

/** @name Unit Test Driver Support 
 *  @ingroup ut_driver
 */

/** The type of a suite registration function. The parameter is the test
 * suite which should be setup 
 */
typedef struct  {
    void (*setup)(Suite*);      /**< suite init function */

    char            *name;      /**< suite name */
    Suite           *suite;     /**< storage for the suite */
} SuiteInfo;

/**@name Unit Test Driver Config
 * @ingroup ut_driver
 * @{
 */

/** 
 * Add a suite to the list for building.
 * @param __func The test suite init function
 * @param __name The test suite name
 */
#define ADD_SUITE(__func, __name) {__func, __name, NULL}

/** vim syntax highlighting breaks on ADD_SUITE ... this fixes it */
#define FIX_BROKEN_VIM_SYNTAX_HIGHLIGHTING }

/** Calculate the number of test_suites with info in \c suties */
#define SUITE_COUNT (sizeof(suites) / sizeof(SuiteInfo))

/** 
 * The list of suites.  This must be defined in one object file that is
 * linked against main.o.  It allows for flexible grouping of tests,
 * with minimal code re-writing.
 * @par
 * Adding a new test suite to this list should be done with the
 * ADD_SUITE macro.
 * @note The generated documentation doesn't fully expand ADD_SUITE, but
 * instead only expands it to the suite name.  Less clutter that way.
 */
extern SuiteInfo    suites[];
/**
 * The number of suites that are to be tested. This must be defined in
 * the same place as \c suites.  And can easily be setup by the
 * SUITE_COUNT macro
 */ 
extern int          suite_count;

/**
 * This is the name of the log file in which data should be stored.  It
 * should give a unique name to the test group, but should not end in
 * '.log' or similar.  Also it should not contain a path.
 */
extern char const *log_name;

/**
 * This controls a call to PRIV_disableChanges().  Each unit test must
 * provide storage space for this, and set it's default value properly.
 */
extern bool     disablePrivs;

/*@}*/

/** Main driver function for testing.  If suites and suite_count are
 * initilized correctly then linking in main.o will be enough to create
 * a working set of test suites.
 * @ingroup ut_driver
 *
 * @param argc The number of command line arguments
 * @param argv String vector with the command line arguments
 * @return 0 if all tests were successful, non-zero otherwise.
 */
int main(int argc, char *argv[]);


#endif /* TEST_UNIT_MAIN_H */
