/**
 * @file acl.c
 *
 * @brief This file implements the various ACL library functions.
 * @ingroup acl
 *
 *
 * $Id: acl.c,v 1.14.2.2 2003/04/30 20:03:58 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Jan 29 11:45:29 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#if HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <stdlib.h>

#include "types.h"
#include "netaddr.h"
#include "error.h"
#include "ipc/buffer.h"
#include "util/netaddr.h"

/* we prefer this over the system string.h b/c this will cover for any
 * functions missing in the system's version.
 */
#include "strcompat.h"

#include "acl.h"

static char const rcsid[] 
    = "$Id: acl.c,v 1.14.2.2 2003/04/30 20:03:58 devink Exp $";

/** Make an actual check of addr against the specified ACL.
 * This does very little preliminary error checking as it was assumed to
 * have been done in a public function.
 * @internal
 * @param acl The ACL to check against
 * @param addr the addr to check against
 * @return TRUE if addr is allowed by acl, FALSE otherwise.
 */
static bool doCheckIP(ACL *aclchk, netaddr_t addr);

/** Check an ACL against the actual link properties.  If props doesn't
 * contain at least all of the bits this denys it.
 * @internal
 * @param acl The ACL to check against
 * @param props The Link Props bitmap to check
 * @return TRUE if props contains all the properties in acl, FALSE
 *  otherwise.
 */
static bool doCheckLinkProps(ACL *acl, int props);

ACLChecker* 
ACL_init(void)
{
    ACLChecker  *chkr;

    chkr = emalloc(sizeof(ACLChecker));
    memset(chkr, 0, sizeof(ACLChecker));

    return chkr;
}

void
ACL_destroy(ACLChecker *aclchk)
{
    if (aclchk != NULL)
        free(aclchk);
}

bool
ACL_setAcl(ACLChecker *aclchk, ACL *acl)
{
    int type;

    if (acl->type == ACLT_NONE || acl->mode == ACLM_NONE) 
        return FALSE;

    type = acl->type;
    if (aclchk->acls[type] != NULL) {
        ACL_free(aclchk->acls[type]);
        aclchk->acls[type] = NULL;
    }

    aclchk->acls[type] = acl;

    return TRUE;
}

bool
ACL_checkHost(ACLChecker *aclchk, char const *host, ACL_type type)
{
    netaddr_t   addr;
    bool        result;

    /* no set ACL == deny */
    if (aclchk->acls[type] == NULL) 
        return FALSE;

    /* this is all we check here */
    if (type != ACLT_GENHOST && type != ACLT_CONHOST) 
        return FALSE;


    /**@note Should we really do all our acl checks on IP? */
	addr=NA_newINetLookup(host);
    if ( NA_isError(addr) )
	{
        /* lookup failed ... deny it */
        return FALSE;
    }

    result = doCheckIP(aclchk->acls[type], addr);

    /* cleanup */
    NA_destroy(addr);

    return result;
}

bool
ACL_checkIP(ACLChecker *aclchk, netaddr_t addr, ACL_type type)
{
    /* no set ACL == deny */
    if (aclchk->acls[type] == NULL)
        return FALSE;

    /* this is all we check here */
    if (type != ACLT_GENHOST && type != ACLT_CONHOST) 
        return FALSE;

    return doCheckIP(aclchk->acls[type], addr);
}



bool
ACL_checkPath(ACLChecker *aclchk, struct path *pHead, ACL_type type)
{
    struct path     *cur;
    bool            result;

    /* Only check if there is such an acl */
    if (aclchk->acls[type] == NULL) 
        return FALSE;

    /* Only checking PATH && PATHSEC here */
    if (type != ACLT_PATH && type != ACLT_PATHSEC)
        return FALSE;

    result = TRUE;  

    cur = pHead;
    while (cur != NULL) {
        if (type == ACLT_PATH) {
            /* Explicit allow Vs. Explicit deny is transitive ... so if
             * doCheckIP() denys any relay, the whole path is denied.
             */
            if (! doCheckIP(aclchk->acls[ACLT_PATH], cur->toIP) 
                || ! doCheckIP(aclchk->acls[ACLT_PATH], cur->fromIP) ) 
            {
                result = FALSE;
                break;
            }
        }
        else {
            /** We only check ACTL_PATH && ACLT_PATHSEC .. so this must 
             * be ACLT_PATHSEC
             */
            assert(type == ACLT_PATHSEC);   
            if (!doCheckLinkProps(aclchk->acls[ACLT_PATHSEC], cur->props) ) {
                result = FALSE;
                break;
            }
        }
    }

    return result;
}

ACL *
ACL_create(ACL_type type, ACL_mode mode)
{
    ACL     *acl;

    acl = emalloc(sizeof(*acl) );
    memset(acl, 0, sizeof(*acl) );

    acl->type = type;
    acl->mode = mode;
    acl->list = LL_create();

    return acl;
}

void
ACL_free(ACL *acl)
{
    LLIterator  itr;
    ACL_data    *elem;

    assert(acl != NULL);

    itr = LL_iterator(acl->list);
    while ( (elem = LLI_next(itr)) != NULL) {
        free(elem);
    }
    LLI_destroy(&itr);
    LL_destroy(&(acl->list));
    free(acl);
}

ACL *
ACL_fromBuffer(struct IPC_buffer *buf)
{
    ACL_type    type;
    ACL_mode    mode;
    uint32_t    count;
    uint32_t    linkProps;
    netaddr_t   addr;
    ACL         *acl;
    bool        error;

    error = FALSE;

    DPRINT(DBG_ACL, "from buf: start, pos = +%d", 
            buf->pos - buf->data);

    /* try to get the header info */
    if (! BUF_getInteger(buf, &type) ) error = TRUE;
    else if (! BUF_getInteger(buf, &mode) ) error = TRUE;
    else if (! BUF_getInteger(buf, &count) ) error = TRUE;

    DPRINT(DBG_ACL, "from buf: type = %d, mode = %d, count = %d",
            type, mode, count);
    DPRINT(DBG_ACL, "from buf: after headers pos = +%d",
            buf->pos - buf->data);

    if (error) {
        return NULL;
    }
    
    /* creat the acl we will return */
    acl = ACL_create(type, mode);

    while (count > 0) {
        if (type == ACLT_PATHSEC) {
            if (! BUF_getInteger(buf, &linkProps) ) break;

            if (! ACL_addLinkProps(acl, linkProps) ) break;
        }
        else {
            addr = NA_newBuffer(buf);
            if (NA_isError(addr) ) {
                NA_destroy(addr);
                break;
            }

            if (! ACL_addIP(acl, addr) ) break;

            NA_destroy(addr);
        }

        count--;
    }

    DPRINT(DBG_ACL, "from buf: end, count = %d",  count);

    if (count > 0) {
        /* ERROR */
        ACL_free(acl);
        return NULL;
    }

    return acl;
}

int
ACL_getBufferSize(ACL *acl)
{
    int     len;


    if (acl->type == ACLT_PATHSEC) {
        len = LL_size(acl->list) * BUF_INTSZ;
    }
    else {
        len = LL_size(acl->list) * NA_MAX_BUFLEN;
    }

    if (len < 0) {
        return -1;
    }

    /* header is 2 INTEGERS + an INTEGER count of data elems */
    len += 3 * BUF_INTSZ;

    return len;
}

bool
ACL_toBuffer(ACL *acl, struct IPC_buffer *out)
{
    ACL_data    *data;
    LLIterator  itr;
    IPC_bufMark countPos, startPos;
    uint32_t    count;
    ACL_type    type;
    int         len;
    int         rtn;

    assert(acl != NULL); 
    assert(out != NULL);
    assert(out->data != NULL);

    DPRINT(DBG_ACL, "to buf: start", NO_VARG);

    /* first there fields are INTEGER, so just initilize len to that */
    len = BUF_INTSZ * 3;            

    /* save start, so we can return on error */
    startPos = BUF_markPos(out);    

    BUF_setInteger(out, (int)acl->type);
    BUF_setInteger(out, (int)acl->mode);

    DPRINT(DBG_ACL, "to buf: type = %d, mode = %d",
            acl->type, acl->mode);

    countPos = BUF_markPos(out);    /* save the postion of the count */
    BUF_setInteger(out, 0);         /* save room for the count */

    itr = LL_iterator(acl->list);
    data = LLI_next(itr);
    type = acl->type;
    count = 0;

    while (data != NULL) {
        if (count >= UINT_MAX) break;   /* to many elements, error */

        count++;
        if (type == ACLT_PATHSEC) {
            if (! BUF_setInteger(out, data->linkProps) ) 
                break; /* error */

            len += BUF_INTSZ;
        }
        else {
            rtn = NA_toBuffer(data->addr, out);
            if (rtn < 0) 
                break; /* error */

            len += rtn;
        }

        data = LLI_next(itr);
    }

    /* data will == NULL even if we overflowed.  But len will be set
     * properly ... and out will not have been written past.
     */

    if (data != NULL) {
        /* error condition */
        BUF_resetToPos(out, startPos);
        return -1;
    }

    /* Set the count */
    BUF_setIntegerAt(out, countPos, count);

    DPRINT(DBG_ACL, "to buf: end, len = %d", len);
    DPRINT_BUF(DBG_ACL, "to buf: end, buf = ", out, -1);

    return len;
}

bool
ACL_addIP(ACL *acl, netaddr_t addr)
{
    ACL_data    *elem;

    /* Not a vaild operation for path security checks */
    if (acl->type == ACLT_PATHSEC) {
        return FALSE;
    }

    elem = emalloc(sizeof(*elem) );
    memset(elem, 0, sizeof(*elem) );

    elem->addr = NA_newNetaddr(addr);

    LL_append(acl->list, elem);

    return TRUE;
}

bool
ACL_addHost(ACL *acl, char const *host)
{
    netaddr_t   addr;
    bool        result;

    /* Not a vaild operation for path security checks */
    if (acl->type == ACLT_PATHSEC) {
        return FALSE;
    }

	addr=NA_newINetLookup(host);
    if ( NA_isError(addr) )
	{
        DPRINT(DBG_ACL, "NA_newINetLookup() failed for %s", host);
        return FALSE;
    }

    result = ACL_addIP(acl, addr);

    NA_destroy(addr);

    return result;
}

bool
ACL_addLinkProps(ACL *acl, int props)
{
    ACL_data        *elem;

    /* Only good for path security checks (or untyped ACLs) */
    if (acl->type != ACLT_PATHSEC && acl->type != ACLT_NONE) {
        return FALSE;
    }

    elem = emalloc(sizeof(*elem));
    memset(elem, 0, sizeof(*elem));

    elem->linkProps = props;

    LL_append(acl->list, elem);

    return TRUE;
}

bool
ACL_addACL(ACL *acl, ACL *add)
{
	LLIterator it;
	ACL_data *elem;

	/* first, error checking */
	if( (add->type==acl->type) && (add->mode==acl->mode) )
		;	/* they match, no error */
	else if( (ACLT_NONE==add->type) && (ACLM_NONE==add->mode) )
		;	/* we're adding a non-type ACL, no error */
	else
		return FALSE;

	it=LL_iterator(add->list);
	while(NULL!=LLI_next(it))
	{
		elem=emalloc(sizeof(ACL_data));
		memcpy(elem,LLI_elem(it),sizeof(ACL_data));

		LL_append(acl->list, elem);
	}

	LLI_destroy(&it);
	return TRUE;
}

static bool 
doCheckLinkProps(ACL *acl, int props)
{
    ACL_data    *curAcl;
    LLIterator  itr;
    int         targetp;
    bool        result;

    assert(acl != NULL);    

    result = TRUE;

    itr = LL_iterator(acl->list);
    while ( (curAcl = LLI_next(itr)) != NULL) {
        /* Mode is ignored here */
        targetp = curAcl->linkProps;
        if ( (props & targetp) != targetp) {
            /* doesn't meet the minimum requirements */
            result = FALSE;
            break;
        }
    }
    return result;
}

static bool
doCheckIP(ACL *acl, netaddr_t addr) 
{
    ACL_data    *curAcl;
    LLIterator  itr;
    bool        found;

    assert(acl != NULL);

    found = FALSE;

    itr = LL_iterator(acl->list);
    while ( (curAcl = LLI_next(itr)) != NULL ) {
        if (NA_compare(curAcl->addr, addr) == 0) {
            found = TRUE;
            break;
        }
    }
    
    if (acl->mode == ACLM_DENY) {
        /* it's a deny list */
        return !found;
    }
    else if (acl->mode == ACLM_ALLOW) {
        /* it's an allow list */
        return found;
    }
    else {
        /* should never be reached */
        throw_fatal("Invalid ACL mode");
        return FALSE;
    }
}

