/**
 * @file acl.h
 *
 * @brief %ACL checking and management code.  
 * @par
 * This file defines the data structures, constants, and functions used 
 * to setup and check ACLs. These ACLs are used by the input modules.
 *
 * @design d2_2
 * @ingroup acl
 *
 * $Id: acl.h,v 1.10.2.1 2003/04/30 20:03:59 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jan 23 11:41:42 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ACL_ACL_H
#define ACL_ACL_H

#include "types.h"
#include "netaddr.h"
#include "message/message.h"
#include "ipc/buffer.h"
#include "debug.h"

/* Overall ACL grouping */
/** @addtogroup acl ACL Library */


/* some document sub-groups */
/**@defgroup acl_modes ACL Mode 
 * @ingroup acl
 */
/**@defgroup acl_types ACL Type 
 * @ingroup acl
 */

/** @name ACL Mode Values
 *  @ingroup acl_modes
 *  @{ 
 */  
typedef enum 
{
    ACLM_NONE,      /**< Undefined mode */
    ACLM_DENY,      /**< Explict deny mode */
    ACLM_ALLOW      /**< Explict allow mode */
} ACL_mode;
/** @} */

/** @name Types of ACL Checks 
 *  @ingroup acl_types
 *  @{
 */
typedef enum 
{
    ACLT_NONE,      /**< not tied to a particular check */
    ACLT_PATH,      /**< path check (@ref d2_2_1) */
    ACLT_PATHSEC,   /**< path security check (@ref d2_2_2) */
    ACLT_GENHOST,   /**< Generating host check (@ref d2_2_3) */
    ACLT_CONHOST,   /**< Connecting host check (@ref d2_2_4) */
    ACLT_COUNT      /**< The number of ACL types */
} ACL_type;
/** @} */

/* our private bits -- must be included after the constants are defined */
#include "acl_private.h"

/**@name ACL Public Interface 
 * @ingroup acl 
 * @{
 */



/**
 * Initilize the %ACL library. This library in re-entrant, and a handle
 * to use will be returned.
 *
 * @return ACLChecker* Handle for one library setup.
 */
ACLChecker* ACL_init(void);

/**
 * Destroy the ACLChecker object.
 *
 * @param aclchk Handle
 */
void ACL_destroy(ACLChecker *aclchk);

/**
 * Add an %ACL to be checked by the ACL_check*() functions.
 * The ACL structure contains all the information
 * needed to know which list gets changed.  This will removed any
 * data that existed previously.
 *
 * @param aclchk Handle
 * @param acl ACL to add to the list
 * @return int TRUE on success, FALSE otherwise.
 */
bool ACL_setAcl(ACLChecker *aclchk, ACL *acl);

/**
 * Match a string against the specified \ref acl_types.  
 *
 * @param aclchk Handle
 * @param host The host to check against an ACL
 * @param type The \ref acl_types to check \c host against
 * @return TRUE if \c host is allowed, FALSE otherwise
 */
bool ACL_checkHost(ACLChecker *aclchk, char const *host, ACL_type type);

/**
 * Match a netaddr address  against the specified \ref acl_types.
 *
 * @param aclchk Handle
 * @param addr The network adddress to check against an ACL
 * @param type The \ref acl_types to check \c ip against
 * @return TRUE if \c ip is allowed, FALSE otherwise
 */
bool ACL_checkIP(ACLChecker *aclchk, netaddr_t addr, ACL_type type);

/**
 * Match an ACL against an entire path.  This only will check ACLT_PATH
 * and ACLT_PATHSEC.
 *
 * @param aclchk Handle
 * @param pathHead The head of the path to check against
 * @return TRUE if the path is allowed, FALSE otherwise.
 */
bool ACL_checkPath(ACLChecker *alcchk, struct path *pathHead, ACL_type type);

/**
 * Construct an ACL. 
 * @warning This MUST be used to construct ACLs, if it isn't bad things
 * may happen when the ACL is destroyed by other functions in this
 * library.
 *
 * @param type The \ref acl_types
 * @param mode The \ref acl_modes
 * @retrun the new ACL
 */
ACL *ACL_create(ACL_type type, ACL_mode mode);

/**
 * Free an ACL.
 *
 * @param acl The ACL to destroy
 *
 * @pre acl != NULL
 */
void ACL_free(ACL *acl);

/**
 * Construct an ACL from a buffer.  The buffer should have been created
 * by ACL_toBuffer().  This is used to serialize ACLs for IPC.  Note
 * that the returned ACL was created with ACL_create() and should be
 * destroyed accordingly.
 *
 * @param buf The buffer to turn into an ACL
 * @return a new ACL
 *
 * @pre buf != NULL
 */
ACL *ACL_fromBuffer(struct IPC_buffer *buf);

/**
 * Get the max amount of space that may be required to put the given ACL
 * in an IPC_buffer.  Note that this will err on the side of being too
 * large.  \c acl may require this much space, but it could also require
 * less space.
 *
 * @param acl The acl to check.
 * @return the max amount of space required, or < 0 on error.
 *
 * @pre acl != NULL
 */
int ACL_getBufferSize(ACL *acl);

/**
 * Serialize an ACL.  This will try to serialize \c acl and place it in
 * \c out.  Errors will make this function reset the position in the
 * buffer \c out so that the buffer will not have garbage data.  Further
 * passing \c out==NULL will result in simply caclulating the buffer
 * size needed and returning that.  
 *
 * @param acl The ACL to serialize.
 * @param out The buffer to write the acl into.
 * @return The number of bytes needed to seralize \c acl, otherwise < 0.  
 *
 * @pre acl != NULL
 * @pre out != NULL
 * @pre out has space allocated for it.
 */
size_t ACL_toBuffer(ACL *acl, struct IPC_buffer *out);

/**
 * This addes IP addresses to the ACL list.
 *
 * @param acl The ACL to work on.
 * @param addr The address to add (will be copied).
 * @return TRUE if \c addr was successfuly added to \c acl, FALSE
 *  otherwise
 */
bool ACL_addIP(ACL *acl, netaddr_t addr);

/**
 * This addes hosts to the ACL list.
 *
 * @param acl The ACL to work on.
 * @param host The host to add (will be copied)
 * @return TRUE if \c host was successfuly added to \c acl, FALSE
 *  otherwise
 */
bool ACL_addHost(ACL *acl, char const *host);

/**
 * This adds link properties to the ACL list.
 * @param acl The ACL to work on.
 * @param linkProps The link properties bit map to add 
 * @return TRUE if \c linkProps were successfuly added to \c acl, FALSE
 *  otherwise
 */
bool ACL_addLinkProps(ACL *acl, int linkProps);

/**
 * This adds in another ACL.  The type and mode of the ACL that is being
 * added must either be ACLT_NONE/ACLM_NONE or match the ACL being
 * updated.
 *
 * @param acl The ACL to work on.
 * @param add the ACL to add to \c acl
 * @return TRUE if \c add was added to \c acl, FALSE otherwise.
 */
bool ACL_addACL(ACL *acl, ACL *add);

/*@}*/

#endif /* ACL_ACL_H */
