/**
 * @file avltree.c
 *
 * @brief An AVL tree container, that uses unsigned integers as the key
 * type.  The data type store is void* for flexibility.
 * @par
 * This code was taken from http://www.purists.org/avltree.
 * Modified by Devin Kowatch <devink@sdsc.edu> for SDSCSyslogd.
 * @par
 * This is not as generic as it could be.  Currently the key must be an
 * usigned integer, however, the stored data is a void*.
 *
 * @todo This needs lots of doxygen cleanup.
 *
 * $Id: avltree.c,v 1.3.2.2 2003/04/30 20:03:59 devink Exp $
 */

#if HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <stdlib.h>

#if WITH_DMALLOC
#   include "dmalloc.h"
#endif

#include "avltree.h"

#include "debug.h"
#include "error.h"

static char const rcsid[] = 
    "$Id: avltree.c,v 1.3.2.2 2003/04/30 20:03:59 devink Exp $";

/**
 *  Which of a given node's subtrees is higher?
 */
enum AVLSKEW 
{
	NONE,	
	LEFT,
	RIGHT
};

/**
 *  Did a given insertion/deletion succeed, and what do we do next?
 */
typedef enum 
{
	ERROR = 0,
	OK,
	BALANCE
} AVLRES;


/**
 *  AVL tree node structure
 */
struct avlnode
{
	struct avlnode *left, *right;
	enum AVLSKEW skew;

    AVL_key_t k;
	void *d;
};

/** AVL Tree object */
struct AVL_tree 
{
    struct avlnode  *root;
};

/*
 *  avlrotleft: perform counterclockwise rotation
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node
 */
void
avlrotleft(struct avlnode **n)
{
	struct avlnode *tmp = *n;

	*n = (*n)->right;
	tmp->right = (*n)->left;
	(*n)->left = tmp;
}

/*
 *  avlrotright: perform clockwise rotation
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node
 */
void
avlrotright(struct avlnode **n)
{
	struct avlnode *tmp = *n;

	*n = (*n)->left;
	tmp->left = (*n)->right;
	(*n)->right = tmp;
}

/*
 *  avlleftgrown: helper function for avlinsert
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node. This node's left 
 *                subtree has just grown due to item insertion; its 
 *                "skew" flag needs adjustment, and the local tree 
 *                (the subtree of which this node is the root node) may 
 *                have become unbalanced.
 *
 *  Return values:
 *
 *    OK          The local tree could be rebalanced or was balanced 
 *                from the start. The parent activations of the avlinsert 
 *                activation that called this function may assume the 
 *                entire tree is valid.
 *
 *    BALANCE     The local tree was balanced, but has grown in height.
 *                Do not assume the entire tree is valid.
 */
AVLRES
avlleftgrown(struct avlnode **n)
{
	switch ((*n)->skew) {
	case LEFT:
		if ((*n)->left->skew == LEFT) {
			(*n)->skew = (*n)->left->skew = NONE;
			avlrotright(n);
		}	
		else {
			switch ((*n)->left->right->skew) {
			case LEFT:
				(*n)->skew = RIGHT;
				(*n)->left->skew = NONE;
				break;

			case RIGHT:
				(*n)->skew = NONE;
				(*n)->left->skew = LEFT;
				break;

			default:
				(*n)->skew = NONE;
				(*n)->left->skew = NONE;
			}
			(*n)->left->right->skew = NONE;
			avlrotleft(& (*n)->left);
			avlrotright(n);
		}
		return OK;

	case RIGHT:
		(*n)->skew = NONE;
		return OK;
	
	default:
		(*n)->skew = LEFT;
		return BALANCE;
	}
}

/*
 *  avlrightgrown: helper function for avlinsert
 *
 *  See avlleftgrown for details.
 */
AVLRES
avlrightgrown(struct avlnode **n)
{
	switch ((*n)->skew) {
	case LEFT:					
		(*n)->skew = NONE;
		return OK;

	case RIGHT:
		if ((*n)->right->skew == RIGHT) {	
			(*n)->skew = (*n)->right->skew = NONE;
			avlrotleft(n);
		}
		else {
			switch ((*n)->right->left->skew) {
			case RIGHT:
				(*n)->skew = LEFT;
				(*n)->right->skew = NONE;
				break;

			case LEFT:
				(*n)->skew = NONE;
				(*n)->right->skew = RIGHT;
				break;

			default:
				(*n)->skew = NONE;
				(*n)->right->skew = NONE;
			}
			(*n)->right->left->skew = NONE;
			avlrotright(& (*n)->right);
			avlrotleft(n);
		}
		return OK;

	default:
		(*n)->skew = RIGHT;
		return BALANCE;
	}
}

/*
 *  avlinsert: insert a node into the AVL tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.
 *
 *    k           Key for the item to be inserted.
 *
 *    d           Item to be inserted.
 *
 *  Return values:
 *
 *    nonzero     The item has been inserted. The excact value of 
 *                nonzero yields is of no concern to user code; when
 *                avlinsert recursively calls itself, the number 
 *                returned tells the parent activation if the AVL tree 
 *                may have become unbalanced; specifically:
 *
 *      OK        None of the subtrees of the node that n points to 
 *                has grown, the AVL tree is valid.
 *
 *      BALANCE   One of the subtrees of the node that n points to 
 *                has grown, the node's "skew" flag needs adjustment,
 *                and the AVL tree may have become unbalanced.
 *
 *    zero        The datum provided could not be inserted, either due 
 *                to AVLKEY collision (the tree already contains another
 *                item with which the same AVLKEY is associated), or
 *                due to insufficient memory.
 */   
AVLRES
avlinsert(struct avlnode **n, AVL_key_t k, void *d)
{
	AVLRES tmp;

	if (!(*n)) {
		(*n) = emalloc(sizeof(struct avlnode));

		(*n)->left = (*n)->right = NULL;
        (*n)->k = k;
		(*n)->d = d;
		(*n)->skew = NONE;
        DPRINT(DBG_AVL, "AVL: inserting new node %p {%u, %p}", 
                *n, (*n)->k, (*n)->d);
		return BALANCE;
	}
	
	if (k < (*n)->k) {
		if ((tmp = avlinsert(& (*n)->left, k, d)) == BALANCE) {
			return avlleftgrown(n);
		}
		return tmp;
	}
	if (k > (*n)->k) {
		if ((tmp = avlinsert(& (*n)->right, k, d)) == BALANCE) {
			return avlrightgrown(n);
		}
		return tmp;
	}
	return ERROR;
}

/*
 *  avlleftshrunk: helper function for avlremove and avlfindlowest
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node. The node's left
 *                subtree has just shrunk due to item removal; its
 *                "skew" flag needs adjustment, and the local tree
 *                (the subtree of which this node is the root node) may
 *                have become unbalanced.
 *
 *   Return values:
 *
 *    OK          The parent activation of the avlremove activation
 *                that called this function may assume the entire
 *                tree is valid.
 *
 *    BALANCE     Do not assume the entire tree is valid.
 */                
AVLRES
avlleftshrunk(struct avlnode **n)
{
	switch ((*n)->skew) {
	case LEFT:
		(*n)->skew = NONE;
		return BALANCE;

	case RIGHT:
		if ((*n)->right->skew == RIGHT) {
			(*n)->skew = (*n)->right->skew = NONE;
			avlrotleft(n);
			return BALANCE;
		}
		else if ((*n)->right->skew == NONE) {
			(*n)->skew = RIGHT;
			(*n)->right->skew = LEFT;
			avlrotleft(n);
			return OK;
		}
		else {
			switch ((*n)->right->left->skew) {
			case LEFT:
				(*n)->skew = NONE;
				(*n)->right->skew = RIGHT;
				break;

			case RIGHT:
				(*n)->skew = LEFT;
				(*n)->right->skew = NONE;
				break;

			default:
				(*n)->skew = NONE;
				(*n)->right->skew = NONE;
			}
			(*n)->right->left->skew = NONE;
			avlrotright(& (*n)->right);
			avlrotleft(n);
			return BALANCE;
		}

	default:
		(*n)->skew = RIGHT;
		return OK;
	}
}

/*
 *  avlrightshrunk: helper function for avlremove and avlfindhighest
 *
 *  See avlleftshrunk for details.
 */
AVLRES
avlrightshrunk(struct avlnode **n)
{
	switch ((*n)->skew) {
	case RIGHT:
		(*n)->skew = NONE;
		return BALANCE;

	case LEFT:
		if ((*n)->left->skew == LEFT) {
			(*n)->skew = (*n)->left->skew = NONE;
			avlrotright(n);
			return BALANCE;
		}
		else if ((*n)->left->skew == NONE) {
			(*n)->skew = LEFT;
			(*n)->left->skew = RIGHT;
			avlrotright(n);
			return OK;
		}
		else {
			switch ((*n)->left->right->skew) {
			case LEFT:
				(*n)->skew = RIGHT;
				(*n)->left->skew = NONE;
				break;

			case RIGHT:
				(*n)->skew = NONE;
				(*n)->left->skew = LEFT;	
				break;
			
			default:
				(*n)->skew = NONE;
				(*n)->left->skew = NONE;
			}
			(*n)->left->right->skew = NONE;
			avlrotleft(& (*n)->left);
			avlrotright(n);
			return BALANCE;
		}

	default:
		(*n)->skew = LEFT;
		return OK;
	}
}

/*
 *  avlfindhighest: replace a node with a subtree's highest-ranking item.
 *
 *  Parameters:
 *
 *    target      Pointer to node to be replaced.
 *
 *    n           Address of pointer to subtree.
 *
 *    res         Pointer to variable used to tell the caller whether
 *                further checks are necessary; analog to the return
 *                values of avlleftgrown and avlleftshrunk (see there). 
 *
 *  Return values:
 *
 *    1           A node was found; the target node has been replaced.
 *
 *    0           The target node could not be replaced because
 *                the subtree provided was empty.
 *
 */
int
avlfindhighest(struct avlnode *target, struct avlnode **n, AVLRES *res)
{
	struct avlnode *tmp;

	*res = BALANCE;
	if (!(*n)) {
		return 0;
	}
	if ((*n)->right) {
		if (!avlfindhighest(target, &(*n)->right, res)) {
			return 0;
		}
		if (*res == BALANCE) {
			*res = avlrightshrunk(n);
		}
		return 1;
	}
    target->k  = (*n)->k;
	target->d  = (*n)->d;
	tmp = *n;
	*n = (*n)->left;
	free(tmp);
	return 1;
}

/*
 *  avlfindlowest: replace node with a subtree's lowest-ranking item.
 *
 *  See avlfindhighest for the details.
 */
int
avlfindlowest(struct avlnode *target, struct avlnode **n, AVLRES *res)
{
	struct avlnode *tmp;

	*res = BALANCE;
	if (!(*n)) {
		return 0;
	}
	if ((*n)->left) {
		if (!avlfindlowest(target, &(*n)->left, res)) {
			return 0;
		}
		if (*res == BALANCE) {
			*res =  avlleftshrunk(n);
		}
		return 1;
	}
    target->k = (*n)->k;
	target->d = (*n)->d;
	tmp = *n;
	*n = (*n)->right;
	free(tmp);
	return 1;
}

/*
 *  avlremove: remove an item from the tree.
 *
 *  Parameters:
 *
 *    n           Address of a pointer to a node.
 *
 *    key         Key of item to be removed.
 *
 *  Return values:
 *
 *    nonzero     The item has been removed. The exact value of 
 *                nonzero yields if of no concern to user code; when
 *                avlremove recursively calls itself, the number
 *                returned tells the parent activation if the AVL tree
 *                may have become unbalanced; specifically:
 *
 *      OK        None of the subtrees of the node that n points to
 *                has shrunk, the AVL tree is valid.
 *
 *      BALANCE   One of the subtrees of the node that n points to
 *                has shrunk, the node's "skew" flag needs adjustment,
 *                and the AVL tree may have become unbalanced.
 *
 *   zero         The tree does not contain an item yielding the
 *                AVLKEY value provided by the caller.
 */
AVLRES
avlremove(struct avlnode **n, AVL_key_t key)
{
	AVLRES tmp = BALANCE;

	if (!(*n)) {
		return ERROR;
	}
	if (key < (*n)->k) {
		if ((tmp = avlremove(& (*n)->left, key)) == BALANCE) {
			return avlleftshrunk(n);
		}
		return tmp;
	}
	if (key > (*n)->k) {
		if ((tmp = avlremove(& (*n)->right, key)) == BALANCE) {
			return avlrightshrunk(n);
		}
		return tmp;
	}
	if ((*n)->left) {
		if (avlfindhighest(*n, &((*n)->left), &tmp)) {
			if (tmp == BALANCE) {
				tmp = avlleftshrunk(n);
			}
			return tmp;
		}
	}
	if ((*n)->right) {
		if (avlfindlowest(*n, &((*n)->right), &tmp)) {
			if (tmp == BALANCE) {
				tmp = avlrightshrunk(n);
			}
			return tmp;
		}
	}
	free(*n);
 	*n = NULL;
	return BALANCE;
}

/*
 *  avlaccess: retrieve the datum corresponding to a given AVLKEY.
 *
 *  Parameters:
 *
 *    n           Pointer to the root node.
 *
 *    key         KEY of item to be accessed.
 *
 *  Return values:
 *
 *    non-NULL    An item yielding the AVLKEY provided has been found,
 *                the return value points to the AVLKEY attached to it.
 *
 *    NULL        The item could not be found.
 */
void *
avlaccess(struct avlnode *n, AVL_key_t key)
{
    
    DPRINT(DBG_AVL, "avlaccess(%p, %u)", n, key);
    if (!n) {
        return NULL;
    }
    if (key < n->k) {
        return avlaccess(n->left, key);
    }
    if (key > n->k) {
        return avlaccess(n->right, key);
    }
    return n->d;
}

/*
 *  avldepthfirst: depth-first tree traversal.
 *
 *  Parameters:
 *
 *    n          Pointer to the root node.
 *
 *    f          Worker function to be called for every node.
 *
 *    param      Additional parameter to be passed to the
 *               worker function
 *
 *    depth      Recursion depth indicator. Allows the worker function
 *               to determine how many levels the node being processed
 *               is below the root node. Can be used, for example,
 *               for selecting the proper indentation width when
 *               avldepthfirst ist used to print a tree dump to
 *               the screen.
 *
 *               Most of the time, you will want to call avldepthfirst
 *               with a "depth" value of zero.
 */
void
avldepthfirst(struct avlnode *n, AVL_worker_t *f, int param, int depth)
{
	if (!n) return;
	avldepthfirst(n->left, f, param, depth + 1);
	(*f)(n->d, param, depth);
	avldepthfirst(n->right, f, param, depth +1);
}

/*
 *  avlbreadthfirst: breadth-first tree traversal.
 * 
 *  See avldepthfirst for details.
 */
/* XXX Need to rewrite with the List structure rather than use the queue
 * that is by the same person.
 */
#if 0
void
avlbreadthfirst(struct avlnode *n, AVLWORKER *f, int param)
{
        struct queue q;

        if (!n) return;
        qinit(&q);
        qinsert(&q, n);
        while (qremove(&q, &n)) {
                (*f)(n, param, 0);
                if (n->left) qinsert(&q, n->left);
                if (n->right) qinsert(&q, n->right);
        }
}
#endif

/* 
 * depth first traversal and free function. This will not free the root
 * node. that is the caller's problem. 
 */
void 
avl_free(struct avlnode *n)
{
    if (n == NULL) return;
    avl_free(n->left);
    free(n->left);
    avl_free(n->right);
    free(n->right);
}

/* Public Intereface functions */
AVLTree
AVL_create_priv(DMS_EMPTY_PARAMS)
{
    AVLTree     tree;

    tree = DMSmalloc(sizeof(struct AVL_tree)); 

    tree->root = NULL;

    return tree;
}

void
AVL_destroy(AVLTree tree)
{

    if (tree->root != NULL) {
        /* depth first recursion to free all the nodes */
        avl_free(tree->root);
        free(tree->root);       /* not free()ed by avl_free() */
    }

    free(tree);             /* now free the whole tree */
}

bool
AVL_insert(AVLTree t, AVL_key_t k, void *d)
{
    DPRINT(DBG_AVL, "AVL_insert(%p, %u, %p)", t, k, d);
    return avlinsert(&(t->root), k, d) != 0;
}

bool
AVL_remove(AVLTree t, AVL_key_t key)
{
    return avlremove(&(t->root), key) != 0;
}

void *
AVL_find(AVLTree t, AVL_key_t key)
{
    DPRINT(DBG_AVL, "AVL_find(%p, %u)", t, key);
    return avlaccess(t->root, key);
}

void
AVL_depthfirst(AVLTree t, AVL_worker_t *f, int param, int depth)
{
    avldepthfirst(t->root, f, param, depth);
}


