/**
 * @file avltree.h
 *
 * @brief An AVL tree container, that uses unsigned integers as the key
 * type.  The data type store is void* for flexibility.
 * @par
 * This code was taken from http://www.purists.org/avltree.
 * Modified by Devin Kowatch <devink@sdsc.edu> for SDSCSyslogd.
 * @par
 * I Made this a lot more generic than it was originally, however, the
 * key type is still expected to be an integer (either signed or
 * unsigned).  I have not changed that because there is no need to yet,
 * and doing so requires extra overhead for function calls/pointers.
 *
 * $Id: avltree.h,v 1.2.2.1 2003/04/02 01:04:38 devink Exp $
 */
#ifndef UTIL_AVLTREE_H
#define UTIL_AVLTREE_H

#include "types.h"
#include "dmalloc_sup.h"

/** AVL Tree Object */
typedef struct AVL_tree *AVLTree;

/** The type of a KEY in the AVL Tree */
typedef uint32_t AVL_key_t; 

/**
 * Create a new AVLTree Object.
 *
 * @return The new AVLTree object.
 */
#define AVL_create()    AVL_create_priv(DMS_EMPTY_ARGS(__FILE__, __LINE__))

/** private function that actually creates the AVL tree */
AVLTree AVL_create_priv(DMS_EMPTY_PARAMS);

/**
 * Destroy an AVLTree Object.  This will free every node, but will not
 * free the data stored in that node.
 *
 * @param tree The AVLTree to destroy.
 */
void AVL_destroy(AVLTree tree);

/**
 * insert a node into the AVL tree.
 *
 * @param tree  AVLTree object
 * @param key   Key to be inserted.
 * @param data  Item to be inserted.
 * @return
 *    TRUE      The item has been inserted. 
 *    FALSE     The datum provided could not be inserted, due 
 *              to \c key collision (the tree already contains another
 *              item with which the same \c key is associated)
 */   
inline bool AVL_insert(AVLTree tree, AVL_key_t k, void *data);

/**
 * remove an item from the tree.
 *
 * @param tree  AVLTree object to work on.
 * @param key   Key of item to be removed.
 * @return
 *    TRUE      The item has been removed. 
 *   FALSE      The tree does not contain an item yielding the
 *              \c key value provided by the caller.
 */
inline bool AVL_remove(AVLTree tree, AVL_key_t key);

/**
 * retrieve the datum corresponding to a given AVLKEY.
 *
 * @param tree     AVL tree object to seach in.
 * @param key      Key of item to be accessed.
 * @return
 *    non-NULL    An item yielding the AVLKEY provided has been found,
 *                the return value points to the AVLKEY attached to it.
 *    NULL        The item could not be found.
 */
inline void * AVL_find(AVLTree tree, AVL_key_t key);


/**
 * Function to be called by the tree traversal functions.
 *
 * @param n       Pointer to the data in a node.
 * @param param   Value provided by the traversal function's caller.
 * @param depth   Recursion depth indicator. Allows the function to
 *                determine how many levels the node bein processed is
 *                below the root node. Can be used, for example,
 *                for selecting the proper indentation width when
 *                avldepthfirst is used to print a tree dump to 
 *                the screen.
 */
typedef void AVL_worker_t(void *n, int param, int depth);

/**
 *  depth-first tree traversal.
 *
 * @param tree   AVLTree object to traverse.
 * @param f      Worker function to be called for every node.
 * @param param  Additional parameter to be passed to the
 *               worker function
 * @param depth  Recursion depth indicator. Allows the worker function
 *               to determine how many levels the node being processed
 *               is below the root node. Can be used, for example,
 *               for selecting the proper indentation width when
 *               avldepthfirst ist used to print a tree dump to
 *               the screen.
 *
 *               Most of the time, you will want to call avldepthfirst
 *               with a "depth" value of zero.
 */
inline void AVL_depthfirst(AVLTree tree, AVL_worker_t *f, int param, int depth);

/* XXX Not imelemented at the moment */
#if 0
/*
 *  avlbreadthfirst: breadth-first tree traversal.
 * 
 *  See avldepthfirst for details.
 */
void
avlbreadthfirst(struct avlnode *n, AVLWORKER *f, int param);
#endif 

#endif /* UTIL_AVLTREE_H */

