/**
 * @file bitvec.h
 *
 * @brief Generic Bit Vector manipulation functions.
 * @par
 * This has functions for a bit map (one integer) and bitvector
 * (multiple integers).
 * There is no realtype for the bitmap, simply use any integer large
 * enough to hold the information.
 *
 * $Id: bitvec.h,v 1.1.2.2 2003/04/30 20:04:00 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Aug 28 18:44:44 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UTIL_BITVEC_H
#define UTIL_BITVEC_H


/**
 * Set a bit in a bit map. 
 * @param map The integer that is the bit map
 * @param bit The bit number to set
 */
#define BM_SET(map, bit)  ((map) |= (1 << (bit)))

/**
 * Is the given bit set?
 * @param map The integer that is the bit map
 * @param bit The bit number to set
 */
#define BM_ISSET(map, bit) (((map) & (1 << (bit))) != 0) 

/**
 * Clear the given bit number from the bitmap
 * @param map The integer that is the bit map
 * @param bit The bit number to set
 */
#define BM_CLR(map, bit)    ((map) &= ~(1 << (bit)))

/**
 * Create an integer with the given bit set.  This is useful for ORing a
 * bunch of bits together for a mask.
 * @param bit The bit to set.
 */
#define BM_MBIT(bit) (1 << (bit))

/**
 * Take the intersection of two bitmaps.
 * @param a The first map
 * @param b The second map
 */
#define BM_ISECT(a, b)  ( (a) & (b) )

/**
 * Take the union of two bitmaps.
 *
 * @param a The first map.
 * @param b The second map.
 */
#define BM_UNION(a, b)  ( (a) | (b) )

#endif /* UTIL_BITVEC_H */
