/**
 * @file debug.c
 *
 * @brief Implement the debug support functions
 * @ingroup debug
 *
 *
 * $Id: debug.c,v 1.18.2.2 2003/04/30 20:04:00 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Apr 08 12:05:12 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include "debug.h"
#include "types.h"
#include "strcompat.h"

static char const rcsid[] 
    = "$Id: debug.c,v 1.18.2.2 2003/04/30 20:04:00 devink Exp $";

/** List of which Debuggig types to print */
static bool Debugging[CHAR_MAX];

/** file to output to */
static int dbgout = -1;

/** Process Name */
static char dbgname[31];

/** Where to start writing the process name. */
static char *procnamestart;

/** Length that the process name can be */
static int procnamelen;

/** The environment */
extern char **environ;

void
DBG_printfatal(char const *msg, char const *f, int l)
{
    char        buf[BUFSIZ];
    size_t      len;

    len = snprintf(buf, BUFSIZ, "%s %s,%d: FATAL: %s\n", dbgname, f, l, msg);

    /* if the string was truncated then adjust len properly, and add a
     * new line
     */ 
    if (len >= BUFSIZ) {
        len = BUFSIZ;
        buf[len - 1] = '\n';
    }

    /* just give up, not much can be done here */
    if (len <= 0) return;   

    /* write it out */
    write(dbgout, buf, len);
}


void 
DBG_print(char mask, char const *msg, char const *f, int l, ...)
{
    va_list     ap;
    char        buf[BUFSIZ];/* space for the message to write */
    char        *p;         /* current start of string in buf */
    size_t      wlen;       /* write length (return value) */
    size_t      blen;       /* length remaining in buf */
    bool        done;       /* loop control */

    if (! Debugging[(int)mask]) return;

    va_start(ap, l);

    /* loose some extra garbage from the file name */
    done = FALSE;
    while (! done) {
        if (f[0] == '.' && f[1] == '.' && f[2] == '/')
            f += 3;
        else if (f[0] == '.' && f[1] == '/')
            f += 2;
        else if (f[0] == 's' && f[1] == 'r' && f[2] == 'c' && f[3] == '/')
            f += 4;
        else 
            done = TRUE;
    }
    
    wlen = snprintf(buf, BUFSIZ, "%s %s,%d: ", dbgname, f, l);
    if (wlen <= 0) return;   /* just give up */
    if (wlen >= BUFSIZ) wlen = BUFSIZ;

    blen = BUFSIZ - wlen;
    p = buf + wlen;

    wlen = vsnprintf(p, blen, msg, ap);
    if (wlen >= blen) wlen = blen;

    blen -= wlen;
    p += wlen; 

    /* 
     * add the last new line, 
     * NIL can be left out as write(2) doesn't care 
     * check the position before p for the newline, *p == '\0'
     */
    if (blen >= 1 && p > buf && *(p-1) != '\n') {
        *p++ = '\n';
        blen--;

        if (blen >= 1) *p = '\0';
    }
    else {
        p--;    /* truncate to fit in the newline */
        *p = '\n';
    }

    /* write it out, blen is free space in the buffer */
    wlen = write(dbgout, buf, BUFSIZ - blen);
}

void 
DBG_printBuf(char mask, char const *msg, char const *f, int l, 
        struct IPC_buffer const *ibuf, int blen)
{
#define USE_BUF(tlen, wlen) \
        if (tlen < wlen) goto done; \
        tlen -= wlen 

    int     i;
    char    tmp[BUFSIZ];
    char    *tp;
    char    *bp;
    size_t  wlen;
    int     tlen;   /* signed so that if it goes < 0 we don't get fucked */
    int     c;

    /* make sure we want to print these */
    if (! Debugging[(int)mask]) return;

    /* figure out how much to print */
    if (blen < 0 || blen > ibuf->size)
        blen = ibuf->size;

    bp = ibuf->data;

    tlen = BUFSIZ - 1;  /* reserve space for the NL */
    tp = &(tmp[0]);
    
    wlen = snprintf(tp, BUFSIZ, "%s %s,%d: %s {", dbgname, f, l, msg);
    if (wlen <= 0) return;  /* give up */

    USE_BUF(tlen, wlen);
    tp += wlen;

    if (ibuf->data == NULL) {
        wlen = strlcpy(tp, "NULL", tlen);
        USE_BUF(tlen, wlen);
        tp += wlen;
    }
    else {
        for (i = 0; i < blen; i++, bp++) {
            if ((i % 32) == 0 ) {
                USE_BUF(tlen, 2);
                *tp++ = '\n';
                *tp++ = '\t';
            }
            else {
                USE_BUF(tlen, 1);
                *tp++ = ' ';
            }

            if (isalnum(*bp) || ispunct(*bp) ) {
                USE_BUF(tlen, 1);
                *tp++ = *bp;
            }
            else {
                USE_BUF(tlen, 2);
                c = ((int)*bp & 0xF0) >> 4;
                *tp++ = (c > 9) ? 'A' + c - 10 : '0' + c;
                c = ((int)*bp & 0x0F);
                *tp++ = (c > 9) ? 'A' + c - 10 : '0' + c;
            }
        }
    }
    USE_BUF(tlen, 1);
    *tp++ = '}';

    /* This is the label that is jumped to when we run out of buffer
     * space.  We add a new line and write the buffer.  
     */
done:
    /* Add the NL */
    *tp = '\n';

    /* Write it, tlen is space _left_ in the buffer */
    write(dbgout, tmp, BUFSIZ - tlen);

#undef USE_BUF
}


bool DBG_init(char const *mask, char const *ofile)
{
    char const  *p;

    memset(Debugging, 0, sizeof(Debugging));
    memset(dbgname, 0, sizeof(dbgname));

    /* only run the rest if there is something to print */
    if (mask == NULL || *mask == '\0') {
        dbgout = STDOUT_FILENO;    /* just in case */
        return TRUE;
    }

    p = mask;
    while (*p != '\0') {
        if (*p < 0) continue;    /* NEXT */

        Debugging[(int)*p] = TRUE;

        p++;
    }

    DBG_newproc("Main");

    Debugging[(int)DBG_DBG] = TRUE; /* always send these */

    if (strcmp(ofile, "-") == 0) {
        dbgout = STDOUT_FILENO;
    }
    else {
        dbgout = open(ofile, O_WRONLY | O_CREAT | O_TRUNC,
                S_IRUSR | S_IWUSR | S_IRGRP);
        if (dbgout == -1) {
            return FALSE; 
        }
    }

    DPRINT(DBG_DBG, "Debugging initilized (mask = \"%s\")", mask);

    return TRUE;
}


void 
DBG_newproc(char const *name)
{
    pid_t   pid;

    pid = getpid();

    snprintf(dbgname, sizeof(dbgname),
            "%s(%u)", name, pid);
}

bool
DBG_isOutputFD(int fd)
{
    return fd >= 0 && fd == dbgout;
}

void
DBG_initProcName(int argc_p,char **argv_p)
{
	int i=0;
	char *endWrite=argv_p[0];

	/* init our static globals */
	procnamelen=0;
	procnamestart=argv_p[0];

	/* try to be a bit smart about this; check to make sure we're not
	 * skipping some memory as we go */
	for(i=0;i<argc_p;i++)
	{
		if(argv_p[i]!=endWrite)
			break;
		
		procnamelen+=strlen(argv_p[i])+1;
		endWrite+=strlen(argv_p[i])+1;
	}
		
	i=0;
	while(environ[i]!=NULL)
	{
		if(environ[i]!=endWrite)
			break;

		procnamelen+=strlen(environ[i])+1;
		endWrite+=strlen(environ[i])+1;

		i++;
	}
}

void
DBG_setProcName(char *name)
{
	memset(procnamestart,0,procnamelen);

	strlcpy(procnamestart,name,procnamelen);
}

