/**
 * @file exiterr.c
 *
 * @brief Functions to exit gracefully in the face of an error.
 * @par
 * This is not in error.c becasue I need to be able to not include these
 * definitions in unit tests.
 *
 * $Id: exiterr.c,v 1.10.2.6 2003/04/30 20:04:01 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue Jun 18 18:42:09 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>

#include "error.h"
#include "test.h"

#include "log.h"
#include "controller/ctlr_destroy.h"
#include "privs.h"

static char const rcsid[] 
    = "$Id: exiterr.c,v 1.10.2.6 2003/04/30 20:04:01 devink Exp $";

/** 
 * A structure to contain all the info needed for cleaning an input
 * module.
 */
struct EXIT_imInfo {
    bool    active;
    INW     *waiter;
    IPC     *ipc;
};

/** The CTLR object to clean up on exit */
static CTLR *exit_ctlr = NULL;

/** The input info for cleaning at exit */
static struct EXIT_imInfo  exit_im = {FALSE, NULL, NULL};

/** The INW object to cleanup on exit */
static INW *exit_inw;

/**
 * Cleanup the exit_ctlr object.
 *
 * @param code The exit code
 */
static void cleanController(int code);

/**
 * Cleanup the exit_im object.
 *
 * @param code The exit code.
 */
static void cleanIM(int code);

void
EXIT(int code)
{
    bool    noclean;

    noclean = (code & EXIT_NOCLEAN) != 0;
    code &= ~EXIT_NOCLEAN;

	/* we don't want to clean; exit ASAP */
	if(noclean)
		exit(code);

    /* log that we are exiting */
    if (code == EXIT_NORMAL) {
        if (! IN_UNIT_TEST_CT) {
            LOG_print(LOG_INFO, "Exiting normally");
        }
    }
    else {
        /* THis relies on the compiler and linke to DTRT and avoid
         * undefined symbol complaints.
         */
        if (IN_UNIT_TEST_CT) {
            LOG_print(LOG_ERROR, "Exiting with error code %d", code);
        }
        else {
            char    buf[1024];
            snprintf(buf, sizeof(buf), "Exiting with error code %d", code);
            UT_FAIL(buf);
        }
    }


    if (exit_ctlr != NULL) {
        cleanController(code);
    }
    else if (exit_im.active) {
        cleanIM(code);
    }

    /* common cleanup actions */
    if (code < EXIT_ERROR) {
        if (exit_inw != NULL)
            INW_destroy(&exit_inw);
    }

    exit(code);
}

void
EXIT_setCTLR(CTLR *ctlr)
{
    exit_ctlr = ctlr;
}

void
EXIT_initIM(void)
{
    exit_ctlr = NULL;

    memset(&exit_im, 0, sizeof(exit_im));
    exit_im.active = TRUE;
}

void
EXIT_setIM_IPC(IPC *ipc)
{
    exit_im.ipc = ipc;
}

void
EXIT_setINW(INW *inw)
{
    exit_inw = inw;
}

static void
cleanIM(int code)
{

    if (exit_im.ipc != NULL) 
        IPC_inputmoduleExit(exit_im.ipc);

    if (code < EXIT_ERROR) {
        if (exit_im.ipc != NULL) 
            IPC_destroy(exit_im.ipc);
    }
}

static void 
cleanController(int code)
{
#if ! IN_UNIT_TEST_CT
    CTLR_killIMs(exit_ctlr);

    CTLR_destroyOMs(exit_ctlr);

    /* remove the PID file */
    PRIV_unlink(exit_ctlr->pidPath);

    if (code < EXIT_IMMED) {
        CTLR_shutdownSockets(exit_ctlr);
        CTLR_shutdownIPC(exit_ctlr);
    }

    if (code < EXIT_ERROR) {
        CTLR_freeMemory(exit_ctlr);
    }
#endif /* ! IN_UNIT_TEST_CT */
}
