/* 
 * File: hashtable.c
 *
 * Implement a hash table
 *
 * Author: Devin Kowatch <devink@webengruven.org>
 * $Id: hashtable.c,v 1.4.2.2 2003/04/30 20:04:01 devink Exp $
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#if HAVE_CONFIG_H
# include "acconfig.h"
#endif

#if HAVE_VALUES_H
# include <values.h>
#endif

#include <stddef.h>
#include <stdlib.h>
#include "types.h"

#include "hashtable.h"

#include "error.h"
#include "dmalloc_sup.h"

#if WITH_DMALLOC
# include <dmalloc.h>
#endif

static char const rcsid[] = 
    "$Id: hashtable.c,v 1.4.2.2 2003/04/30 20:04:01 devink Exp $";

/** An element in the hashtable */
struct HTElem {
    void *key;
    void *data;
    struct HTElem   *next;
}; 

/** 
 * HashTable object type.  This should be treated as opaque by clients of
 * this lirbary.  It may change without warning.
 * @internal
 */ 
struct HT_object {
    struct HTElem   **table;
    size_t          size;
    size_t          elemCount;
    int             hashShift;
    HT_hashFunc     hash;
    HT_keyCmpFunc   keycmp;
};

/** 
 * HashTable Iterator object.  This contains all the data needed to
 * iterate over a hash table.
 * @internal
 */
struct HTI_object {
    struct HTElem       *curElem;   /**< current HTElem */
    int                 nextIndex;  /**< next index into htable->table */
    struct HT_object    *htable;    /**< hashtable that we are iterating */
};

/* Incase this has not been defined yet */
#ifndef BITSPERBYTE
# define BITSPERBYTE CHAR_BIT
#endif

/** Number of bits in a hash table key */
#define KEYBITS     (sizeof(HT_key_t) * BITSPERBYTE)

/** 
 * Used to figure out HT_KEY_MULTIPLIER.
 * Note that KEYMAXPLUSONE is defined on it's own line both for space
 * considerations (fitting HT_KEY_MULTIPLIER on one line) and because it may
 * change due to portability issues (only works on gcc).
 * @warning The scope of this macro is very limited.  It can only be used in
 * the definition of HT_KEY_MULTIPLIER
 */
#define KEYMAXPLUSONE   (uint64_t)UINT_MAX + 1LL

/** 
 * Constant used is hash function.  This was derived from suggestions in
 * _The Art of Computer Programming_ Vol. 3 by Knuth.
 *
 * A = golden ratio = 0.6180339887
 * hashConst = floor(A * 2^KEY_BITS)
 * should be 2654435769;
 */
#define HT_KEY_MULTIPLIER (HT_key_t)(0.61803398875 * KEYMAXPLUSONE)

/**
 * Hash the key.
 *
 * @param ht The hashtable to work on
 * @param key The key.
 * @return the hashtable index.
 */
static HT_key_t doHash(HashTable ht, void *key);

/**
 * Find an element in the hash table by key.
 *
 * @param ht The HashTable to search
 * @param key The key to look for.
 * @param hkout The hash index (output)
 * @param prevOut The previous element (output)
 * @return The element that is mapped to \c key
 */
static struct HTElem * doFind(HashTable ht, void *key, HT_key_t *hkout, 
                              struct HTElem **prevOut);

HashTable 
HT_create(size_t size, HT_hashFunc hash, HT_keyCmpFunc keycmp)
{
    HashTable       ht;
    unsigned int    mask = 1;

    ht = NULL;

    ht = emalloc(sizeof(struct HT_object));
    memset(ht, 0, sizeof(struct HT_object));

    /* check the size to make sure that it is a power of two and setup some
     * options that will be needed later.
     */
    ht->hashShift = KEYBITS;
    while ( (mask & size) == 0 && ht->hashShift >= 0) {
        ht->hashShift--;
        mask <<= 1;
    }

    /* make sure that mask was a power of 2 */
    if ( (~mask & size) != 0) goto error;

    /* setup the private data */
    ht->size = size;
    ht->hash = hash;
    ht->keycmp = keycmp;

    /* allocate space for the table */
    ht->table = emalloc(sizeof(struct HTElem*) * size);
    memset(ht->table, 0, sizeof(struct HTElem*) * size);

    return ht;

error:
    HT_destroy(ht);
    return NULL;
}

void
HT_destroy(HashTable ht)
{
    int             i;
    struct HTElem   *cur, *last;

    if (ht != NULL) {
        if (ht->table != NULL) {
            for (i = 0; i < ht->size; i++) {
                cur = ht->table[i];
                while (cur != NULL) {
                    last = cur;
                    cur = cur->next;
                    free(last);
                }
            }
            free(ht->table);
        }
        free(ht);
    }
}

bool
HT_insert(HashTable ht, void *key, void *data)
{
    HT_key_t        hkey;
    struct HTElem   *elem, *p;

    assert(ht != NULL);


    hkey = doHash(ht, key);
    if (ht->table[hkey] != NULL) {
        for (p = ht->table[hkey]; p != NULL; p = p->next) {
            if (ht->keycmp(key, p->key) == 0) {
                return FALSE;   /* already present */
            }
        }
    }

    /* add the new element */
    elem = emalloc(sizeof(struct HTElem));

    elem->next = ht->table[hkey];
    elem->key = key;
    elem->data = data;

    ht->table[hkey] = elem;
    ht->elemCount++;

    return TRUE;
}

void *
HT_find(HashTable ht, void *key)
{
    struct HTElem   *elem;

    elem = doFind(ht, key, NULL, NULL);
    if (elem == NULL) return NULL;
    
    return elem->data;
}

void *
HT_remove(HashTable ht, void *key)
{
    void            *data;
    HT_key_t        hkey;
    struct HTElem   *elem, *prev;

    elem = doFind(ht, key, &hkey, &prev);

    if (elem == NULL) return NULL;

    data = elem->data;

    if (prev != NULL) prev->next = elem->next;
    else ht->table[hkey] = elem->next;

    ht->elemCount--;

    free(elem);

    return data;
}

bool
HT_isEmpty(HashTable ht)
{
    return ht->elemCount == 0;
}

int
HT_size(HashTable ht)
{
    return ht->elemCount;
}

HTIterator
HT_iterator(HashTable ht)
{
    HTIterator  itr;

    itr = emalloc(sizeof(struct HTI_object));
    memset(itr, 0, sizeof(struct HTI_object));

    itr->htable = ht;

    return itr;
}

void *
HTI_next(HTIterator itr)
{

    /* try advancing down the list in the current bucket first */
    if (itr->curElem != NULL) {
        itr->curElem = itr->curElem->next;
    }

    /* check along the table until we find an element, or reach the end */
    while (itr->curElem == NULL && itr->nextIndex < itr->htable->size) {
        itr->curElem = itr->htable->table[itr->nextIndex];
        itr->nextIndex++;
    }

    /* what even we have now is the return */
    if (itr->curElem == NULL) 
        return NULL;

    return itr->curElem->data;
}

void *
HTI_elem(HTIterator itr)
{
    if (itr->curElem == NULL)
        return NULL;

    return itr->curElem->data;
}

void
HTI_destroy(HTIterator *itr)
{
    free(*itr);
    itr = NULL;
}

static struct HTElem *
doFind(HashTable ht, void *key, HT_key_t *hkout, struct HTElem **prevOut)
{
    struct HTElem   *p, *prev;
    int             hkey;

    hkey = doHash(ht, key);
    
    prev = NULL;
    p = NULL;
    for (p = ht->table[hkey]; p != NULL; p = p->next) {
        if (ht->keycmp(p->key, key) == 0) {
            break;
        }
        prev = p;
    }

    if (hkout != NULL) *hkout = hkey;
    if (prevOut != NULL) *prevOut = prev;

    return p;
}

static HT_key_t
doHash(HashTable ht, void *key)
{
    HT_key_t    hkey; 
    assert(ht != NULL);

    hkey = ht->hash(key);
    hkey *= HT_KEY_MULTIPLIER;
    hkey >>= ht->hashShift;

    return hkey;
}

#define USE_CHAR_BITS       4       /* bits from each char to use in hash */
#define USE_CHAR_BITS_MASK  0x0F    /* mask to get USE_CHAR_BITS */
/* number of characters that can be packed into one word (using only
 * USE_CHAR_BITS from each char).  This number is used to do a MOD operation.
 */
#define CHARS_PER_WORD_MOD  ((KEYBITS / USE_CHAR_BITS) - 1 )
HT_key_t
HT_stringHash(void *keystr)
{
    HT_key_t        key = 0;
    unsigned int    word = 0;
    int             i;
    char const      *p;

    if (keystr == NULL) return -1;

    i = 1;
    p = (char const *)keystr;
    while (*p != '\0') {
        word <<= USE_CHAR_BITS;
        word |= (int)*p & USE_CHAR_BITS_MASK;

        if ( (i & CHARS_PER_WORD_MOD) == 0) {
            key ^= word;
            word = 0;
        }
        i++;
        p++;
    }

    if (word != 0) key ^= word;

    return key;
}

int
HT_stringCmp(void *lhs, void *rhs)
{
    return strcmp((char const*)lhs, (char const*)rhs);
}

