/**
 * @file hashtable.h
 * @brief Funtions and Data object for a HashTable container.
 * @par
 * This implements a hash table container that can store (key, value) pairs of
 * aribtrary types.  For this to work, each hash table must be constructed
 * with a function that takes a key and generates an integer that can then be
 * used to compute the position in the hash table.  It also requires a
 * function to compare two keys for equality.
 * @par
 * The hash table is is allocated at creation time, and will not grow or
 * shrink.  The method for handling collisions is open chaining.
 * @par
 * Key data and Value data are given to the table as pointers.  They are
 * treated as opaque except for the keycmp function and the hash function,
 * which are user supplied.
 * @warning (key, value) data are not deallocated, nor are they copied.  
 * @par 
 * Hashing and comparison functions are supplied for a hash table that uses
 * asciz strings as the key.
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 *
 * $Id: hashtable.h,v 1.3.2.6 2003/04/30 20:04:02 devink Exp $
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UTIL_HASHTABLE_H
#define UTIL_HASHTABLE_H

#include <stddef.h>
#include "types.h"

/**@defgroup hashtbl Hash Table Container
 * @{
 */

/** The type of a hash key. */
typedef uint32_t    HT_key_t;

/** 
 * The type of a hash function. 
 *
 * The hash function must generate a key based on the keydata.  The key should
 * try to be unique for each value of keydata.  They type of keydata is 
 * dependant on the hash function in question (meaning each hash function 
 * must work for a particular type of keydata).
 *
 * @param keydata The keydata to hash
 * @return The hash key.
 */
typedef HT_key_t (*HT_hashFunc)(void *keydata);

/**
 * Compare two hash table keys.
 *
 * @param lhs The lhs key
 * @param rhs The rhs key
 * @return 0 if lhs == rhs
 *         <0 if lhs < rhs
 *         >0 if lhs > rhs
 */
typedef int (*HT_keyCmpFunc)(void *lhs, void *rhs);


/** HashTable data type used by clients.  This must be treated as opaque. */
typedef struct HT_object *HashTable;

/** HashTable Iterator type */
typedef struct HTI_object   *HTIterator;

/** 
 * Create a new hashtable.
 * The \c tblsize is the number of elements in the hash table.  The table it's
 * self cannot hold more elements than this without having at least one slot
 * with a chain depth of two.
 *
 * @param tblsize The total size of the hash table, must be a power of two.
 * @param hash The function used to hash void* keys into HT_key_t keys.
 * @param keycmp The function to comapre two keys.
 * @return The new HashTable object.
 */
HashTable HT_create(size_t tblsize, HT_hashFunc hash, HT_keyCmpFunc keycmp);

/**
 * Destroy a hashtable.
 *
 * @param ht The hash table to destroy.
 */
void HT_destroy(HashTable ht);

/** 
 * Insert an element in to the hash table.  There can be only one
 * element per key, so FALSE is returned if \c key is already present 
 * in the table.
 *
 * @param ht The HashTable to work on.
 * @param key The search key.
 * @param val The value.
 * @return TRUE on sucess, FALSE on failure.
 */
bool HT_insert(HashTable ht, void *key, void *data);

/** 
 * Find an element in the hash table
 *
 * @param key The search key.
 * @return the value for that key, or NULL on error or not found.
 */
void *HT_find(HashTable ht, void *key);

/**
 * Remove an element in the hash table.
 *
 * @param key The key of the element to remove.
 * @return the value that is going to be removed, or NULL on error.
 */
void * HT_remove(HashTable ht, void *key);

/**
 * Is the hash table empty?
 *
 * @parma ht The HashTable object to examine.
 * @return TRUE if the hash table is empty, FALSE if not.
 */
bool HT_isEmpty(HashTable ht);

/**
 * Get the number of elements in the hash table.
 *
 * @parma ht The HashTable object to examine.
 * @return The number of elements in the hash table.
 */
int HT_size(HashTable ht);

/**
 * Get an iterator for the hash table 
 *
 * @param ht The HashTable to iterate over
 * @return A HTIterator for \c ht
 */
HTIterator HT_iterator(HashTable ht);

/*@}*/

/**@name Hash Table Iterator Operations
 * @ingroup hashtable
 * @{
 */

/**
 * Advance the iterator to the next element.
 *
 * @param itr The iterator to work on.
 * @return The next element, or NULL.
 */
void *HTI_next(HTIterator itr);

/** 
 * Get the element the iterator is currently pointing to.
 *
 * @param itr The iterator to work on.
 * @return The current element.
 */
void *HTI_elem(HTIterator itr);

/**
 * Destroy a Hash Table iterator.
 *
 * @param A Pointer to the HTIterator to destroy.
 */
void HTI_destroy(HTIterator *itr);

/*@}*/

/**@name String Keyed HashTable functions
 * @ingroup hashtable
 * @{
 */

/**
 * Create a new hash table that uses strings as the key type.
 *
 * @see HT_create
 * @param tblsize The total size of the hash table, must be a power of two.
 * @return A new HashTable object or NULL on error.
 */
#define HTStr_create(size)  HT_create(size, HT_stringHash, HT_stringCmp)

/** 
 * Hash function for string keys. 
 * 
 * @see HT_hashFunc
 *
 * @param keydata The string to hash
 * @return The hashed value of \c keydata 
 */
HT_key_t HT_stringHash(void *keydata);

/** 
 * Compare function for string keys 
 *
 * @see HT_keyCmpFunc
 *
 * @param lhs The lhs key
 * @param rhs The rhs key
 * @return 0 if lhs == rhs
 *         <0 if lhs < rhs
 *         >0 if lhs > rhs
 */
int HT_stringCmp(void *lhs, void *rhs);


/*@}*/

#endif /* UTIL_HASHTABLE_H */
