/**
 * @file list.c
 *
 * @brief Implementation of the Likned List container.
 * @par
 * This is a double linked list.
 *
 * The list structure is actually circular with a "center" node, in
 * which the next pointer is the head of the list, and the prev pointer
 * is the tail of the list. This center node can be identified by checking
 * LL_state::nodes for that list, there is a macro to make this clear.
 * Further, the head and tail have pointers back to the center node,
 * making some things _much_ easier.
 *
 * @ingroup list
 *
 *
 * $Id: list.c,v 1.10.2.2 2003/04/30 20:04:02 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Feb 01 15:01:43 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <stdlib.h>

#include "strcompat.h"

#include "list.h"
#include "error.h"

static char const rcsid[] 
    = "$Id: list.c,v 1.10.2.2 2003/04/30 20:04:02 devink Exp $";

/**
 * This is the data that is stored for lists (i.e. size).
 */
struct LL_data {
    size_t  size;
};

/** 
 * Create a new node.
 * @param data The data to store in the node.
 * @param file The file this is being called from
 * @param line The line this is being called from.
 * @retrun a new node.
 */
static struct LL_node * makeNode(DMS_PARAMS(void *data) );

/** 
 * Return the address of the terminal sentinal node.  
 * It doesn't matter which direction we are going, the ending node is 
 * always the same. 
 *
 * @param list The list being worked on.
 * @return The address of the last node in the list.
 */
#define END_NODE(list)   (struct LL_node*)(list)

/** 
 * Insert one node AFTER the other.
 * @param pos The position node (in the list)
 * @param add The node to add (not in the list)
 */
static void insertNodeAfter(struct LL_node *pos, struct LL_node *add);

/**
 * Increment the size of the list.
 * @param list The list to change
 */
static inline void incrSize(struct LL_node *list);

/**
 * Decrement the size of the list.
 * @param list The list to change.
 */
static inline void decrSize(struct LL_node *list);

/**@name Direction of last iterator move
 * @{
 */
#define ITR_DIRNONE 0   /**< Just created */
#define ITR_DIRNEXT 1   /**< Last moved via LLI_next() */
#define ITR_DIRPREV 2   /**< Last moved via LLI_prev() */
/*@}*/

/* Get a new list */
List 
LL_create_priv(DMS_EMPTY_PARAMS)
{
    List    list;
    size_t  sz;

    /* make the head node */
    sz = sizeof(struct LL_node) + sizeof(struct LL_data);
    list = DMSmalloc(sz);
    memset(list, 0, sz);

    /* setup variables */
    list->next = list;
    list->prev = list;

    /* space for data is just past the space for the head node */
    list->data = list + 1;

    return list;
}

/* destroy a list */
void
LL_destroy(List *list)
{
    struct LL_node  *next;
    struct LL_node  *cur;

    assert(list != NULL);   

    if (*list == NULL)
        return;  /* nothing to do */

    cur = (*list)->next;
    while (cur != END_NODE(*list)) {
        next = cur->next;
        free(cur);
        cur = next;
    }

    free(*list);
    *list = NULL;

    assert(*list == NULL);
}

/* append a new element */
void
LL_append_priv(DMS_PARAMS(List list, void *elem) )
{
    struct LL_node  *newNode;

    assert(list != NULL);   

    newNode = makeNode(DMS_ARGS(dmfile, dmline, elem));

    /* inserting on list->prev, b/c that will cause newNode to be
     * inserted between list->prev and list.
     */
    insertNodeAfter(list->prev, newNode);
    incrSize(list);
}

/* prepend a new element */
void
LL_prepend_priv(DMS_PARAMS(List list, void *elem) )
{
    struct LL_node  *newNode;

    assert(list != NULL);  

    newNode = makeNode(DMS_ARGS(dmfile, dmline, elem));
    
    /* Inserting after "list" b/c that will cause newNode to be placed
     * on list->next, which is the list head.
     */
    insertNodeAfter(list, newNode);
    incrSize(list);
}

/* get the first element */
void *
LL_first(List list)
{
    assert(list != NULL);   

    if (list->next == END_NODE(list) ) return NULL;
    return list->next->data;
}

/* get the last element */
void *
LL_last(List list)
{
    assert(list != NULL);   

    if (list->prev == END_NODE(list) ) return NULL;
    return list->prev->data;
}

int 
LL_size(List list)
{
    assert(list != NULL);

    return ((struct LL_data*)list->data)->size;
}

bool
LL_isEmpty(List list)
{
    assert(list != NULL);   
    /* if everything is consistent, we can get away with only checking
     * one of these ... but lets be safe.
     */
    return list->next == list && list->prev == list;
}

void*
LL_rmFirst(List list)
{
    struct LL_node  *rmNode;
    void            *rmData;

    assert(list != NULL);

    if (list->next == END_NODE(list) ) return NULL;

    rmNode = list->next;        /* save the node we are removing */
    list->next = rmNode->next;  /* remove the first node from the list */
    list->next->prev = list;    /* fix prev pointer in new first node */

    rmData = rmNode->data;

    free(rmNode);

    decrSize(list);

    return rmData;
}

void*
LL_rmLast(List list)
{
    struct LL_node  *rmNode;
    void            *rmData;

    assert(list != NULL);

    if (list->prev == END_NODE(list) ) return NULL;

    rmNode = list->prev;        /* save removed node */
    list->prev = rmNode->prev;  /* remove last node from list */
    list->prev->next = list;    /* fix next pointer in new last node */

    rmData = rmNode->data;      /* save removed data */
    free(rmNode);               /* free the node */

    decrSize(list);

    return rmData;
}

/* get an iterator */
LLIterator
LL_iterator_priv(DMS_PARAMS(List list) ) 
{
    LLIterator  itr;

    assert(list != NULL);  

    itr = DMSmalloc(sizeof(struct LL_iter));
    memset(itr, 0, sizeof(struct LL_iter));

    itr->cur = list;
    itr->begin = list;

    return itr;
}

/***************** Iterator functions ***************************/

/* next element */
void *
LLI_next(LLIterator itr)
{
    assert(itr != NULL);        /* pre */
    assert(itr->cur != NULL);   /* invariant */

    /* figure out if we are at the end of the line */
    if (itr->cur == itr->begin && itr->dir == ITR_DIRNEXT)
        return NULL;

    /* advance */
    itr->cur = itr->cur->next;

    assert(itr->cur != NULL);  /* invariant */ 

    /* clear a deleted flag */
    itr->deleted = FALSE;

    itr->dir = ITR_DIRNEXT;

    /* if we just moved past the end (this call) we return NULL */
    if (itr->cur == itr->begin)
        return NULL;

    return itr->cur->data;
}

/* previous element */
void *
LLI_prev(LLIterator itr)
{
    assert(itr != NULL);        
    assert(itr->cur != NULL);   

    /* figure out if we are at the end of the line */
    if (itr->cur == itr->begin && itr->dir == ITR_DIRPREV)
        return NULL;

    /* advance, err ... backup. but only if the "real" current node was
     * not deleted.  Remember, a deletion moves the current node back,
     * so the "real" previous node is cur.
     */
    if (! itr->deleted) {
        itr->cur = itr->cur->prev;
    }
    assert(itr->cur != NULL);  

    itr->deleted = FALSE;       /* clear this flag */

    itr->dir = ITR_DIRPREV;

    /* Moving past the begining this call returns NULL */
    if (itr->cur == itr->begin) 
        return NULL;

    return itr->cur->data;
}

/* current element */
void *
LLI_elem(LLIterator itr)
{
    assert(itr != NULL);        
    assert(itr->cur != NULL);  

    if (itr->cur == itr->begin) return NULL;

    assert(itr->cur != NULL);

    return itr->cur->data;
}

/* remove current element */
void *
LLI_remove(LLIterator itr)
{
    struct LL_node      *del;
    void                *rtn;

    assert(itr != NULL);        
    assert(itr->cur != NULL);   

    /* we do NOT delete the center node */
    if (itr->cur == itr->begin) return NULL;

    del = itr->cur;

    (void)LLI_prev(itr);            /* move pos back */
    itr->deleted = TRUE;            /* mark as deleted */

    assert(itr->cur != NULL);   

    /* remove del from the list */
    itr->cur->next = del->next;
    del->next->prev = itr->cur;

    rtn = del->data;
    free(del);

    decrSize(itr->begin);

    return rtn;
}

/* insert elem at the current position of the iterator */
void
LLI_insert_priv(DMS_PARAMS(LLIterator itr, void *elem) )
{
    struct LL_node  *newNode;

    assert(itr != NULL);        /**<@pre itr != NULL */
    assert(itr->cur != NULL);   /**<@pre iterator's curernt node != NULL */

    newNode = makeNode(DMS_ARGS(dmfile, dmline, elem) );
    
    insertNodeAfter(itr->cur, newNode);
    incrSize(itr->begin);
}

/* destroy the iterator */
/**@pre itr != NULL
 * @post *itr == NULL
 */
void
LLI_destroy(LLIterator *itr)
{
    assert(itr != NULL);

    free(*itr);
    *itr = NULL;

    assert(*itr == NULL);
}

/* make a new node */
static struct LL_node *
makeNode(DMS_PARAMS(void *data) ) 
{
    struct LL_node  *newNode;

    newNode = DMSmalloc(sizeof(struct LL_node));
    memset(newNode, 0, sizeof(struct LL_node));

    newNode->data = data;

    return newNode;
}

static void
insertNodeAfter(struct LL_node *pos, struct LL_node *add)
{
    add->next = pos->next;
    add->prev = pos;

    pos->next->prev = add;
    pos->next = add;
}

static inline void
incrSize(struct LL_node *list)
{
    struct LL_data  *data;

    data = (struct LL_data*)list->data;

    data->size++;
}

static inline void
decrSize(struct LL_node *list)
{
    struct LL_data  *data;

    data = (struct LL_data*)list->data;

    data->size--;
}

