/**
 * @file list.h
 *
 * @brief A header for linked list functions.
 * @par
 * This header declares operations on a single linked list.  Ineserts
 * at either end are supported, as are iterators. It also has functions
 * that allow treating the list as a stack or dequeue.  It is upto you
 * to know when that is appropriate though.
 *
 * @ingroup list
 *
 * $Id: list.h,v 1.7.2.2 2003/04/30 20:04:02 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Fri Feb 01 13:54:55 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef LIST_LIST_H
#define LIST_LIST_H

#if HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include "list_private.h"

#include "dmalloc_sup.h"


/**@defgroup list Linked List Container
 * @{
 */

/**@name List Types 
 * @{
 */

/** The type of a list object */
typedef struct LL_node *List;

/** The type of a list iterator object */
typedef struct LL_iter *LLIterator;

/*@}*/


/**@name List Operations
 * @{
 */

/** 
 * Create a new list.
 * @return the new list
 *
 * @test list_create()
 */
#define LL_create()     LL_create_priv(DMS_EMPTY_ARGS(__FILE__, __LINE__) )

/** 
 * Destroy a list.
 * @param l The list to destory.
 * @warning All iterators for this will will be invalidated by
 *          destroying the list.  The client is responsible for handling
 *          this, as this library makes no effort to.
 *
 * @pre l != NULL
 * @post *l == NULL
 */
void LL_destroy(List *l);

/** 
 * Add an element to the end of a list.
 * @param l The List to work on
 * @param elem The element to add
 *
 * @test list_append()
 * @pre l != NULL
 */
#define LL_append(l, elem)  \
    LL_append_priv(DMS_ARGS(__FILE__, __LINE__, l, elem))

/** 
 * Add an element to the begining of a list.
 * @param l The List to work on
 * @param elem The element to add.
 *
 * @test list_prepend()
 * @pre l != NULL
 */
#define LL_prepend(l, elem) \
    LL_prepend_priv(DMS_ARGS(__FILE__, __LINE__, l, elem) )

/** Get the first element in the list 
 * @param l The List to work on
 * @return The first element in \c l
 *
 * @test list_prepend(), list_append()
 * @pre l != NULL
 */
void *LL_first(List l);

/** Get the last element in the list
 * @param l The List to work on.
 * @return The last element in \c l
 *
 * @test list_prepend(), list_append()
 * @pre l != NULL
 */
void *LL_last(List l);

/** Is the list empty?
 * @param l The list to check
 * @return TRUE if the list is empty, FALSE otherwise
 *
 * @test list_create(), list_prepend(), list_append()
 * @pre l != NULL
 */
bool LL_isEmpty(List l);

/**
 * Get the number of elements in this list.
 * @param l The list to use
 * @return The number of elements in the list. < 0 on error.
 *
 * @pre l != NULL
 * @test list_prepend(), list_append(), lliterator_insert(),
 *       lliterator_remove()
 */
int LL_size(List l);


/** 
 * Remove the first element of the list.  
 * This is for making lists act like dqueues/stacks
 * @warning This function will cause list iterators to grossly misbahave
 * in certain circumstances.
 *
 * @param l The list to work on
 * @return data of removed node, or NULL if no node was removed.
 *
 * @pre l != NULL
 * @test list_rmfirst(), list_remove()
 */
void *LL_rmFirst(List l);

/** 
 * Remove the last element of the list.  
 * @warning This function will cause list iterators to grossly misbahave
 * in certain circumstances.
 *
 * @param l The list to work on.
 * @return data of removed node, or NULL if no node was removed.
 *
 * @pre l != NULL
 * @test list_rmlast(), list_remove()
 */
void *LL_rmLast(List l);

/** Get an iterator for the list
 * @param l The list to work on.
 * @return an Iterator for \c l, pointing to the first element.
 *
 * @test lliterator_create
 * @pre l != NULL
 */
#define LL_iterator(l) LL_iterator_priv(DMS_ARGS(__FILE__, __LINE__, l))
/*@}*/

/**@name List Iterator Functions 
 * @{
 */

/** Advance the iterator to the next element.
 * @param itr The iterator to work on.
 * @return The next element.
 *
 * @test lliterator_next()
 * @pre itr != NULL
 * @invariant iterator's current node != NULL
 */
void *LLI_next(LLIterator itr);

/** Advance the iterator to the previous element.
 * @param itr The iterator to work on.
 * @return the previous element.
 *
 * @test lliterator_prev()
 * @pre itr != NULL
 * @invariant iterator's current node != NULL
 */
void *LLI_prev(LLIterator itr);

/** 
 * Get the element the iterator is currently pointing to.
 * @param itr The iterator to work on.
 * @return The current element.;
 *
 * @test lliterator_elem()
 * @pre itr != NULL
 * @invariant iterator's current node != NULL
 */
void *LLI_elem(LLIterator itr);

/** 
 * Remove the element under the iterator. 
 * @param itr The iterator to work on.
 * @return The removed element.
 *
 * @pre itr != NULL
 * @invariant iterator's current node != NULL
 */
void *LLI_remove(LLIterator itr);

/** 
 * Insert an element into the list AFTER the current position.
 * @param itr The iterator to work on.
 * @param elem The element to insert.
 *
 * @pre itr != NULL
 * @invariant iterator's current node != NULL
 */
#define LLI_insert(itr, elem) \
    LLI_insert_priv(DMS_ARGS(__FILE__, __LINE__, itr, elem) )

/** 
 * Destroy an iterator. 
 * @param itr The iterator to destroy
 *
 * @pre itr != NULL
 * @post *itr != NULL
 */
void LLI_destroy(LLIterator *itr);
/*@}*/

/**@name Private functions that need file and line info so that dmalloc
 * will be more useful.
 * @{
 */
void LLI_insert_priv(DMS_PARAMS(LLIterator, void *));
void LL_append_priv(DMS_PARAMS(List, void *));
void LL_prepend_priv(DMS_PARAMS(List, void *));
LLIterator LL_iterator_priv(DMS_PARAMS(List));
List LL_create_priv(DMS_EMPTY_PARAMS);
/*@}*/
/*@}*/

#endif /* LIST_LIST_H */

