/**
 * @file log.c
 *
 * @brief Implement the Internal logging functions
 *
 * $Id: log.c,v 1.15.2.7 2003/04/30 20:04:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Tue May 21 12:08:02 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/time.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>
#include <stdio.h>

#include "test.h"
#include "config/priority.h"
#include "log.h"
#include "controller/controller.h"
#include "ipc/ipc.h"
#include "strcompat.h"
#include "error.h"
#include "util/tsqueue.h"

static char const rcsid[] 
    = "$Id: log.c,v 1.15.2.7 2003/04/30 20:04:03 devink Exp $";

/** Map internal severities to external severities */
static int sevToLogSev[] = {
    0,					/* place holder */
    CONFFS_DEBUG,		/* LOG_DEBUG */
    CONFFS_NOTICE,		/* LOG_INFO */
    CONFFS_ERR,			/* LOG_ERROR */
    CONFFS_EMERG		/* LOG_CRIT */
};

/** Map internal severities to tags for STDOUT messages */
static char *sevTags[] = {
    NULL,
    "DEBUG",
    "INFO",
    "ERROR",
    "FATAL",
};


/** The facility to log to */
#define LOG_fac CONFFS_SYSLOG

/** 
 * The tag format that we should use for logging, this is a printf
 * style format string that expects the pid to be provided. 
 */
char    LOG_tag[MSGTAG_LEN];

/** This host's timezone */
static int LOG_tz;

/** This host's name */
static char LOG_hostname[256];  

/** The length of \c LOG_hostname */
static size_t LOG_hostlen;

/** The three logging outputs we might be using */
enum LOG_TO {
    TO_STDOUT,  /**< Logging to STDOUT */
    TO_OMOD,    /**< In the controller */
    TO_IPC,     /**< In an Input Module */
    TO_TSQ,     /**< A thread safe queue (for threaded IMs) */
    TO_DPRINT   /**< use DPRINT() b/c there is no better place */
};

/** Keep track of where we are logging to */
static enum LOG_TO logdst;

/** The currently configured controller object */
static CTLR *ctlr;

/** The currently configured IPC handle */
static IPC *ipc;

/** The currently configured TSQueue */
static TSQueue *queue;

/** The input channel ID to send messages from */
static uint32_t icSysID;

/** 
 * Mark that we are already in a debugging function. 
 * This MUST BE CALLED on entering all the logging functions, it
 * prevents them from jumping into infinate recursion.
 *
 * @param msg The message that we are trying to log.
 */
#define LOG_ENTER_FUNCTION(msg)\
    if (LOG_in_func) {DPRINT(DBG_DBG, "LOG recursion: %s", msg); return; }\
    LOG_in_func = TRUE

/** 
 * For calling when we are leaving a LOG function.  This will allow log
 * functions to be entered again.
 */
#define LOG_EXIT_FUNCTION() LOG_in_func = FALSE

/** Variable to mark when we are in a LOG function. */
static bool LOG_in_func = FALSE;

/**
 * The function that actually logs a message.  The public functions are
 * just pre-processors.
 *
 * @param sev The severity of the message (one of LOG_*)
 * @param fmt The format string
 * @param ap The variable arguments
 */
static void doLog(int sev, char const *fmt, va_list ap);


/* start the functions */

void
LOG_print(int sev, char const *fmt, ...)
{
    va_list         ap;

    LOG_ENTER_FUNCTION(fmt);

    /* build the message */
    va_start(ap, fmt);

    /* Hope the compiler and linker DTRT */
    if (IN_UNIT_TEST_CT) {
        char    buf[1024];
        snprintf(buf, sizeof(buf), "Error: %s", fmt);
        vprintf(buf, ap);
    }
    else {
        doLog(sev, fmt, ap);
    }


    LOG_EXIT_FUNCTION();
}


void
LOG_eprint(char const *fmt, int err, ...)
{
    va_list     ap;
    char        msgstr[1024];
    size_t      len;

    LOG_ENTER_FUNCTION(fmt);

    va_start(ap, err);

    /* leave half the string for errno message */
    len = strlcpy(msgstr, fmt, sizeof(msgstr) / 2);

    /* drop trailing space */
    len--;
    while (isspace(msgstr[len]) ) {
        len--;
    }

    /* we know that len < sizeof(msgstr)/2 */
    if (msgstr[len] != ':') {
        len++;
        msgstr[len] = ':';
    }
    if (msgstr[len] != ' ') {
        len++;
        msgstr[len] = ' ';
    }
    len++;
    msgstr[len] = '\0';

    strlcat(msgstr, strerror(err), sizeof(msgstr));

    /* Hope the compiler and linker DTRT */
    if (IN_UNIT_TEST_CT) {
        char    buf[1024];
        snprintf(buf, sizeof(buf), "Error: %s", msgstr);
        vprintf(buf, ap);
    }
    else {
        doLog(LOG_ERROR, msgstr, ap);
    }

    LOG_EXIT_FUNCTION();
}

bool
LOG_initCTLR(CTLR *new_ctlr)
{
    ctlr = new_ctlr;
    ipc = NULL;
    queue = NULL;
    logdst = TO_OMOD;

    strlcpy(LOG_tag, "syslog-c", MSGTAG_LEN);

    return TRUE;
}

bool
LOG_initIM(IPC *new_ipc, uint32_t sysid)
{
    ipc = new_ipc;
    ctlr = NULL;
    queue = NULL;
    logdst = TO_IPC;
    icSysID = sysid;

    strlcpy(LOG_tag, "syslog-im", MSGTAG_LEN);

    return TRUE;
}

bool
LOG_initTSQ(TSQueue *newq, uint32_t sysid)
{
    queue = newq;
    ipc = NULL;
    ctlr = NULL;
    logdst = TO_TSQ;
    icSysID = sysid;

    strlcpy(LOG_tag, "syslog-im", MSGTAG_LEN);

    return TRUE;
}

bool
LOG_initLimbo(void)
{
    ctlr = NULL;
    ipc = NULL;
    queue = NULL;
    logdst = TO_DPRINT;

    return TRUE;
}

void
LOG_init(void)
{
    struct timeval  tv;
    struct timezone tz;     /* what we are really trying to get */

    ctlr = NULL;
    ipc = NULL;
    logdst = TO_STDOUT;

    if ( gethostname(LOG_hostname, sizeof(LOG_hostname)) < 0) {
        LOG_hostlen = strlcpy(LOG_hostname, "UNKNOWN", sizeof(LOG_hostname)); 
        LOG_hostlen++;
    }
    else {
        LOG_hostlen = strlen(LOG_hostname) + 1;
    }

    LOG_tz = 0;     /* for lack of a better default */
    if (gettimeofday(&tv, &tz) == 0) {
        LOG_tz = -1 * tz.tz_minuteswest;
    }

    /* all setup */
}

static void
doLog(int sev, char const *fmt, va_list ap)
{
#if ! IN_UNIT_TEST_CT
    char            msgstr[1204];
    struct message  msg;


    /* Make sure sev is reasonable, and fix it if not */
    if (sev > LOG_CRIT) sev = LOG_CRIT;
    if (sev < LOG_DEBUG) sev = LOG_DEBUG;

    /* build the message to print */
    vsnprintf(msgstr, sizeof(msgstr), fmt, ap);

    /* This one is easy */
    if (logdst == TO_STDOUT) {
        printf("%s: %s\n", sevTags[sev], msgstr);
        return;
    }


    /* build the message structure for this message */
    memset(&msg, 0, sizeof(msg));

    msg.sev = sevToLogSev[sev];
    msg.fac = LOG_fac;

    msg.hostlen = LOG_hostlen;
    msg.host = emalloc(LOG_hostlen);
    strlcpy(msg.host, LOG_hostname, LOG_hostlen);

    msg.timestamp = time(NULL);
    msg.tz = LOG_tz;

    strlcpy(msg.tag, LOG_tag, MSGTAG_LEN);

    /* + 16 for some extra space for the pid */
    msg.msglen = strlen(msgstr) + 16;   
    msg.msg = emalloc(msg.msglen);
    snprintf(msg.msg, msg.msglen, "[%d]: %s", getpid(), msgstr);

    if (logdst == TO_OMOD && ctlr != NULL) {
        /* Note that the static icSysID is not used here because the
         * controller exists through reconfigs, which means that icSysID
         * will change without warning.
         */
        CTLR_sendMessage(ctlr, &msg, ctlr->icSysID);
    }
    else if (logdst == TO_IPC && ipc != NULL) {
        IPC_imSendMessage(ipc, icSysID, &msg);
    }
    else if (logdst == TO_TSQ && queue != NULL) {
        struct MSG_qelem *mqe;
        mqe = emalloc(sizeof(struct MSG_qelem));
        memset(mqe, 0, sizeof(struct MSG_qelem));
		mqe->data.msg=emalloc(sizeof(struct message));

        mqe->type = MSGQE_MSG;
        mqe->chan = icSysID;
		memcpy(mqe->data.msg, &msg, sizeof(struct message));
        TSQ_enqueue(queue, mqe);
    }
    /* This is usually triggered when logdst == TO_DPRINT, but also
     * serves as a resonable default destination.
     */
    else {
        /* DBG_DBG an the assumption that if any debugging is being done
         * the lost log messages will be wanted.
         */
        DPRINT(DBG_DBG, "%s: %s\n", sevTags[sev], msgstr);
    }

    MSG_free(&msg);

#endif /* ! IN_UNIT_TEST_CT */
}
