/**
 * @file netaddr.c
 *
 * @brief Implement functions for handling network addresses.
 *
 * $Id: netaddr.c,v 1.19.2.6 2003/04/30 20:04:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon Jan 28 14:59:44 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#if HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "strcompat.h"
#include "netaddr.h"
#include "error.h"
#include "dmalloc_sup.h"

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif



static char const rcsid[] 
    = "$Id: netaddr.c,v 1.19.2.6 2003/04/30 20:04:03 devink Exp $";

/* documentation groups */
/**@defgroup netaddr_impl Network Address Abstration Implementation 
 * @ingroup netaddr
 * @{
 */

/**
 * Make a netaddr_t object that represents an error.  If a valid pointer
 * is provided then this will assume that space for a netaddr_t object
 * has already been allocated and will change that in place, otherwise
 * it will create a new one.
 *
 * @param na The netaddr_t object to change, or NULL
 * @return \c na if provided, or a new netaddr_t object
 */
static netaddr_t mkError(DMS_PARAMS(netaddr_t na));


/**@name create netaddr structures 
 * @{
 */ 
netaddr_t NA_newINet4_priv(DMS_PARAMS(struct in_addr addr))
{
    netaddr_t   rtn;

    rtn = DMSmalloc(sizeof(struct netaddr));
    rtn->type = NAT_IP4;
    rtn->addr.ip4 = addr.s_addr;

    return rtn;
}

netaddr_t NA_newINet6_priv(DMS_PARAMS(struct in6_addr addr))
{
    netaddr_t   rtn;

    rtn = DMSmalloc(sizeof(struct netaddr));
    rtn->type = NAT_IP6;
    memcpy(rtn->addr.ip6, addr.s6_addr, sizeof(rtn->addr.ip6));

    return rtn;
}

netaddr_t NA_newUnix_priv(DMS_PARAMS(char const *path))
{
    netaddr_t   rtn;
    rtn = DMSmalloc(sizeof(struct netaddr));
    rtn->type = NAT_UNIX;
    strlcpy(rtn->addr.sun_path, path, UNIX_PATH_MAX);

    return rtn;
}

netaddr_t NA_newNetaddr_priv(DMS_PARAMS(netaddr_t addr))
{
    netaddr_t   rtn;

    rtn = DMSmalloc(sizeof(struct netaddr));
    memcpy(rtn, addr, sizeof(struct netaddr));

    return rtn;
}

netaddr_t NA_newBuffer_priv(DMS_PARAMS(struct IPC_buffer *buf))
{
    netaddr_t   rtn;
    uint8_t     tbyte;
    int         i;
    IPC_bufMark  startPos;
    
    rtn = DMSmalloc(sizeof(struct netaddr));

    startPos = BUF_markPos(buf);

    if (! BUF_getByte(buf, &tbyte) ) goto error;
    rtn->type = (int)(tbyte); 
    if (rtn->type < 0) goto error;   /* bit of a problem there */

    if (rtn->type == NAT_IP4) {
        if (! BUF_getInteger(buf, &(rtn->addr.ip4)) ) goto error;
    }
    else if (rtn->type == NAT_IP6) {
        for (i = 0; i < sizeof(rtn->addr.ip6); i++) {
            if (! BUF_getByte(buf, &(rtn->addr.ip6[i])) ) goto error;
        }
    }
    else if (rtn->type == NAT_UNIX) {
        if (!BUF_getVarString(buf, NULL, (char**)&(rtn->addr.sun_path),
                    UNIX_PATH_MAX)) 
            goto error;
    }
    else if (rtn->type == NAT_EMPTY) {
        memset(&(rtn->addr), 0, sizeof(rtn->addr));
    }
    else {
        throw_fatal("Unknow Network Address Type");
        goto error;
    }

    return rtn;


    /* This is only reached when an error occurs pulling the netaddr out
     * of the buffer.  It cleans up and returns an error.
     */
    error:
        BUF_resetToPos((struct IPC_buffer*)buf, startPos);
        return mkError(DMS_ARGS(dmfile, dmline, rtn));
}


netaddr_t
NA_newEmpty_priv(DMS_EMPTY_PARAMS)
{
    netaddr_t   rtn;

    rtn = DMSmalloc(sizeof(struct netaddr));
    memset(rtn, 0, sizeof(struct netaddr));
    rtn->type = NAT_EMPTY;

    return rtn;
}

netaddr_t 
NA_newINetLookup_priv(DMS_PARAMS(char const *host))
{
    struct hostent  *phent;

    phent = gethostbyname(host);

    if (phent == NULL) {
        DPRINT(DBG_NA, "gethostbyname failed for \"%s\": (%d) %s",
                host, h_errno, hstrerror(h_errno));
        return mkError(DMS_ARGS(dmfile, dmline, NULL) );
    }

    /* XXX should I do a reverse lookup here ? */

    /**@bug Needs to deal with IPv6 ... but not sure what to look for */
    if (phent->h_addrtype == AF_INET) {
        struct in_addr  inaddr;

        if ( (phent->h_length % sizeof(struct in_addr) ) != 0) {
            return FALSE;
        }
        memcpy(&inaddr, phent->h_addr_list[0], sizeof(struct in_addr));

        return NA_newINet4_priv(DMS_ARGS(dmfile, dmline, inaddr));
    }

    return mkError(DMS_ARGS(dmfile, dmline, NULL));
}

static netaddr_t
mkError(DMS_PARAMS(netaddr_t na))
{
    if (na == NULL) {
        na = DMSmalloc(sizeof(struct netaddr));
    }

    memset(na, 0, sizeof(struct netaddr));
    na->type = NAT_ERROR;

    return na;
}

/*@}*/

bool
NA_isEmpty(netaddr_t addr)
{
    return addr == NULL || addr->type == NAT_EMPTY;
}

bool 
NA_isError(netaddr_t addr)
{
    return (NAT_ERROR == addr->type);
}

/* cleanup a netaddr_t object */
void
NA_destroy(netaddr_t addr)
{
    if (addr != NULL)
        free(addr);
}

/* compare two netaddr_t objects */
int
NA_compare(netaddr_t lhs, netaddr_t rhs)
{
    if (lhs == NULL && rhs == NULL) {
        /* both are empty, so they are equal */
        return 0;
    }
    else if (lhs == NULL || rhs == NULL) {
        /* only one is empty, so they are not equal */
        return 1;
    }

    if (lhs->type != rhs->type) {
        /* the "larger" type is greater */
        return lhs->type - rhs->type;
    }

    /* types are the same, so we only need to check one to figure 
     * out what it is.
     */
    if (lhs->type == NAT_IP4) {
        return lhs->addr.ip4 - rhs->addr.ip4;
    }
    else if (lhs->type == NAT_IP6) {
        return memcmp(lhs->addr.ip6, rhs->addr.ip6, sizeof(lhs->addr.ip6));
    }
    else if (lhs->type == NAT_UNIX) {
        return strcmp(lhs->addr.sun_path, rhs->addr.sun_path);
    }
    else if (lhs->type == NAT_EMPTY || lhs->type == NAT_ERROR) {
        /* XXX should these be considered equal? */
        return 0;
    }

    /* Okay, we have a problem here ... but should return something, and
     * calling lhs == rhs doesn't seem like a good idea.
     */
    throw_fatal("unknown netaddr type");
    return -2;
}

/* create a struct sockaddr */
bool 
NA_mkSockaddr(netaddr_t addr, int port, union NA_sockaddr *sa)
{
    assert(sa != NULL);

    if (addr == NULL || addr->type == NAT_EMPTY || addr->type == NAT_ERROR) {
        /* An error to try to make a sockaddr out of an empty/error net addr */
        return FALSE;
    }
    else if (addr->type == NAT_IP4) {
        sa->sin.sin_family = PF_INET;
        sa->sin.sin_port = htons(port);
        sa->sin.sin_addr.s_addr = addr->addr.ip4;
    }
    else if (addr->type == NAT_IP6) {
        sa->sin6.sin6_family = PF_INET6;
        sa->sin6.sin6_port = htons(port);
        memcpy(&(sa->sin6.sin6_addr), &addr->addr, sizeof(sa->sin6.sin6_addr));
    }
    else if (addr->type == NAT_UNIX) {
        sa->sun.sun_family = PF_UNIX;
        /* ignoring unused port arg */
        strlcpy(sa->sun.sun_path, addr->addr.sun_path, UNIX_PATH_MAX);
    }
    else {
        throw_fatal("unknow netaddr type");
        return FALSE;
    }

    return TRUE;
}

int
NA_getProtFamily(netaddr_t addr) 
{
    if (addr == NULL || addr->type == NAT_EMPTY || addr->type == NAT_ERROR) 
        return -1;

    switch (addr->type) {
        case NAT_IP4: return PF_INET;
        case NAT_IP6: return PF_INET6;
        case NAT_UNIX: return PF_UNIX;
        default: throw_fatal("Unknown netaddr type"); break;
    }

    return -1;
}

char *
NA_getCharAddr(netaddr_t addr)
{
	struct in_addr temp_addr;

    if (addr == NULL || addr->type == NAT_EMPTY || addr->type == NAT_ERROR) 
        return NULL;

	switch (addr->type)
	{
		case NAT_UNIX:
			return addr->addr.sun_path;
			break;

		case NAT_IP4:
			temp_addr.s_addr=addr->addr.ip4;
			return inet_ntoa(temp_addr);
			break;

		case NAT_IP6:
			return "IP6 not implemented (NA_getCharAddr())";
			break;

		default:
			return NULL;
			break;
	}

	/* never reached */
	return NULL;
}

int
NA_toBuffer(netaddr_t addr, struct IPC_buffer *out)
{
    IPC_bufMark startPos;
    bool        success;
    int         len, i;

    startPos = BUF_markPos(out);

    success = FALSE;
    len = 0;

    if (addr == NULL) {
        success = BUF_setByte(out, NAT_EMPTY); 
        len = BUF_BYTESZ;
        goto end_serialize;  /* trying to do more would cause problems */
    }

    /* Don't try to serialize an error */
    if (addr->type == NAT_ERROR) goto end_serialize;

    /* all types are <= 255 */
    if (! BUF_setByte(out, addr->type) ) goto end_serialize;  
    len = BUF_BYTESZ;

    if (addr->type == NAT_IP4) {
        success = BUF_setInteger(out, addr->addr.ip4);
        len += BUF_INTSZ;
    }
    else if (addr->type == NAT_IP6) {
        for (i = 0; i < sizeof(addr->addr.ip6); i++) {
            success = BUF_setByte(out, addr->addr.ip6[i]);

            if (!success) break;    /* stop on error */
            len += BUF_INTSZ;
        }
    }
    else if (addr->type == NAT_UNIX) {
        i = BUF_setVarString(out, addr->addr.sun_path, 0); 
        success = i > 0;
        len += i;
    }
    else if (addr->type == NAT_EMPTY) {
        /* do nothing */
    }
    else {
        throw_fatal("unknown netaddr type");
        success = FALSE;
    }

    /* This marks the end of the deserializing/parsing.  Certain falure
     * modes need to stop early.
     */
    end_serialize:
        if (!success) {
            BUF_resetToPos(out, startPos);
            return -1;
        }

        return len;
}

/*@}*/
