/**
 * @file netaddr.h
 *
 * @brief This file defines several functions for working with network
 * addresses.  
 *
 * Because there are several different types of network
 * address there is a single netaddr struct defined, and various
 * operations on such objects.  These operations allow the rest of the
 * program to ignore the actual type of network address.
 *
 * @par
 * Netaddr objects have two odd values, Empty and Error.  Empty netaddr
 * objects simply have no address in them, this may be because it hasn't
 * been initilized, or was irrelevant.  Empty values can be created with
 * the NA_newEmpty() function, or by simply setting the netaddr_t
 * variable to NULL.  Error values signal that a constructor failed.
 *
 * $Id: netaddr.h,v 1.12.2.2 2003/04/30 20:04:03 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Thu Jan 24 17:56:23 PST 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INC_NETADDR_H
#define INC_NETADDR_H

#ifdef HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "types.h"
#include "ipc/buffer.h"
#include "const.h"

#include "dmalloc_sup.h"

/**@defgroup netaddr Network Address Abstraction */

/**@addtogroup netaddr
 * @{
 */ 

/** @name Network Address Types
 *  @ingroup netaddr
 *  @internal
 *  @{
 */
#define NAT_IP4     1   /**< IPv4 */
#define NAT_IP6     2   /**< IPv6 */
#define NAT_UNIX    3   /**< Unix Domain Sockets */
#define NAT_EMPTY   4   /**< Empty Netaddr */
#define NAT_ERROR   5   /**< Error Value */
/*@}*/


/** 
 * The maximum size needed for the buffer to hold a serialized 
 * network address 
 */
#define NA_MAX_BUFLEN   (BUF_BYTESZ + UNIX_PATH_MAX*BUF_BYTESZ)

/** A structure for storing network addresses.
 * @internal 
 * @ingroup netaddr
 *    
 */    
struct netaddr {
    int     type;   /**< Network Address Type */
    union {
        in_addr_t   ip4;
        uint8_t    ip6[16];
        char        sun_path[UNIX_PATH_MAX];
    } addr;
};

/** 
 * A union filled in by NA_mkSockaddr().  This is large enough to hold
 * the largest sockaddr type supported by Netaddr abstraction, and it
 * can be safely cast to any of the sockaddr* types.
 *
 * Or if it makes you feel better then pass the \c sa member to the
 * socket calls (i.e. bind, listen).  However, if you want to pass the
 * correct type to some other function (i.e. PRIV_bindUnixSocket()) make
 * sure that it actually contains a sockaddr of the correct type (check
 * the protocol family via NA_getProtoFamily()).  
 *
 * Regardless of how you pass it around, always pass the sizeof of the 
 * whole structure unless you have verified the protocol family first.
 *
 * @ingroup netaddr
 */
union NA_sockaddr {
    struct sockaddr     sa;
    struct sockaddr_un  sun;
    struct sockaddr_in  sin;
    struct sockaddr_in6 sin6;
};

/** 
 * Make a type to store network addresses.  This should in all cases be
 * used rather than struct %netaddr
 */
typedef struct netaddr *netaddr_t;

/** @name Create netaddr_t objects 
 *  @ingroup netaddr
 *  @{
 */
/** 
 * Create a new network address. 
 * The return is a netaddr_t object.  If there was an error then a
 * netaddr_t object with an error value will be returned.  This can be
 * checked for with the NA_isError() function.
 *
 * @warning even a netaddr_t object with the error value needs to be
 *          destroyed.
 *
 * @param addr The address to initilize it to
 * @return The netaddr_t* for this address
 *
 * @test netaddr_create()
 */
#define NA_newINet4(addr) NA_newINet4_priv(DMS_ARGS(__FILE__, __LINE__, addr)) 
#define NA_newINet6(addr) NA_newINet6_priv(DMS_ARGS(__FILE__, __LINE__, addr))
#define NA_newINetLookup(host) \
    NA_newINetLookup_priv(DMS_ARGS(__FILE__, __LINE__, host))
#define NA_newUnix(addr)  NA_newUnix_priv(DMS_ARGS(__FILE__, __LINE__, addr))
#define NA_newNetaddr(addr) \
    NA_newNetaddr_priv(DMS_ARGS(__FILE__, __LINE__, addr))
#define NA_newBuffer(buf) NA_newBuffer_priv(DMS_ARGS(__FILE__, __LINE__, buf))

netaddr_t NA_newINet4_priv(DMS_PARAMS(struct in_addr addr));
netaddr_t NA_newINet6_priv(DMS_PARAMS(struct in6_addr addr));
netaddr_t NA_newINetLookup_priv(DMS_PARAMS(char const *host));
netaddr_t NA_newUnix_priv(DMS_PARAMS(char const *addr));
netaddr_t NA_newNetaddr_priv(DMS_PARAMS(netaddr_t addr));
netaddr_t NA_newBuffer_priv(DMS_PARAMS(struct IPC_buffer *buf));
netaddr_t NA_newEmpty_priv(DMS_EMPTY_PARAMS);
/*@}*/

/** 
 * Test a netaddr_t object for being empty.  
 *
 * @param addr The object to check
 * @return TRUE if addr is Empty, FALSE otherwise
 */
inline bool NA_isEmpty(netaddr_t addr);

/**
 * Test a netaddr_t for the error value.
 *
 * @param addr the object to test.
 * @return TRUE if it's an error, FALSE otherwise.
 */
inline bool NA_isError(netaddr_t addr);

/** 
 * Destroy netaddr_t objects.
 * This preforms any needed cleanup on the object.
 *
 *  @param addr The object to cleanup
 */
void NA_destroy(netaddr_t addr);

/** Compare two network addresses.
 *
 *  @test netaddr_compare()
 *
 *  @param lhs The left hand side address to compare
 *  @param rhs The right hand side address to compare
 *  @return 
 *      \li \b 0 if (lhs == rhs) 
 *      \li \b -1 if (lhs < rhs)
 *      \li \b 1 if (lhs > rhs)
 *
 */
int NA_compare(netaddr_t lhs, netaddr_t rhs);

/** Get a byte buffer that represents this address.  As the primary use
 * is for IPC messages, it uses the IPC_buffer functions to fill things
 * in.  This function expects that out is at the right point in the
 * buffer to begin writing.  Because the serialization of a netaddr is
 * variable length, the number of bytes used is returned on success.
 * Further on error the buffer \c out will be reset to the starting
 * position, so it's as if nothing was written to it.
 *
 *  @param addr The object to put in the buffer.
 *  @param out The buffer to store the network address in.
 *  @return the number of bytes written on success, < 0 otherwise
 */
int NA_toBuffer(netaddr_t addr, struct IPC_buffer *out);

/** Build a struct sockaddr from the given netaddr_t object and a port
 *  number.  The port number should be in host byte format.  The family
 *  will be filled in automatically.
 *  \par
 *  if sa does not have enough room for the address then an error will 
 *  be returned.
 *
 *  @param addr The address of the sockaddr
 *  @param port the port number for the sockaddr
 *  @param sa The struct sockaddr to put the data in.
 *  @return TRUE on success, FALSE otherwise 
 */
bool NA_mkSockaddr(netaddr_t addr,int port,union NA_sockaddr *sa);

/**
 * Get the Protocol Family that this address belongs in.
 *
 * @param addr The address to look at.
 * @return the protocol family for the address (PF_*)
 */
int NA_getProtFamily(netaddr_t addr);

/** Get the path (either filename or address) of this NetAddr
 *
 * @param addr The address to return the path of
 * @return The path as a char *, or NULL on failure (or if the object isn't of
 * 			a recognized type).  No free() is necassary on this data, but the
 * 			data _may_ be overwritten by later calls to inet_ntoa().
 */
char *NA_getCharAddr(netaddr_t addr);

/*@}*/

#endif /* INC_NETADDR_H */
