/**
 * @file pathcache.c
 *
 * @brief Implement the Path Cache functions.
 * @ingroup pathcache
 *
 * $Id: pathcache.c,v 1.7.2.6 2003/04/30 20:04:04 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jul 03 18:07:50 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if HAVE_CONFIG_H
#   include "acconfig.h"
#endif

#include <stdlib.h>

#include "pathcache.h"
#include "avltree.h"
#include "list.h"

#include "types.h"
#include "error.h"
#include "message/message.h"

#if WITH_DMALLOC
#   include <dmalloc.h>
#endif


static char const rcsid[] = 
    "$Id: pathcache.c,v 1.7.2.6 2003/04/30 20:04:04 devink Exp $";

/** Path Cache Types */
typedef enum {
    PCT_IM,     /**< Input Module Path cache */
    PCT_CTLR    /**< Controller Path Cache */
} PC_type;

/** Path cache structure used in the controller */
struct PCache {
    PC_type     type;       /**< Type of this Path Cache */
    AVLTree     cache;      /**< the cache */

    uint32_t    nextPID;    /**< Next path id (controller use only) */
    uint32_t    nextOMID;   /**< Next output module id (controller use only) */

    /* This is the path info for a path that only contains the current
     * host.  It is only used by the controller Path Cache.
     */
    struct ctlr_path_info   *localPath;
    AVLTree     omRegs;     /**< Track output module registration */
};

/** Per connection entry in a path cache */
struct conn_ent {
    uint32_t    connID;     /**< Connection ID (key field) */

    /** 
     * A container for all the paths which have come over this
     * connection.  In most cases this will only have one element.  
     * Element Data Type:
     *  - Controller: struct ctlr_path_info*
     *  - input mod: struct im_path_info*
     */
    AVLTree     paths;     
};

/**
 * The information needed for each path in the controller cache.
 */
struct ctlr_path_info
{
    struct path *pathHead;  /**< path head (contains pathID) */

    uint32_t    outPID;     /**< pathID for path with us as latest hop */

    AVLTree     omSent;     /**< oms that have sent this path out */
};

/** 
 * This contains all the information needed about a path for the input
 * modules.
 */
struct im_path_info
{
    uint32_t    pathID;     /**< ID of the path */
    bool        aclPass;    /**< Did this pass it's aclchecks? */
};

/**
 * A struct that keeps track of output module registration info.
 */
struct om_reg_info
{
    uint32_t    id;         /**< output module id */

    uint32_t    curConn;    /**< Id for the current connection */
    List        insData;    /**< List of inserted data */
};

/**
 * find a pathID in the path cache. 
 * This doesn't care what kind of cache this is, it just looks up 
 * \c connID in the first level cache (\c PCache::cache) and then looks up 
 * \c pathID in the second level cache (\c conn_ent::paths) and returns the
 * result.  If either is not found then \c NULL is returned.
 *
 * @param cache  The AVLTree that stores the cache data.
 * @param connID The connection ID to look for.
 * @param pathID The path ID to look for.
 * @return the data found, or NULL if not found.
 */
static void * find_path_info(AVLTree cache, uint32_t connID, uint32_t pathID);

/**
 * Destroy a conn_ent structure.
 * This destroys everything in the conn_ent structure, including
 * clearing out the AVLTrees.  It does need to know what kind of data to
 * expect in \c conn_ent::paths.
 *
 * @param ce The conn_ent structure to clear
 * @param type The PCache type
 */
static void conn_ent_destroy(struct conn_ent *ce, PC_type type);

/** 
 * This is the worker function used by conn_ent_destroy() to actually
 * free the nodes in \c conn_ent::paths.  The type of \c n depends on
 * \c type, which is really a PC_type.
 *
 * @see AVL_worker_t
 *
 * @param n The object to free
 * @param type The PCache type (also gives typeof(\c n)
 * @param depth Ignored
 */
static void conn_ent_free(void *n, int type, int depth);

/**
 * This is the worker function used to free an om_reg_info structure
 * contained in the AVL tree.
 *
 * @see AVL_worker_t
 * @param n The object to free
 * @param param (unused)
 * @param depth (unused)
 */
static void om_reg_info_free(void *n, int param, int depth);

/* Creation functions */
PCache *
PCACHE_ctlrCreate(void)
{
    PCache  *cache;

    cache = emalloc(sizeof(struct PCache));
    memset(cache, 0, sizeof(struct PCache));


    cache->cache = AVL_create();
    cache->omRegs = AVL_create();
    cache->type = PCT_CTLR;

    cache->localPath = emalloc(sizeof(struct ctlr_path_info));
    cache->localPath->pathHead = NULL;
    cache->localPath->outPID = PATH_ID_MIN;
    cache->localPath->omSent = AVL_create();

    return cache;
}

PCache *
PCACHE_imCreate(void)
{
    PCache  *cache;
    
    cache = PCACHE_ctlrCreate();
    cache->type = PCT_IM;

    return cache;
}

uint32_t
PCACHE_regOutputMod(PCache *handle)
{
    struct om_reg_info  *rinfo;
    uint32_t            *data;

    rinfo = emalloc(sizeof(struct om_reg_info));
    rinfo->id = handle->nextOMID;
    rinfo->curConn = 0;
    rinfo->insData = LL_create();

    data = emalloc(sizeof(uint32_t));
    *data = rinfo->curConn;
    LL_prepend(rinfo->insData, data);

    AVL_insert(handle->omRegs, rinfo->id, rinfo);

    handle->nextOMID++;

    return rinfo->id;
}

void
PCACHE_unregOutputMod(PCache *pc, uint32_t omID)
{
    struct om_reg_info  *rinfo;
    uint32_t            *data;

    rinfo = AVL_find(pc->omRegs, omID);
    if (rinfo == NULL) 
        return;         /* don't know about it, so we are done */

    /* Remove it from the registry */
    AVL_remove(pc->omRegs, omID);

    /* free the insert data */
    while ( ! LL_isEmpty(rinfo->insData) ) {
        data = LL_first(rinfo->insData);
        LL_rmFirst(rinfo->insData);
        free(data);
    }
    LL_destroy(&(rinfo->insData));

    /* Free the reginfo struct */
    free(rinfo);
}

void
PCACHE_omNewConn(PCache *pc, uint32_t omID)
{
    uint32_t            *data;
    struct om_reg_info  *rinfo;

    rinfo = AVL_find(pc->omRegs, omID);
    if (rinfo == NULL)
        return; /* nothing we can do */


    /* inc the connection number */
    rinfo->curConn++;

    /* create new insert data to use */
    data = emalloc(sizeof(uint32_t));
    *data = rinfo->curConn;
    LL_prepend(rinfo->insData, data);
}

/* Destructor functions */
void
PCACHE_destroy(PCache *cache)
{
    if (cache == NULL) return;  /* done */
    /* XXX This may require slightly more elaborate cleanup on the data
     * structures in the hash table.
     */
    if (cache->cache != NULL)  
        AVL_destroy(cache->cache);
    if (cache->omRegs != NULL) {
        AVL_depthfirst(cache->omRegs, &om_reg_info_free, 0, 0);
        AVL_destroy(cache->omRegs);
    }
    conn_ent_free(cache->localPath, cache->type, 0);
    free(cache);
}

bool
PCACHE_clearConn(PCache *handle, uint32_t connID)
{
    struct conn_ent     *conn;

    DPRINT(DBG_PC, "PCACHE: clear: start, connection = %u", connID);

    conn = AVL_find(handle->cache, connID);
    if (conn != NULL) {
        if (! AVL_remove(handle->cache, connID) ) {
            return FALSE;
        }
        conn_ent_destroy(conn, handle->type);
    }

    DPRINT(DBG_PC, "PCACHE: clear: done", NO_VARG);

    return TRUE;
}
    
bool
PCACHE_addPath(PCache *handle, struct path *pathHead, bool aclPass)
{
    struct conn_ent         *conn, *new_conn;
    struct im_path_info     *ipinfo;
    struct ctlr_path_info   *cpinfo;
    bool                    error;

    assert(handle != NULL);
    assert(pathHead != NULL);

    DPRINT(DBG_PC, "PCACHE: add(%p, %p, %d)", handle, pathHead, aclPass);

    error = FALSE;

    conn = new_conn = NULL;
    ipinfo = NULL;
    cpinfo = NULL;

    conn = AVL_find(handle->cache, pathHead->connid);
    if (conn == NULL) {
        /* need to add a new one */
        new_conn = emalloc(sizeof(struct conn_ent));
        memset(new_conn, 0, sizeof(struct conn_ent));

        conn = new_conn;
        conn->connID = pathHead->connid;
        conn->paths = AVL_create();

        if (! AVL_insert(handle->cache, pathHead->connid, new_conn) ) {
            /* This really shouldn't happen */
            goto error;
        }
        DPRINT(DBG_PC, "PCACHE: add: inserted new connection %p (%u)", 
                new_conn, new_conn->connID);

    }

    /* conn is valid either way now */
    
    if (handle->type == PCT_CTLR) {
        cpinfo = emalloc(sizeof(struct ctlr_path_info));
        cpinfo->pathHead = pathHead;
        cpinfo->outPID = PATH_UNSET_ID;
        cpinfo->omSent = AVL_create();

        if (! AVL_insert(conn->paths, pathHead->id, cpinfo) ) {
            goto error;
        }
    }
    else if (handle->type == PCT_IM) {
        ipinfo = emalloc(sizeof(struct im_path_info));
        ipinfo->pathID = pathHead->id;
        ipinfo->aclPass = aclPass;

        if (! AVL_insert(conn->paths, pathHead->id, ipinfo) ) {
            goto error;
        }
    }
    else {
        throw_fatal("Unknown PathCache Type");
    }
    /* we don't actually care about the type of [ci]pinfo here */
    DPRINT(DBG_PC, "PCACHE: add: inserted new path %p (%u)", 
            (cpinfo != NULL) ? (void*)cpinfo : (void*)ipinfo, pathHead->id);

    return TRUE;

    /* Cleanup and exit with an error */
error:
    if (new_conn != NULL) {
        /* If this fails it probably means that new_conn insert failed */
        AVL_remove(handle->cache, new_conn->connID);

        /* free the rest of the new connection */
        AVL_destroy(new_conn->paths);
        free(new_conn);     
    }

    if (cpinfo != NULL) 
        free(cpinfo);

    if (ipinfo != NULL)
        free(ipinfo);

    return FALSE;
}

struct path *
PCACHE_findPath(PCache *handle, uint32_t connID, uint32_t pathID)
{
    struct ctlr_path_info   *pinfo;

    DPRINT(DBG_PC, "PCACHE_findPath(%p, %u, %u)", handle, connID, pathID);

    assert(handle != NULL);
    assert(handle->type == PCT_CTLR);

    DPRINT(DBG_PC, "handle = %p", handle);
    DPRINT(DBG_PC, "handle->cache = %p", handle->cache);
    pinfo = find_path_info(handle->cache, connID, pathID);
    if (pinfo == NULL) {
        return NULL;
    }

    return pinfo->pathHead;
}

int
PCACHE_aclCheck(PCache *handle, uint32_t connID, uint32_t pathID)
{
    struct im_path_info *pinfo;

    assert(handle != NULL);
    assert(handle->type == PCT_IM);

    DPRINT(DBG_PC, "aclCheck: start(%p, %u, %u)", handle, connID, pathID);

    pinfo = find_path_info(handle->cache, connID, pathID);
    if (pinfo == NULL) 
        return -1;          /* not found */

    DPRINT(DBG_PC, "aclCheck: end (pinfo == %p, aclPass == %d)", 
            pinfo, pinfo->aclPass);

    return pinfo->aclPass;  /* found, return aclPass */   
}

uint32_t 
PCACHE_findOutgoing(PCache *pc, uint32_t connID, uint32_t pathID, uint32_t omID)
{
    struct ctlr_path_info   *pinfo;
    struct om_reg_info      *rinfo;
    uint32_t                *omConn;

    DPRINT(DBG_PC, "start PCACHE_findOutgoing(%p, %u, %u, %u)",
            pc, connID, pathID, omID);

    assert(pc != NULL);
    assert(pc->type == PCT_CTLR);

    DPRINT(DBG_PC, "now pc = %p", pc); 
    DPRINT(DBG_PC, "now pc->cache = %p", pc->cache);

    /* If we are looking at an unset path Id, then the message had no
     * path previousnly, so we need to generate one with only this host
     * in it.
     */
    if (pathID == PATH_UNSET_ID) 
        pinfo = pc->localPath;
    else 
        pinfo = find_path_info(pc->cache, connID, pathID);

    if (pinfo == NULL) 
        return PATH_UNSET_ID;        /* not found */

    rinfo = AVL_find(pc->omRegs, omID);
    assert(rinfo != NULL);          /* error, no such output module */

    omConn = AVL_find(pinfo->omSent, omID); 
    if (omConn == NULL || *omConn != rinfo->curConn) {
        /* Remove an if there is an outdated entry */
        if (omConn != NULL)
            AVL_remove(pinfo->omSent, omID);

        /* this output module hasn't sent it out yet, so a negative
         * return is in order.
         */
        return PATH_UNSET_ID;
    }

    return pinfo->outPID;
}

struct path*
PCACHE_makeOutgoing(PCache *pc, uint32_t connID, uint32_t pathID, uint32_t omID)
{
    /* New path struct that we are returning */
    struct path             *new_path;           

    struct ctlr_path_info   *pinfo;
    struct om_reg_info      *rinfo;

    assert(pc != NULL);
    assert(pc->type == PCT_CTLR);

    if (pathID == PATH_UNSET_ID) 
        pinfo = pc->localPath;
    else
        pinfo = find_path_info(pc->cache, connID, pathID);

    if (pinfo == NULL) {
        return NULL;
    }

    /* construct the return object */
    if (pinfo->outPID == PATH_UNSET_ID) {
        /* skip the special path id value */
        if (pc->nextPID == PATH_UNSET_ID) pc->nextPID++;

        /* assign the path it's id */
        pinfo->outPID = pc->nextPID;

        /* increment the next path id for next time.
         * XXX We don't check for wrap here ... just assuming that by
         * the time it wraps the old connection will have droped.
         */
        pc->nextPID++;
    }
    new_path = emalloc(sizeof(struct path));
    memset(new_path, 0, sizeof(struct path));
    new_path->id = pinfo->outPID;
    new_path->connid = PATH_UNSET_ID;
    new_path->next = pinfo->pathHead;

    rinfo = AVL_find(pc->omRegs, omID);
    assert(rinfo != NULL);      /* programmer error */

    /* now mark that this OM has sent the path ... failure just means
     * extra copies of the path being sent.
     */
    AVL_insert(pinfo->omSent, omID, LL_first(rinfo->insData));

    return new_path;
}

static void *
find_path_info(AVLTree cache, uint32_t connID, uint32_t pathID)
{
    struct conn_ent     *conn;
    void                *pinfo;

    DPRINT(DBG_PC, "find_path_info(%p, %u, %u)", cache, connID, pathID);

    conn = AVL_find(cache, connID);
    if (conn == NULL) {
        return NULL;  /* not found */
    }

    DPRINT(DBG_PC, "found conn: %p (%u)", conn, conn->connID);
    pinfo = AVL_find(conn->paths, pathID);

    DPRINT(DBG_PC, "found path info: %p", pinfo); 

    return pinfo;   /* will be NULL if not found */
}


static void
conn_ent_destroy(struct conn_ent *ce, PC_type type)
{
    AVL_depthfirst(ce->paths, conn_ent_free, type, 0);
    AVL_destroy(ce->paths);
}

static void
conn_ent_free(void *n, int param, int depth)
{

    if (param == PCT_CTLR) {
        PATH_free(((struct ctlr_path_info*)n)->pathHead);
        AVL_destroy(((struct ctlr_path_info*)n)->omSent);
    }
    free(n);
}

static void
om_reg_info_free(void *n, int param, int depth)
{
    struct om_reg_info  *ori;
    LLIterator          itr;
    void                *data;

    ori = (struct om_reg_info*)n;
    itr = LL_iterator(ori->insData);
    while ( (data = LLI_next(itr)) != NULL) 
        free(data);
    LLI_destroy(&itr);
    LL_destroy(&(ori->insData));
    free(ori);
}
