/**
 * @file pathcache.h
 *
 * @brief Functions for manipulating the path cache.
 * @par
 * This defines functions for manipulating the three path caches, one
 * for Input Modules, one for Output Modules, and one for the
 * Controller.  The Output Module and controller cache are unified as
 * they run in the process and can be; the Input Modules have a separate
 * cache as they run is different processes.  This is only used by the
 * BEEP input/output modules.
 *
 * @ingroup pathcache
 *
 * $Id: pathcache.h,v 1.3.2.7 2003/04/30 20:04:04 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Jul 03 14:33:59 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UTIL_PATHCACHE_H
#define UTIL_PATHCACHE_H

#include "types.h"
#include "message/message.h"

/**@defgroup pathcache Path Caches 
 * @{
 */

/** The path cache object type */
typedef struct PCache PCache;

/**
 * Initilize the controller path cache.  This will return a object which
 * contains all the data in the cache.  It needs to know how many BEEP
 * output channels exist.
 *
 * @return The new path cache object
 */
PCache *PCACHE_ctlrCreate(void);

/** 
 * Create and initilize an Input Module path cache.
 *
 * @return The new input module path cache object.
 */
PCache *PCACHE_imCreate(void);

/**
 * Register that an output module will be using this PCache object to
 * send new paths.  This _only_ needs to be called if the OM is planning
 * on using PCACHE_findOutgoing() and PCACHE_makeOutgoing().
 *
 * The return is an output module ID which the OM will need to use is
 * some of the calls that it makes to the PCACHE functions. 
 *
 * @param cache The PCache object to register with.
 * @return The output module ID 
 */
uint32_t PCACHE_regOutputMod(PCache *cache);

/** 
 * Unregister an output module.  This cleans up the output module
 * registration.  It should be called in the OM destroy function.
 *
 * @param cache The PCache object to register with.
 * @param omID the output module id assgined in PCACHE_regOutputMod()
 */
void PCACHE_unregOutputMod(PCache *cache, uint32_t omID);

/**
 * Tell the patch cache that the output module has dropped it's
 * connection.  This will mean that the patch cache will no longer think
 * that the output module has sent a path over the channel.
 *
 * @param cache The PCache object to register with.
 * @param omID The output module's ID.
 */
void PCACHE_omNewConn(PCache *cache, uint32_t omID);

/**
 * Destroy the given path cache.  
 *
 * @param obj The path cache object to destroy.
 */
void PCACHE_destroy(PCache *cache);

/**
 * Remove a all path cache that match the given connection ID.
 *
 * @param handle The Path cache to work on.
 * @param connID The connection ID for the path
 * @return TRUE on success, FALSE otherwise.
 */
bool PCACHE_clearConn(PCache *handle, uint32_t connID);

/**
 * Add a path to the controller cache.  This takes the head of a path
 * and adds it to the cache for later access.  These entries are keyed
 * on \c pathHead->connid and \c pathHead->id.  Note that \c aclPath
 * will be ignored unless in an input module.
 *
 * @param handle The Path cache to work on.
 * @param pathHead the first struct path (which is a linked list).
 * @param aclPass Does this path pass the ACL checks?
 * @return TRUE if added, FALSE if not (probably a conflict).
 */
bool PCACHE_addPath(PCache *handle, struct path *pathHead, bool aclPass);

/**
 * Find a path in the cache.  This will only work on the controller
 * cache.
 *
 * @param handle The Path cache to work on.
 * @param connID The connection ID to look for.
 * @param pathID The path ID to look for.
 * @return The head of the path, or NULL if not found.
 *
 * @pre handle is a controller cache.
 */
struct path *PCACHE_findPath(PCache *handle, uint32_t connID, uint32_t pathID);

/**
 * Check for the existence of an entry in the path cache.  This returns
 * a tri-state value signalling either not found, aclPass is false, or
 * aclPass is true.  This will only work in the input module cache.
 *
 * @paarm handle The Path cache to work on.
 * @param connID The connection ID to look for.
 * @param pathID The path ID to look for.
 * @return < 0 if not found; 0 if \c aclPass==FALSE; 1 if \c aclPass==TRUE
 * 
 * @pre handle is an input module cache.
 */
int PCACHE_aclCheck(PCache *handle, uint32_t connID, uint32_t pathID);

/**
 * Find a path ID for an outgoing message.
 *
 * This takes the current path for the message (if any) and returns the
 * pathID for a path which has already been sent over the connection and
 * that includes the path already on the message.  If there is no path on 
 * the message then the pathID for a path which consists of only the
 * current host will be returned if already sent.
 *
 * If no appropriate path has been sent over the channel then
 * PATH_UNSET_ID will be returned.  This will also verify that the
 * calling output module has called PCACHE_makeOutgoing() previously.
 * 
 * @param handle The Path cache to work on.
 * @param connID The connection ID to look for.
 * @param pathID The path ID to look for.
 * @param omID   The Output Module ID we are working for.
 * @return pathID for the new path or PATH_UNSET_ID if non has been sent.
 */
uint32_t PCACHE_findOutgoing(PCache *handle, 
        uint32_t connID, uint32_t pathID, uint32_t omID);

/**
 * Create a path for an outgoing message.  This is used when
 * PCACHE_findOutgoing() failed to find an already sent path.  It
 * returns a partially constructed path, which contains a new pathID and
 * is linked to the path specified by \c connID and \c pathID.
 * It will also insert that pathID in the cache, so that it is returned
 * next time PCACHE_findOutgoing() is called with the same connID and
 * pathID.  Thus you must send this path over the channel, it also must
 * be filled in.
 *
 * @warning The returned path object needs to be freed.
 *
 * @param handle The Path cache to work on.
 * @param connID The connection ID to look for.
 * @param omID   The Output Module ID we are working for.
 * @param pathID The path ID to look for.
 */
struct path *PCACHE_makeOutgoing(PCache *handle, 
        uint32_t connID, uint32_t pathID, uint32_t omID);

/*@}*/

#endif /* UTIL_PATHCACHE_H */
