/**
 * @file privs.c
 *
 * @brief Implement the operations to run with privilege.
 *
 * $Id: privs.c,v 1.17.2.6 2003/10/22 04:09:04 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Mon May 20 17:58:20 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#   include <acconfig.h>
#endif

/** Needed to get access to O_NOFOLLOW under linux */
#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

#include "types.h"
#include "log.h"
#include "error.h"
#include "debug.h"
#include "strcompat.h"

#include "privs.h"

static char const rcsid[] 
    = "$Id: privs.c,v 1.17.2.6 2003/10/22 04:09:04 devink Exp $";

/* So the flags constants don't need to worry about this */
#ifndef HAVE_O_NOFOLLOW 
#   define O_NOFOLLOW   0
#endif

/* Not needed on all platforms, so just define to 0 */
#ifndef O_LARGEFILE
#   define O_LARGEFILE 0
#endif

/** allowed flags for an open(2) for read */
#define OPEN_RDFLAGS_ALLOW  (O_RDONLY | O_NOCTTY | O_NONBLOCK)
/** Flags added to every call to open(2) for reading */
#define OPEN_RDFLAGS_ADD    (O_RDONLY)

/** Allowed flags for open(2) for write */
#define OPEN_WRFLAGS_ALLOW  (O_NONBLOCK| O_SYNC| O_LARGEFILE| O_CREAT| O_EXCL)
/** Flags added to every call to open(2) for writing */
#define OPEN_WRFLAGS_ADD    (O_WRONLY | O_NOFOLLOW | O_APPEND)
/** Mode for any file created via PRIV_openWriteFile() */
#define OPEN_WRMODE         (S_IRUSR | S_IWUSR)

/** UID of root */
#define ROOT_UID    0

/** GID of root */
#define ROOT_GID    0

/** The unprivileged user's UID */
static uid_t    unpriv_uid = -1;

/** The unprivileged group's GID */
static gid_t    unpriv_gid = -1;

/** Set to false to disable priv. changing */
static bool     priv_disable = FALSE;

/** 
 * Set to true to signal that privs should be changed. 
 * This is usde by gainPrivs() and dropPrivs() to communicate about what
 * needs to be done.
 */
static bool changing_privs = TRUE;

/**
 * gain privleges.
 *
 * @return TRUE on success, FALSE otherwise.
 */
static bool gainPrivs(void);

/**
 * drop privleges.  This will exit in the event of an error.
 */
static void dropPrivs(void);

/** 
 * Check a struct stat for being an allowable type in PRIV_openWriteFile9) 
 * @param sbuf The struct stat to check
 */
#define WRITE_ALLOWABLE_TYPES(sbuf)\
    (S_ISREG(sbuf) || S_ISCHR(sbuf) || S_ISFIFO(sbuf))

bool
PRIV_setUnprivUser(char const *user, char const *group)
{
    struct passwd   *pwent;
    struct group    *grent;

    DPRINT(DBG_PRIV, "setUnprivUser(%s, %s)", user, group);

    if (user == NULL || *user == '\0') {
        LOG_print(LOG_ERROR, "No unprivleged user is defined.");
        return FALSE;
    }
    if (group == NULL || *group == '\0') {
        LOG_print(LOG_ERROR, "No unprivleged group is defined.");
        return FALSE;
    }

    pwent = getpwnam(user);
    if (pwent == NULL) {
        LOG_print(LOG_ERROR, "Lookup of user \"%s\" failed.", user);
        return FALSE;
    }

    grent = getgrnam(group);
    if (grent == NULL) {
        LOG_print(LOG_ERROR, "Lookup of group \"%s\" failed.", group);
        return FALSE;
    }

    if (pwent->pw_uid == 0) {
        LOG_print(LOG_ERROR, "Attempt to have syslogd always run as root");
        return FALSE;
    }

    if (grent->gr_gid == 0) {
        LOG_print(LOG_ERROR, "Attempt to have syslogd always with GID == 0");
        return FALSE;
    }

    unpriv_uid = pwent->pw_uid;
    unpriv_gid = grent->gr_gid;

    return TRUE;
}

void
PRIV_disableChanges(void)
{
    if (getuid() == ROOT_UID) {
        LOG_print(LOG_ERROR, 
            "Attempt to disable privlege changes while root denied");
        EXIT(EXIT_ERROR);
    }

    priv_disable = TRUE;

    /* this will prevent uninitilized checks from barfing */
    unpriv_uid = getuid();
    unpriv_gid = getgid(); 
}

bool
PRIV_dropTemp(void)
{
    /* if we are disabled then just pretend */
    if (priv_disable) return TRUE;

    assert(unpriv_gid > 0);
    assert(unpriv_uid > 0);

    /* drop GID first */
    if (getgid() != unpriv_gid) {
        /* this check is only to prevent unimportant error messages */
        if (setgid(unpriv_gid) != 0) {
            LOG_eprint("Changing GID to %d failed", errno, unpriv_gid);
            EXIT(EXIT_ERROR);
        }
    }

    /* drop UID next */
    if (getuid() != unpriv_uid) {
        /* again, prevent unneeded error messages */

        /* this is a two stage drop, it should leave:
         * real UID == unpriv_uid
         * effectiv UID == unpriv_uid
         * saved UID == 0
         */
        if (setreuid(unpriv_uid, -1) != 0) {
            LOG_eprint("Changing real UID to %d failed", errno, unpriv_uid);
            EXIT(EXIT_ERROR);
        }

        if (setreuid(-1, unpriv_uid) != 0) {
            LOG_eprint("Changing effective UID to %d failed", errno, unpriv_uid);
            EXIT(EXIT_ERROR);
        }
    }

    return TRUE;
}

bool
PRIV_dropPerm(void)
{
    /* if we are disabled then just pretend */
    if (priv_disable) return TRUE;

    assert(unpriv_gid > 0);
    assert(unpriv_uid > 0);
    /* permantly drop privleges */

    /* start with the group */
    if (getgid() != unpriv_gid) {
        if (setgid(unpriv_gid) != 0) {
            LOG_eprint("Changing GID to %d failed", errno, unpriv_gid);
            EXIT(EXIT_ERROR);
        }
    }

    /* now the user */
    if (getuid() != unpriv_uid) {
        if (setuid(unpriv_uid) != 0) {
            LOG_eprint("Changing UID to %d failed", errno, unpriv_uid);
            EXIT(EXIT_ERROR);
        }
    }

    return TRUE;
}

int
PRIV_openReadFile(char const *name, int flags)
{
    int         rtnFD;
    struct stat statbuf;

    rtnFD = -1;

    /* adjust mode first */
    flags &= OPEN_RDFLAGS_ALLOW;     /* only certain flags can be specified */
    flags |= OPEN_RDFLAGS_ADD;       /* add some others */

    if (! gainPrivs() ) {
        return FALSE;
    }

    /* open the file, then perform the checks needed.  This will avoid
     * race conditions.
     */
    if ( (rtnFD = open(name, flags, 0)) < 0) {
        LOG_eprint("Opening file \"%s\" failed:", errno, name);
        goto drop_privs;    /* just drop privs and return */
    }

    /* stat the file and make sure that it has good permissions */
    if (fstat(rtnFD, &statbuf) < 0) {
        LOG_eprint("Stating file \"%s\" failed:", errno, name);
        goto closefd;   /* need to cleanup before return */
    }

    if (statbuf.st_uid != 0) { 
        LOG_print(LOG_ERROR, 
            "Attempt to read a file (as root) not owned by root");
        goto closefd;   /* need to cleanup before return */
    }

    if ( statbuf.st_gid != 0 && (statbuf.st_mode & S_IWGRP) != 0) {
        LOG_print(LOG_ERROR, 
            "Attempt to read a file (as root) that can be written by a"
            " non-root grop");
        goto closefd;
    }

    if ( (statbuf.st_mode & S_IWOTH) != 0) {
        LOG_print(LOG_ERROR,
                "Attempt to read a file (as root) that is world writable");
        goto closefd;
    }

    /* looks good, skip ahead */
    goto drop_privs;

    /* This closes the FD if a post-open check failed */
closefd:
    if (rtnFD >= 0) close(rtnFD);
    rtnFD = -1;
    /* fall through to drop privs and return */

    /* Here we drop privs */
drop_privs:
    dropPrivs();

    /* Done */
    return rtnFD;
}

int
PRIV_openWriteFile(char const *name, int flags)
{
    int             rtnFD;
    struct stat     statbuf;
#ifndef HAVE_O_NOFOLLOW
    struct stat     lstatbuf;
#endif

    DPRINT(DBG_PRIV, "PRIV_openWriteFile(%s, %x)",
            name, flags);

    /* clean up flags first */
    flags &= OPEN_WRFLAGS_ALLOW;
    flags |= OPEN_WRFLAGS_ADD;

    rtnFD = -1;

    if (! gainPrivs() ) {
        return FALSE;
    }

    /* open first then check, that way we avoid races */
    if ( (rtnFD = open(name, flags, OPEN_WRMODE) ) < 0) {
        LOG_eprint("Opening file \"%s\" for write failed:", errno, name);
        goto drop_privs;
    }

    /* 
     * need to stat the files, this is needed before we can check for
     * sym links.
     */
    if (fstat(rtnFD, &statbuf) < 0) {
        LOG_eprint("fstat failed for file \"%s\":", errno, name);
        goto closefd;
    }

#ifndef HAVE_O_NOFOLLOW
    /* 
     * O_NOFOLLOW doesn't work, so we must explicitly check for the case
     * where name referes to a symlink.
     */
    if (lstat(name, &lstatbuf) < 0) {
        LOG_eprint("lstat failed for file \"%s\":", errno, name);
        goto closefd;
    }

    /* check the lstat results, so a symlink to a normal file is detected. */
    if (! WRITE_ALLOWABLE_TYPES(lstatbuf.st_mode) ) {
        LOG_print(LOG_ERROR, 
            "Attempt to open a special file \"%s\" for write aborted", name);
        goto closefd;
    }

    /* now make sure that the lstat and fstat were on the same file. */
    if (lstatbuf.st_ino != statbuf.st_ino) {
        LOG_print(LOG_ERROR,
            "Attempt to change file \"%s\" between open and check detected",
            name);
        goto closefd;
    }
#else
    /* 
     * If O_NOFOLLOW works, then we don't have to worry about having
     * followed a symlink
     */
    if (! WRITE_ALLOWABLE_TYPES(statbuf.st_mode) ) {
        LOG_print(LOG_ERROR, 
            "Attempt to open a special file \"%s\" for write aborted", name);
        goto closefd;
    }
#endif

    /* No other checks at this time */
    goto drop_privs;    /* skip ahead */

    /*
     * This is an error handler section which is used for cleanup if an
     * error was detected earlier (it closes rtnFD).
     */
closefd:
    if (rtnFD >= 0) close(rtnFD);
    rtnFD = -1;
    /* fall through to priv dropping and return block */

    /*
     * This is hit whenever the function is done, either because of an
     * error or success.
     */
drop_privs:
    dropPrivs();


    return rtnFD;
}

bool
PRIV_bindSocket(int sock, struct sockaddr *sa, socklen_t salen)
{
    bool    success;
    mode_t  oumask;

    DPRINT(DBG_PRIV, "bindSocket: start(%d, %p, %d)", sock, sa, salen);

    if (! gainPrivs() ) {
        DPRINT(DBG_PRIV, "gainPrivs() failed", NO_VARG);
        return FALSE;
    }

    /* actually try the bind */
    success = TRUE;     /* be optimistic */

    /* set the umask incase this is a PF_UNIX socket */
    oumask = umask(S_IRWXG | S_IRWXO);

    if (bind(sock, sa, salen) < 0) {
        LOG_eprint("binding socket:", errno);
        success = FALSE;
        /* fall through and drop privs now */
    }

    /* restore the umask */
    umask(oumask);

    dropPrivs();

    DPRINT(DBG_PRIV, "bindSocket: end (%d)", success);
    return success;
}

bool 
PRIV_bindUnixSocket(int sock, struct sockaddr_un *sun, 
                    bool unprivOwner, mode_t mode)
{
    bool            success;
    bool            compINodes = FALSE;
    struct stat     sbStart, sbEnd;
    char            sock_path[UNIX_PATH_MAX];
    uid_t           owner_uid;
    gid_t           owner_gid;

    DPRINT(DBG_PRIV, "bindUnixSocket: start (%s)", sun->sun_path);

    success = FALSE;

    /* save the socket path 
     * XXX Not sure this is needed, but is here incase bind changes sun
     */
    strlcpy(sock_path, sun->sun_path, UNIX_PATH_MAX);

    /* now gain privs and do the real work */
    if (! gainPrivs() ) {
        DPRINT(DBG_PRIV, "gainPrivs() failed", NO_VARG);
        return FALSE;
    }

    /* log whenever the inode does not change. */
    if (lstat(sock_path, &sbStart) == 0) {
        compINodes = TRUE;
        DPRINT(DBG_PRIV, "sbStart.st_ino = %u", sbStart.st_ino);
    }
    else {
        /* To make sure we have some record ... */
        if (errno != ENOENT) {
            LOG_eprint("bindUnixSocket: stat failed, but file exists", errno);
        }
    }
    DPRINT(DBG_PRIV, "compINodes = %s", compINodes ? "true" : "false");

#if 0
    if (strcmp(sock_path, "/dev/log") ) {
        /*
         * Due to RedHat's minilogd we need to be extra careful that the new
         * socket (if /dev/log) does not have the same inode as the old.  To
         * do this we build a temp file name in the same directory, and rename
         * the socket path to the temp file.
         *
         * Because this is only on RH-linux, some things are more hard
         * coded than they otherwise would be.
         */

        wlen = strlcpy(tmp_path, "/dev/log-XXXXXX", UNIX_PATH_MAX);
        if (wlen <= 0 || wlen >= UNIX_PATH_MAX) {
            LOG_print(LOG_ERROR, "bindUnixSocket: Buffer too small to build temp socket name");
            goto end;
        }

        if (mktemp(tmp_path) == NULL) {
            /* should never happen ... but just in case */
            LOG_print(LOG_ERROR, "bindUnixSocket: mktemp() failed");
            goto end;
        }

        if (rename(sock_path, tmp_path) != 0) {
            if (errno != ENOENT) {
                LOG_eprint("bindUnixSocket: Failde to mave old socket");
                goto end;
            }
            /* no big deal */
        }
        rmTmpPath = TRUE;
    } 
    else {
        /* we can just unlink it and be ok */
        PRIV_unlink(sock_path); 
    }
#endif

    /* remove this so it won't conflict. */
    PRIV_unlink(sock_path);

    /* now bind to the socket */
    if (! PRIV_bindSocket(sock, (struct sockaddr*)sun, sizeof(*sun)) ) {
        /* logged by PRIV_bindSocket() */
        DPRINT(DBG_PRIV, "bindSocket failed", NO_VARG);
        goto end;
    }

    if (compINodes && lstat(sock_path, &sbEnd) == 0) {
        DPRINT(DBG_PRIV, "sbEnd.st_ino = %u", sbEnd.st_ino);
        if (sbEnd.st_ino == sbStart.st_ino) {
            DPRINT(DBG_PRIV, "INode number for unix domain socket %s is the same", sock_path);
            LOG_print(LOG_INFO, "INode number for unix domain socket '%s' did not change", 
                    sock_path);
        }
    }

    /* figure out who should own the socket */
    owner_uid = getuid();
    owner_gid = getgid();
    if (unprivOwner) {
        owner_uid = unpriv_uid;
        owner_gid = unpriv_gid;
    }
    else if (! priv_disable) {
        owner_uid = ROOT_UID;
        owner_gid = ROOT_GID;
    }
    
#ifdef HAVE_LCHOWN
    /* and change the owner and group ... must use lchown so that
     * symbolic links are not followed, and fchown(2) doesn't seem to
     * work correctly. 
     */
    if (lchown(sock_path, owner_uid, owner_gid) < 0) {
        LOG_eprint("lchown()", errno);
        DPRINT(DBG_PRIV, "lchown failed: %d", errno);
        goto end;
    }

#else
#error "Don't know what to do without a lchown function"
#endif

    /* Set the permissions correctly.
     * chmod() must be used here b/c (on Linux at lesat) fchom() doesn't
     * seem to work.
     */
    if (mode != 0 && chmod(sock_path, mode) < 0) {
        LOG_eprint("fchmod()", errno);
        DPRINT(DBG_PRIV, "fchmod() failed: %d", errno);
        goto end;
    }

    success = TRUE; /* we completed successfully */

    /* function footer */
end:
    /* doing this here means it always gets deleted */
#if 0
    if (rmTmpPath)
        PRIV_unlink(tmp_path);
#endif

    dropPrivs();

    DPRINT(DBG_PRIV, "bindUnixSocket: end (%d)", success);
    return success;
}

bool
PRIV_unlink(char const *file)
{
    bool        success;

    if (! gainPrivs() ) {
        return FALSE;
    }

    success = TRUE;

    if (unlink(file) < 0) {
        if (errno != ENOENT) {
            LOG_eprint("unlink()", errno);
            success = FALSE;
        }
    }

    dropPrivs();
    return success;
}

bool
PRIV_fchmod(int fd, mode_t mode)
{
    bool    success;

    if (! gainPrivs() ) {
        return FALSE;
    }

    success = TRUE;

    if (fchmod(fd, mode) < 0) {
        LOG_eprint("fchmod()", errno);
        success = FALSE;
    }

    dropPrivs();

    return success;
}

static bool
gainPrivs(void)
{
    DPRINT(DBG_PRIV, "gainPrivs(): start", NO_VARG);

    changing_privs = TRUE;


    /* if we haven't been initilized then this is an error */
    if (unpriv_uid == -1 || unpriv_gid == -1) {
        DPRINT(DBG_PRIV, "gainPrivs: not initilized", NO_VARG);
        return FALSE;
    }

    /* Are we already root or disabled? */
    if (priv_disable || getuid() == 0 || geteuid() == 0) {
        /* don't loose root permissions when done */
        changing_privs = FALSE;

        DPRINT(DBG_PRIV, "gainPrivs: no need to change: dis %d, uid %u (%u)", 
                priv_disable, getuid(), geteuid());
    }


    if (changing_privs) {
        /* need to gain privs first */
        if (setreuid(-1, ROOT_UID) < 0) {
            LOG_eprint("Changing effective UID to %d failed:", errno, ROOT_UID);

            DPRINT(DBG_PRIV, "gainPrivs: error gaining privs (%d)", errno);

            return FALSE;   /* give up right away */
        }
    }

    DPRINT(DBG_PRIV, "gainPrivs: done", NO_VARG);

    return TRUE;
}

static void
dropPrivs(void)
{
    if (changing_privs) {
        /* don't leave with lower privs than we started with */
        if (setreuid(-1, unpriv_uid) < 0) {
            LOG_eprint("Changing effective UID to %d failed:",errno,unpriv_uid);
            EXIT(EXIT_ERROR);
        }
    }

    changing_privs = TRUE;
}
