/**
 * @file profile-beep.c
 *
 * @brief Provide the implementtion for the BEEP profile (both raw and cooked)
 *
 *
 * $Id: profile-beep.c,v 1.8.2.9 2003/10/22 04:09:04 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Aug 22 16:39:28 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <strcompat.h>
#include <stdlib.h>

#include "types.h"
#include <librr/rr.h>

#include "version.h"
#include "log.h"
#include "error.h"

#include "ipc/ipc.h"
#include "ipc/ipc_err.h"
#include "message/msg_parse_rfc3164.h"
#include "message/msg_parse_cooked.h"

/* for IMOD_setDevFQDN() */
#include "inputmod/inputmod.h"

#include "profile-beep.h"

#define BEEPRAW_URI		"http://xml.resource.org/profiles/syslog/RAW"
#define BEEPCOOKED_URI	"http://xml.resource.org/profiles/syslog/COOKED"

/** The greeting message; use BEEP_setGrtMsg() to set this, or accept
 * the default (see the docs for \ref BEEP_setGrtMsg())
 */
static char *greeting_message = "\r\nSDSC-syslog " VERSION_STR "\r\n";

/** Initialize a new BEEP_prof structure
 *
 * @param object The object to initialize
 */
static void BEEP_prof_init (GObject *object);

/** Initialize a new BEEP_profClass structure
 * This function is responsible for setting the callbacks appropriately
 *
 * @param klass The class object to initialize
 */
static void BEEP_profClass_init (GObjectClass *klass);

static void BEEP_conn_init (GObject *object);

static void BEEP_connClass_init (GObjectClass *class);

/**@group Profile indepedent call backs.
 * @{
 */
/** Called when channel close request in confirmed. */
static void BEEP_close_confirmation (RRChannel *channel, gint code,
		const gchar *xml_lang, const gchar *description);

/** 
 * Called when a channel recieves a close request. 
 * @return TRUE means accept the close request, FALSE means reject it 
 */
static gboolean BEEP_close_indication(RRChannel *, gint code,
        const gchar *xml_lang, const gchar *descr, GError **error);
/*@}*/

/** 
 * Skip mime headers. This is second best to actually acting on the mime
 * headers, but it will be good enough for now.
 *
 * @param frame The frame with data to skip
 * @param newlen Pass back the new length in here.
 * @return the starting location of real content, or NULL on error.
 */
static char *skipMIMEHeaders(const RRFrame *frame, int *newlen);

/* ***************** INDEPENDENT FUNCTIONS **************************** */
static void
BEEP_prof_init (GObject *object)
{
    /* Nothing to do */
}

static void
BEEP_profClass_init (GObjectClass *klass)
{
	/* nothing to do, everything initialized by BEEPRAW_rawClass_init() */
}

GType 
BEEP_getType (void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (BEEP_profClass),
			NULL,
			NULL,
			(GClassInitFunc) BEEP_profClass_init,
			NULL,
			NULL,
			sizeof (BEEP_prof),
			16,
			(GInstanceInitFunc) BEEP_prof_init
		};
		rr_type = g_type_register_static (RR_TYPE_CHANNEL, "BEEPRAW_raw", 
						  &type_info, 0);
	}
	return rr_type;
}

static void
BEEP_conn_init (GObject *object)
{
	BEEP_conn *co = (BEEP_conn*) object;

	co->channel_id=-1;

	co->inmsgs=NULL;

	co->state = BEEP_NOTCON;
	co->syslog_mutex = g_mutex_new();
	co->state_cv = g_cond_new();

    co->syslog_chan = NULL;
}

static void
BEEP_connClass_init (GObjectClass *class)
{
	/* nothing to do */
}

static char *
skipMIMEHeaders(const RRFrame *frame, int *newlen)
{
    int     isLineStart = 1;    
    char    *str = frame->payload;
    int     len = frame->size+1;    
    int     origLen = len;      /* save for final computation. */
    char    *body = NULL;
    int     hlen = -1;

    /* I'd rather my callers always have the length available. */
    assert(newlen != NULL); 

    while (*str != '\0' && len > 1) {
        if (str[0] == '\r' && str[1] == '\n') {
            isLineStart++;
            str += 2;
            len -= 2;
            if (isLineStart > 1)  {
                hlen = origLen - len;    /* found the end of the headers. */
                break;
            }
        }
        else if (isLineStart == 1 && str[0] == '<') {
            /* looks like there are no mime headers here.
             * We should do proper matching on legal mime header chars
             * and format, but this will work for any correctly
             * formmated message (I think)
             */
            DPRINT(DBG_BEEP, "No MIME headors on message, first char is '<'", 
                    NO_VARG);
            hlen = 0;
            break;
        }
        else {
            isLineStart = 0;   /* No longer the start of the line. */
            str++;
            len--;
        }
    }

    if (hlen < 0) {
        DPRINT(DBG_BEEP, "Could not find MIME headers", NO_VARG);
        return NULL;
    }

    body = frame->payload + hlen;
    if (newlen != NULL)
        *newlen = frame->size+1 - hlen;

    return body;
}


GType
BEEP_conn_getType(void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (BEEP_connClass),
			NULL,
			NULL,
			(GClassInitFunc) BEEP_connClass_init,
			NULL,
			NULL,
			sizeof (BEEP_conn),
			16,
			(GInstanceInitFunc) BEEP_conn_init
		};
		rr_type = g_type_register_static (RR_TYPE_TCP_CONNECTION, "BEEP_conn", 
						  &type_info, 0);
	}
	return rr_type;
}

BEEP_conn *
BEEP_conn_new (RRProfileRegistry *profreg, struct iam *iamData)
{
	BEEP_conn *bc;
	RRConnection *conn;

	bc = g_object_new(BEEP_conn_getType(),NULL);
    bc->iamdata = iamData;  /* must be set before a new channel is created */
	conn=RR_CONNECTION(bc);

	if (NULL!=profreg)
	{
		rr_connection_set_profile_registry(conn,profreg);
		/*g_object_unref(G_OBJECT(profreg));*/
	}

	rr_connection_add_channel(conn,RR_CHANNEL(conn->manager));

	return bc;
}

void
BEEP_setGrtMsg(char *message)
{
	greeting_message=message;
}

static gboolean
BEEP_channel_start_check(RRChannel *chan, const gchar *p, GError **error)
{
    bool        has_syslog_chan;

    DPRINT(DBG_BEEP, "BEEP_channel_start_check: role = %s",
        chan->connection->role == RR_ROLE_LISTENER ? "listener" : "initiator");

    if (chan->connection->role != RR_ROLE_LISTENER) {
        /* Opps, only the listener acceps channels. And only the
         * initiator creates them.
         */
        DPRINT(DBG_BEEP, "BEEP_channel_start_check: wrong role", NO_VARG);
        return FALSE;
    }

    has_syslog_chan = FALSE;

    if (BEEP_ISCONN(chan->connection) ) {
        BEEP_conn   *bc = BEEP_CONN(chan->connection);
        g_mutex_lock(bc->syslog_mutex);
            has_syslog_chan = (NULL != bc->syslog_chan);
        g_mutex_unlock(bc->syslog_mutex);
    }

    DPRINT(DBG_BEEP, "BEEP_channel_start_check: has_syslog_chan = %d", 
            has_syslog_chan);

    return !has_syslog_chan;
}


gboolean
BEEP_sendLog (BEEP_conn *conn, char *msg)
{
	RRChannel   *chan;
	RRMessage   *send_msg;
	GError      *error=NULL;
    bool        CTS;        /* clear to send */
    int         msg_type;
    char        *mcopy;
    int         mlen;

	DPRINT(DBG_BEEP, "BEEP_sendLog(): sending a message: \'%s\'", msg);

	chan = RR_CHANNEL (conn->syslog_chan);

	/* make sure that we've got the server info */
	g_mutex_lock (conn->syslog_mutex);
        CTS = TRUE;
        if (BEEP_TRYCON == conn->state) {
            /* We are trying to connect, wait until that changes.
             * We don't wait until BEEP_READY, because we might miss
             * that event.
             */

            while(BEEP_TRYCON == conn->state)
                g_cond_wait(conn->state_cv, conn->syslog_mutex);

        }
        /* We are now either connected or not connected */
        if (BEEP_NOTCON == conn->state) { 
            /* not connected, not trying == return error */
            /* But DO NOT return from a critical section, bad devin */
            CTS = FALSE;
        }
	g_mutex_unlock (conn->syslog_mutex);
    if (! CTS)
        return FALSE;

	/* send the message; this will handle freeing our message */
    if (BEEP_ISRAW(chan)) {
        msg_type = RR_FRAME_TYPE_ANS;
    }
    else if (BEEP_ISCOOK(chan)) {
        msg_type = RR_FRAME_TYPE_MSG;
    }
    else {
        /* Unknown protocol type */
        return FALSE;
    }

    /* the formmatter should have put mime headers in msg already. 
     * But, due to the way errors are handled, we must copy the fucking
     * message anyway.  It allows the caller to free the message buffer,
     * without worrying about RR fucking memory handling and double
     * frees and what not
     */
    mlen = strlen(msg)+1;
    mcopy = emalloc(mlen);
    strlcpy(mcopy, msg, mlen);

    /* create the message. */
    send_msg = rr_message_static_new(msg_type, mcopy, mlen-1, TRUE);

	if(FALSE==rr_channel_send_message (chan, send_msg, &error))
	{
        /* May have disconnected since we last checked. */
        if (g_error_matches(error, RR_ERROR, RR_ERROR_DISCONNECTED) ) {
            /* Make sure that the profile object is marked as
             * unconnected */
            g_mutex_lock(conn->syslog_mutex);
                conn->state = BEEP_NOTCON;
                DPRINT(DBG_BEEP, 
                    "BEEP_sendLog: state change -> BEEP_NOTCON", NO_VARG);

                g_cond_signal(conn->state_cv);   /* state changed */
            g_mutex_unlock(conn->syslog_mutex);
        }
        else {
            /* Not sure what happened ... just log it. */
            LOG_print(LOG_INFO, "error sending BEEP message: %s",
                    error->message);
        }
		return FALSE;
	}

    /* rr_channel_flush dead locks under certain circumstances, so lets
     * just trust RRBEEP to work and send the data eventually
     */
#if 0
	/* block until we're sure it's gone out */
	if(FALSE==rr_channel_flush(chan,&error))
	{
        DPRINT(DBG_BEEP, "BEEP_sendLog(): rr_channel_flush == FALSE", NO_VARG);

        if (g_error_matches(error, RR_ERROR, RR_ERROR_DISCONNECTED) ) {
            /* Make sure that the profile object is marked as
             * unconnected */
            g_mutex_lock(profobj->connState_mutex);
                profobj->connState = BEEP_NOTCON;
                g_cond_signal(profobj->connState_cv);   /* connState changed */
                profobj->connState = BEEP_NOTCON;
            g_mutex_unlock(profobj->connState_mutex);
        }
        else {
            /* Not sure, so just log */
            LOG_print(LOG_INFO, "error flushing BEEP message: %s", 
                    error->message);
        }
		return FALSE;
	}
#endif
	DPRINT(DBG_BEEP, "BEEP_sendLog(): return TRUE", NO_VARG);
	
	return TRUE;
}

BEEP_conn *
BEEP_IN_connect (int sock, struct MFRMT_opt *frmt_opt, 
                 const struct iam *iamData)
{
    BEEP_conn           *conn;
	RRProfileRegistry   *profreg;
    RRChannel           *chan;
    GType               type;
    char const          *type_name;
	GError              *error=NULL;
	BEEP_prof           *profobj=NULL;
    struct iam          *iamCopy = NULL;

	if(!rr_init(NULL,NULL,&error))
	{
		LOG_print(LOG_CRIT,"rr_init() failed: %s",error->message);
		return NULL;
	}

	profreg = rr_profile_registry_new();

    /* skip adding anything to the profile registery */
	if(MFRMT_RFC3164==frmt_opt->type)
	{
		rr_profile_registry_add_profile(profreg, BEEPRAW_getType(), NULL);
        type = BEEPRAW_getType();
        type_name = "RAW";
	}
	else if(MFRMT_COOKED==frmt_opt->type)
	{
		rr_profile_registry_add_profile(profreg, BEEPCOOK_getType(), NULL);
        type = BEEPCOOK_getType();
        type_name = "COOKED";

        /* need to setup the iam data */
        iamCopy = MSG_iamCopy(iamData);
	}
	else
	{
        LOG_print(LOG_ERROR, "Unknown BEEP profile type.");
        return NULL;
	}

    DPRINT(DBG_BEEP, "BEEP_IN_connect: passed iam =  %p, iam copy = %p",
            iamData, iamCopy);

    conn = BEEP_conn_new(profreg, iamCopy);
	if(NULL==conn)
	{

		LOG_print(LOG_CRIT,"connection creation failed"); 
		return NULL;
	}

	if(!rr_tcp_connection_connect_fd (RR_TCP_CONNECTION(conn), 
				sock, RR_ROLE_INITIATOR, &error))
	{
		LOG_print(LOG_CRIT,"connect failed\n");
		return NULL;
	}

    /* Try to start a channel of the desired type */
    chan=rr_connection_start(RR_CONNECTION(conn),NULL,type,NULL,&error);
            
    if(NULL==chan) {
        LOG_print(LOG_ERROR, 
            "Can't start a channel running BEEP/%s: (%d) %s",
            type_name, error->code, error->message);

        return NULL;
    }
            
    profobj=BEEP_PROF(chan);

    /* Okay, we know that we have the profobj now ... put it into
     * BEEP_TRYCON state.
     */
    g_mutex_lock(conn->syslog_mutex);
        /* it seems that it's possible for the channel to be ready
         * before we get here ... so make sure that we are not connected
         * yet, before changing the state.
         */
        DPRINT(DBG_BEEP, "BEEP_IN_connect: current state = %d", conn->state);
        DPRINT(DBG_BEEP, "BEEP_IN_connect: BEEP_READY = %d", BEEP_READY);
        if (BEEP_READY > conn->state) {
            conn->state = BEEP_TRYCON;
            DPRINT(DBG_BEEP, 
                "BEEP_IN_connect: state change to BEEP_TRYCON", NO_VARG);

            g_cond_signal(conn->state_cv);   /* connState changed */
        }
    g_mutex_unlock(conn->syslog_mutex);

	return conn;
}


void
BEEP_IN_shutdown(BEEP_conn *conn)
{
	RRMessage   *send_msg;
	RRChannel   *chan;
    RRManager   *mngr;
    GError      *error = NULL;
    bool        need_send_nul;
    bool        need_close_data;

	DPRINT(DBG_BEEP, "BEEP_IN_shutdown()", NO_VARG);

	chan = RR_CHANNEL (conn->syslog_chan);
    mngr = rr_connection_get_manager(RR_CONNECTION(conn));

    /* figure out if we need to send the NUL packet */
    need_send_nul = FALSE;
    need_close_data = FALSE;
    g_mutex_lock(conn->syslog_mutex);
        if (conn->state != BEEP_CLOSING && conn->state != BEEP_NOTCON) {

            if (BEEP_ISRAW(conn->syslog_chan) )
                need_send_nul = TRUE;

            need_close_data = TRUE;
            conn->state = BEEP_CLOSING;
            DPRINT(DBG_BEEP, 
                "BEEP_IN_shutdown: state change -> BEEP_CLOSING", NO_VARG);
            g_cond_signal(conn->state_cv);
        }
    g_mutex_unlock(conn->syslog_mutex);

    if (need_send_nul) {
        /* Finish our reply with the NUL message */
        send_msg=rr_message_static_new(RR_FRAME_TYPE_NUL, "", 0, FALSE);
        rr_channel_send_message(chan,send_msg,NULL);
    }

    if (need_close_data) {
        rr_manager_close_channel(mngr, chan, 200, NULL, NULL, &error);
    }

    /* wait until the syslog channel has closed */
	g_mutex_lock(conn->syslog_mutex);
		while(BEEP_CLOSING == conn->state) 
			g_cond_wait(conn->state_cv, conn->syslog_mutex);

        /* graceful isn't it :) */
        if (BEEP_NOTCON != conn->state) 
            throw_fatal("BEEP_IN_shutdown: illegal state transition");
	g_mutex_unlock(conn->syslog_mutex);

    /* Now finish shutting down the connection, if we are responsible */
    DPRINT(DBG_BEEP, "Trying connection_disconnect()", NO_VARG);
    if(!rr_connection_disconnect(RR_CONNECTION(conn), &error))
        g_error("disconnect: %s\n",error->message);

    DPRINT(DBG_BEEP, "Trying rr_exit()", NO_VARG);
    if(!rr_exit(&error))
        g_error("rr_exit: %s\n",error->message);

    /*
    if (NULL != conn)
        g_object_unref(G_OBJECT(conn));
    */
}

void
BEEP_LI_init(void)
{
	GError *error=NULL;
	bool	return_val=TRUE;

#ifndef G_THREADS_ENABLED
	LOG_print(LOG_ERROR, "fatal: thread support in glib required");
	EXIT(EXIT_ERROR);
#endif

	return_val=rr_init(NULL,NULL,&error);

	if( !return_val || (NULL!=error) )
	{
		LOG_print(LOG_ERROR, "fatal error in rr_init() (return: %d): %s",
				return_val,error->message);

		if(!g_thread_supported())
			LOG_print(LOG_ERROR, 
                "possible cause of error: no thread support in glib");

		EXIT(EXIT_ERROR);
	}
}

void
BEEP_LI_shutdown(BEEP_conn *conn)
{
    RRManager   *mngr;
    GError      *error = NULL;

    mngr = rr_connection_get_manager(RR_CONNECTION(conn));

    g_mutex_lock(conn->syslog_mutex);
        if (conn->state == BEEP_NOTCON || conn->state == BEEP_CLOSING) {
            g_mutex_unlock(conn->syslog_mutex);
            return;
        }
        conn->state = BEEP_CLOSING;
        DPRINT(DBG_BEEP, 
            "BEEP_LI_shutdown: state change -> BEEP_CLOSING", NO_VARG);
        g_cond_signal(conn->state_cv);
    g_mutex_unlock(conn->syslog_mutex);

    /* Close the syslog channel */
    /* Try a blocking close  */
    rr_manager_close_channel_nonblock(mngr, 
            RR_CHANNEL(conn->syslog_chan), 200, NULL, NULL, &error);

    /* Let BEEP_LI_shutdown_final() finish this */
}

void
BEEP_LI_shutdown_final(BEEP_conn *conn) 
{
    GError *error = NULL;

    /* Now finish shutting down the connection, if we are responsible */
    DPRINT(DBG_BEEP, "Trying connection_disconnect()", NO_VARG);
    if(!rr_connection_disconnect(RR_CONNECTION(conn), &error))
        g_error("disconnect: %s\n",error->message);

    /*
    if (NULL != conn) 
        g_object_unref(G_OBJECT(conn));
    */
}

static void
BEEP_close_confirmation (RRChannel *channel, gint code,
		const gchar *xml_lang, const gchar *description)
{
    BEEP_conn   *conn;

    DPRINT(DBG_BEEP, "BEEP_close_confirmation: chan: %d", channel->id);

    if (! BEEP_ISCONN(channel->connection)) {
        DPRINT(DBG_BEEP, 
            "BEEP_close_confirmation, connection is not a BEEP_CONN", NO_VARG);
        return;
    }
    conn = BEEP_CONN(channel->connection);

	g_mutex_lock (conn->syslog_mutex);
        /* our state is either BEEP_NOTCON, or it should be. */
		conn->state = BEEP_NOTCON;
        conn->syslog_chan = NULL;
        DPRINT(DBG_BEEP, 
            "BEEP_close_confirmation: state change -> BEEP_NOTCON", NO_VARG);
        g_cond_signal(conn->state_cv);
	g_mutex_unlock (conn->syslog_mutex);

	DPRINT(DBG_BEEP,"close confirmation: %d, %s, chan: %d\n",
            code, description, channel->id);

    if (NULL != conn->inmsgs) {
        struct MSG_qelem    *close = emalloc(sizeof(struct MSG_qelem));

        close->type = MSGQE_CLOSE;
        close->chan = conn->channel_id; /* not used */
        close->data.connid = (uint32_t)conn;

        TSQ_enqueue(conn->inmsgs, close);
    }
}

static gboolean
BEEP_close_indication(RRChannel *channel, gint code,
        const gchar *xml_lang, const gchar *description, GError **error)
{
    BEEP_conn   *conn = BEEP_CONN(channel->connection);
    gboolean    need_send_nul = FALSE;
    RRMessage   *send_msg;

    DPRINT(DBG_BEEP, "BEEP_close_indication: chan = %d", channel->id);

    if (! BEEP_ISCONN(channel->connection)) {
        DPRINT(DBG_BEEP, 
            "BEEP_close_indication, connection is not a BEEP_CONN", NO_VARG);
        return TRUE;
    }

    conn = BEEP_CONN(channel->connection);

    g_mutex_lock(conn->syslog_mutex);
        if (BEEP_CLOSING != conn->state
            && BEEP_ISRAW(channel)
            && RR_ROLE_INITIATOR == channel->connection->role) 
        {
            DPRINT(DBG_BEEP, "BEEP_close_indication: will send NUL", NO_VARG);
            need_send_nul = TRUE;
        }

        /* We only move to BEEP_NOTCON if we didn't request the close.
         * If we did request the close, then we are in BEEP_CLOSING, and
         * will move to NOTCON on close_confirmation
         */
        if (BEEP_CLOSING != conn->state) {
            conn->state = BEEP_NOTCON;
            conn->syslog_chan = NULL;
            DPRINT(DBG_BEEP,
                "BEEP_close_indication: state change -> BEEP_NOTCON", NO_VARG);
            g_cond_signal(conn->state_cv);
        }
    g_mutex_unlock(conn->syslog_mutex);

    if (need_send_nul) {
        GError  *error = NULL;
        /* Finish our reply with the NUL message */
        DPRINT(DBG_BEEP, "BEEP_close_indication: sending NUL", NO_VARG);
        send_msg=rr_message_static_new(RR_FRAME_TYPE_NUL, "", 0, FALSE); 
        if (! rr_channel_send_message(channel,send_msg,&error)) {
            DPRINT(DBG_BEEP, "BEEP_close_indication: send NUL failed: (%d) %s",
                    error->code, error->message);
        }
        /* It seems that the NUL never gets sent out ... well fuck it.
         * there is not a damn thing that I can do about it.  I've tried
         * rr_channel_flush() ... but that just hangs forever.  So I
         * give up ... At this point I'm not fucking fixing and errors
         * which seem to be related to the multi-threaded BEEP stack
         * that we are using.  It was a bad idea, it is a bad idea, and
         * my time is better spend writing a BEEP library that is sane.
         *
         * Have a day.
         */
    }

    return TRUE;
}


static void
BEEP_chan_start(RRChannel *chan, BEEP_conST state)
{
    BEEP_conn   *conn;

    /* This means that the session is good to go */
    /* XXX We ignore the piggyback ... please don't use it */

    conn = BEEP_CONN(chan->connection);

    DPRINT(DBG_BEEP, "BEEP_chan_start: role = %s; state = %d",
        RR_CONNECTION(conn)->role == RR_ROLE_INITIATOR 
            ? "initiator" : "listener", state);

    g_mutex_lock(conn->syslog_mutex); 
        conn->syslog_chan = BEEP_PROF(chan);
        conn->state = state;
        DPRINT(DBG_BEEP, "BEEP_chan_start: state change -> (%d)", state);
        g_cond_signal(conn->state_cv);
    g_mutex_unlock(conn->syslog_mutex);
}

/* ***************************** COOKED FUNCTIONS ************************** */
/** Callback: Handles an available frame for syslog/COOKED.  */
static gboolean BEEPCOOK_frame_available(RRChannel *, RRFrame *, GError **);

/** Callback: channel start request (server only) */
static void BEEPCOOK_chan_start_serv(RRChannel *);

/** Callback: channel start (client only) */
static void BEEPCOOK_chan_start_cli(RRChannel *, const gchar *);

/** Initialize a new BEEPcook structure
 *
 * @param object The object to initialize
 */
static void BEEPcook_init (GObject *object);

/** Initialize a new BEEPcookClass structure
 * This function is responsible for setting the callbacks appropriately
 *
 * @param klass The class object to initialize
 */
static void BEEPcookClass_init (GObjectClass *klass);

GType 
BEEPCOOK_getType (void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (BEEPcookClass),
			NULL,
			NULL,
			(GClassInitFunc) BEEPcookClass_init,
			NULL,
			NULL,
			sizeof (BEEPcook),
			16,
			(GInstanceInitFunc) BEEPcook_init
		};
		rr_type = g_type_register_static (BEEP_getType(), "BEEPcook", 
						  &type_info, 0);

		rr_channel_set_uri (rr_type, BEEPCOOKED_URI);
	}
	return rr_type;
}

static void
BEEPcook_init (GObject *object)
{
	/* nothing to do, everything initialized by BEEP_prof_init() */
}

static void
BEEPcookClass_init (GObjectClass *klass)
{
	RRChannelClass *channel_class = (RRChannelClass *)klass;

	channel_class->frame_available = BEEPCOOK_frame_available;

    /* make sure that it's legal to start a channel */
    channel_class->server_init = BEEP_channel_start_check;

	channel_class->server_confirmation = BEEPCOOK_chan_start_serv;
    channel_class->client_confirmation = BEEPCOOK_chan_start_cli;

    /* Handle closing */
	channel_class->close_confirmation = BEEP_close_confirmation;
    channel_class->close_indication = BEEP_close_indication;
}


static void
BEEPCOOK_chan_start_cli(RRChannel *chan, const gchar *piggy)
{
    char        *iam_buf;
    size_t      len, buflen;
    BEEP_conn   *bc;
    RRMessage   *msg;
    bool        sendIam;

    /* XXX This should be in the fuction above this one ... damnit! */

    if(BEEP_ISCONN(chan->connection)) {
        bc=BEEP_CONN(chan->connection);
        /* do we have an iam to send? */
        sendIam = bc->iamdata != NULL;

        DPRINT(DBG_BEEP, 
            "BEEPCOOK_chan_start_cli: sending iamdata: %s",
            sendIam ? "yes" : "no");


        /* Start the channel our initial state depends on if we will
         * send an <iam>
         */
        if (sendIam) {
            BEEP_chan_start(chan, BEEP_TRYCON);
        }
        else {
            BEEP_chan_start(chan, BEEP_READY);
        }

        /* now we send an <iam> if we have the data. */
        if (sendIam) {
            buflen = 1024;  /* should be enough */
            iam_buf = emalloc(buflen);

            len = MFRMT_formatIam(iam_buf, buflen, bc->iamdata);
            if (len == 0) {
                free(iam_buf);  /* cleanup */
                /* XXX shutdown here, should be graceful. */
                throw_fatal("Buffer not big enough to hold IAM element");
            }
            DPRINT(DBG_BEEP, 
                "BEEPCOOK_chan_start_cli: Sending <iam>: '%s'", iam_buf);
            /* note, len-1 to not send the ending NUL, and pass ownership
             * off to RR 
             */
            msg = rr_message_static_new (RR_FRAME_TYPE_MSG, iam_buf,
                    len-1, TRUE);

            rr_channel_send_message (chan , msg, NULL);
        }
    }
    else {
        /* XXX error, but it should never happen */
        throw_fatal("BEEP connection object is not a BEEP_conn");
    }
}

static void
BEEPCOOK_chan_start_serv(RRChannel *chan)
{
    /* we ignore the piggy back */
    BEEP_chan_start(chan, BEEP_READY);
}



static gboolean
BEEPCOOK_frame_available (RRChannel *channel, RRFrame *frame, 
			GError **error)
{
	BEEP_conn *bc;
	int chanid;

    if(BEEP_ISCONN(channel->connection))
    {
        bc=BEEP_CONN(channel->connection);
    }
    else
    {
        LOG_print(LOG_ERROR, 
            "channel->connection was not a BEEP_conn cannot handle frame");
        if (error != NULL && *error == NULL) {
            *error = g_error_new(RR_BEEP_ERROR, 
                        RR_BEEP_CODE_TRANSACTION_FAILED,
                        "something seriously wrong");
        }
        return FALSE;
    }

	switch(frame->type)
	{
		case RR_FRAME_TYPE_MSG:
        {
			/* We recieved a message. */

            struct MPCOOKED_ctx mpc_recvd;
            struct MSG_qelem    *outmsg;
            struct path         *tempPath;
            struct message      *tempMsg;
            RRMessage           *ok_msg;
            char                *content;
            int                 clen;

			DPRINT(DBG_BEEP, 
                "BEEPCOOK_frame_available: received MSG", NO_VARG);

			chanid = bc->channel_id;

			DPRINT(DBG_BEEP, "channel->connection was a BEEP_conn, chanid: %d",
					chanid);


            outmsg=emalloc(sizeof(struct MSG_qelem));
            // skip MIME headers.
            DPRINT(DBG_BEEP, "got message: '%s'", frame->payload);
            content = skipMIMEHeaders(frame, &clen);

            DPRINT(DBG_BEEP, "Skiped MIME headers: content = '%s'", content);

            if (content == NULL
                || !MPCOOKED_parseString(&mpc_recvd,content, clen) 
                || (TRUE==mpc_recvd.error) )
            {
                DPRINT(DBG_BEEP, "Received invalid message: '%s'",
                        frame->payload);

                MPCOOKED_free(&mpc_recvd);
                if (error != NULL && *error == NULL) {
                    *error = g_error_new(RR_BEEP_ERROR, 
                                RR_BEEP_CODE_SYNTAX_ERROR, 
                                "Can't parse message");
                }
                return FALSE;
            }

            switch(mpc_recvd.type)
            {
                case MPC_PATH:
                    /* we need to copy the structure from the static memory,
                     * to avoid segfaults later */
                    tempPath=emalloc(sizeof(struct path));
                    memcpy(tempPath,&(mpc_recvd.data.path),
                            sizeof(struct path));

                    outmsg->data.path=tempPath;
                    outmsg->type=MSGQE_PATH;
                    outmsg->chan=chanid;

                    TSQ_enqueue(bc->inmsgs, (void *)outmsg);

                    break;

                case MPC_IAM:
                    /* XXX we could fill in the BEEP_conn::iamdata
                     * struct, but noone uses it, and it just makes for
                     * another thing to worry about freeing.
                     */
                    DPRINT(DBG_BEEP, "ERROR: MPC_IAM not implemented",
                            NO_VARG);
                    MPCOOKED_free(&mpc_recvd);
                    break;

                case MPC_ENTRY:
                    tempMsg=emalloc(sizeof(struct message));
                    memcpy(tempMsg,&(mpc_recvd.data.message),
                            sizeof(struct message));

                    outmsg->data.msg=tempMsg;
                    outmsg->type=MSGQE_MSG;
                    outmsg->chan=chanid;

                    TSQ_enqueue(bc->inmsgs, (void *)outmsg);

                    break;

                default:
                    DPRINT(DBG_BEEP, 
                        "MPCOOKED_parseString() returned odd value: %d",
                        mpc_recvd.type);

                    MPCOOKED_free(&mpc_recvd);
                    if (error != NULL && *error == NULL) {
                        *error = g_error_new(RR_BEEP_ERROR, 
                                    RR_BEEP_CODE_SYNTAX_ERROR, 
                                    "Unreconized Tag");
                    }
                    return FALSE;
                    break;
            }

            /* If we made it here then we assume that all was well. */
            DPRINT(DBG_BEEP, "BEEPCOOK_frame_available: sending RPY", NO_VARG);
            ok_msg = rr_message_static_new(RR_FRAME_TYPE_RPY, 
                        MFRMT_COOK_OK_MSG, strlen(MFRMT_COOK_OK_MSG), FALSE);
            if (! rr_channel_send_message(channel, ok_msg, error) ) {
                DPRINT(DBG_BEEP, 
                    "BEEPCOOK_frame_available: error sending RPY: (%d) %s",
                    error && *error ? (*error)->code : -1, 
                    error && *error ? (*error)->message : "(null)");
            }

			break;
        }
        case RR_FRAME_TYPE_RPY:
            DPRINT(DBG_BEEP, "BEEPCOOK_frame_available: recieved RPY", NO_VARG);
            /* was this in respnose to our iam? */
            g_mutex_lock(bc->syslog_mutex);
                if (BEEP_TRYCON == bc->state) {
                    bc->state = BEEP_READY; /* iam accepted */
                    DPRINT(DBG_BEEP, 
                        "BEEPCOOK_frame_available: state change -> BEEP_READY",
                        NO_VARG);
                    g_cond_signal(bc->state_cv);
                }
                /* else, good they got our message ... we assumed that
                 * :) 
                 */
            g_mutex_unlock(bc->syslog_mutex);

            break;
        case RR_FRAME_TYPE_ERR:
        {
            bool    is_iam_reject = FALSE;
            char    *content;
            int     clen;
            DPRINT(DBG_BEEP, "BEEPCOOK_frame_available: recieved ERR", NO_VARG);
            /* is this the response to our iam? */
            g_mutex_lock(bc->syslog_mutex);
                if (BEEP_TRYCON == bc->state) {
                    bc->state = BEEP_NOTCON; /* iam rejected */
                    DPRINT(DBG_BEEP,
                        "BEEPCOOK_frame_available: state change -> BEEP_NOTCON",
                        NO_VARG);
                    g_cond_signal(bc->state_cv);
                    is_iam_reject = TRUE;
                }
            g_mutex_unlock(bc->syslog_mutex);

            /* skip mime headers */
            content = skipMIMEHeaders(frame, &clen);
            if (content == NULL) 
                content = "Error parsing error's MIME headers";

            if (is_iam_reject) {
                LOG_print(LOG_ERROR, "Server Rejected our IAM element: %s", 
                        content);
            }
            else {
                LOG_print(LOG_ERROR, "Got BEEP error: %s", content);
            }
            break;
        }
		default:
			DPRINT(DBG_BEEP, 
                "BEEPCOOK_frame_available(): received ?? (default case)", 
                NO_VARG);

			break;
	}

	return TRUE;
}

/* ******************************* RAW FUNCTIONS **************************** */
/**@group 
 * Function to support the syslog/RAW BEEP profile
 * @{
 */

/** Callback: channel start request (server only) */
static void BEEPRAW_chan_start_serv(RRChannel *);

/** Callback: for an incomming channel start (client only). */
static void BEEPRAW_chan_start_cli(RRChannel *, const gchar *);

/** Callback: handle an available frame */
static gboolean BEEPRAW_frame_available (RRChannel *, RRFrame *, GError **);

/** Initialize a new BEEPRAW_raw structure
 *
 * @param object The object to initialize
 */
static void BEEPRAW_raw_init (GObject *object);

/** Initialize a new BEEPRAW_class structure
 * This function is responsible for setting the callbacks appropriately
 *
 * @param klass The class object to initialize
 */
static void BEEPRAW_rawClass_init (GObjectClass *klass);

GType 
BEEPRAW_getType (void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (BEEPRAW_rawClass),
			NULL,
			NULL,
			(GClassInitFunc) BEEPRAW_rawClass_init,
			NULL,
			NULL,
			sizeof (BEEPRAW_raw),
			16,
			(GInstanceInitFunc) BEEPRAW_raw_init
		};
		rr_type = g_type_register_static (BEEP_getType(), "BEEPraw", 
						  &type_info, 0);

		rr_channel_set_uri (rr_type, BEEPRAW_URI);
	}
	return rr_type;
}

static void
BEEPRAW_raw_init (GObject *object)
{
	/* nothing to do, everything initialized by BEEP_prof_init() */
}

static void
BEEPRAW_rawClass_init (GObjectClass *klass)
{
	RRChannelClass *channel_class = (RRChannelClass *)klass;

	channel_class->frame_available = BEEPRAW_frame_available;

    /* Make sure that it's legal to start a channel */
    channel_class->server_init = BEEP_channel_start_check;

    /* handle channel start messages */
	channel_class->server_confirmation = BEEPRAW_chan_start_serv;
    channel_class->client_confirmation = BEEPRAW_chan_start_cli;

    /* handle closing */
	channel_class->close_confirmation = BEEP_close_confirmation;
    channel_class->close_indication = BEEP_close_indication;
}


static void
BEEPRAW_chan_start_serv(RRChannel *channel)
{
	RRMessage *msg;

    /* We need to do the normal channel start stuff, but we also need to 
     * send the first message, so that the client can just ANS it
     * repeatedly.
     */
    BEEP_chan_start(channel, BEEP_READY);

	DPRINT(DBG_BEEP, 
        "BEEP_listener_confirmation(): sending greeting message ('%s')",
        greeting_message);

	msg=rr_message_static_new (RR_FRAME_TYPE_MSG, greeting_message,
			strlen(greeting_message), FALSE);

	rr_channel_send_message (channel, msg, NULL);
}

static void
BEEPRAW_chan_start_cli(RRChannel *chan, const gchar *piggy)
{
    /* Little catch here ... the channel is not actually ready yet */
    BEEP_chan_start(chan, BEEP_TRYCON);
}

static gboolean
BEEPRAW_frame_available (RRChannel *channel, RRFrame *frame, GError **error)
{
	BEEP_conn *bc;
	int chanid;

    if(BEEP_ISCONN(channel->connection))
    {
        bc=BEEP_CONN(channel->connection);
    }
    else
    {
        DPRINT(DBG_BEEP, 
            "BEEPRAW_frame_available: connection is not BEEP_conn", NO_VARG);
        return FALSE;
    }

	switch(frame->type)
	{
		case RR_FRAME_TYPE_MSG:
			/* we've received the server info, so we're ready to start sending
			 * messages
			 */
			DPRINT(DBG_BEEP, "BEEPRAW_frame_available: received MSG", NO_VARG);

			g_mutex_lock (bc->syslog_mutex);
				bc->state = BEEP_READY;
                DPRINT(DBG_BEEP, 
                    "BEEPRAW_frame_available: state change -> BEEP_READY", 
                    NO_VARG);

				g_cond_signal (bc->state_cv);
			g_mutex_unlock (bc->syslog_mutex);
			break;

		case RR_FRAME_TYPE_NUL:
			DPRINT(DBG_BEEP, "BEEP_frame_available: received NUL", NO_VARG);

            /* we are done, but we expect the other side to send the
             * close.
             */

			return TRUE;

			break;

		case RR_FRAME_TYPE_ANS:
        {
            struct MSG_qelem    *msgelem;
            char                *content = NULL;
            int                 clen;

			/* this may only work with a raw profile */
			DPRINT(DBG_BEEP, 
                "BEEPRAW_frame_available: received ANS", NO_VARG);

            
            DPRINT(DBG_BEEP, "Got message with headers: '%s'", frame->payload);

            content = skipMIMEHeaders(frame, &clen);

            DPRINT(DBG_BEEP, "Skiped MIME headers: content = '%s'", content);

			/* calculate chanid */
			chanid=bc->channel_id;
			DPRINT(DBG_BEEP, "channel->connection was a BEEP_conn, chanid: %d",
					chanid);

			/* parse and send the message depending on what type of listener we
			 * are */
            msgelem = emalloc(sizeof(struct MSG_qelem));
            msgelem->data.msg=emalloc(sizeof(struct message));

            if (content == NULL
                || MPOLD_parsein(msgelem->data.msg, content,clen,FALSE, FALSE)<0)
            {
                DPRINT(DBG_BEEP, "Received invalid message: '%s'", content);
                        

                MSG_free(msgelem->data.msg);
                free(msgelem->data.msg);
                free(msgelem);

                return FALSE;
            }

            msgelem->type=MSGQE_MSG;
            msgelem->chan=chanid;

            /** @todo Need to find the peer name */
            DPRINT(DBG_BEEP, "BEEP_frame_available(): peer name is: %s",
                    rr_connection_get_server_name(channel->connection));
            IMOD_setLocalhostInfo(msgelem->data.msg);

            TSQ_enqueue(bc->inmsgs, (void *)msgelem);

            DPRINT(DBG_BEEP, "Enqueued %s, queue is now: %d long",
                    msgelem->data.msg->msg, TSQ_size(bc->inmsgs));

			break;
        }

		default:
			DPRINT(DBG_BEEP, "BR_frame_available(): received ?? (default case)",
					NO_VARG);

			break;
	}

	return TRUE;
}

/*@}*/

