/**
 * @file profile-beep.h
 *
 * @brief Provide an interface to the BEEP profile (both raw and cooked )
 *
 *
 * $Id: profile-beep.h,v 1.8.2.9 2003/10/22 04:09:04 devink Exp $
 *
 * @author Matt Small <small@sdsc.edu>
 * @date   Thu Aug 22 16:39:45 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _PROFILE_BEEPRAW_H
#define _PROFILE_BEEPRAW_H

#include <librr/rr-channel.h>
#include <librr/tcp/rr-tcpconnection.h>

#include "util/tsqueue.h"
#include "message/msg_frmt.h"

G_BEGIN_DECLS

/**@addtogroup profile
 * @{
 */

/** BEEP structure; extends RRChannel to include extra information needed by
 * the BEEP layer */
typedef struct _BEEP_prof {
	RRChannel parent_object;		/**< pointer to the parent object */

} BEEP_prof;

/** Extends RRChannelClass in the same way as BEEP_prof, above */
typedef struct _BEEP_profClass {
	RRChannelClass parent_class;	/**< pointer to the parent class */
} BEEP_profClass;

/** Casts an object to a BEEP_prof, with checking
 *
 * @param obj The object to typecast
 * @return The cast object
 */
#define BEEP_PROF(obj) (G_TYPE_CHECK_INSTANCE_CAST( \
			(obj), BEEP_getType(), BEEP_prof) \
		)

/** Gets the glib type associated with BEEP Raw objects
 *
 * @return The type (constructed through glib voodoo)
 */
GType BEEP_getType (void);
/*@}*/

/**@defgroup profile-conn BEEP Connections
 * @ingroup profile
 * @{
 */

/** Enumerate possible channel states */
typedef enum
{
    BEEP_NOTCON,    /**< No connection established, not trying */
    BEEP_TRYCON,    /**< Trying to establish a connection, not ready yet */
    BEEP_READY,     /**< Clear to send. */
    BEEP_CLOSING    /**< Closing the data channel */
} BEEP_conST;

/** A structure to extend RRTCPConnection to also include a channel_id field */
typedef struct _BEEP_conn {
	RRTCPConnection parent_object;		/**< pointer to the parent object */

	int channel_id;     /**< The channel that this connection came in on */

	TSQueue *inmsgs;	/**< A thread safe queue, containing all incoming 
                             messages (will be sent by the main process, with 
                             its IPC handle, to the controller) */

    BEEP_prof   *syslog_chan;   /**< The channel for syslog in this conn */
    BEEP_conST  state;          /**< Connection state */
    GMutex      *syslog_mutex;  /**< lock access to connState */
    GCond       *state_cv;      /**< signaled when state is changed
                                     Use syslog_mutex as the lock. */

    struct iam  *iamdata;       /**< Data for the IAM message */
} BEEP_conn;

/** A matching structure to extend RRTCPConnectionClass */
typedef struct _BEEP_connClass {
	RRTCPConnectionClass parent_class;	/**< pointer to the parent class */
} BEEP_connClass;

/** Macro to convert an object to a BEEP_conn object
 *
 * @param obj The object to convert
 */
#define BEEP_CONN(obj) (G_TYPE_CHECK_INSTANCE_CAST( \
			(obj), BEEP_conn_getType(), BEEP_conn) \
		)

/** Macro to verify that an object is a BEEP_conn object, or can be 'cast to
 * an object of that type
 *
 * @param obj The object to test
 */
#define BEEP_ISCONN(obj) (G_TYPE_CHECK_INSTANCE_TYPE( \
			(obj), BEEP_conn_getType() ) \
		)

/** Return the type information for a BEEP_conn object
 *
 * @return The unique type ID
 */
GType BEEP_conn_getType(void);

/** 
 * Replaces rr_connection_set_profile_registry(); use this to create
 * connections through our BEEP profile.  This adds a channel_id field in
 * connections, for use in BR_frame_available().
 *
 * @param profreg An optional profile registry
 * @param iamData Data for the IAM element (may be NULL, will be free()ed). 
 * @return A new 
 */
BEEP_conn *BEEP_conn_new (RRProfileRegistry *profreg, struct iam *iamData);

/*@}*/

/**@addtogroup profile
 * @{
 */

/** 
 * Sets the greeting message (sent by the server to the listener, after the
 * listener has connected).
 *
 * If this is not set, the default will be used (SDSC-syslog <version>).
 * The message should both begin or end with "\r\n".
 *
 * @param message The message to send; note that the pointer is just copied, so
 * 					if this is free()ed, then there will most likely be a
 * 					segfault.
 */
void BEEP_setGrtMsg(char *message);


/** 
 * Send a log message over the BEEP connection.  This will copy msg so
 * that error leaks can be prevented in the face of an error (where it's
 * not clear if roadrunner or us should free it).  
 *
 * @note To be used only by the initiator
 *
 * @param conn The BEEP connection object.
 * @param msg The message to send.  BEEP_sendLog _will_not_ free this. 
 *
 * @return TRUE on success, FALSE otherwise
 */
gboolean BEEP_sendLog (BEEP_conn *conn, char *msg);

/** 
 * Start the BEEP layer.  This includes setting up the channel in which
 * syslog messages will be passed over.
 *
 * @param sock A previously created socket to send to
 * @param frmt_opt Format options (used to determine what sort of a sender is
 * 			being created)
 * @param iamData The data for the IAM element it will be copied if used.
 *
 * @return A new connection, should be used with the other BEEP_IN_* functions,
 * 			and destroyed with BEEP_disconnect()
 */
BEEP_conn *BEEP_IN_connect (int sock, struct MFRMT_opt *frmt_opt, 
                            const struct iam *iamData);

/** 
 * Stop the BEEP layer
 *
 * @note See BRI_connect() for notes applicable to this, as well
 *
 * @param raw A raw profile, created with BRI_sendchan_start()
 * @param connection A connection, created with BRI_connect()
 * @param error An error structure, in case of an unepected event
 */
void BEEP_IN_shutdown(BEEP_conn *conn);

/** 
 * Initialize the BEEP listener layer
 * Calls rr_init(), checks the return values
 * Also checks to make sure that gthreads is present and supported
 */
void BEEP_LI_init (void);

/**
 * Close the channel being used for a syslog-reliable profile.  This
 * should be the only channel other than 0 which is open.
 *
 * @param conn The beep connection close.
 */
void BEEP_LI_shutdown(BEEP_conn *conn);

/**
 * Perform the final actions needed to kill the BEEP session.
 *
 * @param conn The beep connection close.
 */
void BEEP_LI_shutdown_final(BEEP_conn *conn);

/*@}*/



/**@defgroup profile-raw BEEP Raw Profile
 * @ingroup profile
 * @{
 */

/** Basic BEEP Raw structure */
typedef struct _BEEPRAW_str {
	BEEP_prof parent_object;		/**< pointer to the parent object */
} BEEPRAW_raw;

/** BEEP Raw class structure */
typedef struct _BEEPRAW_cls {
	BEEP_profClass parent_class;	/**< pointer to the parent class */
} BEEPRAW_rawClass;

/**
 * Check to see if the profile is syslog/RAW
 */
#define BEEP_ISRAW(obj) (G_TYPE_CHECK_INSTANCE_TYPE( \
			(obj), BEEPRAW_getType() ) \
		)

/** Gets the glib type associated with BEEP Raw objects
 *
 * @return The type (constructed through glib voodoo)
 */
GType BEEPRAW_getType (void);

/*@}*/

/**@defgroup profile-cooked BEEP Cooked Profile
 * @ingroup profile
 * @{
 */

/** Basic BEEP Cooked structure */
typedef struct _BEEPcook {
	BEEP_prof parent_object;		/**< pointer to the parent object */
} BEEPcook;

/** Basic BEEP Cooked Class structure */
typedef struct _BEEPcookClass {
	BEEP_profClass parent_class;	/**< pointer to the parent class */
} BEEPcookClass;

/**
 * Check to see if the profile is syslog/COOKED
 */
#define BEEP_ISCOOK(obj) (G_TYPE_CHECK_INSTANCE_TYPE( \
			(obj), BEEPCOOK_getType() ) \
		)

/** Gets the glib type associated with BEEP Cooked objects
 *
 * @return The type (constructed through glib voodoo)
 */
GType BEEPCOOK_getType(void);

/*@}*/

G_END_DECLS

#endif /* _PROFILE_BEEPRAW_H */
