/**
 * @file tslist.c
 *
 * @brief Implement the Thread Safe Queue.
 *
 * $Id: tsqueue.c,v 1.2.2.4 2003/10/22 04:09:04 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Sep 18 14:13:20 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>

#if HAVE_LIBRR
# include <glib/gthread.h>
#endif

#include "tsqueue.h"
#include "list.h"
#include "types.h"
#include "error.h"

/* Be able to loose the locking if we aren't complining with librr
 * (which imples no libgthread).
 */
#if HAVE_LIBRR
# define MUTEX_NEW(mutex)       mutex = g_mutex_new()
# define MUTEX_FREE(mutex)      g_mutex_free(mutex)
# define MUTEX_LOCK(mutex)      g_mutex_lock(mutex)
# define MUTEX_UNLOCK(mutex)    g_mutex_unlock(mutex)

# define COND_NEW(cond)         cond = g_cond_new()
# define COND_FREE(cond)        g_cond_free(cond)
# define COND_WAIT(c, m, to)    g_cond_timed_wait(c, m, to)
# define COND_SIGNAL(cond)      g_cond_signal(cond)

# define TIMEOUT_SET(tv, to_secs) \
    g_get_current_time(tv); (tv)->tv_sec += to_secs

#else
# define GMutex     int
# define GCond      int
# define GTimeVal   int
# define gboolean   bool

# define MUTEX_NEW(mutex)   mutex = NULL
# define MUTEX_FREE(mutex)
# define MUTEX_LOCK(mutex)
# define MUTEX_UNLOCK(mutex)

#define COND_NEW(cond)      cond = NULL
#define COND_FREE(cond)
#define COND_WAIT(c, m, to)
#define COND_SIGNAL(cond)

#define TIMEOUT_SET(tv, to_secs)
#endif /* HAVE_LIBRR */

char const rcsid[]
    = "$Id: tsqueue.c,v 1.2.2.4 2003/10/22 04:09:04 devink Exp $";

/** The Queue data structure */
struct TSQ_object {
    List    list;
    GMutex  *lock;
    GCond   *insert;
};

TSQueue* 
TSQ_create(void)
{
    TSQueue *q;

    q = emalloc(sizeof(struct TSQ_object));

    q->list = LL_create();
    MUTEX_NEW(q->lock);
    COND_NEW(q->insert);

    return q;
}

void
TSQ_destroy(TSQueue **q)
{
    GMutex  *lock;

    if (q == NULL) return;

    MUTEX_LOCK((*q)->lock);

    /* it might be possible to come unlocked with a null queue, which
     * means that some other thread will clean up the lock.
     */
    if (q == NULL) return;

    /* The goal is to do all the clean up we can while locked.  This
     * means that we need to save the lock so that everything else can
     * be freeded.  Then only after unlocking can we free the lock.
     */
    lock = (*q)->lock; 

    LL_destroy(&((*q)->list));
    free(*q);
    *q = NULL;

    MUTEX_UNLOCK(lock);

    MUTEX_FREE(lock);
}

void
TSQ_enqueue(TSQueue *q, void *obj)
{
    assert(q != NULL);

    MUTEX_LOCK(q->lock);

    LL_append(q->list, obj);

    COND_SIGNAL(q->insert);

    MUTEX_UNLOCK(q->lock);
}

void *
TSQ_dequeue(TSQueue *q)
{
    void    *obj;

    assert(q != NULL);

    obj = NULL;

    MUTEX_LOCK(q->lock);

    if (! LL_isEmpty(q->list) ) {
        obj = LL_first(q->list);
        if (! LL_rmFirst(q->list) ) 
            obj = NULL;
    }

    MUTEX_UNLOCK(q->lock);
    return obj;
}

bool
TSQ_wait(TSQueue *q, long timeout)
{
    void        *obj;
    GTimeVal    tv;
    gboolean    haveElem;

    assert(q != NULL);

    obj = NULL;

    /* set up the timeout */
    TIMEOUT_SET(&tv, timeout);

    MUTEX_LOCK(q->lock);

    if (LL_isEmpty(q->list)) {
        /* Wait for something to be inserted */
        COND_WAIT(q->insert, q->lock, &tv);
    }

    /* Check for available elements, COND_WAIT might have timed out.  */
    haveElem = ! LL_isEmpty(q->list);

    MUTEX_UNLOCK(q->lock);

    return haveElem;
}


bool
TSQ_isEmpty(TSQueue *q)
{
    bool rtn;

    assert(q != NULL);

    MUTEX_LOCK(q->lock);
    rtn = LL_isEmpty(q->list);
    MUTEX_UNLOCK(q->lock);

    return rtn;
}

int
TSQ_size(TSQueue *q)
{
    int     rtn;

    assert(q != NULL);

    MUTEX_LOCK(q->lock);
    rtn = LL_size(q->list);
    MUTEX_UNLOCK(q->lock);

    return rtn;
}

