/**
 * @file tsqueue.h
 *
 * @brief A Thread Safe Queue.
 * @par
 * This is a queue which can safely be used from multiple threads at
 * once.  It is built on top of the List structure.
 *
 * $Id: tsqueue.h,v 1.2.2.3 2003/04/30 20:04:06 devink Exp $
 *
 * @author Devin Kowatch <devink@sdsc.edu>
 * @date   Wed Sep 18 14:05:09 PDT 2002
 *
 * Copyright 2002-2003 
 *     The Regents of the University of California. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 * 
 *     * Neither the name of the The Regents of the University of
 *       California nor the names of its contributors may be used to
 *       endorse or promote products derived from this software without
 *       specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UTIL_TSQUEUE_H
#define UTIL_TSQUEUE_H

#if HAVE_LIBRR
#include <glib/gthread.h>
#endif /* HAVE_LIBRR */

#include "list.h"
#include "types.h"

/**@defgroup tsqueue Thread Safe Queue
 * @{
 */

/** The type of a thread safe queue */
typedef struct TSQ_object   TSQueue;

/**
 * Create a new queue.
 * @return a pointer to the new list.
 */
TSQueue* TSQ_create(void);

/**
 * Destroy a queue.
 * @param q The TSQueue to destroy; after the function returns *q will be NULL
 */
void TSQ_destroy(TSQueue **q);

/**
 * Add an object to the queue.
 * @param q The TSQueue object to add to.
 * @param obj The object to data to the queue.
 */
void TSQ_enqueue(TSQueue *q, void *obj);

/**
 * Remove an object from the queue.
 * @param q The TSQueue to take an object from.
 * @return The object that was at the head of the queue.
 */
void * TSQ_dequeue(TSQueue *q);

/**
 * Wait for an insertion to ocurre on the queue.  
 *
 * @param q The TSQueue to wait for an insertion on.
 * @param timeout The max number of seconds to wait.
 * @return TRUE if this is returning due to an insert.
 */
bool TSQ_wait(TSQueue *q, long timeout);

/**
 * Is the queue empty?
 * @param q The TSQueue object to check
 * @return TRUE if \c q is empty, FALSE otherwise.
 */
bool TSQ_isEmpty(TSQueue *q);

/**
 * Get the number of objects currently in the queue.
 * @param q The TSQueue object to look at.
 * @return A count of objectn in the queue.
 */
int TSQ_size(TSQueue *q);

/*@}*/

#endif /* UTIL_TSQUEUE_H */
